/* Generated by Cython 0.29.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "Aspidites._vendor.contracts.inspection",
        "sources": [
            "Aspidites/_vendor/contracts/inspection.py"
        ]
    },
    "module_name": "Aspidites._vendor.contracts.inspection"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__Aspidites___vendor__contracts__inspection
#define __PYX_HAVE_API__Aspidites___vendor__contracts__inspection
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 1
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING "utf8"
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "Aspidites/_vendor/contracts/inspection.py",
};

/*--- Type declarations ---*/
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis;
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;
struct __pyx_defaults2;
typedef struct __pyx_defaults2 __pyx_defaults2;
struct __pyx_defaults3;
typedef struct __pyx_defaults3 __pyx_defaults3;
struct __pyx_defaults4;
typedef struct __pyx_defaults4 __pyx_defaults4;
struct __pyx_defaults5;
typedef struct __pyx_defaults5 __pyx_defaults5;
struct __pyx_defaults6;
typedef struct __pyx_defaults6 __pyx_defaults6;
struct __pyx_defaults7;
typedef struct __pyx_defaults7 __pyx_defaults7;
struct __pyx_defaults8;
typedef struct __pyx_defaults8 __pyx_defaults8;
struct __pyx_defaults {
  PyObject *__pyx_arg_kwonlydefaults;
  PyObject *__pyx_arg_annotations;
  PyObject *__pyx_arg_formatarg;
  PyObject *__pyx_arg_formatvarargs;
  PyObject *__pyx_arg_formatvarkw;
  PyObject *__pyx_arg_formatvalue;
  PyObject *__pyx_arg_formatreturns;
  PyObject *__pyx_arg_formatannotation;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_formatarg;
  PyObject *__pyx_arg_formatvarargs;
  PyObject *__pyx_arg_formatvarkw;
  PyObject *__pyx_arg_formatvalue;
};
struct __pyx_defaults2 {
  PyObject *__pyx_arg_default;
};
struct __pyx_defaults3 {
  PyObject *__pyx_arg_default;
  PyObject *__pyx_arg_annotation;
};
struct __pyx_defaults4 {
  PyObject *__pyx_arg_name;
  PyObject *__pyx_arg_kind;
  PyObject *__pyx_arg_annotation;
  PyObject *__pyx_arg_default;
};
struct __pyx_defaults5 {
  PyObject *__pyx_arg_return_annotation;
};
struct __pyx_defaults6 {
  PyObject *__pyx_arg_parameters;
  PyObject *__pyx_arg_return_annotation;
};
struct __pyx_defaults7 {
  PyObject *__pyx_arg_locals;
  PyObject *__pyx_arg_formatarg;
  PyObject *__pyx_arg_formatvalue;
};
struct __pyx_defaults8 {
  PyObject *__pyx_arg_default;
};

/* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs {
  PyObject_HEAD
  PyObject *__pyx_v_metamro;
};


/* "Aspidites/_vendor/contracts/inspection.py":434
 *     mro = getmro(cls)
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))             # <<<<<<<<<<<<<<
 *     class_bases = (cls,) + mro
 *     all_bases = class_bases + metamro
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *__pyx_outer_scope;
  PyObject *__pyx_v_cls;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap {
  PyObject_HEAD
  PyObject *__pyx_v_stop;
};


/* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile {
  PyObject_HEAD
  PyObject *__pyx_v_all_bytecode_suffixes;
  PyObject *__pyx_v_filename;
  PyObject *__pyx_v_object;
};


/* "Aspidites/_vendor/contracts/inspection.py":743
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):             # <<<<<<<<<<<<<<
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *__pyx_outer_scope;
  PyObject *__pyx_v_s;
};


/* "Aspidites/_vendor/contracts/inspection.py":746
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in             # <<<<<<<<<<<<<<
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *__pyx_outer_scope;
  PyObject *__pyx_v_s;
};


/* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto {
  PyObject_HEAD
  PyObject *__pyx_v_module;
};


/* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec {
  PyObject_HEAD
  PyObject *__pyx_v_annotations;
  PyObject *__pyx_v_formatannotation;
  PyObject *__pyx_v_formatarg;
};


/* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs {
  PyObject_HEAD
  PyObject *__pyx_v_arg2value;
  PyObject *__pyx_v_assign;
  PyObject *__pyx_v_assigned_tuple_params;
};


/* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr {
  PyObject_HEAD
  PyObject *__pyx_v_Parameter;
  PyObject *__pyx_v_RewriteSymbolics;
  PyObject *__pyx_v_ast;
  PyObject *__pyx_v_empty;
  PyObject *__pyx_v_invalid;
  PyObject *__pyx_v_kind;
  PyObject *__pyx_v_module_dict;
  PyObject *__pyx_v_parameters;
  PyObject *__pyx_v_parse_name;
  PyObject *__pyx_v_sys_module_dict;
  PyObject *__pyx_v_wrap_value;
};


/* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_parameters;
};


/* "Aspidites/_vendor/contracts/inspection.py":2914
 *                     params[name] = param
 *             else:
 *                 params = OrderedDict((param.name, param) for param in parameters)             # <<<<<<<<<<<<<<
 * 
 *         self._parameters = types.MappingProxyType(params)
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_param;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/contracts/inspection.py":2973
 * 
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *__pyx_outer_scope;
  PyObject *__pyx_v_param;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* pop_index.proto */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix);
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix);
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (likely(PyList_CheckExact(L) && __Pyx_fits_Py_ssize_t(ix, type, is_signed))) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    __Pyx_fits_Py_ssize_t(ix, type, is_signed) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#else
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)\
    __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
        __Pyx__PyObject_PopIndex(L, py_ix))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* RaiseKeywordRequired.proto */
static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'Aspidites._vendor.contracts.inspection' */
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr = 0;
#define __Pyx_MODULE_NAME "Aspidites._vendor.contracts.inspection"
extern int __pyx_module_is_main_Aspidites___vendor__contracts__inspection;
int __pyx_module_is_main_Aspidites___vendor__contracts__inspection = 0;

/* Implementation of 'Aspidites._vendor.contracts.inspection' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_all;
static PyObject *__pyx_builtin_property;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_classmethod;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_UnicodeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_IndentationError;
static PyObject *__pyx_builtin_DeprecationWarning;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_SyntaxError;
static PyObject *__pyx_builtin_reversed;
static PyObject *__pyx_builtin_eval;
static PyObject *__pyx_builtin_NameError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_NotImplemented;
static const char __pyx_k_[] = "*";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_r[] = "{}={!r}";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_CO[] = "CO_";
static const char __pyx_k_NL[] = "NL";
static const char __pyx_k_OP[] = "OP";
static const char __pyx_k__2[] = "**";
static const char __pyx_k__3[] = "=";
static const char __pyx_k__4[] = " -> ";
static const char __pyx_k_ba[] = "ba";
static const char __pyx_k_co[] = "co";
static const char __pyx_k_eq[] = "__eq__";
static const char __pyx_k_ex[] = "ex";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_to[] = " to ";
static const char __pyx_k_tp[] = "tp";
static const char __pyx_k__10[] = ".";
static const char __pyx_k__12[] = "\n";
static const char __pyx_k__15[] = "<";
static const char __pyx_k__16[] = ">";
static const char __pyx_k__19[] = "";
static const char __pyx_k__25[] = "#!";
static const char __pyx_k__26[] = "#";
static const char __pyx_k__28[] = "@";
static const char __pyx_k__29[] = "(";
static const char __pyx_k__30[] = ")";
static const char __pyx_k__37[] = ": ";
static const char __pyx_k__41[] = ", ";
static const char __pyx_k__44[] = " ";
static const char __pyx_k__51[] = " (";
static const char __pyx_k__52[] = "'";
static const char __pyx_k__55[] = "($";
static const char __pyx_k__56[] = ",";
static const char __pyx_k__57[] = ":";
static const char __pyx_k__58[] = "/";
static const char __pyx_k__59[] = "$";
static const char __pyx_k__73[] = "{}: {}";
static const char __pyx_k__74[] = "{} = {}";
static const char __pyx_k__75[] = "{}={}";
static const char __pyx_k__76[] = "<{} \"{}\">";
static const char __pyx_k__77[] = "<{} ({})>";
static const char __pyx_k__82[] = "<{} {}>";
static const char __pyx_k__83[] = "({})";
static const char __pyx_k__84[] = " -> {}";
static const char __pyx_k_abc[] = "abc";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_all[] = "all";
static const char __pyx_k_and[] = "{} and {}";
static const char __pyx_k_ann[] = "ann";
static const char __pyx_k_arg[] = "arg";
static const char __pyx_k_ast[] = "ast";
static const char __pyx_k_but[] = " but ";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_ctx[] = "ctx";
static const char __pyx_k_def[] = "def";
static const char __pyx_k_dis[] = "dis";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_exc[] = "exc";
static const char __pyx_k_get[] = "__get__";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_mro[] = "__mro__";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pat[] = "pat";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_set[] = "__set__";
static const char __pyx_k_sig[] = "sig";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_var[] = "var";
static const char __pyx_k_was[] = "was";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Load[] = "Load";
static const char __pyx_k_Name[] = "Name";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_True[] = "True";
static const char __pyx_k_anno[] = "anno";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_attr[] = "attr";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_bind[] = "_bind";
static const char __pyx_k_body[] = "body";
static const char __pyx_k_call[] = "__call__";
static const char __pyx_k_cell[] = "cell";
static const char __pyx_k_code[] = "__code__";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_cpos[] = "cpos";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_eval[] = "eval";
static const char __pyx_k_fget[] = "fget";
static const char __pyx_k_file[] = "__file__";
static const char __pyx_k_find[] = "find";
static const char __pyx_k_flag[] = "flag";
static const char __pyx_k_from[] = "from ";
static const char __pyx_k_func[] = "__func__";
static const char __pyx_k_hash[] = "__hash__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_iter[] = "iter";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_last[] = "last";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_lnum[] = "lnum";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_memo[] = "memo";
static const char __pyx_k_meth[] = "meth";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_next[] = "__next__";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_pass[] = ": pass";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "__self__";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_spec[] = "spec";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_tail[] = "tail";
static const char __pyx_k_test[] = "test";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_tree[] = "tree";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_void[] = "_void";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_were[] = "were";
static const char __pyx_k_False[] = "False";
static const char __pyx_k_and_2[] = ", {} and {}";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_bases[] = "__bases__";
static const char __pyx_k_bound[] = "bound";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_chain[] = "chain";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_doc_2[] = "doc";
static const char __pyx_k_empty[] = "_empty";
static const char __pyx_k_entry[] = "entry";
static const char __pyx_k_flags[] = "__flags__";
static const char __pyx_k_fname[] = "fname";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_get_2[] = "get";
static const char __pyx_k_given[] = "given";
static const char __pyx_k_inPy2[] = "inPy2";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_klass[] = "klass";
static const char __pyx_k_kwarg[] = "kwarg";
static const char __pyx_k_lines[] = "lines";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_mro_2[] = "mro";
static const char __pyx_k_named[] = "named";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_nargs[] = "nargs";
static const char __pyx_k_new_2[] = "new";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_param[] = "param";
static const char __pyx_k_parse[] = "parse";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_roots[] = "roots";
static const char __pyx_k_slots[] = "__slots__";
static const char __pyx_k_specs[] = "specs";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_stack[] = "stack";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_takes[] = "() takes ";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_token[] = "token";
static const char __pyx_k_trace[] = "trace";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_varkw[] = "varkw";
static const char __pyx_k_visit[] = "visit";
static const char __pyx_k_DEDENT[] = "DEDENT";
static const char __pyx_k_INDENT[] = "INDENT";
static const char __pyx_k_Module[] = "Module";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_assign[] = "assign";
static const char __pyx_k_call_2[] = "call";
static const char __pyx_k_code_2[] = "code";
static const char __pyx_k_delete[] = "__delete__";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_exists[] = "exists";
static const char __pyx_k_f_back[] = "f_back";
static const char __pyx_k_f_code[] = "f_code";
static const char __pyx_k_f_name[] = "f_name";
static const char __pyx_k_fields[] = "_fields";
static const char __pyx_k_file_2[] = "file";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_func_2[] = "func";
static const char __pyx_k_getdoc[] = "getdoc";
static const char __pyx_k_getmro[] = "getmro";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indent[] = "indent";
static const char __pyx_k_init_2[] = "init";
static const char __pyx_k_iscode[] = "iscode";
static const char __pyx_k_kind_2[] = "_kind";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_kwonly[] = "kwonly";
static const char __pyx_k_lambda[] = "lambda";
static const char __pyx_k_lineno[] = "lineno";
static const char __pyx_k_loader[] = "__loader__";
static const char __pyx_k_locals[] = "<locals>";
static const char __pyx_k_lstrip[] = "lstrip";
static const char __pyx_k_main_2[] = "main";
static const char __pyx_k_margin[] = "margin";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_name_3[] = "_name_";
static const char __pyx_k_name_4[] = "_name";
static const char __pyx_k_neglen[] = "neglen";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_plural[] = "plural";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_self_2[] = "self";
static const char __pyx_k_self_3[] = "_self";
static const char __pyx_k_sigcls[] = "sigcls";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_subarg[] = "subarg";
static const char __pyx_k_suffix[] = "suffix";
static const char __pyx_k_test_2[] = "__test__";
static const char __pyx_k_tokens[] = "tokens";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_unique[] = "unique";
static const char __pyx_k_unwrap[] = "unwrap";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_vararg[] = "vararg";
static const char __pyx_k_ABCMeta[] = "ABCMeta";
static const char __pyx_k_ArgInfo[] = "ArgInfo";
static const char __pyx_k_ArgSpec[] = "ArgSpec";
static const char __pyx_k_COMMENT[] = "COMMENT";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_NEWLINE[] = "NEWLINE";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_abspath[] = "abspath";
static const char __pyx_k_arg_val[] = "arg_val";
static const char __pyx_k_at_most[] = "at most";
static const char __pyx_k_atleast[] = "atleast";
static const char __pyx_k_builtin[] = "builtin";
static const char __pyx_k_class_2[] = "class";
static const char __pyx_k_classes[] = "classes";
static const char __pyx_k_closure[] = "__closure__";
static const char __pyx_k_co_name[] = "co_name";
static const char __pyx_k_comment[] = "comment";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_content[] = "content";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_convert[] = "convert";
static const char __pyx_k_def_foo[] = "def foo";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_empty_2[] = "empty";
static const char __pyx_k_erowcol[] = "erowcol";
static const char __pyx_k_exactly[] = "exactly";
static const char __pyx_k_expline[] = "expline";
static const char __pyx_k_f_lasti[] = "f_lasti";
static const char __pyx_k_finddoc[] = "_finddoc";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_get_obj[] = "get_obj";
static const char __pyx_k_getargs[] = "getargs";
static const char __pyx_k_getattr[] = "__getattr__";
static const char __pyx_k_getfile[] = "getfile";
static const char __pyx_k_gi_code[] = "gi_code";
static const char __pyx_k_given_2[] = " given";
static const char __pyx_k_given_3[] = " given)";
static const char __pyx_k_globals[] = "__globals__";
static const char __pyx_k_homecls[] = "homecls";
static const char __pyx_k_id_func[] = "id_func";
static const char __pyx_k_im_self[] = "im_self";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_invalid[] = "invalid";
static const char __pyx_k_is_type[] = "_is_type";
static const char __pyx_k_isclass[] = "isclass";
static const char __pyx_k_isdigit[] = "isdigit";
static const char __pyx_k_isframe[] = "isframe";
static const char __pyx_k_maxsize[] = "maxsize";
static const char __pyx_k_metamro[] = "metamro";
static const char __pyx_k_missing[] = "missing";
static const char __pyx_k_modname[] = "modname";
static const char __pyx_k_modules[] = "modules";
static const char __pyx_k_nkwargs[] = "nkwargs";
static const char __pyx_k_num_pos[] = "num_pos";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_program[] = "program";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_results[] = "results";
static const char __pyx_k_slots_2[] = "slots";
static const char __pyx_k_srowcol[] = "srowcol";
static const char __pyx_k_started[] = "started";
static const char __pyx_k_tb_next[] = "tb_next";
static const char __pyx_k_token_2[] = "_token";
static const char __pyx_k_value_2[] = "value ";
static const char __pyx_k_varargs[] = "varargs";
static const char __pyx_k_weakref[] = "__weakref__";
static const char __pyx_k_wrapped[] = "__wrapped__";
static const char __pyx_k_Callable[] = "Callable";
static const char __pyx_k_CodeType[] = "CodeType";
static const char __pyx_k_Constant[] = "Constant";
static const char __pyx_k_ENCODING[] = "ENCODING";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_arg2dict[] = "arg2dict";
static const char __pyx_k_arg_vals[] = "arg_vals";
static const char __pyx_k_argnames[] = "argnames";
static const char __pyx_k_argument[] = "argument";
static const char __pyx_k_at_least[] = "at least ";
static const char __pyx_k_basename[] = "basename";
static const char __pyx_k_builtins[] = "builtins";
static const char __pyx_k_children[] = "children";
static const char __pyx_k_cleandoc[] = "cleandoc";
static const char __pyx_k_co_flags[] = "co_flags";
static const char __pyx_k_co_names[] = "co_names";
static const char __pyx_k_comments[] = "comments";
static const char __pyx_k_cr_frame[] = "cr_frame";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_defcount[] = "defcount";
static const char __pyx_k_dict_obj[] = "dict_obj";
static const char __pyx_k_endswith[] = "endswith";
static const char __pyx_k_exc_info[] = "exc_info";
static const char __pyx_k_f_lineno[] = "f_lineno";
static const char __pyx_k_f_locals[] = "f_locals";
static const char __pyx_k_filename[] = "_filename";
static const char __pyx_k_getblock[] = "getblock";
static const char __pyx_k_getframe[] = "_getframe";
static const char __pyx_k_getlines[] = "getlines";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_gi_frame[] = "gi_frame";
static const char __pyx_k_implicit[] = "implicit{}";
static const char __pyx_k_islambda[] = "islambda";
static const char __pyx_k_ismethod[] = "ismethod";
static const char __pyx_k_ismodule[] = "ismodule";
static const char __pyx_k_keywords[] = "keywords";
static const char __pyx_k_last_cls[] = "last_cls";
static const char __pyx_k_locals_2[] = "locals";
static const char __pyx_k_mod_dict[] = "mod_dict";
static const char __pyx_k_module_2[] = "<module>";
static const char __pyx_k_module_3[] = "module";
static const char __pyx_k_normcase[] = "normcase";
static const char __pyx_k_num_args[] = "num_args";
static const char __pyx_k_objclass[] = "__objclass__";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_passline[] = "passline";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_realpath[] = "realpath";
static const char __pyx_k_rendered[] = "rendered";
static const char __pyx_k_reversed[] = "reversed";
static const char __pyx_k_sentinel[] = "_sentinel";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_splitext[] = "splitext";
static const char __pyx_k_srch_cls[] = "srch_cls";
static const char __pyx_k_srch_obj[] = "srch_obj";
static const char __pyx_k_subvalue[] = "subvalue";
static const char __pyx_k_suffixes[] = "suffixes";
static const char __pyx_k_tb_frame[] = "tb_frame";
static const char __pyx_k_text_sig[] = "text_sig";
static const char __pyx_k_tokenize[] = "tokenize";
static const char __pyx_k_too_many[] = "_too_many";
static const char __pyx_k_top_kind[] = "top_kind";
static const char __pyx_k_typing_2[] = "typing.";
static const char __pyx_k_walktree[] = "walktree";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_Arguments[] = "Arguments";
static const char __pyx_k_Attribute[] = "Attribute";
static const char __pyx_k_Awaitable[] = "Awaitable";
static const char __pyx_k_ClassType[] = "ClassType";
static const char __pyx_k_FrameInfo[] = "FrameInfo";
static const char __pyx_k_FrameType[] = "FrameType";
static const char __pyx_k_NameError[] = "NameError";
static const char __pyx_k_Parameter[] = "Parameter";
static const char __pyx_k_Signature[] = "Signature";
static const char __pyx_k_Traceback[] = "Traceback";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_all_bases[] = "all_bases";
static const char __pyx_k_arg2value[] = "arg2value";
static const char __pyx_k_arg_names[] = "arg_names";
static const char __pyx_k_arg_value[] = "arg_value";
static const char __pyx_k_arguments[] = "arguments";
static const char __pyx_k_body_col0[] = "body_col0";
static const char __pyx_k_coroutine[] = "coroutine";
static const char __pyx_k_default_2[] = "_default";
static const char __pyx_k_dict_attr[] = "dict_attr";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_fillvalue[] = "fillvalue";
static const char __pyx_k_findclass[] = "_findclass";
static const char __pyx_k_formatarg[] = "formatarg";
static const char __pyx_k_formatted[] = "formatted";
static const char __pyx_k_frameinfo[] = "frameinfo";
static const char __pyx_k_framelist[] = "framelist";
static const char __pyx_k_func_code[] = "func_code";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_generator[] = "generator";
static const char __pyx_k_getlineno[] = "getlineno";
static const char __pyx_k_getmodule[] = "getmodule";
static const char __pyx_k_getsource[] = "getsource";
static const char __pyx_k_global_ns[] = "global_ns";
static const char __pyx_k_importlib[] = "importlib";
static const char __pyx_k_isbuiltin[] = "isbuiltin";
static const char __pyx_k_isroutine[] = "isroutine";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_kwonlyarg[] = "kwonlyarg";
static const char __pyx_k_linecache[] = "linecache";
static const char __pyx_k_machinery[] = "machinery";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_name_node[] = "name_node";
static const char __pyx_k_new_param[] = "new_param";
static const char __pyx_k_num_total[] = "num_total";
static const char __pyx_k_pos_count[] = "pos_count";
static const char __pyx_k_predicate[] = "predicate";
static const char __pyx_k_processed[] = "processed";
static const char __pyx_k_signature[] = "__signature__";
static const char __pyx_k_tb_lineno[] = "tb_lineno";
static const char __pyx_k_to_unpack[] = " to unpack";
static const char __pyx_k_ArityError[] = "ArityError";
static const char __pyx_k_CO_VARARGS[] = "CO_VARARGS";
static const char __pyx_k_ERRORTOKEN[] = "ERRORTOKEN";
static const char __pyx_k_EndOfBlock[] = "EndOfBlock";
static const char __pyx_k_GEN_CLOSED[] = "GEN_CLOSED";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_MethodType[] = "MethodType";
static const char __pyx_k_ModuleType[] = "ModuleType";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_annotation[] = "annotation";
static const char __pyx_k_at_least_2[] = "at least";
static const char __pyx_k_attrgetter[] = "attrgetter";
static const char __pyx_k_builtin_ns[] = "builtin_ns";
static const char __pyx_k_builtins_2[] = "__builtins__";
static const char __pyx_k_checkcache[] = "checkcache";
static const char __pyx_k_class_dict[] = "class_dict";
static const char __pyx_k_cr_running[] = "cr_running";
static const char __pyx_k_defaults_2[] = "__defaults__";
static const char __pyx_k_expandtabs[] = "expandtabs";
static const char __pyx_k_extra_args[] = "extra_args";
static const char __pyx_k_filename_2[] = "filename";
static const char __pyx_k_findsource[] = "findsource";
static const char __pyx_k_from_bytes[] = "from_bytes";
static const char __pyx_k_getabsfile[] = "getabsfile";
static const char __pyx_k_getargspec[] = "getargspec";
static const char __pyx_k_getmembers[] = "getmembers";
static const char __pyx_k_gi_running[] = "gi_running";
static const char __pyx_k_hash_basis[] = "_hash_basis";
static const char __pyx_k_indentsize[] = "indentsize";
static const char __pyx_k_is_wrapper[] = "_is_wrapper";
static const char __pyx_k_isabstract[] = "isabstract";
static const char __pyx_k_isasyncgen[] = "isasyncgen";
static const char __pyx_k_isfunction[] = "isfunction";
static const char __pyx_k_kwdefaults[] = "__kwdefaults__";
static const char __pyx_k_kwo_params[] = "kwo_params";
static const char __pyx_k_kwonly_sig[] = "kwonly_sig";
static const char __pyx_k_kwonlyargs[] = "kwonlyargs";
static const char __pyx_k_mainobject[] = "mainobject";
static const char __pyx_k_namedtuple[] = "namedtuple";
static const char __pyx_k_new_params[] = "new_params";
static const char __pyx_k_old_params[] = "old_params";
static const char __pyx_k_param_name[] = "param_name";
static const char __pyx_k_parameters[] = "parameters";
static const char __pyx_k_parse_name[] = "parse_name";
static const char __pyx_k_positional[] = "positional";
static const char __pyx_k_qualname_2[] = "qualname";
static const char __pyx_k_sig_params[] = "sig_params";
static const char __pyx_k_stacklevel[] = "stacklevel";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_tokeneater[] = "tokeneater";
static const char __pyx_k_unassigned[] = "unassigned";
static const char __pyx_k_unexpected[] = "unexpected";
static const char __pyx_k_visit_Name[] = "visit_Name";
static const char __pyx_k_wrap_value[] = "wrap_value";
static const char __pyx_k_BlockFinder[] = "BlockFinder";
static const char __pyx_k_CORO_CLOSED[] = "CORO_CLOSED";
static const char __pyx_k_ClassFinder[] = "_ClassFinder";
static const char __pyx_k_ClosureVars[] = "ClosureVars";
static const char __pyx_k_FullArgSpec[] = "FullArgSpec";
static const char __pyx_k_GEN_CREATED[] = "GEN_CREATED";
static const char __pyx_k_GEN_RUNNING[] = "GEN_RUNNING";
static const char __pyx_k_InvalidArgs[] = "InvalidArgs";
static const char __pyx_k_NodeVisitor[] = "NodeVisitor";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_SyntaxError[] = "SyntaxError";
static const char __pyx_k_VAR_KEYWORD[] = "_VAR_KEYWORD";
static const char __pyx_k_annotations[] = "annotations";
static const char __pyx_k_base_module[] = "base_module";
static const char __pyx_k_blockfinder[] = "blockfinder";
static const char __pyx_k_check_class[] = "_check_class";
static const char __pyx_k_class_bases[] = "class_bases";
static const char __pyx_k_classmethod[] = "classmethod";
static const char __pyx_k_co_argcount[] = "co_argcount";
static const char __pyx_k_co_filename[] = "co_filename";
static const char __pyx_k_co_freevars[] = "co_freevars";
static const char __pyx_k_co_varnames[] = "co_varnames";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_description[] = "description";
static const char __pyx_k_formatvalue[] = "formatvalue";
static const char __pyx_k_formatvarkw[] = "formatvarkw";
static const char __pyx_k_getcallargs[] = "getcallargs";
static const char __pyx_k_getcomments[] = "getcomments";
static const char __pyx_k_global_vars[] = "global_vars";
static const char __pyx_k_indecorator[] = "indecorator";
static const char __pyx_k_is_assigned[] = "is_assigned";
static const char __pyx_k_isawaitable[] = "isawaitable";
static const char __pyx_k_iscoroutine[] = "iscoroutine";
static const char __pyx_k_isgenerator[] = "isgenerator";
static const char __pyx_k_istraceback[] = "istraceback";
static const char __pyx_k_kw_defaults[] = "kw_defaults";
static const char __pyx_k_line_number[] = "line_number";
static const char __pyx_k_method_name[] = "method_name";
static const char __pyx_k_module_dict[] = "module_dict";
static const char __pyx_k_module_name[] = "module_name";
static const char __pyx_k_move_to_end[] = "move_to_end";
static const char __pyx_k_signature_2[] = "signature";
static const char __pyx_k_signature_3[] = "_signature";
static const char __pyx_k_wrapped_sig[] = "wrapped_sig";
static const char __pyx_k_zip_longest[] = "zip_longest";
static const char __pyx_k_CORO_CREATED[] = "CORO_CREATED";
static const char __pyx_k_CORO_RUNNING[] = "CORO_RUNNING";
static const char __pyx_k_CO_COROUTINE[] = "CO_COROUTINE";
static const char __pyx_k_CO_GENERATOR[] = "CO_GENERATOR";
static const char __pyx_k_FunctionType[] = "FunctionType";
static const char __pyx_k_KEYWORD_ONLY[] = "_KEYWORD_ONLY";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_UnicodeError[] = "UnicodeError";
static const char __pyx_k_all_suffixes[] = "all_suffixes";
static const char __pyx_k_annotation_2[] = "_annotation";
static const char __pyx_k_bind_partial[] = "bind_partial";
static const char __pyx_k_builtin_vars[] = "builtin_vars";
static const char __pyx_k_class_finder[] = "class_finder";
static const char __pyx_k_class_method[] = "class method";
static const char __pyx_k_currentframe[] = "currentframe";
static const char __pyx_k_default_node[] = "default_node";
static const char __pyx_k_firstdefault[] = "firstdefault";
static const char __pyx_k_from_builtin[] = "from_builtin";
static const char __pyx_k_getargvalues[] = "getargvalues";
static const char __pyx_k_getattribute[] = "__getattribute__";
static const char __pyx_k_getclasstree[] = "getclasstree";
static const char __pyx_k_getframeinfo[] = "getframeinfo";
static const char __pyx_k_isidentifier[] = "isidentifier";
static const char __pyx_k_keyword_only[] = "keyword-only";
static const char __pyx_k_klass_result[] = "klass_result";
static const char __pyx_k_kwargs_param[] = "kwargs_param";
static const char __pyx_k_kwdefaults_2[] = "kwdefaults";
static const char __pyx_k_kwonly_given[] = "kwonly_given";
static const char __pyx_k_literal_eval[] = "literal_eval";
static const char __pyx_k_num_defaults[] = "num_defaults";
static const char __pyx_k_num_required[] = "num_required";
static const char __pyx_k_parameters_2[] = "_parameters";
static const char __pyx_k_partial_args[] = "partial_args";
static const char __pyx_k_posonly_left[] = "posonly_left";
static const char __pyx_k_self_isbound[] = "self_isbound";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_token_stream[] = "token_stream";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_CoroutineType[] = "CoroutineType";
static const char __pyx_k_GEN_SUSPENDED[] = "GEN_SUSPENDED";
static const char __pyx_k_GeneratorType[] = "GeneratorType";
static const char __pyx_k_MethodWrapper[] = "_MethodWrapper";
static const char __pyx_k_ParameterKind[] = "_ParameterKind";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_TracebackType[] = "TracebackType";
static const char __pyx_k_VAR_KEYWORD_2[] = "VAR_KEYWORD";
static const char __pyx_k_annotations_2[] = "__annotations__";
static const char __pyx_k_builtinobject[] = "builtinobject";
static const char __pyx_k_cell_contents[] = "cell_contents";
static const char __pyx_k_delayed_comma[] = "delayed_comma";
static const char __pyx_k_fn_underscore[] = "fn.underscore";
static const char __pyx_k_formatargspec[] = "formatargspec";
static const char __pyx_k_formatreturns[] = "formatreturns";
static const char __pyx_k_formatvarargs[] = "formatvarargs";
static const char __pyx_k_from_callable[] = "from_callable";
static const char __pyx_k_from_function[] = "from_function";
static const char __pyx_k_generic_visit[] = "generic_visit";
static const char __pyx_k_getmodulename[] = "getmodulename";
static const char __pyx_k_getsourcefile[] = "getsourcefile";
static const char __pyx_k_has_code_flag[] = "_has_code_flag";
static const char __pyx_k_instance_dict[] = "instance_dict";
static const char __pyx_k_kind_defaults[] = "kind_defaults";
static const char __pyx_k_modulesbyfile[] = "modulesbyfile";
static const char __pyx_k_new_arguments[] = "new_arguments";
static const char __pyx_k_nonlocal_vars[] = "nonlocal_vars";
static const char __pyx_k_parameter_cls[] = "_parameter_cls";
static const char __pyx_k_parameters_ex[] = "parameters_ex";
static const char __pyx_k_partialmethod[] = "_partialmethod";
static const char __pyx_k_posonly_count[] = "posonly_count";
static const char __pyx_k_self_ismodule[] = "self_ismodule";
static const char __pyx_k_shadowed_dict[] = "_shadowed_dict";
static const char __pyx_k_static_getmro[] = "_static_getmro";
static const char __pyx_k_static_method[] = "static method";
static const char __pyx_k_unbound_names[] = "unbound_names";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_BoundArguments[] = "BoundArguments";
static const char __pyx_k_CORO_SUSPENDED[] = "CORO_SUSPENDED";
static const char __pyx_k_CO_VARKEYWORDS[] = "CO_VARKEYWORDS";
static const char __pyx_k_KEYWORD_ONLY_2[] = "KEYWORD_ONLY";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_Parameter___eq[] = "Parameter.__eq__";
static const char __pyx_k_Parameter_kind[] = "Parameter.kind";
static const char __pyx_k_Parameter_name[] = "Parameter.name";
static const char __pyx_k_Signature___eq[] = "Signature.__eq__";
static const char __pyx_k_VAR_POSITIONAL[] = "_VAR_POSITIONAL";
static const char __pyx_k_apply_defaults[] = "apply_defaults";
static const char __pyx_k_callable_thing[] = "callable_thing";
static const char __pyx_k_check_instance[] = "_check_instance";
static const char __pyx_k_co_firstlineno[] = "co_firstlineno";
static const char __pyx_k_decorator_list[] = "decorator_list";
static const char __pyx_k_filesbymodname[] = "_filesbymodname";
static const char __pyx_k_follow_wrapped[] = "follow_wrapped";
static const char __pyx_k_getattr_static[] = "getattr_static";
static const char __pyx_k_getclosurevars[] = "getclosurevars";
static const char __pyx_k_getfullargspec[] = "getfullargspec";
static const char __pyx_k_getinnerframes[] = "getinnerframes";
static const char __pyx_k_getouterframes[] = "getouterframes";
static const char __pyx_k_getsourcelines[] = "getsourcelines";
static const char __pyx_k_keyword_only_2[] = "keyword_only";
static const char __pyx_k_kwargs_started[] = "kwargs_started";
static const char __pyx_k_kwonlydefaults[] = "kwonlydefaults";
static const char __pyx_k_need_more_than[] = "need more than ";
static const char __pyx_k_self_parameter[] = "self_parameter";
static const char __pyx_k_skip_bound_arg[] = "skip_bound_arg";
static const char __pyx_k_text_signature[] = "__text_signature__";
static const char __pyx_k_unwrap_partial[] = "_unwrap_partial";
static const char __pyx_k_visit_ClassDef[] = "visit_ClassDef";
static const char __pyx_k_NodeTransformer[] = "NodeTransformer";
static const char __pyx_k_POSITIONAL_ONLY[] = "_POSITIONAL_ONLY";
static const char __pyx_k_Parameter___str[] = "Parameter.__str__";
static const char __pyx_k_SOURCE_SUFFIXES[] = "SOURCE_SUFFIXES";
static const char __pyx_k_Signature___str[] = "Signature.__str__";
static const char __pyx_k_Signature__bind[] = "Signature._bind";
static const char __pyx_k_abstractmethods[] = "__abstractmethods__";
static const char __pyx_k_can_accept_self[] = "can_accept_self";
static const char __pyx_k_clean_signature[] = "clean_signature";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_does_not_accept[] = " does not accept ";
static const char __pyx_k_formatargvalues[] = "formatargvalues";
static const char __pyx_k_generate_tokens[] = "generate_tokens";
static const char __pyx_k_instance_result[] = "instance_result";
static const char __pyx_k_partialmethod_2[] = "partialmethod";
static const char __pyx_k_positional_only[] = "positional-only";
static const char __pyx_k_recursion_limit[] = "recursion_limit";
static const char __pyx_k_skip_next_comma[] = "skip_next_comma";
static const char __pyx_k_sys_module_dict[] = "sys_module_dict";
static const char __pyx_k_visit_Attribute[] = "visit_Attribute";
static const char __pyx_k_IndentationError[] = "IndentationError";
static const char __pyx_k_MappingProxyType[] = "MappingProxyType";
static const char __pyx_k_Parameter___hash[] = "Parameter.__hash__";
static const char __pyx_k_Parameter___init[] = "Parameter.__init__";
static const char __pyx_k_Parameter___repr[] = "Parameter.__repr__";
static const char __pyx_k_RewriteSymbolics[] = "RewriteSymbolics";
static const char __pyx_k_Signature___hash[] = "Signature.__hash__";
static const char __pyx_k_Signature___init[] = "Signature.__init__";
static const char __pyx_k_Signature___repr[] = "Signature.__repr__";
static const char __pyx_k_VAR_POSITIONAL_2[] = "VAR_POSITIONAL";
static const char __pyx_k_decoratorhasargs[] = "decoratorhasargs";
static const char __pyx_k_formatannotation[] = "formatannotation";
static const char __pyx_k_is_duck_function[] = "is_duck_function";
static const char __pyx_k_isabstractmethod[] = "__isabstractmethod__";
static const char __pyx_k_isdatadescriptor[] = "isdatadescriptor";
static const char __pyx_k_partial_keywords[] = "partial_keywords";
static const char __pyx_k_variadic_keyword[] = "variadic keyword";
static const char __pyx_k_BuiltinMethodType[] = "BuiltinMethodType";
static const char __pyx_k_POSITIONAL_ONLY_2[] = "POSITIONAL_ONLY";
static const char __pyx_k_Parameter_default[] = "Parameter.default";
static const char __pyx_k_Parameter_replace[] = "Parameter.replace";
static const char __pyx_k_Signature_replace[] = "Signature.replace";
static const char __pyx_k_WrapperDescriptor[] = "_WrapperDescriptor";
static const char __pyx_k_co_kwonlyargcount[] = "co_kwonlyargcount";
static const char __pyx_k_current_parameter[] = "current_parameter";
static const char __pyx_k_getcoroutinestate[] = "getcoroutinestate";
static const char __pyx_k_getgeneratorstate[] = "getgeneratorstate";
static const char __pyx_k_getrecursionlimit[] = "getrecursionlimit";
static const char __pyx_k_missing_arguments[] = "_missing_arguments";
static const char __pyx_k_non_default_count[] = "non_default_count";
static const char __pyx_k_pos_default_count[] = "pos_default_count";
static const char __pyx_k_return_annotation[] = "return_annotation";
static const char __pyx_k_signature_fromstr[] = "_signature_fromstr";
static const char __pyx_k_visit_FunctionDef[] = "visit_FunctionDef";
static const char __pyx_k_AsyncGeneratorType[] = "AsyncGeneratorType";
static const char __pyx_k_BlockFinder___init[] = "BlockFinder.__init__";
static const char __pyx_k_CO_ASYNC_GENERATOR[] = "CO_ASYNC_GENERATOR";
static const char __pyx_k_ClassFinder___init[] = "_ClassFinder.__init__";
static const char __pyx_k_ClassMethodWrapper[] = "_ClassMethodWrapper";
static const char __pyx_k_DeprecationWarning[] = "DeprecationWarning";
static const char __pyx_k_EXTENSION_SUFFIXES[] = "EXTENSION_SUFFIXES";
static const char __pyx_k_PARAM_NAME_MAPPING[] = "_PARAM_NAME_MAPPING";
static const char __pyx_k_Parameter___reduce[] = "Parameter.__reduce__";
static const char __pyx_k_Signature___reduce[] = "Signature.__reduce__";
static const char __pyx_k_args_varargs_varkw[] = "args, varargs, varkw";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_co_posonlyargcount[] = "co_posonlyargcount";
static const char __pyx_k_formatannotation_2[] = "_formatannotation";
static const char __pyx_k_getcoroutinelocals[] = "getcoroutinelocals";
static const char __pyx_k_getdefaultencoding[] = "getdefaultencoding";
static const char __pyx_k_getgeneratorlocals[] = "getgeneratorlocals";
static const char __pyx_k_isasyncgenfunction[] = "isasyncgenfunction";
static const char __pyx_k_isgetsetdescriptor[] = "isgetsetdescriptor";
static const char __pyx_k_ismemberdescriptor[] = "ismemberdescriptor";
static const char __pyx_k_ismethoddescriptor[] = "ismethoddescriptor";
static const char __pyx_k_keyword_only_count[] = "keyword_only_count";
static const char __pyx_k_takes_no_arguments[] = "() takes no arguments (";
static const char __pyx_k_BoundArguments___eq[] = "BoundArguments.__eq__";
static const char __pyx_k_BoundArguments_args[] = "BoundArguments.args";
static const char __pyx_k_BuiltinFunctionType[] = "BuiltinFunctionType";
static const char __pyx_k_COMPILER_FLAG_NAMES[] = "COMPILER_FLAG_NAMES";
static const char __pyx_k_ClassFoundException[] = "ClassFoundException";
static const char __pyx_k_ParameterKind___str[] = "_ParameterKind.__str__";
static const char __pyx_k_TPFLAGS_IS_ABSTRACT[] = "TPFLAGS_IS_ABSTRACT";
static const char __pyx_k_bound_arguments_cls[] = "_bound_arguments_cls";
static const char __pyx_k_first_wrapped_param[] = "first_wrapped_param";
static const char __pyx_k_get_f_from_callable[] = "get_f_from_callable";
static const char __pyx_k_importlib_machinery[] = "importlib.machinery";
static const char __pyx_k_iscoroutinefunction[] = "iscoroutinefunction";
static const char __pyx_k_isgeneratorfunction[] = "isgeneratorfunction";
static const char __pyx_k_positional_argument[] = " positional argument";
static const char __pyx_k_return_annotation_2[] = "_return_annotation";
static const char __pyx_k_transform_to_kwonly[] = "transform_to_kwonly";
static const char __pyx_k_validate_parameters[] = "__validate_parameters__";
static const char __pyx_k_variadic_positional[] = "variadic positional";
static const char __pyx_k_GetSetDescriptorType[] = "GetSetDescriptorType";
static const char __pyx_k_MemberDescriptorType[] = "MemberDescriptorType";
static const char __pyx_k_Parameter___setstate[] = "Parameter.__setstate__";
static const char __pyx_k_Parameter_annotation[] = "Parameter.annotation";
static const char __pyx_k_Signature___setstate[] = "Signature.__setstate__";
static const char __pyx_k_Signature_parameters[] = "Signature.parameters";
static const char __pyx_k_classify_class_attrs[] = "classify_class_attrs";
static const char __pyx_k_last_positional_only[] = "last_positional_only";
static const char __pyx_k_signature_is_builtin[] = "_signature_is_builtin";
static const char __pyx_k_BoundArguments___init[] = "BoundArguments.__init__";
static const char __pyx_k_BoundArguments___repr[] = "BoundArguments.__repr__";
static const char __pyx_k_BoundArguments_kwargs[] = "BoundArguments.kwargs";
static const char __pyx_k_CO_ITERABLE_COROUTINE[] = "CO_ITERABLE_COROUTINE";
static const char __pyx_k_DynamicClassAttribute[] = "DynamicClassAttribute";
static const char __pyx_k_POSITIONAL_OR_KEYWORD[] = "_POSITIONAL_OR_KEYWORD";
static const char __pyx_k_Signature__hash_basis[] = "Signature._hash_basis";
static const char __pyx_k_all_bytecode_suffixes[] = "all_bytecode_suffixes";
static const char __pyx_k_assigned_tuple_params[] = "assigned_tuple_params";
static const char __pyx_k_follow_wrapper_chains[] = "follow_wrapper_chains";
static const char __pyx_k_invalid_argument_type[] = "invalid argument type";
static const char __pyx_k_positional_or_keyword[] = "positional or keyword";
static const char __pyx_k_r_is_a_built_in_class[] = "{!r} is a built-in class";
static const char __pyx_k_signature_get_partial[] = "_signature_get_partial";
static const char __pyx_k_BlockFinder_tokeneater[] = "BlockFinder.tokeneater";
static const char __pyx_k_Signature_from_builtin[] = "Signature.from_builtin";
static const char __pyx_k_formatargandannotation[] = "formatargandannotation";
static const char __pyx_k_r_is_a_built_in_module[] = "{!r} is a built-in module";
static const char __pyx_k_r_is_not_a_code_object[] = "{!r} is not a code object";
static const char __pyx_k_signature_bound_method[] = "_signature_bound_method";
static const char __pyx_k_signature_from_builtin[] = "_signature_from_builtin";
static const char __pyx_k_visit_AsyncFunctionDef[] = "visit_AsyncFunctionDef";
static const char __pyx_k_DEBUG_BYTECODE_SUFFIXES[] = "DEBUG_BYTECODE_SUFFIXES";
static const char __pyx_k_NonUserDefinedCallables[] = "_NonUserDefinedCallables";
static const char __pyx_k_POSITIONAL_OR_KEYWORD_2[] = "POSITIONAL_OR_KEYWORD";
static const char __pyx_k_Signature_from_callable[] = "Signature.from_callable";
static const char __pyx_k_Signature_from_function[] = "Signature.from_function";
static const char __pyx_k_lineno_is_out_of_bounds[] = "lineno is out of bounds";
static const char __pyx_k_signature_from_callable[] = "_signature_from_callable";
static const char __pyx_k_signature_from_function[] = "_signature_from_function";
static const char __pyx_k_BoundArguments_signature[] = "BoundArguments.signature";
static const char __pyx_k_get_callable_fullargspec[] = "get_callable_fullargspec";
static const char __pyx_k_getmembers_locals_lambda[] = "getmembers.<locals>.<lambda>";
static const char __pyx_k_invalid_method_signature[] = "invalid method signature";
static const char __pyx_k_name_must_be_a_str_not_a[] = "name must be a str, not a {}";
static const char __pyx_k_no_signature_found_for_r[] = "no signature found for {!r}";
static const char __pyx_k_render_kw_only_separator[] = "render_kw_only_separator";
static const char __pyx_k_BoundArguments___getstate[] = "BoundArguments.__getstate__";
static const char __pyx_k_BoundArguments___setstate[] = "BoundArguments.__setstate__";
static const char __pyx_k_ClassMethodDescriptorType[] = "ClassMethodDescriptorType";
static const char __pyx_k_ParameterKind_description[] = "_ParameterKind.description";
static const char __pyx_k_could_not_get_source_code[] = "could not get source code";
static const char __pyx_k_getcallargs_locals_assign[] = "getcallargs.<locals>.assign";
static const char __pyx_k_render_pos_only_separator[] = "render_pos_only_separator";
static const char __pyx_k_signature_get_bound_param[] = "_signature_get_bound_param";
static const char __pyx_k_signature_is_functionlike[] = "_signature_is_functionlike";
static const char __pyx_k_source_code_not_available[] = "source code not available";
static const char __pyx_k_too_many_values_to_unpack[] = "too many values to unpack";
static const char __pyx_k_unwrap_locals__is_wrapper[] = "unwrap.<locals>._is_wrapper";
static const char __pyx_k_ClassFinder_visit_ClassDef[] = "_ClassFinder.visit_ClassDef";
static const char __pyx_k_could_not_find_code_object[] = "could not find code object";
static const char __pyx_k_duplicate_parameter_name_r[] = "duplicate parameter name: {!r}";
static const char __pyx_k_formatannotationrelativeto[] = "formatannotationrelativeto";
static const char __pyx_k_r_is_not_a_Python_function[] = "{!r} is not a Python function";
static const char __pyx_k_r_is_not_a_callable_object[] = "{!r} is not a callable object";
static const char __pyx_k_signature_fromstr_locals_p[] = "_signature_fromstr.<locals>.p";
static const char __pyx_k_OPTIMIZED_BYTECODE_SUFFIXES[] = "OPTIMIZED_BYTECODE_SUFFIXES";
static const char __pyx_k_Signature_return_annotation[] = "Signature.return_annotation";
static const char __pyx_k_formatargspec_locals_lambda[] = "formatargspec.<locals>.<lambda>";
static const char __pyx_k_r_is_not_a_Python_generator[] = "{!r} is not a Python generator";
static const char __pyx_k_args_varargs_keywords_locals[] = "args varargs keywords locals";
static const char __pyx_k_getsourcefile_locals_genexpr[] = "getsourcefile.<locals>.genexpr";
static const char __pyx_k_BoundArguments_apply_defaults[] = "BoundArguments.apply_defaults";
static const char __pyx_k_ClassFinder_visit_FunctionDef[] = "_ClassFinder.visit_FunctionDef";
static const char __pyx_k_formatargvalues_locals_lambda[] = "formatargvalues.<locals>.<lambda>";
static const char __pyx_k_is_not_a_valid_Parameter_kind[] = " is not a valid Parameter.kind";
static const char __pyx_k_too_many_positional_arguments[] = "too many positional arguments";
static const char __pyx_k_args_varargs_keywords_defaults[] = "args varargs keywords defaults";
static const char __pyx_k_dict___is_special_don_t_want_t[] = "__dict__ is special, don't want the proxy";
static const char __pyx_k_formatargvalues_locals_convert[] = "formatargvalues.<locals>.convert";
static const char __pyx_k_getcallargs_locals_is_assigned[] = "getcallargs.<locals>.is_assigned";
static const char __pyx_k_parameters_cannot_have_default[] = "{} parameters cannot have default values";
static const char __pyx_k_signature_from_callable_locals[] = "_signature_from_callable.<locals>.<lambda>";
static const char __pyx_k_signature_fromstr_locals_parse[] = "_signature_fromstr.<locals>.parse_name";
static const char __pyx_k_wrapper_loop_when_unwrapping_r[] = "wrapper loop when unwrapping {!r}";
static const char __pyx_k_arg_r_parameter_is_positional_o[] = "{arg!r} parameter is positional only, but was passed as a keyword";
static const char __pyx_k_can_accept_exactly_one_argument[] = "can_accept_exactly_one_argument";
static const char __pyx_k_could_not_find_class_definition[] = "could not find class definition";
static const char __pyx_k_formatargspec_is_deprecated_sin[] = "`formatargspec` is deprecated since Python 3.5. Use `signature` and the `Signature` object directly";
static const char __pyx_k_got_an_unexpected_keyword_argum[] = "() got an unexpected keyword argument '";
static const char __pyx_k_got_multiple_values_for_keyword[] = "() got multiple values for keyword argument '";
static const char __pyx_k_inspect_Signature_from_function[] = "inspect.Signature.from_function() is deprecated since Python 3.5, use Signature.from_callable()";
static const char __pyx_k_missing_a_required_argument_arg[] = "missing a required argument: {arg!r}";
static const char __pyx_k_name_kind_defining_class_object[] = "name kind defining_class object";
static const char __pyx_k_positional_argument_s_and_d_key[] = " positional argument%s (and %d keyword-only argument%s)";
static const char __pyx_k_r_builtin_has_invalid_signature[] = "{!r} builtin has invalid signature";
static const char __pyx_k_r_is_not_a_Python_builtin_funct[] = "{!r} is not a Python builtin function";
static const char __pyx_k_r_is_not_a_frame_or_traceback_o[] = "{!r} is not a frame or traceback object";
static const char __pyx_k_r_is_not_a_valid_parameter_name[] = "{!r} is not a valid parameter name";
static const char __pyx_k_s_def_s_s_async_s_def_s_w_lambd[] = "^(\\s*def\\s)|(\\s*async\\s+def\\s)|(.*(?<!\\w)lambda(:|\\s))|^(\\s*@)";
static const char __pyx_k_s_missing_i_required_s_argument[] = "%s() missing %i required %s argument%s: %s";
static const char __pyx_k_signature_fromstr_locals_Rewrit[] = "_signature_fromstr.<locals>.RewriteSymbolics";
static const char __pyx_k_signature_fromstr_locals_wrap_v[] = "_signature_fromstr.<locals>.wrap_value";
static const char __pyx_k_signature_get_user_defined_meth[] = "_signature_get_user_defined_method";
static const char __pyx_k_signature_strip_non_python_synt[] = "_signature_strip_non_python_syntax";
static const char __pyx_k_wrong_parameter_order_parameter[] = "wrong parameter order: {} parameter before {} parameter";
static const char __pyx_k_A_Signature_object_represents_th[] = "A Signature object represents the overall signature of a function.\n    It stores a Parameter object for each parameter accepted by the\n    function, as well as information specific to the function itself.\n\n    A Signature object has the following public attributes and methods:\n\n    * parameters : OrderedDict\n        An ordered mapping of parameters' names to the corresponding\n        Parameter objects (keyword-only arguments are in the same order\n        as listed in `code.co_varnames`).\n    * return_annotation : object\n        The annotation for the return type of the function if specified.\n        If the function has no annotation for its return type, this\n        attribute is set to `Signature.empty`.\n    * bind(*args, **kwargs) -> BoundArguments\n        Creates a mapping from positional and keyword arguments to\n        parameters.\n    * bind_partial(*args, **kwargs) -> BoundArguments\n        Creates a partial mapping from positional and keyword arguments\n        to parameters (simulating 'functools.partial' behavior.)\n    ";
static const char __pyx_k_A_private_marker_used_in_Paramet[] = "A private marker - used in Parameter & Signature.";
static const char __pyx_k_Annotations_are_not_currently_su[] = "Annotations are not currently supported";
static const char __pyx_k_Aspidites__vendor_contracts_insp[] = "Aspidites._vendor.contracts.inspection";
static const char __pyx_k_Function_has_keyword_only_parame[] = "Function has keyword-only parameters or annotations, use inspect.signature() API which can support them";
static const char __pyx_k_Get_useful_information_from_live[] = "Get useful information from live Python objects.\nThis is a statically compiled version of the inspect module from the standard library.\n\nThis module encapsulates the interface provided by the internal special\nattributes (co_*, im_*, tb_*, etc.) in a friendlier fashion.\nIt also provides some help for examining source code and class layout.\n\nHere are some of the useful functions provided by this module:\n\n    ismodule(), isclass(), ismethod(), isfunction(), isgeneratorfunction(),\n        isgenerator(), istraceback(), isframe(), iscode(), isbuiltin(),\n        isroutine() - check object types\n    getmembers() - get members of an object that satisfy a given condition\n\n    getfile(), getsourcefile(), getsource() - find an object's source code\n    getdoc(), getcomments() - get documentation on an object\n    getmodule() - determine the module that an object came from\n    getclasstree() - arrange classes so as to represent their hierarchy\n\n    getargvalues(), getcallargs() - get info about function arguments\n    getfullargspec() - same, with support for Python 3 features\n    formatargvalues() - format an argument spec\n    getouterframes(), getinnerframes() - get info about frames\n    currentframe() - get the current stack frame\n    stack(), trace() - get info about frames on the stack or in a traceback\n\n    signature() - get a Signature object for the callable\n";
static const char __pyx_k_I_m_sorry_I_don_t_have_the_logic[] = "I'm sorry, I don't have the logic to use kwonlyargs. Perhapse you can help PyContracts and implement this? Thanks.";
static const char __pyx_k_Marker_object_for_Signature_empt[] = "Marker object for Signature.empty and Parameter.empty.";
static const char __pyx_k_Provide_a_tokeneater_method_to_d[] = "Provide a tokeneater() method to detect the end of a code block.";
static const char __pyx_k_Represents_a_parameter_in_a_func[] = "Represents a parameter in a function signature.\n\n    Has the following public attributes:\n\n    * name : str\n        The name of the parameter as a string.\n    * default : object\n        The default value for the parameter if specified.  If the\n        parameter has no default value, this attribute is set to\n        `Parameter.empty`.\n    * annotation\n        The annotation for the parameter if specified.  If the\n        parameter has no annotation, this attribute is set to\n        `Parameter.empty`.\n    * kind : str\n        Describes how argument values are bound to the parameter.\n        Possible values: `Parameter.POSITIONAL_ONLY`,\n        `Parameter.POSITIONAL_OR_KEYWORD`, `Parameter.VAR_POSITIONAL`,\n        `Parameter.KEYWORD_ONLY`, `Parameter.VAR_KEYWORD`.\n    ";
static const char __pyx_k_Result_of_Signature_bind_call_Ho[] = "Result of `Signature.bind` call.  Holds the mapping of arguments\n    to the function's parameters.\n\n    Has the following public attributes:\n\n    * arguments : dict\n        An ordered mutable mapping of parameters' names to arguments' values.\n        Does not contain arguments' default values.\n    * signature : Signature\n        The Signature object that created this instance.\n    * args : tuple\n        Tuple of positional arguments values.\n    * kwargs : dict\n        Dict of keyword arguments values.\n    ";
static const char __pyx_k_Signature___init___locals_genexp[] = "Signature.__init__.<locals>.genexpr";
static const char __pyx_k_Signature__hash_basis_locals_gen[] = "Signature._hash_basis.<locals>.genexpr";
static const char __pyx_k_args_varargs_varkw_defaults_kwon[] = "args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations";
static const char __pyx_k_callable_r_is_not_supported_by_s[] = "callable {!r} is not supported by signature";
static const char __pyx_k_can_accept_at_least_one_argument[] = "can_accept_at_least_one_argument";
static const char __pyx_k_can_accept_self_plus_one_argumen[] = "can_accept_self_plus_one_argument";
static const char __pyx_k_check_callable_accepts_these_arg[] = "check_callable_accepts_these_arguments";
static const char __pyx_k_classify_class_attrs_locals_gene[] = "classify_class_attrs.<locals>.genexpr";
static const char __pyx_k_could_not_find_function_definiti[] = "could not find function definition";
static const char __pyx_k_filename_lineno_function_code_co[] = "filename lineno function code_context index";
static const char __pyx_k_formatannotationrelativeto_local[] = "formatannotationrelativeto.<locals>._formatannotation";
static const char __pyx_k_formatargspec_locals_formatargan[] = "formatargspec.<locals>.formatargandannotation";
static const char __pyx_k_got_an_unexpected_keyword_argume[] = "got an unexpected keyword argument {arg!r}";
static const char __pyx_k_implicit_arguments_must_be_passe[] = "implicit arguments must be passed as positional or keyword arguments, not {}";
static const char __pyx_k_inspect_Signature_from_builtin_i[] = "inspect.Signature.from_builtin() is deprecated since Python 3.5, use Signature.from_callable()";
static const char __pyx_k_inspect_getargspec_is_deprecated[] = "inspect.getargspec() is deprecated since Python 3.0, use inspect.signature() or inspect.getfullargspec()";
static const char __pyx_k_module_class_method_function_tra[] = "module, class, method, function, traceback, frame, or code object was expected, got {}";
static const char __pyx_k_multiple_values_for_argument_arg[] = "multiple values for argument {arg!r}";
static const char __pyx_k_name_is_a_required_attribute_for[] = "name is a required attribute for Parameter";
static const char __pyx_k_no_signature_found_for_builtin_f[] = "no signature found for builtin function {!r}";
static const char __pyx_k_no_signature_found_for_builtin_r[] = "no signature found for builtin {!r}";
static const char __pyx_k_no_signature_found_for_builtin_t[] = "no signature found for builtin type {!r}";
static const char __pyx_k_non_default_argument_follows_def[] = "non-default argument follows default argument";
static const char __pyx_k_nonlocals_globals_builtins_unbou[] = "nonlocals globals builtins unbound";
static const char __pyx_k_partial_object_r_has_incorrect_a[] = "partial object {!r} has incorrect arguments";
static const char __pyx_k_unexpected_object_r_in___signatu[] = "unexpected object {!r} in __signature__ attribute";
static const char __pyx_k_signature_fromstr_locals_Rewrit_2[] = "_signature_fromstr.<locals>.RewriteSymbolics.visit_Attribute";
static const char __pyx_k_signature_fromstr_locals_Rewrit_3[] = "_signature_fromstr.<locals>.RewriteSymbolics.visit_Name";
static const char __pyx_k_Aspidites__vendor_contracts_insp_2[] = "Aspidites/_vendor/contracts/inspection.py";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_n_s_ABCMeta;
static PyObject *__pyx_kp_s_A_Signature_object_represents_th;
static PyObject *__pyx_kp_s_A_private_marker_used_in_Paramet;
static PyObject *__pyx_kp_u_Annotations_are_not_currently_su;
static PyObject *__pyx_n_s_ArgInfo;
static PyObject *__pyx_n_u_ArgInfo;
static PyObject *__pyx_n_s_ArgSpec;
static PyObject *__pyx_n_u_ArgSpec;
static PyObject *__pyx_n_s_Arguments;
static PyObject *__pyx_n_u_Arguments;
static PyObject *__pyx_n_s_ArityError;
static PyObject *__pyx_n_s_Aspidites__vendor_contracts_insp;
static PyObject *__pyx_kp_s_Aspidites__vendor_contracts_insp_2;
static PyObject *__pyx_n_s_AsyncGeneratorType;
static PyObject *__pyx_n_s_Attribute;
static PyObject *__pyx_n_u_Attribute;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_Awaitable;
static PyObject *__pyx_n_s_BlockFinder;
static PyObject *__pyx_n_s_BlockFinder___init;
static PyObject *__pyx_n_s_BlockFinder_tokeneater;
static PyObject *__pyx_n_s_BoundArguments;
static PyObject *__pyx_n_s_BoundArguments___eq;
static PyObject *__pyx_n_s_BoundArguments___getstate;
static PyObject *__pyx_n_s_BoundArguments___init;
static PyObject *__pyx_n_s_BoundArguments___repr;
static PyObject *__pyx_n_s_BoundArguments___setstate;
static PyObject *__pyx_n_s_BoundArguments_apply_defaults;
static PyObject *__pyx_n_s_BoundArguments_args;
static PyObject *__pyx_n_s_BoundArguments_kwargs;
static PyObject *__pyx_n_s_BoundArguments_signature;
static PyObject *__pyx_n_s_BuiltinFunctionType;
static PyObject *__pyx_n_s_BuiltinMethodType;
static PyObject *__pyx_n_u_CO;
static PyObject *__pyx_n_s_COMMENT;
static PyObject *__pyx_n_s_COMPILER_FLAG_NAMES;
static PyObject *__pyx_n_s_CORO_CLOSED;
static PyObject *__pyx_n_u_CORO_CLOSED;
static PyObject *__pyx_n_s_CORO_CREATED;
static PyObject *__pyx_n_u_CORO_CREATED;
static PyObject *__pyx_n_s_CORO_RUNNING;
static PyObject *__pyx_n_u_CORO_RUNNING;
static PyObject *__pyx_n_s_CORO_SUSPENDED;
static PyObject *__pyx_n_u_CORO_SUSPENDED;
static PyObject *__pyx_n_s_CO_ASYNC_GENERATOR;
static PyObject *__pyx_n_s_CO_COROUTINE;
static PyObject *__pyx_n_s_CO_GENERATOR;
static PyObject *__pyx_n_s_CO_ITERABLE_COROUTINE;
static PyObject *__pyx_n_s_CO_VARARGS;
static PyObject *__pyx_n_s_CO_VARKEYWORDS;
static PyObject *__pyx_n_s_Callable;
static PyObject *__pyx_n_s_ClassFinder;
static PyObject *__pyx_n_s_ClassFinder___init;
static PyObject *__pyx_n_s_ClassFinder_visit_ClassDef;
static PyObject *__pyx_n_s_ClassFinder_visit_FunctionDef;
static PyObject *__pyx_n_s_ClassFoundException;
static PyObject *__pyx_n_s_ClassMethodDescriptorType;
static PyObject *__pyx_n_s_ClassMethodWrapper;
static PyObject *__pyx_n_s_ClassType;
static PyObject *__pyx_n_s_ClosureVars;
static PyObject *__pyx_n_u_ClosureVars;
static PyObject *__pyx_n_s_CodeType;
static PyObject *__pyx_n_s_Constant;
static PyObject *__pyx_n_s_CoroutineType;
static PyObject *__pyx_n_s_DEBUG_BYTECODE_SUFFIXES;
static PyObject *__pyx_n_s_DEDENT;
static PyObject *__pyx_n_s_DeprecationWarning;
static PyObject *__pyx_n_s_DynamicClassAttribute;
static PyObject *__pyx_n_s_ENCODING;
static PyObject *__pyx_n_s_ERRORTOKEN;
static PyObject *__pyx_n_s_EXTENSION_SUFFIXES;
static PyObject *__pyx_n_s_EndOfBlock;
static PyObject *__pyx_n_u_False;
static PyObject *__pyx_n_s_FrameInfo;
static PyObject *__pyx_n_u_FrameInfo;
static PyObject *__pyx_n_s_FrameType;
static PyObject *__pyx_n_s_FullArgSpec;
static PyObject *__pyx_n_u_FullArgSpec;
static PyObject *__pyx_n_s_FunctionType;
static PyObject *__pyx_kp_u_Function_has_keyword_only_parame;
static PyObject *__pyx_n_s_GEN_CLOSED;
static PyObject *__pyx_n_u_GEN_CLOSED;
static PyObject *__pyx_n_s_GEN_CREATED;
static PyObject *__pyx_n_u_GEN_CREATED;
static PyObject *__pyx_n_s_GEN_RUNNING;
static PyObject *__pyx_n_u_GEN_RUNNING;
static PyObject *__pyx_n_s_GEN_SUSPENDED;
static PyObject *__pyx_n_u_GEN_SUSPENDED;
static PyObject *__pyx_n_s_GeneratorType;
static PyObject *__pyx_n_s_GetSetDescriptorType;
static PyObject *__pyx_n_u_GetSetDescriptorType;
static PyObject *__pyx_n_s_INDENT;
static PyObject *__pyx_kp_u_I_m_sorry_I_don_t_have_the_logic;
static PyObject *__pyx_n_s_IndentationError;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_n_s_InvalidArgs;
static PyObject *__pyx_n_s_KEYWORD_ONLY;
static PyObject *__pyx_n_s_KEYWORD_ONLY_2;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_Load;
static PyObject *__pyx_n_s_MappingProxyType;
static PyObject *__pyx_kp_s_Marker_object_for_Signature_empt;
static PyObject *__pyx_n_s_MemberDescriptorType;
static PyObject *__pyx_n_u_MemberDescriptorType;
static PyObject *__pyx_n_s_MethodType;
static PyObject *__pyx_n_s_MethodWrapper;
static PyObject *__pyx_n_s_Module;
static PyObject *__pyx_n_s_ModuleType;
static PyObject *__pyx_n_s_NEWLINE;
static PyObject *__pyx_n_s_NL;
static PyObject *__pyx_n_s_Name;
static PyObject *__pyx_n_s_NameError;
static PyObject *__pyx_n_s_NodeTransformer;
static PyObject *__pyx_n_s_NodeVisitor;
static PyObject *__pyx_n_s_NonUserDefinedCallables;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_OP;
static PyObject *__pyx_n_s_OPTIMIZED_BYTECODE_SUFFIXES;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_PARAM_NAME_MAPPING;
static PyObject *__pyx_n_s_POSITIONAL_ONLY;
static PyObject *__pyx_n_s_POSITIONAL_ONLY_2;
static PyObject *__pyx_n_s_POSITIONAL_OR_KEYWORD;
static PyObject *__pyx_n_s_POSITIONAL_OR_KEYWORD_2;
static PyObject *__pyx_n_s_Parameter;
static PyObject *__pyx_n_s_ParameterKind;
static PyObject *__pyx_n_s_ParameterKind___str;
static PyObject *__pyx_n_s_ParameterKind_description;
static PyObject *__pyx_n_s_Parameter___eq;
static PyObject *__pyx_n_s_Parameter___hash;
static PyObject *__pyx_n_s_Parameter___init;
static PyObject *__pyx_n_s_Parameter___reduce;
static PyObject *__pyx_n_s_Parameter___repr;
static PyObject *__pyx_n_s_Parameter___setstate;
static PyObject *__pyx_n_s_Parameter___str;
static PyObject *__pyx_n_s_Parameter_annotation;
static PyObject *__pyx_n_s_Parameter_default;
static PyObject *__pyx_n_s_Parameter_kind;
static PyObject *__pyx_n_s_Parameter_name;
static PyObject *__pyx_n_s_Parameter_replace;
static PyObject *__pyx_kp_s_Provide_a_tokeneater_method_to_d;
static PyObject *__pyx_kp_s_Represents_a_parameter_in_a_func;
static PyObject *__pyx_kp_s_Result_of_Signature_bind_call_Ho;
static PyObject *__pyx_n_s_RewriteSymbolics;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SOURCE_SUFFIXES;
static PyObject *__pyx_n_s_Signature;
static PyObject *__pyx_n_s_Signature___eq;
static PyObject *__pyx_n_s_Signature___hash;
static PyObject *__pyx_n_s_Signature___init;
static PyObject *__pyx_n_s_Signature___init___locals_genexp;
static PyObject *__pyx_n_s_Signature___reduce;
static PyObject *__pyx_n_s_Signature___repr;
static PyObject *__pyx_n_s_Signature___setstate;
static PyObject *__pyx_n_s_Signature___str;
static PyObject *__pyx_n_s_Signature__bind;
static PyObject *__pyx_n_s_Signature__hash_basis;
static PyObject *__pyx_n_s_Signature__hash_basis_locals_gen;
static PyObject *__pyx_n_s_Signature_from_builtin;
static PyObject *__pyx_n_s_Signature_from_callable;
static PyObject *__pyx_n_s_Signature_from_function;
static PyObject *__pyx_n_s_Signature_parameters;
static PyObject *__pyx_n_s_Signature_replace;
static PyObject *__pyx_n_s_Signature_return_annotation;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_SyntaxError;
static PyObject *__pyx_n_s_TPFLAGS_IS_ABSTRACT;
static PyObject *__pyx_n_s_Traceback;
static PyObject *__pyx_n_u_Traceback;
static PyObject *__pyx_n_s_TracebackType;
static PyObject *__pyx_n_u_True;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UnicodeError;
static PyObject *__pyx_n_s_VAR_KEYWORD;
static PyObject *__pyx_n_s_VAR_KEYWORD_2;
static PyObject *__pyx_n_s_VAR_POSITIONAL;
static PyObject *__pyx_n_s_VAR_POSITIONAL_2;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_WrapperDescriptor;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_u__12;
static PyObject *__pyx_kp_u__15;
static PyObject *__pyx_kp_u__16;
static PyObject *__pyx_kp_u__19;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_kp_u__25;
static PyObject *__pyx_kp_u__26;
static PyObject *__pyx_kp_u__28;
static PyObject *__pyx_kp_u__29;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__30;
static PyObject *__pyx_kp_u__37;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__41;
static PyObject *__pyx_kp_u__44;
static PyObject *__pyx_kp_u__51;
static PyObject *__pyx_kp_u__52;
static PyObject *__pyx_kp_u__55;
static PyObject *__pyx_kp_u__56;
static PyObject *__pyx_kp_u__57;
static PyObject *__pyx_kp_u__58;
static PyObject *__pyx_kp_u__59;
static PyObject *__pyx_kp_u__73;
static PyObject *__pyx_kp_u__74;
static PyObject *__pyx_kp_u__75;
static PyObject *__pyx_kp_u__76;
static PyObject *__pyx_kp_u__77;
static PyObject *__pyx_kp_u__82;
static PyObject *__pyx_kp_u__83;
static PyObject *__pyx_kp_u__84;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_abc;
static PyObject *__pyx_n_s_abspath;
static PyObject *__pyx_n_u_abstractmethods;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_all_bases;
static PyObject *__pyx_n_s_all_bytecode_suffixes;
static PyObject *__pyx_n_s_all_suffixes;
static PyObject *__pyx_kp_u_and;
static PyObject *__pyx_kp_u_and_2;
static PyObject *__pyx_n_s_ann;
static PyObject *__pyx_n_s_anno;
static PyObject *__pyx_n_s_annotation;
static PyObject *__pyx_n_s_annotation_2;
static PyObject *__pyx_n_u_annotation_2;
static PyObject *__pyx_n_s_annotations;
static PyObject *__pyx_n_s_annotations_2;
static PyObject *__pyx_n_u_annotations_2;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_apply_defaults;
static PyObject *__pyx_n_s_arg;
static PyObject *__pyx_n_s_arg2dict;
static PyObject *__pyx_n_s_arg2value;
static PyObject *__pyx_n_s_arg_names;
static PyObject *__pyx_kp_u_arg_r_parameter_is_positional_o;
static PyObject *__pyx_n_s_arg_val;
static PyObject *__pyx_n_s_arg_vals;
static PyObject *__pyx_n_s_arg_value;
static PyObject *__pyx_n_s_argnames;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_kp_u_args_varargs_keywords_defaults;
static PyObject *__pyx_kp_u_args_varargs_keywords_locals;
static PyObject *__pyx_kp_u_args_varargs_varkw;
static PyObject *__pyx_kp_u_args_varargs_varkw_defaults_kwon;
static PyObject *__pyx_n_u_argument;
static PyObject *__pyx_n_s_arguments;
static PyObject *__pyx_n_u_arguments;
static PyObject *__pyx_n_u_ascii;
static PyObject *__pyx_n_s_assign;
static PyObject *__pyx_n_s_assigned_tuple_params;
static PyObject *__pyx_n_s_ast;
static PyObject *__pyx_kp_u_at_least;
static PyObject *__pyx_kp_u_at_least_2;
static PyObject *__pyx_kp_u_at_most;
static PyObject *__pyx_n_s_atleast;
static PyObject *__pyx_n_s_attr;
static PyObject *__pyx_n_s_attrgetter;
static PyObject *__pyx_n_s_ba;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_base_module;
static PyObject *__pyx_n_s_basename;
static PyObject *__pyx_n_s_bases;
static PyObject *__pyx_n_s_bind;
static PyObject *__pyx_n_s_bind_partial;
static PyObject *__pyx_n_s_blockfinder;
static PyObject *__pyx_n_s_body;
static PyObject *__pyx_n_s_body_col0;
static PyObject *__pyx_n_s_bound;
static PyObject *__pyx_n_s_bound_arguments_cls;
static PyObject *__pyx_n_s_builtin;
static PyObject *__pyx_n_s_builtin_ns;
static PyObject *__pyx_n_s_builtin_vars;
static PyObject *__pyx_n_s_builtinobject;
static PyObject *__pyx_n_s_builtins;
static PyObject *__pyx_n_u_builtins;
static PyObject *__pyx_n_u_builtins_2;
static PyObject *__pyx_kp_u_but;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cache;
static PyObject *__pyx_n_s_call;
static PyObject *__pyx_n_u_call;
static PyObject *__pyx_n_s_call_2;
static PyObject *__pyx_kp_u_callable_r_is_not_supported_by_s;
static PyObject *__pyx_n_s_callable_thing;
static PyObject *__pyx_n_s_can_accept_at_least_one_argument;
static PyObject *__pyx_n_s_can_accept_exactly_one_argument;
static PyObject *__pyx_n_s_can_accept_self;
static PyObject *__pyx_n_s_can_accept_self_plus_one_argumen;
static PyObject *__pyx_n_s_cell;
static PyObject *__pyx_n_s_cell_contents;
static PyObject *__pyx_n_s_chain;
static PyObject *__pyx_n_s_check_callable_accepts_these_arg;
static PyObject *__pyx_n_s_check_class;
static PyObject *__pyx_n_s_check_instance;
static PyObject *__pyx_n_s_checkcache;
static PyObject *__pyx_n_s_children;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_u_class_2;
static PyObject *__pyx_n_s_class_bases;
static PyObject *__pyx_n_s_class_dict;
static PyObject *__pyx_n_s_class_finder;
static PyObject *__pyx_kp_u_class_method;
static PyObject *__pyx_n_s_classes;
static PyObject *__pyx_n_s_classify_class_attrs;
static PyObject *__pyx_n_s_classify_class_attrs_locals_gene;
static PyObject *__pyx_n_s_classmethod;
static PyObject *__pyx_n_s_clean_signature;
static PyObject *__pyx_n_s_cleandoc;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_closure;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_co;
static PyObject *__pyx_n_s_co_argcount;
static PyObject *__pyx_n_s_co_filename;
static PyObject *__pyx_n_s_co_firstlineno;
static PyObject *__pyx_n_u_co_firstlineno;
static PyObject *__pyx_n_s_co_flags;
static PyObject *__pyx_n_s_co_freevars;
static PyObject *__pyx_n_s_co_kwonlyargcount;
static PyObject *__pyx_n_s_co_name;
static PyObject *__pyx_n_s_co_names;
static PyObject *__pyx_n_s_co_posonlyargcount;
static PyObject *__pyx_n_s_co_varnames;
static PyObject *__pyx_n_s_code;
static PyObject *__pyx_n_u_code;
static PyObject *__pyx_n_s_code_2;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_collections_abc;
static PyObject *__pyx_n_s_comment;
static PyObject *__pyx_n_s_comments;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_content;
static PyObject *__pyx_n_s_context;
static PyObject *__pyx_n_s_convert;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_coroutine;
static PyObject *__pyx_kp_u_could_not_find_class_definition;
static PyObject *__pyx_kp_u_could_not_find_code_object;
static PyObject *__pyx_kp_u_could_not_find_function_definiti;
static PyObject *__pyx_kp_u_could_not_get_source_code;
static PyObject *__pyx_n_s_cpos;
static PyObject *__pyx_n_s_cr_frame;
static PyObject *__pyx_n_u_cr_frame;
static PyObject *__pyx_n_s_cr_running;
static PyObject *__pyx_n_s_ctx;
static PyObject *__pyx_n_s_current_parameter;
static PyObject *__pyx_n_s_currentframe;
static PyObject *__pyx_n_u_d;
static PyObject *__pyx_n_u_data;
static PyObject *__pyx_n_s_decorator_list;
static PyObject *__pyx_n_s_decoratorhasargs;
static PyObject *__pyx_n_u_def;
static PyObject *__pyx_kp_u_def_foo;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_default_2;
static PyObject *__pyx_n_u_default_2;
static PyObject *__pyx_n_s_default_node;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_defaults_2;
static PyObject *__pyx_n_u_defaults_2;
static PyObject *__pyx_n_s_defcount;
static PyObject *__pyx_n_s_delayed_comma;
static PyObject *__pyx_n_u_delete;
static PyObject *__pyx_n_s_description;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_u_dict;
static PyObject *__pyx_kp_u_dict___is_special_don_t_want_t;
static PyObject *__pyx_n_s_dict_attr;
static PyObject *__pyx_n_s_dict_obj;
static PyObject *__pyx_n_s_dis;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_doc_2;
static PyObject *__pyx_kp_u_does_not_accept;
static PyObject *__pyx_kp_u_duplicate_parameter_name_r;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_empty_2;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_endswith;
static PyObject *__pyx_n_s_entry;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_erowcol;
static PyObject *__pyx_n_s_eval;
static PyObject *__pyx_n_s_ex;
static PyObject *__pyx_n_u_exactly;
static PyObject *__pyx_n_s_exc;
static PyObject *__pyx_n_s_exc_info;
static PyObject *__pyx_n_s_exists;
static PyObject *__pyx_n_s_expandtabs;
static PyObject *__pyx_n_s_expline;
static PyObject *__pyx_n_s_extra_args;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_f_back;
static PyObject *__pyx_n_s_f_code;
static PyObject *__pyx_n_s_f_lasti;
static PyObject *__pyx_n_s_f_lineno;
static PyObject *__pyx_n_s_f_locals;
static PyObject *__pyx_n_s_f_name;
static PyObject *__pyx_n_s_fget;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_u_file;
static PyObject *__pyx_n_s_file_2;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_filename_2;
static PyObject *__pyx_kp_u_filename_lineno_function_code_co;
static PyObject *__pyx_n_s_filesbymodname;
static PyObject *__pyx_n_s_fillvalue;
static PyObject *__pyx_n_s_find;
static PyObject *__pyx_n_s_findclass;
static PyObject *__pyx_n_s_finddoc;
static PyObject *__pyx_n_s_findsource;
static PyObject *__pyx_n_s_first_wrapped_param;
static PyObject *__pyx_n_s_firstdefault;
static PyObject *__pyx_n_s_flag;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_fn_underscore;
static PyObject *__pyx_n_s_fname;
static PyObject *__pyx_n_s_follow_wrapped;
static PyObject *__pyx_n_s_follow_wrapper_chains;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_formatannotation;
static PyObject *__pyx_n_s_formatannotation_2;
static PyObject *__pyx_n_s_formatannotationrelativeto;
static PyObject *__pyx_n_s_formatannotationrelativeto_local;
static PyObject *__pyx_n_s_formatarg;
static PyObject *__pyx_n_s_formatargandannotation;
static PyObject *__pyx_n_s_formatargspec;
static PyObject *__pyx_kp_u_formatargspec_is_deprecated_sin;
static PyObject *__pyx_n_s_formatargspec_locals_formatargan;
static PyObject *__pyx_n_s_formatargspec_locals_lambda;
static PyObject *__pyx_n_s_formatargvalues;
static PyObject *__pyx_n_s_formatargvalues_locals_convert;
static PyObject *__pyx_n_s_formatargvalues_locals_lambda;
static PyObject *__pyx_n_s_formatreturns;
static PyObject *__pyx_n_s_formatted;
static PyObject *__pyx_n_s_formatvalue;
static PyObject *__pyx_n_s_formatvarargs;
static PyObject *__pyx_n_s_formatvarkw;
static PyObject *__pyx_n_s_frame;
static PyObject *__pyx_n_u_frame;
static PyObject *__pyx_n_s_frameinfo;
static PyObject *__pyx_n_s_framelist;
static PyObject *__pyx_kp_u_from;
static PyObject *__pyx_n_s_from_builtin;
static PyObject *__pyx_n_u_from_bytes;
static PyObject *__pyx_n_s_from_callable;
static PyObject *__pyx_n_s_from_function;
static PyObject *__pyx_n_s_func;
static PyObject *__pyx_n_u_func;
static PyObject *__pyx_n_s_func_2;
static PyObject *__pyx_n_s_func_code;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_generate_tokens;
static PyObject *__pyx_n_s_generator;
static PyObject *__pyx_n_s_generic_visit;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_u_get;
static PyObject *__pyx_n_s_get_2;
static PyObject *__pyx_n_s_get_callable_fullargspec;
static PyObject *__pyx_n_s_get_f_from_callable;
static PyObject *__pyx_n_s_get_obj;
static PyObject *__pyx_n_s_getabsfile;
static PyObject *__pyx_n_s_getargs;
static PyObject *__pyx_n_s_getargspec;
static PyObject *__pyx_n_s_getargvalues;
static PyObject *__pyx_n_s_getattr;
static PyObject *__pyx_n_s_getattr_static;
static PyObject *__pyx_n_s_getattribute;
static PyObject *__pyx_n_s_getblock;
static PyObject *__pyx_n_s_getcallargs;
static PyObject *__pyx_n_s_getcallargs_locals_assign;
static PyObject *__pyx_n_s_getcallargs_locals_is_assigned;
static PyObject *__pyx_n_s_getclasstree;
static PyObject *__pyx_n_s_getclosurevars;
static PyObject *__pyx_n_s_getcomments;
static PyObject *__pyx_n_s_getcoroutinelocals;
static PyObject *__pyx_n_s_getcoroutinestate;
static PyObject *__pyx_n_s_getdefaultencoding;
static PyObject *__pyx_n_s_getdoc;
static PyObject *__pyx_n_s_getfile;
static PyObject *__pyx_n_s_getframe;
static PyObject *__pyx_n_u_getframe;
static PyObject *__pyx_n_s_getframeinfo;
static PyObject *__pyx_n_s_getfullargspec;
static PyObject *__pyx_n_s_getgeneratorlocals;
static PyObject *__pyx_n_s_getgeneratorstate;
static PyObject *__pyx_n_s_getinnerframes;
static PyObject *__pyx_n_s_getlineno;
static PyObject *__pyx_n_s_getlines;
static PyObject *__pyx_n_s_getmembers;
static PyObject *__pyx_n_s_getmembers_locals_lambda;
static PyObject *__pyx_n_s_getmodule;
static PyObject *__pyx_n_s_getmodulename;
static PyObject *__pyx_n_s_getmro;
static PyObject *__pyx_n_s_getouterframes;
static PyObject *__pyx_n_s_getrecursionlimit;
static PyObject *__pyx_n_s_getsource;
static PyObject *__pyx_n_s_getsourcefile;
static PyObject *__pyx_n_s_getsourcefile_locals_genexpr;
static PyObject *__pyx_n_s_getsourcelines;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_gi_code;
static PyObject *__pyx_n_s_gi_frame;
static PyObject *__pyx_n_u_gi_frame;
static PyObject *__pyx_n_s_gi_running;
static PyObject *__pyx_n_s_given;
static PyObject *__pyx_kp_u_given_2;
static PyObject *__pyx_kp_u_given_3;
static PyObject *__pyx_n_s_global_ns;
static PyObject *__pyx_n_s_global_vars;
static PyObject *__pyx_n_s_globals;
static PyObject *__pyx_kp_u_got_an_unexpected_keyword_argum;
static PyObject *__pyx_kp_u_got_an_unexpected_keyword_argume;
static PyObject *__pyx_kp_u_got_multiple_values_for_keyword;
static PyObject *__pyx_n_s_has_code_flag;
static PyObject *__pyx_n_s_hash;
static PyObject *__pyx_n_s_hash_basis;
static PyObject *__pyx_n_s_homecls;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_id_func;
static PyObject *__pyx_n_s_im_self;
static PyObject *__pyx_kp_u_implicit;
static PyObject *__pyx_kp_u_implicit_arguments_must_be_passe;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_importlib;
static PyObject *__pyx_n_s_importlib_machinery;
static PyObject *__pyx_n_s_inPy2;
static PyObject *__pyx_n_s_indecorator;
static PyObject *__pyx_n_s_indent;
static PyObject *__pyx_n_s_indentsize;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_u_init;
static PyObject *__pyx_n_s_init_2;
static PyObject *__pyx_n_s_inspect;
static PyObject *__pyx_kp_u_inspect_Signature_from_builtin_i;
static PyObject *__pyx_kp_u_inspect_Signature_from_function;
static PyObject *__pyx_kp_u_inspect_getargspec_is_deprecated;
static PyObject *__pyx_n_s_instance_dict;
static PyObject *__pyx_n_s_instance_result;
static PyObject *__pyx_n_s_invalid;
static PyObject *__pyx_kp_u_invalid_argument_type;
static PyObject *__pyx_kp_u_invalid_method_signature;
static PyObject *__pyx_n_s_is_assigned;
static PyObject *__pyx_n_s_is_duck_function;
static PyObject *__pyx_kp_u_is_not_a_valid_Parameter_kind;
static PyObject *__pyx_n_s_is_type;
static PyObject *__pyx_n_s_is_wrapper;
static PyObject *__pyx_n_s_isabstract;
static PyObject *__pyx_n_u_isabstractmethod;
static PyObject *__pyx_n_s_isasyncgen;
static PyObject *__pyx_n_s_isasyncgenfunction;
static PyObject *__pyx_n_s_isawaitable;
static PyObject *__pyx_n_s_isbuiltin;
static PyObject *__pyx_n_s_isclass;
static PyObject *__pyx_n_s_iscode;
static PyObject *__pyx_n_s_iscoroutine;
static PyObject *__pyx_n_s_iscoroutinefunction;
static PyObject *__pyx_n_s_isdatadescriptor;
static PyObject *__pyx_n_s_isdigit;
static PyObject *__pyx_n_s_isframe;
static PyObject *__pyx_n_s_isfunction;
static PyObject *__pyx_n_s_isgenerator;
static PyObject *__pyx_n_s_isgeneratorfunction;
static PyObject *__pyx_n_s_isgetsetdescriptor;
static PyObject *__pyx_n_s_isidentifier;
static PyObject *__pyx_n_s_islambda;
static PyObject *__pyx_n_s_ismemberdescriptor;
static PyObject *__pyx_n_s_ismethod;
static PyObject *__pyx_n_s_ismethoddescriptor;
static PyObject *__pyx_n_s_ismodule;
static PyObject *__pyx_n_s_isroutine;
static PyObject *__pyx_n_s_istraceback;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_kp_u_keyword_only;
static PyObject *__pyx_n_s_keyword_only_2;
static PyObject *__pyx_n_s_keyword_only_count;
static PyObject *__pyx_n_s_keywords;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kind_2;
static PyObject *__pyx_n_u_kind_2;
static PyObject *__pyx_n_s_kind_defaults;
static PyObject *__pyx_n_s_klass;
static PyObject *__pyx_n_s_klass_result;
static PyObject *__pyx_n_s_kw_defaults;
static PyObject *__pyx_n_s_kwarg;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_kwargs_param;
static PyObject *__pyx_n_s_kwargs_started;
static PyObject *__pyx_n_s_kwdefaults;
static PyObject *__pyx_n_u_kwdefaults;
static PyObject *__pyx_n_s_kwdefaults_2;
static PyObject *__pyx_n_s_kwo_params;
static PyObject *__pyx_n_s_kwonly;
static PyObject *__pyx_n_s_kwonly_given;
static PyObject *__pyx_n_s_kwonly_sig;
static PyObject *__pyx_n_s_kwonlyarg;
static PyObject *__pyx_n_s_kwonlyargs;
static PyObject *__pyx_n_s_kwonlydefaults;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_u_lambda;
static PyObject *__pyx_n_s_last;
static PyObject *__pyx_n_s_last_cls;
static PyObject *__pyx_n_s_last_positional_only;
static PyObject *__pyx_n_s_line;
static PyObject *__pyx_n_s_line_number;
static PyObject *__pyx_n_s_linecache;
static PyObject *__pyx_n_s_lineno;
static PyObject *__pyx_kp_u_lineno_is_out_of_bounds;
static PyObject *__pyx_n_s_lines;
static PyObject *__pyx_n_s_literal_eval;
static PyObject *__pyx_n_s_lnum;
static PyObject *__pyx_n_u_loader;
static PyObject *__pyx_kp_u_locals;
static PyObject *__pyx_n_s_locals_2;
static PyObject *__pyx_n_s_lstrip;
static PyObject *__pyx_n_s_machinery;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_u_main;
static PyObject *__pyx_n_s_main_2;
static PyObject *__pyx_n_s_mainobject;
static PyObject *__pyx_n_s_margin;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_maxsize;
static PyObject *__pyx_n_s_memo;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_metamro;
static PyObject *__pyx_n_s_meth;
static PyObject *__pyx_n_u_method;
static PyObject *__pyx_n_s_method_name;
static PyObject *__pyx_n_s_missing;
static PyObject *__pyx_kp_u_missing_a_required_argument_arg;
static PyObject *__pyx_n_s_missing_arguments;
static PyObject *__pyx_n_s_mod_dict;
static PyObject *__pyx_n_s_modname;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_u_module;
static PyObject *__pyx_kp_u_module_2;
static PyObject *__pyx_n_s_module_3;
static PyObject *__pyx_kp_u_module_class_method_function_tra;
static PyObject *__pyx_n_s_module_dict;
static PyObject *__pyx_n_s_module_name;
static PyObject *__pyx_n_s_modules;
static PyObject *__pyx_n_s_modulesbyfile;
static PyObject *__pyx_n_s_move_to_end;
static PyObject *__pyx_n_s_mro;
static PyObject *__pyx_n_u_mro;
static PyObject *__pyx_n_s_mro_2;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_kp_u_multiple_values_for_argument_arg;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_u_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_name_3;
static PyObject *__pyx_n_s_name_4;
static PyObject *__pyx_n_u_name_4;
static PyObject *__pyx_kp_u_name_is_a_required_attribute_for;
static PyObject *__pyx_kp_u_name_kind_defining_class_object;
static PyObject *__pyx_kp_u_name_must_be_a_str_not_a;
static PyObject *__pyx_n_s_name_node;
static PyObject *__pyx_n_s_named;
static PyObject *__pyx_n_s_namedtuple;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_nargs;
static PyObject *__pyx_kp_u_need_more_than;
static PyObject *__pyx_n_s_neglen;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_u_new;
static PyObject *__pyx_n_s_new_2;
static PyObject *__pyx_n_s_new_arguments;
static PyObject *__pyx_n_s_new_param;
static PyObject *__pyx_n_s_new_params;
static PyObject *__pyx_n_s_next;
static PyObject *__pyx_n_s_nkwargs;
static PyObject *__pyx_kp_u_no_signature_found_for_builtin_f;
static PyObject *__pyx_kp_u_no_signature_found_for_builtin_r;
static PyObject *__pyx_kp_u_no_signature_found_for_builtin_t;
static PyObject *__pyx_kp_u_no_signature_found_for_r;
static PyObject *__pyx_n_s_node;
static PyObject *__pyx_kp_u_non_default_argument_follows_def;
static PyObject *__pyx_n_s_non_default_count;
static PyObject *__pyx_n_s_nonlocal_vars;
static PyObject *__pyx_kp_u_nonlocals_globals_builtins_unbou;
static PyObject *__pyx_n_s_normcase;
static PyObject *__pyx_n_s_num_args;
static PyObject *__pyx_n_s_num_defaults;
static PyObject *__pyx_n_s_num_pos;
static PyObject *__pyx_n_s_num_required;
static PyObject *__pyx_n_s_num_total;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_objclass;
static PyObject *__pyx_n_u_objclass;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_old_params;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_other;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_param;
static PyObject *__pyx_n_s_param_name;
static PyObject *__pyx_n_s_parameter_cls;
static PyObject *__pyx_n_s_parameters;
static PyObject *__pyx_n_s_parameters_2;
static PyObject *__pyx_n_u_parameters_2;
static PyObject *__pyx_kp_u_parameters_cannot_have_default;
static PyObject *__pyx_n_s_parameters_ex;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_parse;
static PyObject *__pyx_n_s_parse_name;
static PyObject *__pyx_n_s_partial;
static PyObject *__pyx_n_s_partial_args;
static PyObject *__pyx_n_s_partial_keywords;
static PyObject *__pyx_kp_u_partial_object_r_has_incorrect_a;
static PyObject *__pyx_n_s_partialmethod;
static PyObject *__pyx_n_s_partialmethod_2;
static PyObject *__pyx_kp_u_pass;
static PyObject *__pyx_n_s_passline;
static PyObject *__pyx_n_s_pat;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_plural;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_pos_count;
static PyObject *__pyx_n_s_pos_default_count;
static PyObject *__pyx_n_s_positional;
static PyObject *__pyx_n_u_positional;
static PyObject *__pyx_kp_u_positional_argument;
static PyObject *__pyx_kp_u_positional_argument_s_and_d_key;
static PyObject *__pyx_kp_u_positional_only;
static PyObject *__pyx_kp_u_positional_or_keyword;
static PyObject *__pyx_n_s_posonly_count;
static PyObject *__pyx_n_s_posonly_left;
static PyObject *__pyx_n_s_predicate;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_processed;
static PyObject *__pyx_n_s_program;
static PyObject *__pyx_n_s_property;
static PyObject *__pyx_n_u_property;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_qualname_2;
static PyObject *__pyx_kp_u_r;
static PyObject *__pyx_kp_u_r_builtin_has_invalid_signature;
static PyObject *__pyx_kp_u_r_is_a_built_in_class;
static PyObject *__pyx_kp_u_r_is_a_built_in_module;
static PyObject *__pyx_kp_u_r_is_not_a_Python_builtin_funct;
static PyObject *__pyx_kp_u_r_is_not_a_Python_function;
static PyObject *__pyx_kp_u_r_is_not_a_Python_generator;
static PyObject *__pyx_kp_u_r_is_not_a_callable_object;
static PyObject *__pyx_kp_u_r_is_not_a_code_object;
static PyObject *__pyx_kp_u_r_is_not_a_frame_or_traceback_o;
static PyObject *__pyx_kp_u_r_is_not_a_valid_parameter_name;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_realpath;
static PyObject *__pyx_n_s_recursion_limit;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_render_kw_only_separator;
static PyObject *__pyx_n_s_render_pos_only_separator;
static PyObject *__pyx_n_s_rendered;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_u_replace;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_results;
static PyObject *__pyx_n_u_return;
static PyObject *__pyx_n_s_return_annotation;
static PyObject *__pyx_n_s_return_annotation_2;
static PyObject *__pyx_n_u_return_annotation_2;
static PyObject *__pyx_n_s_reversed;
static PyObject *__pyx_n_s_roots;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_u_s;
static PyObject *__pyx_kp_u_s_def_s_s_async_s_def_s_w_lambd;
static PyObject *__pyx_kp_u_s_missing_i_required_s_argument;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_u_self;
static PyObject *__pyx_n_s_self_2;
static PyObject *__pyx_n_u_self_2;
static PyObject *__pyx_n_s_self_3;
static PyObject *__pyx_n_s_self_isbound;
static PyObject *__pyx_n_s_self_ismodule;
static PyObject *__pyx_n_s_self_parameter;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_sentinel;
static PyObject *__pyx_n_u_set;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_shadowed_dict;
static PyObject *__pyx_n_s_sig;
static PyObject *__pyx_n_s_sig_params;
static PyObject *__pyx_n_s_sigcls;
static PyObject *__pyx_n_s_signature;
static PyObject *__pyx_n_u_signature;
static PyObject *__pyx_n_s_signature_2;
static PyObject *__pyx_n_s_signature_3;
static PyObject *__pyx_n_u_signature_3;
static PyObject *__pyx_n_s_signature_bound_method;
static PyObject *__pyx_n_s_signature_from_builtin;
static PyObject *__pyx_n_s_signature_from_callable;
static PyObject *__pyx_n_s_signature_from_callable_locals;
static PyObject *__pyx_n_s_signature_from_function;
static PyObject *__pyx_n_s_signature_fromstr;
static PyObject *__pyx_n_s_signature_fromstr_locals_Rewrit;
static PyObject *__pyx_n_s_signature_fromstr_locals_Rewrit_2;
static PyObject *__pyx_n_s_signature_fromstr_locals_Rewrit_3;
static PyObject *__pyx_n_s_signature_fromstr_locals_p;
static PyObject *__pyx_n_s_signature_fromstr_locals_parse;
static PyObject *__pyx_n_s_signature_fromstr_locals_wrap_v;
static PyObject *__pyx_n_s_signature_get_bound_param;
static PyObject *__pyx_n_s_signature_get_partial;
static PyObject *__pyx_n_s_signature_get_user_defined_meth;
static PyObject *__pyx_n_s_signature_is_builtin;
static PyObject *__pyx_n_s_signature_is_functionlike;
static PyObject *__pyx_n_s_signature_strip_non_python_synt;
static PyObject *__pyx_n_s_skip_bound_arg;
static PyObject *__pyx_n_s_skip_next_comma;
static PyObject *__pyx_n_s_slots;
static PyObject *__pyx_n_u_slots;
static PyObject *__pyx_n_s_slots_2;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_kp_u_source_code_not_available;
static PyObject *__pyx_n_s_spec;
static PyObject *__pyx_n_s_specs;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_splitext;
static PyObject *__pyx_n_s_srch_cls;
static PyObject *__pyx_n_s_srch_obj;
static PyObject *__pyx_n_s_srowcol;
static PyObject *__pyx_n_s_stack;
static PyObject *__pyx_n_s_stacklevel;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_started;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_static_getmro;
static PyObject *__pyx_kp_u_static_method;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_subarg;
static PyObject *__pyx_n_s_subvalue;
static PyObject *__pyx_n_s_suffix;
static PyObject *__pyx_n_s_suffixes;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_sys_module_dict;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_tail;
static PyObject *__pyx_kp_u_takes;
static PyObject *__pyx_kp_u_takes_no_arguments;
static PyObject *__pyx_n_s_tb;
static PyObject *__pyx_n_s_tb_frame;
static PyObject *__pyx_n_s_tb_lineno;
static PyObject *__pyx_n_s_tb_next;
static PyObject *__pyx_n_u_test;
static PyObject *__pyx_n_s_test_2;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_text_sig;
static PyObject *__pyx_n_s_text_signature;
static PyObject *__pyx_n_u_text_signature;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_kp_u_to;
static PyObject *__pyx_kp_u_to_unpack;
static PyObject *__pyx_n_s_token;
static PyObject *__pyx_n_s_token_2;
static PyObject *__pyx_n_s_token_stream;
static PyObject *__pyx_n_s_tokeneater;
static PyObject *__pyx_n_s_tokenize;
static PyObject *__pyx_n_s_tokens;
static PyObject *__pyx_n_s_too_many;
static PyObject *__pyx_kp_u_too_many_positional_arguments;
static PyObject *__pyx_kp_u_too_many_values_to_unpack;
static PyObject *__pyx_n_s_top_kind;
static PyObject *__pyx_n_s_tp;
static PyObject *__pyx_n_s_trace;
static PyObject *__pyx_n_s_transform_to_kwonly;
static PyObject *__pyx_n_s_tree;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_n_u_typing;
static PyObject *__pyx_kp_u_typing_2;
static PyObject *__pyx_n_s_unassigned;
static PyObject *__pyx_n_s_unbound_names;
static PyObject *__pyx_n_s_unexpected;
static PyObject *__pyx_kp_u_unexpected_object_r_in___signatu;
static PyObject *__pyx_n_s_unique;
static PyObject *__pyx_n_s_unwrap;
static PyObject *__pyx_n_s_unwrap_locals__is_wrapper;
static PyObject *__pyx_n_s_unwrap_partial;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_validate_parameters;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_u_value;
static PyObject *__pyx_kp_u_value_2;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_u_values;
static PyObject *__pyx_n_s_var;
static PyObject *__pyx_n_s_vararg;
static PyObject *__pyx_n_s_varargs;
static PyObject *__pyx_kp_u_variadic_keyword;
static PyObject *__pyx_kp_u_variadic_positional;
static PyObject *__pyx_n_s_varkw;
static PyObject *__pyx_n_s_version_info;
static PyObject *__pyx_n_s_visit;
static PyObject *__pyx_n_s_visit_AsyncFunctionDef;
static PyObject *__pyx_n_s_visit_Attribute;
static PyObject *__pyx_n_s_visit_ClassDef;
static PyObject *__pyx_n_s_visit_FunctionDef;
static PyObject *__pyx_n_s_visit_Name;
static PyObject *__pyx_n_s_void;
static PyObject *__pyx_n_s_walktree;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_u_was;
static PyObject *__pyx_n_u_weakref;
static PyObject *__pyx_n_u_were;
static PyObject *__pyx_n_s_wrap_value;
static PyObject *__pyx_n_s_wrapped;
static PyObject *__pyx_n_u_wrapped;
static PyObject *__pyx_n_s_wrapped_sig;
static PyObject *__pyx_kp_u_wrapper_loop_when_unwrapping_r;
static PyObject *__pyx_kp_u_wrong_parameter_order_parameter;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_n_s_zip_longest;
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda4(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda6(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda7(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_text); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda8(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda9(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda10(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_ismodule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_2isclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_4ismethod(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6ismethoddescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_8isdatadescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_10ismemberdescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12ismemberdescriptor(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14isgetsetdescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_16isgetsetdescriptor(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18isfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20_has_code_flag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f, PyObject *__pyx_v_flag); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_22isgeneratorfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_24iscoroutinefunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26isasyncgenfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_28isasyncgen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_30isgenerator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_32iscoroutine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_34isawaitable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_36istraceback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_38isframe(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_40iscode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_42isbuiltin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_44isroutine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_46isabstract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pair); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_48getmembers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v_predicate); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_50classify_class_attrs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_52getmro(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap__is_wrapper(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_2_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_54unwrap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_56indentsize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_58_findclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_60_finddoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_62getdoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_64cleandoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_doc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_66getfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_68getmodulename(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_70getsourcefile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_72getabsfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v__filename); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_74getmodule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v__filename); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_qualname); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_2visit_FunctionDef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_4visit_ClassDef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_76findsource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_78getcomments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_2tokeneater(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_token, PyObject *__pyx_v_srowcol, PyObject *__pyx_v_erowcol, CYTHON_UNUSED PyObject *__pyx_v_line); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_80getblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lines); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_82getsourcelines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_84getsource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_86walktree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_classes, PyObject *__pyx_v_children, CYTHON_UNUSED PyObject *__pyx_v_parent); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_88getclasstree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_classes, PyObject *__pyx_v_unique); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_90getargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_co); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_92getargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_94getargvalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_96formatannotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_annotation, PyObject *__pyx_v_base_module); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto__formatannotation(PyObject *__pyx_self, PyObject *__pyx_v_annotation); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_98formatannotationrelativeto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_191__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(PyObject *__pyx_self, PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_100formatargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_varargs, PyObject *__pyx_v_varkw, PyObject *__pyx_v_defaults, PyObject *__pyx_v_kwonlyargs, PyObject *__pyx_v_kwonlydefaults, PyObject *__pyx_v_annotations, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvarargs, PyObject *__pyx_v_formatvarkw, PyObject *__pyx_v_formatvalue, PyObject *__pyx_v_formatreturns, PyObject *__pyx_v_formatannotation); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_193__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_205__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_convert(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_locals, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvalue); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_102formatargvalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_varargs, PyObject *__pyx_v_varkw, PyObject *__pyx_v_locals, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvarargs, PyObject *__pyx_v_formatvarkw, PyObject *__pyx_v_formatvalue); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_104_missing_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f_name, PyObject *__pyx_v_argnames, PyObject *__pyx_v_pos, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_106_too_many(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f_name, PyObject *__pyx_v_args, PyObject *__pyx_v_kwonly, PyObject *__pyx_v_varargs, PyObject *__pyx_v_defcount, PyObject *__pyx_v_given, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(PyObject *__pyx_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(PyObject *__pyx_self, PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_108getcallargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_positional, PyObject *__pyx_v_named); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_110getclosurevars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_112getframeinfo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_114getlineno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_116getouterframes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_118getinnerframes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tb, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_120currentframe(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_122stack(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_124trace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_126_static_getmro(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_128_check_instance(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_130_check_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_132_is_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_134_shadowed_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_195__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_136getattr_static(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_attr, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_138getgeneratorstate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_generator); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_140getgeneratorlocals(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_generator); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_142getcoroutinestate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coroutine); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_144getcoroutinelocals(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coroutine); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_146_signature_get_user_defined_method(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_method_name); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_148_signature_get_partial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_wrapped_sig, PyObject *__pyx_v_partial, PyObject *__pyx_v_extra_args); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_150_signature_bound_method(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sig); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_152_signature_is_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_154_signature_is_functionlike(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_156_signature_get_bound_param(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_spec); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_158_signature_strip_non_python_syntax(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signature); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_parse_name(PyObject *__pyx_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_2wrap_value(PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_visit_Attribute(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_2visit_Name(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_207__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_4p(PyObject *__pyx_self, PyObject *__pyx_v_name_node, PyObject *__pyx_v_default_node, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_160_signature_fromstr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_obj, PyObject *__pyx_v_s, PyObject *__pyx_v_skip_bound_arg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_162_signature_from_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_skip_bound_arg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_164_signature_from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_skip_bound_arg); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda11(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_166_signature_from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapper_chains, PyObject *__pyx_v_skip_bound_arg, PyObject *__pyx_v_sigcls); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind___str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_2description(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_197__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_kind, PyObject *__pyx_v_default, PyObject *__pyx_v_annotation); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_4__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_6name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_8default(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_10annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_12kind(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_199__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_14replace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_kind, PyObject *__pyx_v_annotation, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_16__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_18__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_20__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_22__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_arguments); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_2signature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_4args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_6kwargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_8apply_defaults(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_10__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_12__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_14__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_16__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_201__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parameters, PyObject *__pyx_v_return_annotation, PyObject *__pyx_v___validate_parameters__); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_2from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_4from_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_6from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapped); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8parameters(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_10return_annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_203__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_12replace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parameters, PyObject *__pyx_v_return_annotation); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_14_hash_basis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_16__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_18__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_20_bind(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_partial); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_22__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_24__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_26__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_28__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_168signature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapped); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_170can_accept_exactly_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_172get_f_from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_174get_callable_fullargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_176can_accept_at_least_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_178can_accept_self(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_180can_accept_self_plus_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_182check_callable_accepts_these_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get_2, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_update = {0, &__pyx_n_s_update, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_format = {0, &__pyx_n_s_format, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_1048576;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__13;
static PyObject *__pyx_slice__14;
static PyObject *__pyx_slice__24;
static PyObject *__pyx_slice__27;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__121;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__153;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__167;
static PyObject *__pyx_tuple__169;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__187;
static PyObject *__pyx_tuple__189;
static PyObject *__pyx_tuple__191;
static PyObject *__pyx_tuple__192;
static PyObject *__pyx_tuple__193;
static PyObject *__pyx_tuple__195;
static PyObject *__pyx_tuple__196;
static PyObject *__pyx_tuple__198;
static PyObject *__pyx_tuple__199;
static PyObject *__pyx_tuple__200;
static PyObject *__pyx_tuple__202;
static PyObject *__pyx_tuple__204;
static PyObject *__pyx_tuple__205;
static PyObject *__pyx_tuple__207;
static PyObject *__pyx_tuple__209;
static PyObject *__pyx_tuple__211;
static PyObject *__pyx_tuple__213;
static PyObject *__pyx_tuple__215;
static PyObject *__pyx_tuple__217;
static PyObject *__pyx_tuple__218;
static PyObject *__pyx_tuple__220;
static PyObject *__pyx_tuple__221;
static PyObject *__pyx_tuple__223;
static PyObject *__pyx_tuple__224;
static PyObject *__pyx_tuple__226;
static PyObject *__pyx_tuple__227;
static PyObject *__pyx_tuple__229;
static PyObject *__pyx_tuple__230;
static PyObject *__pyx_tuple__232;
static PyObject *__pyx_tuple__234;
static PyObject *__pyx_tuple__236;
static PyObject *__pyx_tuple__237;
static PyObject *__pyx_tuple__239;
static PyObject *__pyx_tuple__240;
static PyObject *__pyx_tuple__242;
static PyObject *__pyx_tuple__244;
static PyObject *__pyx_tuple__246;
static PyObject *__pyx_tuple__248;
static PyObject *__pyx_tuple__250;
static PyObject *__pyx_tuple__252;
static PyObject *__pyx_tuple__254;
static PyObject *__pyx_tuple__256;
static PyObject *__pyx_tuple__258;
static PyObject *__pyx_tuple__260;
static PyObject *__pyx_tuple__262;
static PyObject *__pyx_tuple__264;
static PyObject *__pyx_tuple__265;
static PyObject *__pyx_tuple__267;
static PyObject *__pyx_tuple__269;
static PyObject *__pyx_tuple__271;
static PyObject *__pyx_tuple__273;
static PyObject *__pyx_tuple__275;
static PyObject *__pyx_tuple__277;
static PyObject *__pyx_tuple__278;
static PyObject *__pyx_tuple__280;
static PyObject *__pyx_tuple__281;
static PyObject *__pyx_tuple__283;
static PyObject *__pyx_tuple__284;
static PyObject *__pyx_tuple__286;
static PyObject *__pyx_tuple__288;
static PyObject *__pyx_tuple__290;
static PyObject *__pyx_tuple__291;
static PyObject *__pyx_tuple__293;
static PyObject *__pyx_tuple__295;
static PyObject *__pyx_tuple__297;
static PyObject *__pyx_tuple__299;
static PyObject *__pyx_tuple__301;
static PyObject *__pyx_tuple__303;
static PyObject *__pyx_tuple__305;
static PyObject *__pyx_tuple__307;
static PyObject *__pyx_tuple__309;
static PyObject *__pyx_tuple__311;
static PyObject *__pyx_tuple__313;
static PyObject *__pyx_tuple__315;
static PyObject *__pyx_tuple__316;
static PyObject *__pyx_tuple__318;
static PyObject *__pyx_tuple__320;
static PyObject *__pyx_tuple__322;
static PyObject *__pyx_tuple__324;
static PyObject *__pyx_tuple__326;
static PyObject *__pyx_tuple__328;
static PyObject *__pyx_tuple__330;
static PyObject *__pyx_tuple__332;
static PyObject *__pyx_tuple__334;
static PyObject *__pyx_tuple__335;
static PyObject *__pyx_tuple__337;
static PyObject *__pyx_tuple__339;
static PyObject *__pyx_tuple__341;
static PyObject *__pyx_tuple__343;
static PyObject *__pyx_tuple__345;
static PyObject *__pyx_tuple__347;
static PyObject *__pyx_tuple__349;
static PyObject *__pyx_tuple__351;
static PyObject *__pyx_tuple__353;
static PyObject *__pyx_tuple__355;
static PyObject *__pyx_tuple__357;
static PyObject *__pyx_tuple__359;
static PyObject *__pyx_tuple__361;
static PyObject *__pyx_tuple__363;
static PyObject *__pyx_tuple__365;
static PyObject *__pyx_tuple__367;
static PyObject *__pyx_tuple__369;
static PyObject *__pyx_tuple__371;
static PyObject *__pyx_tuple__373;
static PyObject *__pyx_tuple__375;
static PyObject *__pyx_tuple__377;
static PyObject *__pyx_tuple__379;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__88;
static PyObject *__pyx_codeobj__90;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__98;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__108;
static PyObject *__pyx_codeobj__110;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__116;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__120;
static PyObject *__pyx_codeobj__122;
static PyObject *__pyx_codeobj__124;
static PyObject *__pyx_codeobj__126;
static PyObject *__pyx_codeobj__128;
static PyObject *__pyx_codeobj__130;
static PyObject *__pyx_codeobj__132;
static PyObject *__pyx_codeobj__134;
static PyObject *__pyx_codeobj__136;
static PyObject *__pyx_codeobj__140;
static PyObject *__pyx_codeobj__142;
static PyObject *__pyx_codeobj__144;
static PyObject *__pyx_codeobj__146;
static PyObject *__pyx_codeobj__148;
static PyObject *__pyx_codeobj__150;
static PyObject *__pyx_codeobj__152;
static PyObject *__pyx_codeobj__154;
static PyObject *__pyx_codeobj__156;
static PyObject *__pyx_codeobj__158;
static PyObject *__pyx_codeobj__160;
static PyObject *__pyx_codeobj__162;
static PyObject *__pyx_codeobj__165;
static PyObject *__pyx_codeobj__168;
static PyObject *__pyx_codeobj__170;
static PyObject *__pyx_codeobj__172;
static PyObject *__pyx_codeobj__174;
static PyObject *__pyx_codeobj__176;
static PyObject *__pyx_codeobj__178;
static PyObject *__pyx_codeobj__180;
static PyObject *__pyx_codeobj__182;
static PyObject *__pyx_codeobj__184;
static PyObject *__pyx_codeobj__186;
static PyObject *__pyx_codeobj__188;
static PyObject *__pyx_codeobj__190;
static PyObject *__pyx_codeobj__194;
static PyObject *__pyx_codeobj__197;
static PyObject *__pyx_codeobj__201;
static PyObject *__pyx_codeobj__203;
static PyObject *__pyx_codeobj__206;
static PyObject *__pyx_codeobj__208;
static PyObject *__pyx_codeobj__210;
static PyObject *__pyx_codeobj__212;
static PyObject *__pyx_codeobj__214;
static PyObject *__pyx_codeobj__216;
static PyObject *__pyx_codeobj__219;
static PyObject *__pyx_codeobj__222;
static PyObject *__pyx_codeobj__225;
static PyObject *__pyx_codeobj__228;
static PyObject *__pyx_codeobj__231;
static PyObject *__pyx_codeobj__233;
static PyObject *__pyx_codeobj__235;
static PyObject *__pyx_codeobj__238;
static PyObject *__pyx_codeobj__241;
static PyObject *__pyx_codeobj__243;
static PyObject *__pyx_codeobj__245;
static PyObject *__pyx_codeobj__247;
static PyObject *__pyx_codeobj__249;
static PyObject *__pyx_codeobj__251;
static PyObject *__pyx_codeobj__253;
static PyObject *__pyx_codeobj__255;
static PyObject *__pyx_codeobj__257;
static PyObject *__pyx_codeobj__259;
static PyObject *__pyx_codeobj__261;
static PyObject *__pyx_codeobj__263;
static PyObject *__pyx_codeobj__266;
static PyObject *__pyx_codeobj__268;
static PyObject *__pyx_codeobj__270;
static PyObject *__pyx_codeobj__272;
static PyObject *__pyx_codeobj__274;
static PyObject *__pyx_codeobj__276;
static PyObject *__pyx_codeobj__279;
static PyObject *__pyx_codeobj__282;
static PyObject *__pyx_codeobj__285;
static PyObject *__pyx_codeobj__287;
static PyObject *__pyx_codeobj__289;
static PyObject *__pyx_codeobj__292;
static PyObject *__pyx_codeobj__294;
static PyObject *__pyx_codeobj__296;
static PyObject *__pyx_codeobj__298;
static PyObject *__pyx_codeobj__300;
static PyObject *__pyx_codeobj__302;
static PyObject *__pyx_codeobj__304;
static PyObject *__pyx_codeobj__306;
static PyObject *__pyx_codeobj__308;
static PyObject *__pyx_codeobj__310;
static PyObject *__pyx_codeobj__312;
static PyObject *__pyx_codeobj__314;
static PyObject *__pyx_codeobj__317;
static PyObject *__pyx_codeobj__319;
static PyObject *__pyx_codeobj__321;
static PyObject *__pyx_codeobj__323;
static PyObject *__pyx_codeobj__325;
static PyObject *__pyx_codeobj__327;
static PyObject *__pyx_codeobj__329;
static PyObject *__pyx_codeobj__331;
static PyObject *__pyx_codeobj__333;
static PyObject *__pyx_codeobj__336;
static PyObject *__pyx_codeobj__338;
static PyObject *__pyx_codeobj__340;
static PyObject *__pyx_codeobj__342;
static PyObject *__pyx_codeobj__344;
static PyObject *__pyx_codeobj__346;
static PyObject *__pyx_codeobj__348;
static PyObject *__pyx_codeobj__350;
static PyObject *__pyx_codeobj__352;
static PyObject *__pyx_codeobj__354;
static PyObject *__pyx_codeobj__356;
static PyObject *__pyx_codeobj__358;
static PyObject *__pyx_codeobj__360;
static PyObject *__pyx_codeobj__362;
static PyObject *__pyx_codeobj__364;
static PyObject *__pyx_codeobj__366;
static PyObject *__pyx_codeobj__368;
static PyObject *__pyx_codeobj__370;
static PyObject *__pyx_codeobj__372;
static PyObject *__pyx_codeobj__374;
static PyObject *__pyx_codeobj__376;
static PyObject *__pyx_codeobj__378;
static PyObject *__pyx_codeobj__380;
/* Late includes */

/* "Aspidites/_vendor/contracts/inspection.py":1224
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,             # <<<<<<<<<<<<<<
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_184lambda4(PyObject *__pyx_self, PyObject *__pyx_v_name); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_184lambda4 = {"lambda4", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_184lambda4, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_184lambda4(PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda4 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda4(__pyx_self, ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda4(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u_, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1225
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,
 *                   formatvarkw=lambda name: '**' + name,             # <<<<<<<<<<<<<<
 *                   formatvalue=lambda value: '=' + repr(value),
 *                   formatreturns=lambda text: ' -> ' + text,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_185lambda5(PyObject *__pyx_self, PyObject *__pyx_v_name); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_185lambda5 = {"lambda5", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_185lambda5, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_185lambda5(PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda5 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda5(__pyx_self, ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda5", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u__2, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1226
 *                   formatvarargs=lambda name: '*' + name,
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),             # <<<<<<<<<<<<<<
 *                   formatreturns=lambda text: ' -> ' + text,
 *                   formatannotation=formatannotation):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_186lambda6(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_186lambda6 = {"lambda6", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_186lambda6, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_186lambda6(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda6 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda6(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda6(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda6", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_u__3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1227
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),
 *                   formatreturns=lambda text: ' -> ' + text,             # <<<<<<<<<<<<<<
 *                   formatannotation=formatannotation):
 *     """Format an argument spec from the values returned by getfullargspec.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_187lambda7(PyObject *__pyx_self, PyObject *__pyx_v_text); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_187lambda7 = {"lambda7", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_187lambda7, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_187lambda7(PyObject *__pyx_self, PyObject *__pyx_v_text) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda7 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda7(__pyx_self, ((PyObject *)__pyx_v_text));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda7(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_text) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda7", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u__4, __pyx_v_text); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda7", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1284
 * def formatargvalues(args, varargs, varkw, locals,
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,             # <<<<<<<<<<<<<<
 *                     formatvarkw=lambda name: '**' + name,
 *                     formatvalue=lambda value: '=' + repr(value)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_188lambda8(PyObject *__pyx_self, PyObject *__pyx_v_name); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_188lambda8 = {"lambda8", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_188lambda8, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_188lambda8(PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda8 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda8(__pyx_self, ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda8(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda8", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u_, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1285
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 *                     formatvarkw=lambda name: '**' + name,             # <<<<<<<<<<<<<<
 *                     formatvalue=lambda value: '=' + repr(value)):
 *     """Format an argument spec from the 4 values returned by getargvalues.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_189lambda9(PyObject *__pyx_self, PyObject *__pyx_v_name); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_189lambda9 = {"lambda9", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_189lambda9, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_189lambda9(PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda9 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda9(__pyx_self, ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda9(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda9", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u__2, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda9", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1286
 *                     formatvarargs=lambda name: '*' + name,
 *                     formatvarkw=lambda name: '**' + name,
 *                     formatvalue=lambda value: '=' + repr(value)):             # <<<<<<<<<<<<<<
 *     """Format an argument spec from the 4 values returned by getargvalues.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_190lambda10(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_190lambda10 = {"lambda10", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_190lambda10, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_190lambda10(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda10 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda10(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_9Aspidites_7_vendor_9contracts_10inspection_lambda10(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda10", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_u__3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.lambda10", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":74
 * 
 * # ----------------------------------------------------------- type-checking
 * def ismodule(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a module.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_1ismodule(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_ismodule[] = "Return true if the object is a module.\n\n    Module objects provide these attributes:\n        __cached__      pathname to byte compiled file\n        __doc__         documentation string\n        __file__        filename (missing for built-in modules)";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_1ismodule = {"ismodule", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_1ismodule, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_ismodule};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_1ismodule(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismodule (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_ismodule(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_ismodule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ismodule", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":81
 *         __doc__         documentation string
 *         __file__        filename (missing for built-in modules)"""
 *     return isinstance(obj, types.ModuleType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ModuleType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":74
 * 
 * # ----------------------------------------------------------- type-checking
 * def ismodule(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a module.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.ismodule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":84
 * 
 * 
 * def isclass(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a class.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_3isclass(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_2isclass[] = "Return true if the object is a class.\n\n    Class objects provide these attributes:\n        __doc__         documentation string\n        __module__      name of module in which this class was defined";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_3isclass = {"isclass", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_3isclass, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_2isclass};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_3isclass(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isclass (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_2isclass(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_2isclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isclass", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":90
 *         __doc__         documentation string
 *         __module__      name of module in which this class was defined"""
 *     return isinstance(obj, type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyType_Check(__pyx_v_obj); 
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":84
 * 
 * 
 * def isclass(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a class.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isclass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":93
 * 
 * 
 * def ismethod(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an instance method.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_5ismethod(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_4ismethod[] = "Return true if the object is an instance method.\n\n    Instance method objects provide these attributes:\n        __doc__         documentation string\n        __name__        name with which this method was defined\n        __func__        function object containing implementation of method\n        __self__        instance to which this method is bound";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_5ismethod = {"ismethod", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_5ismethod, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_4ismethod};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_5ismethod(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismethod (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_4ismethod(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_4ismethod(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ismethod", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":101
 *         __func__        function object containing implementation of method
 *         __self__        instance to which this method is bound"""
 *     return isinstance(obj, types.MethodType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":93
 * 
 * 
 * def ismethod(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an instance method.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.ismethod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":104
 * 
 * 
 * def ismethoddescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a method descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_7ismethoddescriptor(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_6ismethoddescriptor[] = "Return true if the object is a method descriptor.\n\n    But not if ismethod() or isclass() or isfunction() are true.\n\n    This is new in Python 2.2, and, for example, is true of int.__add__.\n    An object passing this test has a __get__ attribute but not a __set__\n    attribute, but beyond that the set of attributes varies.  __name__ is\n    usually sensible, and __doc__ often is.\n\n    Methods implemented via descriptors that also pass one of the other\n    tests return false from the ismethoddescriptor() test, simply because\n    the other tests promise more -- you can, e.g., count on having the\n    __func__ attribute (etc) when an object passes ismethod().";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_7ismethoddescriptor = {"ismethoddescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_7ismethoddescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_6ismethoddescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_7ismethoddescriptor(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismethoddescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6ismethoddescriptor(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6ismethoddescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyTypeObject *__pyx_v_tp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ismethoddescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":118
 *     the other tests promise more -- you can, e.g., count on having the
 *     __func__ attribute (etc) when an object passes ismethod()."""
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):             # <<<<<<<<<<<<<<
 *         # mutual exclusion
 *         return False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isclass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":120
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):
 *         # mutual exclusion
 *         return False             # <<<<<<<<<<<<<<
 *     tp = type(obj)
 *     return hasattr(tp, "__get__") and not hasattr(tp, "__set__")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":118
 *     the other tests promise more -- you can, e.g., count on having the
 *     __func__ attribute (etc) when an object passes ismethod()."""
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):             # <<<<<<<<<<<<<<
 *         # mutual exclusion
 *         return False
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":121
 *         # mutual exclusion
 *         return False
 *     tp = type(obj)             # <<<<<<<<<<<<<<
 *     return hasattr(tp, "__get__") and not hasattr(tp, "__set__")
 * 
 */
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
  __pyx_v_tp = ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_obj)));

  /* "Aspidites/_vendor/contracts/inspection.py":122
 *         return False
 *     tp = type(obj)
 *     return hasattr(tp, "__get__") and not hasattr(tp, "__set__")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_HasAttr(((PyObject *)__pyx_v_tp), __pyx_n_u_get); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 122, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_HasAttr(((PyObject *)__pyx_v_tp), __pyx_n_u_set); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_t_5 = (!(__pyx_t_1 != 0));
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L7_bool_binop_done:;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":104
 * 
 * 
 * def ismethoddescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a method descriptor.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.ismethoddescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":125
 * 
 * 
 * def isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a data descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9isdatadescriptor(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_8isdatadescriptor[] = "Return true if the object is a data descriptor.\n\n    Data descriptors have a __set__ or a __delete__ attribute.  Examples are\n    properties (defined in Python) and getsets and members (defined in C).\n    Typically, data descriptors will also have __name__ and __doc__ attributes\n    (properties, getsets, and members have both of these attributes), but this\n    is not guaranteed.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9isdatadescriptor = {"isdatadescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9isdatadescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_8isdatadescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9isdatadescriptor(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isdatadescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_8isdatadescriptor(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_8isdatadescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyTypeObject *__pyx_v_tp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isdatadescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":133
 *     (properties, getsets, and members have both of these attributes), but this
 *     is not guaranteed."""
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):             # <<<<<<<<<<<<<<
 *         # mutual exclusion
 *         return False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isclass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":135
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):
 *         # mutual exclusion
 *         return False             # <<<<<<<<<<<<<<
 *     tp = type(obj)
 *     return hasattr(tp, "__set__") or hasattr(tp, "__delete__")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":133
 *     (properties, getsets, and members have both of these attributes), but this
 *     is not guaranteed."""
 *     if isclass(obj) or ismethod(obj) or isfunction(obj):             # <<<<<<<<<<<<<<
 *         # mutual exclusion
 *         return False
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":136
 *         # mutual exclusion
 *         return False
 *     tp = type(obj)             # <<<<<<<<<<<<<<
 *     return hasattr(tp, "__set__") or hasattr(tp, "__delete__")
 * 
 */
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
  __pyx_v_tp = ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_obj)));

  /* "Aspidites/_vendor/contracts/inspection.py":137
 *         return False
 *     tp = type(obj)
 *     return hasattr(tp, "__set__") or hasattr(tp, "__delete__")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_HasAttr(((PyObject *)__pyx_v_tp), __pyx_n_u_set); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
  if (!__pyx_t_1) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_HasAttr(((PyObject *)__pyx_v_tp), __pyx_n_u_delete); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L7_bool_binop_done:;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":125
 * 
 * 
 * def isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a data descriptor.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isdatadescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":142
 * if hasattr(types, 'MemberDescriptorType'):
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11ismemberdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_10ismemberdescriptor[] = "Return true if the object is a member descriptor.\n\n        Member descriptors are specialized descriptors defined in extension\n        modules.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11ismemberdescriptor = {"ismemberdescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11ismemberdescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_10ismemberdescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11ismemberdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismemberdescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_10ismemberdescriptor(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_10ismemberdescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ismemberdescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":147
 *         Member descriptors are specialized descriptors defined in extension
 *         modules."""
 *         return isinstance(object, types.MemberDescriptorType)             # <<<<<<<<<<<<<<
 * else:
 *     # Other implementations
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MemberDescriptorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":142
 * if hasattr(types, 'MemberDescriptorType'):
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.ismemberdescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":150
 * else:
 *     # Other implementations
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13ismemberdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_12ismemberdescriptor[] = "Return true if the object is a member descriptor.\n\n        Member descriptors are specialized descriptors defined in extension\n        modules.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_13ismemberdescriptor = {"ismemberdescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13ismemberdescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_12ismemberdescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13ismemberdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismemberdescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12ismemberdescriptor(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12ismemberdescriptor(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ismemberdescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":155
 *         Member descriptors are specialized descriptors defined in extension
 *         modules."""
 *         return False             # <<<<<<<<<<<<<<
 * 
 * if hasattr(types, 'GetSetDescriptorType'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":150
 * else:
 *     # Other implementations
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":159
 * if hasattr(types, 'GetSetDescriptorType'):
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15isgetsetdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_14isgetsetdescriptor[] = "Return true if the object is a getset descriptor.\n\n        getset descriptors are specialized descriptors defined in extension\n        modules.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_15isgetsetdescriptor = {"isgetsetdescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15isgetsetdescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_14isgetsetdescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15isgetsetdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isgetsetdescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14isgetsetdescriptor(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14isgetsetdescriptor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isgetsetdescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":164
 *         getset descriptors are specialized descriptors defined in extension
 *         modules."""
 *         return isinstance(object, types.GetSetDescriptorType)             # <<<<<<<<<<<<<<
 * else:
 *     # Other implementations
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_GetSetDescriptorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":159
 * if hasattr(types, 'GetSetDescriptorType'):
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isgetsetdescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":167
 * else:
 *     # Other implementations
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_17isgetsetdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_16isgetsetdescriptor[] = "Return true if the object is a getset descriptor.\n\n        getset descriptors are specialized descriptors defined in extension\n        modules.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_17isgetsetdescriptor = {"isgetsetdescriptor", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_17isgetsetdescriptor, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_16isgetsetdescriptor};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_17isgetsetdescriptor(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isgetsetdescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_16isgetsetdescriptor(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_16isgetsetdescriptor(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isgetsetdescriptor", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":172
 *         getset descriptors are specialized descriptors defined in extension
 *         modules."""
 *         return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":167
 * else:
 *     # Other implementations
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":175
 * 
 * 
 * def isfunction(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_19isfunction(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_18isfunction[] = "Return true if the object is a user-defined function.\n\n    Function objects provide these attributes:\n        __doc__         documentation string\n        __name__        name with which this function was defined\n        __code__        code object containing compiled function bytecode\n        __defaults__    tuple of any default values for arguments\n        __globals__     global namespace in which this function was defined\n        __annotations__ dict of parameter annotations\n        __kwdefaults__  dict of keyword only parameters with defaults";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_19isfunction = {"isfunction", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_19isfunction, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_18isfunction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_19isfunction(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isfunction (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18isfunction(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18isfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isfunction", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":186
 *         __annotations__ dict of parameter annotations
 *         __kwdefaults__  dict of keyword only parameters with defaults"""
 *     return isinstance(object, types.FunctionType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FunctionType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":175
 * 
 * 
 * def isfunction(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isfunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":189
 * 
 * 
 * def _has_code_flag(f, flag):             # <<<<<<<<<<<<<<
 *     """Return true if ``f`` is a function (or a method or functools.partial
 *     wrapper wrapping a function) whose code object has the given ``flag``
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_21_has_code_flag(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_20_has_code_flag[] = "Return true if ``f`` is a function (or a method or functools.partial\n    wrapper wrapping a function) whose code object has the given ``flag``\n    set in its flags.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_21_has_code_flag = {"_has_code_flag", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_21_has_code_flag, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_20_has_code_flag};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_21_has_code_flag(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_f = 0;
  PyObject *__pyx_v_flag = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_has_code_flag (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f,&__pyx_n_s_flag,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_f)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_has_code_flag", 1, 2, 2, 1); __PYX_ERR(0, 189, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_has_code_flag") < 0)) __PYX_ERR(0, 189, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_f = values[0];
    __pyx_v_flag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_has_code_flag", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 189, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._has_code_flag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20_has_code_flag(__pyx_self, __pyx_v_f, __pyx_v_flag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20_has_code_flag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f, PyObject *__pyx_v_flag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_has_code_flag", 0);
  __Pyx_INCREF(__pyx_v_f);

  /* "Aspidites/_vendor/contracts/inspection.py":193
 *     wrapper wrapping a function) whose code object has the given ``flag``
 *     set in its flags."""
 *     while ismethod(f):             # <<<<<<<<<<<<<<
 *         f = f.__func__
 *     f = functools._unwrap_partial(f)
 */
  while (1) {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_f);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/contracts/inspection.py":194
 *     set in its flags."""
 *     while ismethod(f):
 *         f = f.__func__             # <<<<<<<<<<<<<<
 *     f = functools._unwrap_partial(f)
 *     if not isfunction(f):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_f, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":195
 *     while ismethod(f):
 *         f = f.__func__
 *     f = functools._unwrap_partial(f)             # <<<<<<<<<<<<<<
 *     if not isfunction(f):
 *         return False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_functools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unwrap_partial); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_f);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_f, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":196
 *         f = f.__func__
 *     f = functools._unwrap_partial(f)
 *     if not isfunction(f):             # <<<<<<<<<<<<<<
 *         return False
 *     return bool(f.__code__.co_flags & flag)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_f);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":197
 *     f = functools._unwrap_partial(f)
 *     if not isfunction(f):
 *         return False             # <<<<<<<<<<<<<<
 *     return bool(f.__code__.co_flags & flag)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":196
 *         f = f.__func__
 *     f = functools._unwrap_partial(f)
 *     if not isfunction(f):             # <<<<<<<<<<<<<<
 *         return False
 *     return bool(f.__code__.co_flags & flag)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":198
 *     if not isfunction(f):
 *         return False
 *     return bool(f.__code__.co_flags & flag)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_And(__pyx_t_3, __pyx_v_flag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":189
 * 
 * 
 * def _has_code_flag(f, flag):             # <<<<<<<<<<<<<<
 *     """Return true if ``f`` is a function (or a method or functools.partial
 *     wrapper wrapping a function) whose code object has the given ``flag``
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._has_code_flag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":201
 * 
 * 
 * def isgeneratorfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined generator function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_23isgeneratorfunction(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_22isgeneratorfunction[] = "Return true if the object is a user-defined generator function.\n\n    Generator function objects provide the same attributes as functions.\n    See help(isfunction) for a list of attributes.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_23isgeneratorfunction = {"isgeneratorfunction", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_23isgeneratorfunction, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_22isgeneratorfunction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_23isgeneratorfunction(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isgeneratorfunction (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_22isgeneratorfunction(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_22isgeneratorfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isgeneratorfunction", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":206
 *     Generator function objects provide the same attributes as functions.
 *     See help(isfunction) for a list of attributes."""
 *     return _has_code_flag(obj, CO_GENERATOR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_has_code_flag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CO_GENERATOR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_obj);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":201
 * 
 * 
 * def isgeneratorfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined generator function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isgeneratorfunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":209
 * 
 * 
 * def iscoroutinefunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_25iscoroutinefunction(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_24iscoroutinefunction[] = "Return true if the object is a coroutine function.\n\n    Coroutine functions are defined with \"async def\" syntax.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_25iscoroutinefunction = {"iscoroutinefunction", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_25iscoroutinefunction, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_24iscoroutinefunction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_25iscoroutinefunction(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iscoroutinefunction (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_24iscoroutinefunction(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_24iscoroutinefunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iscoroutinefunction", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":214
 *     Coroutine functions are defined with "async def" syntax.
 *     """
 *     return _has_code_flag(obj, CO_COROUTINE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_has_code_flag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CO_COROUTINE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_obj);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":209
 * 
 * 
 * def iscoroutinefunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.iscoroutinefunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":217
 * 
 * 
 * def isasyncgenfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_27isasyncgenfunction(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_26isasyncgenfunction[] = "Return true if the object is an asynchronous generator function.\n\n    Asynchronous generator functions are defined with \"async def\"\n    syntax and have \"yield\" expressions in their body.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_27isasyncgenfunction = {"isasyncgenfunction", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_27isasyncgenfunction, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_26isasyncgenfunction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_27isasyncgenfunction(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isasyncgenfunction (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26isasyncgenfunction(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26isasyncgenfunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isasyncgenfunction", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":223
 *     syntax and have "yield" expressions in their body.
 *     """
 *     return _has_code_flag(obj, CO_ASYNC_GENERATOR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_has_code_flag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CO_ASYNC_GENERATOR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_obj);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":217
 * 
 * 
 * def isasyncgenfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isasyncgenfunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":226
 * 
 * 
 * def isasyncgen(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator."""
 *     return isinstance(object, types.AsyncGeneratorType)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_29isasyncgen(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_28isasyncgen[] = "Return true if the object is an asynchronous generator.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_29isasyncgen = {"isasyncgen", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_29isasyncgen, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_28isasyncgen};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_29isasyncgen(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isasyncgen (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_28isasyncgen(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_28isasyncgen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isasyncgen", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":228
 * def isasyncgen(object):
 *     """Return true if the object is an asynchronous generator."""
 *     return isinstance(object, types.AsyncGeneratorType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_AsyncGeneratorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":226
 * 
 * 
 * def isasyncgen(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator."""
 *     return isinstance(object, types.AsyncGeneratorType)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isasyncgen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":231
 * 
 * 
 * def isgenerator(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a generator.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_31isgenerator(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_30isgenerator[] = "Return true if the object is a generator.\n\n    Generator objects provide these attributes:\n        __iter__        defined to support iteration over container\n        close           raises a new GeneratorExit exception inside the\n                        generator to terminate the iteration\n        gi_code         code object\n        gi_frame        frame object or possibly None once the generator has\n                        been exhausted\n        gi_running      set to 1 when generator is executing, 0 otherwise\n        next            return the next item from the container\n        send            resumes the generator and \"sends\" a value that becomes\n                        the result of the current yield-expression\n        throw           used to raise an exception inside the generator";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_31isgenerator = {"isgenerator", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_31isgenerator, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_30isgenerator};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_31isgenerator(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isgenerator (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_30isgenerator(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_30isgenerator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isgenerator", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":246
 *                         the result of the current yield-expression
 *         throw           used to raise an exception inside the generator"""
 *     return isinstance(object, types.GeneratorType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_GeneratorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":231
 * 
 * 
 * def isgenerator(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a generator.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isgenerator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":249
 * 
 * 
 * def iscoroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine."""
 *     return isinstance(object, types.CoroutineType)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_33iscoroutine(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_32iscoroutine[] = "Return true if the object is a coroutine.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_33iscoroutine = {"iscoroutine", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_33iscoroutine, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_32iscoroutine};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_33iscoroutine(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iscoroutine (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_32iscoroutine(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_32iscoroutine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iscoroutine", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":251
 * def iscoroutine(object):
 *     """Return true if the object is a coroutine."""
 *     return isinstance(object, types.CoroutineType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_CoroutineType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":249
 * 
 * 
 * def iscoroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine."""
 *     return isinstance(object, types.CoroutineType)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.iscoroutine", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":254
 * 
 * 
 * def isawaitable(object):             # <<<<<<<<<<<<<<
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_35isawaitable(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_34isawaitable[] = "Return true if object can be passed to an ``await`` expression.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_35isawaitable = {"isawaitable", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_35isawaitable, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_34isawaitable};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_35isawaitable(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isawaitable (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_34isawaitable(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_34isawaitable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isawaitable", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":256
 * def isawaitable(object):
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or             # <<<<<<<<<<<<<<
 *             isinstance(object, types.GeneratorType) and
 *             bool(object.gi_code.co_flags & CO_ITERABLE_COROUTINE) or
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CoroutineType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_object, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":257
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or
 *             isinstance(object, types.GeneratorType) and             # <<<<<<<<<<<<<<
 *             bool(object.gi_code.co_flags & CO_ITERABLE_COROUTINE) or
 *             isinstance(object, collections.abc.Awaitable))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_GeneratorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_4) {
    goto __pyx_L5_next_or;
  } else {
  }

  /* "Aspidites/_vendor/contracts/inspection.py":258
 *     return (isinstance(object, types.CoroutineType) or
 *             isinstance(object, types.GeneratorType) and
 *             bool(object.gi_code.co_flags & CO_ITERABLE_COROUTINE) or             # <<<<<<<<<<<<<<
 *             isinstance(object, collections.abc.Awaitable))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_gi_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CO_ITERABLE_COROUTINE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyNumber_And(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyBool_FromLong((!(!__pyx_t_4))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_1 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L5_next_or:;

  /* "Aspidites/_vendor/contracts/inspection.py":259
 *             isinstance(object, types.GeneratorType) and
 *             bool(object.gi_code.co_flags & CO_ITERABLE_COROUTINE) or
 *             isinstance(object, collections.abc.Awaitable))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_collections); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_abc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Awaitable); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_object, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":254
 * 
 * 
 * def isawaitable(object):             # <<<<<<<<<<<<<<
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isawaitable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":262
 * 
 * 
 * def istraceback(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a traceback.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_37istraceback(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_36istraceback[] = "Return true if the object is a traceback.\n\n    Traceback objects provide these attributes:\n        tb_frame        frame object at this level\n        tb_lasti        index of last attempted instruction in bytecode\n        tb_lineno       current line number in Python source code\n        tb_next         next inner traceback object (called by this level)";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_37istraceback = {"istraceback", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_37istraceback, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_36istraceback};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_37istraceback(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("istraceback (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_36istraceback(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_36istraceback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("istraceback", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":270
 *         tb_lineno       current line number in Python source code
 *         tb_next         next inner traceback object (called by this level)"""
 *     return isinstance(object, types.TracebackType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_TracebackType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":262
 * 
 * 
 * def istraceback(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a traceback.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.istraceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":273
 * 
 * 
 * def isframe(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a frame object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_39isframe(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_38isframe[] = "Return true if the object is a frame object.\n\n    Frame objects provide these attributes:\n        f_back          next outer frame object (this frame's caller)\n        f_builtins      built-in namespace seen by this frame\n        f_code          code object being executed in this frame\n        f_globals       global namespace seen by this frame\n        f_lasti         index of last attempted instruction in bytecode\n        f_lineno        current line number in Python source code\n        f_locals        local namespace seen by this frame\n        f_trace         tracing function for this frame, or None";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_39isframe = {"isframe", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_39isframe, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_38isframe};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_39isframe(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isframe (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_38isframe(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_38isframe(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isframe", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":285
 *         f_locals        local namespace seen by this frame
 *         f_trace         tracing function for this frame, or None"""
 *     return isinstance(object, types.FrameType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FrameType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":273
 * 
 * 
 * def isframe(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a frame object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":288
 * 
 * 
 * def iscode(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a code object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_41iscode(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_40iscode[] = "Return true if the object is a code object.\n\n    Code objects provide these attributes:\n        co_argcount         number of arguments (not including *, ** args\n                            or keyword only arguments)\n        co_code             string of raw compiled bytecode\n        co_cellvars         tuple of names of cell variables\n        co_consts           tuple of constants used in the bytecode\n        co_filename         name of file in which this code object was created\n        co_firstlineno      number of first line in Python source code\n        co_flags            bitmap: 1=optimized | 2=newlocals | 4=*arg | 8=**arg\n                            | 16=nested | 32=generator | 64=nofree | 128=coroutine\n                            | 256=iterable_coroutine | 512=async_generator\n        co_freevars         tuple of names of free variables\n        co_posonlyargcount  number of positional only arguments\n        co_kwonlyargcount   number of keyword only arguments (not including ** arg)\n        co_lnotab           encoded mapping of line numbers to bytecode indices\n        co_name             name with which this code object was defined\n        co_names            tuple of names of local variables\n        co_nlocals          number of local variables\n        co_stacksize        virtual machine stack space required\n        co_varnames         tuple of names of arguments and local variables";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_41iscode = {"iscode", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_41iscode, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_40iscode};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_41iscode(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iscode (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_40iscode(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_40iscode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iscode", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":311
 *         co_stacksize        virtual machine stack space required
 *         co_varnames         tuple of names of arguments and local variables"""
 *     return isinstance(object, types.CodeType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_CodeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":288
 * 
 * 
 * def iscode(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a code object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.iscode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":314
 * 
 * 
 * def isbuiltin(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a built-in function or method.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_43isbuiltin(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_42isbuiltin[] = "Return true if the object is a built-in function or method.\n\n    Built-in functions and methods provide these attributes:\n        __doc__         documentation string\n        __name__        original name of this function or method\n        __self__        instance to which a method is bound, or None";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_43isbuiltin = {"isbuiltin", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_43isbuiltin, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_42isbuiltin};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_43isbuiltin(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isbuiltin (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_42isbuiltin(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_42isbuiltin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isbuiltin", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":321
 *         __name__        original name of this function or method
 *         __self__        instance to which a method is bound, or None"""
 *     return isinstance(object, types.BuiltinFunctionType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BuiltinFunctionType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_object, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":314
 * 
 * 
 * def isbuiltin(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a built-in function or method.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isbuiltin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":324
 * 
 * 
 * def isroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_45isroutine(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_44isroutine[] = "Return true if the object is any kind of function or method.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_45isroutine = {"isroutine", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_45isroutine, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_44isroutine};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_45isroutine(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isroutine (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_44isroutine(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_44isroutine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isroutine", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":326
 * def isroutine(object):
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)             # <<<<<<<<<<<<<<
 *             or isfunction(object)
 *             or ismethod(object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "Aspidites/_vendor/contracts/inspection.py":327
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 *             or isfunction(object)             # <<<<<<<<<<<<<<
 *             or ismethod(object)
 *             or ismethoddescriptor(object))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isbuiltin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "Aspidites/_vendor/contracts/inspection.py":326
 * def isroutine(object):
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)             # <<<<<<<<<<<<<<
 *             or isfunction(object)
 *             or ismethod(object)
 */
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 326, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":327
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 *             or isfunction(object)             # <<<<<<<<<<<<<<
 *             or ismethod(object)
 *             or ismethoddescriptor(object))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 327, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":328
 *     return (isbuiltin(object)
 *             or isfunction(object)
 *             or ismethod(object)             # <<<<<<<<<<<<<<
 *             or ismethoddescriptor(object))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":329
 *             or isfunction(object)
 *             or ismethod(object)
 *             or ismethoddescriptor(object))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismethoddescriptor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":324
 * 
 * 
 * def isroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isroutine", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":332
 * 
 * 
 * def isabstract(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_47isabstract(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_46isabstract[] = "Return true if the object is an abstract base class (ABC).";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_47isabstract = {"isabstract", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_47isabstract, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_46isabstract};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_47isabstract(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isabstract (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_46isabstract(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_46isabstract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isabstract", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":334
 * def isabstract(object):
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):             # <<<<<<<<<<<<<<
 *         return False
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 */
  __pyx_t_1 = PyType_Check(__pyx_v_object); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":335
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):
 *         return False             # <<<<<<<<<<<<<<
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 *         return True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":334
 * def isabstract(object):
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):             # <<<<<<<<<<<<<<
 *         return False
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":336
 *     if not isinstance(object, type):
 *         return False
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:             # <<<<<<<<<<<<<<
 *         return True
 *     if not issubclass(type(object), abc.ABCMeta):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TPFLAGS_IS_ABSTRACT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_And(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":337
 *         return False
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 *         return True             # <<<<<<<<<<<<<<
 *     if not issubclass(type(object), abc.ABCMeta):
 *         return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":336
 *     if not isinstance(object, type):
 *         return False
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:             # <<<<<<<<<<<<<<
 *         return True
 *     if not issubclass(type(object), abc.ABCMeta):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":338
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 *         return True
 *     if not issubclass(type(object), abc.ABCMeta):             # <<<<<<<<<<<<<<
 *         return False
 *     if hasattr(object, '__abstractmethods__'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_abc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ABCMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = PyObject_IsSubclass(((PyObject *)Py_TYPE(__pyx_v_object)), __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":339
 *         return True
 *     if not issubclass(type(object), abc.ABCMeta):
 *         return False             # <<<<<<<<<<<<<<
 *     if hasattr(object, '__abstractmethods__'):
 *         # It looks like ABCMeta.__new__ has finished running;
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":338
 *     if object.__flags__ & TPFLAGS_IS_ABSTRACT:
 *         return True
 *     if not issubclass(type(object), abc.ABCMeta):             # <<<<<<<<<<<<<<
 *         return False
 *     if hasattr(object, '__abstractmethods__'):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":340
 *     if not issubclass(type(object), abc.ABCMeta):
 *         return False
 *     if hasattr(object, '__abstractmethods__'):             # <<<<<<<<<<<<<<
 *         # It looks like ABCMeta.__new__ has finished running;
 *         # TPFLAGS_IS_ABSTRACT should have been accurate.
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_object, __pyx_n_u_abstractmethods); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":343
 *         # It looks like ABCMeta.__new__ has finished running;
 *         # TPFLAGS_IS_ABSTRACT should have been accurate.
 *         return False             # <<<<<<<<<<<<<<
 *     # It looks like ABCMeta.__new__ has not finished running yet; we're
 *     # probably in __init_subclass__. We'll look for abstractmethods manually.
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":340
 *     if not issubclass(type(object), abc.ABCMeta):
 *         return False
 *     if hasattr(object, '__abstractmethods__'):             # <<<<<<<<<<<<<<
 *         # It looks like ABCMeta.__new__ has finished running;
 *         # TPFLAGS_IS_ABSTRACT should have been accurate.
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":346
 *     # It looks like ABCMeta.__new__ has not finished running yet; we're
 *     # probably in __init_subclass__. We'll look for abstractmethods manually.
 *     for name, value in object.__dict__.items():             # <<<<<<<<<<<<<<
 *         if getattr(value, "__isabstractmethod__", False):
 *             return True
 */
  __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 346, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_4);
  __pyx_t_4 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_7, &__pyx_t_6, &__pyx_t_3, &__pyx_t_5, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":347
 *     # probably in __init_subclass__. We'll look for abstractmethods manually.
 *     for name, value in object.__dict__.items():
 *         if getattr(value, "__isabstractmethod__", False):             # <<<<<<<<<<<<<<
 *             return True
 *     for base in object.__bases__:
 */
    __pyx_t_5 = __Pyx_GetAttr3(__pyx_v_value, __pyx_n_u_isabstractmethod, Py_False); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":348
 *     for name, value in object.__dict__.items():
 *         if getattr(value, "__isabstractmethod__", False):
 *             return True             # <<<<<<<<<<<<<<
 *     for base in object.__bases__:
 *         for name in getattr(base, "__abstractmethods__", ()):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":347
 *     # probably in __init_subclass__. We'll look for abstractmethods manually.
 *     for name, value in object.__dict__.items():
 *         if getattr(value, "__isabstractmethod__", False):             # <<<<<<<<<<<<<<
 *             return True
 *     for base in object.__bases__:
 */
    }
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":349
 *         if getattr(value, "__isabstractmethod__", False):
 *             return True
 *     for base in object.__bases__:             # <<<<<<<<<<<<<<
 *         for name in getattr(base, "__abstractmethods__", ()):
 *             value = getattr(object, name, None)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_bases); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 349, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 349, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 349, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_10(__pyx_t_5);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 349, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":350
 *             return True
 *     for base in object.__bases__:
 *         for name in getattr(base, "__abstractmethods__", ()):             # <<<<<<<<<<<<<<
 *             value = getattr(object, name, None)
 *             if getattr(value, "__isabstractmethod__", False):
 */
    __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_base, __pyx_n_u_abstractmethods, __pyx_empty_tuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 350, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_11(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 350, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":351
 *     for base in object.__bases__:
 *         for name in getattr(base, "__abstractmethods__", ()):
 *             value = getattr(object, name, None)             # <<<<<<<<<<<<<<
 *             if getattr(value, "__isabstractmethod__", False):
 *                 return True
 */
      __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_object, __pyx_v_name, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":352
 *         for name in getattr(base, "__abstractmethods__", ()):
 *             value = getattr(object, name, None)
 *             if getattr(value, "__isabstractmethod__", False):             # <<<<<<<<<<<<<<
 *                 return True
 *     return False
 */
      __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_value, __pyx_n_u_isabstractmethod, Py_False); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/contracts/inspection.py":353
 *             value = getattr(object, name, None)
 *             if getattr(value, "__isabstractmethod__", False):
 *                 return True             # <<<<<<<<<<<<<<
 *     return False
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/contracts/inspection.py":352
 *         for name in getattr(base, "__abstractmethods__", ()):
 *             value = getattr(object, name, None)
 *             if getattr(value, "__isabstractmethod__", False):             # <<<<<<<<<<<<<<
 *                 return True
 *     return False
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":350
 *             return True
 *     for base in object.__bases__:
 *         for name in getattr(base, "__abstractmethods__", ()):             # <<<<<<<<<<<<<<
 *             value = getattr(object, name, None)
 *             if getattr(value, "__isabstractmethod__", False):
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":349
 *         if getattr(value, "__isabstractmethod__", False):
 *             return True
 *     for base in object.__bases__:             # <<<<<<<<<<<<<<
 *         for name in getattr(base, "__abstractmethods__", ()):
 *             value = getattr(object, name, None)
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":354
 *             if getattr(value, "__isabstractmethod__", False):
 *                 return True
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":332
 * 
 * 
 * def isabstract(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.isabstract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":357
 * 
 * 
 * def getmembers(object, predicate=None):             # <<<<<<<<<<<<<<
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_49getmembers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_48getmembers[] = "Return all members of an object as (name, value) pairs sorted by name.\n    Optionally, only return members that satisfy a given predicate.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_49getmembers = {"getmembers", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_49getmembers, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_48getmembers};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_49getmembers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_object = 0;
  PyObject *__pyx_v_predicate = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getmembers (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_object,&__pyx_n_s_predicate,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_object)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predicate);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getmembers") < 0)) __PYX_ERR(0, 357, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_object = values[0];
    __pyx_v_predicate = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getmembers", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 357, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmembers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_48getmembers(__pyx_self, __pyx_v_object, __pyx_v_predicate);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":398
 *             results.append((key, value))
 *         processed.add(key)
 *     results.sort(key=lambda pair: pair[0])             # <<<<<<<<<<<<<<
 *     return results
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_10getmembers_lambda(PyObject *__pyx_self, PyObject *__pyx_v_pair); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_10getmembers_lambda = {"lambda", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_10getmembers_lambda, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_10getmembers_lambda(PyObject *__pyx_self, PyObject *__pyx_v_pair) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, ((PyObject *)__pyx_v_pair));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pair) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_pair, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmembers.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":357
 * 
 * 
 * def getmembers(object, predicate=None):             # <<<<<<<<<<<<<<
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_48getmembers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v_predicate) {
  PyObject *__pyx_v_mro = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_processed = NULL;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getmembers", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":360
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         mro = (object,) + getmro(object)
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":361
 *     Optionally, only return members that satisfy a given predicate."""
 *     if isclass(object):
 *         mro = (object,) + getmro(object)             # <<<<<<<<<<<<<<
 *     else:
 *         mro = ()
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_object);
    __Pyx_GIVEREF(__pyx_v_object);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_object);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getmro); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_mro = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":360
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         mro = (object,) + getmro(object)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":363
 *         mro = (object,) + getmro(object)
 *     else:
 *         mro = ()             # <<<<<<<<<<<<<<
 *     results = []
 *     processed = set()
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_v_mro = __pyx_empty_tuple;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":364
 *     else:
 *         mro = ()
 *     results = []             # <<<<<<<<<<<<<<
 *     processed = set()
 *     names = dir(object)
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_results = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":365
 *         mro = ()
 *     results = []
 *     processed = set()             # <<<<<<<<<<<<<<
 *     names = dir(object)
 *     # :dd any DynamicClassAttributes to the list of names if object is a class;
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_processed = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":366
 *     results = []
 *     processed = set()
 *     names = dir(object)             # <<<<<<<<<<<<<<
 *     # :dd any DynamicClassAttributes to the list of names if object is a class;
 *     # this may result in duplicate entries if, for example, a virtual
 */
  __pyx_t_3 = PyObject_Dir(__pyx_v_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_names = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":370
 *     # this may result in duplicate entries if, for example, a virtual
 *     # attribute with the same name as a DynamicClassAttribute exists
 *     try:             # <<<<<<<<<<<<<<
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":371
 *     # attribute with the same name as a DynamicClassAttribute exists
 *     try:
 *         for base in object.__bases__:             # <<<<<<<<<<<<<<
 *             for k, v in base.__dict__.items():
 *                 if isinstance(v, types.DynamicClassAttribute):
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_bases); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 371, __pyx_L4_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 371, __pyx_L4_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 371, __pyx_L4_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_10(__pyx_t_2);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 371, __pyx_L4_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":372
 *     try:
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():             # <<<<<<<<<<<<<<
 *                 if isinstance(v, types.DynamicClassAttribute):
 *                     names.append(k)
 */
        __pyx_t_11 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__pyx_t_1 == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
          __PYX_ERR(0, 372, __pyx_L4_error)
        }
        __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_1, 0, __pyx_n_s_items, (&__pyx_t_12), (&__pyx_t_13)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 372, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3);
        __pyx_t_3 = __pyx_t_5;
        __pyx_t_5 = 0;
        while (1) {
          __pyx_t_14 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_12, &__pyx_t_11, &__pyx_t_5, &__pyx_t_1, NULL, __pyx_t_13);
          if (unlikely(__pyx_t_14 == 0)) break;
          if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 372, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":373
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():
 *                 if isinstance(v, types.DynamicClassAttribute):             # <<<<<<<<<<<<<<
 *                     names.append(k)
 *     except AttributeError:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DynamicClassAttribute); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 373, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_4 = PyObject_IsInstance(__pyx_v_v, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 373, __pyx_L4_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_15 = (__pyx_t_4 != 0);
          if (__pyx_t_15) {

            /* "Aspidites/_vendor/contracts/inspection.py":374
 *             for k, v in base.__dict__.items():
 *                 if isinstance(v, types.DynamicClassAttribute):
 *                     names.append(k)             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         pass
 */
            __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_names, __pyx_v_k); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 374, __pyx_L4_error)

            /* "Aspidites/_vendor/contracts/inspection.py":373
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():
 *                 if isinstance(v, types.DynamicClassAttribute):             # <<<<<<<<<<<<<<
 *                     names.append(k)
 *     except AttributeError:
 */
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":371
 *     # attribute with the same name as a DynamicClassAttribute exists
 *     try:
 *         for base in object.__bases__:             # <<<<<<<<<<<<<<
 *             for k, v in base.__dict__.items():
 *                 if isinstance(v, types.DynamicClassAttribute):
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":370
 *     # this may result in duplicate entries if, for example, a virtual
 *     # attribute with the same name as a DynamicClassAttribute exists
 *     try:             # <<<<<<<<<<<<<<
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":375
 *                 if isinstance(v, types.DynamicClassAttribute):
 *                     names.append(k)
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         pass
 *     for key in names:
 */
    __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_13) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":370
 *     # this may result in duplicate entries if, for example, a virtual
 *     # attribute with the same name as a DynamicClassAttribute exists
 *     try:             # <<<<<<<<<<<<<<
 *         for base in object.__bases__:
 *             for k, v in base.__dict__.items():
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    __pyx_L9_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":377
 *     except AttributeError:
 *         pass
 *     for key in names:             # <<<<<<<<<<<<<<
 *         # First try to get the value via getattr.  Some descriptors don't
 *         # like calling their __get__ (see bug #1785), so fall back to
 */
  if (likely(PyList_CheckExact(__pyx_v_names)) || PyTuple_CheckExact(__pyx_v_names)) {
    __pyx_t_2 = __pyx_v_names; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 377, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_10(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 377, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":381
 *         # like calling their __get__ (see bug #1785), so fall back to
 *         # looking in the __dict__.
 *         try:             # <<<<<<<<<<<<<<
 *             value = getattr(object, key)
 *             # handle the duplicate key
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":382
 *         # looking in the __dict__.
 *         try:
 *             value = getattr(object, key)             # <<<<<<<<<<<<<<
 *             # handle the duplicate key
 *             if key in processed:
 */
        __pyx_t_3 = __Pyx_GetAttr(__pyx_v_object, __pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":384
 *             value = getattr(object, key)
 *             # handle the duplicate key
 *             if key in processed:             # <<<<<<<<<<<<<<
 *                 raise AttributeError
 *         except AttributeError:
 */
        __pyx_t_15 = (__Pyx_PySet_ContainsTF(__pyx_v_key, __pyx_v_processed, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 384, __pyx_L17_error)
        __pyx_t_4 = (__pyx_t_15 != 0);
        if (unlikely(__pyx_t_4)) {

          /* "Aspidites/_vendor/contracts/inspection.py":385
 *             # handle the duplicate key
 *             if key in processed:
 *                 raise AttributeError             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             for base in mro:
 */
          __Pyx_Raise(__pyx_builtin_AttributeError, 0, 0, 0);
          __PYX_ERR(0, 385, __pyx_L17_error)

          /* "Aspidites/_vendor/contracts/inspection.py":384
 *             value = getattr(object, key)
 *             # handle the duplicate key
 *             if key in processed:             # <<<<<<<<<<<<<<
 *                 raise AttributeError
 *         except AttributeError:
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":381
 *         # like calling their __get__ (see bug #1785), so fall back to
 *         # looking in the __dict__.
 *         try:             # <<<<<<<<<<<<<<
 *             value = getattr(object, key)
 *             # handle the duplicate key
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L24_try_end;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":386
 *             if key in processed:
 *                 raise AttributeError
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             for base in mro:
 *                 if key in base.__dict__:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_13) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmembers", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(0, 386, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Aspidites/_vendor/contracts/inspection.py":387
 *                 raise AttributeError
 *         except AttributeError:
 *             for base in mro:             # <<<<<<<<<<<<<<
 *                 if key in base.__dict__:
 *                     value = base.__dict__[key]
 */
        if (likely(PyList_CheckExact(__pyx_v_mro)) || PyTuple_CheckExact(__pyx_v_mro)) {
          __pyx_t_17 = __pyx_v_mro; __Pyx_INCREF(__pyx_t_17); __pyx_t_12 = 0;
          __pyx_t_18 = NULL;
        } else {
          __pyx_t_12 = -1; __pyx_t_17 = PyObject_GetIter(__pyx_v_mro); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 387, __pyx_L19_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_18 = Py_TYPE(__pyx_t_17)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 387, __pyx_L19_except_error)
        }
        for (;;) {
          if (likely(!__pyx_t_18)) {
            if (likely(PyList_CheckExact(__pyx_t_17))) {
              if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_17)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_19 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_12); __Pyx_INCREF(__pyx_t_19); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 387, __pyx_L19_except_error)
              #else
              __pyx_t_19 = PySequence_ITEM(__pyx_t_17, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 387, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_19);
              #endif
            } else {
              if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_17)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_19 = PyTuple_GET_ITEM(__pyx_t_17, __pyx_t_12); __Pyx_INCREF(__pyx_t_19); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 387, __pyx_L19_except_error)
              #else
              __pyx_t_19 = PySequence_ITEM(__pyx_t_17, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 387, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_19);
              #endif
            }
          } else {
            __pyx_t_19 = __pyx_t_18(__pyx_t_17);
            if (unlikely(!__pyx_t_19)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 387, __pyx_L19_except_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_19);
          }
          __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_19);
          __pyx_t_19 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":388
 *         except AttributeError:
 *             for base in mro:
 *                 if key in base.__dict__:             # <<<<<<<<<<<<<<
 *                     value = base.__dict__[key]
 *                     break
 */
          __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 388, __pyx_L19_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_t_19, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 388, __pyx_L19_except_error)
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __pyx_t_15 = (__pyx_t_4 != 0);
          if (__pyx_t_15) {

            /* "Aspidites/_vendor/contracts/inspection.py":389
 *             for base in mro:
 *                 if key in base.__dict__:
 *                     value = base.__dict__[key]             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
            __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 389, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_20 = __Pyx_PyObject_GetItem(__pyx_t_19, __pyx_v_key); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 389, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_20);
            __pyx_t_20 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":390
 *                 if key in base.__dict__:
 *                     value = base.__dict__[key]
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 # could be a (currently) missing slot member, or a buggy
 */
            goto __pyx_L29_break;

            /* "Aspidites/_vendor/contracts/inspection.py":388
 *         except AttributeError:
 *             for base in mro:
 *                 if key in base.__dict__:             # <<<<<<<<<<<<<<
 *                     value = base.__dict__[key]
 *                     break
 */
          }

          /* "Aspidites/_vendor/contracts/inspection.py":387
 *                 raise AttributeError
 *         except AttributeError:
 *             for base in mro:             # <<<<<<<<<<<<<<
 *                 if key in base.__dict__:
 *                     value = base.__dict__[key]
 */
        }
        /*else*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":394
 *                 # could be a (currently) missing slot member, or a buggy
 *                 # __dir__; discard and move on
 *                 continue             # <<<<<<<<<<<<<<
 *         if not predicate or predicate(value):
 *             results.append((key, value))
 */
          goto __pyx_L31_outer_continue;
        }
        goto __pyx_L29_break;

        /* "Aspidites/_vendor/contracts/inspection.py":387
 *                 raise AttributeError
 *         except AttributeError:
 *             for base in mro:             # <<<<<<<<<<<<<<
 *                 if key in base.__dict__:
 *                     value = base.__dict__[key]
 */
        __pyx_L31_outer_continue:;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        goto __pyx_L27_except_continue;
        __pyx_L29_break:;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L18_exception_handled;
        __pyx_L27_except_continue:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L23_try_continue;
      }
      goto __pyx_L19_except_error;
      __pyx_L19_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":381
 *         # like calling their __get__ (see bug #1785), so fall back to
 *         # looking in the __dict__.
 *         try:             # <<<<<<<<<<<<<<
 *             value = getattr(object, key)
 *             # handle the duplicate key
 */
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L23_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L15_continue;
      __pyx_L18_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      __pyx_L24_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":395
 *                 # __dir__; discard and move on
 *                 continue
 *         if not predicate or predicate(value):             # <<<<<<<<<<<<<<
 *             results.append((key, value))
 *         processed.add(key)
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_predicate); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 395, __pyx_L1_error)
    __pyx_t_21 = ((!__pyx_t_4) != 0);
    if (!__pyx_t_21) {
    } else {
      __pyx_t_15 = __pyx_t_21;
      goto __pyx_L33_bool_binop_done;
    }
    __Pyx_INCREF(__pyx_v_predicate);
    __pyx_t_5 = __pyx_v_predicate; __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_21 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_15 = __pyx_t_21;
    __pyx_L33_bool_binop_done:;
    if (__pyx_t_15) {

      /* "Aspidites/_vendor/contracts/inspection.py":396
 *                 continue
 *         if not predicate or predicate(value):
 *             results.append((key, value))             # <<<<<<<<<<<<<<
 *         processed.add(key)
 *     results.sort(key=lambda pair: pair[0])
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_key);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_value);
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_1); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":395
 *                 # __dir__; discard and move on
 *                 continue
 *         if not predicate or predicate(value):             # <<<<<<<<<<<<<<
 *             results.append((key, value))
 *         processed.add(key)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":397
 *         if not predicate or predicate(value):
 *             results.append((key, value))
 *         processed.add(key)             # <<<<<<<<<<<<<<
 *     results.sort(key=lambda pair: pair[0])
 *     return results
 */
    __pyx_t_16 = PySet_Add(__pyx_v_processed, __pyx_v_key); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 397, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":377
 *     except AttributeError:
 *         pass
 *     for key in names:             # <<<<<<<<<<<<<<
 *         # First try to get the value via getattr.  Some descriptors don't
 *         # like calling their __get__ (see bug #1785), so fall back to
 */
    __pyx_L15_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":398
 *             results.append((key, value))
 *         processed.add(key)
 *     results.sort(key=lambda pair: pair[0])             # <<<<<<<<<<<<<<
 *     return results
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_results, __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_10getmembers_lambda, 0, __pyx_n_s_getmembers_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":399
 *         processed.add(key)
 *     results.sort(key=lambda pair: pair[0])
 *     return results             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_results);
  __pyx_r = __pyx_v_results;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":357
 * 
 * 
 * def getmembers(object, predicate=None):             # <<<<<<<<<<<<<<
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmembers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mro);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_processed);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_51classify_class_attrs(PyObject *__pyx_self, PyObject *__pyx_v_cls); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_50classify_class_attrs[] = "Return list of attribute-descriptor tuples.\n\n    For each name in dir(cls), the return list contains a 4-tuple\n    with these elements:\n\n        0. The name (a string).\n\n        1. The kind of attribute this is, one of these strings:\n               'class method'    created via classmethod()\n               'static method'   created via staticmethod()\n               'property'        created via property()\n               'method'          any other flavor of method or descriptor\n               'data'            not a method\n\n        2. The class which defined this attribute (a class).\n\n        3. The object as obtained by calling getattr; if this fails, or if the\n           resulting object does not live anywhere in the class' mro (including\n           metaclasses) then the object is looked up in the defining class's\n           dict (found by walking the mro).\n\n    If one of the items in dir(cls) is stored in the metaclass it will now\n    be discovered and not have None be listed as the class in which it was\n    defined.  Any items whose home class cannot be discovered are skipped.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_51classify_class_attrs = {"classify_class_attrs", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_51classify_class_attrs, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_50classify_class_attrs};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_51classify_class_attrs(PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("classify_class_attrs (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_50classify_class_attrs(__pyx_self, ((PyObject *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/contracts/inspection.py":434
 *     mro = getmro(cls)
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))             # <<<<<<<<<<<<<<
 *     class_bases = (cls,) + mro
 *     all_bases = class_bases + metamro
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 434, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_classify_class_attrs_locals_gene, __pyx_n_s_Aspidites__vendor_contracts_insp); if (unlikely(!gen)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.classify_class_attrs.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L9_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_metamro)) { __Pyx_RaiseClosureNameError("metamro"); __PYX_ERR(0, 434, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_metamro)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_metamro)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_metamro; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_metamro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 434, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_cls);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_cls, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cls);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_cls;
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, ((PyObject *)(&PyType_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 434, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_builtin_object, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 434, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cls);
      __pyx_r = __pyx_cur_scope->__pyx_v_cls;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L9_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 434, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_50classify_class_attrs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *__pyx_cur_scope;
  PyObject *__pyx_v_mro = NULL;
  PyObject *__pyx_v_class_bases = NULL;
  PyObject *__pyx_v_all_bases = NULL;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_processed = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_homecls = NULL;
  PyObject *__pyx_v_get_obj = NULL;
  PyObject *__pyx_v_dict_obj = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_last_cls = NULL;
  PyObject *__pyx_v_srch_cls = NULL;
  PyObject *__pyx_v_srch_obj = NULL;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *(*__pyx_t_17)(PyObject *);
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("classify_class_attrs", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 405, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/contracts/inspection.py":432
 *     """
 * 
 *     mro = getmro(cls)             # <<<<<<<<<<<<<<
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getmro); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_cls) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_cls);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mro = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":433
 * 
 *     mro = getmro(cls)
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass             # <<<<<<<<<<<<<<
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))
 *     class_bases = (cls,) + mro
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getmro); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_cls))) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)Py_TYPE(__pyx_v_cls)));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_metamro = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":434
 *     mro = getmro(cls)
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))             # <<<<<<<<<<<<<<
 *     class_bases = (cls,) + mro
 *     all_bases = class_bases + metamro
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_metamro);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_metamro, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":435
 *     metamro = getmro(type(cls))  # for attributes stored in the metaclass
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))
 *     class_bases = (cls,) + mro             # <<<<<<<<<<<<<<
 *     all_bases = class_bases + metamro
 *     names = dir(cls)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_cls);
  __Pyx_GIVEREF(__pyx_v_cls);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_cls);
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_mro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_class_bases = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":436
 *     metamro = tuple(cls for cls in metamro if cls not in (type, object))
 *     class_bases = (cls,) + mro
 *     all_bases = class_bases + metamro             # <<<<<<<<<<<<<<
 *     names = dir(cls)
 *     # :dd any DynamicClassAttributes to the list of names;
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_class_bases, __pyx_cur_scope->__pyx_v_metamro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_all_bases = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":437
 *     class_bases = (cls,) + mro
 *     all_bases = class_bases + metamro
 *     names = dir(cls)             # <<<<<<<<<<<<<<
 *     # :dd any DynamicClassAttributes to the list of names;
 *     # this may result in duplicate entries if, for example, a virtual
 */
  __pyx_t_1 = PyObject_Dir(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_names = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":441
 *     # this may result in duplicate entries if, for example, a virtual
 *     # attribute with the same name as a DynamicClassAttribute exists.
 *     for base in mro:             # <<<<<<<<<<<<<<
 *         for k, v in base.__dict__.items():
 *             if isinstance(v, types.DynamicClassAttribute):
 */
  if (likely(PyList_CheckExact(__pyx_v_mro)) || PyTuple_CheckExact(__pyx_v_mro)) {
    __pyx_t_1 = __pyx_v_mro; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_mro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 441, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 441, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":442
 *     # attribute with the same name as a DynamicClassAttribute exists.
 *     for base in mro:
 *         for k, v in base.__dict__.items():             # <<<<<<<<<<<<<<
 *             if isinstance(v, types.DynamicClassAttribute):
 *                 names.append(k)
 */
    __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_t_3 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 442, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_dict_iterator(__pyx_t_3, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_9;
    __pyx_t_9 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_7, &__pyx_t_6, &__pyx_t_9, &__pyx_t_3, NULL, __pyx_t_8);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 442, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":443
 *     for base in mro:
 *         for k, v in base.__dict__.items():
 *             if isinstance(v, types.DynamicClassAttribute):             # <<<<<<<<<<<<<<
 *                 names.append(k)
 *     result = []
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DynamicClassAttribute); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = PyObject_IsInstance(__pyx_v_v, __pyx_t_9); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 443, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Aspidites/_vendor/contracts/inspection.py":444
 *         for k, v in base.__dict__.items():
 *             if isinstance(v, types.DynamicClassAttribute):
 *                 names.append(k)             # <<<<<<<<<<<<<<
 *     result = []
 *     processed = set()
 */
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_names, __pyx_v_k); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 444, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":443
 *     for base in mro:
 *         for k, v in base.__dict__.items():
 *             if isinstance(v, types.DynamicClassAttribute):             # <<<<<<<<<<<<<<
 *                 names.append(k)
 *     result = []
 */
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":441
 *     # this may result in duplicate entries if, for example, a virtual
 *     # attribute with the same name as a DynamicClassAttribute exists.
 *     for base in mro:             # <<<<<<<<<<<<<<
 *         for k, v in base.__dict__.items():
 *             if isinstance(v, types.DynamicClassAttribute):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":445
 *             if isinstance(v, types.DynamicClassAttribute):
 *                 names.append(k)
 *     result = []             # <<<<<<<<<<<<<<
 *     processed = set()
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":446
 *                 names.append(k)
 *     result = []
 *     processed = set()             # <<<<<<<<<<<<<<
 * 
 *     for name in names:
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_processed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":448
 *     processed = set()
 * 
 *     for name in names:             # <<<<<<<<<<<<<<
 *         # Get the object associated with the name, and where it was defined.
 *         # Normal objects will be looked up with both getattr and directly in
 */
  if (likely(PyList_CheckExact(__pyx_v_names)) || PyTuple_CheckExact(__pyx_v_names)) {
    __pyx_t_1 = __pyx_v_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 448, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 448, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 448, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 448, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":458
 *         # Getting an obj from the __dict__ sometimes reveals more than
 *         # using getattr.  Static and class methods are dramatic examples.
 *         homecls = None             # <<<<<<<<<<<<<<
 *         get_obj = None
 *         dict_obj = None
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_homecls, Py_None);

    /* "Aspidites/_vendor/contracts/inspection.py":459
 *         # using getattr.  Static and class methods are dramatic examples.
 *         homecls = None
 *         get_obj = None             # <<<<<<<<<<<<<<
 *         dict_obj = None
 *         if name not in processed:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_get_obj, Py_None);

    /* "Aspidites/_vendor/contracts/inspection.py":460
 *         homecls = None
 *         get_obj = None
 *         dict_obj = None             # <<<<<<<<<<<<<<
 *         if name not in processed:
 *             try:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_dict_obj, Py_None);

    /* "Aspidites/_vendor/contracts/inspection.py":461
 *         get_obj = None
 *         dict_obj = None
 *         if name not in processed:             # <<<<<<<<<<<<<<
 *             try:
 *                 if name == '__dict__':
 */
    __pyx_t_12 = (__Pyx_PySet_ContainsTF(__pyx_v_name, __pyx_v_processed, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 461, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_12 != 0);
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":462
 *         dict_obj = None
 *         if name not in processed:
 *             try:             # <<<<<<<<<<<<<<
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":463
 *         if name not in processed:
 *             try:
 *                 if name == '__dict__':             # <<<<<<<<<<<<<<
 *                     raise Exception("__dict__ is special, don't want the proxy")
 *                 get_obj = getattr(cls, name)
 */
          __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_dict, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 463, __pyx_L11_error)
          if (unlikely(__pyx_t_11)) {

            /* "Aspidites/_vendor/contracts/inspection.py":464
 *             try:
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")             # <<<<<<<<<<<<<<
 *                 get_obj = getattr(cls, name)
 *             except Exception as exc:
 */
            __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 464, __pyx_L11_error)

            /* "Aspidites/_vendor/contracts/inspection.py":463
 *         if name not in processed:
 *             try:
 *                 if name == '__dict__':             # <<<<<<<<<<<<<<
 *                     raise Exception("__dict__ is special, don't want the proxy")
 *                 get_obj = getattr(cls, name)
 */
          }

          /* "Aspidites/_vendor/contracts/inspection.py":465
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")
 *                 get_obj = getattr(cls, name)             # <<<<<<<<<<<<<<
 *             except Exception as exc:
 *                 pass
 */
          __pyx_t_2 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_get_obj, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":462
 *         dict_obj = None
 *         if name not in processed:
 *             try:             # <<<<<<<<<<<<<<
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":469
 *                 pass
 *             else:
 *                 homecls = getattr(get_obj, "__objclass__", homecls)             # <<<<<<<<<<<<<<
 *                 if homecls not in class_bases:
 *                     # if the resulting object does not live somewhere in the
 */
        /*else:*/ {
          __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_get_obj, __pyx_n_u_objclass, __pyx_v_homecls); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_homecls, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":470
 *             else:
 *                 homecls = getattr(get_obj, "__objclass__", homecls)
 *                 if homecls not in class_bases:             # <<<<<<<<<<<<<<
 *                     # if the resulting object does not live somewhere in the
 *                     # mro, drop it and search the mro manually
 */
          __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_v_homecls, __pyx_v_class_bases, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 470, __pyx_L13_except_error)
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "Aspidites/_vendor/contracts/inspection.py":473
 *                     # if the resulting object does not live somewhere in the
 *                     # mro, drop it and search the mro manually
 *                     homecls = None             # <<<<<<<<<<<<<<
 *                     last_cls = None
 *                     # first look in the classes
 */
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_homecls, Py_None);

            /* "Aspidites/_vendor/contracts/inspection.py":474
 *                     # mro, drop it and search the mro manually
 *                     homecls = None
 *                     last_cls = None             # <<<<<<<<<<<<<<
 *                     # first look in the classes
 *                     for srch_cls in class_bases:
 */
            __Pyx_INCREF(Py_None);
            __Pyx_XDECREF_SET(__pyx_v_last_cls, Py_None);

            /* "Aspidites/_vendor/contracts/inspection.py":476
 *                     last_cls = None
 *                     # first look in the classes
 *                     for srch_cls in class_bases:             # <<<<<<<<<<<<<<
 *                         srch_obj = getattr(srch_cls, name, None)
 *                         if srch_obj is get_obj:
 */
            if (likely(PyList_CheckExact(__pyx_v_class_bases)) || PyTuple_CheckExact(__pyx_v_class_bases)) {
              __pyx_t_2 = __pyx_v_class_bases; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
              __pyx_t_17 = NULL;
            } else {
              __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_class_bases); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L13_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_17 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 476, __pyx_L13_except_error)
            }
            for (;;) {
              if (likely(!__pyx_t_17)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_9 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L13_except_error)
                  #else
                  __pyx_t_9 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 476, __pyx_L13_except_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  #endif
                } else {
                  if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L13_except_error)
                  #else
                  __pyx_t_9 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 476, __pyx_L13_except_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  #endif
                }
              } else {
                __pyx_t_9 = __pyx_t_17(__pyx_t_2);
                if (unlikely(!__pyx_t_9)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 476, __pyx_L13_except_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_9);
              }
              __Pyx_XDECREF_SET(__pyx_v_srch_cls, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":477
 *                     # first look in the classes
 *                     for srch_cls in class_bases:
 *                         srch_obj = getattr(srch_cls, name, None)             # <<<<<<<<<<<<<<
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls
 */
              __pyx_t_9 = __Pyx_GetAttr3(__pyx_v_srch_cls, __pyx_v_name, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 477, __pyx_L13_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_XDECREF_SET(__pyx_v_srch_obj, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":478
 *                     for srch_cls in class_bases:
 *                         srch_obj = getattr(srch_cls, name, None)
 *                         if srch_obj is get_obj:             # <<<<<<<<<<<<<<
 *                             last_cls = srch_cls
 *                     # then check the metaclasses
 */
              __pyx_t_12 = (__pyx_v_srch_obj == __pyx_v_get_obj);
              __pyx_t_11 = (__pyx_t_12 != 0);
              if (__pyx_t_11) {

                /* "Aspidites/_vendor/contracts/inspection.py":479
 *                         srch_obj = getattr(srch_cls, name, None)
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls             # <<<<<<<<<<<<<<
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:
 */
                __Pyx_INCREF(__pyx_v_srch_cls);
                __Pyx_DECREF_SET(__pyx_v_last_cls, __pyx_v_srch_cls);

                /* "Aspidites/_vendor/contracts/inspection.py":478
 *                     for srch_cls in class_bases:
 *                         srch_obj = getattr(srch_cls, name, None)
 *                         if srch_obj is get_obj:             # <<<<<<<<<<<<<<
 *                             last_cls = srch_cls
 *                     # then check the metaclasses
 */
              }

              /* "Aspidites/_vendor/contracts/inspection.py":476
 *                     last_cls = None
 *                     # first look in the classes
 *                     for srch_cls in class_bases:             # <<<<<<<<<<<<<<
 *                         srch_obj = getattr(srch_cls, name, None)
 *                         if srch_obj is get_obj:
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":481
 *                             last_cls = srch_cls
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:             # <<<<<<<<<<<<<<
 *                         try:
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 */
            __pyx_t_2 = __pyx_cur_scope->__pyx_v_metamro; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
            for (;;) {
              if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 481, __pyx_L13_except_error)
              #else
              __pyx_t_9 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 481, __pyx_L13_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              #endif
              __Pyx_XDECREF_SET(__pyx_v_srch_cls, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":482
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:
 *                         try:             # <<<<<<<<<<<<<<
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 *                         except AttributeError:
 */
              {
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                /*try:*/ {

                  /* "Aspidites/_vendor/contracts/inspection.py":483
 *                     for srch_cls in metamro:
 *                         try:
 *                             srch_obj = srch_cls.__getattr__(cls, name)             # <<<<<<<<<<<<<<
 *                         except AttributeError:
 *                             continue
 */
                  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_srch_cls, __pyx_n_s_getattr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L26_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_21 = NULL;
                  __pyx_t_8 = 0;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                    __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_3);
                    if (likely(__pyx_t_21)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                      __Pyx_INCREF(__pyx_t_21);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_3, function);
                      __pyx_t_8 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_3)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_v_cls, __pyx_v_name};
                    __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 483, __pyx_L26_error)
                    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __Pyx_GOTREF(__pyx_t_9);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_v_cls, __pyx_v_name};
                    __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 483, __pyx_L26_error)
                    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __Pyx_GOTREF(__pyx_t_9);
                  } else
                  #endif
                  {
                    __pyx_t_22 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 483, __pyx_L26_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    if (__pyx_t_21) {
                      __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_21); __pyx_t_21 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_cls);
                    __Pyx_GIVEREF(__pyx_v_cls);
                    PyTuple_SET_ITEM(__pyx_t_22, 0+__pyx_t_8, __pyx_v_cls);
                    __Pyx_INCREF(__pyx_v_name);
                    __Pyx_GIVEREF(__pyx_v_name);
                    PyTuple_SET_ITEM(__pyx_t_22, 1+__pyx_t_8, __pyx_v_name);
                    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_22, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 483, __pyx_L26_error)
                    __Pyx_GOTREF(__pyx_t_9);
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_srch_obj, __pyx_t_9);
                  __pyx_t_9 = 0;

                  /* "Aspidites/_vendor/contracts/inspection.py":482
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:
 *                         try:             # <<<<<<<<<<<<<<
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 *                         except AttributeError:
 */
                }
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                goto __pyx_L33_try_end;
                __pyx_L26_error:;
                __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

                /* "Aspidites/_vendor/contracts/inspection.py":484
 *                         try:
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 *                         except AttributeError:             # <<<<<<<<<<<<<<
 *                             continue
 *                         if srch_obj is get_obj:
 */
                __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
                if (__pyx_t_8) {
                  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.classify_class_attrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
                  if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_3, &__pyx_t_22) < 0) __PYX_ERR(0, 484, __pyx_L28_except_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GOTREF(__pyx_t_22);

                  /* "Aspidites/_vendor/contracts/inspection.py":485
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 *                         except AttributeError:
 *                             continue             # <<<<<<<<<<<<<<
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls
 */
                  goto __pyx_L35_except_continue;
                  __pyx_L35_except_continue:;
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  goto __pyx_L32_try_continue;
                }
                goto __pyx_L28_except_error;
                __pyx_L28_except_error:;

                /* "Aspidites/_vendor/contracts/inspection.py":482
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:
 *                         try:             # <<<<<<<<<<<<<<
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 *                         except AttributeError:
 */
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                goto __pyx_L13_except_error;
                __pyx_L32_try_continue:;
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                goto __pyx_L24_continue;
                __pyx_L33_try_end:;
              }

              /* "Aspidites/_vendor/contracts/inspection.py":486
 *                         except AttributeError:
 *                             continue
 *                         if srch_obj is get_obj:             # <<<<<<<<<<<<<<
 *                             last_cls = srch_cls
 *                     if last_cls is not None:
 */
              __pyx_t_11 = (__pyx_v_srch_obj == __pyx_v_get_obj);
              __pyx_t_12 = (__pyx_t_11 != 0);
              if (__pyx_t_12) {

                /* "Aspidites/_vendor/contracts/inspection.py":487
 *                             continue
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls             # <<<<<<<<<<<<<<
 *                     if last_cls is not None:
 *                         homecls = last_cls
 */
                __Pyx_INCREF(__pyx_v_srch_cls);
                __Pyx_DECREF_SET(__pyx_v_last_cls, __pyx_v_srch_cls);

                /* "Aspidites/_vendor/contracts/inspection.py":486
 *                         except AttributeError:
 *                             continue
 *                         if srch_obj is get_obj:             # <<<<<<<<<<<<<<
 *                             last_cls = srch_cls
 *                     if last_cls is not None:
 */
              }

              /* "Aspidites/_vendor/contracts/inspection.py":481
 *                             last_cls = srch_cls
 *                     # then check the metaclasses
 *                     for srch_cls in metamro:             # <<<<<<<<<<<<<<
 *                         try:
 *                             srch_obj = srch_cls.__getattr__(cls, name)
 */
              __pyx_L24_continue:;
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":488
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls
 *                     if last_cls is not None:             # <<<<<<<<<<<<<<
 *                         homecls = last_cls
 *         for base in all_bases:
 */
            __pyx_t_12 = (__pyx_v_last_cls != Py_None);
            __pyx_t_11 = (__pyx_t_12 != 0);
            if (__pyx_t_11) {

              /* "Aspidites/_vendor/contracts/inspection.py":489
 *                             last_cls = srch_cls
 *                     if last_cls is not None:
 *                         homecls = last_cls             # <<<<<<<<<<<<<<
 *         for base in all_bases:
 *             if name in base.__dict__:
 */
              __Pyx_INCREF(__pyx_v_last_cls);
              __Pyx_DECREF_SET(__pyx_v_homecls, __pyx_v_last_cls);

              /* "Aspidites/_vendor/contracts/inspection.py":488
 *                         if srch_obj is get_obj:
 *                             last_cls = srch_cls
 *                     if last_cls is not None:             # <<<<<<<<<<<<<<
 *                         homecls = last_cls
 *         for base in all_bases:
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":470
 *             else:
 *                 homecls = getattr(get_obj, "__objclass__", homecls)
 *                 if homecls not in class_bases:             # <<<<<<<<<<<<<<
 *                     # if the resulting object does not live somewhere in the
 *                     # mro, drop it and search the mro manually
 */
          }
        }
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        goto __pyx_L18_try_end;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":466
 *                     raise Exception("__dict__ is special, don't want the proxy")
 *                 get_obj = getattr(cls, name)
 *             except Exception as exc:             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_8) {
          __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.classify_class_attrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_22, &__pyx_t_3) < 0) __PYX_ERR(0, 466, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_22);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_22);
          __pyx_v_exc = __pyx_t_22;
          /*try:*/ {
          }
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_exc);
              __pyx_v_exc = NULL;
              goto __pyx_L44;
            }
            __pyx_L44:;
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L12_exception_handled;
        }
        goto __pyx_L13_except_error;
        __pyx_L13_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":462
 *         dict_obj = None
 *         if name not in processed:
 *             try:             # <<<<<<<<<<<<<<
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")
 */
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        goto __pyx_L1_error;
        __pyx_L12_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        __pyx_L18_try_end:;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":461
 *         get_obj = None
 *         dict_obj = None
 *         if name not in processed:             # <<<<<<<<<<<<<<
 *             try:
 *                 if name == '__dict__':
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":490
 *                     if last_cls is not None:
 *                         homecls = last_cls
 *         for base in all_bases:             # <<<<<<<<<<<<<<
 *             if name in base.__dict__:
 *                 dict_obj = base.__dict__[name]
 */
    if (likely(PyList_CheckExact(__pyx_v_all_bases)) || PyTuple_CheckExact(__pyx_v_all_bases)) {
      __pyx_t_3 = __pyx_v_all_bases; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_17 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_all_bases); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_17 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 490, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_17)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_22 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_22); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
          #else
          __pyx_t_22 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 490, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_22);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_22 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_22); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
          #else
          __pyx_t_22 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 490, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_22);
          #endif
        }
      } else {
        __pyx_t_22 = __pyx_t_17(__pyx_t_3);
        if (unlikely(!__pyx_t_22)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 490, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_22);
      }
      __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_22);
      __pyx_t_22 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":491
 *                         homecls = last_cls
 *         for base in all_bases:
 *             if name in base.__dict__:             # <<<<<<<<<<<<<<
 *                 dict_obj = base.__dict__[name]
 *                 if homecls not in metamro:
 */
      __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_t_22, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Aspidites/_vendor/contracts/inspection.py":492
 *         for base in all_bases:
 *             if name in base.__dict__:
 *                 dict_obj = base.__dict__[name]             # <<<<<<<<<<<<<<
 *                 if homecls not in metamro:
 *                     homecls = base
 */
        __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_dict); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_22);
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_22, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_DECREF_SET(__pyx_v_dict_obj, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":493
 *             if name in base.__dict__:
 *                 dict_obj = base.__dict__[name]
 *                 if homecls not in metamro:             # <<<<<<<<<<<<<<
 *                     homecls = base
 *                 break
 */
        __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_v_homecls, __pyx_cur_scope->__pyx_v_metamro, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "Aspidites/_vendor/contracts/inspection.py":494
 *                 dict_obj = base.__dict__[name]
 *                 if homecls not in metamro:
 *                     homecls = base             # <<<<<<<<<<<<<<
 *                 break
 *         if homecls is None:
 */
          __Pyx_INCREF(__pyx_v_base);
          __Pyx_DECREF_SET(__pyx_v_homecls, __pyx_v_base);

          /* "Aspidites/_vendor/contracts/inspection.py":493
 *             if name in base.__dict__:
 *                 dict_obj = base.__dict__[name]
 *                 if homecls not in metamro:             # <<<<<<<<<<<<<<
 *                     homecls = base
 *                 break
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":495
 *                 if homecls not in metamro:
 *                     homecls = base
 *                 break             # <<<<<<<<<<<<<<
 *         if homecls is None:
 *             # unable to locate the attribute anywhere, most likely due to
 */
        goto __pyx_L46_break;

        /* "Aspidites/_vendor/contracts/inspection.py":491
 *                         homecls = last_cls
 *         for base in all_bases:
 *             if name in base.__dict__:             # <<<<<<<<<<<<<<
 *                 dict_obj = base.__dict__[name]
 *                 if homecls not in metamro:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":490
 *                     if last_cls is not None:
 *                         homecls = last_cls
 *         for base in all_bases:             # <<<<<<<<<<<<<<
 *             if name in base.__dict__:
 *                 dict_obj = base.__dict__[name]
 */
    }
    __pyx_L46_break:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":496
 *                     homecls = base
 *                 break
 *         if homecls is None:             # <<<<<<<<<<<<<<
 *             # unable to locate the attribute anywhere, most likely due to
 *             # buggy custom __dir__; discard and move on
 */
    __pyx_t_11 = (__pyx_v_homecls == Py_None);
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {

      /* "Aspidites/_vendor/contracts/inspection.py":499
 *             # unable to locate the attribute anywhere, most likely due to
 *             # buggy custom __dir__; discard and move on
 *             continue             # <<<<<<<<<<<<<<
 *         obj = get_obj if get_obj is not None else dict_obj
 *         # Classify the object or its descriptor.
 */
      goto __pyx_L8_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":496
 *                     homecls = base
 *                 break
 *         if homecls is None:             # <<<<<<<<<<<<<<
 *             # unable to locate the attribute anywhere, most likely due to
 *             # buggy custom __dir__; discard and move on
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":500
 *             # buggy custom __dir__; discard and move on
 *             continue
 *         obj = get_obj if get_obj is not None else dict_obj             # <<<<<<<<<<<<<<
 *         # Classify the object or its descriptor.
 *         if isinstance(dict_obj, (staticmethod, types.BuiltinMethodType)):
 */
    __pyx_t_12 = (__pyx_v_get_obj != Py_None);
    if ((__pyx_t_12 != 0)) {
      __Pyx_INCREF(__pyx_v_get_obj);
      __pyx_t_3 = __pyx_v_get_obj;
    } else {
      __Pyx_INCREF(__pyx_v_dict_obj);
      __pyx_t_3 = __pyx_v_dict_obj;
    }
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":502
 *         obj = get_obj if get_obj is not None else dict_obj
 *         # Classify the object or its descriptor.
 *         if isinstance(dict_obj, (staticmethod, types.BuiltinMethodType)):             # <<<<<<<<<<<<<<
 *             kind = "static method"
 *             obj = dict_obj
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_BuiltinMethodType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = PyObject_IsInstance(__pyx_v_dict_obj, __pyx_builtin_staticmethod); 
    __pyx_t_23 = (__pyx_t_11 != 0);
    if (!__pyx_t_23) {
    } else {
      __pyx_t_12 = __pyx_t_23;
      goto __pyx_L51_bool_binop_done;
    }
    __pyx_t_23 = PyObject_IsInstance(__pyx_v_dict_obj, __pyx_t_2); 
    __pyx_t_11 = (__pyx_t_23 != 0);
    __pyx_t_12 = __pyx_t_11;
    __pyx_L51_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = (__pyx_t_12 != 0);
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":503
 *         # Classify the object or its descriptor.
 *         if isinstance(dict_obj, (staticmethod, types.BuiltinMethodType)):
 *             kind = "static method"             # <<<<<<<<<<<<<<
 *             obj = dict_obj
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):
 */
      __Pyx_INCREF(__pyx_kp_u_static_method);
      __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_kp_u_static_method);

      /* "Aspidites/_vendor/contracts/inspection.py":504
 *         if isinstance(dict_obj, (staticmethod, types.BuiltinMethodType)):
 *             kind = "static method"
 *             obj = dict_obj             # <<<<<<<<<<<<<<
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):
 *             kind = "class method"
 */
      __Pyx_INCREF(__pyx_v_dict_obj);
      __Pyx_DECREF_SET(__pyx_v_obj, __pyx_v_dict_obj);

      /* "Aspidites/_vendor/contracts/inspection.py":502
 *         obj = get_obj if get_obj is not None else dict_obj
 *         # Classify the object or its descriptor.
 *         if isinstance(dict_obj, (staticmethod, types.BuiltinMethodType)):             # <<<<<<<<<<<<<<
 *             kind = "static method"
 *             obj = dict_obj
 */
      goto __pyx_L50;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":505
 *             kind = "static method"
 *             obj = dict_obj
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):             # <<<<<<<<<<<<<<
 *             kind = "class method"
 *             obj = dict_obj
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ClassMethodDescriptorType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = PyObject_IsInstance(__pyx_v_dict_obj, __pyx_builtin_classmethod); 
    __pyx_t_23 = (__pyx_t_12 != 0);
    if (!__pyx_t_23) {
    } else {
      __pyx_t_11 = __pyx_t_23;
      goto __pyx_L53_bool_binop_done;
    }
    __pyx_t_23 = PyObject_IsInstance(__pyx_v_dict_obj, __pyx_t_3); 
    __pyx_t_12 = (__pyx_t_23 != 0);
    __pyx_t_11 = __pyx_t_12;
    __pyx_L53_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {

      /* "Aspidites/_vendor/contracts/inspection.py":506
 *             obj = dict_obj
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):
 *             kind = "class method"             # <<<<<<<<<<<<<<
 *             obj = dict_obj
 *         elif isinstance(dict_obj, property):
 */
      __Pyx_INCREF(__pyx_kp_u_class_method);
      __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_kp_u_class_method);

      /* "Aspidites/_vendor/contracts/inspection.py":507
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):
 *             kind = "class method"
 *             obj = dict_obj             # <<<<<<<<<<<<<<
 *         elif isinstance(dict_obj, property):
 *             kind = "property"
 */
      __Pyx_INCREF(__pyx_v_dict_obj);
      __Pyx_DECREF_SET(__pyx_v_obj, __pyx_v_dict_obj);

      /* "Aspidites/_vendor/contracts/inspection.py":505
 *             kind = "static method"
 *             obj = dict_obj
 *         elif isinstance(dict_obj, (classmethod, types.ClassMethodDescriptorType)):             # <<<<<<<<<<<<<<
 *             kind = "class method"
 *             obj = dict_obj
 */
      goto __pyx_L50;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":508
 *             kind = "class method"
 *             obj = dict_obj
 *         elif isinstance(dict_obj, property):             # <<<<<<<<<<<<<<
 *             kind = "property"
 *             obj = dict_obj
 */
    __pyx_t_12 = PyObject_IsInstance(__pyx_v_dict_obj, __pyx_builtin_property); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 508, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_12 != 0);
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":509
 *             obj = dict_obj
 *         elif isinstance(dict_obj, property):
 *             kind = "property"             # <<<<<<<<<<<<<<
 *             obj = dict_obj
 *         elif isroutine(obj):
 */
      __Pyx_INCREF(__pyx_n_u_property);
      __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_n_u_property);

      /* "Aspidites/_vendor/contracts/inspection.py":510
 *         elif isinstance(dict_obj, property):
 *             kind = "property"
 *             obj = dict_obj             # <<<<<<<<<<<<<<
 *         elif isroutine(obj):
 *             kind = "method"
 */
      __Pyx_INCREF(__pyx_v_dict_obj);
      __Pyx_DECREF_SET(__pyx_v_obj, __pyx_v_dict_obj);

      /* "Aspidites/_vendor/contracts/inspection.py":508
 *             kind = "class method"
 *             obj = dict_obj
 *         elif isinstance(dict_obj, property):             # <<<<<<<<<<<<<<
 *             kind = "property"
 *             obj = dict_obj
 */
      goto __pyx_L50;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":511
 *             kind = "property"
 *             obj = dict_obj
 *         elif isroutine(obj):             # <<<<<<<<<<<<<<
 *             kind = "method"
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isroutine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_22)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_22);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_22) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_22, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":512
 *             obj = dict_obj
 *         elif isroutine(obj):
 *             kind = "method"             # <<<<<<<<<<<<<<
 *         else:
 *             kind = "data"
 */
      __Pyx_INCREF(__pyx_n_u_method);
      __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_n_u_method);

      /* "Aspidites/_vendor/contracts/inspection.py":511
 *             kind = "property"
 *             obj = dict_obj
 *         elif isroutine(obj):             # <<<<<<<<<<<<<<
 *             kind = "method"
 *         else:
 */
      goto __pyx_L50;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":514
 *             kind = "method"
 *         else:
 *             kind = "data"             # <<<<<<<<<<<<<<
 *         result.append(Attribute(name, kind, homecls, obj))
 *         processed.add(name)
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_u_data);
      __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_n_u_data);
    }
    __pyx_L50:;

    /* "Aspidites/_vendor/contracts/inspection.py":515
 *         else:
 *             kind = "data"
 *         result.append(Attribute(name, kind, homecls, obj))             # <<<<<<<<<<<<<<
 *         processed.add(name)
 *     return result
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Attribute); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_22)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_22);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_22, __pyx_v_name, __pyx_v_kind, __pyx_v_homecls, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_22, __pyx_v_name, __pyx_v_kind, __pyx_v_homecls, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_22) {
        __Pyx_GIVEREF(__pyx_t_22); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_22); __pyx_t_22 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_name);
      __Pyx_INCREF(__pyx_v_kind);
      __Pyx_GIVEREF(__pyx_v_kind);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_kind);
      __Pyx_INCREF(__pyx_v_homecls);
      __Pyx_GIVEREF(__pyx_v_homecls);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_homecls);
      __Pyx_INCREF(__pyx_v_obj);
      __Pyx_GIVEREF(__pyx_v_obj);
      PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_v_obj);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":516
 *             kind = "data"
 *         result.append(Attribute(name, kind, homecls, obj))
 *         processed.add(name)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_t_13 = PySet_Add(__pyx_v_processed, __pyx_v_name); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 516, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":448
 *     processed = set()
 * 
 *     for name in names:             # <<<<<<<<<<<<<<
 *         # Get the object associated with the name, and where it was defined.
 *         # Normal objects will be looked up with both getattr and directly in
 */
    __pyx_L8_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":517
 *         result.append(Attribute(name, kind, homecls, obj))
 *         processed.add(name)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.classify_class_attrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mro);
  __Pyx_XDECREF(__pyx_v_class_bases);
  __Pyx_XDECREF(__pyx_v_all_bases);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_processed);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_homecls);
  __Pyx_XDECREF(__pyx_v_get_obj);
  __Pyx_XDECREF(__pyx_v_dict_obj);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_last_cls);
  __Pyx_XDECREF(__pyx_v_srch_cls);
  __Pyx_XDECREF(__pyx_v_srch_obj);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_20classify_class_attrs_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":522
 * # ----------------------------------------------------------- class helpers
 * 
 * def getmro(cls):             # <<<<<<<<<<<<<<
 *     """Return tuple of base classes (including cls) in method resolution order."""
 *     return cls.__mro__
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_53getmro(PyObject *__pyx_self, PyObject *__pyx_v_cls); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_52getmro[] = "Return tuple of base classes (including cls) in method resolution order.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_53getmro = {"getmro", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_53getmro, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_52getmro};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_53getmro(PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getmro (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_52getmro(__pyx_self, ((PyObject *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_52getmro(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getmro", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":524
 * def getmro(cls):
 *     """Return tuple of base classes (including cls) in method resolution order."""
 *     return cls.__mro__             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_mro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":522
 * # ----------------------------------------------------------- class helpers
 * 
 * def getmro(cls):             # <<<<<<<<<<<<<<
 *     """Return tuple of base classes (including cls) in method resolution order."""
 *     return cls.__mro__
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmro", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_55unwrap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_54unwrap[] = "Get the object wrapped by *func*.\n\n   Follows the chain of :attr:`__wrapped__` attributes returning the last\n   object in the chain.\n\n   *stop* is an optional callback accepting an object in the wrapper chain\n   as its sole argument that allows the unwrapping to be terminated early if\n   the callback returns a true value. If the callback never returns a true\n   value, the last object in the chain is returned as usual. For example,\n   :func:`signature` uses this to stop unwrapping if any object in the\n   chain has a ``__signature__`` attribute defined.\n\n   :exc:`ValueError` is raised if a cycle is encountered.\n\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_55unwrap = {"unwrap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_55unwrap, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_54unwrap};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_55unwrap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_stop = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unwrap (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func_2,&__pyx_n_s_stop,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 1;
        PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
        if (value) { values[index] = value; kw_args--; }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "unwrap") < 0)) __PYX_ERR(0, 529, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
    __pyx_v_stop = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("unwrap", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 529, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.unwrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_54unwrap(__pyx_self, __pyx_v_func, __pyx_v_stop);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":546
 *     """
 *     if stop is None:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__')
 *     else:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_1_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_1_is_wrapper = {"_is_wrapper", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_1_is_wrapper, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_1_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_wrapper (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap__is_wrapper(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap__is_wrapper(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_wrapper", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":547
 *     if stop is None:
 *         def _is_wrapper(f):
 *             return hasattr(f, '__wrapped__')             # <<<<<<<<<<<<<<
 *     else:
 *         def _is_wrapper(f):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_f, __pyx_n_u_wrapped); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 547, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":546
 *     """
 *     if stop is None:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__')
 *     else:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.unwrap._is_wrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":549
 *             return hasattr(f, '__wrapped__')
 *     else:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__') and not stop(f)
 *     f = func  # remember the original func for error reporting
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_3_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_3_is_wrapper = {"_is_wrapper", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_3_is_wrapper, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_3_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_wrapper (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_2_is_wrapper(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_2_is_wrapper(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_wrapper", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":550
 *     else:
 *         def _is_wrapper(f):
 *             return hasattr(f, '__wrapped__') and not stop(f)             # <<<<<<<<<<<<<<
 *     f = func  # remember the original func for error reporting
 *     # Memoise by id to tolerate non-hashable objects, but store objects to
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_f, __pyx_n_u_wrapped); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 550, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_stop)) { __Pyx_RaiseClosureNameError("stop"); __PYX_ERR(0, 550, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stop);
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_stop; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_f);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (!__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":549
 *             return hasattr(f, '__wrapped__')
 *     else:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__') and not stop(f)
 *     f = func  # remember the original func for error reporting
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.unwrap._is_wrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_54unwrap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_stop) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *__pyx_cur_scope;
  PyObject *__pyx_v__is_wrapper = 0;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_memo = NULL;
  PyObject *__pyx_v_recursion_limit = NULL;
  PyObject *__pyx_v_id_func = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("unwrap", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 529, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_stop = __pyx_v_stop;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stop);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_stop);
  __Pyx_INCREF(__pyx_v_func);

  /* "Aspidites/_vendor/contracts/inspection.py":545
 * 
 *     """
 *     if stop is None:             # <<<<<<<<<<<<<<
 *         def _is_wrapper(f):
 *             return hasattr(f, '__wrapped__')
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_stop == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":546
 *     """
 *     if stop is None:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__')
 *     else:
 */
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_1_is_wrapper, 0, __pyx_n_s_unwrap_locals__is_wrapper, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v__is_wrapper = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":545
 * 
 *     """
 *     if stop is None:             # <<<<<<<<<<<<<<
 *         def _is_wrapper(f):
 *             return hasattr(f, '__wrapped__')
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":549
 *             return hasattr(f, '__wrapped__')
 *     else:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__') and not stop(f)
 *     f = func  # remember the original func for error reporting
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_6unwrap_3_is_wrapper, 0, __pyx_n_s_unwrap_locals__is_wrapper, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__9)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v__is_wrapper = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":551
 *         def _is_wrapper(f):
 *             return hasattr(f, '__wrapped__') and not stop(f)
 *     f = func  # remember the original func for error reporting             # <<<<<<<<<<<<<<
 *     # Memoise by id to tolerate non-hashable objects, but store objects to
 *     # ensure they aren't destroyed, which would allow their IDs to be reused.
 */
  __Pyx_INCREF(__pyx_v_func);
  __pyx_v_f = __pyx_v_func;

  /* "Aspidites/_vendor/contracts/inspection.py":554
 *     # Memoise by id to tolerate non-hashable objects, but store objects to
 *     # ensure they aren't destroyed, which would allow their IDs to be reused.
 *     memo = {id(f): f}             # <<<<<<<<<<<<<<
 *     recursion_limit = sys.getrecursionlimit()
 *     while _is_wrapper(func):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_4, __pyx_v_f) < 0) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_memo = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":555
 *     # ensure they aren't destroyed, which would allow their IDs to be reused.
 *     memo = {id(f): f}
 *     recursion_limit = sys.getrecursionlimit()             # <<<<<<<<<<<<<<
 *     while _is_wrapper(func):
 *         func = func.__wrapped__
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getrecursionlimit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_recursion_limit = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":556
 *     memo = {id(f): f}
 *     recursion_limit = sys.getrecursionlimit()
 *     while _is_wrapper(func):             # <<<<<<<<<<<<<<
 *         func = func.__wrapped__
 *         id_func = id(func)
 */
  while (1) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_v__is_wrapper, __pyx_v_func); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_2) break;

    /* "Aspidites/_vendor/contracts/inspection.py":557
 *     recursion_limit = sys.getrecursionlimit()
 *     while _is_wrapper(func):
 *         func = func.__wrapped__             # <<<<<<<<<<<<<<
 *         id_func = id(func)
 *         if (id_func in memo) or (len(memo) >= recursion_limit):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_wrapped); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_func, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":558
 *     while _is_wrapper(func):
 *         func = func.__wrapped__
 *         id_func = id(func)             # <<<<<<<<<<<<<<
 *         if (id_func in memo) or (len(memo) >= recursion_limit):
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_func); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_id_func, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":559
 *         func = func.__wrapped__
 *         id_func = id(func)
 *         if (id_func in memo) or (len(memo) >= recursion_limit):             # <<<<<<<<<<<<<<
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))
 *         memo[id_func] = func
 */
    __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_id_func, __pyx_v_memo, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_1 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = PyDict_Size(__pyx_v_memo); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 559, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_v_recursion_limit, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "Aspidites/_vendor/contracts/inspection.py":560
 *         id_func = id(func)
 *         if (id_func in memo) or (len(memo) >= recursion_limit):
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))             # <<<<<<<<<<<<<<
 *         memo[id_func] = func
 *     return func
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_wrapper_loop_when_unwrapping_r, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_f);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 560, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":559
 *         func = func.__wrapped__
 *         id_func = id(func)
 *         if (id_func in memo) or (len(memo) >= recursion_limit):             # <<<<<<<<<<<<<<
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))
 *         memo[id_func] = func
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":561
 *         if (id_func in memo) or (len(memo) >= recursion_limit):
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))
 *         memo[id_func] = func             # <<<<<<<<<<<<<<
 *     return func
 * 
 */
    if (unlikely(PyDict_SetItem(__pyx_v_memo, __pyx_v_id_func, __pyx_v_func) < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/contracts/inspection.py":562
 *             raise ValueError('wrapper loop when unwrapping {!r}'.format(f))
 *         memo[id_func] = func
 *     return func             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_func);
  __pyx_r = __pyx_v_func;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.unwrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__is_wrapper);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_memo);
  __Pyx_XDECREF(__pyx_v_recursion_limit);
  __Pyx_XDECREF(__pyx_v_id_func);
  __Pyx_XDECREF(__pyx_v_func);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":566
 * 
 * # -------------------------------------------------- source code extraction
 * def indentsize(line):             # <<<<<<<<<<<<<<
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_57indentsize(PyObject *__pyx_self, PyObject *__pyx_v_line); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_56indentsize[] = "Return the indent size, in spaces, at the start of a line of text.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_57indentsize = {"indentsize", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_57indentsize, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_56indentsize};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_57indentsize(PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("indentsize (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_56indentsize(__pyx_self, ((PyObject *)__pyx_v_line));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_56indentsize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_v_expline = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("indentsize", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":568
 * def indentsize(line):
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()             # <<<<<<<<<<<<<<
 *     return len(expline) - len(expline.lstrip())
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_expline = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":569
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()
 *     return len(expline) - len(expline.lstrip())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyObject_Length(__pyx_v_expline); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 569, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_expline, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_t_4 - __pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":566
 * 
 * # -------------------------------------------------- source code extraction
 * def indentsize(line):             # <<<<<<<<<<<<<<
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.indentsize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expline);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":572
 * 
 * 
 * def _findclass(func):             # <<<<<<<<<<<<<<
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_59_findclass(PyObject *__pyx_self, PyObject *__pyx_v_func); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_59_findclass = {"_findclass", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_59_findclass, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_59_findclass(PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_findclass (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_58_findclass(__pyx_self, ((PyObject *)__pyx_v_func));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_58_findclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_v_cls = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_findclass", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":573
 * 
 * def _findclass(func):
 *     cls = sys.modules.get(func.__module__)             # <<<<<<<<<<<<<<
 *     if cls is None:
 *         return None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_modules); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cls = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":574
 * def _findclass(func):
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:             # <<<<<<<<<<<<<<
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:
 */
  __pyx_t_5 = (__pyx_v_cls == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/contracts/inspection.py":575
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:
 *         return None             # <<<<<<<<<<<<<<
 *     for name in func.__qualname__.split('.')[:-1]:
 *         cls = getattr(cls, name)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":574
 * def _findclass(func):
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:             # <<<<<<<<<<<<<<
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":576
 *     if cls is None:
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:             # <<<<<<<<<<<<<<
 *         cls = getattr(cls, name)
 *     if not isclass(cls):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_qualname); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u__10) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u__10);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 576, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 576, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":577
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:
 *         cls = getattr(cls, name)             # <<<<<<<<<<<<<<
 *     if not isclass(cls):
 *         return None
 */
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_cls, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":576
 *     if cls is None:
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:             # <<<<<<<<<<<<<<
 *         cls = getattr(cls, name)
 *     if not isclass(cls):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":578
 *     for name in func.__qualname__.split('.')[:-1]:
 *         cls = getattr(cls, name)
 *     if not isclass(cls):             # <<<<<<<<<<<<<<
 *         return None
 *     return cls
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isclass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_cls) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_cls);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_6) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":579
 *         cls = getattr(cls, name)
 *     if not isclass(cls):
 *         return None             # <<<<<<<<<<<<<<
 *     return cls
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":578
 *     for name in func.__qualname__.split('.')[:-1]:
 *         cls = getattr(cls, name)
 *     if not isclass(cls):             # <<<<<<<<<<<<<<
 *         return None
 *     return cls
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":580
 *     if not isclass(cls):
 *         return None
 *     return cls             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cls);
  __pyx_r = __pyx_v_cls;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":572
 * 
 * 
 * def _findclass(func):             # <<<<<<<<<<<<<<
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._findclass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cls);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":583
 * 
 * 
 * def _finddoc(obj):             # <<<<<<<<<<<<<<
 *     if isclass(obj):
 *         for base in obj.__mro__:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_61_finddoc(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_61_finddoc = {"_finddoc", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_61_finddoc, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_61_finddoc(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_finddoc (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_60_finddoc(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_60_finddoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_v_doc = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_self = NULL;
  PyObject *__pyx_v_cls = NULL;
  PyObject *__pyx_v_func = NULL;
  PyObject *__pyx_v_slots = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_finddoc", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":584
 * 
 * def _finddoc(obj):
 *     if isclass(obj):             # <<<<<<<<<<<<<<
 *         for base in obj.__mro__:
 *             if base is not object:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":585
 * def _finddoc(obj):
 *     if isclass(obj):
 *         for base in obj.__mro__:             # <<<<<<<<<<<<<<
 *             if base is not object:
 *                 try:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_mro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 585, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 585, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":586
 *     if isclass(obj):
 *         for base in obj.__mro__:
 *             if base is not object:             # <<<<<<<<<<<<<<
 *                 try:
 *                     doc = base.__doc__
 */
      __pyx_t_4 = (__pyx_v_base != __pyx_builtin_object);
      __pyx_t_7 = (__pyx_t_4 != 0);
      if (__pyx_t_7) {

        /* "Aspidites/_vendor/contracts/inspection.py":587
 *         for base in obj.__mro__:
 *             if base is not object:
 *                 try:             # <<<<<<<<<<<<<<
 *                     doc = base.__doc__
 *                 except AttributeError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":588
 *             if base is not object:
 *                 try:
 *                     doc = base.__doc__             # <<<<<<<<<<<<<<
 *                 except AttributeError:
 *                     continue
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_doc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_XDECREF_SET(__pyx_v_doc, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":587
 *         for base in obj.__mro__:
 *             if base is not object:
 *                 try:             # <<<<<<<<<<<<<<
 *                     doc = base.__doc__
 *                 except AttributeError:
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L14_try_end;
          __pyx_L7_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":589
 *                 try:
 *                     doc = base.__doc__
 *                 except AttributeError:             # <<<<<<<<<<<<<<
 *                     continue
 *                 if doc is not None:
 */
          __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
          if (__pyx_t_11) {
            __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._finddoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_12) < 0) __PYX_ERR(0, 589, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_12);

            /* "Aspidites/_vendor/contracts/inspection.py":590
 *                     doc = base.__doc__
 *                 except AttributeError:
 *                     continue             # <<<<<<<<<<<<<<
 *                 if doc is not None:
 *                     return doc
 */
            goto __pyx_L16_except_continue;
            __pyx_L16_except_continue:;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            goto __pyx_L13_try_continue;
          }
          goto __pyx_L9_except_error;
          __pyx_L9_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":587
 *         for base in obj.__mro__:
 *             if base is not object:
 *                 try:             # <<<<<<<<<<<<<<
 *                     doc = base.__doc__
 *                 except AttributeError:
 */
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L13_try_continue:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L4_continue;
          __pyx_L14_try_end:;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":591
 *                 except AttributeError:
 *                     continue
 *                 if doc is not None:             # <<<<<<<<<<<<<<
 *                     return doc
 *         return None
 */
        __pyx_t_7 = (__pyx_v_doc != Py_None);
        __pyx_t_4 = (__pyx_t_7 != 0);
        if (__pyx_t_4) {

          /* "Aspidites/_vendor/contracts/inspection.py":592
 *                     continue
 *                 if doc is not None:
 *                     return doc             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_doc);
          __pyx_r = __pyx_v_doc;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L0;

          /* "Aspidites/_vendor/contracts/inspection.py":591
 *                 except AttributeError:
 *                     continue
 *                 if doc is not None:             # <<<<<<<<<<<<<<
 *                     return doc
 *         return None
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":586
 *     if isclass(obj):
 *         for base in obj.__mro__:
 *             if base is not object:             # <<<<<<<<<<<<<<
 *                 try:
 *                     doc = base.__doc__
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":585
 * def _finddoc(obj):
 *     if isclass(obj):
 *         for base in obj.__mro__:             # <<<<<<<<<<<<<<
 *             if base is not object:
 *                 try:
 */
      __pyx_L4_continue:;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":593
 *                 if doc is not None:
 *                     return doc
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     if ismethod(obj):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":584
 * 
 * def _finddoc(obj):
 *     if isclass(obj):             # <<<<<<<<<<<<<<
 *         for base in obj.__mro__:
 *             if base is not object:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":595
 *         return None
 * 
 *     if ismethod(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__func__.__name__
 *         self = obj.__self__
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_12);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_12, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":596
 * 
 *     if ismethod(obj):
 *         name = obj.__func__.__name__             # <<<<<<<<<<<<<<
 *         self = obj.__self__
 *         if (isclass(self) and
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_func); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_name = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":597
 *     if ismethod(obj):
 *         name = obj.__func__.__name__
 *         self = obj.__self__             # <<<<<<<<<<<<<<
 *         if (isclass(self) and
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_v_self = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":598
 *         name = obj.__func__.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):
 *             # classmethod
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_4 = __pyx_t_7;
      goto __pyx_L20_bool_binop_done;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":599
 *         self = obj.__self__
 *         if (isclass(self) and
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):             # <<<<<<<<<<<<<<
 *             # classmethod
 *             cls = self
 */
    __pyx_t_12 = __Pyx_GetAttr3(__pyx_v_self, __pyx_v_name, Py_None); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = __Pyx_GetAttr(__pyx_t_12, __pyx_n_u_func); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_func); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_7 = (__pyx_t_2 == __pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_13 = (__pyx_t_7 != 0);
    __pyx_t_4 = __pyx_t_13;
    __pyx_L20_bool_binop_done:;

    /* "Aspidites/_vendor/contracts/inspection.py":598
 *         name = obj.__func__.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):
 *             # classmethod
 */
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":601
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):
 *             # classmethod
 *             cls = self             # <<<<<<<<<<<<<<
 *         else:
 *             cls = self.__class__
 */
      __Pyx_INCREF(__pyx_v_self);
      __pyx_v_cls = __pyx_v_self;

      /* "Aspidites/_vendor/contracts/inspection.py":598
 *         name = obj.__func__.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 getattr(getattr(self, name, None), '__func__') is obj.__func__):
 *             # classmethod
 */
      goto __pyx_L19;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":603
 *             cls = self
 *         else:
 *             cls = self.__class__             # <<<<<<<<<<<<<<
 *     elif isfunction(obj):
 *         name = obj.__name__
 */
    /*else*/ {
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_cls = __pyx_t_12;
      __pyx_t_12 = 0;
    }
    __pyx_L19:;

    /* "Aspidites/_vendor/contracts/inspection.py":595
 *         return None
 * 
 *     if ismethod(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__func__.__name__
 *         self = obj.__self__
 */
    goto __pyx_L18;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":604
 *         else:
 *             cls = self.__class__
 *     elif isfunction(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         cls = _findclass(obj)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":605
 *             cls = self.__class__
 *     elif isfunction(obj):
 *         name = obj.__name__             # <<<<<<<<<<<<<<
 *         cls = _findclass(obj)
 *         if cls is None or getattr(cls, name) is not obj:
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_v_name = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":606
 *     elif isfunction(obj):
 *         name = obj.__name__
 *         cls = _findclass(obj)             # <<<<<<<<<<<<<<
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_findclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_cls = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":607
 *         name = obj.__name__
 *         cls = _findclass(obj)
 *         if cls is None or getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *     elif isbuiltin(obj):
 */
    __pyx_t_13 = (__pyx_v_cls == Py_None);
    __pyx_t_7 = (__pyx_t_13 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_4 = __pyx_t_7;
      goto __pyx_L23_bool_binop_done;
    }
    __pyx_t_12 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_7 = (__pyx_t_12 != __pyx_v_obj);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_13 = (__pyx_t_7 != 0);
    __pyx_t_4 = __pyx_t_13;
    __pyx_L23_bool_binop_done:;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":608
 *         cls = _findclass(obj)
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None             # <<<<<<<<<<<<<<
 *     elif isbuiltin(obj):
 *         name = obj.__name__
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":607
 *         name = obj.__name__
 *         cls = _findclass(obj)
 *         if cls is None or getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *     elif isbuiltin(obj):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":604
 *         else:
 *             cls = self.__class__
 *     elif isfunction(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         cls = _findclass(obj)
 */
    goto __pyx_L18;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":609
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 *     elif isbuiltin(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         self = obj.__self__
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isbuiltin); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":610
 *             return None
 *     elif isbuiltin(obj):
 *         name = obj.__name__             # <<<<<<<<<<<<<<
 *         self = obj.__self__
 *         if (isclass(self) and
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_v_name = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":611
 *     elif isbuiltin(obj):
 *         name = obj.__name__
 *         self = obj.__self__             # <<<<<<<<<<<<<<
 *         if (isclass(self) and
 *                 self.__qualname__ + '.' + name == obj.__qualname__):
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_v_self = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":612
 *         name = obj.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 self.__qualname__ + '.' + name == obj.__qualname__):
 *             # classmethod
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_12 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (__pyx_t_13) {
    } else {
      __pyx_t_4 = __pyx_t_13;
      goto __pyx_L26_bool_binop_done;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":613
 *         self = obj.__self__
 *         if (isclass(self) and
 *                 self.__qualname__ + '.' + name == obj.__qualname__):             # <<<<<<<<<<<<<<
 *             # classmethod
 *             cls = self
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_qualname); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = PyNumber_Add(__pyx_t_12, __pyx_kp_u__10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyNumber_Add(__pyx_t_2, __pyx_v_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_qualname); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_12, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_t_13;
    __pyx_L26_bool_binop_done:;

    /* "Aspidites/_vendor/contracts/inspection.py":612
 *         name = obj.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 self.__qualname__ + '.' + name == obj.__qualname__):
 *             # classmethod
 */
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":615
 *                 self.__qualname__ + '.' + name == obj.__qualname__):
 *             # classmethod
 *             cls = self             # <<<<<<<<<<<<<<
 *         else:
 *             cls = self.__class__
 */
      __Pyx_INCREF(__pyx_v_self);
      __pyx_v_cls = __pyx_v_self;

      /* "Aspidites/_vendor/contracts/inspection.py":612
 *         name = obj.__name__
 *         self = obj.__self__
 *         if (isclass(self) and             # <<<<<<<<<<<<<<
 *                 self.__qualname__ + '.' + name == obj.__qualname__):
 *             # classmethod
 */
      goto __pyx_L25;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":617
 *             cls = self
 *         else:
 *             cls = self.__class__             # <<<<<<<<<<<<<<
 *     # Should be tested before isdatadescriptor().
 *     elif isinstance(obj, property):
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_cls = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L25:;

    /* "Aspidites/_vendor/contracts/inspection.py":609
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 *     elif isbuiltin(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         self = obj.__self__
 */
    goto __pyx_L18;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":619
 *             cls = self.__class__
 *     # Should be tested before isdatadescriptor().
 *     elif isinstance(obj, property):             # <<<<<<<<<<<<<<
 *         func = obj.fget
 *         name = func.__name__
 */
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_obj, __pyx_builtin_property); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 619, __pyx_L1_error)
  __pyx_t_13 = (__pyx_t_4 != 0);
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":620
 *     # Should be tested before isdatadescriptor().
 *     elif isinstance(obj, property):
 *         func = obj.fget             # <<<<<<<<<<<<<<
 *         name = func.__name__
 *         cls = _findclass(func)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_fget); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_func = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":621
 *     elif isinstance(obj, property):
 *         func = obj.fget
 *         name = func.__name__             # <<<<<<<<<<<<<<
 *         cls = _findclass(func)
 *         if cls is None or getattr(cls, name) is not obj:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":622
 *         func = obj.fget
 *         name = func.__name__
 *         cls = _findclass(func)             # <<<<<<<<<<<<<<
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_findclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_cls = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":623
 *         name = func.__name__
 *         cls = _findclass(func)
 *         if cls is None or getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):
 */
    __pyx_t_4 = (__pyx_v_cls == Py_None);
    __pyx_t_7 = (__pyx_t_4 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_13 = __pyx_t_7;
      goto __pyx_L29_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = (__pyx_t_3 != __pyx_v_obj);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_7 != 0);
    __pyx_t_13 = __pyx_t_4;
    __pyx_L29_bool_binop_done:;
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":624
 *         cls = _findclass(func)
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None             # <<<<<<<<<<<<<<
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):
 *         name = obj.__name__
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":623
 *         name = func.__name__
 *         cls = _findclass(func)
 *         if cls is None or getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":619
 *             cls = self.__class__
 *     # Should be tested before isdatadescriptor().
 *     elif isinstance(obj, property):             # <<<<<<<<<<<<<<
 *         func = obj.fget
 *         name = func.__name__
 */
    goto __pyx_L18;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":625
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         cls = obj.__objclass__
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethoddescriptor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_13 = __pyx_t_4;
    goto __pyx_L31_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isdatadescriptor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __pyx_t_4;
  __pyx_L31_bool_binop_done:;
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":626
 *             return None
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):
 *         name = obj.__name__             # <<<<<<<<<<<<<<
 *         cls = obj.__objclass__
 *         if getattr(cls, name) is not obj:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":627
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):
 *         name = obj.__name__
 *         cls = obj.__objclass__             # <<<<<<<<<<<<<<
 *         if getattr(cls, name) is not obj:
 *             return None
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_objclass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_cls = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":628
 *         name = obj.__name__
 *         cls = obj.__objclass__
 *         if getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *         if ismemberdescriptor(obj):
 */
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = (__pyx_t_3 != __pyx_v_obj);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_13 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":629
 *         cls = obj.__objclass__
 *         if getattr(cls, name) is not obj:
 *             return None             # <<<<<<<<<<<<<<
 *         if ismemberdescriptor(obj):
 *             slots = getattr(cls, '__slots__', None)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":628
 *         name = obj.__name__
 *         cls = obj.__objclass__
 *         if getattr(cls, name) is not obj:             # <<<<<<<<<<<<<<
 *             return None
 *         if ismemberdescriptor(obj):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":630
 *         if getattr(cls, name) is not obj:
 *             return None
 *         if ismemberdescriptor(obj):             # <<<<<<<<<<<<<<
 *             slots = getattr(cls, '__slots__', None)
 *             if isinstance(slots, dict) and name in slots:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismemberdescriptor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":631
 *             return None
 *         if ismemberdescriptor(obj):
 *             slots = getattr(cls, '__slots__', None)             # <<<<<<<<<<<<<<
 *             if isinstance(slots, dict) and name in slots:
 *                 return slots[name]
 */
      __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_cls, __pyx_n_u_slots, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_slots = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":632
 *         if ismemberdescriptor(obj):
 *             slots = getattr(cls, '__slots__', None)
 *             if isinstance(slots, dict) and name in slots:             # <<<<<<<<<<<<<<
 *                 return slots[name]
 *     else:
 */
      __pyx_t_13 = PyDict_Check(__pyx_v_slots); 
      __pyx_t_7 = (__pyx_t_13 != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_4 = __pyx_t_7;
        goto __pyx_L36_bool_binop_done;
      }
      __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_slots, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 632, __pyx_L1_error)
      __pyx_t_13 = (__pyx_t_7 != 0);
      __pyx_t_4 = __pyx_t_13;
      __pyx_L36_bool_binop_done:;
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/contracts/inspection.py":633
 *             slots = getattr(cls, '__slots__', None)
 *             if isinstance(slots, dict) and name in slots:
 *                 return slots[name]             # <<<<<<<<<<<<<<
 *     else:
 *         return None
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_slots, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 633, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_r = __pyx_t_3;
        __pyx_t_3 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/contracts/inspection.py":632
 *         if ismemberdescriptor(obj):
 *             slots = getattr(cls, '__slots__', None)
 *             if isinstance(slots, dict) and name in slots:             # <<<<<<<<<<<<<<
 *                 return slots[name]
 *     else:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":630
 *         if getattr(cls, name) is not obj:
 *             return None
 *         if ismemberdescriptor(obj):             # <<<<<<<<<<<<<<
 *             slots = getattr(cls, '__slots__', None)
 *             if isinstance(slots, dict) and name in slots:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":625
 *         if cls is None or getattr(cls, name) is not obj:
 *             return None
 *     elif ismethoddescriptor(obj) or isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *         name = obj.__name__
 *         cls = obj.__objclass__
 */
    goto __pyx_L18;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":635
 *                 return slots[name]
 *     else:
 *         return None             # <<<<<<<<<<<<<<
 *     for base in cls.__mro__:
 *         try:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }
  __pyx_L18:;

  /* "Aspidites/_vendor/contracts/inspection.py":636
 *     else:
 *         return None
 *     for base in cls.__mro__:             # <<<<<<<<<<<<<<
 *         try:
 *             doc = getattr(base, name).__doc__
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_mro); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 636, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 636, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 636, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 636, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":637
 *         return None
 *     for base in cls.__mro__:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = getattr(base, name).__doc__
 *         except AttributeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":638
 *     for base in cls.__mro__:
 *         try:
 *             doc = getattr(base, name).__doc__             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             continue
 */
        __pyx_t_3 = __Pyx_GetAttr(__pyx_v_base, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_doc); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 638, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_doc, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":637
 *         return None
 *     for base in cls.__mro__:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = getattr(base, name).__doc__
 *         except AttributeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L47_try_end;
      __pyx_L40_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":639
 *         try:
 *             doc = getattr(base, name).__doc__
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             continue
 *         if doc is not None:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._finddoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 639, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Aspidites/_vendor/contracts/inspection.py":640
 *             doc = getattr(base, name).__doc__
 *         except AttributeError:
 *             continue             # <<<<<<<<<<<<<<
 *         if doc is not None:
 *             return doc
 */
        goto __pyx_L49_except_continue;
        __pyx_L49_except_continue:;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L46_try_continue;
      }
      goto __pyx_L42_except_error;
      __pyx_L42_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":637
 *         return None
 *     for base in cls.__mro__:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = getattr(base, name).__doc__
 *         except AttributeError:
 */
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L46_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      goto __pyx_L38_continue;
      __pyx_L47_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":641
 *         except AttributeError:
 *             continue
 *         if doc is not None:             # <<<<<<<<<<<<<<
 *             return doc
 *     return None
 */
    __pyx_t_4 = (__pyx_v_doc != Py_None);
    __pyx_t_13 = (__pyx_t_4 != 0);
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":642
 *             continue
 *         if doc is not None:
 *             return doc             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_doc);
      __pyx_r = __pyx_v_doc;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":641
 *         except AttributeError:
 *             continue
 *         if doc is not None:             # <<<<<<<<<<<<<<
 *             return doc
 *     return None
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":636
 *     else:
 *         return None
 *     for base in cls.__mro__:             # <<<<<<<<<<<<<<
 *         try:
 *             doc = getattr(base, name).__doc__
 */
    __pyx_L38_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":643
 *         if doc is not None:
 *             return doc
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":583
 * 
 * 
 * def _finddoc(obj):             # <<<<<<<<<<<<<<
 *     if isclass(obj):
 *         for base in obj.__mro__:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._finddoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_doc);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_cls);
  __Pyx_XDECREF(__pyx_v_func);
  __Pyx_XDECREF(__pyx_v_slots);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":646
 * 
 * 
 * def getdoc(object):             # <<<<<<<<<<<<<<
 *     """Get the documentation string for an object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_63getdoc(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_62getdoc[] = "Get the documentation string for an object.\n\n    All tabs are expanded to spaces.  To clean up docstrings that are\n    indented to line up with blocks of code, any whitespace than can be\n    uniformly removed from the second line onwards is removed.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_63getdoc = {"getdoc", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_63getdoc, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_62getdoc};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_63getdoc(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getdoc (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_62getdoc(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_62getdoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_doc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getdoc", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":652
 *     indented to line up with blocks of code, any whitespace than can be
 *     uniformly removed from the second line onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         doc = object.__doc__
 *     except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":653
 *     uniformly removed from the second line onwards is removed."""
 *     try:
 *         doc = object.__doc__             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         return None
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_doc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_doc = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":652
 *     indented to line up with blocks of code, any whitespace than can be
 *     uniformly removed from the second line onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         doc = object.__doc__
 *     except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":654
 *     try:
 *         doc = object.__doc__
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         return None
 *     if doc is None:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getdoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 654, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/contracts/inspection.py":655
 *         doc = object.__doc__
 *     except AttributeError:
 *         return None             # <<<<<<<<<<<<<<
 *     if doc is None:
 *         try:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":652
 *     indented to line up with blocks of code, any whitespace than can be
 *     uniformly removed from the second line onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         doc = object.__doc__
 *     except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":656
 *     except AttributeError:
 *         return None
 *     if doc is None:             # <<<<<<<<<<<<<<
 *         try:
 *             doc = _finddoc(object)
 */
  __pyx_t_8 = (__pyx_v_doc == Py_None);
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/contracts/inspection.py":657
 *         return None
 *     if doc is None:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = _finddoc(object)
 *         except (AttributeError, TypeError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":658
 *     if doc is None:
 *         try:
 *             doc = _finddoc(object)             # <<<<<<<<<<<<<<
 *         except (AttributeError, TypeError):
 *             return None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_finddoc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 658, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_object);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 658, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_doc, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":657
 *         return None
 *     if doc is None:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = _finddoc(object)
 *         except (AttributeError, TypeError):
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L17_try_end;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":659
 *         try:
 *             doc = _finddoc(object)
 *         except (AttributeError, TypeError):             # <<<<<<<<<<<<<<
 *             return None
 *     if not isinstance(doc, str):
 */
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_5) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getdoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 659, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_4);

        /* "Aspidites/_vendor/contracts/inspection.py":660
 *             doc = _finddoc(object)
 *         except (AttributeError, TypeError):
 *             return None             # <<<<<<<<<<<<<<
 *     if not isinstance(doc, str):
 *         return None
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L15_except_return;
      }
      goto __pyx_L14_except_error;
      __pyx_L14_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":657
 *         return None
 *     if doc is None:
 *         try:             # <<<<<<<<<<<<<<
 *             doc = _finddoc(object)
 *         except (AttributeError, TypeError):
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L1_error;
      __pyx_L15_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L0;
      __pyx_L17_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":656
 *     except AttributeError:
 *         return None
 *     if doc is None:             # <<<<<<<<<<<<<<
 *         try:
 *             doc = _finddoc(object)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":661
 *         except (AttributeError, TypeError):
 *             return None
 *     if not isinstance(doc, str):             # <<<<<<<<<<<<<<
 *         return None
 *     return cleandoc(doc)
 */
  __pyx_t_9 = PyUnicode_Check(__pyx_v_doc); 
  __pyx_t_8 = ((!(__pyx_t_9 != 0)) != 0);
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/contracts/inspection.py":662
 *             return None
 *     if not isinstance(doc, str):
 *         return None             # <<<<<<<<<<<<<<
 *     return cleandoc(doc)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":661
 *         except (AttributeError, TypeError):
 *             return None
 *     if not isinstance(doc, str):             # <<<<<<<<<<<<<<
 *         return None
 *     return cleandoc(doc)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":663
 *     if not isinstance(doc, str):
 *         return None
 *     return cleandoc(doc)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cleandoc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v_doc) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_doc);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":646
 * 
 * 
 * def getdoc(object):             # <<<<<<<<<<<<<<
 *     """Get the documentation string for an object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getdoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_doc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":666
 * 
 * 
 * def cleandoc(doc):             # <<<<<<<<<<<<<<
 *     """Clean up indentation from docstrings.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_65cleandoc(PyObject *__pyx_self, PyObject *__pyx_v_doc); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_64cleandoc[] = "Clean up indentation from docstrings.\n\n    Any whitespace that can be uniformly removed from the second line\n    onwards is removed.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_65cleandoc = {"cleandoc", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_65cleandoc, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_64cleandoc};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_65cleandoc(PyObject *__pyx_self, PyObject *__pyx_v_doc) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cleandoc (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_64cleandoc(__pyx_self, ((PyObject *)__pyx_v_doc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_64cleandoc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_doc) {
  PyObject *__pyx_v_lines = NULL;
  PyObject *__pyx_v_margin = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_content = NULL;
  PyObject *__pyx_v_indent = NULL;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cleandoc", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":671
 *     Any whitespace that can be uniformly removed from the second line
 *     onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         lines = doc.expandtabs().split('\n')
 *     except UnicodeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":672
 *     onwards is removed."""
 *     try:
 *         lines = doc.expandtabs().split('\n')             # <<<<<<<<<<<<<<
 *     except UnicodeError:
 *         return None
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_doc, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 672, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 672, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_split); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 672, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_kp_u__12) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u__12);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 672, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_lines = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":671
 *     Any whitespace that can be uniformly removed from the second line
 *     onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         lines = doc.expandtabs().split('\n')
 *     except UnicodeError:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":677
 *     else:
 *         # Find minimum indentation of any non-blank lines after first line.
 *         margin = sys.maxsize             # <<<<<<<<<<<<<<
 *         for line in lines[1:]:
 *             content = len(line.lstrip())
 */
    /*else:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 677, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 677, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_margin = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":678
 *         # Find minimum indentation of any non-blank lines after first line.
 *         margin = sys.maxsize
 *         for line in lines[1:]:             # <<<<<<<<<<<<<<
 *             content = len(line.lstrip())
 *             if content:
 */
      __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_lines, 1, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 678, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_4 = __pyx_t_6; __Pyx_INCREF(__pyx_t_4); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 678, __pyx_L5_except_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L5_except_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 678, __pyx_L5_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L5_except_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 678, __pyx_L5_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_9(__pyx_t_4);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 678, __pyx_L5_except_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":679
 *         margin = sys.maxsize
 *         for line in lines[1:]:
 *             content = len(line.lstrip())             # <<<<<<<<<<<<<<
 *             if content:
 *                 indent = len(line) - content
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 679, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 679, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 679, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 679, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_content, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":680
 *         for line in lines[1:]:
 *             content = len(line.lstrip())
 *             if content:             # <<<<<<<<<<<<<<
 *                 indent = len(line) - content
 *                 margin = min(margin, indent)
 */
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_content); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 680, __pyx_L5_except_error)
        if (__pyx_t_11) {

          /* "Aspidites/_vendor/contracts/inspection.py":681
 *             content = len(line.lstrip())
 *             if content:
 *                 indent = len(line) - content             # <<<<<<<<<<<<<<
 *                 margin = min(margin, indent)
 *         # Remove indentation.
 */
          __pyx_t_10 = PyObject_Length(__pyx_v_line); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 681, __pyx_L5_except_error)
          __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = PyNumber_Subtract(__pyx_t_6, __pyx_v_content); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 681, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_indent, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":682
 *             if content:
 *                 indent = len(line) - content
 *                 margin = min(margin, indent)             # <<<<<<<<<<<<<<
 *         # Remove indentation.
 *         if lines:
 */
          __Pyx_INCREF(__pyx_v_indent);
          __pyx_t_5 = __pyx_v_indent;
          __Pyx_INCREF(__pyx_v_margin);
          __pyx_t_6 = __pyx_v_margin;
          __pyx_t_12 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 682, __pyx_L5_except_error)
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 682, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_11) {
            __Pyx_INCREF(__pyx_t_5);
            __pyx_t_7 = __pyx_t_5;
          } else {
            __Pyx_INCREF(__pyx_t_6);
            __pyx_t_7 = __pyx_t_6;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __pyx_t_7;
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_margin, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":680
 *         for line in lines[1:]:
 *             content = len(line.lstrip())
 *             if content:             # <<<<<<<<<<<<<<
 *                 indent = len(line) - content
 *                 margin = min(margin, indent)
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":678
 *         # Find minimum indentation of any non-blank lines after first line.
 *         margin = sys.maxsize
 *         for line in lines[1:]:             # <<<<<<<<<<<<<<
 *             content = len(line.lstrip())
 *             if content:
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":684
 *                 margin = min(margin, indent)
 *         # Remove indentation.
 *         if lines:             # <<<<<<<<<<<<<<
 *             lines[0] = lines[0].lstrip()
 *         if margin < sys.maxsize:
 */
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_lines); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 684, __pyx_L5_except_error)
      if (__pyx_t_11) {

        /* "Aspidites/_vendor/contracts/inspection.py":685
 *         # Remove indentation.
 *         if lines:
 *             lines[0] = lines[0].lstrip()             # <<<<<<<<<<<<<<
 *         if margin < sys.maxsize:
 *             for i in range(1, len(lines)): lines[i] = lines[i][margin:]
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_lines, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 685, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 685, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_v_lines, 0, __pyx_t_4, long, 1, __Pyx_PyInt_From_long, 0, 0, 0) < 0)) __PYX_ERR(0, 685, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":684
 *                 margin = min(margin, indent)
 *         # Remove indentation.
 *         if lines:             # <<<<<<<<<<<<<<
 *             lines[0] = lines[0].lstrip()
 *         if margin < sys.maxsize:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":686
 *         if lines:
 *             lines[0] = lines[0].lstrip()
 *         if margin < sys.maxsize:             # <<<<<<<<<<<<<<
 *             for i in range(1, len(lines)): lines[i] = lines[i][margin:]
 *         # Remove any trailing or leading blank lines.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 686, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_margin, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 686, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_11) {

        /* "Aspidites/_vendor/contracts/inspection.py":687
 *             lines[0] = lines[0].lstrip()
 *         if margin < sys.maxsize:
 *             for i in range(1, len(lines)): lines[i] = lines[i][margin:]             # <<<<<<<<<<<<<<
 *         # Remove any trailing or leading blank lines.
 *         while lines and not lines[-1]:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_lines); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 687, __pyx_L5_except_error)
        __pyx_t_10 = __pyx_t_8;
        for (__pyx_t_13 = 1; __pyx_t_13 < __pyx_t_10; __pyx_t_13+=1) {
          __pyx_v_i = __pyx_t_13;
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_lines, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 687, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 0, &__pyx_v_margin, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 687, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(__Pyx_SetItemInt(__pyx_v_lines, __pyx_v_i, __pyx_t_7, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0) < 0)) __PYX_ERR(0, 687, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":686
 *         if lines:
 *             lines[0] = lines[0].lstrip()
 *         if margin < sys.maxsize:             # <<<<<<<<<<<<<<
 *             for i in range(1, len(lines)): lines[i] = lines[i][margin:]
 *         # Remove any trailing or leading blank lines.
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":689
 *             for i in range(1, len(lines)): lines[i] = lines[i][margin:]
 *         # Remove any trailing or leading blank lines.
 *         while lines and not lines[-1]:             # <<<<<<<<<<<<<<
 *             lines.pop()
 *         while lines and not lines[0]:
 */
      while (1) {
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_lines); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 689, __pyx_L5_except_error)
        if (__pyx_t_14) {
        } else {
          __pyx_t_11 = __pyx_t_14;
          goto __pyx_L18_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_lines, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 689, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 689, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_15 = ((!__pyx_t_14) != 0);
        __pyx_t_11 = __pyx_t_15;
        __pyx_L18_bool_binop_done:;
        if (!__pyx_t_11) break;

        /* "Aspidites/_vendor/contracts/inspection.py":690
 *         # Remove any trailing or leading blank lines.
 *         while lines and not lines[-1]:
 *             lines.pop()             # <<<<<<<<<<<<<<
 *         while lines and not lines[0]:
 *             lines.pop(0)
 */
        __pyx_t_7 = __Pyx_PyObject_Pop(__pyx_v_lines); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 690, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":691
 *         while lines and not lines[-1]:
 *             lines.pop()
 *         while lines and not lines[0]:             # <<<<<<<<<<<<<<
 *             lines.pop(0)
 *         return '\n'.join(lines)
 */
      while (1) {
        __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_lines); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 691, __pyx_L5_except_error)
        if (__pyx_t_15) {
        } else {
          __pyx_t_11 = __pyx_t_15;
          goto __pyx_L22_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_lines, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 691, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 691, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_14 = ((!__pyx_t_15) != 0);
        __pyx_t_11 = __pyx_t_14;
        __pyx_L22_bool_binop_done:;
        if (!__pyx_t_11) break;

        /* "Aspidites/_vendor/contracts/inspection.py":692
 *             lines.pop()
 *         while lines and not lines[0]:
 *             lines.pop(0)             # <<<<<<<<<<<<<<
 *         return '\n'.join(lines)
 * 
 */
        __pyx_t_7 = __Pyx_PyObject_PopIndex(__pyx_v_lines, __pyx_int_0, 0, 1, Py_ssize_t, PyInt_FromSsize_t); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 692, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":693
 *         while lines and not lines[0]:
 *             lines.pop(0)
 *         return '\n'.join(lines)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = PyUnicode_Join(__pyx_kp_u__12, __pyx_v_lines); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 693, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":673
 *     try:
 *         lines = doc.expandtabs().split('\n')
 *     except UnicodeError:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
    __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeError);
    if (__pyx_t_16) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.cleandoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 673, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);

      /* "Aspidites/_vendor/contracts/inspection.py":674
 *         lines = doc.expandtabs().split('\n')
 *     except UnicodeError:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         # Find minimum indentation of any non-blank lines after first line.
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":671
 *     Any whitespace that can be uniformly removed from the second line
 *     onwards is removed."""
 *     try:             # <<<<<<<<<<<<<<
 *         lines = doc.expandtabs().split('\n')
 *     except UnicodeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":666
 * 
 * 
 * def cleandoc(doc):             # <<<<<<<<<<<<<<
 *     """Clean up indentation from docstrings.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.cleandoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_margin);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_content);
  __Pyx_XDECREF(__pyx_v_indent);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":696
 * 
 * 
 * def getfile(object):             # <<<<<<<<<<<<<<
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_67getfile(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_66getfile[] = "Work out which source or compiled file an object was defined in.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_67getfile = {"getfile", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_67getfile, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_66getfile};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_67getfile(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getfile (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_66getfile(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_66getfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_module = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getfile", 0);
  __Pyx_INCREF(__pyx_v_object);

  /* "Aspidites/_vendor/contracts/inspection.py":698
 * def getfile(object):
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         if getattr(object, '__file__', None):
 *             return object.__file__
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":699
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 *         if getattr(object, '__file__', None):             # <<<<<<<<<<<<<<
 *             return object.__file__
 *         raise TypeError('{!r} is a built-in module'.format(object))
 */
    __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_object, __pyx_n_u_file, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":700
 *     if ismodule(object):
 *         if getattr(object, '__file__', None):
 *             return object.__file__             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is a built-in module'.format(object))
 *     if isclass(object):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":699
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 *         if getattr(object, '__file__', None):             # <<<<<<<<<<<<<<
 *             return object.__file__
 *         raise TypeError('{!r} is a built-in module'.format(object))
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":701
 *         if getattr(object, '__file__', None):
 *             return object.__file__
 *         raise TypeError('{!r} is a built-in module'.format(object))             # <<<<<<<<<<<<<<
 *     if isclass(object):
 *         if hasattr(object, '__module__'):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_a_built_in_module, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 701, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":698
 * def getfile(object):
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         if getattr(object, '__file__', None):
 *             return object.__file__
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":702
 *             return object.__file__
 *         raise TypeError('{!r} is a built-in module'.format(object))
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         if hasattr(object, '__module__'):
 *             module = sys.modules.get(object.__module__)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_isclass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":703
 *         raise TypeError('{!r} is a built-in module'.format(object))
 *     if isclass(object):
 *         if hasattr(object, '__module__'):             # <<<<<<<<<<<<<<
 *             module = sys.modules.get(object.__module__)
 *             if getattr(module, '__file__', None):
 */
    __pyx_t_4 = __Pyx_HasAttr(__pyx_v_object, __pyx_n_u_module); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 703, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":704
 *     if isclass(object):
 *         if hasattr(object, '__module__'):
 *             module = sys.modules.get(object.__module__)             # <<<<<<<<<<<<<<
 *             if getattr(module, '__file__', None):
 *                 return module.__file__
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_modules); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_module = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":705
 *         if hasattr(object, '__module__'):
 *             module = sys.modules.get(object.__module__)
 *             if getattr(module, '__file__', None):             # <<<<<<<<<<<<<<
 *                 return module.__file__
 *         raise TypeError('{!r} is a built-in class'.format(object))
 */
      __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_module, __pyx_n_u_file, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/contracts/inspection.py":706
 *             module = sys.modules.get(object.__module__)
 *             if getattr(module, '__file__', None):
 *                 return module.__file__             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is a built-in class'.format(object))
 *     if ismethod(object):
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/contracts/inspection.py":705
 *         if hasattr(object, '__module__'):
 *             module = sys.modules.get(object.__module__)
 *             if getattr(module, '__file__', None):             # <<<<<<<<<<<<<<
 *                 return module.__file__
 *         raise TypeError('{!r} is a built-in class'.format(object))
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":703
 *         raise TypeError('{!r} is a built-in module'.format(object))
 *     if isclass(object):
 *         if hasattr(object, '__module__'):             # <<<<<<<<<<<<<<
 *             module = sys.modules.get(object.__module__)
 *             if getattr(module, '__file__', None):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":707
 *             if getattr(module, '__file__', None):
 *                 return module.__file__
 *         raise TypeError('{!r} is a built-in class'.format(object))             # <<<<<<<<<<<<<<
 *     if ismethod(object):
 *         object = object.__func__
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_a_built_in_class, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 707, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":702
 *             return object.__file__
 *         raise TypeError('{!r} is a built-in module'.format(object))
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         if hasattr(object, '__module__'):
 *             module = sys.modules.get(object.__module__)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":708
 *                 return module.__file__
 *         raise TypeError('{!r} is a built-in class'.format(object))
 *     if ismethod(object):             # <<<<<<<<<<<<<<
 *         object = object.__func__
 *     if isfunction(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":709
 *         raise TypeError('{!r} is a built-in class'.format(object))
 *     if ismethod(object):
 *         object = object.__func__             # <<<<<<<<<<<<<<
 *     if isfunction(object):
 *         object = object.__code__
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":708
 *                 return module.__file__
 *         raise TypeError('{!r} is a built-in class'.format(object))
 *     if ismethod(object):             # <<<<<<<<<<<<<<
 *         object = object.__func__
 *     if isfunction(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":710
 *     if ismethod(object):
 *         object = object.__func__
 *     if isfunction(object):             # <<<<<<<<<<<<<<
 *         object = object.__code__
 *     if istraceback(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":711
 *         object = object.__func__
 *     if isfunction(object):
 *         object = object.__code__             # <<<<<<<<<<<<<<
 *     if istraceback(object):
 *         object = object.tb_frame
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":710
 *     if ismethod(object):
 *         object = object.__func__
 *     if isfunction(object):             # <<<<<<<<<<<<<<
 *         object = object.__code__
 *     if istraceback(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":712
 *     if isfunction(object):
 *         object = object.__code__
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 *     if isframe(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_istraceback); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":713
 *         object = object.__code__
 *     if istraceback(object):
 *         object = object.tb_frame             # <<<<<<<<<<<<<<
 *     if isframe(object):
 *         object = object.f_code
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 713, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":712
 *     if isfunction(object):
 *         object = object.__code__
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 *     if isframe(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":714
 *     if istraceback(object):
 *         object = object.tb_frame
 *     if isframe(object):             # <<<<<<<<<<<<<<
 *         object = object.f_code
 *     if iscode(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isframe); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":715
 *         object = object.tb_frame
 *     if isframe(object):
 *         object = object.f_code             # <<<<<<<<<<<<<<
 *     if iscode(object):
 *         return object.co_filename
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_f_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":714
 *     if istraceback(object):
 *         object = object.tb_frame
 *     if isframe(object):             # <<<<<<<<<<<<<<
 *         object = object.f_code
 *     if iscode(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":716
 *     if isframe(object):
 *         object = object.f_code
 *     if iscode(object):             # <<<<<<<<<<<<<<
 *         return object.co_filename
 *     raise TypeError('module, class, method, function, traceback, frame, or '
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iscode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":717
 *         object = object.f_code
 *     if iscode(object):
 *         return object.co_filename             # <<<<<<<<<<<<<<
 *     raise TypeError('module, class, method, function, traceback, frame, or '
 *                     'code object was expected, got {}'.format(
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_co_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":716
 *     if isframe(object):
 *         object = object.f_code
 *     if iscode(object):             # <<<<<<<<<<<<<<
 *         return object.co_filename
 *     raise TypeError('module, class, method, function, traceback, frame, or '
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":719
 *         return object.co_filename
 *     raise TypeError('module, class, method, function, traceback, frame, or '
 *                     'code object was expected, got {}'.format(             # <<<<<<<<<<<<<<
 *         type(object).__name__))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_module_class_method_function_tra, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/contracts/inspection.py":720
 *     raise TypeError('module, class, method, function, traceback, frame, or '
 *                     'code object was expected, got {}'.format(
 *         type(object).__name__))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_object)), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":718
 *     if iscode(object):
 *         return object.co_filename
 *     raise TypeError('module, class, method, function, traceback, frame, or '             # <<<<<<<<<<<<<<
 *                     'code object was expected, got {}'.format(
 *         type(object).__name__))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 718, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":696
 * 
 * 
 * def getfile(object):             # <<<<<<<<<<<<<<
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_module);
  __Pyx_XDECREF(__pyx_v_object);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":723
 * 
 * 
 * def getmodulename(path):             # <<<<<<<<<<<<<<
 *     """Return the module name for a given file, or None."""
 *     fname = os.path.basename(path)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_69getmodulename(PyObject *__pyx_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_68getmodulename[] = "Return the module name for a given file, or None.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_69getmodulename = {"getmodulename", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_69getmodulename, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_68getmodulename};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_69getmodulename(PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getmodulename (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_68getmodulename(__pyx_self, ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_68getmodulename(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_v_fname = NULL;
  PyObject *__pyx_v_suffixes = NULL;
  PyObject *__pyx_v_neglen = NULL;
  PyObject *__pyx_v_suffix = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_suffix = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getmodulename", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":725
 * def getmodulename(path):
 *     """Return the module name for a given file, or None."""
 *     fname = os.path.basename(path)             # <<<<<<<<<<<<<<
 *     # Check for paths that look like an actual module file
 *     suffixes = [(-len(suffix), suffix)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_basename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_path);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fname = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":727
 *     fname = os.path.basename(path)
 *     # Check for paths that look like an actual module file
 *     suffixes = [(-len(suffix), suffix)             # <<<<<<<<<<<<<<
 *                 for suffix in importlib.machinery.all_suffixes()]
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Aspidites/_vendor/contracts/inspection.py":728
 *     # Check for paths that look like an actual module file
 *     suffixes = [(-len(suffix), suffix)
 *                 for suffix in importlib.machinery.all_suffixes()]             # <<<<<<<<<<<<<<
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_importlib); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_machinery); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_all_suffixes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 728, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 728, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 728, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 728, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_6(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 728, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_suffix, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":727
 *     fname = os.path.basename(path)
 *     # Check for paths that look like an actual module file
 *     suffixes = [(-len(suffix), suffix)             # <<<<<<<<<<<<<<
 *                 for suffix in importlib.machinery.all_suffixes()]
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 */
      __pyx_t_7 = PyObject_Length(__pyx_8genexpr1__pyx_v_suffix); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 727, __pyx_L5_error)
      __pyx_t_2 = PyInt_FromSsize_t((-__pyx_t_7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_INCREF(__pyx_8genexpr1__pyx_v_suffix);
      __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_suffix);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_8genexpr1__pyx_v_suffix);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 727, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":728
 *     # Check for paths that look like an actual module file
 *     suffixes = [(-len(suffix), suffix)
 *                 for suffix in importlib.machinery.all_suffixes()]             # <<<<<<<<<<<<<<
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_suffix); __pyx_8genexpr1__pyx_v_suffix = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_suffix); __pyx_8genexpr1__pyx_v_suffix = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_suffixes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":729
 *     suffixes = [(-len(suffix), suffix)
 *                 for suffix in importlib.machinery.all_suffixes()]
 *     suffixes.sort()  # try longest suffixes first, in case they overlap             # <<<<<<<<<<<<<<
 *     for neglen, suffix in suffixes:
 *         if fname.endswith(suffix):
 */
  __pyx_t_8 = PyList_Sort(__pyx_v_suffixes); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":730
 *                 for suffix in importlib.machinery.all_suffixes()]
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:             # <<<<<<<<<<<<<<
 *         if fname.endswith(suffix):
 *             return fname[:neglen]
 */
  __pyx_t_1 = __pyx_v_suffixes; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 730, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 730, __pyx_L1_error)
      __pyx_L12_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_neglen, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_suffix, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":731
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:
 *         if fname.endswith(suffix):             # <<<<<<<<<<<<<<
 *             return fname[:neglen]
 *     return None
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fname, __pyx_n_s_endswith); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_suffix) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_suffix);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":732
 *     for neglen, suffix in suffixes:
 *         if fname.endswith(suffix):
 *             return fname[:neglen]             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_fname, 0, 0, NULL, &__pyx_v_neglen, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":731
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:
 *         if fname.endswith(suffix):             # <<<<<<<<<<<<<<
 *             return fname[:neglen]
 *     return None
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":730
 *                 for suffix in importlib.machinery.all_suffixes()]
 *     suffixes.sort()  # try longest suffixes first, in case they overlap
 *     for neglen, suffix in suffixes:             # <<<<<<<<<<<<<<
 *         if fname.endswith(suffix):
 *             return fname[:neglen]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":733
 *         if fname.endswith(suffix):
 *             return fname[:neglen]
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":723
 * 
 * 
 * def getmodulename(path):             # <<<<<<<<<<<<<<
 *     """Return the module name for a given file, or None."""
 *     fname = os.path.basename(path)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmodulename", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fname);
  __Pyx_XDECREF(__pyx_v_suffixes);
  __Pyx_XDECREF(__pyx_v_neglen);
  __Pyx_XDECREF(__pyx_v_suffix);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_suffix);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_71getsourcefile(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_70getsourcefile[] = "Return the filename that can be used to locate an object's source.\n    Return None if no way can be identified to get the source.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_71getsourcefile = {"getsourcefile", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_71getsourcefile, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_70getsourcefile};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_71getsourcefile(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsourcefile (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_70getsourcefile(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/contracts/inspection.py":743
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):             # <<<<<<<<<<<<<<
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 743, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_getsourcefile_locals_genexpr, __pyx_n_s_Aspidites__vendor_contracts_insp); if (unlikely(!gen)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getsourcefile.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 743, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_all_bytecode_suffixes)) { __Pyx_RaiseClosureNameError("all_bytecode_suffixes"); __PYX_ERR(0, 743, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_all_bytecode_suffixes)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_all_bytecode_suffixes)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_all_bytecode_suffixes; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_all_bytecode_suffixes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 743, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 743, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 743, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 743, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_filename)) { __Pyx_RaiseClosureNameError("filename"); __PYX_ERR(0, 743, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_filename, __pyx_n_s_endswith); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_cur_scope->__pyx_v_s) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_7) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/contracts/inspection.py":746
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in             # <<<<<<<<<<<<<<
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 746, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_5generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_getsourcefile_locals_genexpr, __pyx_n_s_Aspidites__vendor_contracts_insp); if (unlikely(!gen)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getsourcefile.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 746, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":747
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in
 *              importlib.machinery.EXTENSION_SUFFIXES):             # <<<<<<<<<<<<<<
 *         return None
 *     if os.path.exists(filename):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_importlib); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_machinery); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EXTENSION_SUFFIXES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 747, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 747, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 747, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 747, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":746
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in             # <<<<<<<<<<<<<<
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_filename)) { __Pyx_RaiseClosureNameError("filename"); __PYX_ERR(0, 746, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_filename, __pyx_n_s_endswith); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_cur_scope->__pyx_v_s) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_70getsourcefile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_2generator1 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_5generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsourcefile", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 736, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_object = __pyx_v_object;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_object);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_object);

  /* "Aspidites/_vendor/contracts/inspection.py":740
 *     Return None if no way can be identified to get the source.
 *     """
 *     filename = getfile(object)             # <<<<<<<<<<<<<<
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_cur_scope->__pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_filename = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":741
 *     """
 *     filename = getfile(object)
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]             # <<<<<<<<<<<<<<
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_importlib); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_machinery); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DEBUG_BYTECODE_SUFFIXES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__14, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_all_bytecode_suffixes = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":742
 *     filename = getfile(object)
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]             # <<<<<<<<<<<<<<
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 *         filename = (os.path.splitext(filename)[0] +
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_importlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_machinery); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_OPTIMIZED_BYTECODE_SUFFIXES); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__14, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_cur_scope->__pyx_v_all_bytecode_suffixes, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_all_bytecode_suffixes);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_all_bytecode_suffixes, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":743
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):             # <<<<<<<<<<<<<<
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 */
  __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":744
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 *         filename = (os.path.splitext(filename)[0] +             # <<<<<<<<<<<<<<
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_splitext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_cur_scope->__pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_v_filename);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":745
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])             # <<<<<<<<<<<<<<
 *     elif any(filename.endswith(s) for s in
 *              importlib.machinery.EXTENSION_SUFFIXES):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_importlib); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_machinery); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SOURCE_SUFFIXES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":744
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 *         filename = (os.path.splitext(filename)[0] +             # <<<<<<<<<<<<<<
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in
 */
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_filename);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_filename, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":743
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):             # <<<<<<<<<<<<<<
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":746
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in             # <<<<<<<<<<<<<<
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":748
 *     elif any(filename.endswith(s) for s in
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None             # <<<<<<<<<<<<<<
 *     if os.path.exists(filename):
 *         return filename
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":746
 *         filename = (os.path.splitext(filename)[0] +
 *                     importlib.machinery.SOURCE_SUFFIXES[0])
 *     elif any(filename.endswith(s) for s in             # <<<<<<<<<<<<<<
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 */
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":749
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 *     if os.path.exists(filename):             # <<<<<<<<<<<<<<
 *         return filename
 *     # only return a non-existent filename if the module has a PEP 302 loader
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_exists); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_cur_scope->__pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_cur_scope->__pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":750
 *         return None
 *     if os.path.exists(filename):
 *         return filename             # <<<<<<<<<<<<<<
 *     # only return a non-existent filename if the module has a PEP 302 loader
 *     if getattr(getmodule(object, filename), '__loader__', None) is not None:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_filename);
    __pyx_r = __pyx_cur_scope->__pyx_v_filename;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":749
 *              importlib.machinery.EXTENSION_SUFFIXES):
 *         return None
 *     if os.path.exists(filename):             # <<<<<<<<<<<<<<
 *         return filename
 *     # only return a non-existent filename if the module has a PEP 302 loader
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":752
 *         return filename
 *     # only return a non-existent filename if the module has a PEP 302 loader
 *     if getattr(getmodule(object, filename), '__loader__', None) is not None:             # <<<<<<<<<<<<<<
 *         return filename
 *     # or it is in the linecache
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getmodule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_cur_scope->__pyx_v_object, __pyx_cur_scope->__pyx_v_filename};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_cur_scope->__pyx_v_object, __pyx_cur_scope->__pyx_v_filename};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_object);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_object);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_cur_scope->__pyx_v_object);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_filename);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_filename);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_cur_scope->__pyx_v_filename);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_loader, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_4 != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":753
 *     # only return a non-existent filename if the module has a PEP 302 loader
 *     if getattr(getmodule(object, filename), '__loader__', None) is not None:
 *         return filename             # <<<<<<<<<<<<<<
 *     # or it is in the linecache
 *     if filename in linecache.cache:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_filename);
    __pyx_r = __pyx_cur_scope->__pyx_v_filename;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":752
 *         return filename
 *     # only return a non-existent filename if the module has a PEP 302 loader
 *     if getattr(getmodule(object, filename), '__loader__', None) is not None:             # <<<<<<<<<<<<<<
 *         return filename
 *     # or it is in the linecache
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":755
 *         return filename
 *     # or it is in the linecache
 *     if filename in linecache.cache:             # <<<<<<<<<<<<<<
 *         return filename
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_linecache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_filename, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_7 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":756
 *     # or it is in the linecache
 *     if filename in linecache.cache:
 *         return filename             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_filename);
    __pyx_r = __pyx_cur_scope->__pyx_v_filename;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":755
 *         return filename
 *     # or it is in the linecache
 *     if filename in linecache.cache:             # <<<<<<<<<<<<<<
 *         return filename
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getsourcefile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_2generator1);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_13getsourcefile_5generator2);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":759
 * 
 * 
 * def getabsfile(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return an absolute path to the source or compiled file for an object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_73getabsfile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_72getabsfile[] = "Return an absolute path to the source or compiled file for an object.\n\n    The idea is for each object to have a unique origin, so this routine\n    normalizes the result as much as possible.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_73getabsfile = {"getabsfile", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_73getabsfile, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_72getabsfile};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_73getabsfile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_object = 0;
  PyObject *__pyx_v__filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getabsfile (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_object,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_object)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getabsfile") < 0)) __PYX_ERR(0, 759, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_object = values[0];
    __pyx_v__filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getabsfile", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 759, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getabsfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_72getabsfile(__pyx_self, __pyx_v_object, __pyx_v__filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_72getabsfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v__filename) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getabsfile", 0);
  __Pyx_INCREF(__pyx_v__filename);

  /* "Aspidites/_vendor/contracts/inspection.py":764
 *     The idea is for each object to have a unique origin, so this routine
 *     normalizes the result as much as possible."""
 *     if _filename is None:             # <<<<<<<<<<<<<<
 *         _filename = getsourcefile(object) or getfile(object)
 *     return os.path.normcase(os.path.abspath(_filename))
 */
  __pyx_t_1 = (__pyx_v__filename == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":765
 *     normalizes the result as much as possible."""
 *     if _filename is None:
 *         _filename = getsourcefile(object) or getfile(object)             # <<<<<<<<<<<<<<
 *     return os.path.normcase(os.path.abspath(_filename))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_getsourcefile); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 765, __pyx_L1_error)
    if (!__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_3 = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_getfile); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L4_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v__filename, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":764
 *     The idea is for each object to have a unique origin, so this routine
 *     normalizes the result as much as possible."""
 *     if _filename is None:             # <<<<<<<<<<<<<<
 *         _filename = getsourcefile(object) or getfile(object)
 *     return os.path.normcase(os.path.abspath(_filename))
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":766
 *     if _filename is None:
 *         _filename = getsourcefile(object) or getfile(object)
 *     return os.path.normcase(os.path.abspath(_filename))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_os); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_normcase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_os); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_path); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_abspath); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v__filename) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v__filename);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":759
 * 
 * 
 * def getabsfile(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return an absolute path to the source or compiled file for an object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getabsfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__filename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":773
 * 
 * 
 * def getmodule(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_75getmodule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_74getmodule[] = "Return the module an object was defined in, or None if not found.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_75getmodule = {"getmodule", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_75getmodule, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_74getmodule};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_75getmodule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_object = 0;
  PyObject *__pyx_v__filename = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getmodule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_object,&__pyx_n_s_filename,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_object)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getmodule") < 0)) __PYX_ERR(0, 773, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_object = values[0];
    __pyx_v__filename = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getmodule", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 773, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmodule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_74getmodule(__pyx_self, __pyx_v_object, __pyx_v__filename);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_74getmodule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object, PyObject *__pyx_v__filename) {
  PyObject *__pyx_v_file = NULL;
  PyObject *__pyx_v_modname = NULL;
  PyObject *__pyx_v_module = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_main = NULL;
  PyObject *__pyx_v_mainobject = NULL;
  PyObject *__pyx_v_builtin = NULL;
  PyObject *__pyx_v_builtinobject = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getmodule", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":775
 * def getmodule(object, _filename=None):
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         return object
 *     if hasattr(object, '__module__'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":776
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):
 *         return object             # <<<<<<<<<<<<<<
 *     if hasattr(object, '__module__'):
 *         return sys.modules.get(object.__module__)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_object);
    __pyx_r = __pyx_v_object;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":775
 * def getmodule(object, _filename=None):
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         return object
 *     if hasattr(object, '__module__'):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":777
 *     if ismodule(object):
 *         return object
 *     if hasattr(object, '__module__'):             # <<<<<<<<<<<<<<
 *         return sys.modules.get(object.__module__)
 *     # Try the filename to modulename cache
 */
  __pyx_t_4 = __Pyx_HasAttr(__pyx_v_object, __pyx_n_u_module); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 777, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":778
 *         return object
 *     if hasattr(object, '__module__'):
 *         return sys.modules.get(object.__module__)             # <<<<<<<<<<<<<<
 *     # Try the filename to modulename cache
 *     if _filename is not None and _filename in modulesbyfile:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_modules); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":777
 *     if ismodule(object):
 *         return object
 *     if hasattr(object, '__module__'):             # <<<<<<<<<<<<<<
 *         return sys.modules.get(object.__module__)
 *     # Try the filename to modulename cache
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":780
 *         return sys.modules.get(object.__module__)
 *     # Try the filename to modulename cache
 *     if _filename is not None and _filename in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[_filename])
 *     # Try the cache again with the absolute file name
 */
  __pyx_t_4 = (__pyx_v__filename != Py_None);
  __pyx_t_7 = (__pyx_t_4 != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L6_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v__filename, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_7 != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":781
 *     # Try the filename to modulename cache
 *     if _filename is not None and _filename in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[_filename])             # <<<<<<<<<<<<<<
 *     # Try the cache again with the absolute file name
 *     try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_modules); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v__filename); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":780
 *         return sys.modules.get(object.__module__)
 *     # Try the filename to modulename cache
 *     if _filename is not None and _filename in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[_filename])
 *     # Try the cache again with the absolute file name
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":783
 *         return sys.modules.get(modulesbyfile[_filename])
 *     # Try the cache again with the absolute file name
 *     try:             # <<<<<<<<<<<<<<
 *         file = getabsfile(object, _filename)
 *     except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":784
 *     # Try the cache again with the absolute file name
 *     try:
 *         file = getabsfile(object, _filename)             # <<<<<<<<<<<<<<
 *     except TypeError:
 *         return None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getabsfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 784, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_object, __pyx_v__filename};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L8_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_object, __pyx_v__filename};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L8_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 784, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_object);
        __Pyx_GIVEREF(__pyx_v_object);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_11, __pyx_v_object);
        __Pyx_INCREF(__pyx_v__filename);
        __Pyx_GIVEREF(__pyx_v__filename);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_11, __pyx_v__filename);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_file = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":783
 *         return sys.modules.get(modulesbyfile[_filename])
 *     # Try the cache again with the absolute file name
 *     try:             # <<<<<<<<<<<<<<
 *         file = getabsfile(object, _filename)
 *     except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":785
 *     try:
 *         file = getabsfile(object, _filename)
 *     except TypeError:             # <<<<<<<<<<<<<<
 *         return None
 *     if file in modulesbyfile:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmodule", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 785, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "Aspidites/_vendor/contracts/inspection.py":786
 *         file = getabsfile(object, _filename)
 *     except TypeError:
 *         return None             # <<<<<<<<<<<<<<
 *     if file in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[file])
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L11_except_return;
    }
    goto __pyx_L10_except_error;
    __pyx_L10_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":783
 *         return sys.modules.get(modulesbyfile[_filename])
 *     # Try the cache again with the absolute file name
 *     try:             # <<<<<<<<<<<<<<
 *         file = getabsfile(object, _filename)
 *     except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
    goto __pyx_L1_error;
    __pyx_L11_except_return:;
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
    goto __pyx_L0;
    __pyx_L13_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":787
 *     except TypeError:
 *         return None
 *     if file in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[file])
 *     # Update the filename to module name cache and check yet again
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_file, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":788
 *         return None
 *     if file in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[file])             # <<<<<<<<<<<<<<
 *     # Update the filename to module name cache and check yet again
 *     # Copy sys.modules in order to cope with changes while iterating
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_modules); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_file); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":787
 *     except TypeError:
 *         return None
 *     if file in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[file])
 *     # Update the filename to module name cache and check yet again
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":791
 *     # Update the filename to module name cache and check yet again
 *     # Copy sys.modules in order to cope with changes while iterating
 *     for modname, module in sys.modules.copy().items():             # <<<<<<<<<<<<<<
 *         if ismodule(module) and hasattr(module, '__file__'):
 *             f = module.__file__
 */
  __pyx_t_12 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_modules); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_copy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 791, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_2, 0, __pyx_n_s_items, (&__pyx_t_13), (&__pyx_t_11)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XDECREF(__pyx_t_3);
  __pyx_t_3 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_14 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_13, &__pyx_t_12, &__pyx_t_6, &__pyx_t_2, NULL, __pyx_t_11);
    if (unlikely(__pyx_t_14 == 0)) break;
    if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_modname, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_module, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":792
 *     # Copy sys.modules in order to cope with changes while iterating
 *     for modname, module in sys.modules.copy().items():
 *         if ismodule(module) and hasattr(module, '__file__'):             # <<<<<<<<<<<<<<
 *             f = module.__file__
 *             if f == _filesbymodname.get(modname, None):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_v_module) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_module);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_HasAttr(__pyx_v_module, __pyx_n_u_file); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 792, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_5 != 0);
    __pyx_t_4 = __pyx_t_7;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":793
 *     for modname, module in sys.modules.copy().items():
 *         if ismodule(module) and hasattr(module, '__file__'):
 *             f = module.__file__             # <<<<<<<<<<<<<<
 *             if f == _filesbymodname.get(modname, None):
 *                 # Have already mapped this module, so skip it
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":794
 *         if ismodule(module) and hasattr(module, '__file__'):
 *             f = module.__file__
 *             if f == _filesbymodname.get(modname, None):             # <<<<<<<<<<<<<<
 *                 # Have already mapped this module, so skip it
 *                 continue
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_filesbymodname); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_modname, Py_None};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_modname, Py_None};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_15 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_modname);
        __Pyx_GIVEREF(__pyx_v_modname);
        PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_v_modname);
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, Py_None);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_f, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/contracts/inspection.py":796
 *             if f == _filesbymodname.get(modname, None):
 *                 # Have already mapped this module, so skip it
 *                 continue             # <<<<<<<<<<<<<<
 *             _filesbymodname[modname] = f
 *             f = getabsfile(module)
 */
        goto __pyx_L17_continue;

        /* "Aspidites/_vendor/contracts/inspection.py":794
 *         if ismodule(module) and hasattr(module, '__file__'):
 *             f = module.__file__
 *             if f == _filesbymodname.get(modname, None):             # <<<<<<<<<<<<<<
 *                 # Have already mapped this module, so skip it
 *                 continue
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":797
 *                 # Have already mapped this module, so skip it
 *                 continue
 *             _filesbymodname[modname] = f             # <<<<<<<<<<<<<<
 *             f = getabsfile(module)
 *             # Always map to the name the module knows itself by
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_filesbymodname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_modname, __pyx_v_f) < 0)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":798
 *                 continue
 *             _filesbymodname[modname] = f
 *             f = getabsfile(module)             # <<<<<<<<<<<<<<
 *             # Always map to the name the module knows itself by
 *             modulesbyfile[f] = modulesbyfile[
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getabsfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_15 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_15, __pyx_v_module) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_module);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_f, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":801
 *             # Always map to the name the module knows itself by
 *             modulesbyfile[f] = modulesbyfile[
 *                 os.path.realpath(f)] = module.__name__             # <<<<<<<<<<<<<<
 *     if file in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[file])
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "Aspidites/_vendor/contracts/inspection.py":800
 *             f = getabsfile(module)
 *             # Always map to the name the module knows itself by
 *             modulesbyfile[f] = modulesbyfile[             # <<<<<<<<<<<<<<
 *                 os.path.realpath(f)] = module.__name__
 *     if file in modulesbyfile:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_v_f, __pyx_t_1) < 0)) __PYX_ERR(0, 800, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/contracts/inspection.py":801
 *             # Always map to the name the module knows itself by
 *             modulesbyfile[f] = modulesbyfile[
 *                 os.path.realpath(f)] = module.__name__             # <<<<<<<<<<<<<<
 *     if file in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[file])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_os); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_path); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_realpath); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_15 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_16, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_f);
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_15, __pyx_t_1) < 0)) __PYX_ERR(0, 800, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":792
 *     # Copy sys.modules in order to cope with changes while iterating
 *     for modname, module in sys.modules.copy().items():
 *         if ismodule(module) and hasattr(module, '__file__'):             # <<<<<<<<<<<<<<
 *             f = module.__file__
 *             if f == _filesbymodname.get(modname, None):
 */
    }
    __pyx_L17_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":802
 *             modulesbyfile[f] = modulesbyfile[
 *                 os.path.realpath(f)] = module.__name__
 *     if file in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[file])
 *     # Check the main module
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_file, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = (__pyx_t_4 != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":803
 *                 os.path.realpath(f)] = module.__name__
 *     if file in modulesbyfile:
 *         return sys.modules.get(modulesbyfile[file])             # <<<<<<<<<<<<<<
 *     # Check the main module
 *     main = sys.modules['__main__']
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_modules); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_modulesbyfile); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_15, __pyx_v_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_15, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":802
 *             modulesbyfile[f] = modulesbyfile[
 *                 os.path.realpath(f)] = module.__name__
 *     if file in modulesbyfile:             # <<<<<<<<<<<<<<
 *         return sys.modules.get(modulesbyfile[file])
 *     # Check the main module
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":805
 *         return sys.modules.get(modulesbyfile[file])
 *     # Check the main module
 *     main = sys.modules['__main__']             # <<<<<<<<<<<<<<
 *     if not hasattr(object, '__name__'):
 *         return None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_modules); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_main); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_main = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":806
 *     # Check the main module
 *     main = sys.modules['__main__']
 *     if not hasattr(object, '__name__'):             # <<<<<<<<<<<<<<
 *         return None
 *     if hasattr(main, object.__name__):
 */
  __pyx_t_7 = __Pyx_HasAttr(__pyx_v_object, __pyx_n_u_name); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_t_4 = ((!(__pyx_t_7 != 0)) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":807
 *     main = sys.modules['__main__']
 *     if not hasattr(object, '__name__'):
 *         return None             # <<<<<<<<<<<<<<
 *     if hasattr(main, object.__name__):
 *         mainobject = getattr(main, object.__name__)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":806
 *     # Check the main module
 *     main = sys.modules['__main__']
 *     if not hasattr(object, '__name__'):             # <<<<<<<<<<<<<<
 *         return None
 *     if hasattr(main, object.__name__):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":808
 *     if not hasattr(object, '__name__'):
 *         return None
 *     if hasattr(main, object.__name__):             # <<<<<<<<<<<<<<
 *         mainobject = getattr(main, object.__name__)
 *         if mainobject is object:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_HasAttr(__pyx_v_main, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = (__pyx_t_4 != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":809
 *         return None
 *     if hasattr(main, object.__name__):
 *         mainobject = getattr(main, object.__name__)             # <<<<<<<<<<<<<<
 *         if mainobject is object:
 *             return main
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetAttr(__pyx_v_main, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_mainobject = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":810
 *     if hasattr(main, object.__name__):
 *         mainobject = getattr(main, object.__name__)
 *         if mainobject is object:             # <<<<<<<<<<<<<<
 *             return main
 *     # Check builtins
 */
    __pyx_t_7 = (__pyx_v_mainobject == __pyx_v_object);
    __pyx_t_4 = (__pyx_t_7 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":811
 *         mainobject = getattr(main, object.__name__)
 *         if mainobject is object:
 *             return main             # <<<<<<<<<<<<<<
 *     # Check builtins
 *     builtin = sys.modules['builtins']
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_main);
      __pyx_r = __pyx_v_main;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":810
 *     if hasattr(main, object.__name__):
 *         mainobject = getattr(main, object.__name__)
 *         if mainobject is object:             # <<<<<<<<<<<<<<
 *             return main
 *     # Check builtins
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":808
 *     if not hasattr(object, '__name__'):
 *         return None
 *     if hasattr(main, object.__name__):             # <<<<<<<<<<<<<<
 *         mainobject = getattr(main, object.__name__)
 *         if mainobject is object:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":813
 *             return main
 *     # Check builtins
 *     builtin = sys.modules['builtins']             # <<<<<<<<<<<<<<
 *     if hasattr(builtin, object.__name__):
 *         builtinobject = getattr(builtin, object.__name__)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_modules); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_builtins); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_builtin = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":814
 *     # Check builtins
 *     builtin = sys.modules['builtins']
 *     if hasattr(builtin, object.__name__):             # <<<<<<<<<<<<<<
 *         builtinobject = getattr(builtin, object.__name__)
 *         if builtinobject is object:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_HasAttr(__pyx_v_builtin, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_4 != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":815
 *     builtin = sys.modules['builtins']
 *     if hasattr(builtin, object.__name__):
 *         builtinobject = getattr(builtin, object.__name__)             # <<<<<<<<<<<<<<
 *         if builtinobject is object:
 *             return builtin
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_builtin, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_builtinobject = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":816
 *     if hasattr(builtin, object.__name__):
 *         builtinobject = getattr(builtin, object.__name__)
 *         if builtinobject is object:             # <<<<<<<<<<<<<<
 *             return builtin
 * 
 */
    __pyx_t_7 = (__pyx_v_builtinobject == __pyx_v_object);
    __pyx_t_4 = (__pyx_t_7 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":817
 *         builtinobject = getattr(builtin, object.__name__)
 *         if builtinobject is object:
 *             return builtin             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_builtin);
      __pyx_r = __pyx_v_builtin;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":816
 *     if hasattr(builtin, object.__name__):
 *         builtinobject = getattr(builtin, object.__name__)
 *         if builtinobject is object:             # <<<<<<<<<<<<<<
 *             return builtin
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":814
 *     # Check builtins
 *     builtin = sys.modules['builtins']
 *     if hasattr(builtin, object.__name__):             # <<<<<<<<<<<<<<
 *         builtinobject = getattr(builtin, object.__name__)
 *         if builtinobject is object:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":773
 * 
 * 
 * def getmodule(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getmodule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file);
  __Pyx_XDECREF(__pyx_v_modname);
  __Pyx_XDECREF(__pyx_v_module);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_main);
  __Pyx_XDECREF(__pyx_v_mainobject);
  __Pyx_XDECREF(__pyx_v_builtin);
  __Pyx_XDECREF(__pyx_v_builtinobject);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":826
 * class _ClassFinder(ast.NodeVisitor):
 * 
 *     def __init__(self, qualname):             # <<<<<<<<<<<<<<
 *         self.stack = []
 *         self.qualname = qualname
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_qualname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_qualname_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qualname_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 826, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 826, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_qualname = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder___init__(__pyx_self, __pyx_v_self, __pyx_v_qualname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_qualname) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":827
 * 
 *     def __init__(self, qualname):
 *         self.stack = []             # <<<<<<<<<<<<<<
 *         self.qualname = qualname
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_stack, __pyx_t_1) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":828
 *     def __init__(self, qualname):
 *         self.stack = []
 *         self.qualname = qualname             # <<<<<<<<<<<<<<
 * 
 *     def visit_FunctionDef(self, node):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_qualname_2, __pyx_v_qualname) < 0) __PYX_ERR(0, 828, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":826
 * class _ClassFinder(ast.NodeVisitor):
 * 
 *     def __init__(self, qualname):             # <<<<<<<<<<<<<<
 *         self.stack = []
 *         self.qualname = qualname
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":830
 *         self.qualname = qualname
 * 
 *     def visit_FunctionDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_3visit_FunctionDef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_3visit_FunctionDef = {"visit_FunctionDef", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_3visit_FunctionDef, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_3visit_FunctionDef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_node = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("visit_FunctionDef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_node,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("visit_FunctionDef", 1, 2, 2, 1); __PYX_ERR(0, 830, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "visit_FunctionDef") < 0)) __PYX_ERR(0, 830, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_node = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("visit_FunctionDef", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 830, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.visit_FunctionDef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_2visit_FunctionDef(__pyx_self, __pyx_v_self, __pyx_v_node);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_2visit_FunctionDef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_FunctionDef", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":831
 * 
 *     def visit_FunctionDef(self, node):
 *         self.stack.append(node.name)             # <<<<<<<<<<<<<<
 *         self.stack.append('<locals>')
 *         self.generic_visit(node)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":832
 *     def visit_FunctionDef(self, node):
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')             # <<<<<<<<<<<<<<
 *         self.generic_visit(node)
 *         self.stack.pop()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_kp_u_locals); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":833
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')
 *         self.generic_visit(node)             # <<<<<<<<<<<<<<
 *         self.stack.pop()
 *         self.stack.pop()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_generic_visit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v_node) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_node);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":834
 *         self.stack.append('<locals>')
 *         self.generic_visit(node)
 *         self.stack.pop()             # <<<<<<<<<<<<<<
 *         self.stack.pop()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":835
 *         self.generic_visit(node)
 *         self.stack.pop()
 *         self.stack.pop()             # <<<<<<<<<<<<<<
 * 
 *     visit_AsyncFunctionDef = visit_FunctionDef
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Pop(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":830
 *         self.qualname = qualname
 * 
 *     def visit_FunctionDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.visit_FunctionDef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":839
 *     visit_AsyncFunctionDef = visit_FunctionDef
 * 
 *     def visit_ClassDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_5visit_ClassDef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_5visit_ClassDef = {"visit_ClassDef", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_5visit_ClassDef, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_5visit_ClassDef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_node = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("visit_ClassDef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_node,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("visit_ClassDef", 1, 2, 2, 1); __PYX_ERR(0, 839, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "visit_ClassDef") < 0)) __PYX_ERR(0, 839, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_node = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("visit_ClassDef", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 839, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.visit_ClassDef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_4visit_ClassDef(__pyx_self, __pyx_v_self, __pyx_v_node);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_4visit_ClassDef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_v_line_number = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_ClassDef", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":840
 * 
 *     def visit_ClassDef(self, node):
 *         self.stack.append(node.name)             # <<<<<<<<<<<<<<
 *         if self.qualname == '.'.join(self.stack):
 *             # Return the decorator for the class if present
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":841
 *     def visit_ClassDef(self, node):
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):             # <<<<<<<<<<<<<<
 *             # Return the decorator for the class if present
 *             if node.decorator_list:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_qualname_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyUnicode_Join(__pyx_kp_u__10, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":843
 *         if self.qualname == '.'.join(self.stack):
 *             # Return the decorator for the class if present
 *             if node.decorator_list:             # <<<<<<<<<<<<<<
 *                 line_number = node.decorator_list[0].lineno
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_decorator_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":844
 *             # Return the decorator for the class if present
 *             if node.decorator_list:
 *                 line_number = node.decorator_list[0].lineno             # <<<<<<<<<<<<<<
 *             else:
 *                 line_number = node.lineno
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_decorator_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_lineno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_line_number = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":843
 *         if self.qualname == '.'.join(self.stack):
 *             # Return the decorator for the class if present
 *             if node.decorator_list:             # <<<<<<<<<<<<<<
 *                 line_number = node.decorator_list[0].lineno
 *             else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":846
 *                 line_number = node.decorator_list[0].lineno
 *             else:
 *                 line_number = node.lineno             # <<<<<<<<<<<<<<
 * 
 *             # decrement by one since lines starts with indexing by zero
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_lineno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 846, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_line_number = __pyx_t_4;
      __pyx_t_4 = 0;
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/contracts/inspection.py":849
 * 
 *             # decrement by one since lines starts with indexing by zero
 *             line_number -= 1             # <<<<<<<<<<<<<<
 *             raise ClassFoundException(line_number)
 *         self.generic_visit(node)
 */
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_line_number, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_line_number, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":850
 *             # decrement by one since lines starts with indexing by zero
 *             line_number -= 1
 *             raise ClassFoundException(line_number)             # <<<<<<<<<<<<<<
 *         self.generic_visit(node)
 *         self.stack.pop()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ClassFoundException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_line_number) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_line_number);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 850, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":841
 *     def visit_ClassDef(self, node):
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):             # <<<<<<<<<<<<<<
 *             # Return the decorator for the class if present
 *             if node.decorator_list:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":851
 *             line_number -= 1
 *             raise ClassFoundException(line_number)
 *         self.generic_visit(node)             # <<<<<<<<<<<<<<
 *         self.stack.pop()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_generic_visit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_node) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_node);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":852
 *             raise ClassFoundException(line_number)
 *         self.generic_visit(node)
 *         self.stack.pop()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_Pop(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":839
 *     visit_AsyncFunctionDef = visit_FunctionDef
 * 
 *     def visit_ClassDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ClassFinder.visit_ClassDef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line_number);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":855
 * 
 * 
 * def findsource(object):             # <<<<<<<<<<<<<<
 *     """Return the entire source file and starting line number for an object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_77findsource(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_76findsource[] = "Return the entire source file and starting line number for an object.\n\n    The argument may be a module, class, method, function, traceback, frame,\n    or code object.  The source code is returned as a list of all the lines\n    in the file and the line number indexes a line in that list.  An OSError\n    is raised if the source code cannot be retrieved.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_77findsource = {"findsource", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_77findsource, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_76findsource};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_77findsource(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findsource (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_76findsource(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_76findsource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_file = NULL;
  PyObject *__pyx_v_module = NULL;
  PyObject *__pyx_v_lines = NULL;
  PyObject *__pyx_v_qualname = NULL;
  PyObject *__pyx_v_source = NULL;
  PyObject *__pyx_v_tree = NULL;
  PyObject *__pyx_v_class_finder = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_line_number = NULL;
  PyObject *__pyx_v_lnum = NULL;
  PyObject *__pyx_v_pat = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("findsource", 0);
  __Pyx_INCREF(__pyx_v_object);

  /* "Aspidites/_vendor/contracts/inspection.py":863
 *     is raised if the source code cannot be retrieved."""
 * 
 *     file = getsourcefile(object)             # <<<<<<<<<<<<<<
 *     if file:
 *         # Invalidate cache if needed.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getsourcefile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_file = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":864
 * 
 *     file = getsourcefile(object)
 *     if file:             # <<<<<<<<<<<<<<
 *         # Invalidate cache if needed.
 *         linecache.checkcache(file)
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_file); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 864, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":866
 *     if file:
 *         # Invalidate cache if needed.
 *         linecache.checkcache(file)             # <<<<<<<<<<<<<<
 *     else:
 *         file = getfile(object)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_linecache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_checkcache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_file) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_file);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":864
 * 
 *     file = getsourcefile(object)
 *     if file:             # <<<<<<<<<<<<<<
 *         # Invalidate cache if needed.
 *         linecache.checkcache(file)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":868
 *         linecache.checkcache(file)
 *     else:
 *         file = getfile(object)             # <<<<<<<<<<<<<<
 *         # Allow filenames in form of "<something>" to pass through.
 *         # `doctest` monkeypatches `linecache` module to enable
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_file, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":872
 *         # `doctest` monkeypatches `linecache` module to enable
 *         # inspection, so let `linecache.getlines` to be called.
 *         if not (file.startswith('<') and file.endswith('>')):             # <<<<<<<<<<<<<<
 *             raise OSError('source code not available')
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_file, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u__15) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u__15);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_file, __pyx_n_s_endswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u__16) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u__16);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;
    __pyx_t_5 = ((!__pyx_t_4) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "Aspidites/_vendor/contracts/inspection.py":873
 *         # inspection, so let `linecache.getlines` to be called.
 *         if not (file.startswith('<') and file.endswith('>')):
 *             raise OSError('source code not available')             # <<<<<<<<<<<<<<
 * 
 *     module = getmodule(object, file)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 873, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":872
 *         # `doctest` monkeypatches `linecache` module to enable
 *         # inspection, so let `linecache.getlines` to be called.
 *         if not (file.startswith('<') and file.endswith('>')):             # <<<<<<<<<<<<<<
 *             raise OSError('source code not available')
 * 
 */
    }
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":875
 *             raise OSError('source code not available')
 * 
 *     module = getmodule(object, file)             # <<<<<<<<<<<<<<
 *     if module:
 *         lines = linecache.getlines(file, module.__dict__)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getmodule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_object, __pyx_v_file};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_object, __pyx_v_file};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_object);
    __Pyx_GIVEREF(__pyx_v_object);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_object);
    __Pyx_INCREF(__pyx_v_file);
    __Pyx_GIVEREF(__pyx_v_file);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_file);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_module = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":876
 * 
 *     module = getmodule(object, file)
 *     if module:             # <<<<<<<<<<<<<<
 *         lines = linecache.getlines(file, module.__dict__)
 *     else:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_module); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 876, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":877
 *     module = getmodule(object, file)
 *     if module:
 *         lines = linecache.getlines(file, module.__dict__)             # <<<<<<<<<<<<<<
 *     else:
 *         lines = linecache.getlines(file)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_linecache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getlines); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_file, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_file, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_file);
      __Pyx_GIVEREF(__pyx_v_file);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_v_file);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_lines = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":876
 * 
 *     module = getmodule(object, file)
 *     if module:             # <<<<<<<<<<<<<<
 *         lines = linecache.getlines(file, module.__dict__)
 *     else:
 */
    goto __pyx_L7;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":879
 *         lines = linecache.getlines(file, module.__dict__)
 *     else:
 *         lines = linecache.getlines(file)             # <<<<<<<<<<<<<<
 *     if not lines:
 *         raise OSError('could not get source code')
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_linecache); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_getlines); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_file) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_file);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_lines = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L7:;

  /* "Aspidites/_vendor/contracts/inspection.py":880
 *     else:
 *         lines = linecache.getlines(file)
 *     if not lines:             # <<<<<<<<<<<<<<
 *         raise OSError('could not get source code')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_lines); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 880, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/contracts/inspection.py":881
 *         lines = linecache.getlines(file)
 *     if not lines:
 *         raise OSError('could not get source code')             # <<<<<<<<<<<<<<
 * 
 *     if ismodule(object):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 881, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":880
 *     else:
 *         lines = linecache.getlines(file)
 *     if not lines:             # <<<<<<<<<<<<<<
 *         raise OSError('could not get source code')
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":883
 *         raise OSError('could not get source code')
 * 
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         return lines, 0
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":884
 * 
 *     if ismodule(object):
 *         return lines, 0             # <<<<<<<<<<<<<<
 * 
 *     if isclass(object):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_lines);
    __Pyx_GIVEREF(__pyx_v_lines);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_lines);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_0);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":883
 *         raise OSError('could not get source code')
 * 
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         return lines, 0
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":886
 *         return lines, 0
 * 
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         qualname = object.__qualname__
 *         source = ''.join(lines)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_isclass); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":887
 * 
 *     if isclass(object):
 *         qualname = object.__qualname__             # <<<<<<<<<<<<<<
 *         source = ''.join(lines)
 *         tree = ast.parse(source)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_qualname = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":888
 *     if isclass(object):
 *         qualname = object.__qualname__
 *         source = ''.join(lines)             # <<<<<<<<<<<<<<
 *         tree = ast.parse(source)
 *         class_finder = _ClassFinder(qualname)
 */
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__19, __pyx_v_lines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_source = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":889
 *         qualname = object.__qualname__
 *         source = ''.join(lines)
 *         tree = ast.parse(source)             # <<<<<<<<<<<<<<
 *         class_finder = _ClassFinder(qualname)
 *         try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ast); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_parse); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_source) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_source);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_tree = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":890
 *         source = ''.join(lines)
 *         tree = ast.parse(source)
 *         class_finder = _ClassFinder(qualname)             # <<<<<<<<<<<<<<
 *         try:
 *             class_finder.visit(tree)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ClassFinder); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_qualname) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_qualname);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_class_finder = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":891
 *         tree = ast.parse(source)
 *         class_finder = _ClassFinder(qualname)
 *         try:             # <<<<<<<<<<<<<<
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":892
 *         class_finder = _ClassFinder(qualname)
 *         try:
 *             class_finder.visit(tree)             # <<<<<<<<<<<<<<
 *         except ClassFoundException as e:
 *             line_number = e.args[0]
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_class_finder, __pyx_n_s_visit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 892, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_tree) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_tree);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":891
 *         tree = ast.parse(source)
 *         class_finder = _ClassFinder(qualname)
 *         try:             # <<<<<<<<<<<<<<
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":897
 *             return lines, line_number
 *         else:
 *             raise OSError('could not find class definition')             # <<<<<<<<<<<<<<
 * 
 *     if ismethod(object):
 */
      /*else:*/ {
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 897, __pyx_L13_except_error)
      }
      __pyx_L11_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":893
 *         try:
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:             # <<<<<<<<<<<<<<
 *             line_number = e.args[0]
 *             return lines, line_number
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ClassFoundException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 893, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_7, __pyx_t_8);
      __pyx_t_1 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.findsource", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_1) < 0) __PYX_ERR(0, 893, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_e = __pyx_t_7;
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":894
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:
 *             line_number = e.args[0]             # <<<<<<<<<<<<<<
 *             return lines, line_number
 *         else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_line_number = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":895
 *         except ClassFoundException as e:
 *             line_number = e.args[0]
 *             return lines, line_number             # <<<<<<<<<<<<<<
 *         else:
 *             raise OSError('could not find class definition')
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 895, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_lines);
          __Pyx_GIVEREF(__pyx_v_lines);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_lines);
          __Pyx_INCREF(__pyx_v_line_number);
          __Pyx_GIVEREF(__pyx_v_line_number);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_line_number);
          __pyx_r = __pyx_t_2;
          __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L21_return;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":893
 *         try:
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:             # <<<<<<<<<<<<<<
 *             line_number = e.args[0]
 *             return lines, line_number
 */
        /*finally:*/ {
          __pyx_L22_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e);
              __pyx_v_e = NULL;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L13_except_error;
          }
          __pyx_L21_return: {
            __pyx_t_19 = __pyx_r;
            __pyx_r = 0;
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
            __pyx_r = __pyx_t_19;
            __pyx_t_19 = 0;
            goto __pyx_L14_except_return;
          }
        }
      }
      goto __pyx_L13_except_error;
      __pyx_L13_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":891
 *         tree = ast.parse(source)
 *         class_finder = _ClassFinder(qualname)
 *         try:             # <<<<<<<<<<<<<<
 *             class_finder.visit(tree)
 *         except ClassFoundException as e:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L14_except_return:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":886
 *         return lines, 0
 * 
 *     if isclass(object):             # <<<<<<<<<<<<<<
 *         qualname = object.__qualname__
 *         source = ''.join(lines)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":899
 *             raise OSError('could not find class definition')
 * 
 *     if ismethod(object):             # <<<<<<<<<<<<<<
 *         object = object.__func__
 *     if isfunction(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":900
 * 
 *     if ismethod(object):
 *         object = object.__func__             # <<<<<<<<<<<<<<
 *     if isfunction(object):
 *         object = object.__code__
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":899
 *             raise OSError('could not find class definition')
 * 
 *     if ismethod(object):             # <<<<<<<<<<<<<<
 *         object = object.__func__
 *     if isfunction(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":901
 *     if ismethod(object):
 *         object = object.__func__
 *     if isfunction(object):             # <<<<<<<<<<<<<<
 *         object = object.__code__
 *     if istraceback(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":902
 *         object = object.__func__
 *     if isfunction(object):
 *         object = object.__code__             # <<<<<<<<<<<<<<
 *     if istraceback(object):
 *         object = object.tb_frame
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":901
 *     if ismethod(object):
 *         object = object.__func__
 *     if isfunction(object):             # <<<<<<<<<<<<<<
 *         object = object.__code__
 *     if istraceback(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":903
 *     if isfunction(object):
 *         object = object.__code__
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 *     if isframe(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_istraceback); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":904
 *         object = object.__code__
 *     if istraceback(object):
 *         object = object.tb_frame             # <<<<<<<<<<<<<<
 *     if isframe(object):
 *         object = object.f_code
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":903
 *     if isfunction(object):
 *         object = object.__code__
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 *     if isframe(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":905
 *     if istraceback(object):
 *         object = object.tb_frame
 *     if isframe(object):             # <<<<<<<<<<<<<<
 *         object = object.f_code
 *     if iscode(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_isframe); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":906
 *         object = object.tb_frame
 *     if isframe(object):
 *         object = object.f_code             # <<<<<<<<<<<<<<
 *     if iscode(object):
 *         if not hasattr(object, 'co_firstlineno'):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_f_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":905
 *     if istraceback(object):
 *         object = object.tb_frame
 *     if isframe(object):             # <<<<<<<<<<<<<<
 *         object = object.f_code
 *     if iscode(object):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":907
 *     if isframe(object):
 *         object = object.f_code
 *     if iscode(object):             # <<<<<<<<<<<<<<
 *         if not hasattr(object, 'co_firstlineno'):
 *             raise OSError('could not find function definition')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_iscode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":908
 *         object = object.f_code
 *     if iscode(object):
 *         if not hasattr(object, 'co_firstlineno'):             # <<<<<<<<<<<<<<
 *             raise OSError('could not find function definition')
 *         lnum = object.co_firstlineno - 1
 */
    __pyx_t_4 = __Pyx_HasAttr(__pyx_v_object, __pyx_n_u_co_firstlineno); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 908, __pyx_L1_error)
    __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "Aspidites/_vendor/contracts/inspection.py":909
 *     if iscode(object):
 *         if not hasattr(object, 'co_firstlineno'):
 *             raise OSError('could not find function definition')             # <<<<<<<<<<<<<<
 *         lnum = object.co_firstlineno - 1
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 909, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":908
 *         object = object.f_code
 *     if iscode(object):
 *         if not hasattr(object, 'co_firstlineno'):             # <<<<<<<<<<<<<<
 *             raise OSError('could not find function definition')
 *         lnum = object.co_firstlineno - 1
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":910
 *         if not hasattr(object, 'co_firstlineno'):
 *             raise OSError('could not find function definition')
 *         lnum = object.co_firstlineno - 1             # <<<<<<<<<<<<<<
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 *         while lnum > 0:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_co_firstlineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_lnum = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":911
 *             raise OSError('could not find function definition')
 *         lnum = object.co_firstlineno - 1
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')             # <<<<<<<<<<<<<<
 *         while lnum > 0:
 *             try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_compile); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_1, __pyx_kp_u_s_def_s_s_async_s_def_s_w_lambd) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u_s_def_s_s_async_s_def_s_w_lambd);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_pat = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":912
 *         lnum = object.co_firstlineno - 1
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 *         while lnum > 0:             # <<<<<<<<<<<<<<
 *             try:
 *                 line = lines[lnum]
 */
    while (1) {
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_lnum, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (!__pyx_t_5) break;

      /* "Aspidites/_vendor/contracts/inspection.py":913
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 *         while lnum > 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 line = lines[lnum]
 *             except IndexError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":914
 *         while lnum > 0:
 *             try:
 *                 line = lines[lnum]             # <<<<<<<<<<<<<<
 *             except IndexError:
 *                 raise OSError('lineno is out of bounds')
 */
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_lnum); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 914, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":913
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 *         while lnum > 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 line = lines[lnum]
 *             except IndexError:
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L43_try_end;
        __pyx_L36_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":915
 *             try:
 *                 line = lines[lnum]
 *             except IndexError:             # <<<<<<<<<<<<<<
 *                 raise OSError('lineno is out of bounds')
 *             if pat.match(line):
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
        if (__pyx_t_12) {
          __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.findsource", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(0, 915, __pyx_L38_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_1);

          /* "Aspidites/_vendor/contracts/inspection.py":916
 *                 line = lines[lnum]
 *             except IndexError:
 *                 raise OSError('lineno is out of bounds')             # <<<<<<<<<<<<<<
 *             if pat.match(line):
 *                 break
 */
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 916, __pyx_L38_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 916, __pyx_L38_except_error)
        }
        goto __pyx_L38_except_error;
        __pyx_L38_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":913
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 *         while lnum > 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 line = lines[lnum]
 *             except IndexError:
 */
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L43_try_end:;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":917
 *             except IndexError:
 *                 raise OSError('lineno is out of bounds')
 *             if pat.match(line):             # <<<<<<<<<<<<<<
 *                 break
 *             lnum = lnum - 1
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_pat, __pyx_n_s_match); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_line);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/contracts/inspection.py":918
 *                 raise OSError('lineno is out of bounds')
 *             if pat.match(line):
 *                 break             # <<<<<<<<<<<<<<
 *             lnum = lnum - 1
 *         return lines, lnum
 */
        goto __pyx_L35_break;

        /* "Aspidites/_vendor/contracts/inspection.py":917
 *             except IndexError:
 *                 raise OSError('lineno is out of bounds')
 *             if pat.match(line):             # <<<<<<<<<<<<<<
 *                 break
 *             lnum = lnum - 1
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":919
 *             if pat.match(line):
 *                 break
 *             lnum = lnum - 1             # <<<<<<<<<<<<<<
 *         return lines, lnum
 *     raise OSError('could not find code object')
 */
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_lnum, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_lnum, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L35_break:;

    /* "Aspidites/_vendor/contracts/inspection.py":920
 *                 break
 *             lnum = lnum - 1
 *         return lines, lnum             # <<<<<<<<<<<<<<
 *     raise OSError('could not find code object')
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_lines);
    __Pyx_GIVEREF(__pyx_v_lines);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_lines);
    __Pyx_INCREF(__pyx_v_lnum);
    __Pyx_GIVEREF(__pyx_v_lnum);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_lnum);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":907
 *     if isframe(object):
 *         object = object.f_code
 *     if iscode(object):             # <<<<<<<<<<<<<<
 *         if not hasattr(object, 'co_firstlineno'):
 *             raise OSError('could not find function definition')
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":921
 *             lnum = lnum - 1
 *         return lines, lnum
 *     raise OSError('could not find code object')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 921, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":855
 * 
 * 
 * def findsource(object):             # <<<<<<<<<<<<<<
 *     """Return the entire source file and starting line number for an object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.findsource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file);
  __Pyx_XDECREF(__pyx_v_module);
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_qualname);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_tree);
  __Pyx_XDECREF(__pyx_v_class_finder);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_line_number);
  __Pyx_XDECREF(__pyx_v_lnum);
  __Pyx_XDECREF(__pyx_v_pat);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_object);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":924
 * 
 * 
 * def getcomments(object):             # <<<<<<<<<<<<<<
 *     """Get lines of comments immediately preceding an object's source code.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_79getcomments(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_78getcomments[] = "Get lines of comments immediately preceding an object's source code.\n\n    Returns None when source can't be found.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_79getcomments = {"getcomments", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_79getcomments, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_78getcomments};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_79getcomments(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getcomments (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_78getcomments(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_78getcomments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_lines = NULL;
  PyObject *__pyx_v_lnum = NULL;
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_comments = NULL;
  PyObject *__pyx_v_end = NULL;
  PyObject *__pyx_v_indent = NULL;
  PyObject *__pyx_v_comment = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getcomments", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":929
 *     Returns None when source can't be found.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         lines, lnum = findsource(object)
 *     except (OSError, TypeError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":930
 *     """
 *     try:
 *         lines, lnum = findsource(object)             # <<<<<<<<<<<<<<
 *     except (OSError, TypeError):
 *         return None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_findsource); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_object);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 930, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 930, __pyx_L3_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 930, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 930, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 930, __pyx_L3_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 930, __pyx_L3_error)
        __pyx_L10_unpacking_done:;
      }
      __pyx_v_lines = __pyx_t_5;
      __pyx_t_5 = 0;
      __pyx_v_lnum = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":929
 *     Returns None when source can't be found.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         lines, lnum = findsource(object)
 *     except (OSError, TypeError):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":931
 *     try:
 *         lines, lnum = findsource(object)
 *     except (OSError, TypeError):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcomments", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 931, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);

      /* "Aspidites/_vendor/contracts/inspection.py":932
 *         lines, lnum = findsource(object)
 *     except (OSError, TypeError):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     if ismodule(object):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":929
 *     Returns None when source can't be found.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         lines, lnum = findsource(object)
 *     except (OSError, TypeError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":934
 *         return None
 * 
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         # Look for a comment block at the top of the file.
 *         start = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_10) {

    /* "Aspidites/_vendor/contracts/inspection.py":936
 *     if ismodule(object):
 *         # Look for a comment block at the top of the file.
 *         start = 0             # <<<<<<<<<<<<<<
 *         if lines and lines[0][:2] == '#!':
 *             start = 1
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_start = __pyx_int_0;

    /* "Aspidites/_vendor/contracts/inspection.py":937
 *         # Look for a comment block at the top of the file.
 *         start = 0
 *         if lines and lines[0][:2] == '#!':             # <<<<<<<<<<<<<<
 *             start = 1
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 */
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_lines); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 937, __pyx_L1_error)
    if (__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_lines, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_t_5, 0, 2, NULL, NULL, &__pyx_slice__24, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_kp_u__25, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __pyx_t_11;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/contracts/inspection.py":938
 *         start = 0
 *         if lines and lines[0][:2] == '#!':
 *             start = 1             # <<<<<<<<<<<<<<
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 *             start = start + 1
 */
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_DECREF_SET(__pyx_v_start, __pyx_int_1);

      /* "Aspidites/_vendor/contracts/inspection.py":937
 *         # Look for a comment block at the top of the file.
 *         start = 0
 *         if lines and lines[0][:2] == '#!':             # <<<<<<<<<<<<<<
 *             start = 1
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":939
 *         if lines and lines[0][:2] == '#!':
 *             start = 1
 *         while start < len(lines) and lines[start].strip() in ('', '#'):             # <<<<<<<<<<<<<<
 *             start = start + 1
 *         if start < len(lines) and lines[start][:1] == '#':
 */
    while (1) {
      __pyx_t_12 = PyObject_Length(__pyx_v_lines); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 939, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_start, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_11) {
      } else {
        __pyx_t_10 = __pyx_t_11;
        goto __pyx_L19_bool_binop_done;
      }
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_strip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_kp_u__19, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
      if (!__pyx_t_13) {
      } else {
        __pyx_t_11 = __pyx_t_13;
        goto __pyx_L21_bool_binop_done;
      }
      __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
      __pyx_t_11 = __pyx_t_13;
      __pyx_L21_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = (__pyx_t_11 != 0);
      __pyx_t_10 = __pyx_t_13;
      __pyx_L19_bool_binop_done:;
      if (!__pyx_t_10) break;

      /* "Aspidites/_vendor/contracts/inspection.py":940
 *             start = 1
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 *             start = start + 1             # <<<<<<<<<<<<<<
 *         if start < len(lines) and lines[start][:1] == '#':
 *             comments = []
 */
      __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_start, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_start, __pyx_t_5);
      __pyx_t_5 = 0;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":941
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 *             start = start + 1
 *         if start < len(lines) and lines[start][:1] == '#':             # <<<<<<<<<<<<<<
 *             comments = []
 *             end = start
 */
    __pyx_t_12 = PyObject_Length(__pyx_v_lines); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 941, __pyx_L1_error)
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_start, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_13) {
    } else {
      __pyx_t_10 = __pyx_t_13;
      goto __pyx_L24_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 1, NULL, NULL, &__pyx_slice__27, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __pyx_t_13;
    __pyx_L24_bool_binop_done:;
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/contracts/inspection.py":942
 *             start = start + 1
 *         if start < len(lines) and lines[start][:1] == '#':
 *             comments = []             # <<<<<<<<<<<<<<
 *             end = start
 *             while end < len(lines) and lines[end][:1] == '#':
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_comments = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":943
 *         if start < len(lines) and lines[start][:1] == '#':
 *             comments = []
 *             end = start             # <<<<<<<<<<<<<<
 *             while end < len(lines) and lines[end][:1] == '#':
 *                 comments.append(lines[end].expandtabs())
 */
      __Pyx_INCREF(__pyx_v_start);
      __pyx_v_end = __pyx_v_start;

      /* "Aspidites/_vendor/contracts/inspection.py":944
 *             comments = []
 *             end = start
 *             while end < len(lines) and lines[end][:1] == '#':             # <<<<<<<<<<<<<<
 *                 comments.append(lines[end].expandtabs())
 *                 end = end + 1
 */
      while (1) {
        __pyx_t_12 = PyObject_Length(__pyx_v_lines); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 944, __pyx_L1_error)
        __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PyObject_RichCompare(__pyx_v_end, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_13) {
        } else {
          __pyx_t_10 = __pyx_t_13;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 1, NULL, NULL, &__pyx_slice__27, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 944, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = __pyx_t_13;
        __pyx_L28_bool_binop_done:;
        if (!__pyx_t_10) break;

        /* "Aspidites/_vendor/contracts/inspection.py":945
 *             end = start
 *             while end < len(lines) and lines[end][:1] == '#':
 *                 comments.append(lines[end].expandtabs())             # <<<<<<<<<<<<<<
 *                 end = end + 1
 *             return ''.join(comments)
 */
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 945, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 945, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 945, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_comments, __pyx_t_5); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 945, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":946
 *             while end < len(lines) and lines[end][:1] == '#':
 *                 comments.append(lines[end].expandtabs())
 *                 end = end + 1             # <<<<<<<<<<<<<<
 *             return ''.join(comments)
 * 
 */
        __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_end, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_end, __pyx_t_5);
        __pyx_t_5 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":947
 *                 comments.append(lines[end].expandtabs())
 *                 end = end + 1
 *             return ''.join(comments)             # <<<<<<<<<<<<<<
 * 
 *     # Look for a preceding block of comments at the same indentation.
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__19, __pyx_v_comments); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":941
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 *             start = start + 1
 *         if start < len(lines) and lines[start][:1] == '#':             # <<<<<<<<<<<<<<
 *             comments = []
 *             end = start
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":934
 *         return None
 * 
 *     if ismodule(object):             # <<<<<<<<<<<<<<
 *         # Look for a comment block at the top of the file.
 *         start = 0
 */
    goto __pyx_L13;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":950
 * 
 *     # Look for a preceding block of comments at the same indentation.
 *     elif lnum > 0:             # <<<<<<<<<<<<<<
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1
 */
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_lnum, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 950, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_10) {

    /* "Aspidites/_vendor/contracts/inspection.py":951
 *     # Look for a preceding block of comments at the same indentation.
 *     elif lnum > 0:
 *         indent = indentsize(lines[lnum])             # <<<<<<<<<<<<<<
 *         end = lnum - 1
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_indentsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_lnum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_indent = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":952
 *     elif lnum > 0:
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1             # <<<<<<<<<<<<<<
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \
 *                 indentsize(lines[end]) == indent:
 */
    __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_v_lnum, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_end = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":953
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \             # <<<<<<<<<<<<<<
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_end, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_13) {
    } else {
      __pyx_t_10 = __pyx_t_13;
      goto __pyx_L31_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_5, 0, 1, NULL, NULL, &__pyx_slice__27, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_13) {
    } else {
      __pyx_t_10 = __pyx_t_13;
      goto __pyx_L31_bool_binop_done;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":954
 *         end = lnum - 1
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \
 *                 indentsize(lines[end]) == indent:             # <<<<<<<<<<<<<<
 *             comments = [lines[end].expandtabs().lstrip()]
 *             if end > 0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_indentsize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_v_indent, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __pyx_t_13;
    __pyx_L31_bool_binop_done:;

    /* "Aspidites/_vendor/contracts/inspection.py":953
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \             # <<<<<<<<<<<<<<
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]
 */
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/contracts/inspection.py":955
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]             # <<<<<<<<<<<<<<
 *             if end > 0:
 *                 end = end - 1
 */
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5);
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_v_comments = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":956
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]
 *             if end > 0:             # <<<<<<<<<<<<<<
 *                 end = end - 1
 *                 comment = lines[end].expandtabs().lstrip()
 */
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_end, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 956, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_10) {

        /* "Aspidites/_vendor/contracts/inspection.py":957
 *             comments = [lines[end].expandtabs().lstrip()]
 *             if end > 0:
 *                 end = end - 1             # <<<<<<<<<<<<<<
 *                 comment = lines[end].expandtabs().lstrip()
 *                 while comment[:1] == '#' and indentsize(lines[end]) == indent:
 */
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_v_end, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 957, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_end, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":958
 *             if end > 0:
 *                 end = end - 1
 *                 comment = lines[end].expandtabs().lstrip()             # <<<<<<<<<<<<<<
 *                 while comment[:1] == '#' and indentsize(lines[end]) == indent:
 *                     comments[:0] = [comment]
 */
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_comment = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":959
 *                 end = end - 1
 *                 comment = lines[end].expandtabs().lstrip()
 *                 while comment[:1] == '#' and indentsize(lines[end]) == indent:             # <<<<<<<<<<<<<<
 *                     comments[:0] = [comment]
 *                     end = end - 1
 */
        while (1) {
          __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_comment, 0, 1, NULL, NULL, &__pyx_slice__27, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_13) {
          } else {
            __pyx_t_10 = __pyx_t_13;
            goto __pyx_L37_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_indentsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_v_indent, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 959, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_10 = __pyx_t_13;
          __pyx_L37_bool_binop_done:;
          if (!__pyx_t_10) break;

          /* "Aspidites/_vendor/contracts/inspection.py":960
 *                 comment = lines[end].expandtabs().lstrip()
 *                 while comment[:1] == '#' and indentsize(lines[end]) == indent:
 *                     comments[:0] = [comment]             # <<<<<<<<<<<<<<
 *                     end = end - 1
 *                     if end < 0: break
 */
          __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 960, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_v_comment);
          __Pyx_GIVEREF(__pyx_v_comment);
          PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_comment);
          if (__Pyx_PyObject_SetSlice(__pyx_v_comments, __pyx_t_6, 0, 0, NULL, NULL, NULL, 0, 1, 1) < 0) __PYX_ERR(0, 960, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":961
 *                 while comment[:1] == '#' and indentsize(lines[end]) == indent:
 *                     comments[:0] = [comment]
 *                     end = end - 1             # <<<<<<<<<<<<<<
 *                     if end < 0: break
 *                     comment = lines[end].expandtabs().lstrip()
 */
          __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_end, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF_SET(__pyx_v_end, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":962
 *                     comments[:0] = [comment]
 *                     end = end - 1
 *                     if end < 0: break             # <<<<<<<<<<<<<<
 *                     comment = lines[end].expandtabs().lstrip()
 *             while comments and comments[0].strip() == '#':
 */
          __pyx_t_6 = PyObject_RichCompare(__pyx_v_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 962, __pyx_L1_error)
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_10) {
            goto __pyx_L36_break;
          }

          /* "Aspidites/_vendor/contracts/inspection.py":963
 *                     end = end - 1
 *                     if end < 0: break
 *                     comment = lines[end].expandtabs().lstrip()             # <<<<<<<<<<<<<<
 *             while comments and comments[0].strip() == '#':
 *                 comments[:1] = []
 */
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_lines, __pyx_v_end); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 963, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 963, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 963, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_comment, __pyx_t_6);
          __pyx_t_6 = 0;
        }
        __pyx_L36_break:;

        /* "Aspidites/_vendor/contracts/inspection.py":956
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]
 *             if end > 0:             # <<<<<<<<<<<<<<
 *                 end = end - 1
 *                 comment = lines[end].expandtabs().lstrip()
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":964
 *                     if end < 0: break
 *                     comment = lines[end].expandtabs().lstrip()
 *             while comments and comments[0].strip() == '#':             # <<<<<<<<<<<<<<
 *                 comments[:1] = []
 *             while comments and comments[-1].strip() == '#':
 */
      while (1) {
        __pyx_t_13 = (PyList_GET_SIZE(__pyx_v_comments) != 0);
        if (__pyx_t_13) {
        } else {
          __pyx_t_10 = __pyx_t_13;
          goto __pyx_L42_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(PyList_GET_ITEM(__pyx_v_comments, 0), __pyx_n_s_strip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 964, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 964, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 964, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __pyx_t_13;
        __pyx_L42_bool_binop_done:;
        if (!__pyx_t_10) break;

        /* "Aspidites/_vendor/contracts/inspection.py":965
 *                     comment = lines[end].expandtabs().lstrip()
 *             while comments and comments[0].strip() == '#':
 *                 comments[:1] = []             # <<<<<<<<<<<<<<
 *             while comments and comments[-1].strip() == '#':
 *                 comments[-1:] = []
 */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_PyObject_SetSlice(__pyx_v_comments, __pyx_t_6, 0, 1, NULL, NULL, NULL, 0, 1, 1) < 0) __PYX_ERR(0, 965, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":966
 *             while comments and comments[0].strip() == '#':
 *                 comments[:1] = []
 *             while comments and comments[-1].strip() == '#':             # <<<<<<<<<<<<<<
 *                 comments[-1:] = []
 *             return ''.join(comments)
 */
      while (1) {
        __pyx_t_13 = (PyList_GET_SIZE(__pyx_v_comments) != 0);
        if (__pyx_t_13) {
        } else {
          __pyx_t_10 = __pyx_t_13;
          goto __pyx_L46_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_comments, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_strip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_kp_u__26, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 966, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __pyx_t_13;
        __pyx_L46_bool_binop_done:;
        if (!__pyx_t_10) break;

        /* "Aspidites/_vendor/contracts/inspection.py":967
 *                 comments[:1] = []
 *             while comments and comments[-1].strip() == '#':
 *                 comments[-1:] = []             # <<<<<<<<<<<<<<
 *             return ''.join(comments)
 * 
 */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 967, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_PyObject_SetSlice(__pyx_v_comments, __pyx_t_6, -1L, 0, NULL, NULL, NULL, 1, 0, 1) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":968
 *             while comments and comments[-1].strip() == '#':
 *                 comments[-1:] = []
 *             return ''.join(comments)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__19, __pyx_v_comments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":953
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1
 *         if end >= 0 and lines[end].lstrip()[:1] == '#' and \             # <<<<<<<<<<<<<<
 *                 indentsize(lines[end]) == indent:
 *             comments = [lines[end].expandtabs().lstrip()]
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":950
 * 
 *     # Look for a preceding block of comments at the same indentation.
 *     elif lnum > 0:             # <<<<<<<<<<<<<<
 *         indent = indentsize(lines[lnum])
 *         end = lnum - 1
 */
  }
  __pyx_L13:;

  /* "Aspidites/_vendor/contracts/inspection.py":924
 * 
 * 
 * def getcomments(object):             # <<<<<<<<<<<<<<
 *     """Get lines of comments immediately preceding an object's source code.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcomments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_lnum);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_comments);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_indent);
  __Pyx_XDECREF(__pyx_v_comment);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":977
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.indent = 0
 *         self.islambda = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":978
 * 
 *     def __init__(self):
 *         self.indent = 0             # <<<<<<<<<<<<<<
 *         self.islambda = False
 *         self.started = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indent, __pyx_int_0) < 0) __PYX_ERR(0, 978, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":979
 *     def __init__(self):
 *         self.indent = 0
 *         self.islambda = False             # <<<<<<<<<<<<<<
 *         self.started = False
 *         self.passline = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_islambda, Py_False) < 0) __PYX_ERR(0, 979, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":980
 *         self.indent = 0
 *         self.islambda = False
 *         self.started = False             # <<<<<<<<<<<<<<
 *         self.passline = False
 *         self.indecorator = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_started, Py_False) < 0) __PYX_ERR(0, 980, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":981
 *         self.islambda = False
 *         self.started = False
 *         self.passline = False             # <<<<<<<<<<<<<<
 *         self.indecorator = False
 *         self.decoratorhasargs = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_passline, Py_False) < 0) __PYX_ERR(0, 981, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":982
 *         self.started = False
 *         self.passline = False
 *         self.indecorator = False             # <<<<<<<<<<<<<<
 *         self.decoratorhasargs = False
 *         self.last = 1
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indecorator, Py_False) < 0) __PYX_ERR(0, 982, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":983
 *         self.passline = False
 *         self.indecorator = False
 *         self.decoratorhasargs = False             # <<<<<<<<<<<<<<
 *         self.last = 1
 *         self.body_col0 = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_decoratorhasargs, Py_False) < 0) __PYX_ERR(0, 983, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":984
 *         self.indecorator = False
 *         self.decoratorhasargs = False
 *         self.last = 1             # <<<<<<<<<<<<<<
 *         self.body_col0 = None
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_last, __pyx_int_1) < 0) __PYX_ERR(0, 984, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":985
 *         self.decoratorhasargs = False
 *         self.last = 1
 *         self.body_col0 = None             # <<<<<<<<<<<<<<
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_body_col0, Py_None) < 0) __PYX_ERR(0, 985, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":977
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.indent = 0
 *         self.islambda = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BlockFinder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":987
 *         self.body_col0 = None
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):             # <<<<<<<<<<<<<<
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_3tokeneater(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_3tokeneater = {"tokeneater", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_3tokeneater, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_3tokeneater(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_type = 0;
  PyObject *__pyx_v_token = 0;
  PyObject *__pyx_v_srowcol = 0;
  PyObject *__pyx_v_erowcol = 0;
  CYTHON_UNUSED PyObject *__pyx_v_line = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tokeneater (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_type,&__pyx_n_s_token,&__pyx_n_s_srowcol,&__pyx_n_s_erowcol,&__pyx_n_s_line,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, 1); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_token)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, 2); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_srowcol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, 3); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_erowcol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, 4); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_line)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, 5); __PYX_ERR(0, 987, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "tokeneater") < 0)) __PYX_ERR(0, 987, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_self = values[0];
    __pyx_v_type = values[1];
    __pyx_v_token = values[2];
    __pyx_v_srowcol = values[3];
    __pyx_v_erowcol = values[4];
    __pyx_v_line = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("tokeneater", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 987, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BlockFinder.tokeneater", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_2tokeneater(__pyx_self, __pyx_v_self, __pyx_v_type, __pyx_v_token, __pyx_v_srowcol, __pyx_v_erowcol, __pyx_v_line);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_2tokeneater(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_token, PyObject *__pyx_v_srowcol, PyObject *__pyx_v_erowcol, CYTHON_UNUSED PyObject *__pyx_v_line) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tokeneater", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":988
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):
 *         if not self.started and not self.indecorator:             # <<<<<<<<<<<<<<
 *             # skip any decorators
 *             if token == "@":
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_started); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indecorator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":990
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 *             if token == "@":             # <<<<<<<<<<<<<<
 *                 self.indecorator = True
 *             # look for the first "def", "class" or "lambda"
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_token, __pyx_kp_u__28, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 990, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/contracts/inspection.py":991
 *             # skip any decorators
 *             if token == "@":
 *                 self.indecorator = True             # <<<<<<<<<<<<<<
 *             # look for the first "def", "class" or "lambda"
 *             elif token in ("def", "class", "lambda"):
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indecorator, Py_True) < 0) __PYX_ERR(0, 991, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":990
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 *             if token == "@":             # <<<<<<<<<<<<<<
 *                 self.indecorator = True
 *             # look for the first "def", "class" or "lambda"
 */
      goto __pyx_L6;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":993
 *                 self.indecorator = True
 *             # look for the first "def", "class" or "lambda"
 *             elif token in ("def", "class", "lambda"):             # <<<<<<<<<<<<<<
 *                 if token == "lambda":
 *                     self.islambda = True
 */
    __Pyx_INCREF(__pyx_v_token);
    __pyx_t_2 = __pyx_v_token;
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_def, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_class_2, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_lambda, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":994
 *             # look for the first "def", "class" or "lambda"
 *             elif token in ("def", "class", "lambda"):
 *                 if token == "lambda":             # <<<<<<<<<<<<<<
 *                     self.islambda = True
 *                 self.started = True
 */
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_token, __pyx_n_u_lambda, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 994, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/contracts/inspection.py":995
 *             elif token in ("def", "class", "lambda"):
 *                 if token == "lambda":
 *                     self.islambda = True             # <<<<<<<<<<<<<<
 *                 self.started = True
 *             self.passline = True  # skip to the end of the line
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_islambda, Py_True) < 0) __PYX_ERR(0, 995, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":994
 *             # look for the first "def", "class" or "lambda"
 *             elif token in ("def", "class", "lambda"):
 *                 if token == "lambda":             # <<<<<<<<<<<<<<
 *                     self.islambda = True
 *                 self.started = True
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":996
 *                 if token == "lambda":
 *                     self.islambda = True
 *                 self.started = True             # <<<<<<<<<<<<<<
 *             self.passline = True  # skip to the end of the line
 *         elif token == "(":
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_started, Py_True) < 0) __PYX_ERR(0, 996, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":993
 *                 self.indecorator = True
 *             # look for the first "def", "class" or "lambda"
 *             elif token in ("def", "class", "lambda"):             # <<<<<<<<<<<<<<
 *                 if token == "lambda":
 *                     self.islambda = True
 */
    }
    __pyx_L6:;

    /* "Aspidites/_vendor/contracts/inspection.py":997
 *                     self.islambda = True
 *                 self.started = True
 *             self.passline = True  # skip to the end of the line             # <<<<<<<<<<<<<<
 *         elif token == "(":
 *             if self.indecorator:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_passline, Py_True) < 0) __PYX_ERR(0, 997, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":988
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):
 *         if not self.started and not self.indecorator:             # <<<<<<<<<<<<<<
 *             # skip any decorators
 *             if token == "@":
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":998
 *                 self.started = True
 *             self.passline = True  # skip to the end of the line
 *         elif token == "(":             # <<<<<<<<<<<<<<
 *             if self.indecorator:
 *                 self.decoratorhasargs = True
 */
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_token, __pyx_kp_u__29, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 998, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":999
 *             self.passline = True  # skip to the end of the line
 *         elif token == "(":
 *             if self.indecorator:             # <<<<<<<<<<<<<<
 *                 self.decoratorhasargs = True
 *         elif token == ")":
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indecorator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1000
 *         elif token == "(":
 *             if self.indecorator:
 *                 self.decoratorhasargs = True             # <<<<<<<<<<<<<<
 *         elif token == ")":
 *             if self.indecorator:
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_decoratorhasargs, Py_True) < 0) __PYX_ERR(0, 1000, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":999
 *             self.passline = True  # skip to the end of the line
 *         elif token == "(":
 *             if self.indecorator:             # <<<<<<<<<<<<<<
 *                 self.decoratorhasargs = True
 *         elif token == ")":
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":998
 *                 self.started = True
 *             self.passline = True  # skip to the end of the line
 *         elif token == "(":             # <<<<<<<<<<<<<<
 *             if self.indecorator:
 *                 self.decoratorhasargs = True
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1001
 *             if self.indecorator:
 *                 self.decoratorhasargs = True
 *         elif token == ")":             # <<<<<<<<<<<<<<
 *             if self.indecorator:
 *                 self.indecorator = False
 */
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_token, __pyx_kp_u__30, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1001, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1002
 *                 self.decoratorhasargs = True
 *         elif token == ")":
 *             if self.indecorator:             # <<<<<<<<<<<<<<
 *                 self.indecorator = False
 *                 self.decoratorhasargs = False
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indecorator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1003
 *         elif token == ")":
 *             if self.indecorator:
 *                 self.indecorator = False             # <<<<<<<<<<<<<<
 *                 self.decoratorhasargs = False
 *         elif type == tokenize.NEWLINE:
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indecorator, Py_False) < 0) __PYX_ERR(0, 1003, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1004
 *             if self.indecorator:
 *                 self.indecorator = False
 *                 self.decoratorhasargs = False             # <<<<<<<<<<<<<<
 *         elif type == tokenize.NEWLINE:
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_decoratorhasargs, Py_False) < 0) __PYX_ERR(0, 1004, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1002
 *                 self.decoratorhasargs = True
 *         elif token == ")":
 *             if self.indecorator:             # <<<<<<<<<<<<<<
 *                 self.indecorator = False
 *                 self.decoratorhasargs = False
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1001
 *             if self.indecorator:
 *                 self.decoratorhasargs = True
 *         elif token == ")":             # <<<<<<<<<<<<<<
 *             if self.indecorator:
 *                 self.indecorator = False
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1005
 *                 self.indecorator = False
 *                 self.decoratorhasargs = False
 *         elif type == tokenize.NEWLINE:             # <<<<<<<<<<<<<<
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 *             self.last = srowcol[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_NEWLINE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_type, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1006
 *                 self.decoratorhasargs = False
 *         elif type == tokenize.NEWLINE:
 *             self.passline = False  # stop skipping when a NEWLINE is seen             # <<<<<<<<<<<<<<
 *             self.last = srowcol[0]
 *             if self.islambda:  # lambdas always end at the first NEWLINE
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_passline, Py_False) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1007
 *         elif type == tokenize.NEWLINE:
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 *             self.last = srowcol[0]             # <<<<<<<<<<<<<<
 *             if self.islambda:  # lambdas always end at the first NEWLINE
 *                 raise EndOfBlock
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_srowcol, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_last, __pyx_t_2) < 0) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1008
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 *             self.last = srowcol[0]
 *             if self.islambda:  # lambdas always end at the first NEWLINE             # <<<<<<<<<<<<<<
 *                 raise EndOfBlock
 *             # hitting a NEWLINE when in a decorator without args
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_islambda); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_3)) {

      /* "Aspidites/_vendor/contracts/inspection.py":1009
 *             self.last = srowcol[0]
 *             if self.islambda:  # lambdas always end at the first NEWLINE
 *                 raise EndOfBlock             # <<<<<<<<<<<<<<
 *             # hitting a NEWLINE when in a decorator without args
 *             # ends the decorator
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_EndOfBlock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1009, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1008
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 *             self.last = srowcol[0]
 *             if self.islambda:  # lambdas always end at the first NEWLINE             # <<<<<<<<<<<<<<
 *                 raise EndOfBlock
 *             # hitting a NEWLINE when in a decorator without args
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1012
 *             # hitting a NEWLINE when in a decorator without args
 *             # ends the decorator
 *             if self.indecorator and not self.decoratorhasargs:             # <<<<<<<<<<<<<<
 *                 self.indecorator = False
 *         elif self.passline:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indecorator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __pyx_t_1;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_decoratorhasargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = ((!__pyx_t_1) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1013
 *             # ends the decorator
 *             if self.indecorator and not self.decoratorhasargs:
 *                 self.indecorator = False             # <<<<<<<<<<<<<<
 *         elif self.passline:
 *             pass
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indecorator, Py_False) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1012
 *             # hitting a NEWLINE when in a decorator without args
 *             # ends the decorator
 *             if self.indecorator and not self.decoratorhasargs:             # <<<<<<<<<<<<<<
 *                 self.indecorator = False
 *         elif self.passline:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1005
 *                 self.indecorator = False
 *                 self.decoratorhasargs = False
 *         elif type == tokenize.NEWLINE:             # <<<<<<<<<<<<<<
 *             self.passline = False  # stop skipping when a NEWLINE is seen
 *             self.last = srowcol[0]
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1014
 *             if self.indecorator and not self.decoratorhasargs:
 *                 self.indecorator = False
 *         elif self.passline:             # <<<<<<<<<<<<<<
 *             pass
 *         elif type == tokenize.INDENT:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_passline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1016
 *         elif self.passline:
 *             pass
 *         elif type == tokenize.INDENT:             # <<<<<<<<<<<<<<
 *             if self.body_col0 is None and self.started:
 *                 self.body_col0 = erowcol[1]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_type, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1017
 *             pass
 *         elif type == tokenize.INDENT:
 *             if self.body_col0 is None and self.started:             # <<<<<<<<<<<<<<
 *                 self.body_col0 = erowcol[1]
 *             self.indent = self.indent + 1
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_body_col0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_t_2 == Py_None);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = (__pyx_t_4 != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __pyx_t_1;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_started); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = __pyx_t_1;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1018
 *         elif type == tokenize.INDENT:
 *             if self.body_col0 is None and self.started:
 *                 self.body_col0 = erowcol[1]             # <<<<<<<<<<<<<<
 *             self.indent = self.indent + 1
 *             self.passline = True
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_erowcol, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_body_col0, __pyx_t_2) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1017
 *             pass
 *         elif type == tokenize.INDENT:
 *             if self.body_col0 is None and self.started:             # <<<<<<<<<<<<<<
 *                 self.body_col0 = erowcol[1]
 *             self.indent = self.indent + 1
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1019
 *             if self.body_col0 is None and self.started:
 *                 self.body_col0 = erowcol[1]
 *             self.indent = self.indent + 1             # <<<<<<<<<<<<<<
 *             self.passline = True
 *         elif type == tokenize.DEDENT:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indent, __pyx_t_5) < 0) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1020
 *                 self.body_col0 = erowcol[1]
 *             self.indent = self.indent + 1
 *             self.passline = True             # <<<<<<<<<<<<<<
 *         elif type == tokenize.DEDENT:
 *             self.indent = self.indent - 1
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_passline, Py_True) < 0) __PYX_ERR(0, 1020, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1016
 *         elif self.passline:
 *             pass
 *         elif type == tokenize.INDENT:             # <<<<<<<<<<<<<<
 *             if self.body_col0 is None and self.started:
 *                 self.body_col0 = erowcol[1]
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1021
 *             self.indent = self.indent + 1
 *             self.passline = True
 *         elif type == tokenize.DEDENT:             # <<<<<<<<<<<<<<
 *             self.indent = self.indent - 1
 *             # the end of matching indent/dedent pairs end a block
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_DEDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_type, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1022
 *             self.passline = True
 *         elif type == tokenize.DEDENT:
 *             self.indent = self.indent - 1             # <<<<<<<<<<<<<<
 *             # the end of matching indent/dedent pairs end a block
 *             # (note that this only works for "def"/"class" blocks,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indent); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_indent, __pyx_t_2) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1026
 *             # (note that this only works for "def"/"class" blocks,
 *             #  not e.g. for "if: else:" or "try: finally:" blocks)
 *             if self.indent <= 0:             # <<<<<<<<<<<<<<
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_3)) {

      /* "Aspidites/_vendor/contracts/inspection.py":1027
 *             #  not e.g. for "if: else:" or "try: finally:" blocks)
 *             if self.indent <= 0:
 *                 raise EndOfBlock             # <<<<<<<<<<<<<<
 *         elif type == tokenize.COMMENT:
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_EndOfBlock); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1027, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1026
 *             # (note that this only works for "def"/"class" blocks,
 *             #  not e.g. for "if: else:" or "try: finally:" blocks)
 *             if self.indent <= 0:             # <<<<<<<<<<<<<<
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1021
 *             self.indent = self.indent + 1
 *             self.passline = True
 *         elif type == tokenize.DEDENT:             # <<<<<<<<<<<<<<
 *             self.indent = self.indent - 1
 *             # the end of matching indent/dedent pairs end a block
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1028
 *             if self.indent <= 0:
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:             # <<<<<<<<<<<<<<
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:
 *                 # Include comments if indented at least as much as the block
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_COMMENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_type, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1029
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:             # <<<<<<<<<<<<<<
 *                 # Include comments if indented at least as much as the block
 *                 self.last = srowcol[0]
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_body_col0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = (__pyx_t_5 != Py_None);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = (__pyx_t_1 != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_srowcol, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_body_col0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_5, __pyx_t_2, Py_GE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __pyx_t_4;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1031
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:
 *                 # Include comments if indented at least as much as the block
 *                 self.last = srowcol[0]             # <<<<<<<<<<<<<<
 *         elif self.indent == 0 and type not in (tokenize.COMMENT, tokenize.NL):
 *             # any other token on the same indentation level end the previous
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_srowcol, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_last, __pyx_t_6) < 0) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1029
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:             # <<<<<<<<<<<<<<
 *                 # Include comments if indented at least as much as the block
 *                 self.last = srowcol[0]
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1028
 *             if self.indent <= 0:
 *                 raise EndOfBlock
 *         elif type == tokenize.COMMENT:             # <<<<<<<<<<<<<<
 *             if self.body_col0 is not None and srowcol[1] >= self.body_col0:
 *                 # Include comments if indented at least as much as the block
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1032
 *                 # Include comments if indented at least as much as the block
 *                 self.last = srowcol[0]
 *         elif self.indent == 0 and type not in (tokenize.COMMENT, tokenize.NL):             # <<<<<<<<<<<<<<
 *             # any other token on the same indentation level end the previous
 *             # block as well, except the pseudo-tokens COMMENT and NL.
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_indent); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_6, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L24_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_v_type);
  __pyx_t_2 = __pyx_v_type;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_COMMENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L26_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_NL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = __pyx_t_1;
  __pyx_L26_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__pyx_t_4 != 0);
  __pyx_t_3 = __pyx_t_1;
  __pyx_L24_bool_binop_done:;
  if (unlikely(__pyx_t_3)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1035
 *             # any other token on the same indentation level end the previous
 *             # block as well, except the pseudo-tokens COMMENT and NL.
 *             raise EndOfBlock             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_EndOfBlock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1035, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1032
 *                 # Include comments if indented at least as much as the block
 *                 self.last = srowcol[0]
 *         elif self.indent == 0 and type not in (tokenize.COMMENT, tokenize.NL):             # <<<<<<<<<<<<<<
 *             # any other token on the same indentation level end the previous
 *             # block as well, except the pseudo-tokens COMMENT and NL.
 */
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":987
 *         self.body_col0 = None
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):             # <<<<<<<<<<<<<<
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BlockFinder.tokeneater", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1038
 * 
 * 
 * def getblock(lines):             # <<<<<<<<<<<<<<
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_81getblock(PyObject *__pyx_self, PyObject *__pyx_v_lines); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_80getblock[] = "Extract the block of code at the top of the given list of lines.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_81getblock = {"getblock", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_81getblock, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_80getblock};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_81getblock(PyObject *__pyx_self, PyObject *__pyx_v_lines) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getblock (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_80getblock(__pyx_self, ((PyObject *)__pyx_v_lines));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_80getblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lines) {
  PyObject *__pyx_v_blockfinder = NULL;
  PyObject *__pyx_v_tokens = NULL;
  PyObject *__pyx_v__token = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getblock", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1040
 * def getblock(lines):
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()             # <<<<<<<<<<<<<<
 *     try:
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlockFinder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_blockfinder = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1041
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 *     try:             # <<<<<<<<<<<<<<
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":1042
 *     blockfinder = BlockFinder()
 *     try:
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)             # <<<<<<<<<<<<<<
 *         for _token in tokens:
 *             blockfinder.tokeneater(*_token)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_generate_tokens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_GetIter(__pyx_v_lines); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_next); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1042, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_tokens = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1043
 *     try:
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:             # <<<<<<<<<<<<<<
 *             blockfinder.tokeneater(*_token)
 *     except (EndOfBlock, IndentationError):
 */
      if (likely(PyList_CheckExact(__pyx_v_tokens)) || PyTuple_CheckExact(__pyx_v_tokens)) {
        __pyx_t_1 = __pyx_v_tokens; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_tokens); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1043, __pyx_L3_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1043, __pyx_L3_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1043, __pyx_L3_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_9(__pyx_t_1);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1043, __pyx_L3_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v__token, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1044
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:
 *             blockfinder.tokeneater(*_token)             # <<<<<<<<<<<<<<
 *     except (EndOfBlock, IndentationError):
 *         pass
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_blockfinder, __pyx_n_s_tokeneater); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_v__token); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1044, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1043
 *     try:
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:             # <<<<<<<<<<<<<<
 *             blockfinder.tokeneater(*_token)
 *     except (EndOfBlock, IndentationError):
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1041
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 *     try:             # <<<<<<<<<<<<<<
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1045
 *         for _token in tokens:
 *             blockfinder.tokeneater(*_token)
 *     except (EndOfBlock, IndentationError):             # <<<<<<<<<<<<<<
 *         pass
 *     return lines[:blockfinder.last]
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_EndOfBlock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1045, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_3) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_builtin_IndentationError);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_2, __pyx_t_7);
    __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_7 = 0;
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":1041
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 *     try:             # <<<<<<<<<<<<<<
 *         tokens = tokenize.generate_tokens(iter(lines).__next__)
 *         for _token in tokens:
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1047
 *     except (EndOfBlock, IndentationError):
 *         pass
 *     return lines[:blockfinder.last]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_blockfinder, __pyx_n_s_last); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_lines, 0, 0, NULL, &__pyx_t_7, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1038
 * 
 * 
 * def getblock(lines):             # <<<<<<<<<<<<<<
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_blockfinder);
  __Pyx_XDECREF(__pyx_v_tokens);
  __Pyx_XDECREF(__pyx_v__token);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1050
 * 
 * 
 * def getsourcelines(object):             # <<<<<<<<<<<<<<
 *     """Return a list of source lines and starting line number for an object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_83getsourcelines(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_82getsourcelines[] = "Return a list of source lines and starting line number for an object.\n\n    The argument may be a module, class, method, function, traceback, frame,\n    or code object.  The source code is returned as a list of the lines\n    corresponding to the object and the line number indicates where in the\n    original source file the first line of code was found.  An OSError is\n    raised if the source code cannot be retrieved.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_83getsourcelines = {"getsourcelines", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_83getsourcelines, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_82getsourcelines};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_83getsourcelines(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsourcelines (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_82getsourcelines(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_82getsourcelines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_lines = NULL;
  PyObject *__pyx_v_lnum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsourcelines", 0);
  __Pyx_INCREF(__pyx_v_object);

  /* "Aspidites/_vendor/contracts/inspection.py":1058
 *     original source file the first line of code was found.  An OSError is
 *     raised if the source code cannot be retrieved."""
 *     object = unwrap(object)             # <<<<<<<<<<<<<<
 *     lines, lnum = findsource(object)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unwrap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1059
 *     raised if the source code cannot be retrieved."""
 *     object = unwrap(object)
 *     lines, lnum = findsource(object)             # <<<<<<<<<<<<<<
 * 
 *     if istraceback(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_findsource); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1059, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1059, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1059, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_lines = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_lnum = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1061
 *     lines, lnum = findsource(object)
 * 
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_istraceback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/contracts/inspection.py":1062
 * 
 *     if istraceback(object):
 *         object = object.tb_frame             # <<<<<<<<<<<<<<
 * 
 *     # for module or frame that corresponds to module, return all source lines
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_object, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1061
 *     lines, lnum = findsource(object)
 * 
 *     if istraceback(object):             # <<<<<<<<<<<<<<
 *         object = object.tb_frame
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1065
 * 
 *     # for module or frame that corresponds to module, return all source lines
 *     if (ismodule(object) or             # <<<<<<<<<<<<<<
 *             (isframe(object) and object.f_code.co_name == "<module>")):
 *         return lines, 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L7_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1066
 *     # for module or frame that corresponds to module, return all source lines
 *     if (ismodule(object) or
 *             (isframe(object) and object.f_code.co_name == "<module>")):             # <<<<<<<<<<<<<<
 *         return lines, 0
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isframe); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_object, __pyx_n_s_f_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_co_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_kp_u_module_2, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L7_bool_binop_done:;

  /* "Aspidites/_vendor/contracts/inspection.py":1065
 * 
 *     # for module or frame that corresponds to module, return all source lines
 *     if (ismodule(object) or             # <<<<<<<<<<<<<<
 *             (isframe(object) and object.f_code.co_name == "<module>")):
 *         return lines, 0
 */
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/contracts/inspection.py":1067
 *     if (ismodule(object) or
 *             (isframe(object) and object.f_code.co_name == "<module>")):
 *         return lines, 0             # <<<<<<<<<<<<<<
 *     else:
 *         return getblock(lines[lnum:]), lnum + 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_lines);
    __Pyx_GIVEREF(__pyx_v_lines);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_lines);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1065
 * 
 *     # for module or frame that corresponds to module, return all source lines
 *     if (ismodule(object) or             # <<<<<<<<<<<<<<
 *             (isframe(object) and object.f_code.co_name == "<module>")):
 *         return lines, 0
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1069
 *         return lines, 0
 *     else:
 *         return getblock(lines[lnum:]), lnum + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getblock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_lines, 0, 0, &__pyx_v_lnum, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_lnum, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1050
 * 
 * 
 * def getsourcelines(object):             # <<<<<<<<<<<<<<
 *     """Return a list of source lines and starting line number for an object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getsourcelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_lnum);
  __Pyx_XDECREF(__pyx_v_object);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1072
 * 
 * 
 * def getsource(object):             # <<<<<<<<<<<<<<
 *     """Return the text of the source code for an object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_85getsource(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_84getsource[] = "Return the text of the source code for an object.\n\n    The argument may be a module, class, method, function, traceback, frame,\n    or code object.  The source code is returned as a single string.  An\n    OSError is raised if the source code cannot be retrieved.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_85getsource = {"getsource", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_85getsource, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_84getsource};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_85getsource(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsource (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_84getsource(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_84getsource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_v_lines = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_lnum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsource", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1078
 *     or code object.  The source code is returned as a single string.  An
 *     OSError is raised if the source code cannot be retrieved."""
 *     lines, lnum = getsourcelines(object)             # <<<<<<<<<<<<<<
 *     return ''.join(lines)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getsourcelines); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_object) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_object);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1078, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1078, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1078, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_lines = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_lnum = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1079
 *     OSError is raised if the source code cannot be retrieved."""
 *     lines, lnum = getsourcelines(object)
 *     return ''.join(lines)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__19, __pyx_v_lines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1072
 * 
 * 
 * def getsource(object):             # <<<<<<<<<<<<<<
 *     """Return the text of the source code for an object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getsource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_lnum);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1083
 * 
 * # --------------------------------------------------- class tree extraction
 * def walktree(classes, children, parent):             # <<<<<<<<<<<<<<
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_87walktree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_86walktree[] = "Recursive helper function for getclasstree().";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_87walktree = {"walktree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_87walktree, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_86walktree};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_87walktree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_classes = 0;
  PyObject *__pyx_v_children = 0;
  CYTHON_UNUSED PyObject *__pyx_v_parent = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walktree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_classes,&__pyx_n_s_children,&__pyx_n_s_parent,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("walktree", 1, 3, 3, 1); __PYX_ERR(0, 1083, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("walktree", 1, 3, 3, 2); __PYX_ERR(0, 1083, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "walktree") < 0)) __PYX_ERR(0, 1083, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_classes = values[0];
    __pyx_v_children = values[1];
    __pyx_v_parent = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walktree", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1083, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.walktree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_86walktree(__pyx_self, __pyx_v_classes, __pyx_v_children, __pyx_v_parent);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_86walktree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_classes, PyObject *__pyx_v_children, CYTHON_UNUSED PyObject *__pyx_v_parent) {
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walktree", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1085
 * def walktree(classes, children, parent):
 *     """Recursive helper function for getclasstree()."""
 *     results = []             # <<<<<<<<<<<<<<
 *     classes.sort(key=attrgetter('__module__', '__name__'))
 *     for c in classes:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1086
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 *     classes.sort(key=attrgetter('__module__', '__name__'))             # <<<<<<<<<<<<<<
 *     for c in classes:
 *         results.append((c, c.__bases__))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_classes, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_4) < 0) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1087
 *     results = []
 *     classes.sort(key=attrgetter('__module__', '__name__'))
 *     for c in classes:             # <<<<<<<<<<<<<<
 *         results.append((c, c.__bases__))
 *         if c in children:
 */
  if (likely(PyList_CheckExact(__pyx_v_classes)) || PyTuple_CheckExact(__pyx_v_classes)) {
    __pyx_t_4 = __pyx_v_classes; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_classes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1087, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1087, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1087, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1087, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1087, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1088
 *     classes.sort(key=attrgetter('__module__', '__name__'))
 *     for c in classes:
 *         results.append((c, c.__bases__))             # <<<<<<<<<<<<<<
 *         if c in children:
 *             results.append(walktree(children[c], children, c))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_c, __pyx_n_s_bases); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_c);
    __Pyx_GIVEREF(__pyx_v_c);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_c);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1088, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1089
 *     for c in classes:
 *         results.append((c, c.__bases__))
 *         if c in children:             # <<<<<<<<<<<<<<
 *             results.append(walktree(children[c], children, c))
 *     return results
 */
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_c, __pyx_v_children, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1089, __pyx_L1_error)
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/contracts/inspection.py":1090
 *         results.append((c, c.__bases__))
 *         if c in children:
 *             results.append(walktree(children[c], children, c))             # <<<<<<<<<<<<<<
 *     return results
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_walktree); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_children, __pyx_v_c); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1090, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_3, __pyx_v_children, __pyx_v_c};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 3+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_3, __pyx_v_children, __pyx_v_c};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 3+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(3+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_children);
        __Pyx_GIVEREF(__pyx_v_children);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_children);
        __Pyx_INCREF(__pyx_v_c);
        __Pyx_GIVEREF(__pyx_v_c);
        PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_11, __pyx_v_c);
        __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1090, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1089
 *     for c in classes:
 *         results.append((c, c.__bases__))
 *         if c in children:             # <<<<<<<<<<<<<<
 *             results.append(walktree(children[c], children, c))
 *     return results
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1087
 *     results = []
 *     classes.sort(key=attrgetter('__module__', '__name__'))
 *     for c in classes:             # <<<<<<<<<<<<<<
 *         results.append((c, c.__bases__))
 *         if c in children:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1091
 *         if c in children:
 *             results.append(walktree(children[c], children, c))
 *     return results             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_results);
  __pyx_r = __pyx_v_results;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1083
 * 
 * # --------------------------------------------------- class tree extraction
 * def walktree(classes, children, parent):             # <<<<<<<<<<<<<<
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.walktree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1094
 * 
 * 
 * def getclasstree(classes, unique=False):             # <<<<<<<<<<<<<<
 *     """Arrange the given list of classes into a hierarchy of nested lists.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_89getclasstree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_88getclasstree[] = "Arrange the given list of classes into a hierarchy of nested lists.\n\n    Where a nested list appears, it contains classes derived from the class\n    whose entry immediately precedes the list.  Each entry is a 2-tuple\n    containing a class and a tuple of its base classes.  If the 'unique'\n    argument is true, exactly one entry appears in the returned structure\n    for each class in the given list.  Otherwise, classes using multiple\n    inheritance and their descendants will appear multiple times.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_89getclasstree = {"getclasstree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_89getclasstree, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_88getclasstree};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_89getclasstree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_classes = 0;
  PyObject *__pyx_v_unique = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getclasstree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_classes,&__pyx_n_s_unique,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_unique);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getclasstree") < 0)) __PYX_ERR(0, 1094, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_classes = values[0];
    __pyx_v_unique = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getclasstree", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1094, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getclasstree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_88getclasstree(__pyx_self, __pyx_v_classes, __pyx_v_unique);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_88getclasstree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_classes, PyObject *__pyx_v_unique) {
  PyObject *__pyx_v_children = NULL;
  PyObject *__pyx_v_roots = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_parent = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getclasstree", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1103
 *     for each class in the given list.  Otherwise, classes using multiple
 *     inheritance and their descendants will appear multiple times."""
 *     children = {}             # <<<<<<<<<<<<<<
 *     roots = []
 *     for c in classes:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_children = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1104
 *     inheritance and their descendants will appear multiple times."""
 *     children = {}
 *     roots = []             # <<<<<<<<<<<<<<
 *     for c in classes:
 *         if c.__bases__:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_roots = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1105
 *     children = {}
 *     roots = []
 *     for c in classes:             # <<<<<<<<<<<<<<
 *         if c.__bases__:
 *             for parent in c.__bases__:
 */
  if (likely(PyList_CheckExact(__pyx_v_classes)) || PyTuple_CheckExact(__pyx_v_classes)) {
    __pyx_t_1 = __pyx_v_classes; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_classes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1105, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1105, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1105, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1106
 *     roots = []
 *     for c in classes:
 *         if c.__bases__:             # <<<<<<<<<<<<<<
 *             for parent in c.__bases__:
 *                 if parent not in children:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_c, __pyx_n_s_bases); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":1107
 *     for c in classes:
 *         if c.__bases__:
 *             for parent in c.__bases__:             # <<<<<<<<<<<<<<
 *                 if parent not in children:
 *                     children[parent] = []
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_c, __pyx_n_s_bases); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_6 = __pyx_t_4; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1107, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1107, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1107, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1107, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1107, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1107, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_v_parent, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1108
 *         if c.__bases__:
 *             for parent in c.__bases__:
 *                 if parent not in children:             # <<<<<<<<<<<<<<
 *                     children[parent] = []
 *                 if c not in children[parent]:
 */
        __pyx_t_5 = (__Pyx_PyDict_ContainsTF(__pyx_v_parent, __pyx_v_children, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1108, __pyx_L1_error)
        __pyx_t_9 = (__pyx_t_5 != 0);
        if (__pyx_t_9) {

          /* "Aspidites/_vendor/contracts/inspection.py":1109
 *             for parent in c.__bases__:
 *                 if parent not in children:
 *                     children[parent] = []             # <<<<<<<<<<<<<<
 *                 if c not in children[parent]:
 *                     children[parent].append(c)
 */
          __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1109, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (unlikely(PyDict_SetItem(__pyx_v_children, __pyx_v_parent, __pyx_t_4) < 0)) __PYX_ERR(0, 1109, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1108
 *         if c.__bases__:
 *             for parent in c.__bases__:
 *                 if parent not in children:             # <<<<<<<<<<<<<<
 *                     children[parent] = []
 *                 if c not in children[parent]:
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1110
 *                 if parent not in children:
 *                     children[parent] = []
 *                 if c not in children[parent]:             # <<<<<<<<<<<<<<
 *                     children[parent].append(c)
 *                 if unique and parent in classes: break
 */
        __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_children, __pyx_v_parent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_c, __pyx_t_4, Py_NE)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1110, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_5 = (__pyx_t_9 != 0);
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/contracts/inspection.py":1111
 *                     children[parent] = []
 *                 if c not in children[parent]:
 *                     children[parent].append(c)             # <<<<<<<<<<<<<<
 *                 if unique and parent in classes: break
 *         elif c not in roots:
 */
          __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_children, __pyx_v_parent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyObject_Append(__pyx_t_4, __pyx_v_c); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1110
 *                 if parent not in children:
 *                     children[parent] = []
 *                 if c not in children[parent]:             # <<<<<<<<<<<<<<
 *                     children[parent].append(c)
 *                 if unique and parent in classes: break
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1112
 *                 if c not in children[parent]:
 *                     children[parent].append(c)
 *                 if unique and parent in classes: break             # <<<<<<<<<<<<<<
 *         elif c not in roots:
 *             roots.append(c)
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_unique); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1112, __pyx_L1_error)
        if (__pyx_t_9) {
        } else {
          __pyx_t_5 = __pyx_t_9;
          goto __pyx_L11_bool_binop_done;
        }
        __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_parent, __pyx_v_classes, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1112, __pyx_L1_error)
        __pyx_t_11 = (__pyx_t_9 != 0);
        __pyx_t_5 = __pyx_t_11;
        __pyx_L11_bool_binop_done:;
        if (__pyx_t_5) {
          goto __pyx_L7_break;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1107
 *     for c in classes:
 *         if c.__bases__:
 *             for parent in c.__bases__:             # <<<<<<<<<<<<<<
 *                 if parent not in children:
 *                     children[parent] = []
 */
      }
      __pyx_L7_break:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1106
 *     roots = []
 *     for c in classes:
 *         if c.__bases__:             # <<<<<<<<<<<<<<
 *             for parent in c.__bases__:
 *                 if parent not in children:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1113
 *                     children[parent].append(c)
 *                 if unique and parent in classes: break
 *         elif c not in roots:             # <<<<<<<<<<<<<<
 *             roots.append(c)
 *     for parent in children:
 */
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_c, __pyx_v_roots, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_5 != 0);
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":1114
 *                 if unique and parent in classes: break
 *         elif c not in roots:
 *             roots.append(c)             # <<<<<<<<<<<<<<
 *     for parent in children:
 *         if parent not in classes:
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_roots, __pyx_v_c); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1114, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1113
 *                     children[parent].append(c)
 *                 if unique and parent in classes: break
 *         elif c not in roots:             # <<<<<<<<<<<<<<
 *             roots.append(c)
 *     for parent in children:
 */
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/contracts/inspection.py":1105
 *     children = {}
 *     roots = []
 *     for c in classes:             # <<<<<<<<<<<<<<
 *         if c.__bases__:
 *             for parent in c.__bases__:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1115
 *         elif c not in roots:
 *             roots.append(c)
 *     for parent in children:             # <<<<<<<<<<<<<<
 *         if parent not in classes:
 *             roots.append(parent)
 */
  __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_children, 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_12)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_13 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_2, &__pyx_t_6, NULL, NULL, __pyx_t_12);
    if (unlikely(__pyx_t_13 == 0)) break;
    if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 1115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_parent, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1116
 *             roots.append(c)
 *     for parent in children:
 *         if parent not in classes:             # <<<<<<<<<<<<<<
 *             roots.append(parent)
 *     return walktree(roots, children, None)
 */
    __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_v_parent, __pyx_v_classes, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1116, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_11 != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":1117
 *     for parent in children:
 *         if parent not in classes:
 *             roots.append(parent)             # <<<<<<<<<<<<<<
 *     return walktree(roots, children, None)
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_roots, __pyx_v_parent); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1117, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1116
 *             roots.append(c)
 *     for parent in children:
 *         if parent not in classes:             # <<<<<<<<<<<<<<
 *             roots.append(parent)
 *     return walktree(roots, children, None)
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1118
 *         if parent not in classes:
 *             roots.append(parent)
 *     return walktree(roots, children, None)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_walktree); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_roots, __pyx_v_children, Py_None};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_roots, __pyx_v_children, Py_None};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_14 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_roots);
    __Pyx_GIVEREF(__pyx_v_roots);
    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_12, __pyx_v_roots);
    __Pyx_INCREF(__pyx_v_children);
    __Pyx_GIVEREF(__pyx_v_children);
    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_12, __pyx_v_children);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_12, Py_None);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1094
 * 
 * 
 * def getclasstree(classes, unique=False):             # <<<<<<<<<<<<<<
 *     """Arrange the given list of classes into a hierarchy of nested lists.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getclasstree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_children);
  __Pyx_XDECREF(__pyx_v_roots);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_parent);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1125
 * 
 * 
 * def getargs(co):             # <<<<<<<<<<<<<<
 *     """Get information about the arguments accepted by a code object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_91getargs(PyObject *__pyx_self, PyObject *__pyx_v_co); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_90getargs[] = "Get information about the arguments accepted by a code object.\n\n    Three things are returned: (args, varargs, varkw), where\n    'args' is the list of argument names. Keyword-only arguments are\n    appended. 'varargs' and 'varkw' are the names of the * and **\n    arguments or None.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_91getargs = {"getargs", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_91getargs, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_90getargs};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_91getargs(PyObject *__pyx_self, PyObject *__pyx_v_co) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getargs (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_90getargs(__pyx_self, ((PyObject *)__pyx_v_co));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_90getargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_co) {
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_nargs = NULL;
  PyObject *__pyx_v_nkwargs = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_kwonlyargs = NULL;
  CYTHON_UNUSED long __pyx_v_step;
  PyObject *__pyx_v_varargs = NULL;
  PyObject *__pyx_v_varkw = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getargs", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1132
 *     appended. 'varargs' and 'varkw' are the names of the * and **
 *     arguments or None."""
 *     if not iscode(co):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a code object'.format(co))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iscode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_co) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_co);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1133
 *     arguments or None."""
 *     if not iscode(co):
 *         raise TypeError('{!r} is not a code object'.format(co))             # <<<<<<<<<<<<<<
 * 
 *     names = co.co_varnames
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_code_object, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_co) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_co);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1133, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1132
 *     appended. 'varargs' and 'varkw' are the names of the * and **
 *     arguments or None."""
 *     if not iscode(co):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a code object'.format(co))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1135
 *         raise TypeError('{!r} is not a code object'.format(co))
 * 
 *     names = co.co_varnames             # <<<<<<<<<<<<<<
 *     nargs = co.co_argcount
 *     nkwargs = co.co_kwonlyargcount
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_varnames); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_names = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1136
 * 
 *     names = co.co_varnames
 *     nargs = co.co_argcount             # <<<<<<<<<<<<<<
 *     nkwargs = co.co_kwonlyargcount
 *     args = list(names[:nargs])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_argcount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_nargs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1137
 *     names = co.co_varnames
 *     nargs = co.co_argcount
 *     nkwargs = co.co_kwonlyargcount             # <<<<<<<<<<<<<<
 *     args = list(names[:nargs])
 *     kwonlyargs = list(names[nargs:nargs + nkwargs])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_kwonlyargcount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_nkwargs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1138
 *     nargs = co.co_argcount
 *     nkwargs = co.co_kwonlyargcount
 *     args = list(names[:nargs])             # <<<<<<<<<<<<<<
 *     kwonlyargs = list(names[nargs:nargs + nkwargs])
 *     step = 0
 */
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_names, 0, 0, NULL, &__pyx_v_nargs, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1139
 *     nkwargs = co.co_kwonlyargcount
 *     args = list(names[:nargs])
 *     kwonlyargs = list(names[nargs:nargs + nkwargs])             # <<<<<<<<<<<<<<
 *     step = 0
 * 
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_nargs, __pyx_v_nkwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_names, 0, 0, &__pyx_v_nargs, &__pyx_t_1, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_kwonlyargs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1140
 *     args = list(names[:nargs])
 *     kwonlyargs = list(names[nargs:nargs + nkwargs])
 *     step = 0             # <<<<<<<<<<<<<<
 * 
 *     nargs += nkwargs
 */
  __pyx_v_step = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1142
 *     step = 0
 * 
 *     nargs += nkwargs             # <<<<<<<<<<<<<<
 *     varargs = None
 *     if co.co_flags & CO_VARARGS:
 */
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_nargs, __pyx_v_nkwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_nargs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1143
 * 
 *     nargs += nkwargs
 *     varargs = None             # <<<<<<<<<<<<<<
 *     if co.co_flags & CO_VARARGS:
 *         varargs = co.co_varnames[nargs]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_varargs = Py_None;

  /* "Aspidites/_vendor/contracts/inspection.py":1144
 *     nargs += nkwargs
 *     varargs = None
 *     if co.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *         varargs = co.co_varnames[nargs]
 *         nargs = nargs + 1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CO_VARARGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1145
 *     varargs = None
 *     if co.co_flags & CO_VARARGS:
 *         varargs = co.co_varnames[nargs]             # <<<<<<<<<<<<<<
 *         nargs = nargs + 1
 *     varkw = None
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_varnames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_nargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_varargs, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1146
 *     if co.co_flags & CO_VARARGS:
 *         varargs = co.co_varnames[nargs]
 *         nargs = nargs + 1             # <<<<<<<<<<<<<<
 *     varkw = None
 *     if co.co_flags & CO_VARKEYWORDS:
 */
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_nargs, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_nargs, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1144
 *     nargs += nkwargs
 *     varargs = None
 *     if co.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *         varargs = co.co_varnames[nargs]
 *         nargs = nargs + 1
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1147
 *         varargs = co.co_varnames[nargs]
 *         nargs = nargs + 1
 *     varkw = None             # <<<<<<<<<<<<<<
 *     if co.co_flags & CO_VARKEYWORDS:
 *         varkw = co.co_varnames[nargs]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_varkw = Py_None;

  /* "Aspidites/_vendor/contracts/inspection.py":1148
 *         nargs = nargs + 1
 *     varkw = None
 *     if co.co_flags & CO_VARKEYWORDS:             # <<<<<<<<<<<<<<
 *         varkw = co.co_varnames[nargs]
 *     return Arguments(args + kwonlyargs, varargs, varkw)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CO_VARKEYWORDS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_And(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1149
 *     varkw = None
 *     if co.co_flags & CO_VARKEYWORDS:
 *         varkw = co.co_varnames[nargs]             # <<<<<<<<<<<<<<
 *     return Arguments(args + kwonlyargs, varargs, varkw)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_co, __pyx_n_s_co_varnames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_nargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_varkw, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1148
 *         nargs = nargs + 1
 *     varkw = None
 *     if co.co_flags & CO_VARKEYWORDS:             # <<<<<<<<<<<<<<
 *         varkw = co.co_varnames[nargs]
 *     return Arguments(args + kwonlyargs, varargs, varkw)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1150
 *     if co.co_flags & CO_VARKEYWORDS:
 *         varkw = co.co_varnames[nargs]
 *     return Arguments(args + kwonlyargs, varargs, varkw)             # <<<<<<<<<<<<<<
 * 
 * ArgSpec = namedtuple('ArgSpec', 'args varargs keywords defaults')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Arguments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_v_args, __pyx_v_kwonlyargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_2, __pyx_v_varargs, __pyx_v_varkw};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1150, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_2, __pyx_v_varargs, __pyx_v_varkw};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1150, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_varargs);
    __Pyx_GIVEREF(__pyx_v_varargs);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_varargs);
    __Pyx_INCREF(__pyx_v_varkw);
    __Pyx_GIVEREF(__pyx_v_varkw);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_v_varkw);
    __pyx_t_2 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1125
 * 
 * 
 * def getargs(co):             # <<<<<<<<<<<<<<
 *     """Get information about the arguments accepted by a code object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_nargs);
  __Pyx_XDECREF(__pyx_v_nkwargs);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwonlyargs);
  __Pyx_XDECREF(__pyx_v_varargs);
  __Pyx_XDECREF(__pyx_v_varkw);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1155
 * 
 * 
 * def getargspec(func):             # <<<<<<<<<<<<<<
 *     """Get the names and default values of a function's parameters.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_93getargspec(PyObject *__pyx_self, PyObject *__pyx_v_func); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_92getargspec[] = "Get the names and default values of a function's parameters.\n\n    A tuple of four things is returned: (args, varargs, keywords, defaults).\n    'args' is a list of the argument names, including keyword-only argument names.\n    'varargs' and 'keywords' are the names of the * and ** parameters or None.\n    'defaults' is an n-tuple of the default values of the last n parameters.\n\n    This function is deprecated, as it does not support annotations or\n    keyword-only parameters and will raise ValueError if either is present\n    on the supplied callable.\n\n    For a more structured introspection API, use inspect.signature() instead.\n\n    Alternatively, use getfullargspec() for an API with a similar namedtuple\n    based interface, but full support for annotations and keyword-only\n    parameters.\n\n    Deprecated since Python 3.5, use `inspect.getfullargspec()`.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_93getargspec = {"getargspec", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_93getargspec, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_92getargspec};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_93getargspec(PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getargspec (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_92getargspec(__pyx_self, ((PyObject *)__pyx_v_func));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_92getargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_varargs = NULL;
  PyObject *__pyx_v_varkw = NULL;
  PyObject *__pyx_v_defaults = NULL;
  PyObject *__pyx_v_kwonlyargs = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_kwonlydefaults = NULL;
  PyObject *__pyx_v_ann = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getargspec", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1175
 *     Deprecated since Python 3.5, use `inspect.getfullargspec()`.
 *     """
 *     warnings.warn("inspect.getargspec() is deprecated since Python 3.0, "             # <<<<<<<<<<<<<<
 *                   "use inspect.signature() or inspect.getfullargspec()",
 *                   DeprecationWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1177
 *     warnings.warn("inspect.getargspec() is deprecated since Python 3.0, "
 *                   "use inspect.signature() or inspect.getfullargspec()",
 *                   DeprecationWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *     args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \
 *         getfullargspec(func)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1177, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1175
 *     Deprecated since Python 3.5, use `inspect.getfullargspec()`.
 *     """
 *     warnings.warn("inspect.getargspec() is deprecated since Python 3.0, "             # <<<<<<<<<<<<<<
 *                   "use inspect.signature() or inspect.getfullargspec()",
 *                   DeprecationWarning, stacklevel=2)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__32, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1179
 *                   DeprecationWarning, stacklevel=2)
 *     args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \
 *         getfullargspec(func)             # <<<<<<<<<<<<<<
 *     if kwonlyargs or ann:
 *         raise ValueError("Function has keyword-only parameters or annotations"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getfullargspec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 7)) {
      if (size > 7) __Pyx_RaiseTooManyValuesError(7);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1178, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 6); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 6); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[7] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8};
      for (i=0; i < 7; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[7] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8};
    __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
    for (index=0; index < 7; index++) {
      PyObject* item = __pyx_t_10(__pyx_t_9); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 7) < 0) __PYX_ERR(0, 1178, __pyx_L1_error)
    __pyx_t_10 = NULL;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1178, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1178
 *                   "use inspect.signature() or inspect.getfullargspec()",
 *                   DeprecationWarning, stacklevel=2)
 *     args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \             # <<<<<<<<<<<<<<
 *         getfullargspec(func)
 *     if kwonlyargs or ann:
 */
  __pyx_v_args = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_varargs = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_varkw = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_defaults = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_kwonlyargs = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_kwonlydefaults = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_ann = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1180
 *     args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \
 *         getfullargspec(func)
 *     if kwonlyargs or ann:             # <<<<<<<<<<<<<<
 *         raise ValueError("Function has keyword-only parameters or annotations"
 *                          ", use inspect.signature() API which can support them")
 */
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_kwonlyargs); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
  if (!__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_ann); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __pyx_t_11 = __pyx_t_12;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_11)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1181
 *         getfullargspec(func)
 *     if kwonlyargs or ann:
 *         raise ValueError("Function has keyword-only parameters or annotations"             # <<<<<<<<<<<<<<
 *                          ", use inspect.signature() API which can support them")
 *     return ArgSpec(args, varargs, varkw, defaults)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1181, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1180
 *     args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \
 *         getfullargspec(func)
 *     if kwonlyargs or ann:             # <<<<<<<<<<<<<<
 *         raise ValueError("Function has keyword-only parameters or annotations"
 *                          ", use inspect.signature() API which can support them")
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1183
 *         raise ValueError("Function has keyword-only parameters or annotations"
 *                          ", use inspect.signature() API which can support them")
 *     return ArgSpec(args, varargs, varkw, defaults)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ArgSpec); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = NULL;
  __pyx_t_13 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_13 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_v_defaults};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_v_defaults};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_args);
    __Pyx_GIVEREF(__pyx_v_args);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_13, __pyx_v_args);
    __Pyx_INCREF(__pyx_v_varargs);
    __Pyx_GIVEREF(__pyx_v_varargs);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_13, __pyx_v_varargs);
    __Pyx_INCREF(__pyx_v_varkw);
    __Pyx_GIVEREF(__pyx_v_varkw);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_13, __pyx_v_varkw);
    __Pyx_INCREF(__pyx_v_defaults);
    __Pyx_GIVEREF(__pyx_v_defaults);
    PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_13, __pyx_v_defaults);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1155
 * 
 * 
 * def getargspec(func):             # <<<<<<<<<<<<<<
 *     """Get the names and default values of a function's parameters.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getargspec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_varargs);
  __Pyx_XDECREF(__pyx_v_varkw);
  __Pyx_XDECREF(__pyx_v_defaults);
  __Pyx_XDECREF(__pyx_v_kwonlyargs);
  __Pyx_XDECREF(__pyx_v_kwonlydefaults);
  __Pyx_XDECREF(__pyx_v_ann);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1191
 * 
 * 
 * def getargvalues(frame):             # <<<<<<<<<<<<<<
 *     """Get information about arguments passed into a particular frame.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_95getargvalues(PyObject *__pyx_self, PyObject *__pyx_v_frame); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_94getargvalues[] = "Get information about arguments passed into a particular frame.\n\n    A tuple of four things is returned: (args, varargs, varkw, locals).\n    'args' is a list of the argument names.\n    'varargs' and 'varkw' are the names of the * and ** arguments or None.\n    'locals' is the locals dictionary of the given frame.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_95getargvalues = {"getargvalues", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_95getargvalues, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_94getargvalues};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_95getargvalues(PyObject *__pyx_self, PyObject *__pyx_v_frame) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getargvalues (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_94getargvalues(__pyx_self, ((PyObject *)__pyx_v_frame));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_94getargvalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_varargs = NULL;
  PyObject *__pyx_v_varkw = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getargvalues", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1198
 *     'varargs' and 'varkw' are the names of the * and ** arguments or None.
 *     'locals' is the locals dictionary of the given frame."""
 *     args, varargs, varkw = getargs(frame.f_code)             # <<<<<<<<<<<<<<
 *     return ArgInfo(args, varargs, varkw, frame.f_locals)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1198, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 1198, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1198, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_args = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_varargs = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_varkw = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1199
 *     'locals' is the locals dictionary of the given frame."""
 *     args, varargs, varkw = getargs(frame.f_code)
 *     return ArgInfo(args, varargs, varkw, frame.f_locals)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ArgInfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_locals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_args);
    __Pyx_GIVEREF(__pyx_v_args);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_7, __pyx_v_args);
    __Pyx_INCREF(__pyx_v_varargs);
    __Pyx_GIVEREF(__pyx_v_varargs);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_7, __pyx_v_varargs);
    __Pyx_INCREF(__pyx_v_varkw);
    __Pyx_GIVEREF(__pyx_v_varkw);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_7, __pyx_v_varkw);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_7, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1191
 * 
 * 
 * def getargvalues(frame):             # <<<<<<<<<<<<<<
 *     """Get information about arguments passed into a particular frame.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getargvalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_varargs);
  __Pyx_XDECREF(__pyx_v_varkw);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1202
 * 
 * 
 * def formatannotation(annotation, base_module=None):             # <<<<<<<<<<<<<<
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_97formatannotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_97formatannotation = {"formatannotation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_97formatannotation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_97formatannotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_annotation = 0;
  PyObject *__pyx_v_base_module = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("formatannotation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_annotation,&__pyx_n_s_base_module,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_annotation)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_module);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "formatannotation") < 0)) __PYX_ERR(0, 1202, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_annotation = values[0];
    __pyx_v_base_module = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("formatannotation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1202, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatannotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_96formatannotation(__pyx_self, __pyx_v_annotation, __pyx_v_base_module);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_96formatannotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_annotation, PyObject *__pyx_v_base_module) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("formatannotation", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1203
 * 
 * def formatannotation(annotation, base_module=None):
 *     if getattr(annotation, '__module__', None) == 'typing':             # <<<<<<<<<<<<<<
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_annotation, __pyx_n_u_module, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_typing, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1204
 * def formatannotation(annotation, base_module=None):
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')             # <<<<<<<<<<<<<<
 *     if isinstance(annotation, type):
 *         if annotation.__module__ in ('builtins', base_module):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_Repr(__pyx_v_annotation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1203
 * 
 * def formatannotation(annotation, base_module=None):
 *     if getattr(annotation, '__module__', None) == 'typing':             # <<<<<<<<<<<<<<
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1205
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):             # <<<<<<<<<<<<<<
 *         if annotation.__module__ in ('builtins', base_module):
 *             return annotation.__qualname__
 */
  __pyx_t_2 = PyType_Check(__pyx_v_annotation); 
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1206
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):
 *         if annotation.__module__ in ('builtins', base_module):             # <<<<<<<<<<<<<<
 *             return annotation.__qualname__
 *         return annotation.__module__ + '.' + annotation.__qualname__
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotation, __pyx_n_s_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_builtins, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1206, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_4 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_base_module, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_4 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":1207
 *     if isinstance(annotation, type):
 *         if annotation.__module__ in ('builtins', base_module):
 *             return annotation.__qualname__             # <<<<<<<<<<<<<<
 *         return annotation.__module__ + '.' + annotation.__qualname__
 *     return repr(annotation)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotation, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":1206
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):
 *         if annotation.__module__ in ('builtins', base_module):             # <<<<<<<<<<<<<<
 *             return annotation.__qualname__
 *         return annotation.__module__ + '.' + annotation.__qualname__
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1208
 *         if annotation.__module__ in ('builtins', base_module):
 *             return annotation.__qualname__
 *         return annotation.__module__ + '.' + annotation.__qualname__             # <<<<<<<<<<<<<<
 *     return repr(annotation)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotation, __pyx_n_s_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_kp_u__10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotation, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1205
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 *     if isinstance(annotation, type):             # <<<<<<<<<<<<<<
 *         if annotation.__module__ in ('builtins', base_module):
 *             return annotation.__qualname__
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1209
 *             return annotation.__qualname__
 *         return annotation.__module__ + '.' + annotation.__qualname__
 *     return repr(annotation)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyObject_Repr(__pyx_v_annotation); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1202
 * 
 * 
 * def formatannotation(annotation, base_module=None):             # <<<<<<<<<<<<<<
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatannotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_99formatannotationrelativeto(PyObject *__pyx_self, PyObject *__pyx_v_object); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_99formatannotationrelativeto = {"formatannotationrelativeto", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_99formatannotationrelativeto, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_99formatannotationrelativeto(PyObject *__pyx_self, PyObject *__pyx_v_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("formatannotationrelativeto (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_98formatannotationrelativeto(__pyx_self, ((PyObject *)__pyx_v_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1215
 *     module = getattr(object, '__module__', None)
 * 
 *     def _formatannotation(annotation):             # <<<<<<<<<<<<<<
 *         return formatannotation(annotation, module)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto_1_formatannotation(PyObject *__pyx_self, PyObject *__pyx_v_annotation); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto_1_formatannotation = {"_formatannotation", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto_1_formatannotation, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto_1_formatannotation(PyObject *__pyx_self, PyObject *__pyx_v_annotation) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_formatannotation (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto__formatannotation(__pyx_self, ((PyObject *)__pyx_v_annotation));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto__formatannotation(PyObject *__pyx_self, PyObject *__pyx_v_annotation) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_formatannotation", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":1216
 * 
 *     def _formatannotation(annotation):
 *         return formatannotation(annotation, module)             # <<<<<<<<<<<<<<
 * 
 *     return _formatannotation
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_formatannotation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_module)) { __Pyx_RaiseClosureNameError("module"); __PYX_ERR(0, 1216, __pyx_L1_error) }
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_annotation, __pyx_cur_scope->__pyx_v_module};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_annotation, __pyx_cur_scope->__pyx_v_module};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_annotation);
    __Pyx_GIVEREF(__pyx_v_annotation);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_annotation);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_module);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_module);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_cur_scope->__pyx_v_module);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1215
 *     module = getattr(object, '__module__', None)
 * 
 *     def _formatannotation(annotation):             # <<<<<<<<<<<<<<
 *         return formatannotation(annotation, module)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatannotationrelativeto._formatannotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_98formatannotationrelativeto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_object) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *__pyx_cur_scope;
  PyObject *__pyx_v__formatannotation = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("formatannotationrelativeto", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1212, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1213
 * 
 * def formatannotationrelativeto(object):
 *     module = getattr(object, '__module__', None)             # <<<<<<<<<<<<<<
 * 
 *     def _formatannotation(annotation):
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_object, __pyx_n_u_module, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_module = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1215
 *     module = getattr(object, '__module__', None)
 * 
 *     def _formatannotation(annotation):             # <<<<<<<<<<<<<<
 *         return formatannotation(annotation, module)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_26formatannotationrelativeto_1_formatannotation, 0, __pyx_n_s_formatannotationrelativeto_local, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__formatannotation = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1218
 *         return formatannotation(annotation, module)
 * 
 *     return _formatannotation             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v__formatannotation);
  __pyx_r = __pyx_v__formatannotation;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatannotationrelativeto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__formatannotation);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_191__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);

  /* "Aspidites/_vendor/contracts/inspection.py":1222
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},             # <<<<<<<<<<<<<<
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,
 */
  __pyx_t_1 = PyTuple_New(12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject*)__pyx_empty_tuple));
  __Pyx_GIVEREF(((PyObject*)__pyx_empty_tuple));
  PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject*)__pyx_empty_tuple));
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_kwonlydefaults);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_kwonlydefaults);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_kwonlydefaults);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_annotations);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_annotations);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_annotations);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatarg);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarargs);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarargs);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarargs);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarkw);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarkw);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvarkw);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvalue);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvalue);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatvalue);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatreturns);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatreturns);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatreturns);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatannotation);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatannotation);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_formatannotation);

  /* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_101formatargspec(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_100formatargspec[] = "Format an argument spec from the values returned by getfullargspec.\n\n    The first seven arguments are (args, varargs, varkw, defaults,\n    kwonlyargs, kwonlydefaults, annotations).  The other five arguments\n    are the corresponding optional formatting functions that are called to\n    turn names and values into strings.  The last argument is an optional\n    function to format the sequence of arguments.\n\n    Deprecated since Python 3.5: use the `signature` function and `Signature`\n    objects.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_101formatargspec = {"formatargspec", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_101formatargspec, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_100formatargspec};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_101formatargspec(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_varargs = 0;
  PyObject *__pyx_v_varkw = 0;
  PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_v_kwonlyargs = 0;
  PyObject *__pyx_v_kwonlydefaults = 0;
  PyObject *__pyx_v_annotations = 0;
  PyObject *__pyx_v_formatarg = 0;
  PyObject *__pyx_v_formatvarargs = 0;
  PyObject *__pyx_v_formatvarkw = 0;
  PyObject *__pyx_v_formatvalue = 0;
  PyObject *__pyx_v_formatreturns = 0;
  PyObject *__pyx_v_formatannotation = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("formatargspec (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_varargs,&__pyx_n_s_varkw,&__pyx_n_s_defaults,&__pyx_n_s_kwonlyargs,&__pyx_n_s_kwonlydefaults,&__pyx_n_s_annotations,&__pyx_n_s_formatarg,&__pyx_n_s_formatvarargs,&__pyx_n_s_formatvarkw,&__pyx_n_s_formatvalue,&__pyx_n_s_formatreturns,&__pyx_n_s_formatannotation,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject*)__pyx_empty_tuple));
    values[5] = __pyx_dynamic_args->__pyx_arg_kwonlydefaults;
    values[6] = __pyx_dynamic_args->__pyx_arg_annotations;
    values[7] = __pyx_dynamic_args->__pyx_arg_formatarg;
    values[8] = __pyx_dynamic_args->__pyx_arg_formatvarargs;
    values[9] = __pyx_dynamic_args->__pyx_arg_formatvarkw;
    values[10] = __pyx_dynamic_args->__pyx_arg_formatvalue;
    values[11] = __pyx_dynamic_args->__pyx_arg_formatreturns;
    values[12] = __pyx_dynamic_args->__pyx_arg_formatannotation;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_varargs);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_varkw);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwonlyargs);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwonlydefaults);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_annotations);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatarg);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvarargs);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvarkw);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvalue);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatreturns);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatannotation);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "formatargspec") < 0)) __PYX_ERR(0, 1221, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_args = values[0];
    __pyx_v_varargs = values[1];
    __pyx_v_varkw = values[2];
    __pyx_v_defaults = values[3];
    __pyx_v_kwonlyargs = values[4];
    __pyx_v_kwonlydefaults = values[5];
    __pyx_v_annotations = values[6];
    __pyx_v_formatarg = values[7];
    __pyx_v_formatvarargs = values[8];
    __pyx_v_formatvarkw = values[9];
    __pyx_v_formatvalue = values[10];
    __pyx_v_formatreturns = values[11];
    __pyx_v_formatannotation = values[12];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("formatargspec", 0, 1, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1221, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargspec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_100formatargspec(__pyx_self, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_v_defaults, __pyx_v_kwonlyargs, __pyx_v_kwonlydefaults, __pyx_v_annotations, __pyx_v_formatarg, __pyx_v_formatvarargs, __pyx_v_formatvarkw, __pyx_v_formatvalue, __pyx_v_formatreturns, __pyx_v_formatannotation);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1248
 *          stacklevel=2)
 * 
 *     def formatargandannotation(arg):             # <<<<<<<<<<<<<<
 *         result = formatarg(arg)
 *         if arg in annotations:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_1formatargandannotation(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_1formatargandannotation = {"formatargandannotation", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_1formatargandannotation, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_1formatargandannotation(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("formatargandannotation (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *__pyx_outer_scope;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("formatargandannotation", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":1249
 * 
 *     def formatargandannotation(arg):
 *         result = formatarg(arg)             # <<<<<<<<<<<<<<
 *         if arg in annotations:
 *             result += ': ' + formatannotation(annotations[arg])
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_formatarg)) { __Pyx_RaiseClosureNameError("formatarg"); __PYX_ERR(0, 1249, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_formatarg);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_formatarg; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_arg) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_arg);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1250
 *     def formatargandannotation(arg):
 *         result = formatarg(arg)
 *         if arg in annotations:             # <<<<<<<<<<<<<<
 *             result += ': ' + formatannotation(annotations[arg])
 *         return result
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_annotations)) { __Pyx_RaiseClosureNameError("annotations"); __PYX_ERR(0, 1250, __pyx_L1_error) }
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_arg, __pyx_cur_scope->__pyx_v_annotations, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1251
 *         result = formatarg(arg)
 *         if arg in annotations:
 *             result += ': ' + formatannotation(annotations[arg])             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_formatannotation)) { __Pyx_RaiseClosureNameError("formatannotation"); __PYX_ERR(0, 1251, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_annotations)) { __Pyx_RaiseClosureNameError("annotations"); __PYX_ERR(0, 1251, __pyx_L1_error) }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_annotations, __pyx_v_arg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_formatannotation);
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_formatannotation; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_kp_u__37, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1250
 *     def formatargandannotation(arg):
 *         result = formatarg(arg)
 *         if arg in annotations:             # <<<<<<<<<<<<<<
 *             result += ': ' + formatannotation(annotations[arg])
 *         return result
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1252
 *         if arg in annotations:
 *             result += ': ' + formatannotation(annotations[arg])
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     specs = []
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1248
 *          stacklevel=2)
 * 
 *     def formatargandannotation(arg):             # <<<<<<<<<<<<<<
 *         result = formatarg(arg)
 *         if arg in annotations:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargspec.formatargandannotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_100formatargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_varargs, PyObject *__pyx_v_varkw, PyObject *__pyx_v_defaults, PyObject *__pyx_v_kwonlyargs, PyObject *__pyx_v_kwonlydefaults, PyObject *__pyx_v_annotations, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvarargs, PyObject *__pyx_v_formatvarkw, PyObject *__pyx_v_formatvalue, PyObject *__pyx_v_formatreturns, PyObject *__pyx_v_formatannotation) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *__pyx_cur_scope;
  PyObject *__pyx_v_warn = NULL;
  PyObject *__pyx_v_formatargandannotation = 0;
  PyObject *__pyx_v_specs = NULL;
  PyObject *__pyx_v_firstdefault = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_spec = NULL;
  PyObject *__pyx_v_kwonlyarg = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("formatargspec", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1221, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_annotations = __pyx_v_annotations;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_annotations);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_annotations);
  __pyx_cur_scope->__pyx_v_formatarg = __pyx_v_formatarg;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_formatarg);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_formatarg);
  __pyx_cur_scope->__pyx_v_formatannotation = __pyx_v_formatannotation;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_formatannotation);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_formatannotation);

  /* "Aspidites/_vendor/contracts/inspection.py":1241
 *     """
 * 
 *     from warnings import warn             # <<<<<<<<<<<<<<
 * 
 *     warn("`formatargspec` is deprecated since Python 3.5. Use `signature` and "
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_warn);
  __Pyx_GIVEREF(__pyx_n_s_warn);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_warn);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_warnings, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_warn = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1246
 *          "the `Signature` object directly",
 *          DeprecationWarning,
 *          stacklevel=2)             # <<<<<<<<<<<<<<
 * 
 *     def formatargandannotation(arg):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1246, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1243
 *     from warnings import warn
 * 
 *     warn("`formatargspec` is deprecated since Python 3.5. Use `signature` and "             # <<<<<<<<<<<<<<
 *          "the `Signature` object directly",
 *          DeprecationWarning,
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_warn, __pyx_tuple__38, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1248
 *          stacklevel=2)
 * 
 *     def formatargandannotation(arg):             # <<<<<<<<<<<<<<
 *         result = formatarg(arg)
 *         if arg in annotations:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_1formatargandannotation, 0, __pyx_n_s_formatargspec_locals_formatargan, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_formatargandannotation = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1254
 *         return result
 * 
 *     specs = []             # <<<<<<<<<<<<<<
 *     if defaults:
 *         firstdefault = len(args) - len(defaults)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_specs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1255
 * 
 *     specs = []
 *     if defaults:             # <<<<<<<<<<<<<<
 *         firstdefault = len(args) - len(defaults)
 *     for i, arg in enumerate(args):
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1255, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1256
 *     specs = []
 *     if defaults:
 *         firstdefault = len(args) - len(defaults)             # <<<<<<<<<<<<<<
 *     for i, arg in enumerate(args):
 *         spec = formatargandannotation(arg)
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1256, __pyx_L1_error)
    __pyx_t_5 = PyObject_Length(__pyx_v_defaults); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1256, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t((__pyx_t_4 - __pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_firstdefault = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1255
 * 
 *     specs = []
 *     if defaults:             # <<<<<<<<<<<<<<
 *         firstdefault = len(args) - len(defaults)
 *     for i, arg in enumerate(args):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1257
 *     if defaults:
 *         firstdefault = len(args) - len(defaults)
 *     for i, arg in enumerate(args):             # <<<<<<<<<<<<<<
 *         spec = formatargandannotation(arg)
 *         # noinspection PyUnboundLocalVariable
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_args)) || PyTuple_CheckExact(__pyx_v_args)) {
    __pyx_t_2 = __pyx_v_args; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1257, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1257, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1257, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1257, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1258
 *         firstdefault = len(args) - len(defaults)
 *     for i, arg in enumerate(args):
 *         spec = formatargandannotation(arg)             # <<<<<<<<<<<<<<
 *         # noinspection PyUnboundLocalVariable
 *         if defaults and i >= firstdefault:
 */
    __pyx_t_7 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(__pyx_v_formatargandannotation, __pyx_v_arg); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_spec, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1260
 *         spec = formatargandannotation(arg)
 *         # noinspection PyUnboundLocalVariable
 *         if defaults and i >= firstdefault:             # <<<<<<<<<<<<<<
 *             spec = spec + formatvalue(defaults[i - firstdefault])
 *         specs.append(spec)
 */
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1260, __pyx_L1_error)
    if (__pyx_t_8) {
    } else {
      __pyx_t_3 = __pyx_t_8;
      goto __pyx_L7_bool_binop_done;
    }
    if (unlikely(!__pyx_v_firstdefault)) { __Pyx_RaiseUnboundLocalError("firstdefault"); __PYX_ERR(0, 1260, __pyx_L1_error) }
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_i, __pyx_v_firstdefault, Py_GE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1260, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1260, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_3 = __pyx_t_8;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/contracts/inspection.py":1261
 *         # noinspection PyUnboundLocalVariable
 *         if defaults and i >= firstdefault:
 *             spec = spec + formatvalue(defaults[i - firstdefault])             # <<<<<<<<<<<<<<
 *         specs.append(spec)
 *     if varargs is not None:
 */
      if (unlikely(!__pyx_v_firstdefault)) { __Pyx_RaiseUnboundLocalError("firstdefault"); __PYX_ERR(0, 1261, __pyx_L1_error) }
      __pyx_t_9 = PyNumber_Subtract(__pyx_v_i, __pyx_v_firstdefault); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_defaults, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_INCREF(__pyx_v_formatvalue);
      __pyx_t_9 = __pyx_v_formatvalue; __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Add(__pyx_v_spec, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_spec, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1260
 *         spec = formatargandannotation(arg)
 *         # noinspection PyUnboundLocalVariable
 *         if defaults and i >= firstdefault:             # <<<<<<<<<<<<<<
 *             spec = spec + formatvalue(defaults[i - firstdefault])
 *         specs.append(spec)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1262
 *         if defaults and i >= firstdefault:
 *             spec = spec + formatvalue(defaults[i - firstdefault])
 *         specs.append(spec)             # <<<<<<<<<<<<<<
 *     if varargs is not None:
 *         specs.append(formatvarargs(formatargandannotation(varargs)))
 */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_v_spec); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1262, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1257
 *     if defaults:
 *         firstdefault = len(args) - len(defaults)
 *     for i, arg in enumerate(args):             # <<<<<<<<<<<<<<
 *         spec = formatargandannotation(arg)
 *         # noinspection PyUnboundLocalVariable
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1263
 *             spec = spec + formatvalue(defaults[i - firstdefault])
 *         specs.append(spec)
 *     if varargs is not None:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarargs(formatargandannotation(varargs)))
 *     else:
 */
  __pyx_t_3 = (__pyx_v_varargs != Py_None);
  __pyx_t_8 = (__pyx_t_3 != 0);
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/contracts/inspection.py":1264
 *         specs.append(spec)
 *     if varargs is not None:
 *         specs.append(formatvarargs(formatargandannotation(varargs)))             # <<<<<<<<<<<<<<
 *     else:
 *         if kwonlyargs:
 */
    __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(__pyx_v_formatargandannotation, __pyx_v_varargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_formatvarargs);
    __pyx_t_9 = __pyx_v_formatvarargs; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1263
 *             spec = spec + formatvalue(defaults[i - firstdefault])
 *         specs.append(spec)
 *     if varargs is not None:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarargs(formatargandannotation(varargs)))
 *     else:
 */
    goto __pyx_L9;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1266
 *         specs.append(formatvarargs(formatargandannotation(varargs)))
 *     else:
 *         if kwonlyargs:             # <<<<<<<<<<<<<<
 *             specs.append('*')
 *     if kwonlyargs:
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_kwonlyargs); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1266, __pyx_L1_error)
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":1267
 *     else:
 *         if kwonlyargs:
 *             specs.append('*')             # <<<<<<<<<<<<<<
 *     if kwonlyargs:
 *         for kwonlyarg in kwonlyargs:
 */
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_kp_u_); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1267, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1266
 *         specs.append(formatvarargs(formatargandannotation(varargs)))
 *     else:
 *         if kwonlyargs:             # <<<<<<<<<<<<<<
 *             specs.append('*')
 *     if kwonlyargs:
 */
    }
  }
  __pyx_L9:;

  /* "Aspidites/_vendor/contracts/inspection.py":1268
 *         if kwonlyargs:
 *             specs.append('*')
 *     if kwonlyargs:             # <<<<<<<<<<<<<<
 *         for kwonlyarg in kwonlyargs:
 *             spec = formatargandannotation(kwonlyarg)
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_kwonlyargs); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/contracts/inspection.py":1269
 *             specs.append('*')
 *     if kwonlyargs:
 *         for kwonlyarg in kwonlyargs:             # <<<<<<<<<<<<<<
 *             spec = formatargandannotation(kwonlyarg)
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:
 */
    if (likely(PyList_CheckExact(__pyx_v_kwonlyargs)) || PyTuple_CheckExact(__pyx_v_kwonlyargs)) {
      __pyx_t_1 = __pyx_v_kwonlyargs; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_kwonlyargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1269, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1269, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1269, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1269, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1269, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_6(__pyx_t_1);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1269, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_kwonlyarg, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1270
 *     if kwonlyargs:
 *         for kwonlyarg in kwonlyargs:
 *             spec = formatargandannotation(kwonlyarg)             # <<<<<<<<<<<<<<
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 */
      __pyx_t_9 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(__pyx_v_formatargandannotation, __pyx_v_kwonlyarg); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_spec, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1271
 *         for kwonlyarg in kwonlyargs:
 *             spec = formatargandannotation(kwonlyarg)
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:             # <<<<<<<<<<<<<<
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 *             specs.append(spec)
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_kwonlydefaults); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1271, __pyx_L1_error)
      if (__pyx_t_3) {
      } else {
        __pyx_t_8 = __pyx_t_3;
        goto __pyx_L15_bool_binop_done;
      }
      __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_kwonlyarg, __pyx_v_kwonlydefaults, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1271, __pyx_L1_error)
      __pyx_t_13 = (__pyx_t_3 != 0);
      __pyx_t_8 = __pyx_t_13;
      __pyx_L15_bool_binop_done:;
      if (__pyx_t_8) {

        /* "Aspidites/_vendor/contracts/inspection.py":1272
 *             spec = formatargandannotation(kwonlyarg)
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])             # <<<<<<<<<<<<<<
 *             specs.append(spec)
 *     if varkw is not None:
 */
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_kwonlydefaults, __pyx_v_kwonlyarg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_formatvalue);
        __pyx_t_7 = __pyx_v_formatvalue; __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1272, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_spec, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1272, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF_SET(__pyx_v_spec, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1271
 *         for kwonlyarg in kwonlyargs:
 *             spec = formatargandannotation(kwonlyarg)
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:             # <<<<<<<<<<<<<<
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 *             specs.append(spec)
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1273
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 *             specs.append(spec)             # <<<<<<<<<<<<<<
 *     if varkw is not None:
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 */
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_v_spec); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1273, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1269
 *             specs.append('*')
 *     if kwonlyargs:
 *         for kwonlyarg in kwonlyargs:             # <<<<<<<<<<<<<<
 *             spec = formatargandannotation(kwonlyarg)
 *             if kwonlydefaults and kwonlyarg in kwonlydefaults:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1268
 *         if kwonlyargs:
 *             specs.append('*')
 *     if kwonlyargs:             # <<<<<<<<<<<<<<
 *         for kwonlyarg in kwonlyargs:
 *             spec = formatargandannotation(kwonlyarg)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1274
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 *             specs.append(spec)
 *     if varkw is not None:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 *     result = '(' + ', '.join(specs) + ')'
 */
  __pyx_t_8 = (__pyx_v_varkw != Py_None);
  __pyx_t_13 = (__pyx_t_8 != 0);
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":1275
 *             specs.append(spec)
 *     if varkw is not None:
 *         specs.append(formatvarkw(formatargandannotation(varkw)))             # <<<<<<<<<<<<<<
 *     result = '(' + ', '.join(specs) + ')'
 *     if 'return' in annotations:
 */
    __pyx_t_7 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_13formatargspec_formatargandannotation(__pyx_v_formatargandannotation, __pyx_v_varkw); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_formatvarkw);
    __pyx_t_9 = __pyx_v_formatvarkw; __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1274
 *                 spec += formatvalue(kwonlydefaults[kwonlyarg])
 *             specs.append(spec)
 *     if varkw is not None:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 *     result = '(' + ', '.join(specs) + ')'
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1276
 *     if varkw is not None:
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 *     result = '(' + ', '.join(specs) + ')'             # <<<<<<<<<<<<<<
 *     if 'return' in annotations:
 *         result += formatreturns(formatannotation(annotations['return']))
 */
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__41, __pyx_v_specs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyUnicode_Concat(__pyx_kp_u__29, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_t_9, __pyx_kp_u__30); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1277
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 *     result = '(' + ', '.join(specs) + ')'
 *     if 'return' in annotations:             # <<<<<<<<<<<<<<
 *         result += formatreturns(formatannotation(annotations['return']))
 *     return result
 */
  __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_return, __pyx_cur_scope->__pyx_v_annotations, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_13 != 0);
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/contracts/inspection.py":1278
 *     result = '(' + ', '.join(specs) + ')'
 *     if 'return' in annotations:
 *         result += formatreturns(formatannotation(annotations['return']))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_cur_scope->__pyx_v_annotations, __pyx_n_u_return); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_formatannotation);
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_formatannotation; __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v_formatreturns);
    __pyx_t_2 = __pyx_v_formatreturns; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1277
 *         specs.append(formatvarkw(formatargandannotation(varkw)))
 *     result = '(' + ', '.join(specs) + ')'
 *     if 'return' in annotations:             # <<<<<<<<<<<<<<
 *         result += formatreturns(formatannotation(annotations['return']))
 *     return result
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1279
 *     if 'return' in annotations:
 *         result += formatreturns(formatannotation(annotations['return']))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargspec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_warn);
  __Pyx_XDECREF(__pyx_v_formatargandannotation);
  __Pyx_XDECREF(__pyx_v_specs);
  __Pyx_XDECREF(__pyx_v_firstdefault);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_spec);
  __Pyx_XDECREF(__pyx_v_kwonlyarg);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1282
 * 
 * 
 * def formatargvalues(args, varargs, varkw, locals,             # <<<<<<<<<<<<<<
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_193__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatarg);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarargs);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarargs);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarargs);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarkw);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarkw);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvarkw);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvalue);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvalue);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_formatvalue);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_103formatargvalues(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_102formatargvalues[] = "Format an argument spec from the 4 values returned by getargvalues.\n\n    The first four arguments are (args, varargs, varkw, locals).  The\n    next four arguments are the corresponding optional formatting functions\n    that are called to turn names and values into strings.  The ninth\n    argument is an optional function to format the sequence of arguments.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_103formatargvalues = {"formatargvalues", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_103formatargvalues, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_102formatargvalues};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_103formatargvalues(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_varargs = 0;
  PyObject *__pyx_v_varkw = 0;
  PyObject *__pyx_v_locals = 0;
  PyObject *__pyx_v_formatarg = 0;
  PyObject *__pyx_v_formatvarargs = 0;
  PyObject *__pyx_v_formatvarkw = 0;
  PyObject *__pyx_v_formatvalue = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("formatargvalues (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_varargs,&__pyx_n_s_varkw,&__pyx_n_s_locals_2,&__pyx_n_s_formatarg,&__pyx_n_s_formatvarargs,&__pyx_n_s_formatvarkw,&__pyx_n_s_formatvalue,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    __pyx_defaults1 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self);
    values[4] = __pyx_dynamic_args->__pyx_arg_formatarg;
    values[5] = __pyx_dynamic_args->__pyx_arg_formatvarargs;
    values[6] = __pyx_dynamic_args->__pyx_arg_formatvarkw;
    values[7] = __pyx_dynamic_args->__pyx_arg_formatvalue;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_varargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("formatargvalues", 0, 4, 8, 1); __PYX_ERR(0, 1282, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_varkw)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("formatargvalues", 0, 4, 8, 2); __PYX_ERR(0, 1282, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_locals_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("formatargvalues", 0, 4, 8, 3); __PYX_ERR(0, 1282, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatarg);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvarargs);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvarkw);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvalue);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "formatargvalues") < 0)) __PYX_ERR(0, 1282, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_args = values[0];
    __pyx_v_varargs = values[1];
    __pyx_v_varkw = values[2];
    __pyx_v_locals = values[3];
    __pyx_v_formatarg = values[4];
    __pyx_v_formatvarargs = values[5];
    __pyx_v_formatvarkw = values[6];
    __pyx_v_formatvalue = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("formatargvalues", 0, 4, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargvalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_102formatargvalues(__pyx_self, __pyx_v_args, __pyx_v_varargs, __pyx_v_varkw, __pyx_v_locals, __pyx_v_formatarg, __pyx_v_formatvarargs, __pyx_v_formatvarkw, __pyx_v_formatvalue);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1294
 *     argument is an optional function to format the sequence of arguments."""
 * 
 *     def convert(name, locals=locals,             # <<<<<<<<<<<<<<
 *                 formatarg=formatarg, formatvalue=formatvalue):
 *         return formatarg(name) + formatvalue(locals[name])
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_205__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_locals);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_locals);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_locals);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatarg);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatarg);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatvalue);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatvalue);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_formatvalue);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_1convert(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_1convert = {"convert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_1convert, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_1convert(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_locals = 0;
  PyObject *__pyx_v_formatarg = 0;
  PyObject *__pyx_v_formatvalue = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_locals_2,&__pyx_n_s_formatarg,&__pyx_n_s_formatvalue,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults7 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_locals;
    values[2] = __pyx_dynamic_args->__pyx_arg_formatarg;
    values[3] = __pyx_dynamic_args->__pyx_arg_formatvalue;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_locals_2);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatarg);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatvalue);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convert") < 0)) __PYX_ERR(0, 1294, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_locals = values[1];
    __pyx_v_formatarg = values[2];
    __pyx_v_formatvalue = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1294, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargvalues.convert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_convert(__pyx_self, __pyx_v_name, __pyx_v_locals, __pyx_v_formatarg, __pyx_v_formatvalue);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_convert(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_locals, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvalue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1296
 *     def convert(name, locals=locals,
 *                 formatarg=formatarg, formatvalue=formatvalue):
 *         return formatarg(name) + formatvalue(locals[name])             # <<<<<<<<<<<<<<
 * 
 *     specs = []
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_formatarg);
  __pyx_t_2 = __pyx_v_formatarg; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_name);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_locals, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_formatvalue);
  __pyx_t_4 = __pyx_v_formatvalue; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1294
 *     argument is an optional function to format the sequence of arguments."""
 * 
 *     def convert(name, locals=locals,             # <<<<<<<<<<<<<<
 *                 formatarg=formatarg, formatvalue=formatvalue):
 *         return formatarg(name) + formatvalue(locals[name])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargvalues.convert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1282
 * 
 * 
 * def formatargvalues(args, varargs, varkw, locals,             # <<<<<<<<<<<<<<
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_102formatargvalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_varargs, PyObject *__pyx_v_varkw, PyObject *__pyx_v_locals, PyObject *__pyx_v_formatarg, PyObject *__pyx_v_formatvarargs, PyObject *__pyx_v_formatvarkw, PyObject *__pyx_v_formatvalue) {
  PyObject *__pyx_v_convert = 0;
  PyObject *__pyx_v_specs = NULL;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("formatargvalues", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1294
 *     argument is an optional function to format the sequence of arguments."""
 * 
 *     def convert(name, locals=locals,             # <<<<<<<<<<<<<<
 *                 formatarg=formatarg, formatvalue=formatvalue):
 *         return formatarg(name) + formatvalue(locals[name])
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_15formatargvalues_1convert, 0, __pyx_n_s_formatargvalues_locals_convert, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults7), 3)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_locals);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_1)->__pyx_arg_locals = __pyx_v_locals;
  __Pyx_GIVEREF(__pyx_v_locals);

  /* "Aspidites/_vendor/contracts/inspection.py":1295
 * 
 *     def convert(name, locals=locals,
 *                 formatarg=formatarg, formatvalue=formatvalue):             # <<<<<<<<<<<<<<
 *         return formatarg(name) + formatvalue(locals[name])
 * 
 */
  __Pyx_INCREF(__pyx_v_formatarg);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_1)->__pyx_arg_formatarg = __pyx_v_formatarg;
  __Pyx_GIVEREF(__pyx_v_formatarg);
  __Pyx_INCREF(__pyx_v_formatvalue);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_1)->__pyx_arg_formatvalue = __pyx_v_formatvalue;
  __Pyx_GIVEREF(__pyx_v_formatvalue);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_205__defaults__);
  __pyx_v_convert = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1298
 *         return formatarg(name) + formatvalue(locals[name])
 * 
 *     specs = []             # <<<<<<<<<<<<<<
 *     for i in range(len(args)):
 *         specs.append(convert(args[i]))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_specs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1299
 * 
 *     specs = []
 *     for i in range(len(args)):             # <<<<<<<<<<<<<<
 *         specs.append(convert(args[i]))
 *     if varargs:
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1299, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "Aspidites/_vendor/contracts/inspection.py":1300
 *     specs = []
 *     for i in range(len(args)):
 *         specs.append(convert(args[i]))             # <<<<<<<<<<<<<<
 *     if varargs:
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_args, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_v_convert, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1301
 *     for i in range(len(args)):
 *         specs.append(convert(args[i]))
 *     if varargs:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 *     if varkw:
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_varargs); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1301, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":1302
 *         specs.append(convert(args[i]))
 *     if varargs:
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))             # <<<<<<<<<<<<<<
 *     if varkw:
 *         specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))
 */
    __Pyx_INCREF(__pyx_v_formatvarargs);
    __pyx_t_1 = __pyx_v_formatvarargs; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_v_varargs) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_varargs);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_locals, __pyx_v_varargs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_formatvalue);
    __pyx_t_9 = __pyx_v_formatvalue; __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Add(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1302, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1301
 *     for i in range(len(args)):
 *         specs.append(convert(args[i]))
 *     if varargs:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 *     if varkw:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1303
 *     if varargs:
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 *     if varkw:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))
 *     return '(' + ', '.join(specs) + ')'
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_varkw); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1303, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":1304
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 *     if varkw:
 *         specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))             # <<<<<<<<<<<<<<
 *     return '(' + ', '.join(specs) + ')'
 * 
 */
    __Pyx_INCREF(__pyx_v_formatvarkw);
    __pyx_t_1 = __pyx_v_formatvarkw; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_9 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_varkw) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_varkw);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_locals, __pyx_v_varkw); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_formatvalue);
    __pyx_t_8 = __pyx_v_formatvalue; __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_10, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_Add(__pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_specs, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1303
 *     if varargs:
 *         specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
 *     if varkw:             # <<<<<<<<<<<<<<
 *         specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))
 *     return '(' + ', '.join(specs) + ')'
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1305
 *     if varkw:
 *         specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))
 *     return '(' + ', '.join(specs) + ')'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyUnicode_Join(__pyx_kp_u__41, __pyx_v_specs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u__29, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_kp_u__30); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1282
 * 
 * 
 * def formatargvalues(args, varargs, varkw, locals,             # <<<<<<<<<<<<<<
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.formatargvalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_convert);
  __Pyx_XDECREF(__pyx_v_specs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1308
 * 
 * 
 * def _missing_arguments(f_name, argnames, pos, values):             # <<<<<<<<<<<<<<
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_105_missing_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_105_missing_arguments = {"_missing_arguments", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_105_missing_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_105_missing_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_f_name = 0;
  PyObject *__pyx_v_argnames = 0;
  PyObject *__pyx_v_pos = 0;
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_missing_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f_name,&__pyx_n_s_argnames,&__pyx_n_s_pos,&__pyx_n_s_values,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_f_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_argnames)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_missing_arguments", 1, 4, 4, 1); __PYX_ERR(0, 1308, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_missing_arguments", 1, 4, 4, 2); __PYX_ERR(0, 1308, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_missing_arguments", 1, 4, 4, 3); __PYX_ERR(0, 1308, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_missing_arguments") < 0)) __PYX_ERR(0, 1308, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_f_name = values[0];
    __pyx_v_argnames = values[1];
    __pyx_v_pos = values[2];
    __pyx_v_values = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_missing_arguments", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1308, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._missing_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_104_missing_arguments(__pyx_self, __pyx_v_f_name, __pyx_v_argnames, __pyx_v_pos, __pyx_v_values);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_104_missing_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f_name, PyObject *__pyx_v_argnames, PyObject *__pyx_v_pos, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_names = NULL;
  Py_ssize_t __pyx_v_missing;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_tail = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_missing_arguments", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1309
 * 
 * def _missing_arguments(f_name, argnames, pos, values):
 *     names = [repr(name) for name in argnames if name not in values]             # <<<<<<<<<<<<<<
 *     missing = len(names)
 *     if missing == 1:
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1309, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_argnames)) || PyTuple_CheckExact(__pyx_v_argnames)) {
      __pyx_t_2 = __pyx_v_argnames; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_argnames); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1309, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1309, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1309, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1309, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1309, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1309, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1309, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_name, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr4__pyx_v_name, __pyx_v_values, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1309, __pyx_L5_error)
      __pyx_t_7 = (__pyx_t_6 != 0);
      if (__pyx_t_7) {
        __pyx_t_5 = PyObject_Repr(__pyx_8genexpr4__pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1309, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1309, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_name); __pyx_8genexpr4__pyx_v_name = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_name); __pyx_8genexpr4__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1310
 * def _missing_arguments(f_name, argnames, pos, values):
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)             # <<<<<<<<<<<<<<
 *     if missing == 1:
 *         s = names[0]
 */
  __pyx_t_3 = PyList_GET_SIZE(__pyx_v_names); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1310, __pyx_L1_error)
  __pyx_v_missing = __pyx_t_3;

  /* "Aspidites/_vendor/contracts/inspection.py":1311
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 *     if missing == 1:             # <<<<<<<<<<<<<<
 *         s = names[0]
 *     elif missing == 2:
 */
  switch (__pyx_v_missing) {
    case 1:

    /* "Aspidites/_vendor/contracts/inspection.py":1312
 *     missing = len(names)
 *     if missing == 1:
 *         s = names[0]             # <<<<<<<<<<<<<<
 *     elif missing == 2:
 *         s = "{} and {}".format(*names)
 */
    __pyx_t_1 = PyList_GET_ITEM(__pyx_v_names, 0);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_s = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1311
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 *     if missing == 1:             # <<<<<<<<<<<<<<
 *         s = names[0]
 *     elif missing == 2:
 */
    break;
    case 2:

    /* "Aspidites/_vendor/contracts/inspection.py":1314
 *         s = names[0]
 *     elif missing == 2:
 *         s = "{} and {}".format(*names)             # <<<<<<<<<<<<<<
 *     else:
 *         tail = ", {} and {}".format(*names[-2:])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_and, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_Tuple(__pyx_v_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_s = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1313
 *     if missing == 1:
 *         s = names[0]
 *     elif missing == 2:             # <<<<<<<<<<<<<<
 *         s = "{} and {}".format(*names)
 *     else:
 */
    break;
    default:

    /* "Aspidites/_vendor/contracts/inspection.py":1316
 *         s = "{} and {}".format(*names)
 *     else:
 *         tail = ", {} and {}".format(*names[-2:])             # <<<<<<<<<<<<<<
 *         del names[-2:]
 *         s = ", ".join(names) + tail
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_and_2, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyList_GetSlice(__pyx_v_names, -2L, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_tail = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1317
 *     else:
 *         tail = ", {} and {}".format(*names[-2:])
 *         del names[-2:]             # <<<<<<<<<<<<<<
 *         s = ", ".join(names) + tail
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %
 */
    if (__Pyx_PyObject_DelSlice(__pyx_v_names, -2L, 0, NULL, NULL, NULL, 1, 0, 1) < 0) __PYX_ERR(0, 1317, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1318
 *         tail = ", {} and {}".format(*names[-2:])
 *         del names[-2:]
 *         s = ", ".join(names) + tail             # <<<<<<<<<<<<<<
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %
 *                     (f_name, missing,
 */
    __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__41, __pyx_v_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_tail); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_s = __pyx_t_1;
    __pyx_t_1 = 0;
    break;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1320
 *         s = ", ".join(names) + tail
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %
 *                     (f_name, missing,             # <<<<<<<<<<<<<<
 *                      "positional" if pos else "keyword-only",
 *                      "" if missing == 1 else "s", s))
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_missing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":1321
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %
 *                     (f_name, missing,
 *                      "positional" if pos else "keyword-only",             # <<<<<<<<<<<<<<
 *                      "" if missing == 1 else "s", s))
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_pos); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1321, __pyx_L1_error)
  if (__pyx_t_7) {
    __Pyx_INCREF(__pyx_n_u_positional);
    __pyx_t_2 = __pyx_n_u_positional;
  } else {
    __Pyx_INCREF(__pyx_kp_u_keyword_only);
    __pyx_t_2 = __pyx_kp_u_keyword_only;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1322
 *                     (f_name, missing,
 *                      "positional" if pos else "keyword-only",
 *                      "" if missing == 1 else "s", s))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (((__pyx_v_missing == 1) != 0)) {
    __Pyx_INCREF(__pyx_kp_u__19);
    __pyx_t_5 = __pyx_kp_u__19;
  } else {
    __Pyx_INCREF(__pyx_n_u_s);
    __pyx_t_5 = __pyx_n_u_s;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1320
 *         s = ", ".join(names) + tail
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %
 *                     (f_name, missing,             # <<<<<<<<<<<<<<
 *                      "positional" if pos else "keyword-only",
 *                      "" if missing == 1 else "s", s))
 */
  __pyx_t_8 = PyTuple_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_f_name);
  __Pyx_GIVEREF(__pyx_v_f_name);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_f_name);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_s);
  __Pyx_GIVEREF(__pyx_v_s);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_v_s);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1319
 *         del names[-2:]
 *         s = ", ".join(names) + tail
 *     raise TypeError("%s() missing %i required %s argument%s: %s" %             # <<<<<<<<<<<<<<
 *                     (f_name, missing,
 *                      "positional" if pos else "keyword-only",
 */
  __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_s_missing_i_required_s_argument, __pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_Raise(__pyx_t_8, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __PYX_ERR(0, 1319, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1308
 * 
 * 
 * def _missing_arguments(f_name, argnames, pos, values):             # <<<<<<<<<<<<<<
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._missing_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_tail);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1325
 * 
 * 
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):             # <<<<<<<<<<<<<<
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_107_too_many(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_107_too_many = {"_too_many", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_107_too_many, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_107_too_many(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_f_name = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwonly = 0;
  PyObject *__pyx_v_varargs = 0;
  PyObject *__pyx_v_defcount = 0;
  PyObject *__pyx_v_given = 0;
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_too_many (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f_name,&__pyx_n_s_args,&__pyx_n_s_kwonly,&__pyx_n_s_varargs,&__pyx_n_s_defcount,&__pyx_n_s_given,&__pyx_n_s_values,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_f_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 1); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwonly)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 2); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_varargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 3); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defcount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 4); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_given)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 5); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, 6); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_too_many") < 0)) __PYX_ERR(0, 1325, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_f_name = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwonly = values[2];
    __pyx_v_varargs = values[3];
    __pyx_v_defcount = values[4];
    __pyx_v_given = values[5];
    __pyx_v_values = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_too_many", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._too_many", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_106_too_many(__pyx_self, __pyx_v_f_name, __pyx_v_args, __pyx_v_kwonly, __pyx_v_varargs, __pyx_v_defcount, __pyx_v_given, __pyx_v_values);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_106_too_many(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f_name, PyObject *__pyx_v_args, PyObject *__pyx_v_kwonly, PyObject *__pyx_v_varargs, PyObject *__pyx_v_defcount, PyObject *__pyx_v_given, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_atleast = NULL;
  Py_ssize_t __pyx_v_kwonly_given;
  PyObject *__pyx_v_plural = NULL;
  PyObject *__pyx_v_sig = NULL;
  PyObject *__pyx_v_kwonly_sig = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_too_many", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1326
 * 
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):
 *     atleast = len(args) - defcount             # <<<<<<<<<<<<<<
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 *     if varargs:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1326, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_defcount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_atleast = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1327
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])             # <<<<<<<<<<<<<<
 *     if varargs:
 *         plural = atleast != 1
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1327, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_v_kwonly)) || PyTuple_CheckExact(__pyx_v_kwonly)) {
      __pyx_t_2 = __pyx_v_kwonly; __Pyx_INCREF(__pyx_t_2); __pyx_t_1 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_1 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_kwonly); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1327, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1327, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 1327, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1327, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 1327, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1327, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1327, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_arg, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr5__pyx_v_arg, __pyx_v_values, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1327, __pyx_L5_error)
      __pyx_t_7 = (__pyx_t_6 != 0);
      if (__pyx_t_7) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_8genexpr5__pyx_v_arg))) __PYX_ERR(0, 1327, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_arg); __pyx_8genexpr5__pyx_v_arg = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_arg); __pyx_8genexpr5__pyx_v_arg = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_t_1 = PyList_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_kwonly_given = __pyx_t_1;

  /* "Aspidites/_vendor/contracts/inspection.py":1328
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 *     if varargs:             # <<<<<<<<<<<<<<
 *         plural = atleast != 1
 *         sig = "at least %d" % (atleast,)
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_varargs); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1328, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":1329
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 *     if varargs:
 *         plural = atleast != 1             # <<<<<<<<<<<<<<
 *         sig = "at least %d" % (atleast,)
 *     elif defcount:
 */
    __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_v_atleast, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_plural = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1330
 *     if varargs:
 *         plural = atleast != 1
 *         sig = "at least %d" % (atleast,)             # <<<<<<<<<<<<<<
 *     elif defcount:
 *         plural = True
 */
    __pyx_t_3 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_v_atleast), __pyx_n_u_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_at_least, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_sig = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1328
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 *     if varargs:             # <<<<<<<<<<<<<<
 *         plural = atleast != 1
 *         sig = "at least %d" % (atleast,)
 */
    goto __pyx_L10;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1331
 *         plural = atleast != 1
 *         sig = "at least %d" % (atleast,)
 *     elif defcount:             # <<<<<<<<<<<<<<
 *         plural = True
 *         sig = "from %d to %d" % (atleast, len(args))
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_defcount); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1331, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":1332
 *         sig = "at least %d" % (atleast,)
 *     elif defcount:
 *         plural = True             # <<<<<<<<<<<<<<
 *         sig = "from %d to %d" % (atleast, len(args))
 *     else:
 */
    __Pyx_INCREF(Py_True);
    __pyx_v_plural = Py_True;

    /* "Aspidites/_vendor/contracts/inspection.py":1333
 *     elif defcount:
 *         plural = True
 *         sig = "from %d to %d" % (atleast, len(args))             # <<<<<<<<<<<<<<
 *     else:
 *         plural = len(args) != 1
 */
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_from);
    __pyx_t_1 += 5;
    __Pyx_GIVEREF(__pyx_kp_u_from);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_from);
    __pyx_t_3 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_v_atleast), __pyx_n_u_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_to);
    __pyx_t_1 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_to);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_to);
    __pyx_t_9 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1333, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_9, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_sig = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1331
 *         plural = atleast != 1
 *         sig = "at least %d" % (atleast,)
 *     elif defcount:             # <<<<<<<<<<<<<<
 *         plural = True
 *         sig = "from %d to %d" % (atleast, len(args))
 */
    goto __pyx_L10;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1335
 *         sig = "from %d to %d" % (atleast, len(args))
 *     else:
 *         plural = len(args) != 1             # <<<<<<<<<<<<<<
 *         sig = str(len(args))
 *     kwonly_sig = ""
 */
  /*else*/ {
    __pyx_t_1 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1335, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_t_1 != 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_plural = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1336
 *     else:
 *         plural = len(args) != 1
 *         sig = str(len(args))             # <<<<<<<<<<<<<<
 *     kwonly_sig = ""
 *     if kwonly_given:
 */
    __pyx_t_1 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1336, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_sig = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L10:;

  /* "Aspidites/_vendor/contracts/inspection.py":1337
 *         plural = len(args) != 1
 *         sig = str(len(args))
 *     kwonly_sig = ""             # <<<<<<<<<<<<<<
 *     if kwonly_given:
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 */
  __Pyx_INCREF(__pyx_kp_u__19);
  __pyx_v_kwonly_sig = __pyx_kp_u__19;

  /* "Aspidites/_vendor/contracts/inspection.py":1338
 *         sig = str(len(args))
 *     kwonly_sig = ""
 *     if kwonly_given:             # <<<<<<<<<<<<<<
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 */
  __pyx_t_7 = (__pyx_v_kwonly_given != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/contracts/inspection.py":1339
 *     kwonly_sig = ""
 *     if kwonly_given:
 *         msg = " positional argument%s (and %d keyword-only argument%s)"             # <<<<<<<<<<<<<<
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 *                              "s" if kwonly_given != 1 else ""))
 */
    __Pyx_INCREF(__pyx_kp_u_positional_argument_s_and_d_key);
    __pyx_v_msg = __pyx_kp_u_positional_argument_s_and_d_key;

    /* "Aspidites/_vendor/contracts/inspection.py":1340
 *     if kwonly_given:
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,             # <<<<<<<<<<<<<<
 *                              "s" if kwonly_given != 1 else ""))
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %
 */
    __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_v_given, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_n_u_s);
      __pyx_t_2 = __pyx_n_u_s;
    } else {
      __Pyx_INCREF(__pyx_kp_u__19);
      __pyx_t_2 = __pyx_kp_u__19;
    }
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_kwonly_given); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/contracts/inspection.py":1341
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 *                              "s" if kwonly_given != 1 else ""))             # <<<<<<<<<<<<<<
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %
 *                     (f_name, sig, "s" if plural else "", given, kwonly_sig,
 */
    if (((__pyx_v_kwonly_given != 1) != 0)) {
      __Pyx_INCREF(__pyx_n_u_s);
      __pyx_t_5 = __pyx_n_u_s;
    } else {
      __Pyx_INCREF(__pyx_kp_u__19);
      __pyx_t_5 = __pyx_kp_u__19;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1340
 *     if kwonly_given:
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,             # <<<<<<<<<<<<<<
 *                              "s" if kwonly_given != 1 else ""))
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %
 */
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyUnicode_Format(__pyx_v_msg, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF_SET(__pyx_v_kwonly_sig, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1338
 *         sig = str(len(args))
 *     kwonly_sig = ""
 *     if kwonly_given:             # <<<<<<<<<<<<<<
 *         msg = " positional argument%s (and %d keyword-only argument%s)"
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1342
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 *                              "s" if kwonly_given != 1 else ""))
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %             # <<<<<<<<<<<<<<
 *                     (f_name, sig, "s" if plural else "", given, kwonly_sig,
 *                      "was" if given == 1 and not kwonly_given else "were"))
 */
  __pyx_t_5 = PyTuple_New(11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_8 = 127;

  /* "Aspidites/_vendor/contracts/inspection.py":1343
 *                              "s" if kwonly_given != 1 else ""))
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %
 *                     (f_name, sig, "s" if plural else "", given, kwonly_sig,             # <<<<<<<<<<<<<<
 *                      "was" if given == 1 and not kwonly_given else "were"))
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_f_name), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_INCREF(__pyx_kp_u_takes);
  __pyx_t_1 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_takes);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_takes);
  __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_sig), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_INCREF(__pyx_kp_u_positional_argument);
  __pyx_t_1 += 20;
  __Pyx_GIVEREF(__pyx_kp_u_positional_argument);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u_positional_argument);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_plural); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1343, __pyx_L1_error)
  if (__pyx_t_7) {
    __Pyx_INCREF(__pyx_n_u_s);
    __pyx_t_10 = __pyx_n_u_s;
  } else {
    __Pyx_INCREF(__pyx_kp_u__19);
    __pyx_t_10 = __pyx_kp_u__19;
  }
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_but);
  __pyx_t_1 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_but);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u_but);
  __pyx_t_3 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_v_given), __pyx_n_u_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_kwonly_sig);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_kwonly_sig) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_kwonly_sig) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_kwonly_sig);
  __Pyx_GIVEREF(__pyx_v_kwonly_sig);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_v_kwonly_sig);
  __Pyx_INCREF(__pyx_kp_u__44);
  __pyx_t_1 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__44);
  PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_kp_u__44);

  /* "Aspidites/_vendor/contracts/inspection.py":1344
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %
 *                     (f_name, sig, "s" if plural else "", given, kwonly_sig,
 *                      "was" if given == 1 and not kwonly_given else "were"))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_EqObjC(__pyx_v_given, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_7 = __pyx_t_6;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_6 = ((!(__pyx_v_kwonly_given != 0)) != 0);
  __pyx_t_7 = __pyx_t_6;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_7) {
    __Pyx_INCREF(__pyx_n_u_was);
    __pyx_t_3 = __pyx_n_u_was;
  } else {
    __Pyx_INCREF(__pyx_n_u_were);
    __pyx_t_3 = __pyx_n_u_were;
  }
  __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_8;
  __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_5, 9, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_INCREF(__pyx_kp_u_given_2);
  __pyx_t_1 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_given_2);
  PyTuple_SET_ITEM(__pyx_t_5, 10, __pyx_kp_u_given_2);

  /* "Aspidites/_vendor/contracts/inspection.py":1342
 *         kwonly_sig = (msg % ("s" if given != 1 else "", kwonly_given,
 *                              "s" if kwonly_given != 1 else ""))
 *     raise TypeError("%s() takes %s positional argument%s but %d%s %s given" %             # <<<<<<<<<<<<<<
 *                     (f_name, sig, "s" if plural else "", given, kwonly_sig,
 *                      "was" if given == 1 and not kwonly_given else "were"))
 */
  __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_5, 11, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_Raise(__pyx_t_5, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __PYX_ERR(0, 1342, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1325
 * 
 * 
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):             # <<<<<<<<<<<<<<
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._too_many", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XDECREF(__pyx_v_atleast);
  __Pyx_XDECREF(__pyx_v_plural);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_kwonly_sig);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_arg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_109getcallargs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_109getcallargs = {"getcallargs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_109getcallargs, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_109getcallargs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_positional = 0;
  PyObject *__pyx_v_named = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getcallargs (wrapper)", 0);
  __pyx_v_named = PyDict_New(); if (unlikely(!__pyx_v_named)) return NULL;
  __Pyx_GOTREF(__pyx_v_named);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_positional = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_positional)) {
      __Pyx_DECREF(__pyx_v_named); __pyx_v_named = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_positional);
  } else {
    __pyx_v_positional = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func_2,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_named, values, used_pos_args, "getcallargs") < 0)) __PYX_ERR(0, 1347, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getcallargs", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1347, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_positional); __pyx_v_positional = 0;
  __Pyx_DECREF(__pyx_v_named); __pyx_v_named = 0;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_108getcallargs(__pyx_self, __pyx_v_func, __pyx_v_positional, __pyx_v_named);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_positional);
  __Pyx_XDECREF(__pyx_v_named);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1383
 *     assigned_tuple_params = []
 * 
 *     def assign(arg, value):             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         arg: cython.p_char
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_1assign(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_1assign = {"assign", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_1assign, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_1assign(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("assign (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s_value,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("assign", 1, 2, 2, 1); __PYX_ERR(0, 1383, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "assign") < 0)) __PYX_ERR(0, 1383, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v_value = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("assign", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1383, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs.assign", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_self, __pyx_v_arg, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(PyObject *__pyx_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_value) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_outer_scope;
  int __pyx_v_i;
  PyObject *__pyx_v_subarg = NULL;
  PyObject *__pyx_v_subvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_INCREF(__pyx_v_value);

  /* "Aspidites/_vendor/contracts/inspection.py":1389
 *         i: cython.int
 *         nonlocal assigned_tuple_params
 *         if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *             arg2value[arg] = value
 *         else:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_arg); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1390
 *         nonlocal assigned_tuple_params
 *         if isinstance(arg, str):
 *             arg2value[arg] = value             # <<<<<<<<<<<<<<
 *         else:
 *             assigned_tuple_params.append(arg)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_arg2value)) { __Pyx_RaiseClosureNameError("arg2value"); __PYX_ERR(0, 1390, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_arg2value == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1390, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_arg2value, __pyx_v_arg, __pyx_v_value) < 0)) __PYX_ERR(0, 1390, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1389
 *         i: cython.int
 *         nonlocal assigned_tuple_params
 *         if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *             arg2value[arg] = value
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1392
 *             arg2value[arg] = value
 *         else:
 *             assigned_tuple_params.append(arg)             # <<<<<<<<<<<<<<
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):
 */
  /*else*/ {
    if (unlikely(!__pyx_cur_scope->__pyx_v_assigned_tuple_params)) { __Pyx_RaiseClosureNameError("assigned_tuple_params"); __PYX_ERR(0, 1392, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_assigned_tuple_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 1392, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_assigned_tuple_params, __pyx_v_arg); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1392, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1393
 *         else:
 *             assigned_tuple_params.append(arg)
 *             value = iter(value)             # <<<<<<<<<<<<<<
 *             for i, subarg in enumerate(arg):
 *                 try:
 */
    __pyx_t_4 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1394
 *             assigned_tuple_params.append(arg)
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):             # <<<<<<<<<<<<<<
 *                 try:
 *                     subvalue = next(value)
 */
    __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_v_arg)) || PyTuple_CheckExact(__pyx_v_arg)) {
      __pyx_t_4 = __pyx_v_arg; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_arg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1394, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1394, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1394, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1394, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_subarg, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_v_i = __pyx_t_5;
      __pyx_t_5 = (__pyx_t_5 + 1);

      /* "Aspidites/_vendor/contracts/inspection.py":1395
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):
 *                 try:             # <<<<<<<<<<<<<<
 *                     subvalue = next(value)
 *                 except StopIteration:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":1396
 *             for i, subarg in enumerate(arg):
 *                 try:
 *                     subvalue = next(value)             # <<<<<<<<<<<<<<
 *                 except StopIteration:
 *                     raise ValueError('need more than %d %s to unpack' %
 */
          __pyx_t_8 = __Pyx_PyIter_Next(__pyx_v_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1396, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_XDECREF_SET(__pyx_v_subvalue, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1395
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):
 *                 try:             # <<<<<<<<<<<<<<
 *                     subvalue = next(value)
 *                 except StopIteration:
 */
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L13_try_end;
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1397
 *                 try:
 *                     subvalue = next(value)
 *                 except StopIteration:             # <<<<<<<<<<<<<<
 *                     raise ValueError('need more than %d %s to unpack' %
 *                                      (i, 'values' if i > 1 else 'value'))
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
        if (__pyx_t_12) {
          __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs.assign", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_13, &__pyx_t_14) < 0) __PYX_ERR(0, 1397, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GOTREF(__pyx_t_14);

          /* "Aspidites/_vendor/contracts/inspection.py":1398
 *                     subvalue = next(value)
 *                 except StopIteration:
 *                     raise ValueError('need more than %d %s to unpack' %             # <<<<<<<<<<<<<<
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)
 */
          __pyx_t_15 = PyTuple_New(5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1398, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = 0;
          __pyx_t_17 = 127;
          __Pyx_INCREF(__pyx_kp_u_need_more_than);
          __pyx_t_16 += 15;
          __Pyx_GIVEREF(__pyx_kp_u_need_more_than);
          PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_kp_u_need_more_than);

          /* "Aspidites/_vendor/contracts/inspection.py":1399
 *                 except StopIteration:
 *                     raise ValueError('need more than %d %s to unpack' %
 *                                      (i, 'values' if i > 1 else 'value'))             # <<<<<<<<<<<<<<
 *                 assign(subarg, subvalue)
 *             try:
 */
          __pyx_t_18 = __Pyx_PyUnicode_From_int(__pyx_v_i, 0, ' ', 'd'); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1399, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_16 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18);
          __Pyx_GIVEREF(__pyx_t_18);
          PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_18);
          __pyx_t_18 = 0;
          __Pyx_INCREF(__pyx_kp_u__44);
          __pyx_t_16 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__44);
          PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_kp_u__44);
          if (((__pyx_v_i > 1) != 0)) {
            __Pyx_INCREF(__pyx_n_u_values);
            __pyx_t_18 = __pyx_n_u_values;
          } else {
            __Pyx_INCREF(__pyx_n_u_value);
            __pyx_t_18 = __pyx_n_u_value;
          }
          __pyx_t_19 = __Pyx_PyUnicode_Unicode(__pyx_t_18); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1399, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) : __pyx_t_17;
          __pyx_t_16 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_19);
          __Pyx_GIVEREF(__pyx_t_19);
          PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_19);
          __pyx_t_19 = 0;
          __Pyx_INCREF(__pyx_kp_u_to_unpack);
          __pyx_t_16 += 10;
          __Pyx_GIVEREF(__pyx_kp_u_to_unpack);
          PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_kp_u_to_unpack);

          /* "Aspidites/_vendor/contracts/inspection.py":1398
 *                     subvalue = next(value)
 *                 except StopIteration:
 *                     raise ValueError('need more than %d %s to unpack' %             # <<<<<<<<<<<<<<
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)
 */
          __pyx_t_19 = __Pyx_PyUnicode_Join(__pyx_t_15, 5, __pyx_t_16, __pyx_t_17); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1398, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_19); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1398, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_Raise(__pyx_t_15, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __PYX_ERR(0, 1398, __pyx_L8_except_error)
        }
        goto __pyx_L8_except_error;
        __pyx_L8_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":1395
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):
 *                 try:             # <<<<<<<<<<<<<<
 *                     subvalue = next(value)
 *                 except StopIteration:
 */
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L13_try_end:;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1400
 *                     raise ValueError('need more than %d %s to unpack' %
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)             # <<<<<<<<<<<<<<
 *             try:
 *                 next(value)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_assign)) { __Pyx_RaiseClosureNameError("assign"); __PYX_ERR(0, 1400, __pyx_L1_error) }
      __pyx_t_14 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_subarg, __pyx_v_subvalue); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1394
 *             assigned_tuple_params.append(arg)
 *             value = iter(value)
 *             for i, subarg in enumerate(arg):             # <<<<<<<<<<<<<<
 *                 try:
 *                     subvalue = next(value)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1401
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)
 *             try:             # <<<<<<<<<<<<<<
 *                 next(value)
 *             except StopIteration:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1402
 *                 assign(subarg, subvalue)
 *             try:
 *                 next(value)             # <<<<<<<<<<<<<<
 *             except StopIteration:
 *                 pass
 */
        __pyx_t_4 = __Pyx_PyIter_Next(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1402, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1401
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)
 *             try:             # <<<<<<<<<<<<<<
 *                 next(value)
 *             except StopIteration:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1406
 *                 pass
 *             else:
 *                 raise ValueError('too many values to unpack')             # <<<<<<<<<<<<<<
 * 
 *     def is_assigned(arg):
 */
      /*else:*/ {
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1406, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1406, __pyx_L18_except_error)
      }
      __pyx_L16_error:;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1403
 *             try:
 *                 next(value)
 *             except StopIteration:             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
      if (__pyx_t_5) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L17_exception_handled;
      }
      goto __pyx_L18_except_error;
      __pyx_L18_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":1401
 *                                      (i, 'values' if i > 1 else 'value'))
 *                 assign(subarg, subvalue)
 *             try:             # <<<<<<<<<<<<<<
 *                 next(value)
 *             except StopIteration:
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L17_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
    }
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":1383
 *     assigned_tuple_params = []
 * 
 *     def assign(arg, value):             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         arg: cython.p_char
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs.assign", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subarg);
  __Pyx_XDECREF(__pyx_v_subvalue);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1408
 *                 raise ValueError('too many values to unpack')
 * 
 *     def is_assigned(arg):             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str):
 *             return arg in arg2value
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_3is_assigned(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_3is_assigned = {"is_assigned", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_3is_assigned, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_3is_assigned(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_assigned (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_assigned", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":1409
 * 
 *     def is_assigned(arg):
 *         if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *             return arg in arg2value
 *         return arg in assigned_tuple_params
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_arg); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1410
 *     def is_assigned(arg):
 *         if isinstance(arg, str):
 *             return arg in arg2value             # <<<<<<<<<<<<<<
 *         return arg in assigned_tuple_params
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_arg2value)) { __Pyx_RaiseClosureNameError("arg2value"); __PYX_ERR(0, 1410, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_arg2value == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1410, __pyx_L1_error)
    }
    __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_arg, __pyx_cur_scope->__pyx_v_arg2value, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1409
 * 
 *     def is_assigned(arg):
 *         if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *             return arg in arg2value
 *         return arg in assigned_tuple_params
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1411
 *         if isinstance(arg, str):
 *             return arg in arg2value
 *         return arg in assigned_tuple_params             # <<<<<<<<<<<<<<
 * 
 *     im_self = getattr(func, '__self__', None)
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_assigned_tuple_params)) { __Pyx_RaiseClosureNameError("assigned_tuple_params"); __PYX_ERR(0, 1411, __pyx_L1_error) }
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_arg, __pyx_cur_scope->__pyx_v_assigned_tuple_params, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1408
 *                 raise ValueError('too many values to unpack')
 * 
 *     def is_assigned(arg):             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str):
 *             return arg in arg2value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs.is_assigned", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_108getcallargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_positional, PyObject *__pyx_v_named) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_cur_scope;
  PyObject *__pyx_v_f_name = 0;
  int __pyx_v_num_args;
  int __pyx_v_num_pos;
  int __pyx_v_num_total;
  int __pyx_v_num_defaults;
  int __pyx_v_num_required;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_varargs = NULL;
  PyObject *__pyx_v_varkw = NULL;
  PyObject *__pyx_v_defaults = NULL;
  PyObject *__pyx_v_kwonlyargs = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_kwonlydefaults = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_annotations = NULL;
  PyObject *__pyx_v_is_assigned = 0;
  PyObject *__pyx_v_im_self = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_unexpected = NULL;
  Py_ssize_t __pyx_v_unassigned;
  PyObject *__pyx_8genexpr6__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  Py_UCS4 __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getcallargs", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1347, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_INCREF(__pyx_v_positional);

  /* "Aspidites/_vendor/contracts/inspection.py":1370
 *     """
 *     args, varargs, varkw, defaults, \
 *     kwonlyargs, kwonlydefaults, annotations = getfullargspec(func)             # <<<<<<<<<<<<<<
 * 
 *     if kwonlyargs:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getfullargspec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 7)) {
      if (size > 7) __Pyx_RaiseTooManyValuesError(7);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1369, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 6); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 6); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[7] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8};
      for (i=0; i < 7; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 1369, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[7] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8};
    __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
    for (index=0; index < 7; index++) {
      PyObject* item = __pyx_t_10(__pyx_t_9); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 7) < 0) __PYX_ERR(0, 1369, __pyx_L1_error)
    __pyx_t_10 = NULL;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1369, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1369
 *     values from 'positional' and 'named'.
 *     """
 *     args, varargs, varkw, defaults, \             # <<<<<<<<<<<<<<
 *     kwonlyargs, kwonlydefaults, annotations = getfullargspec(func)
 * 
 */
  __pyx_v_args = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_varargs = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_varkw = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_defaults = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_kwonlyargs = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_kwonlydefaults = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_annotations = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1372
 *     kwonlyargs, kwonlydefaults, annotations = getfullargspec(func)
 * 
 *     if kwonlyargs:             # <<<<<<<<<<<<<<
 *         raise ValueError("I'm sorry, I don't have the logic to use kwonlyargs. "
 *                          "Perhapse you can help PyContracts and implement this? Thanks.")
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_kwonlyargs); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1372, __pyx_L1_error)
  if (unlikely(__pyx_t_11)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1373
 * 
 *     if kwonlyargs:
 *         raise ValueError("I'm sorry, I don't have the logic to use kwonlyargs. "             # <<<<<<<<<<<<<<
 *                          "Perhapse you can help PyContracts and implement this? Thanks.")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1373, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1372
 *     kwonlyargs, kwonlydefaults, annotations = getfullargspec(func)
 * 
 *     if kwonlyargs:             # <<<<<<<<<<<<<<
 *         raise ValueError("I'm sorry, I don't have the logic to use kwonlyargs. "
 *                          "Perhapse you can help PyContracts and implement this? Thanks.")
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1376
 *                          "Perhapse you can help PyContracts and implement this? Thanks.")
 * 
 *     f_name = func.__name__             # <<<<<<<<<<<<<<
 *     arg2value = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1376, __pyx_L1_error)
  __pyx_v_f_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1377
 * 
 *     f_name = func.__name__
 *     arg2value = {}             # <<<<<<<<<<<<<<
 * 
 *     # The following closures are basically because of tuple
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_arg2value = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1381
 *     # The following closures are basically because of tuple
 *     # parameter unpacking.
 *     assigned_tuple_params = []             # <<<<<<<<<<<<<<
 * 
 *     def assign(arg, value):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_assigned_tuple_params = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1383
 *     assigned_tuple_params = []
 * 
 *     def assign(arg, value):             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         arg: cython.p_char
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_1assign, 0, __pyx_n_s_getcallargs_locals_assign, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_assign = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1408
 *                 raise ValueError('too many values to unpack')
 * 
 *     def is_assigned(arg):             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str):
 *             return arg in arg2value
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_3is_assigned, 0, __pyx_n_s_getcallargs_locals_is_assigned, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_is_assigned = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1413
 *         return arg in assigned_tuple_params
 * 
 *     im_self = getattr(func, '__self__', None)             # <<<<<<<<<<<<<<
 * 
 *     if ismethod(func) and im_self is not None:
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_func, __pyx_n_u_self, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_im_self = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1415
 *     im_self = getattr(func, '__self__', None)
 * 
 *     if ismethod(func) and im_self is not None:             # <<<<<<<<<<<<<<
 *         # implicit 'self' (or 'cls' for classmethods) argument
 *         positional = (im_self,) + positional
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_12 = (__pyx_v_im_self != Py_None);
  __pyx_t_13 = (__pyx_t_12 != 0);
  __pyx_t_11 = __pyx_t_13;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":1417
 *     if ismethod(func) and im_self is not None:
 *         # implicit 'self' (or 'cls' for classmethods) argument
 *         positional = (im_self,) + positional             # <<<<<<<<<<<<<<
 *     num_pos = len(positional)
 *     num_total = num_pos + len(named)
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_im_self);
    __Pyx_GIVEREF(__pyx_v_im_self);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_im_self);
    __pyx_t_8 = PyNumber_Add(__pyx_t_1, __pyx_v_positional); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_positional, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1415
 *     im_self = getattr(func, '__self__', None)
 * 
 *     if ismethod(func) and im_self is not None:             # <<<<<<<<<<<<<<
 *         # implicit 'self' (or 'cls' for classmethods) argument
 *         positional = (im_self,) + positional
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1418
 *         # implicit 'self' (or 'cls' for classmethods) argument
 *         positional = (im_self,) + positional
 *     num_pos = len(positional)             # <<<<<<<<<<<<<<
 *     num_total = num_pos + len(named)
 *     num_args = len(args)
 */
  __pyx_t_14 = PyTuple_GET_SIZE(__pyx_v_positional); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1418, __pyx_L1_error)
  __pyx_v_num_pos = __pyx_t_14;

  /* "Aspidites/_vendor/contracts/inspection.py":1419
 *         positional = (im_self,) + positional
 *     num_pos = len(positional)
 *     num_total = num_pos + len(named)             # <<<<<<<<<<<<<<
 *     num_args = len(args)
 *     num_defaults = len(defaults) if defaults else 0
 */
  __pyx_t_14 = PyDict_Size(__pyx_v_named); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1419, __pyx_L1_error)
  __pyx_v_num_total = (__pyx_v_num_pos + __pyx_t_14);

  /* "Aspidites/_vendor/contracts/inspection.py":1420
 *     num_pos = len(positional)
 *     num_total = num_pos + len(named)
 *     num_args = len(args)             # <<<<<<<<<<<<<<
 *     num_defaults = len(defaults) if defaults else 0
 *     for arg, value in zip(args, positional):
 */
  __pyx_t_14 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1420, __pyx_L1_error)
  __pyx_v_num_args = __pyx_t_14;

  /* "Aspidites/_vendor/contracts/inspection.py":1421
 *     num_total = num_pos + len(named)
 *     num_args = len(args)
 *     num_defaults = len(defaults) if defaults else 0             # <<<<<<<<<<<<<<
 *     for arg, value in zip(args, positional):
 *         assign(arg, value)
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1421, __pyx_L1_error)
  if (__pyx_t_11) {
    __pyx_t_15 = PyObject_Length(__pyx_v_defaults); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1421, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_15;
  } else {
    __pyx_t_14 = 0;
  }
  __pyx_v_num_defaults = __pyx_t_14;

  /* "Aspidites/_vendor/contracts/inspection.py":1422
 *     num_args = len(args)
 *     num_defaults = len(defaults) if defaults else 0
 *     for arg, value in zip(args, positional):             # <<<<<<<<<<<<<<
 *         assign(arg, value)
 *     if varargs:
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_args);
  __Pyx_INCREF(__pyx_v_positional);
  __Pyx_GIVEREF(__pyx_v_positional);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_positional);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_8 = __pyx_t_1; __Pyx_INCREF(__pyx_t_8); __pyx_t_14 = 0;
    __pyx_t_16 = NULL;
  } else {
    __pyx_t_14 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1422, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_16)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1422, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1422, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1422, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1422, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_16(__pyx_t_8);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1422, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1422, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_5); if (unlikely(!__pyx_t_7)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_6 = __pyx_t_10(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1422, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1422, __pyx_L1_error)
      __pyx_L12_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1423
 *     num_defaults = len(defaults) if defaults else 0
 *     for arg, value in zip(args, positional):
 *         assign(arg, value)             # <<<<<<<<<<<<<<
 *     if varargs:
 *         if num_pos > num_args:
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_arg, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1422
 *     num_args = len(args)
 *     num_defaults = len(defaults) if defaults else 0
 *     for arg, value in zip(args, positional):             # <<<<<<<<<<<<<<
 *         assign(arg, value)
 *     if varargs:
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1424
 *     for arg, value in zip(args, positional):
 *         assign(arg, value)
 *     if varargs:             # <<<<<<<<<<<<<<
 *         if num_pos > num_args:
 *             assign(varargs, positional[-(num_pos - num_args):])
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_varargs); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1424, __pyx_L1_error)
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":1425
 *         assign(arg, value)
 *     if varargs:
 *         if num_pos > num_args:             # <<<<<<<<<<<<<<
 *             assign(varargs, positional[-(num_pos - num_args):])
 *         else:
 */
    __pyx_t_11 = ((__pyx_v_num_pos > __pyx_v_num_args) != 0);
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":1426
 *     if varargs:
 *         if num_pos > num_args:
 *             assign(varargs, positional[-(num_pos - num_args):])             # <<<<<<<<<<<<<<
 *         else:
 *             assign(varargs, ())
 */
      __pyx_t_8 = __Pyx_PyTuple_GetSlice(__pyx_v_positional, (-(__pyx_v_num_pos - __pyx_v_num_args)), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_varargs, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1425
 *         assign(arg, value)
 *     if varargs:
 *         if num_pos > num_args:             # <<<<<<<<<<<<<<
 *             assign(varargs, positional[-(num_pos - num_args):])
 *         else:
 */
      goto __pyx_L14;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1428
 *             assign(varargs, positional[-(num_pos - num_args):])
 *         else:
 *             assign(varargs, ())             # <<<<<<<<<<<<<<
 *     elif 0 < num_args < num_pos:
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 */
    /*else*/ {
      __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_varargs, __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L14:;

    /* "Aspidites/_vendor/contracts/inspection.py":1424
 *     for arg, value in zip(args, positional):
 *         assign(arg, value)
 *     if varargs:             # <<<<<<<<<<<<<<
 *         if num_pos > num_args:
 *             assign(varargs, positional[-(num_pos - num_args):])
 */
    goto __pyx_L13;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1429
 *         else:
 *             assign(varargs, ())
 *     elif 0 < num_args < num_pos:             # <<<<<<<<<<<<<<
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 */
  __pyx_t_11 = (0 < __pyx_v_num_args);
  if (__pyx_t_11) {
    __pyx_t_11 = (__pyx_v_num_args < __pyx_v_num_pos);
  }
  __pyx_t_13 = (__pyx_t_11 != 0);
  if (unlikely(__pyx_t_13)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1430
 *             assign(varargs, ())
 *     elif 0 < num_args < num_pos:
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (             # <<<<<<<<<<<<<<
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 *             'arguments' if num_args > 1 else 'argument', num_total))
 */
    __pyx_t_1 = PyTuple_New(10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = 0;
    __pyx_t_17 = 127;

    /* "Aspidites/_vendor/contracts/inspection.py":1431
 *     elif 0 < num_args < num_pos:
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at most' if defaults else 'exactly', num_args,             # <<<<<<<<<<<<<<
 *             'arguments' if num_args > 1 else 'argument', num_total))
 *     elif num_args == 0 and num_total:
 */
    __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_v_f_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_takes);
    __pyx_t_14 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_takes);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_takes);
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1431, __pyx_L1_error)
    if (__pyx_t_13) {
      __Pyx_INCREF(__pyx_kp_u_at_most);
      __pyx_t_8 = __pyx_kp_u_at_most;
    } else {
      __Pyx_INCREF(__pyx_n_u_exactly);
      __pyx_t_8 = __pyx_n_u_exactly;
    }
    __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__44);
    __pyx_t_14 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__44);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__44);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_num_args, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__44);
    __pyx_t_14 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__44);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__44);

    /* "Aspidites/_vendor/contracts/inspection.py":1432
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 *             'arguments' if num_args > 1 else 'argument', num_total))             # <<<<<<<<<<<<<<
 *     elif num_args == 0 and num_total:
 *         raise TypeError('%s() takes no arguments (%d given)' %
 */
    if (((__pyx_v_num_args > 1) != 0)) {
      __Pyx_INCREF(__pyx_n_u_arguments);
      __pyx_t_6 = __pyx_n_u_arguments;
    } else {
      __Pyx_INCREF(__pyx_n_u_argument);
      __pyx_t_6 = __pyx_n_u_argument;
    }
    __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u__51);
    __pyx_t_14 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__51);
    PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u__51);
    __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_v_num_total, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_given_3);
    __pyx_t_14 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_given_3);
    PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_kp_u_given_3);

    /* "Aspidites/_vendor/contracts/inspection.py":1430
 *             assign(varargs, ())
 *     elif 0 < num_args < num_pos:
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (             # <<<<<<<<<<<<<<
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 *             'arguments' if num_args > 1 else 'argument', num_total))
 */
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_1, 10, __pyx_t_14, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1430, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1429
 *         else:
 *             assign(varargs, ())
 *     elif 0 < num_args < num_pos:             # <<<<<<<<<<<<<<
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1433
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 *             'arguments' if num_args > 1 else 'argument', num_total))
 *     elif num_args == 0 and num_total:             # <<<<<<<<<<<<<<
 *         raise TypeError('%s() takes no arguments (%d given)' %
 *                         (f_name, num_total))
 */
  __pyx_t_11 = ((__pyx_v_num_args == 0) != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_13 = __pyx_t_11;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_11 = (__pyx_v_num_total != 0);
  __pyx_t_13 = __pyx_t_11;
  __pyx_L15_bool_binop_done:;
  if (unlikely(__pyx_t_13)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1434
 *             'arguments' if num_args > 1 else 'argument', num_total))
 *     elif num_args == 0 and num_total:
 *         raise TypeError('%s() takes no arguments (%d given)' %             # <<<<<<<<<<<<<<
 *                         (f_name, num_total))
 *     for arg in args:
 */
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = 0;
    __pyx_t_17 = 127;

    /* "Aspidites/_vendor/contracts/inspection.py":1435
 *     elif num_args == 0 and num_total:
 *         raise TypeError('%s() takes no arguments (%d given)' %
 *                         (f_name, num_total))             # <<<<<<<<<<<<<<
 *     for arg in args:
 *         if isinstance(arg, str) and arg in named:
 */
    __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_v_f_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_takes_no_arguments);
    __pyx_t_14 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_takes_no_arguments);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_takes_no_arguments);
    __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_v_num_total, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_given_3);
    __pyx_t_14 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_given_3);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_given_3);

    /* "Aspidites/_vendor/contracts/inspection.py":1434
 *             'arguments' if num_args > 1 else 'argument', num_total))
 *     elif num_args == 0 and num_total:
 *         raise TypeError('%s() takes no arguments (%d given)' %             # <<<<<<<<<<<<<<
 *                         (f_name, num_total))
 *     for arg in args:
 */
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_14, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1434, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1433
 *             f_name, 'at most' if defaults else 'exactly', num_args,
 *             'arguments' if num_args > 1 else 'argument', num_total))
 *     elif num_args == 0 and num_total:             # <<<<<<<<<<<<<<
 *         raise TypeError('%s() takes no arguments (%d given)' %
 *                         (f_name, num_total))
 */
  }
  __pyx_L13:;

  /* "Aspidites/_vendor/contracts/inspection.py":1436
 *         raise TypeError('%s() takes no arguments (%d given)' %
 *                         (f_name, num_total))
 *     for arg in args:             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):
 */
  if (likely(PyList_CheckExact(__pyx_v_args)) || PyTuple_CheckExact(__pyx_v_args)) {
    __pyx_t_1 = __pyx_v_args; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
    __pyx_t_16 = NULL;
  } else {
    __pyx_t_14 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_16 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1436, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_16)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_16(__pyx_t_1);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1436, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1437
 *                         (f_name, num_total))
 *     for arg in args:
 *         if isinstance(arg, str) and arg in named:             # <<<<<<<<<<<<<<
 *             if is_assigned(arg):
 *                 raise TypeError("%s() got multiple values for keyword "
 */
    __pyx_t_11 = PyUnicode_Check(__pyx_v_arg); 
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_13 = __pyx_t_12;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_12 = (__Pyx_PyDict_ContainsTF(__pyx_v_arg, __pyx_v_named, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1437, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_12 != 0);
    __pyx_t_13 = __pyx_t_11;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":1438
 *     for arg in args:
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%s() got multiple values for keyword "
 *                                 "argument '%s'" % (f_name, arg))
 */
      __pyx_t_8 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(__pyx_v_is_assigned, __pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1438, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__pyx_t_13)) {

        /* "Aspidites/_vendor/contracts/inspection.py":1439
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):
 *                 raise TypeError("%s() got multiple values for keyword "             # <<<<<<<<<<<<<<
 *                                 "argument '%s'" % (f_name, arg))
 *             else:
 */
        __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_15 = 0;
        __pyx_t_17 = 127;

        /* "Aspidites/_vendor/contracts/inspection.py":1440
 *             if is_assigned(arg):
 *                 raise TypeError("%s() got multiple values for keyword "
 *                                 "argument '%s'" % (f_name, arg))             # <<<<<<<<<<<<<<
 *             else:
 *                 assign(arg, named.pop(arg))
 */
        __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_f_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
        __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_INCREF(__pyx_kp_u_got_multiple_values_for_keyword);
        __pyx_t_15 += 45;
        __Pyx_GIVEREF(__pyx_kp_u_got_multiple_values_for_keyword);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_kp_u_got_multiple_values_for_keyword);
        __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_arg), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
        __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_INCREF(__pyx_kp_u__52);
        __pyx_t_15 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__52);
        PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_kp_u__52);

        /* "Aspidites/_vendor/contracts/inspection.py":1439
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):
 *                 raise TypeError("%s() got multiple values for keyword "             # <<<<<<<<<<<<<<
 *                                 "argument '%s'" % (f_name, arg))
 *             else:
 */
        __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_8, 4, __pyx_t_15, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 1439, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":1438
 *     for arg in args:
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%s() got multiple values for keyword "
 *                                 "argument '%s'" % (f_name, arg))
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1442
 *                                 "argument '%s'" % (f_name, arg))
 *             else:
 *                 assign(arg, named.pop(arg))             # <<<<<<<<<<<<<<
 *     if defaults:  # fill in any missing values with the defaults
 *         for arg, value in zip(args[-num_defaults:], defaults):
 */
      /*else*/ {
        __pyx_t_8 = __Pyx_PyDict_Pop(__pyx_v_named, __pyx_v_arg, ((PyObject *)NULL)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_arg, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1437
 *                         (f_name, num_total))
 *     for arg in args:
 *         if isinstance(arg, str) and arg in named:             # <<<<<<<<<<<<<<
 *             if is_assigned(arg):
 *                 raise TypeError("%s() got multiple values for keyword "
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1436
 *         raise TypeError('%s() takes no arguments (%d given)' %
 *                         (f_name, num_total))
 *     for arg in args:             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str) and arg in named:
 *             if is_assigned(arg):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1443
 *             else:
 *                 assign(arg, named.pop(arg))
 *     if defaults:  # fill in any missing values with the defaults             # <<<<<<<<<<<<<<
 *         for arg, value in zip(args[-num_defaults:], defaults):
 *             if not is_assigned(arg):
 */
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1443, __pyx_L1_error)
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":1444
 *                 assign(arg, named.pop(arg))
 *     if defaults:  # fill in any missing values with the defaults
 *         for arg, value in zip(args[-num_defaults:], defaults):             # <<<<<<<<<<<<<<
 *             if not is_assigned(arg):
 *                 assign(arg, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_args, (-__pyx_v_num_defaults), 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_defaults);
    __Pyx_GIVEREF(__pyx_v_defaults);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_defaults);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6); __pyx_t_14 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1444, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1444, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1444, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1444, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_16(__pyx_t_6);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1444, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1444, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_5)->tp_iternext;
        index = 0; __pyx_t_8 = __pyx_t_10(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L26_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        index = 1; __pyx_t_7 = __pyx_t_10(__pyx_t_5); if (unlikely(!__pyx_t_7)) goto __pyx_L26_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1444, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L27_unpacking_done;
        __pyx_L26_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1444, __pyx_L1_error)
        __pyx_L27_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1445
 *     if defaults:  # fill in any missing values with the defaults
 *         for arg, value in zip(args[-num_defaults:], defaults):
 *             if not is_assigned(arg):             # <<<<<<<<<<<<<<
 *                 assign(arg, value)
 *     if varkw:
 */
      __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(__pyx_v_is_assigned, __pyx_v_arg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1445, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_11 = ((!__pyx_t_13) != 0);
      if (__pyx_t_11) {

        /* "Aspidites/_vendor/contracts/inspection.py":1446
 *         for arg, value in zip(args[-num_defaults:], defaults):
 *             if not is_assigned(arg):
 *                 assign(arg, value)             # <<<<<<<<<<<<<<
 *     if varkw:
 *         assign(varkw, named)
 */
        __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_arg, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1446, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1445
 *     if defaults:  # fill in any missing values with the defaults
 *         for arg, value in zip(args[-num_defaults:], defaults):
 *             if not is_assigned(arg):             # <<<<<<<<<<<<<<
 *                 assign(arg, value)
 *     if varkw:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1444
 *                 assign(arg, named.pop(arg))
 *     if defaults:  # fill in any missing values with the defaults
 *         for arg, value in zip(args[-num_defaults:], defaults):             # <<<<<<<<<<<<<<
 *             if not is_assigned(arg):
 *                 assign(arg, value)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1443
 *             else:
 *                 assign(arg, named.pop(arg))
 *     if defaults:  # fill in any missing values with the defaults             # <<<<<<<<<<<<<<
 *         for arg, value in zip(args[-num_defaults:], defaults):
 *             if not is_assigned(arg):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1447
 *             if not is_assigned(arg):
 *                 assign(arg, value)
 *     if varkw:             # <<<<<<<<<<<<<<
 *         assign(varkw, named)
 *     elif named:
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_varkw); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1447, __pyx_L1_error)
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":1448
 *                 assign(arg, value)
 *     if varkw:
 *         assign(varkw, named)             # <<<<<<<<<<<<<<
 *     elif named:
 *         unexpected = next(iter(named))
 */
    __pyx_t_6 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_assign(__pyx_cur_scope->__pyx_v_assign, __pyx_v_varkw, __pyx_v_named); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1447
 *             if not is_assigned(arg):
 *                 assign(arg, value)
 *     if varkw:             # <<<<<<<<<<<<<<
 *         assign(varkw, named)
 *     elif named:
 */
    goto __pyx_L29;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1449
 *     if varkw:
 *         assign(varkw, named)
 *     elif named:             # <<<<<<<<<<<<<<
 *         unexpected = next(iter(named))
 *         if isinstance(unexpected, str):
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_named); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1449, __pyx_L1_error)
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":1450
 *         assign(varkw, named)
 *     elif named:
 *         unexpected = next(iter(named))             # <<<<<<<<<<<<<<
 *         if isinstance(unexpected, str):
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 */
    __pyx_t_6 = PyObject_GetIter(__pyx_v_named); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_unexpected = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1451
 *     elif named:
 *         unexpected = next(iter(named))
 *         if isinstance(unexpected, str):             # <<<<<<<<<<<<<<
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %
 */
    __pyx_t_11 = PyUnicode_Check(__pyx_v_unexpected); 
    __pyx_t_13 = (__pyx_t_11 != 0);
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":1452
 *         unexpected = next(iter(named))
 *         if isinstance(unexpected, str):
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')             # <<<<<<<<<<<<<<
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %
 *                         (f_name, unexpected))
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_unexpected, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_sys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_getdefaultencoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_18 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_18 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_7, __pyx_n_u_replace};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_18, 2+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_7, __pyx_n_u_replace};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_18, 2+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_18); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_18, __pyx_t_7);
        __Pyx_INCREF(__pyx_n_u_replace);
        __Pyx_GIVEREF(__pyx_n_u_replace);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_18, __pyx_n_u_replace);
        __pyx_t_7 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_unexpected, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1451
 *     elif named:
 *         unexpected = next(iter(named))
 *         if isinstance(unexpected, str):             # <<<<<<<<<<<<<<
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1453
 *         if isinstance(unexpected, str):
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %             # <<<<<<<<<<<<<<
 *                         (f_name, unexpected))
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 */
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = 0;
    __pyx_t_17 = 127;

    /* "Aspidites/_vendor/contracts/inspection.py":1454
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %
 *                         (f_name, unexpected))             # <<<<<<<<<<<<<<
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 *     if unassigned:
 */
    __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_f_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_got_an_unexpected_keyword_argum);
    __pyx_t_14 += 39;
    __Pyx_GIVEREF(__pyx_kp_u_got_an_unexpected_keyword_argum);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_got_an_unexpected_keyword_argum);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_unexpected), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__52);
    __pyx_t_14 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__52);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__52);

    /* "Aspidites/_vendor/contracts/inspection.py":1453
 *         if isinstance(unexpected, str):
 *             unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %             # <<<<<<<<<<<<<<
 *                         (f_name, unexpected))
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 */
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_14, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1453, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1449
 *     if varkw:
 *         assign(varkw, named)
 *     elif named:             # <<<<<<<<<<<<<<
 *         unexpected = next(iter(named))
 *         if isinstance(unexpected, str):
 */
  }
  __pyx_L29:;

  /* "Aspidites/_vendor/contracts/inspection.py":1455
 *         raise TypeError("%s() got an unexpected keyword argument '%s'" %
 *                         (f_name, unexpected))
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])             # <<<<<<<<<<<<<<
 *     if unassigned:
 *         num_required = num_args - num_defaults
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1455, __pyx_L33_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_args)) || PyTuple_CheckExact(__pyx_v_args)) {
      __pyx_t_6 = __pyx_v_args; __Pyx_INCREF(__pyx_t_6); __pyx_t_14 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1455, __pyx_L33_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1455, __pyx_L33_error)
    }
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1455, __pyx_L33_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1455, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1455, __pyx_L33_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1455, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_16(__pyx_t_6);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1455, __pyx_L33_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_arg, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_11getcallargs_2is_assigned(__pyx_v_is_assigned, __pyx_8genexpr6__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1455, __pyx_L33_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1455, __pyx_L33_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_13) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr6__pyx_v_arg))) __PYX_ERR(0, 1455, __pyx_L33_error)
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_arg); __pyx_8genexpr6__pyx_v_arg = 0;
    goto __pyx_L37_exit_scope;
    __pyx_L33_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_arg); __pyx_8genexpr6__pyx_v_arg = 0;
    goto __pyx_L1_error;
    __pyx_L37_exit_scope:;
  } /* exit inner scope */
  __pyx_t_14 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_unassigned = (__pyx_v_num_args - __pyx_t_14);

  /* "Aspidites/_vendor/contracts/inspection.py":1456
 *                         (f_name, unexpected))
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 *     if unassigned:             # <<<<<<<<<<<<<<
 *         num_required = num_args - num_defaults
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 */
  __pyx_t_13 = (__pyx_v_unassigned != 0);
  if (unlikely(__pyx_t_13)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1457
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 *     if unassigned:
 *         num_required = num_args - num_defaults             # <<<<<<<<<<<<<<
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at least' if defaults else 'exactly', num_required,
 */
    __pyx_v_num_required = (__pyx_v_num_args - __pyx_v_num_defaults);

    /* "Aspidites/_vendor/contracts/inspection.py":1458
 *     if unassigned:
 *         num_required = num_args - num_defaults
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (             # <<<<<<<<<<<<<<
 *             f_name, 'at least' if defaults else 'exactly', num_required,
 *             'arguments' if num_required > 1 else 'argument', num_total))
 */
    __pyx_t_1 = PyTuple_New(10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = 0;
    __pyx_t_17 = 127;

    /* "Aspidites/_vendor/contracts/inspection.py":1459
 *         num_required = num_args - num_defaults
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at least' if defaults else 'exactly', num_required,             # <<<<<<<<<<<<<<
 *             'arguments' if num_required > 1 else 'argument', num_total))
 *     return arg2value
 */
    __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_f_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_takes);
    __pyx_t_14 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_takes);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_takes);
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1459, __pyx_L1_error)
    if (__pyx_t_13) {
      __Pyx_INCREF(__pyx_kp_u_at_least_2);
      __pyx_t_6 = __pyx_kp_u_at_least_2;
    } else {
      __Pyx_INCREF(__pyx_n_u_exactly);
      __pyx_t_6 = __pyx_n_u_exactly;
    }
    __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u__44);
    __pyx_t_14 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__44);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__44);
    __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_v_num_required, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u__44);
    __pyx_t_14 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__44);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__44);

    /* "Aspidites/_vendor/contracts/inspection.py":1460
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 *             f_name, 'at least' if defaults else 'exactly', num_required,
 *             'arguments' if num_required > 1 else 'argument', num_total))             # <<<<<<<<<<<<<<
 *     return arg2value
 * 
 */
    if (((__pyx_v_num_required > 1) != 0)) {
      __Pyx_INCREF(__pyx_n_u_arguments);
      __pyx_t_8 = __pyx_n_u_arguments;
    } else {
      __Pyx_INCREF(__pyx_n_u_argument);
      __pyx_t_8 = __pyx_n_u_argument;
    }
    __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__51);
    __pyx_t_14 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__51);
    PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u__51);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_num_total, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_given_3);
    __pyx_t_14 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_given_3);
    PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_kp_u_given_3);

    /* "Aspidites/_vendor/contracts/inspection.py":1458
 *     if unassigned:
 *         num_required = num_args - num_defaults
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (             # <<<<<<<<<<<<<<
 *             f_name, 'at least' if defaults else 'exactly', num_required,
 *             'arguments' if num_required > 1 else 'argument', num_total))
 */
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_1, 10, __pyx_t_14, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1458, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1456
 *                         (f_name, unexpected))
 *     unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
 *     if unassigned:             # <<<<<<<<<<<<<<
 *         num_required = num_args - num_defaults
 *         raise TypeError('%s() takes %s %d %s (%d given)' % (
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1461
 *             f_name, 'at least' if defaults else 'exactly', num_required,
 *             'arguments' if num_required > 1 else 'argument', num_total))
 *     return arg2value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_arg2value);
  __pyx_r = __pyx_cur_scope->__pyx_v_arg2value;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcallargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positional);
  __Pyx_XDECREF(__pyx_v_f_name);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_varargs);
  __Pyx_XDECREF(__pyx_v_varkw);
  __Pyx_XDECREF(__pyx_v_defaults);
  __Pyx_XDECREF(__pyx_v_kwonlyargs);
  __Pyx_XDECREF(__pyx_v_kwonlydefaults);
  __Pyx_XDECREF(__pyx_v_annotations);
  __Pyx_XDECREF(__pyx_v_is_assigned);
  __Pyx_XDECREF(__pyx_v_im_self);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_unexpected);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_arg);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1467
 * 
 * 
 * def getclosurevars(func):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of free variables to their current values.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_111getclosurevars(PyObject *__pyx_self, PyObject *__pyx_v_func); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_110getclosurevars[] = "\n    Get the mapping of free variables to their current values.\n\n    Returns a named tuple of dicts mapping the current nonlocal, global\n    and builtin references as seen by the body of the function. A final\n    set of unbound names that could not be resolved is also provided.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_111getclosurevars = {"getclosurevars", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_111getclosurevars, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_110getclosurevars};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_111getclosurevars(PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getclosurevars (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_110getclosurevars(__pyx_self, ((PyObject *)__pyx_v_func));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_110getclosurevars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_v_code = NULL;
  PyObject *__pyx_v_nonlocal_vars = NULL;
  PyObject *__pyx_v_global_ns = NULL;
  PyObject *__pyx_v_builtin_ns = NULL;
  PyObject *__pyx_v_global_vars = NULL;
  PyObject *__pyx_v_builtin_vars = NULL;
  PyObject *__pyx_v_unbound_names = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_var = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_cell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getclosurevars", 0);
  __Pyx_INCREF(__pyx_v_func);

  /* "Aspidites/_vendor/contracts/inspection.py":1476
 *     """
 * 
 *     if ismethod(func):             # <<<<<<<<<<<<<<
 *         func = func.__func__
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1477
 * 
 *     if ismethod(func):
 *         func = func.__func__             # <<<<<<<<<<<<<<
 * 
 *     if not isfunction(func):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_func, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1476
 *     """
 * 
 *     if ismethod(func):             # <<<<<<<<<<<<<<
 *         func = func.__func__
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1479
 *         func = func.__func__
 * 
 *     if not isfunction(func):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python function".format(func))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1480
 * 
 *     if not isfunction(func):
 *         raise TypeError("{!r} is not a Python function".format(func))             # <<<<<<<<<<<<<<
 * 
 *     code = func.__code__
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_Python_function, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1480, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1479
 *         func = func.__func__
 * 
 *     if not isfunction(func):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python function".format(func))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1482
 *         raise TypeError("{!r} is not a Python function".format(func))
 * 
 *     code = func.__code__             # <<<<<<<<<<<<<<
 *     # Nonlocal references are named in co_freevars and resolved
 *     # by looking them up in __closure__ by positional index
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_code = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1485
 *     # Nonlocal references are named in co_freevars and resolved
 *     # by looking them up in __closure__ by positional index
 *     if func.__closure__ is None:             # <<<<<<<<<<<<<<
 *         nonlocal_vars = {}
 *     else:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_closure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = (__pyx_t_2 == Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1486
 *     # by looking them up in __closure__ by positional index
 *     if func.__closure__ is None:
 *         nonlocal_vars = {}             # <<<<<<<<<<<<<<
 *     else:
 *         nonlocal_vars = {
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_nonlocal_vars = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1485
 *     # Nonlocal references are named in co_freevars and resolved
 *     # by looking them up in __closure__ by positional index
 *     if func.__closure__ is None:             # <<<<<<<<<<<<<<
 *         nonlocal_vars = {}
 *     else:
 */
    goto __pyx_L5;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1488
 *         nonlocal_vars = {}
 *     else:
 *         nonlocal_vars = {             # <<<<<<<<<<<<<<
 *             var: cell.cell_contents
 *             for var, cell in zip(code.co_freevars, func.__closure__)
 */
  /*else*/ {
    { /* enter inner scope */
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1488, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/contracts/inspection.py":1490
 *         nonlocal_vars = {
 *             var: cell.cell_contents
 *             for var, cell in zip(code.co_freevars, func.__closure__)             # <<<<<<<<<<<<<<
 *         }
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_n_s_co_freevars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_closure); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1490, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1490, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1490, __pyx_L8_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1490, __pyx_L8_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1490, __pyx_L8_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1490, __pyx_L8_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
          PyObject* sequence = __pyx_t_3;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1490, __pyx_L8_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_9);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1490, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1490, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
          index = 0; __pyx_t_1 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_1)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 1490, __pyx_L8_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L12_unpacking_done;
          __pyx_L11_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1490, __pyx_L8_error)
          __pyx_L12_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_var, __pyx_t_1);
        __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_cell, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1489
 *     else:
 *         nonlocal_vars = {
 *             var: cell.cell_contents             # <<<<<<<<<<<<<<
 *             for var, cell in zip(code.co_freevars, func.__closure__)
 *         }
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr7__pyx_v_cell, __pyx_n_s_cell_contents); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1489, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_8genexpr7__pyx_v_var, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 1489, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1490
 *         nonlocal_vars = {
 *             var: cell.cell_contents
 *             for var, cell in zip(code.co_freevars, func.__closure__)             # <<<<<<<<<<<<<<
 *         }
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_cell); __pyx_8genexpr7__pyx_v_cell = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_var); __pyx_8genexpr7__pyx_v_var = 0;
      goto __pyx_L13_exit_scope;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_cell); __pyx_8genexpr7__pyx_v_cell = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_var); __pyx_8genexpr7__pyx_v_var = 0;
      goto __pyx_L1_error;
      __pyx_L13_exit_scope:;
    } /* exit inner scope */
    __pyx_v_nonlocal_vars = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L5:;

  /* "Aspidites/_vendor/contracts/inspection.py":1495
 *     # Global and builtin references are named in co_names and resolved
 *     # by looking them up in __globals__ or __builtins__
 *     global_ns = func.__globals__             # <<<<<<<<<<<<<<
 *     builtin_ns = global_ns.get("__builtins__", builtins.__dict__)
 *     if ismodule(builtin_ns):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_globals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_global_ns = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1496
 *     # by looking them up in __globals__ or __builtins__
 *     global_ns = func.__globals__
 *     builtin_ns = global_ns.get("__builtins__", builtins.__dict__)             # <<<<<<<<<<<<<<
 *     if ismodule(builtin_ns):
 *         builtin_ns = builtin_ns.__dict__
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_global_ns, __pyx_n_s_get_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_builtins); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_u_builtins_2, __pyx_t_9};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_u_builtins_2, __pyx_t_9};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_n_u_builtins_2);
    __Pyx_GIVEREF(__pyx_n_u_builtins_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_12, __pyx_n_u_builtins_2);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_12, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_builtin_ns = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1497
 *     global_ns = func.__globals__
 *     builtin_ns = global_ns.get("__builtins__", builtins.__dict__)
 *     if ismodule(builtin_ns):             # <<<<<<<<<<<<<<
 *         builtin_ns = builtin_ns.__dict__
 *     global_vars = {}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_v_builtin_ns) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_builtin_ns);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1498
 *     builtin_ns = global_ns.get("__builtins__", builtins.__dict__)
 *     if ismodule(builtin_ns):
 *         builtin_ns = builtin_ns.__dict__             # <<<<<<<<<<<<<<
 *     global_vars = {}
 *     builtin_vars = {}
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_builtin_ns, __pyx_n_s_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_builtin_ns, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1497
 *     global_ns = func.__globals__
 *     builtin_ns = global_ns.get("__builtins__", builtins.__dict__)
 *     if ismodule(builtin_ns):             # <<<<<<<<<<<<<<
 *         builtin_ns = builtin_ns.__dict__
 *     global_vars = {}
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1499
 *     if ismodule(builtin_ns):
 *         builtin_ns = builtin_ns.__dict__
 *     global_vars = {}             # <<<<<<<<<<<<<<
 *     builtin_vars = {}
 *     unbound_names = set()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_global_vars = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1500
 *         builtin_ns = builtin_ns.__dict__
 *     global_vars = {}
 *     builtin_vars = {}             # <<<<<<<<<<<<<<
 *     unbound_names = set()
 *     for name in code.co_names:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_builtin_vars = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1501
 *     global_vars = {}
 *     builtin_vars = {}
 *     unbound_names = set()             # <<<<<<<<<<<<<<
 *     for name in code.co_names:
 *         if name in ("None", "True", "False"):
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_unbound_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1502
 *     builtin_vars = {}
 *     unbound_names = set()
 *     for name in code.co_names:             # <<<<<<<<<<<<<<
 *         if name in ("None", "True", "False"):
 *             # Because these used to be builtins instead of keywords, they
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_n_s_co_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_6 = __pyx_t_2; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1502, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1502, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1502, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1502, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1502, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1502, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1503
 *     unbound_names = set()
 *     for name in code.co_names:
 *         if name in ("None", "True", "False"):             # <<<<<<<<<<<<<<
 *             # Because these used to be builtins instead of keywords, they
 *             # may still show up as name references. We ignore them.
 */
    __Pyx_INCREF(__pyx_v_name);
    __pyx_t_2 = __pyx_v_name;
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_kp_u_None, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1503, __pyx_L1_error)
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_True, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1503, __pyx_L1_error)
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_False, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_5;
    __pyx_L18_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":1506
 *             # Because these used to be builtins instead of keywords, they
 *             # may still show up as name references. We ignore them.
 *             continue             # <<<<<<<<<<<<<<
 *         try:
 *             global_vars[name] = global_ns[name]
 */
      goto __pyx_L15_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":1503
 *     unbound_names = set()
 *     for name in code.co_names:
 *         if name in ("None", "True", "False"):             # <<<<<<<<<<<<<<
 *             # Because these used to be builtins instead of keywords, they
 *             # may still show up as name references. We ignore them.
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1507
 *             # may still show up as name references. We ignore them.
 *             continue
 *         try:             # <<<<<<<<<<<<<<
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1508
 *             continue
 *         try:
 *             global_vars[name] = global_ns[name]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             try:
 */
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_global_ns, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1508, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(PyDict_SetItem(__pyx_v_global_vars, __pyx_v_name, __pyx_t_2) < 0)) __PYX_ERR(0, 1508, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1507
 *             # may still show up as name references. We ignore them.
 *             continue
 *         try:             # <<<<<<<<<<<<<<
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      goto __pyx_L28_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1509
 *         try:
 *             global_vars[name] = global_ns[name]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             try:
 *                 builtin_vars[name] = builtin_ns[name]
 */
      __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_12) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getclosurevars", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_9) < 0) __PYX_ERR(0, 1509, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_9);

        /* "Aspidites/_vendor/contracts/inspection.py":1510
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 builtin_vars[name] = builtin_ns[name]
 *             except KeyError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":1511
 *         except KeyError:
 *             try:
 *                 builtin_vars[name] = builtin_ns[name]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 unbound_names.add(name)
 */
            __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_builtin_ns, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (unlikely(PyDict_SetItem(__pyx_v_builtin_vars, __pyx_v_name, __pyx_t_3) < 0)) __PYX_ERR(0, 1511, __pyx_L31_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":1510
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 builtin_vars[name] = builtin_ns[name]
 *             except KeyError:
 */
          }
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          goto __pyx_L38_try_end;
          __pyx_L31_error:;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1512
 *             try:
 *                 builtin_vars[name] = builtin_ns[name]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 unbound_names.add(name)
 * 
 */
          __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
          if (__pyx_t_12) {
            __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getclosurevars", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_10, &__pyx_t_19) < 0) __PYX_ERR(0, 1512, __pyx_L33_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GOTREF(__pyx_t_19);

            /* "Aspidites/_vendor/contracts/inspection.py":1513
 *                 builtin_vars[name] = builtin_ns[name]
 *             except KeyError:
 *                 unbound_names.add(name)             # <<<<<<<<<<<<<<
 * 
 *     return ClosureVars(nonlocal_vars, global_vars,
 */
            __pyx_t_20 = PySet_Add(__pyx_v_unbound_names, __pyx_v_name); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1513, __pyx_L33_except_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            goto __pyx_L32_exception_handled;
          }
          goto __pyx_L33_except_error;
          __pyx_L33_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":1510
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 builtin_vars[name] = builtin_ns[name]
 *             except KeyError:
 */
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          goto __pyx_L23_except_error;
          __pyx_L32_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          __pyx_L38_try_end:;
        }
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L22_exception_handled;
      }
      goto __pyx_L23_except_error;
      __pyx_L23_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":1507
 *             # may still show up as name references. We ignore them.
 *             continue
 *         try:             # <<<<<<<<<<<<<<
 *             global_vars[name] = global_ns[name]
 *         except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      goto __pyx_L1_error;
      __pyx_L22_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      __pyx_L28_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1502
 *     builtin_vars = {}
 *     unbound_names = set()
 *     for name in code.co_names:             # <<<<<<<<<<<<<<
 *         if name in ("None", "True", "False"):
 *             # Because these used to be builtins instead of keywords, they
 */
    __pyx_L15_continue:;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1515
 *                 unbound_names.add(name)
 * 
 *     return ClosureVars(nonlocal_vars, global_vars,             # <<<<<<<<<<<<<<
 *                        builtin_vars, unbound_names)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ClosureVars); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":1516
 * 
 *     return ClosureVars(nonlocal_vars, global_vars,
 *                        builtin_vars, unbound_names)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_nonlocal_vars, __pyx_v_global_vars, __pyx_v_builtin_vars, __pyx_v_unbound_names};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_6);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_nonlocal_vars, __pyx_v_global_vars, __pyx_v_builtin_vars, __pyx_v_unbound_names};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_6);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_nonlocal_vars);
    __Pyx_GIVEREF(__pyx_v_nonlocal_vars);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_12, __pyx_v_nonlocal_vars);
    __Pyx_INCREF(__pyx_v_global_vars);
    __Pyx_GIVEREF(__pyx_v_global_vars);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_12, __pyx_v_global_vars);
    __Pyx_INCREF(__pyx_v_builtin_vars);
    __Pyx_GIVEREF(__pyx_v_builtin_vars);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_12, __pyx_v_builtin_vars);
    __Pyx_INCREF(__pyx_v_unbound_names);
    __Pyx_GIVEREF(__pyx_v_unbound_names);
    PyTuple_SET_ITEM(__pyx_t_2, 3+__pyx_t_12, __pyx_v_unbound_names);
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1467
 * 
 * 
 * def getclosurevars(func):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of free variables to their current values.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getclosurevars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_code);
  __Pyx_XDECREF(__pyx_v_nonlocal_vars);
  __Pyx_XDECREF(__pyx_v_global_ns);
  __Pyx_XDECREF(__pyx_v_builtin_ns);
  __Pyx_XDECREF(__pyx_v_global_vars);
  __Pyx_XDECREF(__pyx_v_builtin_vars);
  __Pyx_XDECREF(__pyx_v_unbound_names);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_var);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_cell);
  __Pyx_XDECREF(__pyx_v_func);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1524
 * 
 * 
 * def getframeinfo(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get information about a frame or traceback object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_113getframeinfo(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_112getframeinfo[] = "Get information about a frame or traceback object.\n\n    A tuple of five things is returned: the filename, the line number of\n    the current line, the function name, a list of lines of context from\n    the source code, and the index of the current line within that list.\n    The optional second argument specifies the number of lines of context\n    to return, which are centered around the current line.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_113getframeinfo = {"getframeinfo", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_113getframeinfo, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_112getframeinfo};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_113getframeinfo(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_frame = 0;
  PyObject *__pyx_v_context = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getframeinfo (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_frame,&__pyx_n_s_context,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_frame)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getframeinfo") < 0)) __PYX_ERR(0, 1524, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_frame = values[0];
    __pyx_v_context = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getframeinfo", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getframeinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_112getframeinfo(__pyx_self, __pyx_v_frame, __pyx_v_context);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_112getframeinfo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_lineno = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_lines = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_lnum = NULL;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  Py_ssize_t __pyx_t_11;
  long __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getframeinfo", 0);
  __Pyx_INCREF(__pyx_v_frame);

  /* "Aspidites/_vendor/contracts/inspection.py":1532
 *     The optional second argument specifies the number of lines of context
 *     to return, which are centered around the current line."""
 *     if istraceback(frame):             # <<<<<<<<<<<<<<
 *         lineno = frame.tb_lineno
 *         frame = frame.tb_frame
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_istraceback); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_frame);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1533
 *     to return, which are centered around the current line."""
 *     if istraceback(frame):
 *         lineno = frame.tb_lineno             # <<<<<<<<<<<<<<
 *         frame = frame.tb_frame
 *     else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_tb_lineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_lineno = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1534
 *     if istraceback(frame):
 *         lineno = frame.tb_lineno
 *         frame = frame.tb_frame             # <<<<<<<<<<<<<<
 *     else:
 *         lineno = frame.f_lineno
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_frame, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1532
 *     The optional second argument specifies the number of lines of context
 *     to return, which are centered around the current line."""
 *     if istraceback(frame):             # <<<<<<<<<<<<<<
 *         lineno = frame.tb_lineno
 *         frame = frame.tb_frame
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1536
 *         frame = frame.tb_frame
 *     else:
 *         lineno = frame.f_lineno             # <<<<<<<<<<<<<<
 *     if not isframe(frame):
 *         raise TypeError('{!r} is not a frame or traceback object'.format(frame))
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_lineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_lineno = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":1537
 *     else:
 *         lineno = frame.f_lineno
 *     if not isframe(frame):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a frame or traceback object'.format(frame))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isframe); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_frame);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1538
 *         lineno = frame.f_lineno
 *     if not isframe(frame):
 *         raise TypeError('{!r} is not a frame or traceback object'.format(frame))             # <<<<<<<<<<<<<<
 * 
 *     filename = getsourcefile(frame) or getfile(frame)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_frame_or_traceback_o, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_frame);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1538, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1537
 *     else:
 *         lineno = frame.f_lineno
 *     if not isframe(frame):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a frame or traceback object'.format(frame))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1540
 *         raise TypeError('{!r} is not a frame or traceback object'.format(frame))
 * 
 *     filename = getsourcefile(frame) or getfile(frame)             # <<<<<<<<<<<<<<
 *     if context > 0:
 *         start = lineno - 1 - context // 2
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getsourcefile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_frame);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1540, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_frame);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_v_filename = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1541
 * 
 *     filename = getsourcefile(frame) or getfile(frame)
 *     if context > 0:             # <<<<<<<<<<<<<<
 *         start = lineno - 1 - context // 2
 *         try:
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_context, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1542
 *     filename = getsourcefile(frame) or getfile(frame)
 *     if context > 0:
 *         start = lineno - 1 - context // 2             # <<<<<<<<<<<<<<
 *         try:
 *             lines, lnum = findsource(frame)
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_lineno, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_context, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_start = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1543
 *     if context > 0:
 *         start = lineno - 1 - context // 2
 *         try:             # <<<<<<<<<<<<<<
 *             lines, lnum = findsource(frame)
 *         except OSError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1544
 *         start = lineno - 1 - context // 2
 *         try:
 *             lines, lnum = findsource(frame)             # <<<<<<<<<<<<<<
 *         except OSError:
 *             lines = index = None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_findsource); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_frame) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_frame);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1544, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
          PyObject* sequence = __pyx_t_3;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1544, __pyx_L8_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1544, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1544, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_10 = Py_TYPE(__pyx_t_6)->tp_iternext;
          index = 0; __pyx_t_1 = __pyx_t_10(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_2 = __pyx_t_10(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L14_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_6), 2) < 0) __PYX_ERR(0, 1544, __pyx_L8_error)
          __pyx_t_10 = NULL;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L15_unpacking_done;
          __pyx_L14_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_10 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1544, __pyx_L8_error)
          __pyx_L15_unpacking_done:;
        }
        __pyx_v_lines = __pyx_t_1;
        __pyx_t_1 = 0;
        __pyx_v_lnum = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1543
 *     if context > 0:
 *         start = lineno - 1 - context // 2
 *         try:             # <<<<<<<<<<<<<<
 *             lines, lnum = findsource(frame)
 *         except OSError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1548
 *             lines = index = None
 *         else:
 *             start = max(0, min(start, len(lines) - context))             # <<<<<<<<<<<<<<
 *             lines = lines[start:start + context]
 *             index = lineno - 1 - start
 */
      /*else:*/ {
        __pyx_t_11 = PyObject_Length(__pyx_v_lines); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_v_start);
        __pyx_t_3 = __pyx_v_start;
        __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_5) {
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_1 = __pyx_t_2;
        } else {
          __Pyx_INCREF(__pyx_t_3);
          __pyx_t_1 = __pyx_t_3;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_2 = __pyx_t_1;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = 0;
        __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_5) {
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_1 = __pyx_t_2;
        } else {
          __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1548, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_1 = __pyx_t_6;
          __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __pyx_t_1;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_start, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1549
 *         else:
 *             start = max(0, min(start, len(lines) - context))
 *             lines = lines[start:start + context]             # <<<<<<<<<<<<<<
 *             index = lineno - 1 - start
 *     else:
 */
        __pyx_t_2 = PyNumber_Add(__pyx_v_start, __pyx_v_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1549, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_lines, 0, 0, &__pyx_v_start, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1549, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v_lines, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1550
 *             start = max(0, min(start, len(lines) - context))
 *             lines = lines[start:start + context]
 *             index = lineno - 1 - start             # <<<<<<<<<<<<<<
 *     else:
 *         lines = index = None
 */
        __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_lineno, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1550, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_index = __pyx_t_2;
        __pyx_t_2 = 0;
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1545
 *         try:
 *             lines, lnum = findsource(frame)
 *         except OSError:             # <<<<<<<<<<<<<<
 *             lines = index = None
 *         else:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
      if (__pyx_t_13) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getframeinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 1545, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);

        /* "Aspidites/_vendor/contracts/inspection.py":1546
 *             lines, lnum = findsource(frame)
 *         except OSError:
 *             lines = index = None             # <<<<<<<<<<<<<<
 *         else:
 *             start = max(0, min(start, len(lines) - context))
 */
        __Pyx_INCREF(Py_None);
        __Pyx_XDECREF_SET(__pyx_v_lines, Py_None);
        __Pyx_INCREF(Py_None);
        __pyx_v_index = Py_None;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L9_exception_handled;
      }
      goto __pyx_L10_except_error;
      __pyx_L10_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":1543
 *     if context > 0:
 *         start = lineno - 1 - context // 2
 *         try:             # <<<<<<<<<<<<<<
 *             lines, lnum = findsource(frame)
 *         except OSError:
 */
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L9_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L13_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1541
 * 
 *     filename = getsourcefile(frame) or getfile(frame)
 *     if context > 0:             # <<<<<<<<<<<<<<
 *         start = lineno - 1 - context // 2
 *         try:
 */
    goto __pyx_L7;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1552
 *             index = lineno - 1 - start
 *     else:
 *         lines = index = None             # <<<<<<<<<<<<<<
 * 
 *     return Traceback(filename, lineno, frame.f_code.co_name, lines, index)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_lines = Py_None;
    __Pyx_INCREF(Py_None);
    __pyx_v_index = Py_None;
  }
  __pyx_L7:;

  /* "Aspidites/_vendor/contracts/inspection.py":1554
 *         lines = index = None
 * 
 *     return Traceback(filename, lineno, frame.f_code.co_name, lines, index)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Traceback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_co_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_13 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_13 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[6] = {__pyx_t_2, __pyx_v_filename, __pyx_v_lineno, __pyx_t_3, __pyx_v_lines, __pyx_v_index};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_13, 5+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[6] = {__pyx_t_2, __pyx_v_filename, __pyx_v_lineno, __pyx_t_3, __pyx_v_lines, __pyx_v_index};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_13, 5+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_14 = PyTuple_New(5+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_filename);
    __Pyx_GIVEREF(__pyx_v_filename);
    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_v_filename);
    __Pyx_INCREF(__pyx_v_lineno);
    __Pyx_GIVEREF(__pyx_v_lineno);
    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_v_lineno);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_13, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_lines);
    __Pyx_GIVEREF(__pyx_v_lines);
    PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_13, __pyx_v_lines);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_14, 4+__pyx_t_13, __pyx_v_index);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1524
 * 
 * 
 * def getframeinfo(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get information about a frame or traceback object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getframeinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lineno);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_lnum);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1557
 * 
 * 
 * def getlineno(frame):             # <<<<<<<<<<<<<<
 *     """Get the line number from a frame object, allowing for optimization."""
 *     # FrameType.f_lineno is now a descriptor that grovels co_lnotab
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_115getlineno(PyObject *__pyx_self, PyObject *__pyx_v_frame); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_114getlineno[] = "Get the line number from a frame object, allowing for optimization.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_115getlineno = {"getlineno", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_115getlineno, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_114getlineno};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_115getlineno(PyObject *__pyx_self, PyObject *__pyx_v_frame) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getlineno (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_114getlineno(__pyx_self, ((PyObject *)__pyx_v_frame));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_114getlineno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getlineno", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1560
 *     """Get the line number from a frame object, allowing for optimization."""
 *     # FrameType.f_lineno is now a descriptor that grovels co_lnotab
 *     return frame.f_lineno             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_lineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1557
 * 
 * 
 * def getlineno(frame):             # <<<<<<<<<<<<<<
 *     """Get the line number from a frame object, allowing for optimization."""
 *     # FrameType.f_lineno is now a descriptor that grovels co_lnotab
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getlineno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1566
 * 
 * 
 * def getouterframes(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a frame and all higher (calling) frames.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_117getouterframes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_116getouterframes[] = "Get a list of records for a frame and all higher (calling) frames.\n\n    Each record contains a frame object, filename, line number, function\n    name, a list of lines of context, and index within the context.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_117getouterframes = {"getouterframes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_117getouterframes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_116getouterframes};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_117getouterframes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_frame = 0;
  PyObject *__pyx_v_context = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getouterframes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_frame,&__pyx_n_s_context,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_frame)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getouterframes") < 0)) __PYX_ERR(0, 1566, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_frame = values[0];
    __pyx_v_context = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getouterframes", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1566, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getouterframes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_116getouterframes(__pyx_self, __pyx_v_frame, __pyx_v_context);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_116getouterframes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_framelist = NULL;
  PyObject *__pyx_v_frameinfo = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getouterframes", 0);
  __Pyx_INCREF(__pyx_v_frame);

  /* "Aspidites/_vendor/contracts/inspection.py":1571
 *     Each record contains a frame object, filename, line number, function
 *     name, a list of lines of context, and index within the context."""
 *     framelist = []             # <<<<<<<<<<<<<<
 *     while frame:
 *         frameinfo = (frame,) + getframeinfo(frame, context)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_framelist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1572
 *     name, a list of lines of context, and index within the context."""
 *     framelist = []
 *     while frame:             # <<<<<<<<<<<<<<
 *         frameinfo = (frame,) + getframeinfo(frame, context)
 *         framelist.append(FrameInfo(*frameinfo))
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_frame); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1572, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "Aspidites/_vendor/contracts/inspection.py":1573
 *     framelist = []
 *     while frame:
 *         frameinfo = (frame,) + getframeinfo(frame, context)             # <<<<<<<<<<<<<<
 *         framelist.append(FrameInfo(*frameinfo))
 *         frame = frame.f_back
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_frame);
    __Pyx_GIVEREF(__pyx_v_frame);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_frame);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getframeinfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_frame, __pyx_v_context};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_frame, __pyx_v_context};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_frame);
      __Pyx_GIVEREF(__pyx_v_frame);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_frame);
      __Pyx_INCREF(__pyx_v_context);
      __Pyx_GIVEREF(__pyx_v_context);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_context);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_frameinfo, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1574
 *     while frame:
 *         frameinfo = (frame,) + getframeinfo(frame, context)
 *         framelist.append(FrameInfo(*frameinfo))             # <<<<<<<<<<<<<<
 *         frame = frame.f_back
 *     return framelist
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FrameInfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_frameinfo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_framelist, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1574, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1575
 *         frameinfo = (frame,) + getframeinfo(frame, context)
 *         framelist.append(FrameInfo(*frameinfo))
 *         frame = frame.f_back             # <<<<<<<<<<<<<<
 *     return framelist
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_back); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_frame, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1576
 *         framelist.append(FrameInfo(*frameinfo))
 *         frame = frame.f_back
 *     return framelist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_framelist);
  __pyx_r = __pyx_v_framelist;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1566
 * 
 * 
 * def getouterframes(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a frame and all higher (calling) frames.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getouterframes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_framelist);
  __Pyx_XDECREF(__pyx_v_frameinfo);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1579
 * 
 * 
 * def getinnerframes(tb, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a traceback's frame and all lower frames.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_119getinnerframes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_118getinnerframes[] = "Get a list of records for a traceback's frame and all lower frames.\n\n    Each record contains a frame object, filename, line number, function\n    name, a list of lines of context, and index within the context.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_119getinnerframes = {"getinnerframes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_119getinnerframes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_118getinnerframes};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_119getinnerframes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_tb = 0;
  PyObject *__pyx_v_context = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getinnerframes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tb,&__pyx_n_s_context,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tb)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getinnerframes") < 0)) __PYX_ERR(0, 1579, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_tb = values[0];
    __pyx_v_context = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getinnerframes", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1579, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getinnerframes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_118getinnerframes(__pyx_self, __pyx_v_tb, __pyx_v_context);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_118getinnerframes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tb, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_framelist = NULL;
  PyObject *__pyx_v_frameinfo = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getinnerframes", 0);
  __Pyx_INCREF(__pyx_v_tb);

  /* "Aspidites/_vendor/contracts/inspection.py":1584
 *     Each record contains a frame object, filename, line number, function
 *     name, a list of lines of context, and index within the context."""
 *     framelist = []             # <<<<<<<<<<<<<<
 *     while tb:
 *         frameinfo = (tb.tb_frame,) + getframeinfo(tb, context)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_framelist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1585
 *     name, a list of lines of context, and index within the context."""
 *     framelist = []
 *     while tb:             # <<<<<<<<<<<<<<
 *         frameinfo = (tb.tb_frame,) + getframeinfo(tb, context)
 *         framelist.append(FrameInfo(*frameinfo))
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_tb); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1585, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "Aspidites/_vendor/contracts/inspection.py":1586
 *     framelist = []
 *     while tb:
 *         frameinfo = (tb.tb_frame,) + getframeinfo(tb, context)             # <<<<<<<<<<<<<<
 *         framelist.append(FrameInfo(*frameinfo))
 *         tb = tb.tb_next
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getframeinfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_tb, __pyx_v_context};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_tb, __pyx_v_context};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_tb);
      __Pyx_GIVEREF(__pyx_v_tb);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_tb);
      __Pyx_INCREF(__pyx_v_context);
      __Pyx_GIVEREF(__pyx_v_context);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_context);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_frameinfo, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1587
 *     while tb:
 *         frameinfo = (tb.tb_frame,) + getframeinfo(tb, context)
 *         framelist.append(FrameInfo(*frameinfo))             # <<<<<<<<<<<<<<
 *         tb = tb.tb_next
 *     return framelist
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FrameInfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_frameinfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_framelist, __pyx_t_3); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1588
 *         frameinfo = (tb.tb_frame,) + getframeinfo(tb, context)
 *         framelist.append(FrameInfo(*frameinfo))
 *         tb = tb.tb_next             # <<<<<<<<<<<<<<
 *     return framelist
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_next); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_tb, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1589
 *         framelist.append(FrameInfo(*frameinfo))
 *         tb = tb.tb_next
 *     return framelist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_framelist);
  __pyx_r = __pyx_v_framelist;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1579
 * 
 * 
 * def getinnerframes(tb, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a traceback's frame and all lower frames.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getinnerframes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_framelist);
  __Pyx_XDECREF(__pyx_v_frameinfo);
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1592
 * 
 * 
 * def currentframe():             # <<<<<<<<<<<<<<
 *     """Return the frame of the caller or None if this is not possible."""
 *     return sys._getframe(1) if hasattr(sys, "_getframe") else None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_121currentframe(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_120currentframe[] = "Return the frame of the caller or None if this is not possible.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_121currentframe = {"currentframe", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_121currentframe, METH_NOARGS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_120currentframe};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_121currentframe(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("currentframe (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_120currentframe(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_120currentframe(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("currentframe", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1594
 * def currentframe():
 *     """Return the frame of the caller or None if this is not possible."""
 *     return sys._getframe(1) if hasattr(sys, "_getframe") else None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_HasAttr(__pyx_t_2, __pyx_n_u_getframe); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((__pyx_t_3 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getframe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_int_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1592
 * 
 * 
 * def currentframe():             # <<<<<<<<<<<<<<
 *     """Return the frame of the caller or None if this is not possible."""
 *     return sys._getframe(1) if hasattr(sys, "_getframe") else None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.currentframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1597
 * 
 * 
 * def stack(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack above the caller's frame."""
 *     # noinspection PyProtectedMember, PyUnresolvedReferences
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_123stack(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_122stack[] = "Return a list of records for the stack above the caller's frame.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_123stack = {"stack", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_123stack, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_122stack};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_123stack(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_context = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stack (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "stack") < 0)) __PYX_ERR(0, 1597, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("stack", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1597, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.stack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_122stack(__pyx_self, __pyx_v_context);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_122stack(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stack", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1600
 *     """Return a list of records for the stack above the caller's frame."""
 *     # noinspection PyProtectedMember, PyUnresolvedReferences
 *     return getouterframes(sys._getframe(1), context)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getouterframes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getframe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_int_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_context};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_context};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_context);
    __Pyx_GIVEREF(__pyx_v_context);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_context);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1597
 * 
 * 
 * def stack(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack above the caller's frame."""
 *     # noinspection PyProtectedMember, PyUnresolvedReferences
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.stack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1603
 * 
 * 
 * def trace(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack below the current exception."""
 *     # noinspection PyTypeChecker
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_125trace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_124trace[] = "Return a list of records for the stack below the current exception.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_125trace = {"trace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_125trace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_124trace};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_125trace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_context = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "trace") < 0)) __PYX_ERR(0, 1603, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("trace", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1603, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_124trace(__pyx_self, __pyx_v_context);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_124trace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trace", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1606
 *     """Return a list of records for the stack below the current exception."""
 *     # noinspection PyTypeChecker
 *     return getinnerframes(sys.exc_info()[2], context)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getinnerframes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_exc_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_context};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1606, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_context};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1606, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_context);
    __Pyx_GIVEREF(__pyx_v_context);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_context);
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1603
 * 
 * 
 * def trace(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack below the current exception."""
 *     # noinspection PyTypeChecker
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1614
 * 
 * 
 * def _static_getmro(klass):             # <<<<<<<<<<<<<<
 *     return type.__dict__['__mro__'].__get__(klass)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_127_static_getmro(PyObject *__pyx_self, PyObject *__pyx_v_klass); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_127_static_getmro = {"_static_getmro", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_127_static_getmro, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_127_static_getmro(PyObject *__pyx_self, PyObject *__pyx_v_klass) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_static_getmro (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_126_static_getmro(__pyx_self, ((PyObject *)__pyx_v_klass));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_126_static_getmro(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_static_getmro", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1615
 * 
 * def _static_getmro(klass):
 *     return type.__dict__['__mro__'].__get__(klass)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_mro); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_klass) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_klass);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1614
 * 
 * 
 * def _static_getmro(klass):             # <<<<<<<<<<<<<<
 *     return type.__dict__['__mro__'].__get__(klass)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._static_getmro", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1618
 * 
 * 
 * def _check_instance(obj, attr):             # <<<<<<<<<<<<<<
 *     instance_dict = {}
 *     try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_129_check_instance(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_129_check_instance = {"_check_instance", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_129_check_instance, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_129_check_instance(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_attr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_check_instance (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_attr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_check_instance", 1, 2, 2, 1); __PYX_ERR(0, 1618, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_check_instance") < 0)) __PYX_ERR(0, 1618, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_obj = values[0];
    __pyx_v_attr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_check_instance", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1618, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._check_instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_128_check_instance(__pyx_self, __pyx_v_obj, __pyx_v_attr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_128_check_instance(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_attr) {
  PyObject *__pyx_v_instance_dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_instance", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1619
 * 
 * def _check_instance(obj, attr):
 *     instance_dict = {}             # <<<<<<<<<<<<<<
 *     try:
 *         instance_dict = object.__getattribute__(obj, "__dict__")
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instance_dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1620
 * def _check_instance(obj, attr):
 *     instance_dict = {}
 *     try:             # <<<<<<<<<<<<<<
 *         instance_dict = object.__getattribute__(obj, "__dict__")
 *     except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":1621
 *     instance_dict = {}
 *     try:
 *         instance_dict = object.__getattribute__(obj, "__dict__")             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         pass
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_object, __pyx_n_s_getattribute); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1621, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_obj, __pyx_n_u_dict};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1621, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_obj, __pyx_n_u_dict};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1621, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1621, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_obj);
        __Pyx_INCREF(__pyx_n_u_dict);
        __Pyx_GIVEREF(__pyx_n_u_dict);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_n_u_dict);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1621, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_instance_dict, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1620
 * def _check_instance(obj, attr):
 *     instance_dict = {}
 *     try:             # <<<<<<<<<<<<<<
 *         instance_dict = object.__getattribute__(obj, "__dict__")
 *     except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1622
 *     try:
 *         instance_dict = object.__getattribute__(obj, "__dict__")
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         pass
 *     return dict.get(instance_dict, attr, _sentinel)
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":1620
 * def _check_instance(obj, attr):
 *     instance_dict = {}
 *     try:             # <<<<<<<<<<<<<<
 *         instance_dict = object.__getattribute__(obj, "__dict__")
 *     except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1624
 *     except AttributeError:
 *         pass
 *     return dict.get(instance_dict, attr, _sentinel)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_instance_dict == Py_None)) {
    PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "get", "object");
    __PYX_ERR(0, 1624, __pyx_L1_error)
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_instance_dict, __pyx_v_attr, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1618
 * 
 * 
 * def _check_instance(obj, attr):             # <<<<<<<<<<<<<<
 *     instance_dict = {}
 *     try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._check_instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_instance_dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1627
 * 
 * 
 * def _check_class(klass, attr):             # <<<<<<<<<<<<<<
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_131_check_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_131_check_class = {"_check_class", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_131_check_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_131_check_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_klass = 0;
  PyObject *__pyx_v_attr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_check_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_klass,&__pyx_n_s_attr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_klass)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_check_class", 1, 2, 2, 1); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_check_class") < 0)) __PYX_ERR(0, 1627, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_klass = values[0];
    __pyx_v_attr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_check_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1627, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._check_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_130_check_class(__pyx_self, __pyx_v_klass, __pyx_v_attr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_130_check_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass, PyObject *__pyx_v_attr) {
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_class", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1628
 * 
 * def _check_class(klass, attr):
 *     for entry in _static_getmro(klass):             # <<<<<<<<<<<<<<
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_static_getmro); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_klass) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_klass);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1628, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1628, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1628, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1628, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1629
 * def _check_class(klass, attr):
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:             # <<<<<<<<<<<<<<
 *             try:
 *                 return entry.__dict__[attr]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_shadowed_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, ((PyObject *)Py_TYPE(__pyx_v_entry))) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_entry)));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = (__pyx_t_1 == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":1630
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:             # <<<<<<<<<<<<<<
 *                 return entry.__dict__[attr]
 *             except KeyError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":1631
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:
 *                 return entry.__dict__[attr]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 pass
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1631, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1631, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_r = __pyx_t_1;
          __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L10_try_return;

          /* "Aspidites/_vendor/contracts/inspection.py":1630
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:             # <<<<<<<<<<<<<<
 *                 return entry.__dict__[attr]
 *             except KeyError:
 */
        }
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1632
 *             try:
 *                 return entry.__dict__[attr]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 pass
 *     return _sentinel
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
        if (__pyx_t_12) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L7_exception_handled;
        }
        goto __pyx_L8_except_error;
        __pyx_L8_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":1630
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:             # <<<<<<<<<<<<<<
 *                 return entry.__dict__[attr]
 *             except KeyError:
 */
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L10_try_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L0;
        __pyx_L7_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1629
 * def _check_class(klass, attr):
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:             # <<<<<<<<<<<<<<
 *             try:
 *                 return entry.__dict__[attr]
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1628
 * 
 * def _check_class(klass, attr):
 *     for entry in _static_getmro(klass):             # <<<<<<<<<<<<<<
 *         if _shadowed_dict(type(entry)) is _sentinel:
 *             try:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1634
 *             except KeyError:
 *                 pass
 *     return _sentinel             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1627
 * 
 * 
 * def _check_class(klass, attr):             # <<<<<<<<<<<<<<
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._check_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1637
 * 
 * 
 * def _is_type(obj):             # <<<<<<<<<<<<<<
 *     try:
 *         _static_getmro(obj)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_133_is_type(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_133_is_type = {"_is_type", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_133_is_type, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_133_is_type(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_type (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_132_is_type(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_132_is_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_type", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1638
 * 
 * def _is_type(obj):
 *     try:             # <<<<<<<<<<<<<<
 *         _static_getmro(obj)
 *     except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":1639
 * def _is_type(obj):
 *     try:
 *         _static_getmro(obj)             # <<<<<<<<<<<<<<
 *     except TypeError:
 *         return False
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_static_getmro); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1639, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_obj);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1639, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1638
 * 
 * def _is_type(obj):
 *     try:             # <<<<<<<<<<<<<<
 *         _static_getmro(obj)
 *     except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1640
 *     try:
 *         _static_getmro(obj)
 *     except TypeError:             # <<<<<<<<<<<<<<
 *         return False
 *     return True
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._is_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 1640, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/contracts/inspection.py":1641
 *         _static_getmro(obj)
 *     except TypeError:
 *         return False             # <<<<<<<<<<<<<<
 *     return True
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":1638
 * 
 * def _is_type(obj):
 *     try:             # <<<<<<<<<<<<<<
 *         _static_getmro(obj)
 *     except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1642
 *     except TypeError:
 *         return False
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1637
 * 
 * 
 * def _is_type(obj):             # <<<<<<<<<<<<<<
 *     try:
 *         _static_getmro(obj)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._is_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1645
 * 
 * 
 * def _shadowed_dict(klass):             # <<<<<<<<<<<<<<
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_135_shadowed_dict(PyObject *__pyx_self, PyObject *__pyx_v_klass); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_135_shadowed_dict = {"_shadowed_dict", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_135_shadowed_dict, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_135_shadowed_dict(PyObject *__pyx_self, PyObject *__pyx_v_klass) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_shadowed_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_134_shadowed_dict(__pyx_self, ((PyObject *)__pyx_v_klass));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_134_shadowed_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_klass) {
  PyObject *__pyx_v_dict_attr = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_class_dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_shadowed_dict", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1646
 * 
 * def _shadowed_dict(klass):
 *     dict_attr = type.__dict__["__dict__"]             # <<<<<<<<<<<<<<
 *     for entry in _static_getmro(klass):
 *         try:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dict_attr = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1647
 * def _shadowed_dict(klass):
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):             # <<<<<<<<<<<<<<
 *         try:
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_static_getmro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_klass) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_klass);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1647, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1647, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1647, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1647, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1648
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 *         try:             # <<<<<<<<<<<<<<
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 *         except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1649
 *     for entry in _static_getmro(klass):
 *         try:
 *             class_dict = dict_attr.__get__(entry)["__dict__"]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             pass
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dict_attr, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_entry) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_entry);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1649, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_class_dict, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1648
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 *         try:             # <<<<<<<<<<<<<<
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 *         except KeyError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1653
 *             pass
 *         else:
 *             if not (type(class_dict) is types.GetSetDescriptorType and             # <<<<<<<<<<<<<<
 *                     class_dict.__name__ == "__dict__" and
 *                     class_dict.__objclass__ is entry):
 */
      /*else:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1653, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_GetSetDescriptorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1653, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = (((PyObject *)Py_TYPE(__pyx_v_class_dict)) == __pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L14_bool_binop_done;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1654
 *         else:
 *             if not (type(class_dict) is types.GetSetDescriptorType and
 *                     class_dict.__name__ == "__dict__" and             # <<<<<<<<<<<<<<
 *                     class_dict.__objclass__ is entry):
 *                 return class_dict
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_class_dict, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1654, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_dict, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1654, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L14_bool_binop_done;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1655
 *             if not (type(class_dict) is types.GetSetDescriptorType and
 *                     class_dict.__name__ == "__dict__" and
 *                     class_dict.__objclass__ is entry):             # <<<<<<<<<<<<<<
 *                 return class_dict
 *     return _sentinel
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_class_dict, __pyx_n_s_objclass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1655, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_12 = (__pyx_t_2 == __pyx_v_entry);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        __pyx_t_10 = __pyx_t_11;
        __pyx_L14_bool_binop_done:;

        /* "Aspidites/_vendor/contracts/inspection.py":1653
 *             pass
 *         else:
 *             if not (type(class_dict) is types.GetSetDescriptorType and             # <<<<<<<<<<<<<<
 *                     class_dict.__name__ == "__dict__" and
 *                     class_dict.__objclass__ is entry):
 */
        __pyx_t_11 = ((!__pyx_t_10) != 0);
        if (__pyx_t_11) {

          /* "Aspidites/_vendor/contracts/inspection.py":1656
 *                     class_dict.__name__ == "__dict__" and
 *                     class_dict.__objclass__ is entry):
 *                 return class_dict             # <<<<<<<<<<<<<<
 *     return _sentinel
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_class_dict);
          __pyx_r = __pyx_v_class_dict;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L8_except_return;

          /* "Aspidites/_vendor/contracts/inspection.py":1653
 *             pass
 *         else:
 *             if not (type(class_dict) is types.GetSetDescriptorType and             # <<<<<<<<<<<<<<
 *                     class_dict.__name__ == "__dict__" and
 *                     class_dict.__objclass__ is entry):
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1650
 *         try:
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_13) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":1648
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 *         try:             # <<<<<<<<<<<<<<
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 *         except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L8_except_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L0;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L12_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1647
 * def _shadowed_dict(klass):
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):             # <<<<<<<<<<<<<<
 *         try:
 *             class_dict = dict_attr.__get__(entry)["__dict__"]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1657
 *                     class_dict.__objclass__ is entry):
 *                 return class_dict
 *     return _sentinel             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1645
 * 
 * 
 * def _shadowed_dict(klass):             # <<<<<<<<<<<<<<
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._shadowed_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dict_attr);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_class_dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1660
 * 
 * 
 * def getattr_static(obj, attr, default=_sentinel):             # <<<<<<<<<<<<<<
 *     """Retrieve attributes without triggering dynamic lookup via the
 *        descriptor protocol,  __getattr__ or __getattribute__.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_195__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_default);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_default);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_default);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_137getattr_static(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_136getattr_static[] = "Retrieve attributes without triggering dynamic lookup via the\n       descriptor protocol,  __getattr__ or __getattribute__.\n\n       Note: this function may not be able to retrieve all attributes\n       that getattr can fetch (like dynamically created attributes)\n       and may find attributes that getattr can't (like descriptors\n       that raise AttributeError). It can also return descriptor objects\n       instead of instance members in some cases. See the\n       documentation for details.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_137getattr_static = {"getattr_static", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_137getattr_static, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_136getattr_static};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_137getattr_static(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_attr = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getattr_static (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_attr,&__pyx_n_s_default,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults2 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_default;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getattr_static", 0, 2, 3, 1); __PYX_ERR(0, 1660, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getattr_static") < 0)) __PYX_ERR(0, 1660, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_attr = values[1];
    __pyx_v_default = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getattr_static", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1660, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getattr_static", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_136getattr_static(__pyx_self, __pyx_v_obj, __pyx_v_attr, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_136getattr_static(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_attr, PyObject *__pyx_v_default) {
  PyObject *__pyx_v_instance_result = NULL;
  PyObject *__pyx_v_klass = NULL;
  PyObject *__pyx_v_dict_attr = NULL;
  PyObject *__pyx_v_klass_result = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getattr_static", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1671
 *        documentation for details.
 *     """
 *     instance_result = _sentinel             # <<<<<<<<<<<<<<
 *     if not _is_type(obj):
 *         klass = type(obj)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instance_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1672
 *     """
 *     instance_result = _sentinel
 *     if not _is_type(obj):             # <<<<<<<<<<<<<<
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_is_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1673
 *     instance_result = _sentinel
 *     if not _is_type(obj):
 *         klass = type(obj)             # <<<<<<<<<<<<<<
 *         dict_attr = _shadowed_dict(klass)
 *         if (dict_attr is _sentinel or
 */
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
    __pyx_v_klass = ((PyObject *)Py_TYPE(__pyx_v_obj));

    /* "Aspidites/_vendor/contracts/inspection.py":1674
 *     if not _is_type(obj):
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)             # <<<<<<<<<<<<<<
 *         if (dict_attr is _sentinel or
 *                 type(dict_attr) is types.MemberDescriptorType):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_shadowed_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_klass) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_klass);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_dict_attr = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1675
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)
 *         if (dict_attr is _sentinel or             # <<<<<<<<<<<<<<
 *                 type(dict_attr) is types.MemberDescriptorType):
 *             instance_result = _check_instance(obj, attr)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__pyx_v_dict_attr == __pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L5_bool_binop_done;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1676
 *         dict_attr = _shadowed_dict(klass)
 *         if (dict_attr is _sentinel or
 *                 type(dict_attr) is types.MemberDescriptorType):             # <<<<<<<<<<<<<<
 *             instance_result = _check_instance(obj, attr)
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MemberDescriptorType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (((PyObject *)Py_TYPE(__pyx_v_dict_attr)) == __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L5_bool_binop_done:;

    /* "Aspidites/_vendor/contracts/inspection.py":1675
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)
 *         if (dict_attr is _sentinel or             # <<<<<<<<<<<<<<
 *                 type(dict_attr) is types.MemberDescriptorType):
 *             instance_result = _check_instance(obj, attr)
 */
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":1677
 *         if (dict_attr is _sentinel or
 *                 type(dict_attr) is types.MemberDescriptorType):
 *             instance_result = _check_instance(obj, attr)             # <<<<<<<<<<<<<<
 *     else:
 *         klass = obj
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_check_instance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_obj, __pyx_v_attr};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_obj, __pyx_v_attr};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_obj);
        __Pyx_INCREF(__pyx_v_attr);
        __Pyx_GIVEREF(__pyx_v_attr);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_attr);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_instance_result, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1675
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)
 *         if (dict_attr is _sentinel or             # <<<<<<<<<<<<<<
 *                 type(dict_attr) is types.MemberDescriptorType):
 *             instance_result = _check_instance(obj, attr)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1672
 *     """
 *     instance_result = _sentinel
 *     if not _is_type(obj):             # <<<<<<<<<<<<<<
 *         klass = type(obj)
 *         dict_attr = _shadowed_dict(klass)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1679
 *             instance_result = _check_instance(obj, attr)
 *     else:
 *         klass = obj             # <<<<<<<<<<<<<<
 * 
 *     klass_result = _check_class(klass, attr)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_obj);
    __pyx_v_klass = __pyx_v_obj;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":1681
 *         klass = obj
 * 
 *     klass_result = _check_class(klass, attr)             # <<<<<<<<<<<<<<
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_check_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_klass, __pyx_v_attr};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_klass, __pyx_v_attr};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_klass);
    __Pyx_GIVEREF(__pyx_v_klass);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_klass);
    __Pyx_INCREF(__pyx_v_attr);
    __Pyx_GIVEREF(__pyx_v_attr);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_v_attr);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_klass_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1683
 *     klass_result = _check_class(klass, attr)
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_instance_result != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (__pyx_t_4 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L8_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = (__pyx_v_klass_result != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_6 != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1684
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and             # <<<<<<<<<<<<<<
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 *             return klass_result
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_check_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_klass_result)), __pyx_n_u_get};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1684, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_klass_result)), __pyx_n_u_get};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1684, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, ((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      __Pyx_INCREF(__pyx_n_u_get);
      __Pyx_GIVEREF(__pyx_n_u_get);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_n_u_get);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__pyx_t_2 != __pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L11_bool_binop_done;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1685
 *     if instance_result is not _sentinel and klass_result is not _sentinel:
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):             # <<<<<<<<<<<<<<
 *             return klass_result
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_check_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)Py_TYPE(__pyx_v_klass_result)), __pyx_n_u_set};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1685, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)Py_TYPE(__pyx_v_klass_result)), __pyx_n_u_set};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1685, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, ((PyObject *)Py_TYPE(__pyx_v_klass_result)));
      __Pyx_INCREF(__pyx_n_u_set);
      __Pyx_GIVEREF(__pyx_n_u_set);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_n_u_set);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = (__pyx_t_1 != __pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L11_bool_binop_done:;

    /* "Aspidites/_vendor/contracts/inspection.py":1684
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and             # <<<<<<<<<<<<<<
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 *             return klass_result
 */
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":1686
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 *             return klass_result             # <<<<<<<<<<<<<<
 * 
 *     if instance_result is not _sentinel:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_klass_result);
      __pyx_r = __pyx_v_klass_result;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":1684
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and             # <<<<<<<<<<<<<<
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 *             return klass_result
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1683
 *     klass_result = _check_class(klass, attr)
 * 
 *     if instance_result is not _sentinel and klass_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         if (_check_class(type(klass_result), '__get__') is not _sentinel and
 *                 _check_class(type(klass_result), '__set__') is not _sentinel):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1688
 *             return klass_result
 * 
 *     if instance_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         return instance_result
 *     if klass_result is not _sentinel:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = (__pyx_v_instance_result != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1689
 * 
 *     if instance_result is not _sentinel:
 *         return instance_result             # <<<<<<<<<<<<<<
 *     if klass_result is not _sentinel:
 *         return klass_result
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_instance_result);
    __pyx_r = __pyx_v_instance_result;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1688
 *             return klass_result
 * 
 *     if instance_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         return instance_result
 *     if klass_result is not _sentinel:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1690
 *     if instance_result is not _sentinel:
 *         return instance_result
 *     if klass_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         return klass_result
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_klass_result != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1691
 *         return instance_result
 *     if klass_result is not _sentinel:
 *         return klass_result             # <<<<<<<<<<<<<<
 * 
 *     if obj is klass:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_klass_result);
    __pyx_r = __pyx_v_klass_result;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1690
 *     if instance_result is not _sentinel:
 *         return instance_result
 *     if klass_result is not _sentinel:             # <<<<<<<<<<<<<<
 *         return klass_result
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1693
 *         return klass_result
 * 
 *     if obj is klass:             # <<<<<<<<<<<<<<
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):
 */
  __pyx_t_5 = (__pyx_v_obj == __pyx_v_klass);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1695
 *     if obj is klass:
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):             # <<<<<<<<<<<<<<
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_static_getmro); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_klass))) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)Py_TYPE(__pyx_v_klass)));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1695, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1695, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1695, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1695, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1695, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_10(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1695, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1696
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):
 *             if _shadowed_dict(type(entry)) is _sentinel:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return entry.__dict__[attr]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_shadowed_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, ((PyObject *)Py_TYPE(__pyx_v_entry))) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_entry)));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = (__pyx_t_2 == __pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/contracts/inspection.py":1697
 *         for entry in _static_getmro(type(klass)):
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return entry.__dict__[attr]
 *                 except KeyError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":1698
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:
 *                     return entry.__dict__[attr]             # <<<<<<<<<<<<<<
 *                 except KeyError:
 *                     pass
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1698, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1698, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_r = __pyx_t_2;
            __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L23_try_return;

            /* "Aspidites/_vendor/contracts/inspection.py":1697
 *         for entry in _static_getmro(type(klass)):
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return entry.__dict__[attr]
 *                 except KeyError:
 */
          }
          __pyx_L19_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1699
 *                 try:
 *                     return entry.__dict__[attr]
 *                 except KeyError:             # <<<<<<<<<<<<<<
 *                     pass
 *     if default is not _sentinel:
 */
          __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
          if (__pyx_t_7) {
            __Pyx_ErrRestore(0,0,0);
            goto __pyx_L20_exception_handled;
          }
          goto __pyx_L21_except_error;
          __pyx_L21_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":1697
 *         for entry in _static_getmro(type(klass)):
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return entry.__dict__[attr]
 *                 except KeyError:
 */
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          goto __pyx_L1_error;
          __pyx_L23_try_return:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          goto __pyx_L0;
          __pyx_L20_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1696
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):
 *             if _shadowed_dict(type(entry)) is _sentinel:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return entry.__dict__[attr]
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1695
 *     if obj is klass:
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):             # <<<<<<<<<<<<<<
 *             if _shadowed_dict(type(entry)) is _sentinel:
 *                 try:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1693
 *         return klass_result
 * 
 *     if obj is klass:             # <<<<<<<<<<<<<<
 *         # for types we check the metaclass too
 *         for entry in _static_getmro(type(klass)):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1701
 *                 except KeyError:
 *                     pass
 *     if default is not _sentinel:             # <<<<<<<<<<<<<<
 *         return default
 *     raise AttributeError(attr)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__pyx_v_default != __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1702
 *                     pass
 *     if default is not _sentinel:
 *         return default             # <<<<<<<<<<<<<<
 *     raise AttributeError(attr)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_default);
    __pyx_r = __pyx_v_default;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1701
 *                 except KeyError:
 *                     pass
 *     if default is not _sentinel:             # <<<<<<<<<<<<<<
 *         return default
 *     raise AttributeError(attr)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1703
 *     if default is not _sentinel:
 *         return default
 *     raise AttributeError(attr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_v_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 1703, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1660
 * 
 * 
 * def getattr_static(obj, attr, default=_sentinel):             # <<<<<<<<<<<<<<
 *     """Retrieve attributes without triggering dynamic lookup via the
 *        descriptor protocol,  __getattr__ or __getattribute__.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getattr_static", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_instance_result);
  __Pyx_XDECREF(__pyx_v_klass);
  __Pyx_XDECREF(__pyx_v_dict_attr);
  __Pyx_XDECREF(__pyx_v_klass_result);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1714
 * 
 * 
 * def getgeneratorstate(generator):             # <<<<<<<<<<<<<<
 *     """Get current state of a generator-iterator.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_139getgeneratorstate(PyObject *__pyx_self, PyObject *__pyx_v_generator); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_138getgeneratorstate[] = "Get current state of a generator-iterator.\n\n    Possible states are:\n      GEN_CREATED: Waiting to start execution.\n      GEN_RUNNING: Currently being executed by the interpreter.\n      GEN_SUSPENDED: Currently suspended at a yield expression.\n      GEN_CLOSED: Execution has completed.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_139getgeneratorstate = {"getgeneratorstate", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_139getgeneratorstate, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_138getgeneratorstate};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_139getgeneratorstate(PyObject *__pyx_self, PyObject *__pyx_v_generator) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getgeneratorstate (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_138getgeneratorstate(__pyx_self, ((PyObject *)__pyx_v_generator));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_138getgeneratorstate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_generator) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getgeneratorstate", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1723
 *       GEN_CLOSED: Execution has completed.
 *     """
 *     if generator.gi_running:             # <<<<<<<<<<<<<<
 *         return GEN_RUNNING
 *     if generator.gi_frame is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_generator, __pyx_n_s_gi_running); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1724
 *     """
 *     if generator.gi_running:
 *         return GEN_RUNNING             # <<<<<<<<<<<<<<
 *     if generator.gi_frame is None:
 *         return GEN_CLOSED
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_GEN_RUNNING); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1723
 *       GEN_CLOSED: Execution has completed.
 *     """
 *     if generator.gi_running:             # <<<<<<<<<<<<<<
 *         return GEN_RUNNING
 *     if generator.gi_frame is None:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1725
 *     if generator.gi_running:
 *         return GEN_RUNNING
 *     if generator.gi_frame is None:             # <<<<<<<<<<<<<<
 *         return GEN_CLOSED
 *     if generator.gi_frame.f_lasti == -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_generator, __pyx_n_s_gi_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1726
 *         return GEN_RUNNING
 *     if generator.gi_frame is None:
 *         return GEN_CLOSED             # <<<<<<<<<<<<<<
 *     if generator.gi_frame.f_lasti == -1:
 *         return GEN_CREATED
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_GEN_CLOSED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1725
 *     if generator.gi_running:
 *         return GEN_RUNNING
 *     if generator.gi_frame is None:             # <<<<<<<<<<<<<<
 *         return GEN_CLOSED
 *     if generator.gi_frame.f_lasti == -1:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1727
 *     if generator.gi_frame is None:
 *         return GEN_CLOSED
 *     if generator.gi_frame.f_lasti == -1:             # <<<<<<<<<<<<<<
 *         return GEN_CREATED
 *     return GEN_SUSPENDED
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_generator, __pyx_n_s_gi_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_f_lasti); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1728
 *         return GEN_CLOSED
 *     if generator.gi_frame.f_lasti == -1:
 *         return GEN_CREATED             # <<<<<<<<<<<<<<
 *     return GEN_SUSPENDED
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_GEN_CREATED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1727
 *     if generator.gi_frame is None:
 *         return GEN_CLOSED
 *     if generator.gi_frame.f_lasti == -1:             # <<<<<<<<<<<<<<
 *         return GEN_CREATED
 *     return GEN_SUSPENDED
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1729
 *     if generator.gi_frame.f_lasti == -1:
 *         return GEN_CREATED
 *     return GEN_SUSPENDED             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_GEN_SUSPENDED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1714
 * 
 * 
 * def getgeneratorstate(generator):             # <<<<<<<<<<<<<<
 *     """Get current state of a generator-iterator.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getgeneratorstate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1732
 * 
 * 
 * def getgeneratorlocals(generator):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of generator local variables to their current values.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_141getgeneratorlocals(PyObject *__pyx_self, PyObject *__pyx_v_generator); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_140getgeneratorlocals[] = "\n    Get the mapping of generator local variables to their current values.\n\n    A dict is returned, with the keys the local variable names and values the\n    bound values.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_141getgeneratorlocals = {"getgeneratorlocals", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_141getgeneratorlocals, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_140getgeneratorlocals};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_141getgeneratorlocals(PyObject *__pyx_self, PyObject *__pyx_v_generator) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getgeneratorlocals (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_140getgeneratorlocals(__pyx_self, ((PyObject *)__pyx_v_generator));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_140getgeneratorlocals(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_generator) {
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getgeneratorlocals", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1739
 *     bound values."""
 * 
 *     if not isgenerator(generator):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python generator".format(generator))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isgenerator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_generator) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_generator);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1740
 * 
 *     if not isgenerator(generator):
 *         raise TypeError("{!r} is not a Python generator".format(generator))             # <<<<<<<<<<<<<<
 * 
 *     frame = getattr(generator, "gi_frame", None)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_Python_generator, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_generator) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_generator);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1740, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1739
 *     bound values."""
 * 
 *     if not isgenerator(generator):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python generator".format(generator))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1742
 *         raise TypeError("{!r} is not a Python generator".format(generator))
 * 
 *     frame = getattr(generator, "gi_frame", None)             # <<<<<<<<<<<<<<
 *     if frame is not None:
 *         return generator.gi_frame.f_locals
 */
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_generator, __pyx_n_u_gi_frame, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_frame = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1743
 * 
 *     frame = getattr(generator, "gi_frame", None)
 *     if frame is not None:             # <<<<<<<<<<<<<<
 *         return generator.gi_frame.f_locals
 *     else:
 */
  __pyx_t_5 = (__pyx_v_frame != Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1744
 *     frame = getattr(generator, "gi_frame", None)
 *     if frame is not None:
 *         return generator.gi_frame.f_locals             # <<<<<<<<<<<<<<
 *     else:
 *         return {}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_generator, __pyx_n_s_gi_frame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_f_locals); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1743
 * 
 *     frame = getattr(generator, "gi_frame", None)
 *     if frame is not None:             # <<<<<<<<<<<<<<
 *         return generator.gi_frame.f_locals
 *     else:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1746
 *         return generator.gi_frame.f_locals
 *     else:
 *         return {}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1732
 * 
 * 
 * def getgeneratorlocals(generator):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of generator local variables to their current values.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getgeneratorlocals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1757
 * 
 * 
 * def getcoroutinestate(coroutine):             # <<<<<<<<<<<<<<
 *     """Get current state of a coroutine object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_143getcoroutinestate(PyObject *__pyx_self, PyObject *__pyx_v_coroutine); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_142getcoroutinestate[] = "Get current state of a coroutine object.\n\n    Possible states are:\n      CORO_CREATED: Waiting to start execution.\n      CORO_RUNNING: Currently being executed by the interpreter.\n      CORO_SUSPENDED: Currently suspended at an await expression.\n      CORO_CLOSED: Execution has completed.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_143getcoroutinestate = {"getcoroutinestate", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_143getcoroutinestate, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_142getcoroutinestate};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_143getcoroutinestate(PyObject *__pyx_self, PyObject *__pyx_v_coroutine) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getcoroutinestate (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_142getcoroutinestate(__pyx_self, ((PyObject *)__pyx_v_coroutine));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_142getcoroutinestate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coroutine) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getcoroutinestate", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1766
 *       CORO_CLOSED: Execution has completed.
 *     """
 *     if coroutine.cr_running:             # <<<<<<<<<<<<<<
 *         return CORO_RUNNING
 *     if coroutine.cr_frame is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coroutine, __pyx_n_s_cr_running); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1767
 *     """
 *     if coroutine.cr_running:
 *         return CORO_RUNNING             # <<<<<<<<<<<<<<
 *     if coroutine.cr_frame is None:
 *         return CORO_CLOSED
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CORO_RUNNING); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1766
 *       CORO_CLOSED: Execution has completed.
 *     """
 *     if coroutine.cr_running:             # <<<<<<<<<<<<<<
 *         return CORO_RUNNING
 *     if coroutine.cr_frame is None:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1768
 *     if coroutine.cr_running:
 *         return CORO_RUNNING
 *     if coroutine.cr_frame is None:             # <<<<<<<<<<<<<<
 *         return CORO_CLOSED
 *     if coroutine.cr_frame.f_lasti == -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coroutine, __pyx_n_s_cr_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1769
 *         return CORO_RUNNING
 *     if coroutine.cr_frame is None:
 *         return CORO_CLOSED             # <<<<<<<<<<<<<<
 *     if coroutine.cr_frame.f_lasti == -1:
 *         return CORO_CREATED
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CORO_CLOSED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1768
 *     if coroutine.cr_running:
 *         return CORO_RUNNING
 *     if coroutine.cr_frame is None:             # <<<<<<<<<<<<<<
 *         return CORO_CLOSED
 *     if coroutine.cr_frame.f_lasti == -1:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1770
 *     if coroutine.cr_frame is None:
 *         return CORO_CLOSED
 *     if coroutine.cr_frame.f_lasti == -1:             # <<<<<<<<<<<<<<
 *         return CORO_CREATED
 *     return CORO_SUSPENDED
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coroutine, __pyx_n_s_cr_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_f_lasti); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1771
 *         return CORO_CLOSED
 *     if coroutine.cr_frame.f_lasti == -1:
 *         return CORO_CREATED             # <<<<<<<<<<<<<<
 *     return CORO_SUSPENDED
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CORO_CREATED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1770
 *     if coroutine.cr_frame is None:
 *         return CORO_CLOSED
 *     if coroutine.cr_frame.f_lasti == -1:             # <<<<<<<<<<<<<<
 *         return CORO_CREATED
 *     return CORO_SUSPENDED
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1772
 *     if coroutine.cr_frame.f_lasti == -1:
 *         return CORO_CREATED
 *     return CORO_SUSPENDED             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CORO_SUSPENDED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1757
 * 
 * 
 * def getcoroutinestate(coroutine):             # <<<<<<<<<<<<<<
 *     """Get current state of a coroutine object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcoroutinestate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1775
 * 
 * 
 * def getcoroutinelocals(coroutine):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of coroutine local variables to their current values.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_145getcoroutinelocals(PyObject *__pyx_self, PyObject *__pyx_v_coroutine); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_144getcoroutinelocals[] = "\n    Get the mapping of coroutine local variables to their current values.\n\n    A dict is returned, with the keys the local variable names and values the\n    bound values.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_145getcoroutinelocals = {"getcoroutinelocals", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_145getcoroutinelocals, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_144getcoroutinelocals};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_145getcoroutinelocals(PyObject *__pyx_self, PyObject *__pyx_v_coroutine) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getcoroutinelocals (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_144getcoroutinelocals(__pyx_self, ((PyObject *)__pyx_v_coroutine));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_144getcoroutinelocals(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coroutine) {
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getcoroutinelocals", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1781
 *     A dict is returned, with the keys the local variable names and values the
 *     bound values."""
 *     frame = getattr(coroutine, "cr_frame", None)             # <<<<<<<<<<<<<<
 *     if frame is not None:
 *         return frame.f_locals
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_coroutine, __pyx_n_u_cr_frame, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_frame = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1782
 *     bound values."""
 *     frame = getattr(coroutine, "cr_frame", None)
 *     if frame is not None:             # <<<<<<<<<<<<<<
 *         return frame.f_locals
 *     else:
 */
  __pyx_t_2 = (__pyx_v_frame != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":1783
 *     frame = getattr(coroutine, "cr_frame", None)
 *     if frame is not None:
 *         return frame.f_locals             # <<<<<<<<<<<<<<
 *     else:
 *         return {}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_n_s_f_locals); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1782
 *     bound values."""
 *     frame = getattr(coroutine, "cr_frame", None)
 *     if frame is not None:             # <<<<<<<<<<<<<<
 *         return frame.f_locals
 *     else:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1785
 *         return frame.f_locals
 *     else:
 *         return {}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1775
 * 
 * 
 * def getcoroutinelocals(coroutine):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of coroutine local variables to their current values.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.getcoroutinelocals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1803
 * 
 * 
 * def _signature_get_user_defined_method(cls, method_name):             # <<<<<<<<<<<<<<
 *     """Private helper. Checks if ``cls`` has an attribute
 *     named ``method_name`` and returns it only if it is a
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_147_signature_get_user_defined_method(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_146_signature_get_user_defined_method[] = "Private helper. Checks if ``cls`` has an attribute\n    named ``method_name`` and returns it only if it is a\n    pure python function.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_147_signature_get_user_defined_method = {"_signature_get_user_defined_method", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_147_signature_get_user_defined_method, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_146_signature_get_user_defined_method};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_147_signature_get_user_defined_method(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_method_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_get_user_defined_method (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_method_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_get_user_defined_method", 1, 2, 2, 1); __PYX_ERR(0, 1803, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_get_user_defined_method") < 0)) __PYX_ERR(0, 1803, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_method_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_get_user_defined_method", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1803, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_user_defined_method", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_146_signature_get_user_defined_method(__pyx_self, __pyx_v_cls, __pyx_v_method_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_146_signature_get_user_defined_method(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_method_name) {
  PyObject *__pyx_v_meth = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_get_user_defined_method", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1808
 *     pure python function.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         meth = getattr(cls, method_name)
 *     except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":1809
 *     """
 *     try:
 *         meth = getattr(cls, method_name)             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         return
 */
      __pyx_t_4 = __Pyx_GetAttr(__pyx_v_cls, __pyx_v_method_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1809, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_meth = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1808
 *     pure python function.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         meth = getattr(cls, method_name)
 *     except AttributeError:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1813
 *         return
 *     else:
 *         if not isinstance(meth, _NonUserDefinedCallables):             # <<<<<<<<<<<<<<
 *             # Once '__signature__' will be added to 'C'-level
 *             # callables, this check won't be necessary
 */
    /*else:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_NonUserDefinedCallables); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1813, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_IsInstance(__pyx_v_meth, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1813, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
      if (__pyx_t_6) {

        /* "Aspidites/_vendor/contracts/inspection.py":1816
 *             # Once '__signature__' will be added to 'C'-level
 *             # callables, this check won't be necessary
 *             return meth             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_meth);
        __pyx_r = __pyx_v_meth;
        goto __pyx_L6_except_return;

        /* "Aspidites/_vendor/contracts/inspection.py":1813
 *         return
 *     else:
 *         if not isinstance(meth, _NonUserDefinedCallables):             # <<<<<<<<<<<<<<
 *             # Once '__signature__' will be added to 'C'-level
 *             # callables, this check won't be necessary
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1810
 *     try:
 *         meth = getattr(cls, method_name)
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         return
 *     else:
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_user_defined_method", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 1810, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);

      /* "Aspidites/_vendor/contracts/inspection.py":1811
 *         meth = getattr(cls, method_name)
 *     except AttributeError:
 *         return             # <<<<<<<<<<<<<<
 *     else:
 *         if not isinstance(meth, _NonUserDefinedCallables):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":1808
 *     pure python function.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         meth = getattr(cls, method_name)
 *     except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1803
 * 
 * 
 * def _signature_get_user_defined_method(cls, method_name):             # <<<<<<<<<<<<<<
 *     """Private helper. Checks if ``cls`` has an attribute
 *     named ``method_name`` and returns it only if it is a
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_user_defined_method", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_meth);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1819
 * 
 * 
 * def _signature_get_partial(wrapped_sig, partial, extra_args=()):             # <<<<<<<<<<<<<<
 *     """Private helper to calculate how 'wrapped_sig' signature will
 *     look like after applying a 'functools.partial' object (or alike)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_149_signature_get_partial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_148_signature_get_partial[] = "Private helper to calculate how 'wrapped_sig' signature will\n    look like after applying a 'functools.partial' object (or alike)\n    on it.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_149_signature_get_partial = {"_signature_get_partial", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_149_signature_get_partial, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_148_signature_get_partial};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_149_signature_get_partial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_wrapped_sig = 0;
  PyObject *__pyx_v_partial = 0;
  PyObject *__pyx_v_extra_args = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_get_partial (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_wrapped_sig,&__pyx_n_s_partial,&__pyx_n_s_extra_args,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject*)__pyx_empty_tuple));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wrapped_sig)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partial)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_get_partial", 0, 2, 3, 1); __PYX_ERR(0, 1819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_extra_args);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_get_partial") < 0)) __PYX_ERR(0, 1819, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_wrapped_sig = values[0];
    __pyx_v_partial = values[1];
    __pyx_v_extra_args = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_get_partial", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1819, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_partial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_148_signature_get_partial(__pyx_self, __pyx_v_wrapped_sig, __pyx_v_partial, __pyx_v_extra_args);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_148_signature_get_partial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_wrapped_sig, PyObject *__pyx_v_partial, PyObject *__pyx_v_extra_args) {
  PyObject *__pyx_v_old_params = NULL;
  PyObject *__pyx_v_new_params = NULL;
  PyObject *__pyx_v_partial_args = NULL;
  PyObject *__pyx_v_partial_keywords = NULL;
  PyObject *__pyx_v_ba = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_msg = NULL;
  int __pyx_v_transform_to_kwonly;
  PyObject *__pyx_v_param_name = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_arg_value = NULL;
  PyObject *__pyx_v_new_param = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_get_partial", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1825
 *     """
 * 
 *     old_params = wrapped_sig.parameters             # <<<<<<<<<<<<<<
 *     new_params = OrderedDict(old_params.items())
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wrapped_sig, __pyx_n_s_parameters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_old_params = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1826
 * 
 *     old_params = wrapped_sig.parameters
 *     new_params = OrderedDict(old_params.items())             # <<<<<<<<<<<<<<
 * 
 *     partial_args = partial.args or ()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_old_params, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_new_params = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1828
 *     new_params = OrderedDict(old_params.items())
 * 
 *     partial_args = partial.args or ()             # <<<<<<<<<<<<<<
 *     partial_keywords = partial.keywords or {}
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_partial, __pyx_n_s_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1828, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_t_1 = __pyx_empty_tuple;
  __pyx_L3_bool_binop_done:;
  __pyx_v_partial_args = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1829
 * 
 *     partial_args = partial.args or ()
 *     partial_keywords = partial.keywords or {}             # <<<<<<<<<<<<<<
 * 
 *     if extra_args:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_partial, __pyx_n_s_keywords); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1829, __pyx_L1_error)
  if (!__pyx_t_6) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_v_partial_keywords = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1831
 *     partial_keywords = partial.keywords or {}
 * 
 *     if extra_args:             # <<<<<<<<<<<<<<
 *         partial_args = extra_args + partial_args
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_extra_args); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1831, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/contracts/inspection.py":1832
 * 
 *     if extra_args:
 *         partial_args = extra_args + partial_args             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_1 = PyNumber_Add(__pyx_v_extra_args, __pyx_v_partial_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_partial_args, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1831
 *     partial_keywords = partial.keywords or {}
 * 
 *     if extra_args:             # <<<<<<<<<<<<<<
 *         partial_args = extra_args + partial_args
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1834
 *         partial_args = extra_args + partial_args
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":1835
 * 
 *     try:
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)             # <<<<<<<<<<<<<<
 *     except TypeError as ex:
 *         msg = 'partial object {!r} has incorrect arguments'.format(partial)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wrapped_sig, __pyx_n_s_bind_partial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1835, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_v_partial_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1835, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_partial_keywords == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
        __PYX_ERR(0, 1835, __pyx_L8_error)
      }
      if (likely(PyDict_CheckExact(__pyx_v_partial_keywords))) {
        __pyx_t_3 = PyDict_Copy(__pyx_v_partial_keywords); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_3 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_partial_keywords, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1835, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_ba = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1834
 *         partial_args = extra_args + partial_args
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1836
 *     try:
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:             # <<<<<<<<<<<<<<
 *         msg = 'partial object {!r} has incorrect arguments'.format(partial)
 *         raise ValueError(msg) from ex
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_partial", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 1836, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1837
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:
 *         msg = 'partial object {!r} has incorrect arguments'.format(partial)             # <<<<<<<<<<<<<<
 *         raise ValueError(msg) from ex
 * 
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_partial_object_r_has_incorrect_a, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1837, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_11, __pyx_v_partial) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_partial);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1837, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_msg = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1838
 *     except TypeError as ex:
 *         msg = 'partial object {!r} has incorrect arguments'.format(partial)
 *         raise ValueError(msg) from ex             # <<<<<<<<<<<<<<
 * 
 *     transform_to_kwonly = False
 */
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1838, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, __pyx_v_ex);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 1838, __pyx_L19_error)
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1836
 *     try:
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:             # <<<<<<<<<<<<<<
 *         msg = 'partial object {!r} has incorrect arguments'.format(partial)
 *         raise ValueError(msg) from ex
 */
      /*finally:*/ {
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_10 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L10_except_error;
        }
      }
    }
    goto __pyx_L10_except_error;
    __pyx_L10_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":1834
 *         partial_args = extra_args + partial_args
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         ba = wrapped_sig.bind_partial(*partial_args, **partial_keywords)
 *     except TypeError as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L13_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1840
 *         raise ValueError(msg) from ex
 * 
 *     transform_to_kwonly = False             # <<<<<<<<<<<<<<
 *     for param_name, param in old_params.items():
 *         try:
 */
  __pyx_v_transform_to_kwonly = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1841
 * 
 *     transform_to_kwonly = False
 *     for param_name, param in old_params.items():             # <<<<<<<<<<<<<<
 *         try:
 *             arg_value = ba.arguments[param_name]
 */
  __pyx_t_20 = 0;
  if (unlikely(__pyx_v_old_params == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1841, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_old_params, 0, __pyx_n_s_items, (&__pyx_t_21), (&__pyx_t_12)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_t_2 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_21, &__pyx_t_20, &__pyx_t_3, &__pyx_t_4, NULL, __pyx_t_12);
    if (unlikely(__pyx_t_10 == 0)) break;
    if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 1841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_param_name, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1842
 *     transform_to_kwonly = False
 *     for param_name, param in old_params.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arg_value = ba.arguments[param_name]
 *         except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":1843
 *     for param_name, param in old_params.items():
 *         try:
 *             arg_value = ba.arguments[param_name]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             pass
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ba, __pyx_n_s_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1843, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_param_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1843, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_arg_value, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1842
 *     transform_to_kwonly = False
 *     for param_name, param in old_params.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arg_value = ba.arguments[param_name]
 *         except KeyError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1847
 *             pass
 *         else:
 *             if param.kind is _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                 # If positional-only parameter is bound by partial,
 *                 # it effectively disappears from the signature
 */
      /*else:*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1847, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1847, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = (__pyx_t_3 == __pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_22 = (__pyx_t_6 != 0);
        if (__pyx_t_22) {

          /* "Aspidites/_vendor/contracts/inspection.py":1850
 *                 # If positional-only parameter is bound by partial,
 *                 # it effectively disappears from the signature
 *                 new_params.pop(param_name)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1850, __pyx_L29_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_param_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_param_name);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1850, __pyx_L29_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1851
 *                 # it effectively disappears from the signature
 *                 new_params.pop(param_name)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 */
          goto __pyx_L33_try_continue;

          /* "Aspidites/_vendor/contracts/inspection.py":1847
 *             pass
 *         else:
 *             if param.kind is _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                 # If positional-only parameter is bound by partial,
 *                 # it effectively disappears from the signature
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1853
 *                 continue
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 if param_name in partial_keywords:
 *                     # This means that this parameter, and all parameters
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1853, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1853, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_22 = (__pyx_t_4 == __pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_6 = (__pyx_t_22 != 0);
        if (__pyx_t_6) {

          /* "Aspidites/_vendor/contracts/inspection.py":1854
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 *                 if param_name in partial_keywords:             # <<<<<<<<<<<<<<
 *                     # This means that this parameter, and all parameters
 *                     # after it should be keyword-only (and var-positional
 */
          __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_param_name, __pyx_v_partial_keywords, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1854, __pyx_L29_except_error)
          __pyx_t_22 = (__pyx_t_6 != 0);
          if (__pyx_t_22) {

            /* "Aspidites/_vendor/contracts/inspection.py":1867
 *                     # raise a TypeError, saying that "a" argument received
 *                     # multiple values.
 *                     transform_to_kwonly = True             # <<<<<<<<<<<<<<
 *                     # Set the new default value
 *                     new_params[param_name] = param.replace(default=arg_value)
 */
            __pyx_v_transform_to_kwonly = 1;

            /* "Aspidites/_vendor/contracts/inspection.py":1869
 *                     transform_to_kwonly = True
 *                     # Set the new default value
 *                     new_params[param_name] = param.replace(default=arg_value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     # was passed as a positional argument
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1869, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1869, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, __pyx_v_arg_value) < 0) __PYX_ERR(0, 1869, __pyx_L29_except_error)
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1869, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(PyObject_SetItem(__pyx_v_new_params, __pyx_v_param_name, __pyx_t_1) < 0)) __PYX_ERR(0, 1869, __pyx_L29_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":1854
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 *                 if param_name in partial_keywords:             # <<<<<<<<<<<<<<
 *                     # This means that this parameter, and all parameters
 *                     # after it should be keyword-only (and var-positional
 */
            goto __pyx_L37;
          }

          /* "Aspidites/_vendor/contracts/inspection.py":1872
 *                 else:
 *                     # was passed as a positional argument
 *                     new_params.pop(param.name)             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
          /*else*/ {
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1872, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1872, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":1873
 *                     # was passed as a positional argument
 *                     new_params.pop(param.name)
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *             if param.kind is _KEYWORD_ONLY:
 */
            goto __pyx_L33_try_continue;
          }
          __pyx_L37:;

          /* "Aspidites/_vendor/contracts/inspection.py":1853
 *                 continue
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 if param_name in partial_keywords:
 *                     # This means that this parameter, and all parameters
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":1875
 *                     continue
 * 
 *             if param.kind is _KEYWORD_ONLY:             # <<<<<<<<<<<<<<
 *                 # Set the new default value
 *                 new_params[param_name] = param.replace(default=arg_value)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1875, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1875, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_22 = (__pyx_t_1 == __pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = (__pyx_t_22 != 0);
        if (__pyx_t_6) {

          /* "Aspidites/_vendor/contracts/inspection.py":1877
 *             if param.kind is _KEYWORD_ONLY:
 *                 # Set the new default value
 *                 new_params[param_name] = param.replace(default=arg_value)             # <<<<<<<<<<<<<<
 * 
 *         if transform_to_kwonly:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1877, __pyx_L29_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1877, __pyx_L29_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_default, __pyx_v_arg_value) < 0) __PYX_ERR(0, 1877, __pyx_L29_except_error)
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1877, __pyx_L29_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_new_params, __pyx_v_param_name, __pyx_t_3) < 0)) __PYX_ERR(0, 1877, __pyx_L29_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":1875
 *                     continue
 * 
 *             if param.kind is _KEYWORD_ONLY:             # <<<<<<<<<<<<<<
 *                 # Set the new default value
 *                 new_params[param_name] = param.replace(default=arg_value)
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L34_try_end;
      __pyx_L27_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":1844
 *         try:
 *             arg_value = ba.arguments[param_name]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_10) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L28_exception_handled;
      }
      goto __pyx_L29_except_error;
      __pyx_L29_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":1842
 *     transform_to_kwonly = False
 *     for param_name, param in old_params.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arg_value = ba.arguments[param_name]
 *         except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L33_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L25_continue;
      __pyx_L28_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      __pyx_L34_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":1879
 *                 new_params[param_name] = param.replace(default=arg_value)
 * 
 *         if transform_to_kwonly:             # <<<<<<<<<<<<<<
 *             assert param.kind is not _POSITIONAL_ONLY
 * 
 */
    __pyx_t_6 = (__pyx_v_transform_to_kwonly != 0);
    if (__pyx_t_6) {

      /* "Aspidites/_vendor/contracts/inspection.py":1880
 * 
 *         if transform_to_kwonly:
 *             assert param.kind is not _POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = (__pyx_t_3 != __pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!(__pyx_t_6 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 1880, __pyx_L1_error)
        }
      }
      #endif

      /* "Aspidites/_vendor/contracts/inspection.py":1882
 *             assert param.kind is not _POSITIONAL_ONLY
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 new_param = new_params[param_name].replace(kind=_KEYWORD_ONLY)
 *                 new_params[param_name] = new_param
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = (__pyx_t_1 == __pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_22 = (__pyx_t_6 != 0);
      if (__pyx_t_22) {

        /* "Aspidites/_vendor/contracts/inspection.py":1883
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 *                 new_param = new_params[param_name].replace(kind=_KEYWORD_ONLY)             # <<<<<<<<<<<<<<
 *                 new_params[param_name] = new_param
 *                 new_params.move_to_end(param_name)
 */
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_new_params, __pyx_v_param_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_kind, __pyx_t_4) < 0) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_new_param, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1884
 *             if param.kind is _POSITIONAL_OR_KEYWORD:
 *                 new_param = new_params[param_name].replace(kind=_KEYWORD_ONLY)
 *                 new_params[param_name] = new_param             # <<<<<<<<<<<<<<
 *                 new_params.move_to_end(param_name)
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):
 */
        if (unlikely(PyObject_SetItem(__pyx_v_new_params, __pyx_v_param_name, __pyx_v_new_param) < 0)) __PYX_ERR(0, 1884, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":1885
 *                 new_param = new_params[param_name].replace(kind=_KEYWORD_ONLY)
 *                 new_params[param_name] = new_param
 *                 new_params.move_to_end(param_name)             # <<<<<<<<<<<<<<
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):
 *                 new_params.move_to_end(param_name)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_move_to_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_param_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_param_name);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1882
 *             assert param.kind is not _POSITIONAL_ONLY
 * 
 *             if param.kind is _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 new_param = new_params[param_name].replace(kind=_KEYWORD_ONLY)
 *                 new_params[param_name] = new_param
 */
        goto __pyx_L40;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1886
 *                 new_params[param_name] = new_param
 *                 new_params.move_to_end(param_name)
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):             # <<<<<<<<<<<<<<
 *                 new_params.move_to_end(param_name)
 *             elif param.kind is _VAR_POSITIONAL:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_6) {
      } else {
        __pyx_t_22 = __pyx_t_6;
        goto __pyx_L41_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_22 = __pyx_t_6;
      __pyx_L41_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = (__pyx_t_22 != 0);
      if (__pyx_t_6) {

        /* "Aspidites/_vendor/contracts/inspection.py":1887
 *                 new_params.move_to_end(param_name)
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):
 *                 new_params.move_to_end(param_name)             # <<<<<<<<<<<<<<
 *             elif param.kind is _VAR_POSITIONAL:
 *                 new_params.pop(param.name)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_move_to_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_param_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_param_name);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1886
 *                 new_params[param_name] = new_param
 *                 new_params.move_to_end(param_name)
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):             # <<<<<<<<<<<<<<
 *                 new_params.move_to_end(param_name)
 *             elif param.kind is _VAR_POSITIONAL:
 */
        goto __pyx_L40;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":1888
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):
 *                 new_params.move_to_end(param_name)
 *             elif param.kind is _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 new_params.pop(param.name)
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = (__pyx_t_4 == __pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_22 = (__pyx_t_6 != 0);
      if (__pyx_t_22) {

        /* "Aspidites/_vendor/contracts/inspection.py":1889
 *                 new_params.move_to_end(param_name)
 *             elif param.kind is _VAR_POSITIONAL:
 *                 new_params.pop(param.name)             # <<<<<<<<<<<<<<
 * 
 *     return wrapped_sig.replace(parameters=new_params.values())
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":1888
 *             elif param.kind in (_KEYWORD_ONLY, _VAR_KEYWORD):
 *                 new_params.move_to_end(param_name)
 *             elif param.kind is _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 new_params.pop(param.name)
 * 
 */
      }
      __pyx_L40:;

      /* "Aspidites/_vendor/contracts/inspection.py":1879
 *                 new_params[param_name] = param.replace(default=arg_value)
 * 
 *         if transform_to_kwonly:             # <<<<<<<<<<<<<<
 *             assert param.kind is not _POSITIONAL_ONLY
 * 
 */
    }
    __pyx_L25_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1891
 *                 new_params.pop(param.name)
 * 
 *     return wrapped_sig.replace(parameters=new_params.values())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_wrapped_sig, __pyx_n_s_replace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_params, __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_parameters, __pyx_t_4) < 0) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1819
 * 
 * 
 * def _signature_get_partial(wrapped_sig, partial, extra_args=()):             # <<<<<<<<<<<<<<
 *     """Private helper to calculate how 'wrapped_sig' signature will
 *     look like after applying a 'functools.partial' object (or alike)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_partial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_old_params);
  __Pyx_XDECREF(__pyx_v_new_params);
  __Pyx_XDECREF(__pyx_v_partial_args);
  __Pyx_XDECREF(__pyx_v_partial_keywords);
  __Pyx_XDECREF(__pyx_v_ba);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_param_name);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_arg_value);
  __Pyx_XDECREF(__pyx_v_new_param);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1894
 * 
 * 
 * def _signature_bound_method(sig):             # <<<<<<<<<<<<<<
 *     """Private helper to transform signatures for unbound
 *     functions to bound methods.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_151_signature_bound_method(PyObject *__pyx_self, PyObject *__pyx_v_sig); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_150_signature_bound_method[] = "Private helper to transform signatures for unbound\n    functions to bound methods.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_151_signature_bound_method = {"_signature_bound_method", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_151_signature_bound_method, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_150_signature_bound_method};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_151_signature_bound_method(PyObject *__pyx_self, PyObject *__pyx_v_sig) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_bound_method (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_150_signature_bound_method(__pyx_self, ((PyObject *)__pyx_v_sig));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_150_signature_bound_method(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sig) {
  PyObject *__pyx_v_params = NULL;
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_bound_method", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1899
 *     """
 * 
 *     params = tuple(sig.parameters.values())             # <<<<<<<<<<<<<<
 * 
 *     if not params or params[0].kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_parameters); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_params = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1901
 *     params = tuple(sig.parameters.values())
 * 
 *     if not params or params[0].kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid method signature')
 * 
 */
  __pyx_t_5 = (PyTuple_GET_SIZE(__pyx_v_params) != 0);
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(PyTuple_GET_ITEM(__pyx_v_params, 0), __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_6 = __pyx_t_5;
    goto __pyx_L6_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __pyx_t_5;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/contracts/inspection.py":1902
 * 
 *     if not params or params[0].kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *         raise ValueError('invalid method signature')             # <<<<<<<<<<<<<<
 * 
 *     kind = params[0].kind
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1902, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":1901
 *     params = tuple(sig.parameters.values())
 * 
 *     if not params or params[0].kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid method signature')
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1904
 *         raise ValueError('invalid method signature')
 * 
 *     kind = params[0].kind             # <<<<<<<<<<<<<<
 *     if kind in (_POSITIONAL_OR_KEYWORD, _POSITIONAL_ONLY):
 *         # Drop first parameter:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(PyTuple_GET_ITEM(__pyx_v_params, 0), __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_kind = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1905
 * 
 *     kind = params[0].kind
 *     if kind in (_POSITIONAL_OR_KEYWORD, _POSITIONAL_ONLY):             # <<<<<<<<<<<<<<
 *         # Drop first parameter:
 *         # '(p1, p2[, ...])' -> '(p2[, ...])'
 */
  __Pyx_INCREF(__pyx_v_kind);
  __pyx_t_3 = __pyx_v_kind;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L9_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L9_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":1908
 *         # Drop first parameter:
 *         # '(p1, p2[, ...])' -> '(p2[, ...])'
 *         params = params[1:]             # <<<<<<<<<<<<<<
 *     else:
 *         if kind is not _VAR_POSITIONAL:
 */
    __pyx_t_3 = __Pyx_PyTuple_GetSlice(__pyx_v_params, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_params, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1905
 * 
 *     kind = params[0].kind
 *     if kind in (_POSITIONAL_OR_KEYWORD, _POSITIONAL_ONLY):             # <<<<<<<<<<<<<<
 *         # Drop first parameter:
 *         # '(p1, p2[, ...])' -> '(p2[, ...])'
 */
    goto __pyx_L8;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1910
 *         params = params[1:]
 *     else:
 *         if kind is not _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *             # Unless we add a new parameter type we never
 *             # get here
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__pyx_v_kind != __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (unlikely(__pyx_t_4)) {

      /* "Aspidites/_vendor/contracts/inspection.py":1913
 *             # Unless we add a new parameter type we never
 *             # get here
 *             raise ValueError('invalid argument type')             # <<<<<<<<<<<<<<
 *         # It's a var-positional parameter.
 *         # Do nothing. '(*args[, ...])' -> '(*args[, ...])'
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1913, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1913, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":1910
 *         params = params[1:]
 *     else:
 *         if kind is not _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *             # Unless we add a new parameter type we never
 *             # get here
 */
    }
  }
  __pyx_L8:;

  /* "Aspidites/_vendor/contracts/inspection.py":1917
 *         # Do nothing. '(*args[, ...])' -> '(*args[, ...])'
 * 
 *     return sig.replace(parameters=params)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_parameters, __pyx_v_params) < 0) __PYX_ERR(0, 1917, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1894
 * 
 * 
 * def _signature_bound_method(sig):             # <<<<<<<<<<<<<<
 *     """Private helper to transform signatures for unbound
 *     functions to bound methods.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_bound_method", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1920
 * 
 * 
 * def _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a callable that might
 *     support Argument Clinic's __text_signature__ protocol.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_153_signature_is_builtin(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_152_signature_is_builtin[] = "Private helper to test if `obj` is a callable that might\n    support Argument Clinic's __text_signature__ protocol.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_153_signature_is_builtin = {"_signature_is_builtin", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_153_signature_is_builtin, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_152_signature_is_builtin};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_153_signature_is_builtin(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_is_builtin (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_152_signature_is_builtin(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_152_signature_is_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_is_builtin", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1924
 *     support Argument Clinic's __text_signature__ protocol.
 *     """
 *     return (isbuiltin(obj) or             # <<<<<<<<<<<<<<
 *             ismethoddescriptor(obj) or
 *             isinstance(obj, _NonUserDefinedCallables) or
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isbuiltin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1924, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1925
 *     """
 *     return (isbuiltin(obj) or
 *             ismethoddescriptor(obj) or             # <<<<<<<<<<<<<<
 *             isinstance(obj, _NonUserDefinedCallables) or
 *             # Can't test 'isinstance(type)' here, as it would
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ismethoddescriptor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1925, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1926
 *     return (isbuiltin(obj) or
 *             ismethoddescriptor(obj) or
 *             isinstance(obj, _NonUserDefinedCallables) or             # <<<<<<<<<<<<<<
 *             # Can't test 'isinstance(type)' here, as it would
 *             # also be True for regular python classes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NonUserDefinedCallables); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1929
 *             # Can't test 'isinstance(type)' here, as it would
 *             # also be True for regular python classes
 *             obj in (type, object))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_t_2 = __pyx_v_obj;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, ((PyObject *)(&PyType_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_builtin_object, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_5;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1920
 * 
 * 
 * def _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a callable that might
 *     support Argument Clinic's __text_signature__ protocol.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_is_builtin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1932
 * 
 * 
 * def _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a duck type of FunctionType.
 *     A good example of such objects are functions compiled with
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_155_signature_is_functionlike(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_154_signature_is_functionlike[] = "Private helper to test if `obj` is a duck type of FunctionType.\n    A good example of such objects are functions compiled with\n    Cython, which have all attributes that a pure Python function\n    would have, but have their code statically compiled.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_155_signature_is_functionlike = {"_signature_is_functionlike", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_155_signature_is_functionlike, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_154_signature_is_functionlike};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_155_signature_is_functionlike(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_is_functionlike (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_154_signature_is_functionlike(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_154_signature_is_functionlike(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_code = NULL;
  PyObject *__pyx_v_defaults = NULL;
  PyObject *__pyx_v_kwdefaults = NULL;
  PyObject *__pyx_v_annotations = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_is_functionlike", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1939
 *     """
 * 
 *     if not callable(obj) or isclass(obj):             # <<<<<<<<<<<<<<
 *         # All function-like objects are obviously callables,
 *         # and not classes.
 */
  __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_v_obj); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1939, __pyx_L1_error)
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_isclass); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":1942
 *         # All function-like objects are obviously callables,
 *         # and not classes.
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     name = getattr(obj, '__name__', None)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1939
 *     """
 * 
 *     if not callable(obj) or isclass(obj):             # <<<<<<<<<<<<<<
 *         # All function-like objects are obviously callables,
 *         # and not classes.
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1944
 *         return False
 * 
 *     name = getattr(obj, '__name__', None)             # <<<<<<<<<<<<<<
 *     code = getattr(obj, '__code__', None)
 *     defaults = getattr(obj, '__defaults__', _void)  # Important to use _void ...
 */
  __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_name, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_name = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1945
 * 
 *     name = getattr(obj, '__name__', None)
 *     code = getattr(obj, '__code__', None)             # <<<<<<<<<<<<<<
 *     defaults = getattr(obj, '__defaults__', _void)  # Important to use _void ...
 *     kwdefaults = getattr(obj, '__kwdefaults__', _void)  # ... and not None here
 */
  __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_code, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_code = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1946
 *     name = getattr(obj, '__name__', None)
 *     code = getattr(obj, '__code__', None)
 *     defaults = getattr(obj, '__defaults__', _void)  # Important to use _void ...             # <<<<<<<<<<<<<<
 *     kwdefaults = getattr(obj, '__kwdefaults__', _void)  # ... and not None here
 *     annotations = getattr(obj, '__annotations__', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_void); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_defaults_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_defaults = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1947
 *     code = getattr(obj, '__code__', None)
 *     defaults = getattr(obj, '__defaults__', _void)  # Important to use _void ...
 *     kwdefaults = getattr(obj, '__kwdefaults__', _void)  # ... and not None here             # <<<<<<<<<<<<<<
 *     annotations = getattr(obj, '__annotations__', None)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_void); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_kwdefaults, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_kwdefaults = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1948
 *     defaults = getattr(obj, '__defaults__', _void)  # Important to use _void ...
 *     kwdefaults = getattr(obj, '__kwdefaults__', _void)  # ... and not None here
 *     annotations = getattr(obj, '__annotations__', None)             # <<<<<<<<<<<<<<
 * 
 *     return (isinstance(code, types.CodeType) and
 */
  __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_annotations_2, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_annotations = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1950
 *     annotations = getattr(obj, '__annotations__', None)
 * 
 *     return (isinstance(code, types.CodeType) and             # <<<<<<<<<<<<<<
 *             isinstance(name, str) and
 *             (defaults is None or isinstance(defaults, tuple)) and
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_CodeType); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_code, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1951
 * 
 *     return (isinstance(code, types.CodeType) and
 *             isinstance(name, str) and             # <<<<<<<<<<<<<<
 *             (defaults is None or isinstance(defaults, tuple)) and
 *             (kwdefaults is None or isinstance(kwdefaults, dict)) and
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_name); 
  if (__pyx_t_1) {
  } else {
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1952
 *     return (isinstance(code, types.CodeType) and
 *             isinstance(name, str) and
 *             (defaults is None or isinstance(defaults, tuple)) and             # <<<<<<<<<<<<<<
 *             (kwdefaults is None or isinstance(kwdefaults, dict)) and
 *             isinstance(annotations, dict))
 */
  __pyx_t_1 = (__pyx_v_defaults == Py_None);
  if (!__pyx_t_1) {
  } else {
    goto __pyx_L9_next_and;
  }
  __pyx_t_1 = PyTuple_Check(__pyx_v_defaults); 
  if (__pyx_t_1) {
  } else {
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_L9_next_and:;

  /* "Aspidites/_vendor/contracts/inspection.py":1953
 *             isinstance(name, str) and
 *             (defaults is None or isinstance(defaults, tuple)) and
 *             (kwdefaults is None or isinstance(kwdefaults, dict)) and             # <<<<<<<<<<<<<<
 *             isinstance(annotations, dict))
 * 
 */
  __pyx_t_1 = (__pyx_v_kwdefaults == Py_None);
  if (!__pyx_t_1) {
  } else {
    goto __pyx_L11_next_and;
  }
  __pyx_t_1 = PyDict_Check(__pyx_v_kwdefaults); 
  if (__pyx_t_1) {
  } else {
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_L11_next_and:;

  /* "Aspidites/_vendor/contracts/inspection.py":1954
 *             (defaults is None or isinstance(defaults, tuple)) and
 *             (kwdefaults is None or isinstance(kwdefaults, dict)) and
 *             isinstance(annotations, dict))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_annotations); 
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_L6_bool_binop_done:;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1932
 * 
 * 
 * def _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a duck type of FunctionType.
 *     A good example of such objects are functions compiled with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_is_functionlike", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_code);
  __Pyx_XDECREF(__pyx_v_defaults);
  __Pyx_XDECREF(__pyx_v_kwdefaults);
  __Pyx_XDECREF(__pyx_v_annotations);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1957
 * 
 * 
 * def _signature_get_bound_param(spec):             # <<<<<<<<<<<<<<
 *     """ Private helper to get first parameter name from a
 *     __text_signature__ of a builtin method, which should
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_157_signature_get_bound_param(PyObject *__pyx_self, PyObject *__pyx_v_spec); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_156_signature_get_bound_param[] = " Private helper to get first parameter name from a\n    __text_signature__ of a builtin method, which should\n    be in the following format: '($param1, ...)'.\n    Assumptions are that the first argument won't have\n    a default value or an annotation.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_157_signature_get_bound_param = {"_signature_get_bound_param", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_157_signature_get_bound_param, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_156_signature_get_bound_param};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_157_signature_get_bound_param(PyObject *__pyx_self, PyObject *__pyx_v_spec) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_get_bound_param (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_156_signature_get_bound_param(__pyx_self, ((PyObject *)__pyx_v_spec));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_156_signature_get_bound_param(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_spec) {
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_v_cpos = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_get_bound_param", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1965
 *     """
 * 
 *     assert spec.startswith('($')             # <<<<<<<<<<<<<<
 * 
 *     pos = spec.find(',')
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_startswith); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__55) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__55);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1965, __pyx_L1_error)
    }
  }
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":1967
 *     assert spec.startswith('($')
 * 
 *     pos = spec.find(',')             # <<<<<<<<<<<<<<
 *     if pos == -1:
 *         pos = spec.find(')')
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__56) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__56);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pos = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1968
 * 
 *     pos = spec.find(',')
 *     if pos == -1:             # <<<<<<<<<<<<<<
 *         pos = spec.find(')')
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_pos, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1968, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":1969
 *     pos = spec.find(',')
 *     if pos == -1:
 *         pos = spec.find(')')             # <<<<<<<<<<<<<<
 * 
 *     cpos = spec.find(':')
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__30) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__30);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_pos, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":1968
 * 
 *     pos = spec.find(',')
 *     if pos == -1:             # <<<<<<<<<<<<<<
 *         pos = spec.find(')')
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1971
 *         pos = spec.find(')')
 * 
 *     cpos = spec.find(':')             # <<<<<<<<<<<<<<
 *     assert cpos == -1 or cpos > pos
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__57) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__57);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cpos = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1972
 * 
 *     cpos = spec.find(':')
 *     assert cpos == -1 or cpos > pos             # <<<<<<<<<<<<<<
 * 
 *     cpos = spec.find('=')
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_cpos, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1972, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1972, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_cpos, __pyx_v_pos, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1972, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1972, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1972, __pyx_L1_error)
    }
  }
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":1974
 *     assert cpos == -1 or cpos > pos
 * 
 *     cpos = spec.find('=')             # <<<<<<<<<<<<<<
 *     assert cpos == -1 or cpos > pos
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__3);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_cpos, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1975
 * 
 *     cpos = spec.find('=')
 *     assert cpos == -1 or cpos > pos             # <<<<<<<<<<<<<<
 * 
 *     return spec[2:pos]
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_cpos, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1975, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_cpos, __pyx_v_pos, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1975, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1975, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1975, __pyx_L1_error)
    }
  }
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":1977
 *     assert cpos == -1 or cpos > pos
 * 
 *     return spec[2:pos]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_spec, 2, 0, NULL, &__pyx_v_pos, NULL, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1957
 * 
 * 
 * def _signature_get_bound_param(spec):             # <<<<<<<<<<<<<<
 *     """ Private helper to get first parameter name from a
 *     __text_signature__ of a builtin method, which should
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_get_bound_param", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_cpos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":1980
 * 
 * 
 * def _signature_strip_non_python_syntax(signature):             # <<<<<<<<<<<<<<
 *     """
 *     Private helper function. Takes a signature in Argument Clinic's
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_159_signature_strip_non_python_syntax(PyObject *__pyx_self, PyObject *__pyx_v_signature); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_158_signature_strip_non_python_syntax[] = "\n    Private helper function. Takes a signature in Argument Clinic's\n    extended signature format.\n\n    Returns a tuple of three things:\n      * that signature re-rendered in standard Python syntax,\n      * the index of the \"self\" parameter (generally 0), or None if\n        the function does not have a \"self\" parameter, and\n      * the index of the last \"positional only\" parameter,\n        or None if the signature has no positional-only parameters.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_159_signature_strip_non_python_syntax = {"_signature_strip_non_python_syntax", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_159_signature_strip_non_python_syntax, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_158_signature_strip_non_python_syntax};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_159_signature_strip_non_python_syntax(PyObject *__pyx_self, PyObject *__pyx_v_signature) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_strip_non_python_syntax (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_158_signature_strip_non_python_syntax(__pyx_self, ((PyObject *)__pyx_v_signature));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_158_signature_strip_non_python_syntax(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signature) {
  PyObject *__pyx_v_self_parameter = NULL;
  PyObject *__pyx_v_last_positional_only = NULL;
  PyObject *__pyx_v_lines = NULL;
  PyObject *__pyx_v_generator = NULL;
  PyObject *__pyx_v_token_stream = NULL;
  int __pyx_v_delayed_comma;
  int __pyx_v_skip_next_comma;
  PyObject *__pyx_v_text = NULL;
  PyObject *__pyx_v_add = NULL;
  PyObject *__pyx_v_current_parameter = NULL;
  PyObject *__pyx_v_OP = NULL;
  PyObject *__pyx_v_ERRORTOKEN = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_type = NULL;
  PyObject *__pyx_v_string = NULL;
  PyObject *__pyx_v_clean_signature = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_l = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_strip_non_python_syntax", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":1993
 *     """
 * 
 *     if not signature:             # <<<<<<<<<<<<<<
 *         return signature, None, None
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_signature); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":1994
 * 
 *     if not signature:
 *         return signature, None, None             # <<<<<<<<<<<<<<
 * 
 *     self_parameter = None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_signature);
    __Pyx_GIVEREF(__pyx_v_signature);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_signature);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":1993
 *     """
 * 
 *     if not signature:             # <<<<<<<<<<<<<<
 *         return signature, None, None
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":1996
 *         return signature, None, None
 * 
 *     self_parameter = None             # <<<<<<<<<<<<<<
 *     last_positional_only = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_self_parameter = Py_None;

  /* "Aspidites/_vendor/contracts/inspection.py":1997
 * 
 *     self_parameter = None
 *     last_positional_only = None             # <<<<<<<<<<<<<<
 * 
 *     lines = [l.encode('ascii') for l in signature.split('\n')]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_last_positional_only = Py_None;

  /* "Aspidites/_vendor/contracts/inspection.py":1999
 *     last_positional_only = None
 * 
 *     lines = [l.encode('ascii') for l in signature.split('\n')]             # <<<<<<<<<<<<<<
 *     generator = iter(lines).__next__
 *     token_stream = tokenize.tokenize(generator)
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_signature, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1999, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u__12) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__12);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1999, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1999, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1999, __pyx_L6_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1999, __pyx_L6_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_8(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1999, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_l, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr8__pyx_v_l, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1999, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_n_u_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_ascii);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1999, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_l); __pyx_8genexpr8__pyx_v_l = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_l); __pyx_8genexpr8__pyx_v_l = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_lines = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2000
 * 
 *     lines = [l.encode('ascii') for l in signature.split('\n')]
 *     generator = iter(lines).__next__             # <<<<<<<<<<<<<<
 *     token_stream = tokenize.tokenize(generator)
 * 
 */
  __pyx_t_3 = PyObject_GetIter(__pyx_v_lines); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_next); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_generator = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2001
 *     lines = [l.encode('ascii') for l in signature.split('\n')]
 *     generator = iter(lines).__next__
 *     token_stream = tokenize.tokenize(generator)             # <<<<<<<<<<<<<<
 * 
 *     delayed_comma = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_generator) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_generator);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_token_stream = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2003
 *     token_stream = tokenize.tokenize(generator)
 * 
 *     delayed_comma = False             # <<<<<<<<<<<<<<
 *     skip_next_comma = False
 *     text = []
 */
  __pyx_v_delayed_comma = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2004
 * 
 *     delayed_comma = False
 *     skip_next_comma = False             # <<<<<<<<<<<<<<
 *     text = []
 *     add = text.append
 */
  __pyx_v_skip_next_comma = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2005
 *     delayed_comma = False
 *     skip_next_comma = False
 *     text = []             # <<<<<<<<<<<<<<
 *     add = text.append
 * 
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_text = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2006
 *     skip_next_comma = False
 *     text = []
 *     add = text.append             # <<<<<<<<<<<<<<
 * 
 *     current_parameter = 0
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_text, __pyx_n_s_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_add = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2008
 *     add = text.append
 * 
 *     current_parameter = 0             # <<<<<<<<<<<<<<
 *     OP = token.OP
 *     ERRORTOKEN = token.ERRORTOKEN
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_current_parameter = __pyx_int_0;

  /* "Aspidites/_vendor/contracts/inspection.py":2009
 * 
 *     current_parameter = 0
 *     OP = token.OP             # <<<<<<<<<<<<<<
 *     ERRORTOKEN = token.ERRORTOKEN
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_token); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_OP); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_OP = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2010
 *     current_parameter = 0
 *     OP = token.OP
 *     ERRORTOKEN = token.ERRORTOKEN             # <<<<<<<<<<<<<<
 * 
 *     # token stream always starts with ENCODING token, skip it
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_token); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ERRORTOKEN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_ERRORTOKEN = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2013
 * 
 *     # token stream always starts with ENCODING token, skip it
 *     t = next(token_stream)             # <<<<<<<<<<<<<<
 *     assert t.type == tokenize.ENCODING
 * 
 */
  __pyx_t_5 = __Pyx_PyIter_Next(__pyx_v_token_stream); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_t = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2014
 *     # token stream always starts with ENCODING token, skip it
 *     t = next(token_stream)
 *     assert t.type == tokenize.ENCODING             # <<<<<<<<<<<<<<
 * 
 *     for t in token_stream:
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_tokenize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ENCODING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 2014, __pyx_L1_error)
    }
  }
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":2016
 *     assert t.type == tokenize.ENCODING
 * 
 *     for t in token_stream:             # <<<<<<<<<<<<<<
 *         type, string = t.type, t.string
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_token_stream)) || PyTuple_CheckExact(__pyx_v_token_stream)) {
    __pyx_t_4 = __pyx_v_token_stream; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_token_stream); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2016, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2016, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 2016, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2016, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2017
 * 
 *     for t in token_stream:
 *         type, string = t.type, t.string             # <<<<<<<<<<<<<<
 * 
 *         if type == OP:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_type, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_string, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2019
 *         type, string = t.type, t.string
 * 
 *         if type == OP:             # <<<<<<<<<<<<<<
 *             if string == ',':
 *                 if skip_next_comma:
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_type, __pyx_v_OP, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2019, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2019, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2020
 * 
 *         if type == OP:
 *             if string == ',':             # <<<<<<<<<<<<<<
 *                 if skip_next_comma:
 *                     skip_next_comma = False
 */
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_string, __pyx_kp_u__56, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2020, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/contracts/inspection.py":2021
 *         if type == OP:
 *             if string == ',':
 *                 if skip_next_comma:             # <<<<<<<<<<<<<<
 *                     skip_next_comma = False
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_skip_next_comma != 0);
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/contracts/inspection.py":2022
 *             if string == ',':
 *                 if skip_next_comma:
 *                     skip_next_comma = False             # <<<<<<<<<<<<<<
 *                 else:
 *                     assert not delayed_comma
 */
          __pyx_v_skip_next_comma = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2021
 *         if type == OP:
 *             if string == ',':
 *                 if skip_next_comma:             # <<<<<<<<<<<<<<
 *                     skip_next_comma = False
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2024
 *                     skip_next_comma = False
 *                 else:
 *                     assert not delayed_comma             # <<<<<<<<<<<<<<
 *                     delayed_comma = True
 *                     current_parameter += 1
 */
        /*else*/ {
          #ifndef CYTHON_WITHOUT_ASSERTIONS
          if (unlikely(!Py_OptimizeFlag)) {
            if (unlikely(!((!(__pyx_v_delayed_comma != 0)) != 0))) {
              PyErr_SetNone(PyExc_AssertionError);
              __PYX_ERR(0, 2024, __pyx_L1_error)
            }
          }
          #endif

          /* "Aspidites/_vendor/contracts/inspection.py":2025
 *                 else:
 *                     assert not delayed_comma
 *                     delayed_comma = True             # <<<<<<<<<<<<<<
 *                     current_parameter += 1
 *                 continue
 */
          __pyx_v_delayed_comma = 1;

          /* "Aspidites/_vendor/contracts/inspection.py":2026
 *                     assert not delayed_comma
 *                     delayed_comma = True
 *                     current_parameter += 1             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
          __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_current_parameter, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2026, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF_SET(__pyx_v_current_parameter, __pyx_t_5);
          __pyx_t_5 = 0;
        }
        __pyx_L14:;

        /* "Aspidites/_vendor/contracts/inspection.py":2027
 *                     delayed_comma = True
 *                     current_parameter += 1
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             if string == '/':
 */
        goto __pyx_L10_continue;

        /* "Aspidites/_vendor/contracts/inspection.py":2020
 * 
 *         if type == OP:
 *             if string == ',':             # <<<<<<<<<<<<<<
 *                 if skip_next_comma:
 *                     skip_next_comma = False
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2029
 *                 continue
 * 
 *             if string == '/':             # <<<<<<<<<<<<<<
 *                 assert not skip_next_comma
 *                 assert last_positional_only is None
 */
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_string, __pyx_kp_u__58, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2029, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/contracts/inspection.py":2030
 * 
 *             if string == '/':
 *                 assert not skip_next_comma             # <<<<<<<<<<<<<<
 *                 assert last_positional_only is None
 *                 skip_next_comma = True
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(!Py_OptimizeFlag)) {
          if (unlikely(!((!(__pyx_v_skip_next_comma != 0)) != 0))) {
            PyErr_SetNone(PyExc_AssertionError);
            __PYX_ERR(0, 2030, __pyx_L1_error)
          }
        }
        #endif

        /* "Aspidites/_vendor/contracts/inspection.py":2031
 *             if string == '/':
 *                 assert not skip_next_comma
 *                 assert last_positional_only is None             # <<<<<<<<<<<<<<
 *                 skip_next_comma = True
 *                 last_positional_only = current_parameter - 1
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(!Py_OptimizeFlag)) {
          __pyx_t_2 = (__pyx_v_last_positional_only == Py_None);
          if (unlikely(!(__pyx_t_2 != 0))) {
            PyErr_SetNone(PyExc_AssertionError);
            __PYX_ERR(0, 2031, __pyx_L1_error)
          }
        }
        #endif

        /* "Aspidites/_vendor/contracts/inspection.py":2032
 *                 assert not skip_next_comma
 *                 assert last_positional_only is None
 *                 skip_next_comma = True             # <<<<<<<<<<<<<<
 *                 last_positional_only = current_parameter - 1
 *                 continue
 */
        __pyx_v_skip_next_comma = 1;

        /* "Aspidites/_vendor/contracts/inspection.py":2033
 *                 assert last_positional_only is None
 *                 skip_next_comma = True
 *                 last_positional_only = current_parameter - 1             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_v_current_parameter, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_last_positional_only, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2034
 *                 skip_next_comma = True
 *                 last_positional_only = current_parameter - 1
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *         if (type == ERRORTOKEN) and (string == '$'):
 */
        goto __pyx_L10_continue;

        /* "Aspidites/_vendor/contracts/inspection.py":2029
 *                 continue
 * 
 *             if string == '/':             # <<<<<<<<<<<<<<
 *                 assert not skip_next_comma
 *                 assert last_positional_only is None
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2019
 *         type, string = t.type, t.string
 * 
 *         if type == OP:             # <<<<<<<<<<<<<<
 *             if string == ',':
 *                 if skip_next_comma:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2036
 *                 continue
 * 
 *         if (type == ERRORTOKEN) and (string == '$'):             # <<<<<<<<<<<<<<
 *             assert self_parameter is None
 *             self_parameter = current_parameter
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_type, __pyx_v_ERRORTOKEN, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2036, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2036, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_string, __pyx_kp_u__59, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2036, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_1;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2037
 * 
 *         if (type == ERRORTOKEN) and (string == '$'):
 *             assert self_parameter is None             # <<<<<<<<<<<<<<
 *             self_parameter = current_parameter
 *             continue
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_2 = (__pyx_v_self_parameter == Py_None);
        if (unlikely(!(__pyx_t_2 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 2037, __pyx_L1_error)
        }
      }
      #endif

      /* "Aspidites/_vendor/contracts/inspection.py":2038
 *         if (type == ERRORTOKEN) and (string == '$'):
 *             assert self_parameter is None
 *             self_parameter = current_parameter             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
      __Pyx_INCREF(__pyx_v_current_parameter);
      __Pyx_DECREF_SET(__pyx_v_self_parameter, __pyx_v_current_parameter);

      /* "Aspidites/_vendor/contracts/inspection.py":2039
 *             assert self_parameter is None
 *             self_parameter = current_parameter
 *             continue             # <<<<<<<<<<<<<<
 * 
 *         if delayed_comma:
 */
      goto __pyx_L10_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":2036
 *                 continue
 * 
 *         if (type == ERRORTOKEN) and (string == '$'):             # <<<<<<<<<<<<<<
 *             assert self_parameter is None
 *             self_parameter = current_parameter
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2041
 *             continue
 * 
 *         if delayed_comma:             # <<<<<<<<<<<<<<
 *             delayed_comma = False
 *             if not ((type == OP) and (string == ')')):
 */
    __pyx_t_2 = (__pyx_v_delayed_comma != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2042
 * 
 *         if delayed_comma:
 *             delayed_comma = False             # <<<<<<<<<<<<<<
 *             if not ((type == OP) and (string == ')')):
 *                 add(', ')
 */
      __pyx_v_delayed_comma = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2043
 *         if delayed_comma:
 *             delayed_comma = False
 *             if not ((type == OP) and (string == ')')):             # <<<<<<<<<<<<<<
 *                 add(', ')
 *         add(string)
 */
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_type, __pyx_v_OP, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2043, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2043, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L21_bool_binop_done;
      }
      __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_string, __pyx_kp_u__30, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2043, __pyx_L1_error)
      __pyx_t_2 = __pyx_t_1;
      __pyx_L21_bool_binop_done:;
      __pyx_t_1 = ((!__pyx_t_2) != 0);
      if (__pyx_t_1) {

        /* "Aspidites/_vendor/contracts/inspection.py":2044
 *             delayed_comma = False
 *             if not ((type == OP) and (string == ')')):
 *                 add(', ')             # <<<<<<<<<<<<<<
 *         add(string)
 *         if (string == ','):
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_text, __pyx_kp_u__41); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 2044, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2043
 *         if delayed_comma:
 *             delayed_comma = False
 *             if not ((type == OP) and (string == ')')):             # <<<<<<<<<<<<<<
 *                 add(', ')
 *         add(string)
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2041
 *             continue
 * 
 *         if delayed_comma:             # <<<<<<<<<<<<<<
 *             delayed_comma = False
 *             if not ((type == OP) and (string == ')')):
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2045
 *             if not ((type == OP) and (string == ')')):
 *                 add(', ')
 *         add(string)             # <<<<<<<<<<<<<<
 *         if (string == ','):
 *             add(' ')
 */
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_text, __pyx_v_string); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 2045, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2046
 *                 add(', ')
 *         add(string)
 *         if (string == ','):             # <<<<<<<<<<<<<<
 *             add(' ')
 *     clean_signature = ''.join(text)
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_string, __pyx_kp_u__56, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2046, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/contracts/inspection.py":2047
 *         add(string)
 *         if (string == ','):
 *             add(' ')             # <<<<<<<<<<<<<<
 *     clean_signature = ''.join(text)
 *     return clean_signature, self_parameter, last_positional_only
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_text, __pyx_kp_u__44); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 2047, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2046
 *                 add(', ')
 *         add(string)
 *         if (string == ','):             # <<<<<<<<<<<<<<
 *             add(' ')
 *     clean_signature = ''.join(text)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2016
 *     assert t.type == tokenize.ENCODING
 * 
 *     for t in token_stream:             # <<<<<<<<<<<<<<
 *         type, string = t.type, t.string
 * 
 */
    __pyx_L10_continue:;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2048
 *         if (string == ','):
 *             add(' ')
 *     clean_signature = ''.join(text)             # <<<<<<<<<<<<<<
 *     return clean_signature, self_parameter, last_positional_only
 * 
 */
  __pyx_t_4 = PyUnicode_Join(__pyx_kp_u__19, __pyx_v_text); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_clean_signature = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2049
 *             add(' ')
 *     clean_signature = ''.join(text)
 *     return clean_signature, self_parameter, last_positional_only             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_clean_signature);
  __Pyx_GIVEREF(__pyx_v_clean_signature);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_clean_signature);
  __Pyx_INCREF(__pyx_v_self_parameter);
  __Pyx_GIVEREF(__pyx_v_self_parameter);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self_parameter);
  __Pyx_INCREF(__pyx_v_last_positional_only);
  __Pyx_GIVEREF(__pyx_v_last_positional_only);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_last_positional_only);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":1980
 * 
 * 
 * def _signature_strip_non_python_syntax(signature):             # <<<<<<<<<<<<<<
 *     """
 *     Private helper function. Takes a signature in Argument Clinic's
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_strip_non_python_syntax", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_parameter);
  __Pyx_XDECREF(__pyx_v_last_positional_only);
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_generator);
  __Pyx_XDECREF(__pyx_v_token_stream);
  __Pyx_XDECREF(__pyx_v_text);
  __Pyx_XDECREF(__pyx_v_add);
  __Pyx_XDECREF(__pyx_v_current_parameter);
  __Pyx_XDECREF(__pyx_v_OP);
  __Pyx_XDECREF(__pyx_v_ERRORTOKEN);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_type);
  __Pyx_XDECREF(__pyx_v_string);
  __Pyx_XDECREF(__pyx_v_clean_signature);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_l);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_161_signature_fromstr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_160_signature_fromstr[] = "Private helper to parse content of '__text_signature__'\n    and return a Signature based on it.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_161_signature_fromstr = {"_signature_fromstr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_161_signature_fromstr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_160_signature_fromstr};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_161_signature_fromstr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_skip_bound_arg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_fromstr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_obj,&__pyx_n_s_s,&__pyx_n_s_skip_bound_arg,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_fromstr", 0, 3, 4, 1); __PYX_ERR(0, 2052, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_fromstr", 0, 3, 4, 2); __PYX_ERR(0, 2052, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_skip_bound_arg);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_fromstr") < 0)) __PYX_ERR(0, 2052, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_obj = values[1];
    __pyx_v_s = values[2];
    __pyx_v_skip_bound_arg = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_fromstr", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2052, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_160_signature_fromstr(__pyx_self, __pyx_v_cls, __pyx_v_obj, __pyx_v_s, __pyx_v_skip_bound_arg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2090
 *     sys_module_dict = sys.modules.copy()
 * 
 *     def parse_name(node):             # <<<<<<<<<<<<<<
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_1parse_name(PyObject *__pyx_self, PyObject *__pyx_v_node); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_1parse_name = {"parse_name", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_1parse_name, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_1parse_name(PyObject *__pyx_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse_name (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_parse_name(__pyx_self, ((PyObject *)__pyx_v_node));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_parse_name(PyObject *__pyx_self, PyObject *__pyx_v_node) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_name", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":2091
 * 
 *     def parse_name(node):
 *         assert isinstance(node, ast.arg)             # <<<<<<<<<<<<<<
 *         if node.annotation is not None:
 *             raise ValueError("Annotations are not currently supported")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2091, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_arg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_node, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2091, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!(__pyx_t_2 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 2091, __pyx_L1_error)
    }
  }
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":2092
 *     def parse_name(node):
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Annotations are not currently supported")
 *         return node.arg
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_annotation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2093
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 *             raise ValueError("Annotations are not currently supported")             # <<<<<<<<<<<<<<
 *         return node.arg
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2093, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2092
 *     def parse_name(node):
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Annotations are not currently supported")
 *         return node.arg
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2094
 *         if node.annotation is not None:
 *             raise ValueError("Annotations are not currently supported")
 *         return node.arg             # <<<<<<<<<<<<<<
 * 
 *     def wrap_value(s):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_arg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2090
 *     sys_module_dict = sys.modules.copy()
 * 
 *     def parse_name(node):             # <<<<<<<<<<<<<<
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.parse_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2096
 *         return node.arg
 * 
 *     def wrap_value(s):             # <<<<<<<<<<<<<<
 *         try:
 *             value = eval(s, module_dict)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_3wrap_value(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_3wrap_value = {"wrap_value", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_3wrap_value, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_3wrap_value(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap_value (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_2wrap_value(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_2wrap_value(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_outer_scope;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap_value", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":2097
 * 
 *     def wrap_value(s):
 *         try:             # <<<<<<<<<<<<<<
 *             value = eval(s, module_dict)
 *         except NameError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":2098
 *     def wrap_value(s):
 *         try:
 *             value = eval(s, module_dict)             # <<<<<<<<<<<<<<
 *         except NameError:
 *             try:
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_module_dict)) { __Pyx_RaiseClosureNameError("module_dict"); __PYX_ERR(0, 2098, __pyx_L3_error) }
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2098, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_s);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_module_dict);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_module_dict);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_cur_scope->__pyx_v_module_dict);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_eval, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2098, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_value = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2097
 * 
 *     def wrap_value(s):
 *         try:             # <<<<<<<<<<<<<<
 *             value = eval(s, module_dict)
 *         except NameError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2099
 *         try:
 *             value = eval(s, module_dict)
 *         except NameError:             # <<<<<<<<<<<<<<
 *             try:
 *                 value = eval(s, sys_module_dict)
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_NameError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.wrap_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 2099, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/contracts/inspection.py":2100
 *             value = eval(s, module_dict)
 *         except NameError:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = eval(s, sys_module_dict)
 *             except NameError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":2101
 *         except NameError:
 *             try:
 *                 value = eval(s, sys_module_dict)             # <<<<<<<<<<<<<<
 *             except NameError:
 *                 raise RuntimeError()
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_sys_module_dict)) { __Pyx_RaiseClosureNameError("sys_module_dict"); __PYX_ERR(0, 2101, __pyx_L11_error) }
          __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2101, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_v_s);
          __Pyx_GIVEREF(__pyx_v_s);
          PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_s);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sys_module_dict);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sys_module_dict);
          PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_cur_scope->__pyx_v_sys_module_dict);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_builtin_eval, __pyx_t_11, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2101, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_12);
          __pyx_t_12 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2100
 *             value = eval(s, module_dict)
 *         except NameError:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = eval(s, sys_module_dict)
 *             except NameError:
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L18_try_end;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2102
 *             try:
 *                 value = eval(s, sys_module_dict)
 *             except NameError:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError()
 * 
 */
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_NameError);
        if (__pyx_t_6) {
          __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.wrap_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_11, &__pyx_t_13) < 0) __PYX_ERR(0, 2102, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_GOTREF(__pyx_t_13);

          /* "Aspidites/_vendor/contracts/inspection.py":2103
 *                 value = eval(s, sys_module_dict)
 *             except NameError:
 *                 raise RuntimeError()             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(value, (str, int, float, bytes, bool, type(None))):
 */
          __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_builtin_RuntimeError); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2103, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_Raise(__pyx_t_14, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __PYX_ERR(0, 2103, __pyx_L13_except_error)
        }
        goto __pyx_L13_except_error;
        __pyx_L13_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":2100
 *             value = eval(s, module_dict)
 *         except NameError:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = eval(s, sys_module_dict)
 *             except NameError:
 */
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L5_except_error;
        __pyx_L18_try_end:;
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":2097
 * 
 *     def wrap_value(s):
 *         try:             # <<<<<<<<<<<<<<
 *             value = eval(s, module_dict)
 *         except NameError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2105
 *                 raise RuntimeError()
 * 
 *         if isinstance(value, (str, int, float, bytes, bool, type(None))):             # <<<<<<<<<<<<<<
 *             return ast.Constant(value)
 *         raise RuntimeError()
 */
  __Pyx_INCREF(((PyObject*)&PyBool_Type));
  __pyx_t_7 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(((PyObject *)Py_TYPE(Py_None)));
  __pyx_t_4 = ((PyObject *)Py_TYPE(Py_None));
  __pyx_t_16 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (!__pyx_t_17) {
  } else {
    __pyx_t_15 = __pyx_t_17;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_17 = PyInt_Check(__pyx_v_value); 
  __pyx_t_16 = (__pyx_t_17 != 0);
  if (!__pyx_t_16) {
  } else {
    __pyx_t_15 = __pyx_t_16;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_16 = PyFloat_Check(__pyx_v_value); 
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (!__pyx_t_17) {
  } else {
    __pyx_t_15 = __pyx_t_17;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_17 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_16 = (__pyx_t_17 != 0);
  if (!__pyx_t_16) {
  } else {
    __pyx_t_15 = __pyx_t_16;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_16 = PyObject_IsInstance(__pyx_v_value, __pyx_t_7); 
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (!__pyx_t_17) {
  } else {
    __pyx_t_15 = __pyx_t_17;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_17 = PyObject_IsInstance(__pyx_v_value, __pyx_t_4); 
  __pyx_t_16 = (__pyx_t_17 != 0);
  __pyx_t_15 = __pyx_t_16;
  __pyx_L22_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_16 = (__pyx_t_15 != 0);
  if (__pyx_t_16) {

    /* "Aspidites/_vendor/contracts/inspection.py":2106
 * 
 *         if isinstance(value, (str, int, float, bytes, bool, type(None))):
 *             return ast.Constant(value)             # <<<<<<<<<<<<<<
 *         raise RuntimeError()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2106, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_Constant); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2105
 *                 raise RuntimeError()
 * 
 *         if isinstance(value, (str, int, float, bytes, bool, type(None))):             # <<<<<<<<<<<<<<
 *             return ast.Constant(value)
 *         raise RuntimeError()
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2107
 *         if isinstance(value, (str, int, float, bytes, bool, type(None))):
 *             return ast.Constant(value)
 *         raise RuntimeError()             # <<<<<<<<<<<<<<
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):
 */
  __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_builtin_RuntimeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_Raise(__pyx_t_7, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __PYX_ERR(0, 2107, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2096
 *         return node.arg
 * 
 *     def wrap_value(s):             # <<<<<<<<<<<<<<
 *         try:
 *             value = eval(s, module_dict)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.wrap_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2110
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):
 *         def visit_Attribute(self, node):             # <<<<<<<<<<<<<<
 *             a = []
 *             n = node
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_1visit_Attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_1visit_Attribute = {"visit_Attribute", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_1visit_Attribute, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_1visit_Attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_node = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("visit_Attribute (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_node,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("visit_Attribute", 1, 2, 2, 1); __PYX_ERR(0, 2110, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "visit_Attribute") < 0)) __PYX_ERR(0, 2110, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_node = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("visit_Attribute", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2110, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.RewriteSymbolics.visit_Attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_visit_Attribute(__pyx_self, __pyx_v_self, __pyx_v_node);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_visit_Attribute(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_outer_scope;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_Attribute", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":2111
 *     class RewriteSymbolics(ast.NodeTransformer):
 *         def visit_Attribute(self, node):
 *             a = []             # <<<<<<<<<<<<<<
 *             n = node
 *             while isinstance(n, ast.Attribute):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_a = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2112
 *         def visit_Attribute(self, node):
 *             a = []
 *             n = node             # <<<<<<<<<<<<<<
 *             while isinstance(n, ast.Attribute):
 *                 a.append(n.attr)
 */
  __Pyx_INCREF(__pyx_v_node);
  __pyx_v_n = __pyx_v_node;

  /* "Aspidites/_vendor/contracts/inspection.py":2113
 *             a = []
 *             n = node
 *             while isinstance(n, ast.Attribute):             # <<<<<<<<<<<<<<
 *                 a.append(n.attr)
 *                 n = n.value
 */
  while (1) {
    if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2113, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_Attribute); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_n, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (!__pyx_t_3) break;

    /* "Aspidites/_vendor/contracts/inspection.py":2114
 *             n = node
 *             while isinstance(n, ast.Attribute):
 *                 a.append(n.attr)             # <<<<<<<<<<<<<<
 *                 n = n.value
 *             if not isinstance(n, ast.Name):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_n, __pyx_n_s_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_a, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2114, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2115
 *             while isinstance(n, ast.Attribute):
 *                 a.append(n.attr)
 *                 n = n.value             # <<<<<<<<<<<<<<
 *             if not isinstance(n, ast.Name):
 *                 raise RuntimeError()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_n, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_n, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2116
 *                 a.append(n.attr)
 *                 n = n.value
 *             if not isinstance(n, ast.Name):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError()
 *             a.append(n.id)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2116, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_Name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_n, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2117
 *                 n = n.value
 *             if not isinstance(n, ast.Name):
 *                 raise RuntimeError()             # <<<<<<<<<<<<<<
 *             a.append(n.id)
 *             value = ".".join(reversed(a))
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_RuntimeError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2117, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2116
 *                 a.append(n.attr)
 *                 n = n.value
 *             if not isinstance(n, ast.Name):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError()
 *             a.append(n.id)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2118
 *             if not isinstance(n, ast.Name):
 *                 raise RuntimeError()
 *             a.append(n.id)             # <<<<<<<<<<<<<<
 *             value = ".".join(reversed(a))
 *             return wrap_value(value)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_n, __pyx_n_s_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_a, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2119
 *                 raise RuntimeError()
 *             a.append(n.id)
 *             value = ".".join(reversed(a))             # <<<<<<<<<<<<<<
 *             return wrap_value(value)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_reversed, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__10, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_value = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2120
 *             a.append(n.id)
 *             value = ".".join(reversed(a))
 *             return wrap_value(value)             # <<<<<<<<<<<<<<
 * 
 *         def visit_Name(self, node):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_wrap_value)) { __Pyx_RaiseClosureNameError("wrap_value"); __PYX_ERR(0, 2120, __pyx_L1_error) }
  __pyx_t_5 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_2wrap_value(__pyx_cur_scope->__pyx_v_wrap_value, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2110
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):
 *         def visit_Attribute(self, node):             # <<<<<<<<<<<<<<
 *             a = []
 *             n = node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.RewriteSymbolics.visit_Attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2122
 *             return wrap_value(value)
 * 
 *         def visit_Name(self, node):             # <<<<<<<<<<<<<<
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_3visit_Name(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_3visit_Name = {"visit_Name", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_3visit_Name, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_3visit_Name(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_node = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("visit_Name (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_node,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("visit_Name", 1, 2, 2, 1); __PYX_ERR(0, 2122, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "visit_Name") < 0)) __PYX_ERR(0, 2122, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_node = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("visit_Name", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.RewriteSymbolics.visit_Name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_2visit_Name(__pyx_self, __pyx_v_self, __pyx_v_node);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_2visit_Name(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_Name", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/contracts/inspection.py":2123
 * 
 *         def visit_Name(self, node):
 *             if not isinstance(node.ctx, ast.Load):             # <<<<<<<<<<<<<<
 *                 raise ValueError()
 *             return wrap_value(node.id)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2123, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_Load); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2124
 *         def visit_Name(self, node):
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()             # <<<<<<<<<<<<<<
 *             return wrap_value(node.id)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2124, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2123
 * 
 *         def visit_Name(self, node):
 *             if not isinstance(node.ctx, ast.Load):             # <<<<<<<<<<<<<<
 *                 raise ValueError()
 *             return wrap_value(node.id)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2125
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()
 *             return wrap_value(node.id)             # <<<<<<<<<<<<<<
 * 
 *     def p(name_node, default_node, default=empty):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_wrap_value)) { __Pyx_RaiseClosureNameError("wrap_value"); __PYX_ERR(0, 2125, __pyx_L1_error) }
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_2wrap_value(__pyx_cur_scope->__pyx_v_wrap_value, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2122
 *             return wrap_value(value)
 * 
 *         def visit_Name(self, node):             # <<<<<<<<<<<<<<
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.RewriteSymbolics.visit_Name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2127
 *             return wrap_value(node.id)
 * 
 *     def p(name_node, default_node, default=empty):             # <<<<<<<<<<<<<<
 *         name = parse_name(name_node)
 *         if name is invalid:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_207__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_default);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_default);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_default);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_5p(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_5p = {"p", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_5p, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_5p(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name_node = 0;
  PyObject *__pyx_v_default_node = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("p (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_node,&__pyx_n_s_default_node,&__pyx_n_s_default,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults8 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_default;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default_node)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("p", 0, 2, 3, 1); __PYX_ERR(0, 2127, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "p") < 0)) __PYX_ERR(0, 2127, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name_node = values[0];
    __pyx_v_default_node = values[1];
    __pyx_v_default = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("p", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2127, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.p", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_4p(__pyx_self, __pyx_v_name_node, __pyx_v_default_node, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_4p(PyObject *__pyx_self, PyObject *__pyx_v_name_node, PyObject *__pyx_v_default_node, PyObject *__pyx_v_default) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_outer_scope;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("p", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_INCREF(__pyx_v_default_node);
  __Pyx_INCREF(__pyx_v_default);

  /* "Aspidites/_vendor/contracts/inspection.py":2128
 * 
 *     def p(name_node, default_node, default=empty):
 *         name = parse_name(name_node)             # <<<<<<<<<<<<<<
 *         if name is invalid:
 *             return None
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_parse_name)) { __Pyx_RaiseClosureNameError("parse_name"); __PYX_ERR(0, 2128, __pyx_L1_error) }
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_parse_name(__pyx_cur_scope->__pyx_v_parse_name, __pyx_v_name_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2129
 *     def p(name_node, default_node, default=empty):
 *         name = parse_name(name_node)
 *         if name is invalid:             # <<<<<<<<<<<<<<
 *             return None
 *         if default_node and default_node is not _empty:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_invalid)) { __Pyx_RaiseClosureNameError("invalid"); __PYX_ERR(0, 2129, __pyx_L1_error) }
  __pyx_t_2 = (__pyx_v_name == __pyx_cur_scope->__pyx_v_invalid);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2130
 *         name = parse_name(name_node)
 *         if name is invalid:
 *             return None             # <<<<<<<<<<<<<<
 *         if default_node and default_node is not _empty:
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2129
 *     def p(name_node, default_node, default=empty):
 *         name = parse_name(name_node)
 *         if name is invalid:             # <<<<<<<<<<<<<<
 *             return None
 *         if default_node and default_node is not _empty:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2131
 *         if name is invalid:
 *             return None
 *         if default_node and default_node is not _empty:             # <<<<<<<<<<<<<<
 *             try:
 *                 default_node = RewriteSymbolics().visit(default_node)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_default_node); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2131, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_default_node != __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_2 != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2132
 *             return None
 *         if default_node and default_node is not _empty:
 *             try:             # <<<<<<<<<<<<<<
 *                 default_node = RewriteSymbolics().visit(default_node)
 *                 o = ast.literal_eval(default_node)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":2133
 *         if default_node and default_node is not _empty:
 *             try:
 *                 default_node = RewriteSymbolics().visit(default_node)             # <<<<<<<<<<<<<<
 *                 o = ast.literal_eval(default_node)
 *             except ValueError:
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_RewriteSymbolics)) { __Pyx_RaiseClosureNameError("RewriteSymbolics"); __PYX_ERR(0, 2133, __pyx_L7_error) }
        __pyx_t_8 = __Pyx_PyObject_CallNoArg(__pyx_cur_scope->__pyx_v_RewriteSymbolics); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2133, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_visit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2133, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_v_default_node) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_default_node);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2133, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF_SET(__pyx_v_default_node, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2134
 *             try:
 *                 default_node = RewriteSymbolics().visit(default_node)
 *                 o = ast.literal_eval(default_node)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 o = invalid
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_ast)) { __Pyx_RaiseClosureNameError("ast"); __PYX_ERR(0, 2134, __pyx_L7_error) }
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_literal_eval); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2134, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_v_default_node) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_default_node);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2134, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_v_o = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2132
 *             return None
 *         if default_node and default_node is not _empty:
 *             try:             # <<<<<<<<<<<<<<
 *                 default_node = RewriteSymbolics().visit(default_node)
 *                 o = ast.literal_eval(default_node)
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2135
 *                 default_node = RewriteSymbolics().visit(default_node)
 *                 o = ast.literal_eval(default_node)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 o = invalid
 *             if o is invalid:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.p", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_9, &__pyx_t_8) < 0) __PYX_ERR(0, 2135, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_t_8);

        /* "Aspidites/_vendor/contracts/inspection.py":2136
 *                 o = ast.literal_eval(default_node)
 *             except ValueError:
 *                 o = invalid             # <<<<<<<<<<<<<<
 *             if o is invalid:
 *                 return None
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_invalid)) { __Pyx_RaiseClosureNameError("invalid"); __PYX_ERR(0, 2136, __pyx_L9_except_error) }
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_invalid);
        __Pyx_XDECREF_SET(__pyx_v_o, __pyx_cur_scope->__pyx_v_invalid);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;
      __pyx_L9_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":2132
 *             return None
 *         if default_node and default_node is not _empty:
 *             try:             # <<<<<<<<<<<<<<
 *                 default_node = RewriteSymbolics().visit(default_node)
 *                 o = ast.literal_eval(default_node)
 */
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L12_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2137
 *             except ValueError:
 *                 o = invalid
 *             if o is invalid:             # <<<<<<<<<<<<<<
 *                 return None
 *             default = o if o is not invalid else default
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_invalid)) { __Pyx_RaiseClosureNameError("invalid"); __PYX_ERR(0, 2137, __pyx_L1_error) }
    __pyx_t_3 = (__pyx_v_o == __pyx_cur_scope->__pyx_v_invalid);
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":2138
 *                 o = invalid
 *             if o is invalid:
 *                 return None             # <<<<<<<<<<<<<<
 *             default = o if o is not invalid else default
 *         parameters.append(Parameter(name, kind, default=default, annotation=empty))
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":2137
 *             except ValueError:
 *                 o = invalid
 *             if o is invalid:             # <<<<<<<<<<<<<<
 *                 return None
 *             default = o if o is not invalid else default
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2139
 *             if o is invalid:
 *                 return None
 *             default = o if o is not invalid else default             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, kind, default=default, annotation=empty))
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_invalid)) { __Pyx_RaiseClosureNameError("invalid"); __PYX_ERR(0, 2139, __pyx_L1_error) }
    __pyx_t_4 = (__pyx_v_o != __pyx_cur_scope->__pyx_v_invalid);
    if ((__pyx_t_4 != 0)) {
      __Pyx_INCREF(__pyx_v_o);
      __pyx_t_8 = __pyx_v_o;
    } else {
      __Pyx_INCREF(__pyx_v_default);
      __pyx_t_8 = __pyx_v_default;
    }
    __Pyx_DECREF_SET(__pyx_v_default, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2131
 *         if name is invalid:
 *             return None
 *         if default_node and default_node is not _empty:             # <<<<<<<<<<<<<<
 *             try:
 *                 default_node = RewriteSymbolics().visit(default_node)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2140
 *                 return None
 *             default = o if o is not invalid else default
 *         parameters.append(Parameter(name, kind, default=default, annotation=empty))             # <<<<<<<<<<<<<<
 * 
 *     # non-keyword-only parameters
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_parameters)) { __Pyx_RaiseClosureNameError("parameters"); __PYX_ERR(0, 2140, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_parameters == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 2140, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_Parameter)) { __Pyx_RaiseClosureNameError("Parameter"); __PYX_ERR(0, 2140, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_v_kind)) { __Pyx_RaiseClosureNameError("kind"); __PYX_ERR(0, 2140, __pyx_L1_error) }
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_name);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_kind);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_kind);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_cur_scope->__pyx_v_kind);
  __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, __pyx_v_default) < 0) __PYX_ERR(0, 2140, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_v_empty)) { __Pyx_RaiseClosureNameError("empty"); __PYX_ERR(0, 2140, __pyx_L1_error) }
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_annotation, __pyx_cur_scope->__pyx_v_empty) < 0) __PYX_ERR(0, 2140, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v_Parameter, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_parameters, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2127
 *             return wrap_value(node.id)
 * 
 *     def p(name_node, default_node, default=empty):             # <<<<<<<<<<<<<<
 *         name = parse_name(name_node)
 *         if name is invalid:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr.p", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XDECREF(__pyx_v_default_node);
  __Pyx_XDECREF(__pyx_v_default);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_160_signature_fromstr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_obj, PyObject *__pyx_v_s, PyObject *__pyx_v_skip_bound_arg) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_cur_scope;
  PyObject *__pyx_v_clean_signature = NULL;
  PyObject *__pyx_v_self_parameter = NULL;
  PyObject *__pyx_v_last_positional_only = NULL;
  PyObject *__pyx_v_program = NULL;
  PyObject *__pyx_v_module = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_module_name = NULL;
  PyObject *__pyx_v_p = 0;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_defaults = NULL;
  PyObject *__pyx_v_iter = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_default = NULL;
  PyObject *__pyx_v__self = NULL;
  PyObject *__pyx_v_self_isbound = NULL;
  PyObject *__pyx_v_self_ismodule = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_fromstr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2052, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2058
 *     # Lazy import ast because it's relatively heavy and
 *     # it's not used for other than this function.
 *     import ast             # <<<<<<<<<<<<<<
 * 
 *     Parameter = cls._parameter_cls
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ast, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_ast = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2060
 *     import ast
 * 
 *     Parameter = cls._parameter_cls             # <<<<<<<<<<<<<<
 * 
 *     clean_signature, self_parameter, last_positional_only = \
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_parameter_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_Parameter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2063
 * 
 *     clean_signature, self_parameter, last_positional_only = \
 *         _signature_strip_non_python_syntax(s)             # <<<<<<<<<<<<<<
 * 
 *     program = "def foo" + clean_signature + ": pass"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_signature_strip_non_python_synt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_s) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_s);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 2062, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 2062, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 2062, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2062
 *     Parameter = cls._parameter_cls
 * 
 *     clean_signature, self_parameter, last_positional_only = \             # <<<<<<<<<<<<<<
 *         _signature_strip_non_python_syntax(s)
 * 
 */
  __pyx_v_clean_signature = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_self_parameter = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_last_positional_only = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2065
 *         _signature_strip_non_python_syntax(s)
 * 
 *     program = "def foo" + clean_signature + ": pass"             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u_def_foo, __pyx_v_clean_signature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_kp_u_pass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_program = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2067
 *     program = "def foo" + clean_signature + ": pass"
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         module = ast.parse(program)
 *     except SyntaxError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":2068
 * 
 *     try:
 *         module = ast.parse(program)             # <<<<<<<<<<<<<<
 *     except SyntaxError:
 *         module = None
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_parse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2068, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_program) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_program);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2068, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_module = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2067
 *     program = "def foo" + clean_signature + ": pass"
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         module = ast.parse(program)
 *     except SyntaxError:
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2069
 *     try:
 *         module = ast.parse(program)
 *     except SyntaxError:             # <<<<<<<<<<<<<<
 *         module = None
 * 
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_SyntaxError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 2069, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_3);

      /* "Aspidites/_vendor/contracts/inspection.py":2070
 *         module = ast.parse(program)
 *     except SyntaxError:
 *         module = None             # <<<<<<<<<<<<<<
 * 
 *     if not isinstance(module, ast.Module):
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_module, Py_None);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_exception_handled;
    }
    goto __pyx_L7_except_error;
    __pyx_L7_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":2067
 *     program = "def foo" + clean_signature + ": pass"
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         module = ast.parse(program)
 *     except SyntaxError:
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L6_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L10_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2072
 *         module = None
 * 
 *     if not isinstance(module, ast.Module):             # <<<<<<<<<<<<<<
 *         raise ValueError("{!r} builtin has invalid signature".format(obj))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_Module); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = PyObject_IsInstance(__pyx_v_module, __pyx_t_3); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = ((!(__pyx_t_11 != 0)) != 0);
  if (unlikely(__pyx_t_12)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2073
 * 
 *     if not isinstance(module, ast.Module):
 *         raise ValueError("{!r} builtin has invalid signature".format(obj))             # <<<<<<<<<<<<<<
 * 
 *     f = module.body[0]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_builtin_has_invalid_signature, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2073, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2072
 *         module = None
 * 
 *     if not isinstance(module, ast.Module):             # <<<<<<<<<<<<<<
 *         raise ValueError("{!r} builtin has invalid signature".format(obj))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2075
 *         raise ValueError("{!r} builtin has invalid signature".format(obj))
 * 
 *     f = module.body[0]             # <<<<<<<<<<<<<<
 * 
 *     parameters = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_body); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_f = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2077
 *     f = module.body[0]
 * 
 *     parameters = []             # <<<<<<<<<<<<<<
 *     empty = Parameter.empty
 *     invalid = object()
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_parameters = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2078
 * 
 *     parameters = []
 *     empty = Parameter.empty             # <<<<<<<<<<<<<<
 *     invalid = object()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_empty_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_empty = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2079
 *     parameters = []
 *     empty = Parameter.empty
 *     invalid = object()             # <<<<<<<<<<<<<<
 * 
 *     module = None
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_invalid = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2081
 *     invalid = object()
 * 
 *     module = None             # <<<<<<<<<<<<<<
 *     module_dict = {}
 *     module_name = getattr(obj, '__module__', None)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_DECREF_SET(__pyx_v_module, Py_None);

  /* "Aspidites/_vendor/contracts/inspection.py":2082
 * 
 *     module = None
 *     module_dict = {}             # <<<<<<<<<<<<<<
 *     module_name = getattr(obj, '__module__', None)
 *     if module_name:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_module_dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2083
 *     module = None
 *     module_dict = {}
 *     module_name = getattr(obj, '__module__', None)             # <<<<<<<<<<<<<<
 *     if module_name:
 *         module = sys.modules.get(module_name, None)
 */
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_module, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_module_name = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2084
 *     module_dict = {}
 *     module_name = getattr(obj, '__module__', None)
 *     if module_name:             # <<<<<<<<<<<<<<
 *         module = sys.modules.get(module_name, None)
 *         if module:
 */
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_module_name); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2084, __pyx_L1_error)
  if (__pyx_t_12) {

    /* "Aspidites/_vendor/contracts/inspection.py":2085
 *     module_name = getattr(obj, '__module__', None)
 *     if module_name:
 *         module = sys.modules.get(module_name, None)             # <<<<<<<<<<<<<<
 *         if module:
 *             module_dict = module.__dict__
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_modules); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_module_name, Py_None};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2085, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_module_name, Py_None};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2085, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_module_name);
      __Pyx_GIVEREF(__pyx_v_module_name);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_10, __pyx_v_module_name);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_10, Py_None);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_module, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2086
 *     if module_name:
 *         module = sys.modules.get(module_name, None)
 *         if module:             # <<<<<<<<<<<<<<
 *             module_dict = module.__dict__
 *     sys_module_dict = sys.modules.copy()
 */
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_module); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2086, __pyx_L1_error)
    if (__pyx_t_12) {

      /* "Aspidites/_vendor/contracts/inspection.py":2087
 *         module = sys.modules.get(module_name, None)
 *         if module:
 *             module_dict = module.__dict__             # <<<<<<<<<<<<<<
 *     sys_module_dict = sys.modules.copy()
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_module, __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2087, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_module_dict);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_module_dict, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2086
 *     if module_name:
 *         module = sys.modules.get(module_name, None)
 *         if module:             # <<<<<<<<<<<<<<
 *             module_dict = module.__dict__
 *     sys_module_dict = sys.modules.copy()
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2084
 *     module_dict = {}
 *     module_name = getattr(obj, '__module__', None)
 *     if module_name:             # <<<<<<<<<<<<<<
 *         module = sys.modules.get(module_name, None)
 *         if module:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2088
 *         if module:
 *             module_dict = module.__dict__
 *     sys_module_dict = sys.modules.copy()             # <<<<<<<<<<<<<<
 * 
 *     def parse_name(node):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_modules); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_sys_module_dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2090
 *     sys_module_dict = sys.modules.copy()
 * 
 *     def parse_name(node):             # <<<<<<<<<<<<<<
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_1parse_name, 0, __pyx_n_s_signature_fromstr_locals_parse, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_parse_name = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2096
 *         return node.arg
 * 
 *     def wrap_value(s):             # <<<<<<<<<<<<<<
 *         try:
 *             value = eval(s, module_dict)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_3wrap_value, 0, __pyx_n_s_signature_fromstr_locals_wrap_v, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_wrap_value = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2109
 *         raise RuntimeError()
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):             # <<<<<<<<<<<<<<
 *         def visit_Attribute(self, node):
 *             a = []
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ast, __pyx_n_s_NodeTransformer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_1, __pyx_n_s_RewriteSymbolics, __pyx_n_s_signature_fromstr_locals_Rewrit, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/contracts/inspection.py":2110
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):
 *         def visit_Attribute(self, node):             # <<<<<<<<<<<<<<
 *             a = []
 *             n = node
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_1visit_Attribute, 0, __pyx_n_s_signature_fromstr_locals_Rewrit_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_visit_Attribute, __pyx_t_4) < 0) __PYX_ERR(0, 2110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2122
 *             return wrap_value(value)
 * 
 *         def visit_Name(self, node):             # <<<<<<<<<<<<<<
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_16RewriteSymbolics_3visit_Name, 0, __pyx_n_s_signature_fromstr_locals_Rewrit_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_visit_Name, __pyx_t_4) < 0) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2109
 *         raise RuntimeError()
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):             # <<<<<<<<<<<<<<
 *         def visit_Attribute(self, node):
 *             a = []
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_RewriteSymbolics, __pyx_t_1, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_RewriteSymbolics = __pyx_t_4;
  __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2127
 *             return wrap_value(node.id)
 * 
 *     def p(name_node, default_node, default=empty):             # <<<<<<<<<<<<<<
 *         name = parse_name(name_node)
 *         if name is invalid:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_18_signature_fromstr_5p, 0, __pyx_n_s_signature_fromstr_locals_p, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults8), 1)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_empty);
  __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_t_1)->__pyx_arg_default = __pyx_cur_scope->__pyx_v_empty;
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_empty);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_207__defaults__);
  __pyx_v_p = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2144
 *     # non-keyword-only parameters
 *     # noinspection PyUnresolvedReferences
 *     args = reversed(f.args.args)             # <<<<<<<<<<<<<<
 *     # noinspection PyUnresolvedReferences
 *     defaults = reversed(f.args.defaults)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_reversed, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_args = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2146
 *     args = reversed(f.args.args)
 *     # noinspection PyUnresolvedReferences
 *     defaults = reversed(f.args.defaults)             # <<<<<<<<<<<<<<
 *     iter = itertools.zip_longest(args, defaults, fillvalue=None)
 *     if last_positional_only is not None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_defaults); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_reversed, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_defaults = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2147
 *     # noinspection PyUnresolvedReferences
 *     defaults = reversed(f.args.defaults)
 *     iter = itertools.zip_longest(args, defaults, fillvalue=None)             # <<<<<<<<<<<<<<
 *     if last_positional_only is not None:
 *         kind = Parameter.POSITIONAL_ONLY
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zip_longest); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_args);
  __Pyx_INCREF(__pyx_v_defaults);
  __Pyx_GIVEREF(__pyx_v_defaults);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_defaults);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fillvalue, Py_None) < 0) __PYX_ERR(0, 2147, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_iter = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2148
 *     defaults = reversed(f.args.defaults)
 *     iter = itertools.zip_longest(args, defaults, fillvalue=None)
 *     if last_positional_only is not None:             # <<<<<<<<<<<<<<
 *         kind = Parameter.POSITIONAL_ONLY
 *     else:
 */
  __pyx_t_12 = (__pyx_v_last_positional_only != Py_None);
  __pyx_t_11 = (__pyx_t_12 != 0);
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":2149
 *     iter = itertools.zip_longest(args, defaults, fillvalue=None)
 *     if last_positional_only is not None:
 *         kind = Parameter.POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 *     else:
 *         kind = Parameter.POSITIONAL_OR_KEYWORD
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_POSITIONAL_ONLY_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_kind = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2148
 *     defaults = reversed(f.args.defaults)
 *     iter = itertools.zip_longest(args, defaults, fillvalue=None)
 *     if last_positional_only is not None:             # <<<<<<<<<<<<<<
 *         kind = Parameter.POSITIONAL_ONLY
 *     else:
 */
    goto __pyx_L16;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2151
 *         kind = Parameter.POSITIONAL_ONLY
 *     else:
 *         kind = Parameter.POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 *     for i, (name, default) in enumerate(reversed(list(iter))):
 *         p(name, default)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_POSITIONAL_OR_KEYWORD_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_kind = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L16:;

  /* "Aspidites/_vendor/contracts/inspection.py":2152
 *     else:
 *         kind = Parameter.POSITIONAL_OR_KEYWORD
 *     for i, (name, default) in enumerate(reversed(list(iter))):             # <<<<<<<<<<<<<<
 *         p(name, default)
 *         if i == last_positional_only:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_4 = __pyx_int_0;
  __pyx_t_2 = PySequence_List(__pyx_v_iter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = PyList_GET_SIZE(__pyx_t_1) - 1;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_13 < 0) break;
    if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_2); __pyx_t_13--; if (unlikely(0 < 0)) __PYX_ERR(0, 2152, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13--; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 2152, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_14 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_14)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_14); if (unlikely(!__pyx_t_3)) goto __pyx_L19_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_5 = __pyx_t_6(__pyx_t_14); if (unlikely(!__pyx_t_5)) goto __pyx_L19_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_14), 2) < 0) __PYX_ERR(0, 2152, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L20_unpacking_done;
      __pyx_L19_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 2152, __pyx_L1_error)
      __pyx_L20_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_default, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2153
 *         kind = Parameter.POSITIONAL_OR_KEYWORD
 *     for i, (name, default) in enumerate(reversed(list(iter))):
 *         p(name, default)             # <<<<<<<<<<<<<<
 *         if i == last_positional_only:
 *             kind = Parameter.POSITIONAL_OR_KEYWORD
 */
    __Pyx_INCREF(__pyx_v_p);
    __pyx_t_5 = __pyx_v_p; __pyx_t_3 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_default};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_default};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, __pyx_v_name);
      __Pyx_INCREF(__pyx_v_default);
      __Pyx_GIVEREF(__pyx_v_default);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_v_default);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2154
 *     for i, (name, default) in enumerate(reversed(list(iter))):
 *         p(name, default)
 *         if i == last_positional_only:             # <<<<<<<<<<<<<<
 *             kind = Parameter.POSITIONAL_OR_KEYWORD
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_i, __pyx_v_last_positional_only, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "Aspidites/_vendor/contracts/inspection.py":2155
 *         p(name, default)
 *         if i == last_positional_only:
 *             kind = Parameter.POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 * 
 *     # *args
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_POSITIONAL_OR_KEYWORD_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2155, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_kind);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_kind, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2154
 *     for i, (name, default) in enumerate(reversed(list(iter))):
 *         p(name, default)
 *         if i == last_positional_only:             # <<<<<<<<<<<<<<
 *             kind = Parameter.POSITIONAL_OR_KEYWORD
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2152
 *     else:
 *         kind = Parameter.POSITIONAL_OR_KEYWORD
 *     for i, (name, default) in enumerate(reversed(list(iter))):             # <<<<<<<<<<<<<<
 *         p(name, default)
 *         if i == last_positional_only:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2159
 *     # *args
 *     # noinspection PyUnresolvedReferences
 *     if f.args.vararg:             # <<<<<<<<<<<<<<
 *         kind = Parameter.VAR_POSITIONAL
 *         # noinspection PyUnresolvedReferences
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vararg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":2160
 *     # noinspection PyUnresolvedReferences
 *     if f.args.vararg:
 *         kind = Parameter.VAR_POSITIONAL             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         p(f.args.vararg, empty)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_VAR_POSITIONAL_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_kind);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_kind, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2162
 *         kind = Parameter.VAR_POSITIONAL
 *         # noinspection PyUnresolvedReferences
 *         p(f.args.vararg, empty)             # <<<<<<<<<<<<<<
 * 
 *     # keyword-only arguments
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_vararg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_v_p);
    __pyx_t_4 = __pyx_v_p; __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_2, __pyx_cur_scope->__pyx_v_empty};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2162, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_2, __pyx_cur_scope->__pyx_v_empty};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2162, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, __pyx_t_2);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_empty);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_empty);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_cur_scope->__pyx_v_empty);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2159
 *     # *args
 *     # noinspection PyUnresolvedReferences
 *     if f.args.vararg:             # <<<<<<<<<<<<<<
 *         kind = Parameter.VAR_POSITIONAL
 *         # noinspection PyUnresolvedReferences
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2165
 * 
 *     # keyword-only arguments
 *     kind = Parameter.KEYWORD_ONLY             # <<<<<<<<<<<<<<
 *     # noinspection PyUnresolvedReferences
 *     for name, default in zip(f.args.kwonlyargs, f.args.kw_defaults):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_KEYWORD_ONLY_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_kind);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_kind, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2167
 *     kind = Parameter.KEYWORD_ONLY
 *     # noinspection PyUnresolvedReferences
 *     for name, default in zip(f.args.kwonlyargs, f.args.kw_defaults):             # <<<<<<<<<<<<<<
 *         p(name, default)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kwonlyargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kw_defaults); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_14);
  __pyx_t_4 = 0;
  __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
    __pyx_t_1 = __pyx_t_14; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
    __pyx_t_15 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2167, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  for (;;) {
    if (likely(!__pyx_t_15)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_14); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 2167, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2167, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_14); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 2167, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2167, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      }
    } else {
      __pyx_t_14 = __pyx_t_15(__pyx_t_1);
      if (unlikely(!__pyx_t_14)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2167, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_14);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
      PyObject* sequence = __pyx_t_14;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 2167, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 2167, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L26_unpacking_done;
      __pyx_L25_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 2167, __pyx_L1_error)
      __pyx_L26_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_default, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2168
 *     # noinspection PyUnresolvedReferences
 *     for name, default in zip(f.args.kwonlyargs, f.args.kw_defaults):
 *         p(name, default)             # <<<<<<<<<<<<<<
 * 
 *     # **kwargs
 */
    __Pyx_INCREF(__pyx_v_p);
    __pyx_t_2 = __pyx_v_p; __pyx_t_4 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_name, __pyx_v_default};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_14);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_name, __pyx_v_default};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_14);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_10, __pyx_v_name);
      __Pyx_INCREF(__pyx_v_default);
      __Pyx_GIVEREF(__pyx_v_default);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_10, __pyx_v_default);
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2167
 *     kind = Parameter.KEYWORD_ONLY
 *     # noinspection PyUnresolvedReferences
 *     for name, default in zip(f.args.kwonlyargs, f.args.kw_defaults):             # <<<<<<<<<<<<<<
 *         p(name, default)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2172
 *     # **kwargs
 *     # noinspection PyUnresolvedReferences
 *     if f.args.kwarg:             # <<<<<<<<<<<<<<
 *         kind = Parameter.VAR_KEYWORD
 *         # noinspection PyUnresolvedReferences
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kwarg); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":2173
 *     # noinspection PyUnresolvedReferences
 *     if f.args.kwarg:
 *         kind = Parameter.VAR_KEYWORD             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         p(f.args.kwarg, empty)
 */
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_VAR_KEYWORD_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_kind);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_kind, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_14);
    __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2175
 *         kind = Parameter.VAR_KEYWORD
 *         # noinspection PyUnresolvedReferences
 *         p(f.args.kwarg, empty)             # <<<<<<<<<<<<<<
 * 
 *     if self_parameter is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kwarg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v_p);
    __pyx_t_1 = __pyx_v_p; __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_2, __pyx_cur_scope->__pyx_v_empty};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2175, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_2, __pyx_cur_scope->__pyx_v_empty};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2175, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_10, __pyx_t_2);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_empty);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_empty);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_10, __pyx_cur_scope->__pyx_v_empty);
      __pyx_t_2 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2172
 *     # **kwargs
 *     # noinspection PyUnresolvedReferences
 *     if f.args.kwarg:             # <<<<<<<<<<<<<<
 *         kind = Parameter.VAR_KEYWORD
 *         # noinspection PyUnresolvedReferences
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2177
 *         p(f.args.kwarg, empty)
 * 
 *     if self_parameter is not None:             # <<<<<<<<<<<<<<
 *         # Possibly strip the bound argument:
 *         #    - We *always* strip first bound argument if
 */
  __pyx_t_11 = (__pyx_v_self_parameter != Py_None);
  __pyx_t_12 = (__pyx_t_11 != 0);
  if (__pyx_t_12) {

    /* "Aspidites/_vendor/contracts/inspection.py":2183
 *         #    - We don't strip first bound argument if
 *         #      skip_bound_arg is False.
 *         assert parameters             # <<<<<<<<<<<<<<
 *         _self = getattr(obj, '__self__', None)
 *         self_isbound = _self is not None
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_12 = (PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_parameters) != 0);
      if (unlikely(!__pyx_t_12)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 2183, __pyx_L1_error)
      }
    }
    #endif

    /* "Aspidites/_vendor/contracts/inspection.py":2184
 *         #      skip_bound_arg is False.
 *         assert parameters
 *         _self = getattr(obj, '__self__', None)             # <<<<<<<<<<<<<<
 *         self_isbound = _self is not None
 *         self_ismodule = ismodule(_self)
 */
    __pyx_t_14 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_n_u_self, Py_None); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_v__self = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2185
 *         assert parameters
 *         _self = getattr(obj, '__self__', None)
 *         self_isbound = _self is not None             # <<<<<<<<<<<<<<
 *         self_ismodule = ismodule(_self)
 *         if self_isbound and (self_ismodule or skip_bound_arg):
 */
    __pyx_t_12 = (__pyx_v__self != Py_None);
    __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_t_12); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_v_self_isbound = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2186
 *         _self = getattr(obj, '__self__', None)
 *         self_isbound = _self is not None
 *         self_ismodule = ismodule(_self)             # <<<<<<<<<<<<<<
 *         if self_isbound and (self_ismodule or skip_bound_arg):
 *             parameters.pop(0)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ismodule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_14 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v__self) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v__self);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self_ismodule = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2187
 *         self_isbound = _self is not None
 *         self_ismodule = ismodule(_self)
 *         if self_isbound and (self_ismodule or skip_bound_arg):             # <<<<<<<<<<<<<<
 *             parameters.pop(0)
 *         else:
 */
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self_isbound); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2187, __pyx_L1_error)
    if (__pyx_t_11) {
    } else {
      __pyx_t_12 = __pyx_t_11;
      goto __pyx_L30_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self_ismodule); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2187, __pyx_L1_error)
    if (!__pyx_t_11) {
    } else {
      __pyx_t_12 = __pyx_t_11;
      goto __pyx_L30_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_skip_bound_arg); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2187, __pyx_L1_error)
    __pyx_t_12 = __pyx_t_11;
    __pyx_L30_bool_binop_done:;
    if (__pyx_t_12) {

      /* "Aspidites/_vendor/contracts/inspection.py":2188
 *         self_ismodule = ismodule(_self)
 *         if self_isbound and (self_ismodule or skip_bound_arg):
 *             parameters.pop(0)             # <<<<<<<<<<<<<<
 *         else:
 *             # for builtins, self parameter is always positional-only!
 */
      __pyx_t_14 = __Pyx_PyList_PopIndex(__pyx_cur_scope->__pyx_v_parameters, __pyx_int_0, 0, 1, Py_ssize_t, PyInt_FromSsize_t); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2187
 *         self_isbound = _self is not None
 *         self_ismodule = ismodule(_self)
 *         if self_isbound and (self_ismodule or skip_bound_arg):             # <<<<<<<<<<<<<<
 *             parameters.pop(0)
 *         else:
 */
      goto __pyx_L29;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2191
 *         else:
 *             # for builtins, self parameter is always positional-only!
 *             p = parameters[0].replace(kind=Parameter.POSITIONAL_ONLY)             # <<<<<<<<<<<<<<
 *             parameters[0] = p
 * 
 */
    /*else*/ {
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(PyList_GET_ITEM(__pyx_cur_scope->__pyx_v_parameters, 0), __pyx_n_s_replace); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_Parameter, __pyx_n_s_POSITIONAL_ONLY_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_kind, __pyx_t_4) < 0) __PYX_ERR(0, 2191, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2192
 *             # for builtins, self parameter is always positional-only!
 *             p = parameters[0].replace(kind=Parameter.POSITIONAL_ONLY)
 *             parameters[0] = p             # <<<<<<<<<<<<<<
 * 
 *     return cls(parameters, return_annotation=cls.empty)
 */
      if (unlikely(__Pyx_SetItemInt(__pyx_cur_scope->__pyx_v_parameters, 0, __pyx_v_p, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 2192, __pyx_L1_error)
    }
    __pyx_L29:;

    /* "Aspidites/_vendor/contracts/inspection.py":2177
 *         p(f.args.kwarg, empty)
 * 
 *     if self_parameter is not None:             # <<<<<<<<<<<<<<
 *         # Possibly strip the bound argument:
 *         #    - We *always* strip first bound argument if
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2194
 *             parameters[0] = p
 * 
 *     return cls(parameters, return_annotation=cls.empty)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_parameters);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_cur_scope->__pyx_v_parameters);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_empty_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_annotation, __pyx_t_14) < 0) __PYX_ERR(0, 2194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_v_cls, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_14;
  __pyx_t_14 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_fromstr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_clean_signature);
  __Pyx_XDECREF(__pyx_v_self_parameter);
  __Pyx_XDECREF(__pyx_v_last_positional_only);
  __Pyx_XDECREF(__pyx_v_program);
  __Pyx_XDECREF(__pyx_v_module);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_module_name);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_defaults);
  __Pyx_XDECREF(__pyx_v_iter);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_default);
  __Pyx_XDECREF(__pyx_v__self);
  __Pyx_XDECREF(__pyx_v_self_isbound);
  __Pyx_XDECREF(__pyx_v_self_ismodule);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2197
 * 
 * 
 * def _signature_from_builtin(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper function to get signature for
 *     builtin callables.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_163_signature_from_builtin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_162_signature_from_builtin[] = "Private helper function to get signature for\n    builtin callables.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_163_signature_from_builtin = {"_signature_from_builtin", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_163_signature_from_builtin, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_162_signature_from_builtin};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_163_signature_from_builtin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_skip_bound_arg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_from_builtin (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_func_2,&__pyx_n_s_skip_bound_arg,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_from_builtin", 0, 2, 3, 1); __PYX_ERR(0, 2197, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_skip_bound_arg);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_from_builtin") < 0)) __PYX_ERR(0, 2197, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_func = values[1];
    __pyx_v_skip_bound_arg = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_from_builtin", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2197, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_builtin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_162_signature_from_builtin(__pyx_self, __pyx_v_cls, __pyx_v_func, __pyx_v_skip_bound_arg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_162_signature_from_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_skip_bound_arg) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_from_builtin", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2202
 *     """
 * 
 *     if not _signature_is_builtin(func):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python builtin "
 *                         "function".format(func))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_signature_is_builtin); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2204
 *     if not _signature_is_builtin(func):
 *         raise TypeError("{!r} is not a Python builtin "
 *                         "function".format(func))             # <<<<<<<<<<<<<<
 * 
 *     s = getattr(func, "__text_signature__", None)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_Python_builtin_funct, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2203
 * 
 *     if not _signature_is_builtin(func):
 *         raise TypeError("{!r} is not a Python builtin "             # <<<<<<<<<<<<<<
 *                         "function".format(func))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2203, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2202
 *     """
 * 
 *     if not _signature_is_builtin(func):             # <<<<<<<<<<<<<<
 *         raise TypeError("{!r} is not a Python builtin "
 *                         "function".format(func))
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2206
 *                         "function".format(func))
 * 
 *     s = getattr(func, "__text_signature__", None)             # <<<<<<<<<<<<<<
 *     if not s:
 *         raise ValueError("no signature found for builtin {!r}".format(func))
 */
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_func, __pyx_n_u_text_signature, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_s = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2207
 * 
 *     s = getattr(func, "__text_signature__", None)
 *     if not s:             # <<<<<<<<<<<<<<
 *         raise ValueError("no signature found for builtin {!r}".format(func))
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_s); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2207, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2208
 *     s = getattr(func, "__text_signature__", None)
 *     if not s:
 *         raise ValueError("no signature found for builtin {!r}".format(func))             # <<<<<<<<<<<<<<
 * 
 *     return _signature_fromstr(cls, func, s, skip_bound_arg)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_no_signature_found_for_builtin_r, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_func);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2208, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2207
 * 
 *     s = getattr(func, "__text_signature__", None)
 *     if not s:             # <<<<<<<<<<<<<<
 *         raise ValueError("no signature found for builtin {!r}".format(func))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2210
 *         raise ValueError("no signature found for builtin {!r}".format(func))
 * 
 *     return _signature_fromstr(cls, func, s, skip_bound_arg)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_signature_fromstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_cls, __pyx_v_func, __pyx_v_s, __pyx_v_skip_bound_arg};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_cls, __pyx_v_func, __pyx_v_s, __pyx_v_skip_bound_arg};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_func);
    __Pyx_INCREF(__pyx_v_s);
    __Pyx_GIVEREF(__pyx_v_s);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_s);
    __Pyx_INCREF(__pyx_v_skip_bound_arg);
    __Pyx_GIVEREF(__pyx_v_skip_bound_arg);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_skip_bound_arg);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2197
 * 
 * 
 * def _signature_from_builtin(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper function to get signature for
 *     builtin callables.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_builtin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2213
 * 
 * 
 * def _signature_from_function(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper: constructs Signature for the given python function."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_165_signature_from_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_164_signature_from_function[] = "Private helper: constructs Signature for the given python function.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_165_signature_from_function = {"_signature_from_function", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_165_signature_from_function, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_164_signature_from_function};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_165_signature_from_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_skip_bound_arg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_from_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_func_2,&__pyx_n_s_skip_bound_arg,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_signature_from_function", 0, 2, 3, 1); __PYX_ERR(0, 2213, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_skip_bound_arg);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_from_function") < 0)) __PYX_ERR(0, 2213, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_func = values[1];
    __pyx_v_skip_bound_arg = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_from_function", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2213, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_164_signature_from_function(__pyx_self, __pyx_v_cls, __pyx_v_func, __pyx_v_skip_bound_arg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_164_signature_from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_skip_bound_arg) {
  int __pyx_v_is_duck_function;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_Parameter = NULL;
  PyObject *__pyx_v_func_code = NULL;
  PyObject *__pyx_v_pos_count = NULL;
  PyObject *__pyx_v_arg_names = NULL;
  PyObject *__pyx_v_posonly_count = NULL;
  PyObject *__pyx_v_positional = NULL;
  PyObject *__pyx_v_keyword_only_count = NULL;
  PyObject *__pyx_v_keyword_only = NULL;
  PyObject *__pyx_v_annotations = NULL;
  PyObject *__pyx_v_defaults = NULL;
  PyObject *__pyx_v_kwdefaults = NULL;
  PyObject *__pyx_v_pos_default_count = NULL;
  PyObject *__pyx_v_parameters = NULL;
  PyObject *__pyx_v_non_default_count = NULL;
  PyObject *__pyx_v_posonly_left = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_v_annotation = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_v_default = NULL;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_from_function", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2216
 *     """Private helper: constructs Signature for the given python function."""
 * 
 *     is_duck_function = False             # <<<<<<<<<<<<<<
 *     if not isfunction(func):
 *         if _signature_is_functionlike(func):
 */
  __pyx_v_is_duck_function = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2217
 * 
 *     is_duck_function = False
 *     if not isfunction(func):             # <<<<<<<<<<<<<<
 *         if _signature_is_functionlike(func):
 *             is_duck_function = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":2218
 *     is_duck_function = False
 *     if not isfunction(func):
 *         if _signature_is_functionlike(func):             # <<<<<<<<<<<<<<
 *             is_duck_function = True
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_signature_is_functionlike); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(__pyx_t_5)) {

      /* "Aspidites/_vendor/contracts/inspection.py":2219
 *     if not isfunction(func):
 *         if _signature_is_functionlike(func):
 *             is_duck_function = True             # <<<<<<<<<<<<<<
 *         else:
 *             # If it's not a pure Python function, and not a duck type
 */
      __pyx_v_is_duck_function = 1;

      /* "Aspidites/_vendor/contracts/inspection.py":2218
 *     is_duck_function = False
 *     if not isfunction(func):
 *         if _signature_is_functionlike(func):             # <<<<<<<<<<<<<<
 *             is_duck_function = True
 *         else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2223
 *             # If it's not a pure Python function, and not a duck type
 *             # of pure function:
 *             raise TypeError('{!r} is not a Python function'.format(func))             # <<<<<<<<<<<<<<
 * 
 *     s = getattr(func, "__text_signature__", None)
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_Python_function, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_func) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_func);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2223, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/contracts/inspection.py":2217
 * 
 *     is_duck_function = False
 *     if not isfunction(func):             # <<<<<<<<<<<<<<
 *         if _signature_is_functionlike(func):
 *             is_duck_function = True
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2225
 *             raise TypeError('{!r} is not a Python function'.format(func))
 * 
 *     s = getattr(func, "__text_signature__", None)             # <<<<<<<<<<<<<<
 *     if s:
 *         return _signature_fromstr(cls, func, s, skip_bound_arg)
 */
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_func, __pyx_n_u_text_signature, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_s = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2226
 * 
 *     s = getattr(func, "__text_signature__", None)
 *     if s:             # <<<<<<<<<<<<<<
 *         return _signature_fromstr(cls, func, s, skip_bound_arg)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_s); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2226, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":2227
 *     s = getattr(func, "__text_signature__", None)
 *     if s:
 *         return _signature_fromstr(cls, func, s, skip_bound_arg)             # <<<<<<<<<<<<<<
 * 
 *     Parameter = cls._parameter_cls
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_signature_fromstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_cls, __pyx_v_func, __pyx_v_s, __pyx_v_skip_bound_arg};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_cls, __pyx_v_func, __pyx_v_s, __pyx_v_skip_bound_arg};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_cls);
      __Pyx_GIVEREF(__pyx_v_cls);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_cls);
      __Pyx_INCREF(__pyx_v_func);
      __Pyx_GIVEREF(__pyx_v_func);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_func);
      __Pyx_INCREF(__pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_s);
      __Pyx_INCREF(__pyx_v_skip_bound_arg);
      __Pyx_GIVEREF(__pyx_v_skip_bound_arg);
      PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_skip_bound_arg);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2226
 * 
 *     s = getattr(func, "__text_signature__", None)
 *     if s:             # <<<<<<<<<<<<<<
 *         return _signature_fromstr(cls, func, s, skip_bound_arg)
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2229
 *         return _signature_fromstr(cls, func, s, skip_bound_arg)
 * 
 *     Parameter = cls._parameter_cls             # <<<<<<<<<<<<<<
 * 
 *     # Parameter information.
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_parameter_cls); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_Parameter = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2232
 * 
 *     # Parameter information.
 *     func_code = func.__code__             # <<<<<<<<<<<<<<
 *     pos_count = func_code.co_argcount
 *     arg_names = func_code.co_varnames
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_func_code = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2233
 *     # Parameter information.
 *     func_code = func.__code__
 *     pos_count = func_code.co_argcount             # <<<<<<<<<<<<<<
 *     arg_names = func_code.co_varnames
 *     posonly_count = func_code.co_posonlyargcount
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_argcount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_pos_count = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2234
 *     func_code = func.__code__
 *     pos_count = func_code.co_argcount
 *     arg_names = func_code.co_varnames             # <<<<<<<<<<<<<<
 *     posonly_count = func_code.co_posonlyargcount
 *     positional = arg_names[:pos_count]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_varnames); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_arg_names = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2235
 *     pos_count = func_code.co_argcount
 *     arg_names = func_code.co_varnames
 *     posonly_count = func_code.co_posonlyargcount             # <<<<<<<<<<<<<<
 *     positional = arg_names[:pos_count]
 *     keyword_only_count = func_code.co_kwonlyargcount
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_posonlyargcount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_posonly_count = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2236
 *     arg_names = func_code.co_varnames
 *     posonly_count = func_code.co_posonlyargcount
 *     positional = arg_names[:pos_count]             # <<<<<<<<<<<<<<
 *     keyword_only_count = func_code.co_kwonlyargcount
 *     keyword_only = arg_names[pos_count:pos_count + keyword_only_count]
 */
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_arg_names, 0, 0, NULL, &__pyx_v_pos_count, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_positional = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2237
 *     posonly_count = func_code.co_posonlyargcount
 *     positional = arg_names[:pos_count]
 *     keyword_only_count = func_code.co_kwonlyargcount             # <<<<<<<<<<<<<<
 *     keyword_only = arg_names[pos_count:pos_count + keyword_only_count]
 *     annotations = func.__annotations__
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_kwonlyargcount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_keyword_only_count = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2238
 *     positional = arg_names[:pos_count]
 *     keyword_only_count = func_code.co_kwonlyargcount
 *     keyword_only = arg_names[pos_count:pos_count + keyword_only_count]             # <<<<<<<<<<<<<<
 *     annotations = func.__annotations__
 *     defaults = func.__defaults__
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_pos_count, __pyx_v_keyword_only_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_arg_names, 0, 0, &__pyx_v_pos_count, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_keyword_only = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2239
 *     keyword_only_count = func_code.co_kwonlyargcount
 *     keyword_only = arg_names[pos_count:pos_count + keyword_only_count]
 *     annotations = func.__annotations__             # <<<<<<<<<<<<<<
 *     defaults = func.__defaults__
 *     kwdefaults = func.__kwdefaults__
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_annotations_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_annotations = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2240
 *     keyword_only = arg_names[pos_count:pos_count + keyword_only_count]
 *     annotations = func.__annotations__
 *     defaults = func.__defaults__             # <<<<<<<<<<<<<<
 *     kwdefaults = func.__kwdefaults__
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_defaults_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_defaults = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2241
 *     annotations = func.__annotations__
 *     defaults = func.__defaults__
 *     kwdefaults = func.__kwdefaults__             # <<<<<<<<<<<<<<
 * 
 *     if defaults:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_kwdefaults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_kwdefaults = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2243
 *     kwdefaults = func.__kwdefaults__
 * 
 *     if defaults:             # <<<<<<<<<<<<<<
 *         pos_default_count = len(defaults)
 *     else:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_defaults); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2243, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":2244
 * 
 *     if defaults:
 *         pos_default_count = len(defaults)             # <<<<<<<<<<<<<<
 *     else:
 *         pos_default_count = 0
 */
    __pyx_t_8 = PyObject_Length(__pyx_v_defaults); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2244, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_pos_default_count = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2243
 *     kwdefaults = func.__kwdefaults__
 * 
 *     if defaults:             # <<<<<<<<<<<<<<
 *         pos_default_count = len(defaults)
 *     else:
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2246
 *         pos_default_count = len(defaults)
 *     else:
 *         pos_default_count = 0             # <<<<<<<<<<<<<<
 * 
 *     parameters = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_pos_default_count = __pyx_int_0;
  }
  __pyx_L6:;

  /* "Aspidites/_vendor/contracts/inspection.py":2248
 *         pos_default_count = 0
 * 
 *     parameters = []             # <<<<<<<<<<<<<<
 * 
 *     non_default_count = pos_count - pos_default_count
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parameters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2250
 *     parameters = []
 * 
 *     non_default_count = pos_count - pos_default_count             # <<<<<<<<<<<<<<
 *     posonly_left = posonly_count
 * 
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_pos_count, __pyx_v_pos_default_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_non_default_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2251
 * 
 *     non_default_count = pos_count - pos_default_count
 *     posonly_left = posonly_count             # <<<<<<<<<<<<<<
 * 
 *     # Non-keyword-only parameters w/o defaults.
 */
  __Pyx_INCREF(__pyx_v_posonly_count);
  __pyx_v_posonly_left = __pyx_v_posonly_count;

  /* "Aspidites/_vendor/contracts/inspection.py":2254
 * 
 *     # Non-keyword-only parameters w/o defaults.
 *     for name in positional[:non_default_count]:             # <<<<<<<<<<<<<<
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_positional, 0, 0, NULL, &__pyx_v_non_default_count, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2254, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2254, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2254, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2254, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2255
 *     # Non-keyword-only parameters w/o defaults.
 *     for name in positional[:non_default_count]:
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_posonly_left); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2255, __pyx_L1_error)
    if (__pyx_t_5) {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __pyx_t_7;
      __pyx_t_7 = 0;
    } else {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __pyx_t_7;
      __pyx_t_7 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2256
 *     for name in positional[:non_default_count]:
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind))
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_name, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_name, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_annotation, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2257
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=kind))
 *         if posonly_left:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
    __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2257, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2258
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind))             # <<<<<<<<<<<<<<
 *         if posonly_left:
 *             posonly_left -= 1
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_kind, __pyx_v_kind) < 0) __PYX_ERR(0, 2257, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2257
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=kind))
 *         if posonly_left:
 */
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_v_Parameter, __pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_11); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 2257, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2259
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind))
 *         if posonly_left:             # <<<<<<<<<<<<<<
 *             posonly_left -= 1
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_posonly_left); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2259, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":2260
 *                                     kind=kind))
 *         if posonly_left:
 *             posonly_left -= 1             # <<<<<<<<<<<<<<
 * 
 *     # ... w/ defaults.
 */
      __pyx_t_11 = __Pyx_PyInt_SubtractObjC(__pyx_v_posonly_left, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF_SET(__pyx_v_posonly_left, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2259
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind))
 *         if posonly_left:             # <<<<<<<<<<<<<<
 *             posonly_left -= 1
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2254
 * 
 *     # Non-keyword-only parameters w/o defaults.
 *     for name in positional[:non_default_count]:             # <<<<<<<<<<<<<<
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2263
 * 
 *     # ... w/ defaults.
 *     for offset, name in enumerate(positional[non_default_count:]):             # <<<<<<<<<<<<<<
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_v_positional, 0, 0, &__pyx_v_non_default_count, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
    __pyx_t_7 = __pyx_t_11; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2263, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_11 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2263, __pyx_L1_error)
        #else
        __pyx_t_11 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2263, __pyx_L1_error)
        #else
        __pyx_t_11 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
      }
    } else {
      __pyx_t_11 = __pyx_t_9(__pyx_t_7);
      if (unlikely(!__pyx_t_11)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2263, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_11);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_11);
    __pyx_t_11 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_offset, __pyx_t_2);
    __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_11;
    __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2264
 *     # ... w/ defaults.
 *     for offset, name in enumerate(positional[non_default_count:]):
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_posonly_left); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2264, __pyx_L1_error)
    if (__pyx_t_5) {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __pyx_t_1;
      __pyx_t_1 = 0;
    } else {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2265
 *     for offset, name in enumerate(positional[non_default_count:]):
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind,
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_name, __pyx_t_3};
      __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2265, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_name, __pyx_t_3};
      __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2265, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_annotation, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2266
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=kind,
 *                                     default=defaults[offset]))
 */
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_name);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2266, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2267
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind,             # <<<<<<<<<<<<<<
 *                                     default=defaults[offset]))
 *         if posonly_left:
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_kind, __pyx_v_kind) < 0) __PYX_ERR(0, 2266, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2268
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=kind,
 *                                     default=defaults[offset]))             # <<<<<<<<<<<<<<
 *         if posonly_left:
 *             posonly_left -= 1
 */
    __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_v_defaults, __pyx_v_offset); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_default, __pyx_t_13) < 0) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2266
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=kind,
 *                                     default=defaults[offset]))
 */
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_v_Parameter, __pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_13); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2269
 *                                     kind=kind,
 *                                     default=defaults[offset]))
 *         if posonly_left:             # <<<<<<<<<<<<<<
 *             posonly_left -= 1
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_posonly_left); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2269, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":2270
 *                                     default=defaults[offset]))
 *         if posonly_left:
 *             posonly_left -= 1             # <<<<<<<<<<<<<<
 * 
 *     # *args
 */
      __pyx_t_13 = __Pyx_PyInt_SubtractObjC(__pyx_v_posonly_left, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF_SET(__pyx_v_posonly_left, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2269
 *                                     kind=kind,
 *                                     default=defaults[offset]))
 *         if posonly_left:             # <<<<<<<<<<<<<<
 *             posonly_left -= 1
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2263
 * 
 *     # ... w/ defaults.
 *     for offset, name in enumerate(positional[non_default_count:]):             # <<<<<<<<<<<<<<
 *         kind = _POSITIONAL_ONLY if posonly_left else _POSITIONAL_OR_KEYWORD
 *         annotation = annotations.get(name, _empty)
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2273
 * 
 *     # *args
 *     if func_code.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *         name = arg_names[pos_count + keyword_only_count]
 *         annotation = annotations.get(name, _empty)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_CO_VARARGS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = PyNumber_And(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/contracts/inspection.py":2274
 *     # *args
 *     if func_code.co_flags & CO_VARARGS:
 *         name = arg_names[pos_count + keyword_only_count]             # <<<<<<<<<<<<<<
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 */
    __pyx_t_13 = PyNumber_Add(__pyx_v_pos_count, __pyx_v_keyword_only_count); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_arg_names, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2275
 *     if func_code.co_flags & CO_VARARGS:
 *         name = arg_names[pos_count + keyword_only_count]
 *         annotation = annotations.get(name, _empty)             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_VAR_POSITIONAL))
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_t_2};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2275, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_t_2};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2275, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_annotation, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2276
 *         name = arg_names[pos_count + keyword_only_count]
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_VAR_POSITIONAL))
 * 
 */
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_name);
    __pyx_t_13 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2276, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2277
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_VAR_POSITIONAL))             # <<<<<<<<<<<<<<
 * 
 *     # Keyword-only parameters.
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_kind, __pyx_t_11) < 0) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2276
 *         name = arg_names[pos_count + keyword_only_count]
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_VAR_POSITIONAL))
 * 
 */
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_v_Parameter, __pyx_t_7, __pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_11); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 2276, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2273
 * 
 *     # *args
 *     if func_code.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *         name = arg_names[pos_count + keyword_only_count]
 *         annotation = annotations.get(name, _empty)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2280
 * 
 *     # Keyword-only parameters.
 *     for name in keyword_only:             # <<<<<<<<<<<<<<
 *         default = _empty
 *         if kwdefaults is not None:
 */
  if (likely(PyList_CheckExact(__pyx_v_keyword_only)) || PyTuple_CheckExact(__pyx_v_keyword_only)) {
    __pyx_t_11 = __pyx_v_keyword_only; __Pyx_INCREF(__pyx_t_11); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_v_keyword_only); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_9 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2280, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_11))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_11)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_13 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_8); __Pyx_INCREF(__pyx_t_13); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2280, __pyx_L1_error)
        #else
        __pyx_t_13 = PySequence_ITEM(__pyx_t_11, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_8); __Pyx_INCREF(__pyx_t_13); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 2280, __pyx_L1_error)
        #else
        __pyx_t_13 = PySequence_ITEM(__pyx_t_11, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        #endif
      }
    } else {
      __pyx_t_13 = __pyx_t_9(__pyx_t_11);
      if (unlikely(!__pyx_t_13)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2280, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_13);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_13);
    __pyx_t_13 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2281
 *     # Keyword-only parameters.
 *     for name in keyword_only:
 *         default = _empty             # <<<<<<<<<<<<<<
 *         if kwdefaults is not None:
 *             default = kwdefaults.get(name, _empty)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_empty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_XDECREF_SET(__pyx_v_default, __pyx_t_13);
    __pyx_t_13 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2282
 *     for name in keyword_only:
 *         default = _empty
 *         if kwdefaults is not None:             # <<<<<<<<<<<<<<
 *             default = kwdefaults.get(name, _empty)
 * 
 */
    __pyx_t_5 = (__pyx_v_kwdefaults != Py_None);
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":2283
 *         default = _empty
 *         if kwdefaults is not None:
 *             default = kwdefaults.get(name, _empty)             # <<<<<<<<<<<<<<
 * 
 *         annotation = annotations.get(name, _empty)
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwdefaults, __pyx_n_s_get_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_t_2};
        __pyx_t_13 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_t_2};
        __pyx_t_13 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_default, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2282
 *     for name in keyword_only:
 *         default = _empty
 *         if kwdefaults is not None:             # <<<<<<<<<<<<<<
 *             default = kwdefaults.get(name, _empty)
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2285
 *             default = kwdefaults.get(name, _empty)
 * 
 *         annotation = annotations.get(name, _empty)             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_KEYWORD_ONLY,
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_name, __pyx_t_3};
      __pyx_t_13 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2285, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_name, __pyx_t_3};
      __pyx_t_13 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2285, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_annotation, __pyx_t_13);
    __pyx_t_13 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2286
 * 
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_KEYWORD_ONLY,
 *                                     default=default))
 */
    __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_name);
    __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2286, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2287
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_KEYWORD_ONLY,             # <<<<<<<<<<<<<<
 *                                     default=default))
 *     # **kwargs
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_kind, __pyx_t_1) < 0) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2288
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_KEYWORD_ONLY,
 *                                     default=default))             # <<<<<<<<<<<<<<
 *     # **kwargs
 *     if func_code.co_flags & CO_VARKEYWORDS:
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, __pyx_v_default) < 0) __PYX_ERR(0, 2286, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2286
 * 
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_KEYWORD_ONLY,
 *                                     default=default))
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_Parameter, __pyx_t_13, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2280
 * 
 *     # Keyword-only parameters.
 *     for name in keyword_only:             # <<<<<<<<<<<<<<
 *         default = _empty
 *         if kwdefaults is not None:
 */
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2290
 *                                     default=default))
 *     # **kwargs
 *     if func_code.co_flags & CO_VARKEYWORDS:             # <<<<<<<<<<<<<<
 *         index = pos_count + keyword_only_count
 *         if func_code.co_flags & CO_VARARGS:
 */
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CO_VARKEYWORDS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyNumber_And(__pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":2291
 *     # **kwargs
 *     if func_code.co_flags & CO_VARKEYWORDS:
 *         index = pos_count + keyword_only_count             # <<<<<<<<<<<<<<
 *         if func_code.co_flags & CO_VARARGS:
 *             index += 1
 */
    __pyx_t_7 = PyNumber_Add(__pyx_v_pos_count, __pyx_v_keyword_only_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_index = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2292
 *     if func_code.co_flags & CO_VARKEYWORDS:
 *         index = pos_count + keyword_only_count
 *         if func_code.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *             index += 1
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_func_code, __pyx_n_s_co_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CO_VARARGS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PyNumber_And(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":2293
 *         index = pos_count + keyword_only_count
 *         if func_code.co_flags & CO_VARARGS:
 *             index += 1             # <<<<<<<<<<<<<<
 * 
 *         name = arg_names[index]
 */
      __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_v_index, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2292
 *     if func_code.co_flags & CO_VARKEYWORDS:
 *         index = pos_count + keyword_only_count
 *         if func_code.co_flags & CO_VARARGS:             # <<<<<<<<<<<<<<
 *             index += 1
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2295
 *             index += 1
 * 
 *         name = arg_names[index]             # <<<<<<<<<<<<<<
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 */
    __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_arg_names, __pyx_v_index); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2296
 * 
 *         name = arg_names[index]
 *         annotation = annotations.get(name, _empty)             # <<<<<<<<<<<<<<
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_VAR_KEYWORD))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_name, __pyx_t_7};
      __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2296, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_name, __pyx_t_7};
      __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2296, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_annotation, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2297
 *         name = arg_names[index]
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_VAR_KEYWORD))
 * 
 */
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_name);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2297, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2298
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,
 *                                     kind=_VAR_KEYWORD))             # <<<<<<<<<<<<<<
 * 
 *     # Is 'func' is a pure Python function - don't validate the
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_kind, __pyx_t_3) < 0) __PYX_ERR(0, 2297, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2297
 *         name = arg_names[index]
 *         annotation = annotations.get(name, _empty)
 *         parameters.append(Parameter(name, annotation=annotation,             # <<<<<<<<<<<<<<
 *                                     kind=_VAR_KEYWORD))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_Parameter, __pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 2297, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2290
 *                                     default=default))
 *     # **kwargs
 *     if func_code.co_flags & CO_VARKEYWORDS:             # <<<<<<<<<<<<<<
 *         index = pos_count + keyword_only_count
 *         if func_code.co_flags & CO_VARARGS:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2302
 *     # Is 'func' is a pure Python function - don't validate the
 *     # parameters list (for correct order and defaults), it should be OK.
 *     return cls(parameters,             # <<<<<<<<<<<<<<
 *                return_annotation=annotations.get('return', _empty),
 *                __validate_parameters__=is_duck_function)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_v_parameters);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_parameters);

  /* "Aspidites/_vendor/contracts/inspection.py":2303
 *     # parameters list (for correct order and defaults), it should be OK.
 *     return cls(parameters,
 *                return_annotation=annotations.get('return', _empty),             # <<<<<<<<<<<<<<
 *                __validate_parameters__=is_duck_function)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_annotations, __pyx_n_s_get_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_empty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_n_u_return, __pyx_t_13};
    __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2303, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_n_u_return, __pyx_t_13};
    __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2303, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_n_u_return);
    __Pyx_GIVEREF(__pyx_n_u_return);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_6, __pyx_n_u_return);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_6, __pyx_t_13);
    __pyx_t_13 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_annotation, __pyx_t_11) < 0) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2304
 *     return cls(parameters,
 *                return_annotation=annotations.get('return', _empty),
 *                __validate_parameters__=is_duck_function)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_v_is_duck_function); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_validate_parameters, __pyx_t_11) < 0) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2302
 *     # Is 'func' is a pure Python function - don't validate the
 *     # parameters list (for correct order and defaults), it should be OK.
 *     return cls(parameters,             # <<<<<<<<<<<<<<
 *                return_annotation=annotations.get('return', _empty),
 *                __validate_parameters__=is_duck_function)
 */
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_v_cls, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2213
 * 
 * 
 * def _signature_from_function(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper: constructs Signature for the given python function."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_Parameter);
  __Pyx_XDECREF(__pyx_v_func_code);
  __Pyx_XDECREF(__pyx_v_pos_count);
  __Pyx_XDECREF(__pyx_v_arg_names);
  __Pyx_XDECREF(__pyx_v_posonly_count);
  __Pyx_XDECREF(__pyx_v_positional);
  __Pyx_XDECREF(__pyx_v_keyword_only_count);
  __Pyx_XDECREF(__pyx_v_keyword_only);
  __Pyx_XDECREF(__pyx_v_annotations);
  __Pyx_XDECREF(__pyx_v_defaults);
  __Pyx_XDECREF(__pyx_v_kwdefaults);
  __Pyx_XDECREF(__pyx_v_pos_default_count);
  __Pyx_XDECREF(__pyx_v_parameters);
  __Pyx_XDECREF(__pyx_v_non_default_count);
  __Pyx_XDECREF(__pyx_v_posonly_left);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_annotation);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XDECREF(__pyx_v_default);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_167_signature_from_callable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_166_signature_from_callable[] = "Private helper function to get signature for arbitrary\n    callable objects.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_167_signature_from_callable = {"_signature_from_callable", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_167_signature_from_callable, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_166_signature_from_callable};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_167_signature_from_callable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_follow_wrapper_chains = 0;
  PyObject *__pyx_v_skip_bound_arg = 0;
  PyObject *__pyx_v_sigcls = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_signature_from_callable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_sigcls,&__pyx_n_s_follow_wrapper_chains,&__pyx_n_s_skip_bound_arg,0};
    PyObject* values[4] = {0,0,0,0};

    /* "Aspidites/_vendor/contracts/inspection.py":2308
 * 
 * def _signature_from_callable(obj, *,
 *                              follow_wrapper_chains=True,             # <<<<<<<<<<<<<<
 *                              skip_bound_arg=True,
 *                              sigcls):
 */
    values[2] = ((PyObject *)((PyObject *)Py_True));

    /* "Aspidites/_vendor/contracts/inspection.py":2309
 * def _signature_from_callable(obj, *,
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,             # <<<<<<<<<<<<<<
 *                              sigcls):
 *     """Private helper function to get signature for arbitrary
 */
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (likely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sigcls)) != 0)) kw_args--;
        else {
          __Pyx_RaiseKeywordRequired("_signature_from_callable", __pyx_n_s_sigcls); __PYX_ERR(0, 2307, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 2; index < 4 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_signature_from_callable") < 0)) __PYX_ERR(0, 2307, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      __Pyx_RaiseKeywordRequired("_signature_from_callable", __pyx_n_s_sigcls); __PYX_ERR(0, 2307, __pyx_L3_error)
    }
    __pyx_v_obj = values[0];
    __pyx_v_sigcls = values[1];
    __pyx_v_follow_wrapper_chains = values[2];
    __pyx_v_skip_bound_arg = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_signature_from_callable", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2307, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_166_signature_from_callable(__pyx_self, __pyx_v_obj, __pyx_v_follow_wrapper_chains, __pyx_v_skip_bound_arg, __pyx_v_sigcls);

  /* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2335
 *     # Was this function wrapped by a decorator?
 *     if follow_wrapper_chains:
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))             # <<<<<<<<<<<<<<
 *         if isinstance(obj, types.MethodType):
 *             # If the unwrapped object is a *method*, we might want to
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_24_signature_from_callable_lambda11(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_24_signature_from_callable_lambda11 = {"lambda11", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_24_signature_from_callable_lambda11, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_24_signature_from_callable_lambda11(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda11 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda11(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda11(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda11", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_f, __pyx_n_u_signature); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2335, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_callable.lambda11", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_166_signature_from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapper_chains, PyObject *__pyx_v_skip_bound_arg, PyObject *__pyx_v_sigcls) {
  PyObject *__pyx_v_sig = NULL;
  PyObject *__pyx_v_partialmethod = NULL;
  PyObject *__pyx_v_wrapped_sig = NULL;
  PyObject *__pyx_v_first_wrapped_param = NULL;
  PyObject *__pyx_v_sig_params = NULL;
  PyObject *__pyx_v_new_params = NULL;
  PyObject *__pyx_v_call = NULL;
  PyObject *__pyx_v_new = NULL;
  PyObject *__pyx_v_init = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_v_text_sig = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_signature_from_callable", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "Aspidites/_vendor/contracts/inspection.py":2315
 *     """
 * 
 *     if not callable(obj):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a callable object'.format(obj))
 * 
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2315, __pyx_L1_error)
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2316
 * 
 *     if not callable(obj):
 *         raise TypeError('{!r} is not a callable object'.format(obj))             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(obj, types.MethodType):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_callable_object, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2316, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2315
 *     """
 * 
 *     if not callable(obj):             # <<<<<<<<<<<<<<
 *         raise TypeError('{!r} is not a callable object'.format(obj))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2318
 *         raise TypeError('{!r} is not a callable object'.format(obj))
 * 
 *     if isinstance(obj, types.MethodType):             # <<<<<<<<<<<<<<
 *         # In this case we skip the first parameter of the underlying
 *         # function (usually `self` or `cls`).
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2322
 *         # function (usually `self` or `cls`).
 *         # noinspection PyTypeChecker
 *         sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.__func__,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/contracts/inspection.py":2323
 *         # noinspection PyTypeChecker
 *         sig = _signature_from_callable(
 *             obj.__func__,             # <<<<<<<<<<<<<<
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_func); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "Aspidites/_vendor/contracts/inspection.py":2322
 *         # function (usually `self` or `cls`).
 *         # noinspection PyTypeChecker
 *         sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.__func__,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2324
 *         sig = _signature_from_callable(
 *             obj.__func__,
 *             follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *             skip_bound_arg=skip_bound_arg,
 *             sigcls=sigcls)
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2324, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2325
 *             obj.__func__,
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *             sigcls=sigcls)
 * 
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2324, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2326
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,
 *             sigcls=sigcls)             # <<<<<<<<<<<<<<
 * 
 *         if skip_bound_arg:
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2324, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2322
 *         # function (usually `self` or `cls`).
 *         # noinspection PyTypeChecker
 *         sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.__func__,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_sig = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2328
 *             sigcls=sigcls)
 * 
 *         if skip_bound_arg:             # <<<<<<<<<<<<<<
 *             return _signature_bound_method(sig)
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_skip_bound_arg); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2328, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/contracts/inspection.py":2329
 * 
 *         if skip_bound_arg:
 *             return _signature_bound_method(sig)             # <<<<<<<<<<<<<<
 *         else:
 *             return sig
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_signature_bound_method); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_sig) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_sig);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2329, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":2328
 *             sigcls=sigcls)
 * 
 *         if skip_bound_arg:             # <<<<<<<<<<<<<<
 *             return _signature_bound_method(sig)
 *         else:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2331
 *             return _signature_bound_method(sig)
 *         else:
 *             return sig             # <<<<<<<<<<<<<<
 * 
 *     # Was this function wrapped by a decorator?
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_sig);
      __pyx_r = __pyx_v_sig;
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2318
 *         raise TypeError('{!r} is not a callable object'.format(obj))
 * 
 *     if isinstance(obj, types.MethodType):             # <<<<<<<<<<<<<<
 *         # In this case we skip the first parameter of the underlying
 *         # function (usually `self` or `cls`).
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2334
 * 
 *     # Was this function wrapped by a decorator?
 *     if follow_wrapper_chains:             # <<<<<<<<<<<<<<
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))
 *         if isinstance(obj, types.MethodType):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_follow_wrapper_chains); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2334, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2335
 *     # Was this function wrapped by a decorator?
 *     if follow_wrapper_chains:
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))             # <<<<<<<<<<<<<<
 *         if isinstance(obj, types.MethodType):
 *             # If the unwrapped object is a *method*, we might want to
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_unwrap); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_obj);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_24_signature_from_callable_lambda11, 0, __pyx_n_s_signature_from_callable_locals, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_stop, __pyx_t_3) < 0) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2336
 *     if follow_wrapper_chains:
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))
 *         if isinstance(obj, types.MethodType):             # <<<<<<<<<<<<<<
 *             # If the unwrapped object is a *method*, we might want to
 *             # skip its first parameter (self).
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_5); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2341
 *             # See test_signature_wrapped_bound_method for details.
 *             # noinspection PyTypeChecker
 *             return _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 obj,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "Aspidites/_vendor/contracts/inspection.py":2342
 *             # noinspection PyTypeChecker
 *             return _signature_from_callable(
 *                 obj,             # <<<<<<<<<<<<<<
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 */
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_obj);
      __Pyx_GIVEREF(__pyx_v_obj);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_obj);

      /* "Aspidites/_vendor/contracts/inspection.py":2343
 *             return _signature_from_callable(
 *                 obj,
 *                 follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2344
 *                 obj,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                 sigcls=sigcls)
 * 
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2345
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2341
 *             # See test_signature_wrapped_bound_method for details.
 *             # noinspection PyTypeChecker
 *             return _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 obj,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":2336
 *     if follow_wrapper_chains:
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))
 *         if isinstance(obj, types.MethodType):             # <<<<<<<<<<<<<<
 *             # If the unwrapped object is a *method*, we might want to
 *             # skip its first parameter (self).
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2334
 * 
 *     # Was this function wrapped by a decorator?
 *     if follow_wrapper_chains:             # <<<<<<<<<<<<<<
 *         obj = unwrap(obj, stop=(lambda f: hasattr(f, "__signature__")))
 *         if isinstance(obj, types.MethodType):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2347
 *                 sigcls=sigcls)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         sig = obj.__signature__
 *     except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":2348
 * 
 *     try:
 *         sig = obj.__signature__             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         pass
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_signature); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2348, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_sig = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2347
 *                 sigcls=sigcls)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         sig = obj.__signature__
 *     except AttributeError:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2352
 *         pass
 *     else:
 *         if sig is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(sig, Signature):
 *                 raise TypeError(
 */
    /*else:*/ {
      __pyx_t_2 = (__pyx_v_sig != Py_None);
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "Aspidites/_vendor/contracts/inspection.py":2353
 *     else:
 *         if sig is not None:
 *             if not isinstance(sig, Signature):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'unexpected object {!r} in __signature__ '
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Signature); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2353, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = PyObject_IsInstance(__pyx_v_sig, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2353, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
        if (unlikely(__pyx_t_2)) {

          /* "Aspidites/_vendor/contracts/inspection.py":2356
 *                 raise TypeError(
 *                     'unexpected object {!r} in __signature__ '
 *                     'attribute'.format(sig))             # <<<<<<<<<<<<<<
 *             return sig
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unexpected_object_r_in___signatu, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2356, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_sig) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_sig);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2356, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2354
 *         if sig is not None:
 *             if not isinstance(sig, Signature):
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'unexpected object {!r} in __signature__ '
 *                     'attribute'.format(sig))
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2354, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 2354, __pyx_L10_except_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2353
 *     else:
 *         if sig is not None:
 *             if not isinstance(sig, Signature):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'unexpected object {!r} in __signature__ '
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2357
 *                     'unexpected object {!r} in __signature__ '
 *                     'attribute'.format(sig))
 *             return sig             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_sig);
        __pyx_r = __pyx_v_sig;
        goto __pyx_L11_except_return;

        /* "Aspidites/_vendor/contracts/inspection.py":2352
 *         pass
 *     else:
 *         if sig is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(sig, Signature):
 *                 raise TypeError(
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2349
 *     try:
 *         sig = obj.__signature__
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         pass
 *     else:
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L9_exception_handled;
    }
    goto __pyx_L10_except_error;
    __pyx_L10_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":2347
 *                 sigcls=sigcls)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         sig = obj.__signature__
 *     except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L11_except_return:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L0;
    __pyx_L9_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L13_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2359
 *             return sig
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # noinspection PyProtectedMember
 *         partialmethod = obj._partialmethod
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":2361
 *     try:
 *         # noinspection PyProtectedMember
 *         partialmethod = obj._partialmethod             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         pass
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_partialmethod); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2361, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_partialmethod = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2359
 *             return sig
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # noinspection PyProtectedMember
 *         partialmethod = obj._partialmethod
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2365
 *         pass
 *     else:
 *         if isinstance(partialmethod, functools.partialmethod):             # <<<<<<<<<<<<<<
 *             # Unbound partialmethod (see functools.partialmethod)
 *             # This means, that we need to calculate the signature
 */
    /*else:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_functools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2365, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_partialmethod_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2365, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = PyObject_IsInstance(__pyx_v_partialmethod, __pyx_t_6); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2365, __pyx_L18_except_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "Aspidites/_vendor/contracts/inspection.py":2373
 *             # automatically (as for boundmethods)
 * 
 *             wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 partialmethod.func,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2373, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "Aspidites/_vendor/contracts/inspection.py":2374
 * 
 *             wrapped_sig = _signature_from_callable(
 *                 partialmethod.func,             # <<<<<<<<<<<<<<
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_partialmethod, __pyx_n_s_func_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2374, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "Aspidites/_vendor/contracts/inspection.py":2373
 *             # automatically (as for boundmethods)
 * 
 *             wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 partialmethod.func,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2373, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2375
 *             wrapped_sig = _signature_from_callable(
 *                 partialmethod.func,
 *                 follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)
 */
        __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2375, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2375, __pyx_L18_except_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2376
 *                 partialmethod.func,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                 sigcls=sigcls)
 * 
 */
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2375, __pyx_L18_except_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2377
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)             # <<<<<<<<<<<<<<
 * 
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))
 */
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2375, __pyx_L18_except_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2373
 *             # automatically (as for boundmethods)
 * 
 *             wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 partialmethod.func,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2373, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_wrapped_sig = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2379
 *                 sigcls=sigcls)
 * 
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))             # <<<<<<<<<<<<<<
 *             first_wrapped_param = tuple(wrapped_sig.parameters.values())[0]
 *             if first_wrapped_param.kind is Parameter.VAR_POSITIONAL:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_signature_get_partial); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2379, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_wrapped_sig, __pyx_v_partialmethod, __pyx_tuple__71};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2379, __pyx_L18_except_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_wrapped_sig, __pyx_v_partialmethod, __pyx_tuple__71};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2379, __pyx_L18_except_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2379, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(__pyx_v_wrapped_sig);
          __Pyx_GIVEREF(__pyx_v_wrapped_sig);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_v_wrapped_sig);
          __Pyx_INCREF(__pyx_v_partialmethod);
          __Pyx_GIVEREF(__pyx_v_partialmethod);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_v_partialmethod);
          __Pyx_INCREF(__pyx_tuple__71);
          __Pyx_GIVEREF(__pyx_tuple__71);
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_10, __pyx_tuple__71);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2379, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2380
 * 
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))
 *             first_wrapped_param = tuple(wrapped_sig.parameters.values())[0]             # <<<<<<<<<<<<<<
 *             if first_wrapped_param.kind is Parameter.VAR_POSITIONAL:
 *                 # First argument of the wrapped callable is `*args`, as in
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_wrapped_sig, __pyx_n_s_parameters); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2380, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2380, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2380, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2380, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, 0);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_first_wrapped_param = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2381
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))
 *             first_wrapped_param = tuple(wrapped_sig.parameters.values())[0]
 *             if first_wrapped_param.kind is Parameter.VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # First argument of the wrapped callable is `*args`, as in
 *                 # `partialmethod(lambda *args)`.
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_first_wrapped_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2381, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Parameter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2381, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_VAR_POSITIONAL_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2381, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_1 = (__pyx_t_5 == __pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_2 = (__pyx_t_1 != 0);
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/contracts/inspection.py":2384
 *                 # First argument of the wrapped callable is `*args`, as in
 *                 # `partialmethod(lambda *args)`.
 *                 return sig             # <<<<<<<<<<<<<<
 *             else:
 *                 sig_params = tuple(sig.parameters.values())
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_sig);
          __pyx_r = __pyx_v_sig;
          goto __pyx_L19_except_return;

          /* "Aspidites/_vendor/contracts/inspection.py":2381
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))
 *             first_wrapped_param = tuple(wrapped_sig.parameters.values())[0]
 *             if first_wrapped_param.kind is Parameter.VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # First argument of the wrapped callable is `*args`, as in
 *                 # `partialmethod(lambda *args)`.
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2386
 *                 return sig
 *             else:
 *                 sig_params = tuple(sig.parameters.values())             # <<<<<<<<<<<<<<
 *                 assert (not sig_params or
 *                         first_wrapped_param is not sig_params[0])
 */
        /*else*/ {
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_parameters); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2386, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2386, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2386, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2386, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_v_sig_params = ((PyObject*)__pyx_t_6);
          __pyx_t_6 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2387
 *             else:
 *                 sig_params = tuple(sig.parameters.values())
 *                 assert (not sig_params or             # <<<<<<<<<<<<<<
 *                         first_wrapped_param is not sig_params[0])
 *                 new_params = (first_wrapped_param,) + sig_params
 */
          #ifndef CYTHON_WITHOUT_ASSERTIONS
          if (unlikely(!Py_OptimizeFlag)) {
            __pyx_t_1 = (PyTuple_GET_SIZE(__pyx_v_sig_params) != 0);
            __pyx_t_11 = ((!__pyx_t_1) != 0);
            if (!__pyx_t_11) {
            } else {
              __pyx_t_2 = __pyx_t_11;
              goto __pyx_L24_bool_binop_done;
            }

            /* "Aspidites/_vendor/contracts/inspection.py":2388
 *                 sig_params = tuple(sig.parameters.values())
 *                 assert (not sig_params or
 *                         first_wrapped_param is not sig_params[0])             # <<<<<<<<<<<<<<
 *                 new_params = (first_wrapped_param,) + sig_params
 *                 return sig.replace(parameters=new_params)
 */
            __pyx_t_11 = (__pyx_v_first_wrapped_param != PyTuple_GET_ITEM(__pyx_v_sig_params, 0));
            __pyx_t_1 = (__pyx_t_11 != 0);
            __pyx_t_2 = __pyx_t_1;
            __pyx_L24_bool_binop_done:;
            if (unlikely(!__pyx_t_2)) {
              PyErr_SetNone(PyExc_AssertionError);
              __PYX_ERR(0, 2387, __pyx_L18_except_error)
            }
          }
          #endif

          /* "Aspidites/_vendor/contracts/inspection.py":2389
 *                 assert (not sig_params or
 *                         first_wrapped_param is not sig_params[0])
 *                 new_params = (first_wrapped_param,) + sig_params             # <<<<<<<<<<<<<<
 *                 return sig.replace(parameters=new_params)
 * 
 */
          __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2389, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_v_first_wrapped_param);
          __Pyx_GIVEREF(__pyx_v_first_wrapped_param);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_first_wrapped_param);
          __pyx_t_4 = PyNumber_Add(__pyx_t_6, __pyx_v_sig_params); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2389, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_v_new_params = ((PyObject*)__pyx_t_4);
          __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2390
 *                         first_wrapped_param is not sig_params[0])
 *                 new_params = (first_wrapped_param,) + sig_params
 *                 return sig.replace(parameters=new_params)             # <<<<<<<<<<<<<<
 * 
 *     if isfunction(obj) or _signature_is_functionlike(obj):
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2390, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2390, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_parameters, __pyx_v_new_params) < 0) __PYX_ERR(0, 2390, __pyx_L18_except_error)
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2390, __pyx_L18_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L19_except_return;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2365
 *         pass
 *     else:
 *         if isinstance(partialmethod, functools.partialmethod):             # <<<<<<<<<<<<<<
 *             # Unbound partialmethod (see functools.partialmethod)
 *             # This means, that we need to calculate the signature
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L21_try_end;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2362
 *         # noinspection PyProtectedMember
 *         partialmethod = obj._partialmethod
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         pass
 *     else:
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L17_exception_handled;
    }
    goto __pyx_L18_except_error;
    __pyx_L18_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":2359
 *             return sig
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # noinspection PyProtectedMember
 *         partialmethod = obj._partialmethod
 */
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L19_except_return:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
    goto __pyx_L0;
    __pyx_L17_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
    __pyx_L21_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2392
 *                 return sig.replace(parameters=new_params)
 * 
 *     if isfunction(obj) or _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *         # If it's a pure Python function, or an object that is duck type
 *         # of a Python function (Cython functions, for instance), then:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L27_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signature_is_functionlike); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L27_bool_binop_done:;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2395
 *         # If it's a pure Python function, or an object that is duck type
 *         # of a Python function (Cython functions, for instance), then:
 *         return _signature_from_function(sigcls, obj,             # <<<<<<<<<<<<<<
 *                                         skip_bound_arg=skip_bound_arg)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_signature_from_function); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_sigcls);
    __Pyx_GIVEREF(__pyx_v_sigcls);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_sigcls);
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_obj);

    /* "Aspidites/_vendor/contracts/inspection.py":2396
 *         # of a Python function (Cython functions, for instance), then:
 *         return _signature_from_function(sigcls, obj,
 *                                         skip_bound_arg=skip_bound_arg)             # <<<<<<<<<<<<<<
 * 
 *     if _signature_is_builtin(obj):
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2396, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2395
 *         # If it's a pure Python function, or an object that is duck type
 *         # of a Python function (Cython functions, for instance), then:
 *         return _signature_from_function(sigcls, obj,             # <<<<<<<<<<<<<<
 *                                         skip_bound_arg=skip_bound_arg)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2392
 *                 return sig.replace(parameters=new_params)
 * 
 *     if isfunction(obj) or _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *         # If it's a pure Python function, or an object that is duck type
 *         # of a Python function (Cython functions, for instance), then:
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2398
 *                                         skip_bound_arg=skip_bound_arg)
 * 
 *     if _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *         return _signature_from_builtin(sigcls, obj,
 *                                        skip_bound_arg=skip_bound_arg)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_signature_is_builtin); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2399
 * 
 *     if _signature_is_builtin(obj):
 *         return _signature_from_builtin(sigcls, obj,             # <<<<<<<<<<<<<<
 *                                        skip_bound_arg=skip_bound_arg)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signature_from_builtin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_sigcls);
    __Pyx_GIVEREF(__pyx_v_sigcls);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_sigcls);
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_obj);

    /* "Aspidites/_vendor/contracts/inspection.py":2400
 *     if _signature_is_builtin(obj):
 *         return _signature_from_builtin(sigcls, obj,
 *                                        skip_bound_arg=skip_bound_arg)             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(obj, functools.partial):
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2400, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2399
 * 
 *     if _signature_is_builtin(obj):
 *         return _signature_from_builtin(sigcls, obj,             # <<<<<<<<<<<<<<
 *                                        skip_bound_arg=skip_bound_arg)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2398
 *                                         skip_bound_arg=skip_bound_arg)
 * 
 *     if _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *         return _signature_from_builtin(sigcls, obj,
 *                                        skip_bound_arg=skip_bound_arg)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2402
 *                                        skip_bound_arg=skip_bound_arg)
 * 
 *     if isinstance(obj, functools.partial):             # <<<<<<<<<<<<<<
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_functools); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_partial); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_6); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2404
 *     if isinstance(obj, functools.partial):
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.func,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "Aspidites/_vendor/contracts/inspection.py":2405
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(
 *             obj.func,             # <<<<<<<<<<<<<<
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_func_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "Aspidites/_vendor/contracts/inspection.py":2404
 *     if isinstance(obj, functools.partial):
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.func,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2406
 *         wrapped_sig = _signature_from_callable(
 *             obj.func,
 *             follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *             skip_bound_arg=skip_bound_arg,
 *             sigcls=sigcls)
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2406, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2407
 *             obj.func,
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *             sigcls=sigcls)
 *         return _signature_get_partial(wrapped_sig, obj)
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2406, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2408
 *             follow_wrapper_chains=follow_wrapper_chains,
 *             skip_bound_arg=skip_bound_arg,
 *             sigcls=sigcls)             # <<<<<<<<<<<<<<
 *         return _signature_get_partial(wrapped_sig, obj)
 * 
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2406, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2404
 *     if isinstance(obj, functools.partial):
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *             obj.func,
 *             follow_wrapper_chains=follow_wrapper_chains,
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_wrapped_sig = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2409
 *             skip_bound_arg=skip_bound_arg,
 *             sigcls=sigcls)
 *         return _signature_get_partial(wrapped_sig, obj)             # <<<<<<<<<<<<<<
 * 
 *     sig = None
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_signature_get_partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_wrapped_sig, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_wrapped_sig, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_wrapped_sig);
      __Pyx_GIVEREF(__pyx_v_wrapped_sig);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_v_wrapped_sig);
      __Pyx_INCREF(__pyx_v_obj);
      __Pyx_GIVEREF(__pyx_v_obj);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_v_obj);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2402
 *                                        skip_bound_arg=skip_bound_arg)
 * 
 *     if isinstance(obj, functools.partial):             # <<<<<<<<<<<<<<
 *         # noinspection PyTypeChecker
 *         wrapped_sig = _signature_from_callable(
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2411
 *         return _signature_get_partial(wrapped_sig, obj)
 * 
 *     sig = None             # <<<<<<<<<<<<<<
 *     if isinstance(obj, type):
 *         # obj is a class or a metaclass
 */
  __Pyx_INCREF(Py_None);
  __Pyx_XDECREF_SET(__pyx_v_sig, Py_None);

  /* "Aspidites/_vendor/contracts/inspection.py":2412
 * 
 *     sig = None
 *     if isinstance(obj, type):             # <<<<<<<<<<<<<<
 *         # obj is a class or a metaclass
 * 
 */
  __pyx_t_1 = PyType_Check(__pyx_v_obj); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2417
 *         # First, let's see if it has an overloaded __call__ defined
 *         # in its metaclass
 *         call = _signature_get_user_defined_method(type(obj), '__call__')             # <<<<<<<<<<<<<<
 *         if call is not None:
 *             sig = _signature_from_callable(
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_signature_get_user_defined_meth); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, ((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_n_u_call};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, ((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_n_u_call};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_10, ((PyObject *)Py_TYPE(__pyx_v_obj)));
      __Pyx_INCREF(__pyx_n_u_call);
      __Pyx_GIVEREF(__pyx_n_u_call);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_10, __pyx_n_u_call);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_call = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2418
 *         # in its metaclass
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:             # <<<<<<<<<<<<<<
 *             sig = _signature_from_callable(
 *                 call,
 */
    __pyx_t_2 = (__pyx_v_call != Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/contracts/inspection.py":2419
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:
 *             sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 call,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2419, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "Aspidites/_vendor/contracts/inspection.py":2420
 *         if call is not None:
 *             sig = _signature_from_callable(
 *                 call,             # <<<<<<<<<<<<<<
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 */
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2419, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_call);
      __Pyx_GIVEREF(__pyx_v_call);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_call);

      /* "Aspidites/_vendor/contracts/inspection.py":2421
 *             sig = _signature_from_callable(
 *                 call,
 *                 follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2421, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2422
 *                 call,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                 sigcls=sigcls)
 *         else:
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2421, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2423
 *                 follow_wrapper_chains=follow_wrapper_chains,
 *                 skip_bound_arg=skip_bound_arg,
 *                 sigcls=sigcls)             # <<<<<<<<<<<<<<
 *         else:
 *             # Now we check if the 'obj' class has a '__new__' method
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2421, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2419
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:
 *             sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                 call,
 *                 follow_wrapper_chains=follow_wrapper_chains,
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2419, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_sig, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2418
 *         # in its metaclass
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:             # <<<<<<<<<<<<<<
 *             sig = _signature_from_callable(
 *                 call,
 */
      goto __pyx_L32;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2426
 *         else:
 *             # Now we check if the 'obj' class has a '__new__' method
 *             new = _signature_get_user_defined_method(obj, '__new__')             # <<<<<<<<<<<<<<
 *             if new is not None:
 *                 sig = _signature_from_callable(
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_signature_get_user_defined_meth); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_obj, __pyx_n_u_new};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_obj, __pyx_n_u_new};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_10, __pyx_v_obj);
        __Pyx_INCREF(__pyx_n_u_new);
        __Pyx_GIVEREF(__pyx_n_u_new);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_10, __pyx_n_u_new);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_new = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2427
 *             # Now we check if the 'obj' class has a '__new__' method
 *             new = _signature_get_user_defined_method(obj, '__new__')
 *             if new is not None:             # <<<<<<<<<<<<<<
 *                 sig = _signature_from_callable(
 *                     new,
 */
      __pyx_t_1 = (__pyx_v_new != Py_None);
      __pyx_t_2 = (__pyx_t_1 != 0);
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/contracts/inspection.py":2428
 *             new = _signature_get_user_defined_method(obj, '__new__')
 *             if new is not None:
 *                 sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                     new,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "Aspidites/_vendor/contracts/inspection.py":2429
 *             if new is not None:
 *                 sig = _signature_from_callable(
 *                     new,             # <<<<<<<<<<<<<<
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,
 */
        __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_v_new);
        __Pyx_GIVEREF(__pyx_v_new);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_new);

        /* "Aspidites/_vendor/contracts/inspection.py":2430
 *                 sig = _signature_from_callable(
 *                     new,
 *                     follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)
 */
        __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2430, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2431
 *                     new,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                     sigcls=sigcls)
 *             else:
 */
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2430, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2432
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)             # <<<<<<<<<<<<<<
 *             else:
 *                 # Finally, we should have at least __init__ implemented
 */
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2430, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2428
 *             new = _signature_get_user_defined_method(obj, '__new__')
 *             if new is not None:
 *                 sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                     new,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 */
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_sig, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2427
 *             # Now we check if the 'obj' class has a '__new__' method
 *             new = _signature_get_user_defined_method(obj, '__new__')
 *             if new is not None:             # <<<<<<<<<<<<<<
 *                 sig = _signature_from_callable(
 *                     new,
 */
        goto __pyx_L33;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2435
 *             else:
 *                 # Finally, we should have at least __init__ implemented
 *                 init = _signature_get_user_defined_method(obj, '__init__')             # <<<<<<<<<<<<<<
 *                 if init is not None:
 *                     sig = _signature_from_callable(
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signature_get_user_defined_meth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2435, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_n_u_init};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2435, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_obj, __pyx_n_u_init};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2435, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2435, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_obj);
          __Pyx_GIVEREF(__pyx_v_obj);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_v_obj);
          __Pyx_INCREF(__pyx_n_u_init);
          __Pyx_GIVEREF(__pyx_n_u_init);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_n_u_init);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2435, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_init = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2436
 *                 # Finally, we should have at least __init__ implemented
 *                 init = _signature_get_user_defined_method(obj, '__init__')
 *                 if init is not None:             # <<<<<<<<<<<<<<
 *                     sig = _signature_from_callable(
 *                         init,
 */
        __pyx_t_2 = (__pyx_v_init != Py_None);
        __pyx_t_1 = (__pyx_t_2 != 0);
        if (__pyx_t_1) {

          /* "Aspidites/_vendor/contracts/inspection.py":2437
 *                 init = _signature_get_user_defined_method(obj, '__init__')
 *                 if init is not None:
 *                     sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                         init,
 *                         follow_wrapper_chains=follow_wrapper_chains,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2437, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "Aspidites/_vendor/contracts/inspection.py":2438
 *                 if init is not None:
 *                     sig = _signature_from_callable(
 *                         init,             # <<<<<<<<<<<<<<
 *                         follow_wrapper_chains=follow_wrapper_chains,
 *                         skip_bound_arg=skip_bound_arg,
 */
          __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2437, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_v_init);
          __Pyx_GIVEREF(__pyx_v_init);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_init);

          /* "Aspidites/_vendor/contracts/inspection.py":2439
 *                     sig = _signature_from_callable(
 *                         init,
 *                         follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                         skip_bound_arg=skip_bound_arg,
 *                         sigcls=sigcls)
 */
          __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2439, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2440
 *                         init,
 *                         follow_wrapper_chains=follow_wrapper_chains,
 *                         skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                         sigcls=sigcls)
 * 
 */
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2441
 *                         follow_wrapper_chains=follow_wrapper_chains,
 *                         skip_bound_arg=skip_bound_arg,
 *                         sigcls=sigcls)             # <<<<<<<<<<<<<<
 * 
 *         if sig is None:
 */
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2437
 *                 init = _signature_get_user_defined_method(obj, '__init__')
 *                 if init is not None:
 *                     sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                         init,
 *                         follow_wrapper_chains=follow_wrapper_chains,
 */
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2437, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF_SET(__pyx_v_sig, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2436
 *                 # Finally, we should have at least __init__ implemented
 *                 init = _signature_get_user_defined_method(obj, '__init__')
 *                 if init is not None:             # <<<<<<<<<<<<<<
 *                     sig = _signature_from_callable(
 *                         init,
 */
        }
      }
      __pyx_L33:;
    }
    __pyx_L32:;

    /* "Aspidites/_vendor/contracts/inspection.py":2443
 *                         sigcls=sigcls)
 * 
 *         if sig is None:             # <<<<<<<<<<<<<<
 *             # At this point we know, that `obj` is a class, with no user-
 *             # defined '__init__', '__new__', or class-level '__call__'
 */
    __pyx_t_1 = (__pyx_v_sig == Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2447
 *             # defined '__init__', '__new__', or class-level '__call__'
 * 
 *             for base in obj.__mro__[:-1]:             # <<<<<<<<<<<<<<
 *                 # Since '__text_signature__' is implemented as a
 *                 # descriptor that extracts text signature from the
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_mro); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, -1L, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_4 = __pyx_t_6; __Pyx_INCREF(__pyx_t_4); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2447, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2447, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2447, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 2447, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2447, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_13(__pyx_t_4);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 2447, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_base, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2455
 *                 # class in there, which is 'object') to find the first
 *                 # class with non-empty text signature.
 *                 try:             # <<<<<<<<<<<<<<
 *                     text_sig = base.__text_signature__
 *                 except AttributeError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":2456
 *                 # class with non-empty text signature.
 *                 try:
 *                     text_sig = base.__text_signature__             # <<<<<<<<<<<<<<
 *                 except AttributeError:
 *                     pass
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_text_signature); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2456, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_XDECREF_SET(__pyx_v_text_sig, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":2455
 *                 # class in there, which is 'object') to find the first
 *                 # class with non-empty text signature.
 *                 try:             # <<<<<<<<<<<<<<
 *                     text_sig = base.__text_signature__
 *                 except AttributeError:
 */
          }

          /* "Aspidites/_vendor/contracts/inspection.py":2460
 *                     pass
 *                 else:
 *                     if text_sig:             # <<<<<<<<<<<<<<
 *                         # If 'obj' class has a __text_signature__ attribute:
 *                         # return a signature based on it
 */
          /*else:*/ {
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_text_sig); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2460, __pyx_L40_except_error)
            if (__pyx_t_2) {

              /* "Aspidites/_vendor/contracts/inspection.py":2463
 *                         # If 'obj' class has a __text_signature__ attribute:
 *                         # return a signature based on it
 *                         return _signature_fromstr(sigcls, obj, text_sig)             # <<<<<<<<<<<<<<
 * 
 *             # No '__text_signature__' was found for the 'obj' class.
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signature_fromstr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2463, __pyx_L40_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = NULL;
              __pyx_t_10 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_10 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_sigcls, __pyx_v_obj, __pyx_v_text_sig};
                __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2463, __pyx_L40_except_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_sigcls, __pyx_v_obj, __pyx_v_text_sig};
                __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2463, __pyx_L40_except_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              {
                __pyx_t_14 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2463, __pyx_L40_except_error)
                __Pyx_GOTREF(__pyx_t_14);
                if (__pyx_t_5) {
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_5); __pyx_t_5 = NULL;
                }
                __Pyx_INCREF(__pyx_v_sigcls);
                __Pyx_GIVEREF(__pyx_v_sigcls);
                PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, __pyx_v_sigcls);
                __Pyx_INCREF(__pyx_v_obj);
                __Pyx_GIVEREF(__pyx_v_obj);
                PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_v_obj);
                __Pyx_INCREF(__pyx_v_text_sig);
                __Pyx_GIVEREF(__pyx_v_text_sig);
                PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_10, __pyx_v_text_sig);
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2463, __pyx_L40_except_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_r = __pyx_t_6;
              __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L41_except_return;

              /* "Aspidites/_vendor/contracts/inspection.py":2460
 *                     pass
 *                 else:
 *                     if text_sig:             # <<<<<<<<<<<<<<
 *                         # If 'obj' class has a __text_signature__ attribute:
 *                         # return a signature based on it
 */
            }
          }
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L45_try_end;
          __pyx_L38_error:;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2457
 *                 try:
 *                     text_sig = base.__text_signature__
 *                 except AttributeError:             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
          __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
          if (__pyx_t_10) {
            __Pyx_ErrRestore(0,0,0);
            goto __pyx_L39_exception_handled;
          }
          goto __pyx_L40_except_error;
          __pyx_L40_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":2455
 *                 # class in there, which is 'object') to find the first
 *                 # class with non-empty text signature.
 *                 try:             # <<<<<<<<<<<<<<
 *                     text_sig = base.__text_signature__
 *                 except AttributeError:
 */
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L1_error;
          __pyx_L41_except_return:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L0;
          __pyx_L39_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          __pyx_L45_try_end:;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2447
 *             # defined '__init__', '__new__', or class-level '__call__'
 * 
 *             for base in obj.__mro__[:-1]:             # <<<<<<<<<<<<<<
 *                 # Since '__text_signature__' is implemented as a
 *                 # descriptor that extracts text signature from the
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2468
 *             # Last option is to check if its '__init__' is
 *             # object.__init__ or type.__init__.
 *             if type not in obj.__mro__:             # <<<<<<<<<<<<<<
 *                 # We have a class (not metaclass), but no user-defined
 *                 # __init__ or __new__ for it
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_mro); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(((PyObject *)(&PyType_Type)), __pyx_t_4, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2468, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "Aspidites/_vendor/contracts/inspection.py":2471
 *                 # We have a class (not metaclass), but no user-defined
 *                 # __init__ or __new__ for it
 *                 if (obj.__init__ is object.__init__ and             # <<<<<<<<<<<<<<
 *                         obj.__new__ is object.__new__):
 *                     # Return a signature of 'object' builtin.
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_init); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2471, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_object, __pyx_n_s_init); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2471, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = (__pyx_t_4 == __pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_11 = (__pyx_t_2 != 0);
        if (__pyx_t_11) {
        } else {
          __pyx_t_1 = __pyx_t_11;
          goto __pyx_L49_bool_binop_done;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2472
 *                 # __init__ or __new__ for it
 *                 if (obj.__init__ is object.__init__ and
 *                         obj.__new__ is object.__new__):             # <<<<<<<<<<<<<<
 *                     # Return a signature of 'object' builtin.
 *                     return sigcls.from_callable(object)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_new); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_object, __pyx_n_s_new); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = (__pyx_t_6 == __pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_2 = (__pyx_t_11 != 0);
        __pyx_t_1 = __pyx_t_2;
        __pyx_L49_bool_binop_done:;

        /* "Aspidites/_vendor/contracts/inspection.py":2471
 *                 # We have a class (not metaclass), but no user-defined
 *                 # __init__ or __new__ for it
 *                 if (obj.__init__ is object.__init__ and             # <<<<<<<<<<<<<<
 *                         obj.__new__ is object.__new__):
 *                     # Return a signature of 'object' builtin.
 */
        if (likely(__pyx_t_1)) {

          /* "Aspidites/_vendor/contracts/inspection.py":2474
 *                         obj.__new__ is object.__new__):
 *                     # Return a signature of 'object' builtin.
 *                     return sigcls.from_callable(object)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_sigcls, __pyx_n_s_from_callable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2474, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_builtin_object) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_builtin_object);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2474, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_r = __pyx_t_4;
          __pyx_t_4 = 0;
          goto __pyx_L0;

          /* "Aspidites/_vendor/contracts/inspection.py":2471
 *                 # We have a class (not metaclass), but no user-defined
 *                 # __init__ or __new__ for it
 *                 if (obj.__init__ is object.__init__ and             # <<<<<<<<<<<<<<
 *                         obj.__new__ is object.__new__):
 *                     # Return a signature of 'object' builtin.
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2476
 *                     return sigcls.from_callable(object)
 *                 else:
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         'no signature found for builtin type {!r}'.format(obj))
 * 
 */
        /*else*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":2477
 *                 else:
 *                     raise ValueError(
 *                         'no signature found for builtin type {!r}'.format(obj))             # <<<<<<<<<<<<<<
 * 
 *     elif not isinstance(obj, _NonUserDefinedCallables):
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_no_signature_found_for_builtin_t, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2477, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_obj);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2477, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2476
 *                     return sigcls.from_callable(object)
 *                 else:
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         'no signature found for builtin type {!r}'.format(obj))
 * 
 */
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_6, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __PYX_ERR(0, 2476, __pyx_L1_error)
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2468
 *             # Last option is to check if its '__init__' is
 *             # object.__init__ or type.__init__.
 *             if type not in obj.__mro__:             # <<<<<<<<<<<<<<
 *                 # We have a class (not metaclass), but no user-defined
 *                 # __init__ or __new__ for it
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2443
 *                         sigcls=sigcls)
 * 
 *         if sig is None:             # <<<<<<<<<<<<<<
 *             # At this point we know, that `obj` is a class, with no user-
 *             # defined '__init__', '__new__', or class-level '__call__'
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2412
 * 
 *     sig = None
 *     if isinstance(obj, type):             # <<<<<<<<<<<<<<
 *         # obj is a class or a metaclass
 * 
 */
    goto __pyx_L31;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2479
 *                         'no signature found for builtin type {!r}'.format(obj))
 * 
 *     elif not isinstance(obj, _NonUserDefinedCallables):             # <<<<<<<<<<<<<<
 *         # An object with __call__
 *         # We also check that the 'obj' is not an instance of
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_NonUserDefinedCallables); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2484
 *         # _WrapperDescriptor or _MethodWrapper to avoid
 *         # infinite recursion (and even potential segfault)
 *         call = _signature_get_user_defined_method(type(obj), '__call__')             # <<<<<<<<<<<<<<
 *         if call is not None:
 *             try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_signature_get_user_defined_meth); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_n_u_call};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2484, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)Py_TYPE(__pyx_v_obj)), __pyx_n_u_call};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2484, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, ((PyObject *)Py_TYPE(__pyx_v_obj)));
      __Pyx_INCREF(__pyx_n_u_call);
      __Pyx_GIVEREF(__pyx_n_u_call);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_n_u_call);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_call = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2485
 *         # infinite recursion (and even potential segfault)
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 sig = _signature_from_callable(
 */
    __pyx_t_2 = (__pyx_v_call != Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/contracts/inspection.py":2486
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 sig = _signature_from_callable(
 *                     call,
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        /*try:*/ {

          /* "Aspidites/_vendor/contracts/inspection.py":2487
 *         if call is not None:
 *             try:
 *                 sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                     call,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2487, __pyx_L52_error)
          __Pyx_GOTREF(__pyx_t_6);

          /* "Aspidites/_vendor/contracts/inspection.py":2488
 *             try:
 *                 sig = _signature_from_callable(
 *                     call,             # <<<<<<<<<<<<<<
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,
 */
          __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2487, __pyx_L52_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_v_call);
          __Pyx_GIVEREF(__pyx_v_call);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_call);

          /* "Aspidites/_vendor/contracts/inspection.py":2489
 *                 sig = _signature_from_callable(
 *                     call,
 *                     follow_wrapper_chains=follow_wrapper_chains,             # <<<<<<<<<<<<<<
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)
 */
          __pyx_t_14 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2489, __pyx_L52_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapper_chains) < 0) __PYX_ERR(0, 2489, __pyx_L52_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2490
 *                     call,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,             # <<<<<<<<<<<<<<
 *                     sigcls=sigcls)
 *             except ValueError as ex:
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_skip_bound_arg, __pyx_v_skip_bound_arg) < 0) __PYX_ERR(0, 2489, __pyx_L52_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2491
 *                     follow_wrapper_chains=follow_wrapper_chains,
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)             # <<<<<<<<<<<<<<
 *             except ValueError as ex:
 *                 msg = 'no signature found for {!r}'.format(obj)
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_sigcls, __pyx_v_sigcls) < 0) __PYX_ERR(0, 2489, __pyx_L52_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2487
 *         if call is not None:
 *             try:
 *                 sig = _signature_from_callable(             # <<<<<<<<<<<<<<
 *                     call,
 *                     follow_wrapper_chains=follow_wrapper_chains,
 */
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2487, __pyx_L52_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF_SET(__pyx_v_sig, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2486
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 sig = _signature_from_callable(
 *                     call,
 */
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L57_try_end;
        __pyx_L52_error:;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2492
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)
 *             except ValueError as ex:             # <<<<<<<<<<<<<<
 *                 msg = 'no signature found for {!r}'.format(obj)
 *                 raise ValueError(msg) from ex
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_14, &__pyx_t_4) < 0) __PYX_ERR(0, 2492, __pyx_L54_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_14);
          __pyx_v_ex = __pyx_t_14;
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":2493
 *                     sigcls=sigcls)
 *             except ValueError as ex:
 *                 msg = 'no signature found for {!r}'.format(obj)             # <<<<<<<<<<<<<<
 *                 raise ValueError(msg) from ex
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_no_signature_found_for_r, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2493, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_15 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_6 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_15, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_obj);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2493, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_v_msg = __pyx_t_6;
            __pyx_t_6 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":2494
 *             except ValueError as ex:
 *                 msg = 'no signature found for {!r}'.format(obj)
 *                 raise ValueError(msg) from ex             # <<<<<<<<<<<<<<
 * 
 *     if sig is not None:
 */
            __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2494, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_Raise(__pyx_t_6, 0, 0, __pyx_v_ex);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __PYX_ERR(0, 2494, __pyx_L63_error)
          }

          /* "Aspidites/_vendor/contracts/inspection.py":2492
 *                     skip_bound_arg=skip_bound_arg,
 *                     sigcls=sigcls)
 *             except ValueError as ex:             # <<<<<<<<<<<<<<
 *                 msg = 'no signature found for {!r}'.format(obj)
 *                 raise ValueError(msg) from ex
 */
          /*finally:*/ {
            __pyx_L63_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_23);
              __pyx_t_10 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex);
                __pyx_v_ex = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
              }
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
              __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
              __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
              goto __pyx_L54_except_error;
            }
          }
        }
        goto __pyx_L54_except_error;
        __pyx_L54_except_error:;

        /* "Aspidites/_vendor/contracts/inspection.py":2486
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 sig = _signature_from_callable(
 *                     call,
 */
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
        goto __pyx_L1_error;
        __pyx_L57_try_end:;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2485
 *         # infinite recursion (and even potential segfault)
 *         call = _signature_get_user_defined_method(type(obj), '__call__')
 *         if call is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 sig = _signature_from_callable(
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2479
 *                         'no signature found for builtin type {!r}'.format(obj))
 * 
 *     elif not isinstance(obj, _NonUserDefinedCallables):             # <<<<<<<<<<<<<<
 *         # An object with __call__
 *         # We also check that the 'obj' is not an instance of
 */
  }
  __pyx_L31:;

  /* "Aspidites/_vendor/contracts/inspection.py":2496
 *                 raise ValueError(msg) from ex
 * 
 *     if sig is not None:             # <<<<<<<<<<<<<<
 *         # For classes and objects we skip the first parameter of their
 *         # __call__, __new__, or __init__ methods
 */
  __pyx_t_1 = (__pyx_v_sig != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2499
 *         # For classes and objects we skip the first parameter of their
 *         # __call__, __new__, or __init__ methods
 *         if skip_bound_arg:             # <<<<<<<<<<<<<<
 *             return _signature_bound_method(sig)
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_skip_bound_arg); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2499, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2500
 *         # __call__, __new__, or __init__ methods
 *         if skip_bound_arg:
 *             return _signature_bound_method(sig)             # <<<<<<<<<<<<<<
 *         else:
 *             return sig
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_signature_bound_method); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
        }
      }
      __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_3, __pyx_v_sig) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_sig);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/contracts/inspection.py":2499
 *         # For classes and objects we skip the first parameter of their
 *         # __call__, __new__, or __init__ methods
 *         if skip_bound_arg:             # <<<<<<<<<<<<<<
 *             return _signature_bound_method(sig)
 *         else:
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2502
 *             return _signature_bound_method(sig)
 *         else:
 *             return sig             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(obj, types.BuiltinFunctionType):
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_sig);
      __pyx_r = __pyx_v_sig;
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2496
 *                 raise ValueError(msg) from ex
 * 
 *     if sig is not None:             # <<<<<<<<<<<<<<
 *         # For classes and objects we skip the first parameter of their
 *         # __call__, __new__, or __init__ methods
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2504
 *             return sig
 * 
 *     if isinstance(obj, types.BuiltinFunctionType):             # <<<<<<<<<<<<<<
 *         # Raise a nicer error message for builtins
 *         msg = 'no signature found for builtin function {!r}'.format(obj)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_BuiltinFunctionType); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_14); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_1)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2506
 *     if isinstance(obj, types.BuiltinFunctionType):
 *         # Raise a nicer error message for builtins
 *         msg = 'no signature found for builtin function {!r}'.format(obj)             # <<<<<<<<<<<<<<
 *         raise ValueError(msg)
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_no_signature_found_for_builtin_f, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_msg = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2507
 *         # Raise a nicer error message for builtins
 *         msg = 'no signature found for builtin function {!r}'.format(obj)
 *         raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *     raise ValueError('callable {!r} is not supported by signature'.format(obj))
 */
    __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_Raise(__pyx_t_14, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __PYX_ERR(0, 2507, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2504
 *             return sig
 * 
 *     if isinstance(obj, types.BuiltinFunctionType):             # <<<<<<<<<<<<<<
 *         # Raise a nicer error message for builtins
 *         msg = 'no signature found for builtin function {!r}'.format(obj)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2509
 *         raise ValueError(msg)
 * 
 *     raise ValueError('callable {!r} is not supported by signature'.format(obj))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_callable_r_is_not_supported_by_s, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_obj);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(0, 2509, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._signature_from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_partialmethod);
  __Pyx_XDECREF(__pyx_v_wrapped_sig);
  __Pyx_XDECREF(__pyx_v_first_wrapped_param);
  __Pyx_XDECREF(__pyx_v_sig_params);
  __Pyx_XDECREF(__pyx_v_new_params);
  __Pyx_XDECREF(__pyx_v_call);
  __Pyx_XDECREF(__pyx_v_new);
  __Pyx_XDECREF(__pyx_v_init);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_text_sig);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2527
 *     VAR_KEYWORD = 4
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._name_
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_1__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_1__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_1__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_1__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind___str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind___str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2528
 * 
 *     def __str__(self):
 *         return self._name_             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2527
 *     VAR_KEYWORD = 4
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._name_
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ParameterKind.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2531
 * 
 *     @property
 *     def description(self):             # <<<<<<<<<<<<<<
 *         return _PARAM_NAME_MAPPING[self]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_3description(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_3description = {"description", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_3description, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_3description(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("description (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_2description(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_2description(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("description", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2532
 *     @property
 *     def description(self):
 *         return _PARAM_NAME_MAPPING[self]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PARAM_NAME_MAPPING); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2531
 * 
 *     @property
 *     def description(self):             # <<<<<<<<<<<<<<
 *         return _PARAM_NAME_MAPPING[self]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection._ParameterKind.description", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2582
 *     empty = _empty
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):             # <<<<<<<<<<<<<<
 *         try:
 *             self._kind = _ParameterKind(kind)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_197__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_default, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_default) < 0) __PYX_ERR(0, 2582, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_annotation, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_annotation) < 0) __PYX_ERR(0, 2582, __pyx_L1_error)
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 0, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_default = 0;
  PyObject *__pyx_v_annotation = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_name_2,&__pyx_n_s_kind,&__pyx_n_s_default,&__pyx_n_s_annotation,0};
    PyObject* values[5] = {0,0,0,0,0};
    __pyx_defaults3 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_default;
    values[4] = __pyx_dynamic_args->__pyx_arg_annotation;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 2582, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kind)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 2582, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 3; index < 5 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2582, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_kind = values[2];
    __pyx_v_default = values[3];
    __pyx_v_annotation = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2582, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter___init__(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_kind, __pyx_v_default, __pyx_v_annotation);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_kind, PyObject *__pyx_v_default, PyObject *__pyx_v_annotation) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_name);

  /* "Aspidites/_vendor/contracts/inspection.py":2583
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):
 *         try:             # <<<<<<<<<<<<<<
 *             self._kind = _ParameterKind(kind)
 *         except ValueError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":2584
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):
 *         try:
 *             self._kind = _ParameterKind(kind)             # <<<<<<<<<<<<<<
 *         except ValueError:
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2584, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_kind) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_kind);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2584, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_kind_2, __pyx_t_4) < 0) __PYX_ERR(0, 2584, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2583
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):
 *         try:             # <<<<<<<<<<<<<<
 *             self._kind = _ParameterKind(kind)
 *         except ValueError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2585
 *         try:
 *             self._kind = _ParameterKind(kind)
 *         except ValueError:             # <<<<<<<<<<<<<<
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 *         if default is not _empty:
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 2585, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/contracts/inspection.py":2586
 *             self._kind = _ParameterKind(kind)
 *         except ValueError:
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')             # <<<<<<<<<<<<<<
 *         if default is not _empty:
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):
 */
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2586, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_value_2);
      __pyx_t_9 += 6;
      __Pyx_GIVEREF(__pyx_kp_u_value_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_u_value_2);
      __pyx_t_11 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_kind), __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2586, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_Parameter_kind);
      __pyx_t_9 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_Parameter_kind);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_kp_u_is_not_a_valid_Parameter_kind);
      __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_8, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2586, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2586, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 2586, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":2583
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):
 *         try:             # <<<<<<<<<<<<<<
 *             self._kind = _ParameterKind(kind)
 *         except ValueError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2587
 *         except ValueError:
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 *         if default is not _empty:             # <<<<<<<<<<<<<<
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):
 *                 msg = '{} parameters cannot have default values'
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = (__pyx_v_default != __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_13 = (__pyx_t_12 != 0);
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":2588
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 *         if default is not _empty:
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):             # <<<<<<<<<<<<<<
 *                 msg = '{} parameters cannot have default values'
 *                 msg = msg.format(self._kind.description)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_12) {
    } else {
      __pyx_t_13 = __pyx_t_12;
      goto __pyx_L13_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_13 = __pyx_t_12;
    __pyx_L13_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_12 = (__pyx_t_13 != 0);
    if (unlikely(__pyx_t_12)) {

      /* "Aspidites/_vendor/contracts/inspection.py":2589
 *         if default is not _empty:
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):
 *                 msg = '{} parameters cannot have default values'             # <<<<<<<<<<<<<<
 *                 msg = msg.format(self._kind.description)
 *                 raise ValueError(msg)
 */
      __Pyx_INCREF(__pyx_kp_u_parameters_cannot_have_default);
      __pyx_v_msg = __pyx_kp_u_parameters_cannot_have_default;

      /* "Aspidites/_vendor/contracts/inspection.py":2590
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):
 *                 msg = '{} parameters cannot have default values'
 *                 msg = msg.format(self._kind.description)             # <<<<<<<<<<<<<<
 *                 raise ValueError(msg)
 *         self._default = default
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_description); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyUnicode_Type_format, __pyx_v_msg, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2591
 *                 msg = '{} parameters cannot have default values'
 *                 msg = msg.format(self._kind.description)
 *                 raise ValueError(msg)             # <<<<<<<<<<<<<<
 *         self._default = default
 *         self._annotation = annotation
 */
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 2591, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2588
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 *         if default is not _empty:
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):             # <<<<<<<<<<<<<<
 *                 msg = '{} parameters cannot have default values'
 *                 msg = msg.format(self._kind.description)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2587
 *         except ValueError:
 *             raise ValueError(f'value {kind!r} is not a valid Parameter.kind')
 *         if default is not _empty:             # <<<<<<<<<<<<<<
 *             if self._kind in (_VAR_POSITIONAL, _VAR_KEYWORD):
 *                 msg = '{} parameters cannot have default values'
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2592
 *                 msg = msg.format(self._kind.description)
 *                 raise ValueError(msg)
 *         self._default = default             # <<<<<<<<<<<<<<
 *         self._annotation = annotation
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_default_2, __pyx_v_default) < 0) __PYX_ERR(0, 2592, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2593
 *                 raise ValueError(msg)
 *         self._default = default
 *         self._annotation = annotation             # <<<<<<<<<<<<<<
 * 
 *         if name is _empty:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2, __pyx_v_annotation) < 0) __PYX_ERR(0, 2593, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2595
 *         self._annotation = annotation
 * 
 *         if name is _empty:             # <<<<<<<<<<<<<<
 *             raise ValueError('name is a required attribute for Parameter')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = (__pyx_v_name == __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_13 = (__pyx_t_12 != 0);
  if (unlikely(__pyx_t_13)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2596
 * 
 *         if name is _empty:
 *             raise ValueError('name is a required attribute for Parameter')             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(name, str):
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 2596, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2595
 *         self._annotation = annotation
 * 
 *         if name is _empty:             # <<<<<<<<<<<<<<
 *             raise ValueError('name is a required attribute for Parameter')
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2598
 *             raise ValueError('name is a required attribute for Parameter')
 * 
 *         if not isinstance(name, str):             # <<<<<<<<<<<<<<
 *             msg = 'name must be a str, not a {}'.format(type(name).__name__)
 *             raise TypeError(msg)
 */
  __pyx_t_13 = PyUnicode_Check(__pyx_v_name); 
  __pyx_t_12 = ((!(__pyx_t_13 != 0)) != 0);
  if (unlikely(__pyx_t_12)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2599
 * 
 *         if not isinstance(name, str):
 *             msg = 'name must be a str, not a {}'.format(type(name).__name__)             # <<<<<<<<<<<<<<
 *             raise TypeError(msg)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_name_must_be_a_str_not_a, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_name)), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_msg = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2600
 *         if not isinstance(name, str):
 *             msg = 'name must be a str, not a {}'.format(type(name).__name__)
 *             raise TypeError(msg)             # <<<<<<<<<<<<<<
 * 
 *         if name[0] == '.' and name[1:].isdigit():
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_v_msg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 2600, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2598
 *             raise ValueError('name is a required attribute for Parameter')
 * 
 *         if not isinstance(name, str):             # <<<<<<<<<<<<<<
 *             msg = 'name must be a str, not a {}'.format(type(name).__name__)
 *             raise TypeError(msg)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2602
 *             raise TypeError(msg)
 * 
 *         if name[0] == '.' and name[1:].isdigit():             # <<<<<<<<<<<<<<
 *             # These are implicit arguments generated by comprehensions. In
 *             # order to provide a friendlier interface to users, we recast
 */
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_name, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_kp_u__10, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_13) {
  } else {
    __pyx_t_12 = __pyx_t_13;
    goto __pyx_L18_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_name, 1, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_isdigit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_12 = __pyx_t_13;
  __pyx_L18_bool_binop_done:;
  if (__pyx_t_12) {

    /* "Aspidites/_vendor/contracts/inspection.py":2607
 *             # their name as "implicitN" and treat them as positional-only.
 *             # See issue 19611.
 *             if self._kind != _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 msg = (
 *                     'implicit arguments must be passed as '
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_t_12)) {

      /* "Aspidites/_vendor/contracts/inspection.py":2609
 *             if self._kind != _POSITIONAL_OR_KEYWORD:
 *                 msg = (
 *                     'implicit arguments must be passed as '             # <<<<<<<<<<<<<<
 *                     'positional or keyword arguments, not {}'
 *                 )
 */
      __Pyx_INCREF(__pyx_kp_u_implicit_arguments_must_be_passe);
      __pyx_v_msg = __pyx_kp_u_implicit_arguments_must_be_passe;

      /* "Aspidites/_vendor/contracts/inspection.py":2612
 *                     'positional or keyword arguments, not {}'
 *                 )
 *                 msg = msg.format(self._kind.description)             # <<<<<<<<<<<<<<
 *                 raise ValueError(msg)
 *             self._kind = _POSITIONAL_ONLY
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_description); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyUnicode_Type_format, __pyx_v_msg, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2613
 *                 )
 *                 msg = msg.format(self._kind.description)
 *                 raise ValueError(msg)             # <<<<<<<<<<<<<<
 *             self._kind = _POSITIONAL_ONLY
 *             name = 'implicit{}'.format(name[1:])
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2613, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 2613, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2607
 *             # their name as "implicitN" and treat them as positional-only.
 *             # See issue 19611.
 *             if self._kind != _POSITIONAL_OR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 msg = (
 *                     'implicit arguments must be passed as '
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2614
 *                 msg = msg.format(self._kind.description)
 *                 raise ValueError(msg)
 *             self._kind = _POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 *             name = 'implicit{}'.format(name[1:])
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_kind_2, __pyx_t_5) < 0) __PYX_ERR(0, 2614, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2615
 *                 raise ValueError(msg)
 *             self._kind = _POSITIONAL_ONLY
 *             name = 'implicit{}'.format(name[1:])             # <<<<<<<<<<<<<<
 * 
 *         if not name.isidentifier():
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_implicit, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_name, 1, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2602
 *             raise TypeError(msg)
 * 
 *         if name[0] == '.' and name[1:].isdigit():             # <<<<<<<<<<<<<<
 *             # These are implicit arguments generated by comprehensions. In
 *             # order to provide a friendlier interface to users, we recast
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2617
 *             name = 'implicit{}'.format(name[1:])
 * 
 *         if not name.isidentifier():             # <<<<<<<<<<<<<<
 *             raise ValueError('{!r} is not a valid parameter name'.format(name))
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_isidentifier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_13 = ((!__pyx_t_12) != 0);
  if (unlikely(__pyx_t_13)) {

    /* "Aspidites/_vendor/contracts/inspection.py":2618
 * 
 *         if not name.isidentifier():
 *             raise ValueError('{!r} is not a valid parameter name'.format(name))             # <<<<<<<<<<<<<<
 * 
 *         self._name = name
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_is_not_a_valid_parameter_name, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2618, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":2617
 *             name = 'implicit{}'.format(name[1:])
 * 
 *         if not name.isidentifier():             # <<<<<<<<<<<<<<
 *             raise ValueError('{!r} is not a valid parameter name'.format(name))
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2620
 *             raise ValueError('{!r} is not a valid parameter name'.format(name))
 * 
 *         self._name = name             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name_4, __pyx_v_name) < 0) __PYX_ERR(0, 2620, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2582
 *     empty = _empty
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):             # <<<<<<<<<<<<<<
 *         try:
 *             self._kind = _ParameterKind(kind)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2622
 *         self._name = name
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (self._name, self._kind),
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_3__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_3__reduce__ = {"__reduce__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_3__reduce__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_3__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_2__reduce__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2623
 * 
 *     def __reduce__(self):
 *         return (type(self),             # <<<<<<<<<<<<<<
 *                 (self._name, self._kind),
 *                 {'_default': self._default,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "Aspidites/_vendor/contracts/inspection.py":2624
 *     def __reduce__(self):
 *         return (type(self),
 *                 (self._name, self._kind),             # <<<<<<<<<<<<<<
 *                 {'_default': self._default,
 *                  '_annotation': self._annotation})
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2625
 *         return (type(self),
 *                 (self._name, self._kind),
 *                 {'_default': self._default,             # <<<<<<<<<<<<<<
 *                  '_annotation': self._annotation})
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_default_2, __pyx_t_1) < 0) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2626
 *                 (self._name, self._kind),
 *                 {'_default': self._default,
 *                  '_annotation': self._annotation})             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_annotation_2, __pyx_t_1) < 0) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2623
 * 
 *     def __reduce__(self):
 *         return (type(self),             # <<<<<<<<<<<<<<
 *                 (self._name, self._kind),
 *                 {'_default': self._default,
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2622
 *         self._name = name
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (self._name, self._kind),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2628
 *                  '_annotation': self._annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._default = state['_default']
 *         self._annotation = state['_annotation']
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_5__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_5__setstate__ = {"__setstate__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_5__setstate__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_5__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, 1); __PYX_ERR(0, 2628, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate__") < 0)) __PYX_ERR(0, 2628, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2628, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_4__setstate__(__pyx_self, __pyx_v_self, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_4__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2629
 * 
 *     def __setstate__(self, state):
 *         self._default = state['_default']             # <<<<<<<<<<<<<<
 *         self._annotation = state['_annotation']
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_u_default_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_default_2, __pyx_t_1) < 0) __PYX_ERR(0, 2629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2630
 *     def __setstate__(self, state):
 *         self._default = state['_default']
 *         self._annotation = state['_annotation']             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_u_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2, __pyx_t_1) < 0) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2628
 *                  '_annotation': self._annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._default = state['_default']
 *         self._annotation = state['_annotation']
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2633
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self._name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_7name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_7name = {"name", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_7name, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_7name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_6name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_6name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("name", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2634
 *     @property
 *     def name(self):
 *         return self._name             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2633
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self._name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2637
 * 
 *     @property
 *     def default(self):             # <<<<<<<<<<<<<<
 *         return self._default
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_9default(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_9default = {"default", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_9default, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_9default(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("default (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_8default(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_8default(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("default", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2638
 *     @property
 *     def default(self):
 *         return self._default             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2637
 * 
 *     @property
 *     def default(self):             # <<<<<<<<<<<<<<
 *         return self._default
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2641
 * 
 *     @property
 *     def annotation(self):             # <<<<<<<<<<<<<<
 *         return self._annotation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_11annotation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_11annotation = {"annotation", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_11annotation, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_11annotation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("annotation (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_10annotation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_10annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("annotation", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2642
 *     @property
 *     def annotation(self):
 *         return self._annotation             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2641
 * 
 *     @property
 *     def annotation(self):             # <<<<<<<<<<<<<<
 *         return self._annotation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2645
 * 
 *     @property
 *     def kind(self):             # <<<<<<<<<<<<<<
 *         return self._kind
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_13kind(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_13kind = {"kind", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_13kind, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_13kind(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("kind (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_12kind(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_12kind(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("kind", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2646
 *     @property
 *     def kind(self):
 *         return self._kind             # <<<<<<<<<<<<<<
 * 
 *     def replace(self, *, name=_void, kind=_void,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2645
 * 
 *     @property
 *     def kind(self):             # <<<<<<<<<<<<<<
 *         return self._kind
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.kind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2648
 *         return self._kind
 * 
 *     def replace(self, *, name=_void, kind=_void,             # <<<<<<<<<<<<<<
 *                 annotation=_void, default=_void):
 *         """Creates a customized copy of the Parameter."""
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_199__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_name_2, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_name) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_kind, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_kind) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_annotation, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_annotation) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_default, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_default) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 0, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_15replace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_14replace[] = "Creates a customized copy of the Parameter.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_15replace = {"replace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_15replace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_14replace};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_15replace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_annotation = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_name_2,&__pyx_n_s_kind,&__pyx_n_s_annotation,&__pyx_n_s_default,0};
    PyObject* values[5] = {0,0,0,0,0};
    __pyx_defaults4 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_name;
    values[2] = __pyx_dynamic_args->__pyx_arg_kind;
    values[3] = __pyx_dynamic_args->__pyx_arg_annotation;
    values[4] = __pyx_dynamic_args->__pyx_arg_default;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args > 0 && likely(kw_args <= 4)) {
        Py_ssize_t index;
        for (index = 1; index < 5 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "replace") < 0)) __PYX_ERR(0, 2648, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_kind = values[2];
    __pyx_v_annotation = values[3];
    __pyx_v_default = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("replace", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2648, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_14replace(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_kind, __pyx_v_annotation, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_14replace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_kind, PyObject *__pyx_v_annotation, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 0);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_INCREF(__pyx_v_kind);
  __Pyx_INCREF(__pyx_v_annotation);
  __Pyx_INCREF(__pyx_v_default);

  /* "Aspidites/_vendor/contracts/inspection.py":2652
 *         """Creates a customized copy of the Parameter."""
 * 
 *         if name is _void:             # <<<<<<<<<<<<<<
 *             name = self._name
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_name == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2653
 * 
 *         if name is _void:
 *             name = self._name             # <<<<<<<<<<<<<<
 * 
 *         if kind is _void:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2652
 *         """Creates a customized copy of the Parameter."""
 * 
 *         if name is _void:             # <<<<<<<<<<<<<<
 *             name = self._name
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2655
 *             name = self._name
 * 
 *         if kind is _void:             # <<<<<<<<<<<<<<
 *             kind = self._kind
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_kind == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2656
 * 
 *         if kind is _void:
 *             kind = self._kind             # <<<<<<<<<<<<<<
 * 
 *         if annotation is _void:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_kind, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2655
 *             name = self._name
 * 
 *         if kind is _void:             # <<<<<<<<<<<<<<
 *             kind = self._kind
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2658
 *             kind = self._kind
 * 
 *         if annotation is _void:             # <<<<<<<<<<<<<<
 *             annotation = self._annotation
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_annotation == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2659
 * 
 *         if annotation is _void:
 *             annotation = self._annotation             # <<<<<<<<<<<<<<
 * 
 *         if default is _void:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_annotation, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2658
 *             kind = self._kind
 * 
 *         if annotation is _void:             # <<<<<<<<<<<<<<
 *             annotation = self._annotation
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2661
 *             annotation = self._annotation
 * 
 *         if default is _void:             # <<<<<<<<<<<<<<
 *             default = self._default
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_default == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2662
 * 
 *         if default is _void:
 *             default = self._default             # <<<<<<<<<<<<<<
 * 
 *         return type(self)(name, kind, default=default, annotation=annotation)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_default, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2661
 *             annotation = self._annotation
 * 
 *         if default is _void:             # <<<<<<<<<<<<<<
 *             default = self._default
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2664
 *             default = self._default
 * 
 *         return type(self)(name, kind, default=default, annotation=annotation)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
  __Pyx_INCREF(__pyx_v_kind);
  __Pyx_GIVEREF(__pyx_v_kind);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_kind);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, __pyx_v_default) < 0) __PYX_ERR(0, 2664, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_annotation, __pyx_v_annotation) < 0) __PYX_ERR(0, 2664, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)Py_TYPE(__pyx_v_self)), __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2648
 *         return self._kind
 * 
 *     def replace(self, *, name=_void, kind=_void,             # <<<<<<<<<<<<<<
 *                 annotation=_void, default=_void):
 *         """Creates a customized copy of the Parameter."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_annotation);
  __Pyx_XDECREF(__pyx_v_default);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2666
 *         return type(self)(name, kind, default=default, annotation=annotation)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         kind = self.kind
 *         formatted = self._name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_17__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_17__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_17__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_17__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_16__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_16__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_v_formatted = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2667
 * 
 *     def __str__(self):
 *         kind = self.kind             # <<<<<<<<<<<<<<
 *         formatted = self._name
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_kind = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2668
 *     def __str__(self):
 *         kind = self.kind
 *         formatted = self._name             # <<<<<<<<<<<<<<
 * 
 *         # Add annotation and default value
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_formatted = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2671
 * 
 *         # Add annotation and default value
 *         if self._annotation is not _empty:             # <<<<<<<<<<<<<<
 *             formatted = '{}: {}'.format(formatted,
 *                                         formatannotation(self._annotation))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":2672
 *         # Add annotation and default value
 *         if self._annotation is not _empty:
 *             formatted = '{}: {}'.format(formatted,             # <<<<<<<<<<<<<<
 *                                         formatannotation(self._annotation))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__73, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Aspidites/_vendor/contracts/inspection.py":2673
 *         if self._annotation is not _empty:
 *             formatted = '{}: {}'.format(formatted,
 *                                         formatannotation(self._annotation))             # <<<<<<<<<<<<<<
 * 
 *         if self._default is not _empty:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_formatannotation); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_formatted, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2672, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_formatted, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2672, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_formatted);
      __Pyx_GIVEREF(__pyx_v_formatted);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_formatted);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_formatted, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2671
 * 
 *         # Add annotation and default value
 *         if self._annotation is not _empty:             # <<<<<<<<<<<<<<
 *             formatted = '{}: {}'.format(formatted,
 *                                         formatannotation(self._annotation))
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2675
 *                                         formatannotation(self._annotation))
 * 
 *         if self._default is not _empty:             # <<<<<<<<<<<<<<
 *             if self._annotation is not _empty:
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__pyx_t_2 != __pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2676
 * 
 *         if self._default is not _empty:
 *             if self._annotation is not _empty:             # <<<<<<<<<<<<<<
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/contracts/inspection.py":2677
 *         if self._default is not _empty:
 *             if self._annotation is not _empty:
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))             # <<<<<<<<<<<<<<
 *             else:
 *                 formatted = '{}={}'.format(formatted, repr(self._default))
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__74, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = PyObject_Repr(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_formatted, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_formatted, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_formatted);
        __Pyx_GIVEREF(__pyx_v_formatted);
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_v_formatted);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_formatted, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2676
 * 
 *         if self._default is not _empty:
 *             if self._annotation is not _empty:             # <<<<<<<<<<<<<<
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))
 *             else:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2679
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))
 *             else:
 *                 formatted = '{}={}'.format(formatted, repr(self._default))             # <<<<<<<<<<<<<<
 * 
 *         if kind == _VAR_POSITIONAL:
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__75, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyObject_Repr(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_formatted, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2679, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_formatted, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2679, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_formatted);
        __Pyx_GIVEREF(__pyx_v_formatted);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_formatted);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_formatted, __pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/contracts/inspection.py":2675
 *                                         formatannotation(self._annotation))
 * 
 *         if self._default is not _empty:             # <<<<<<<<<<<<<<
 *             if self._annotation is not _empty:
 *                 formatted = '{} = {}'.format(formatted, repr(self._default))
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2681
 *                 formatted = '{}={}'.format(formatted, repr(self._default))
 * 
 *         if kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *             formatted = '*' + formatted
 *         elif kind == _VAR_KEYWORD:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_kind, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":2682
 * 
 *         if kind == _VAR_POSITIONAL:
 *             formatted = '*' + formatted             # <<<<<<<<<<<<<<
 *         elif kind == _VAR_KEYWORD:
 *             formatted = '**' + formatted
 */
    __pyx_t_1 = PyNumber_Add(__pyx_kp_u_, __pyx_v_formatted); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_formatted, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2681
 *                 formatted = '{}={}'.format(formatted, repr(self._default))
 * 
 *         if kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *             formatted = '*' + formatted
 *         elif kind == _VAR_KEYWORD:
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2683
 *         if kind == _VAR_POSITIONAL:
 *             formatted = '*' + formatted
 *         elif kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *             formatted = '**' + formatted
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_kind, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":2684
 *             formatted = '*' + formatted
 *         elif kind == _VAR_KEYWORD:
 *             formatted = '**' + formatted             # <<<<<<<<<<<<<<
 * 
 *         return formatted
 */
    __pyx_t_2 = PyNumber_Add(__pyx_kp_u__2, __pyx_v_formatted); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_formatted, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2683
 *         if kind == _VAR_POSITIONAL:
 *             formatted = '*' + formatted
 *         elif kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *             formatted = '**' + formatted
 * 
 */
  }
  __pyx_L6:;

  /* "Aspidites/_vendor/contracts/inspection.py":2686
 *             formatted = '**' + formatted
 * 
 *         return formatted             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_formatted);
  __pyx_r = __pyx_v_formatted;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2666
 *         return type(self)(name, kind, default=default, annotation=annotation)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         kind = self.kind
 *         formatted = self._name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_formatted);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2688
 *         return formatted
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_19__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_19__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_19__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_19__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_18__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_18__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2689
 * 
 *     def __repr__(self):
 *         return '<{} "{}">'.format(self.__class__.__name__, self)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__76, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2688
 *         return formatted
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2691
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_21__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_21__hash__ = {"__hash__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_21__hash__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_21__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_20__hash__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_20__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_hash_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2692
 * 
 *     def __hash__(self):
 *         return hash((self.name, self.kind, self.annotation, self.default))             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_6 = PyObject_Hash(__pyx_t_5); if (unlikely(__pyx_t_6 == ((Py_hash_t)-1))) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromHash_t(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2691
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2694
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_23__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_23__eq__ = {"__eq__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_23__eq__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_23__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, 1); __PYX_ERR(0, 2694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__eq__") < 0)) __PYX_ERR(0, 2694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_22__eq__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_22__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2695
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, Parameter):
 */
  __pyx_t_1 = (__pyx_v_self == __pyx_v_other);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2696
 *     def __eq__(self, other):
 *         if self is other:
 *             return True             # <<<<<<<<<<<<<<
 *         if not isinstance(other, Parameter):
 *             return NotImplemented
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2695
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, Parameter):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2697
 *         if self is other:
 *             return True
 *         if not isinstance(other, Parameter):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return (self._name == other._name and
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Parameter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2698
 *             return True
 *         if not isinstance(other, Parameter):
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return (self._name == other._name and
 *                 self._kind == other._kind and
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2697
 *         if self is other:
 *             return True
 *         if not isinstance(other, Parameter):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return (self._name == other._name and
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2699
 *         if not isinstance(other, Parameter):
 *             return NotImplemented
 *         return (self._name == other._name and             # <<<<<<<<<<<<<<
 *                 self._kind == other._kind and
 *                 self._default == other._default and
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_name_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2699, __pyx_L1_error)
  if (__pyx_t_1) {
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_3 = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2700
 *             return NotImplemented
 *         return (self._name == other._name and
 *                 self._kind == other._kind and             # <<<<<<<<<<<<<<
 *                 self._default == other._default and
 *                 self._annotation == other._annotation)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_kind_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2700, __pyx_L1_error)
  if (__pyx_t_1) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2701
 *         return (self._name == other._name and
 *                 self._kind == other._kind and
 *                 self._default == other._default and             # <<<<<<<<<<<<<<
 *                 self._annotation == other._annotation)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_default_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_default_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2701, __pyx_L1_error)
  if (__pyx_t_1) {
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_3 = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2702
 *                 self._kind == other._kind and
 *                 self._default == other._default and
 *                 self._annotation == other._annotation)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_annotation_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2694
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Parameter.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2724
 *     __slots__ = ('arguments', '_signature', '__weakref__')
 * 
 *     def __init__(self, signature, arguments):             # <<<<<<<<<<<<<<
 *         self.arguments = arguments
 *         self._signature = signature
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_signature = 0;
  PyObject *__pyx_v_arguments = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_signature_2,&__pyx_n_s_arguments,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signature_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 2724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arguments)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 2724, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2724, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_signature = values[1];
    __pyx_v_arguments = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2724, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments___init__(__pyx_self, __pyx_v_self, __pyx_v_signature, __pyx_v_arguments);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_arguments) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2725
 * 
 *     def __init__(self, signature, arguments):
 *         self.arguments = arguments             # <<<<<<<<<<<<<<
 *         self._signature = signature
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_arguments, __pyx_v_arguments) < 0) __PYX_ERR(0, 2725, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2726
 *     def __init__(self, signature, arguments):
 *         self.arguments = arguments
 *         self._signature = signature             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_signature_3, __pyx_v_signature) < 0) __PYX_ERR(0, 2726, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2724
 *     __slots__ = ('arguments', '_signature', '__weakref__')
 * 
 *     def __init__(self, signature, arguments):             # <<<<<<<<<<<<<<
 *         self.arguments = arguments
 *         self._signature = signature
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2729
 * 
 *     @property
 *     def signature(self):             # <<<<<<<<<<<<<<
 *         return self._signature
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_3signature(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_3signature = {"signature", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_3signature, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_3signature(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("signature (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_2signature(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_2signature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("signature", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2730
 *     @property
 *     def signature(self):
 *         return self._signature             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2729
 * 
 *     @property
 *     def signature(self):             # <<<<<<<<<<<<<<
 *         return self._signature
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2733
 * 
 *     @property
 *     def args(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_5args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_5args = {"args", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_5args, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_5args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("args (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_4args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_4args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_param_name = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("args", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2734
 *     @property
 *     def args(self):
 *         args = []             # <<<<<<<<<<<<<<
 *         for param_name, param in self._signature.parameters.items():
 *             if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2735
 *     def args(self):
 *         args = []
 *         for param_name, param in self._signature.parameters.items():             # <<<<<<<<<<<<<<
 *             if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *                 break
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_parameters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_6 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2735, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_param_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2736
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 *             if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (!__pyx_t_10) {
    } else {
      __pyx_t_8 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2736, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = (__pyx_t_8 != 0);
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/contracts/inspection.py":2737
 *         for param_name, param in self._signature.parameters.items():
 *             if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      goto __pyx_L4_break;

      /* "Aspidites/_vendor/contracts/inspection.py":2736
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 *             if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2739
 *                 break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":2740
 * 
 *             try:
 *                 arg = self.arguments[param_name]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 # We're done here. Other arguments
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2740, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_param_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2740, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2739
 *                 break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2746
 *                 break
 *             else:
 *                 if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                     # *args
 *                     args.extend(arg)
 */
      /*else:*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2746, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2746, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2746, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2746, __pyx_L10_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_10) {

          /* "Aspidites/_vendor/contracts/inspection.py":2748
 *                 if param.kind == _VAR_POSITIONAL:
 *                     # *args
 *                     args.extend(arg)             # <<<<<<<<<<<<<<
 *                 else:
 *                     # plain argument
 */
          __pyx_t_14 = __Pyx_PyList_Extend(__pyx_v_args, __pyx_v_arg); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 2748, __pyx_L10_except_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2746
 *                 break
 *             else:
 *                 if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                     # *args
 *                     args.extend(arg)
 */
          goto __pyx_L16;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2751
 *                 else:
 *                     # plain argument
 *                     args.append(arg)             # <<<<<<<<<<<<<<
 * 
 *         return tuple(args)
 */
        /*else*/ {
          __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_args, __pyx_v_arg); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 2751, __pyx_L10_except_error)
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L15_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2741
 *             try:
 *                 arg = self.arguments[param_name]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 # We're done here. Other arguments
 *                 # will be mapped in 'BoundArguments.kwargs'
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.args", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 2741, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_5);

        /* "Aspidites/_vendor/contracts/inspection.py":2744
 *                 # We're done here. Other arguments
 *                 # will be mapped in 'BoundArguments.kwargs'
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 if param.kind == _VAR_POSITIONAL:
 */
        goto __pyx_L17_except_break;
        __pyx_L17_except_break:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L13_try_break;
      }
      goto __pyx_L10_except_error;
      __pyx_L10_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":2739
 *                 break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L13_try_break:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L4_break;
      __pyx_L15_try_end:;
    }
  }
  __pyx_L4_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2753
 *                     args.append(arg)
 * 
 *         return tuple(args)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_AsTuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2733
 * 
 *     @property
 *     def args(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_param_name);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2756
 * 
 *     @property
 *     def kwargs(self):             # <<<<<<<<<<<<<<
 *         kwargs = {}
 *         kwargs_started = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_7kwargs(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_7kwargs = {"kwargs", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_7kwargs, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_7kwargs(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("kwargs (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_6kwargs(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_6kwargs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_kwargs = NULL;
  int __pyx_v_kwargs_started;
  PyObject *__pyx_v_param_name = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("kwargs", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2757
 *     @property
 *     def kwargs(self):
 *         kwargs = {}             # <<<<<<<<<<<<<<
 *         kwargs_started = False
 *         for param_name, param in self._signature.parameters.items():
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_kwargs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2758
 *     def kwargs(self):
 *         kwargs = {}
 *         kwargs_started = False             # <<<<<<<<<<<<<<
 *         for param_name, param in self._signature.parameters.items():
 *             if not kwargs_started:
 */
  __pyx_v_kwargs_started = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2759
 *         kwargs = {}
 *         kwargs_started = False
 *         for param_name, param in self._signature.parameters.items():             # <<<<<<<<<<<<<<
 *             if not kwargs_started:
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_parameters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_6 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2759, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_param_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2760
 *         kwargs_started = False
 *         for param_name, param in self._signature.parameters.items():
 *             if not kwargs_started:             # <<<<<<<<<<<<<<
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *                     kwargs_started = True
 */
    __pyx_t_8 = ((!(__pyx_v_kwargs_started != 0)) != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":2761
 *         for param_name, param in self._signature.parameters.items():
 *             if not kwargs_started:
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                     kwargs_started = True
 *                 else:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (!__pyx_t_10) {
      } else {
        __pyx_t_8 = __pyx_t_10;
        goto __pyx_L7_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2761, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __pyx_t_10;
      __pyx_L7_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = (__pyx_t_8 != 0);
      if (__pyx_t_10) {

        /* "Aspidites/_vendor/contracts/inspection.py":2762
 *             if not kwargs_started:
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *                     kwargs_started = True             # <<<<<<<<<<<<<<
 *                 else:
 *                     if param_name not in self.arguments:
 */
        __pyx_v_kwargs_started = 1;

        /* "Aspidites/_vendor/contracts/inspection.py":2761
 *         for param_name, param in self._signature.parameters.items():
 *             if not kwargs_started:
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                     kwargs_started = True
 *                 else:
 */
        goto __pyx_L6;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2764
 *                     kwargs_started = True
 *                 else:
 *                     if param_name not in self.arguments:             # <<<<<<<<<<<<<<
 *                         kwargs_started = True
 *                         continue
 */
      /*else*/ {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2764, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_v_param_name, __pyx_t_6, Py_NE)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2764, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_8 = (__pyx_t_10 != 0);
        if (__pyx_t_8) {

          /* "Aspidites/_vendor/contracts/inspection.py":2765
 *                 else:
 *                     if param_name not in self.arguments:
 *                         kwargs_started = True             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
          __pyx_v_kwargs_started = 1;

          /* "Aspidites/_vendor/contracts/inspection.py":2766
 *                     if param_name not in self.arguments:
 *                         kwargs_started = True
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *             if not kwargs_started:
 */
          goto __pyx_L3_continue;

          /* "Aspidites/_vendor/contracts/inspection.py":2764
 *                     kwargs_started = True
 *                 else:
 *                     if param_name not in self.arguments:             # <<<<<<<<<<<<<<
 *                         kwargs_started = True
 *                         continue
 */
        }
      }
      __pyx_L6:;

      /* "Aspidites/_vendor/contracts/inspection.py":2760
 *         kwargs_started = False
 *         for param_name, param in self._signature.parameters.items():
 *             if not kwargs_started:             # <<<<<<<<<<<<<<
 *                 if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *                     kwargs_started = True
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2768
 *                         continue
 * 
 *             if not kwargs_started:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_8 = ((!(__pyx_v_kwargs_started != 0)) != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":2769
 * 
 *             if not kwargs_started:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      goto __pyx_L3_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":2768
 *                         continue
 * 
 *             if not kwargs_started:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2771
 *                 continue
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":2772
 * 
 *             try:
 *                 arg = self.arguments[param_name]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 pass
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2772, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_param_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2772, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2771
 *                 continue
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":2776
 *                 pass
 *             else:
 *                 if param.kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                     # **kwargs
 *                     kwargs.update(arg)
 */
      /*else:*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2776, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2776, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2776, __pyx_L13_except_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 2776, __pyx_L13_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_8) {

          /* "Aspidites/_vendor/contracts/inspection.py":2778
 *                 if param.kind == _VAR_KEYWORD:
 *                     # **kwargs
 *                     kwargs.update(arg)             # <<<<<<<<<<<<<<
 *                 else:
 *                     # plain keyword argument
 */
          __pyx_t_9 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_kwargs, __pyx_v_arg); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2778, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2776
 *                 pass
 *             else:
 *                 if param.kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                     # **kwargs
 *                     kwargs.update(arg)
 */
          goto __pyx_L19;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2781
 *                 else:
 *                     # plain keyword argument
 *                     kwargs[param_name] = arg             # <<<<<<<<<<<<<<
 * 
 *         return kwargs
 */
        /*else*/ {
          if (unlikely(PyDict_SetItem(__pyx_v_kwargs, __pyx_v_param_name, __pyx_v_arg) < 0)) __PYX_ERR(0, 2781, __pyx_L13_except_error)
        }
        __pyx_L19:;
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L18_try_end;
      __pyx_L11_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2773
 *             try:
 *                 arg = self.arguments[param_name]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_7) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L12_exception_handled;
      }
      goto __pyx_L13_except_error;
      __pyx_L13_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":2771
 *                 continue
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 arg = self.arguments[param_name]
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L12_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __pyx_L18_try_end:;
    }
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2783
 *                     kwargs[param_name] = arg
 * 
 *         return kwargs             # <<<<<<<<<<<<<<
 * 
 *     def apply_defaults(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_r = __pyx_v_kwargs;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2756
 * 
 *     @property
 *     def kwargs(self):             # <<<<<<<<<<<<<<
 *         kwargs = {}
 *         kwargs_started = False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.kwargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XDECREF(__pyx_v_param_name);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2785
 *         return kwargs
 * 
 *     def apply_defaults(self):             # <<<<<<<<<<<<<<
 *         """Set default values for missing arguments.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_9apply_defaults(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_8apply_defaults[] = "Set default values for missing arguments.\n\n        For variable-positional arguments (*args) the default is an\n        empty tuple.\n\n        For variable-keyword arguments (**kwargs) the default is an\n        empty dict.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_9apply_defaults = {"apply_defaults", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_9apply_defaults, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_8apply_defaults};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_9apply_defaults(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_defaults (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_8apply_defaults(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_8apply_defaults(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_arguments = NULL;
  PyObject *__pyx_v_new_arguments = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_defaults", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2794
 *         empty dict.
 *         """
 *         arguments = self.arguments             # <<<<<<<<<<<<<<
 *         new_arguments = []
 *         for name, param in self._signature.parameters.items():
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_arguments = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2795
 *         """
 *         arguments = self.arguments
 *         new_arguments = []             # <<<<<<<<<<<<<<
 *         for name, param in self._signature.parameters.items():
 *             try:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_arguments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2796
 *         arguments = self.arguments
 *         new_arguments = []
 *         for name, param in self._signature.parameters.items():             # <<<<<<<<<<<<<<
 *             try:
 *                 new_arguments.append((name, arguments[name]))
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_parameters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_6 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2796, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2797
 *         new_arguments = []
 *         for name, param in self._signature.parameters.items():
 *             try:             # <<<<<<<<<<<<<<
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":2798
 *         for name, param in self._signature.parameters.items():
 *             try:
 *                 new_arguments.append((name, arguments[name]))             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 if param.default is not _empty:
 */
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_arguments, __pyx_v_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2798, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2798, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_new_arguments, __pyx_t_5); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2798, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2797
 *         new_arguments = []
 *         for name, param in self._signature.parameters.items():
 *             try:             # <<<<<<<<<<<<<<
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2799
 *             try:
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 if param.default is not _empty:
 *                     val = param.default
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.apply_defaults", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_12) < 0) __PYX_ERR(0, 2799, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_12);

        /* "Aspidites/_vendor/contracts/inspection.py":2800
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:
 *                 if param.default is not _empty:             # <<<<<<<<<<<<<<
 *                     val = param.default
 *                 elif param.kind is _VAR_POSITIONAL:
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_default); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2800, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_empty); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2800, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = (__pyx_t_13 != __pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_16 = (__pyx_t_15 != 0);
        if (__pyx_t_16) {

          /* "Aspidites/_vendor/contracts/inspection.py":2801
 *             except KeyError:
 *                 if param.default is not _empty:
 *                     val = param.default             # <<<<<<<<<<<<<<
 *                 elif param.kind is _VAR_POSITIONAL:
 *                     val = ()
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_default); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2801, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2800
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:
 *                 if param.default is not _empty:             # <<<<<<<<<<<<<<
 *                     val = param.default
 *                 elif param.kind is _VAR_POSITIONAL:
 */
          goto __pyx_L15;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2802
 *                 if param.default is not _empty:
 *                     val = param.default
 *                 elif param.kind is _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                     val = ()
 *                 elif param.kind is _VAR_KEYWORD:
 */
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2802, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2802, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_16 = (__pyx_t_14 == __pyx_t_13);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_15 = (__pyx_t_16 != 0);
        if (__pyx_t_15) {

          /* "Aspidites/_vendor/contracts/inspection.py":2803
 *                     val = param.default
 *                 elif param.kind is _VAR_POSITIONAL:
 *                     val = ()             # <<<<<<<<<<<<<<
 *                 elif param.kind is _VAR_KEYWORD:
 *                     val = {}
 */
          __Pyx_INCREF(__pyx_empty_tuple);
          __Pyx_XDECREF_SET(__pyx_v_val, __pyx_empty_tuple);

          /* "Aspidites/_vendor/contracts/inspection.py":2802
 *                 if param.default is not _empty:
 *                     val = param.default
 *                 elif param.kind is _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                     val = ()
 *                 elif param.kind is _VAR_KEYWORD:
 */
          goto __pyx_L15;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2804
 *                 elif param.kind is _VAR_POSITIONAL:
 *                     val = ()
 *                 elif param.kind is _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                     val = {}
 *                 else:
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2804, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2804, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = (__pyx_t_13 == __pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_16 = (__pyx_t_15 != 0);
        if (__pyx_t_16) {

          /* "Aspidites/_vendor/contracts/inspection.py":2805
 *                     val = ()
 *                 elif param.kind is _VAR_KEYWORD:
 *                     val = {}             # <<<<<<<<<<<<<<
 *                 else:
 *                     # This BoundArguments was likely produced by
 */
          __pyx_t_14 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2805, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2804
 *                 elif param.kind is _VAR_POSITIONAL:
 *                     val = ()
 *                 elif param.kind is _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                     val = {}
 *                 else:
 */
          goto __pyx_L15;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2809
 *                     # This BoundArguments was likely produced by
 *                     # Signature.bind_partial().
 *                     continue             # <<<<<<<<<<<<<<
 *                 new_arguments.append((name, val))
 *         self.arguments = dict(new_arguments)
 */
        /*else*/ {
          goto __pyx_L14_except_continue;
        }
        __pyx_L15:;

        /* "Aspidites/_vendor/contracts/inspection.py":2810
 *                     # Signature.bind_partial().
 *                     continue
 *                 new_arguments.append((name, val))             # <<<<<<<<<<<<<<
 *         self.arguments = dict(new_arguments)
 * 
 */
        __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2810, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_name);
        __Pyx_INCREF(__pyx_v_val);
        __Pyx_GIVEREF(__pyx_v_val);
        PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_val);
        __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_new_arguments, __pyx_t_14); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2810, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L6_exception_handled;
        __pyx_L14_except_continue:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L11_try_continue;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":2797
 *         new_arguments = []
 *         for name, param in self._signature.parameters.items():
 *             try:             # <<<<<<<<<<<<<<
 *                 new_arguments.append((name, arguments[name]))
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L11_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L3_continue;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L12_try_end:;
    }
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2811
 *                     continue
 *                 new_arguments.append((name, val))
 *         self.arguments = dict(new_arguments)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_new_arguments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_arguments, __pyx_t_1) < 0) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2785
 *         return kwargs
 * 
 *     def apply_defaults(self):             # <<<<<<<<<<<<<<
 *         """Set default values for missing arguments.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.apply_defaults", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arguments);
  __Pyx_XDECREF(__pyx_v_new_arguments);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2813
 *         self.arguments = dict(new_arguments)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_11__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_11__eq__ = {"__eq__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_11__eq__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_11__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, 1); __PYX_ERR(0, 2813, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__eq__") < 0)) __PYX_ERR(0, 2813, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2813, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_10__eq__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_10__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2814
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, BoundArguments):
 */
  __pyx_t_1 = (__pyx_v_self == __pyx_v_other);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2815
 *     def __eq__(self, other):
 *         if self is other:
 *             return True             # <<<<<<<<<<<<<<
 *         if not isinstance(other, BoundArguments):
 *             return NotImplemented
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2814
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, BoundArguments):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2816
 *         if self is other:
 *             return True
 *         if not isinstance(other, BoundArguments):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return (self.signature == other.signature and
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_BoundArguments); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2817
 *             return True
 *         if not isinstance(other, BoundArguments):
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return (self.signature == other.signature and
 *                 self.arguments == other.arguments)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2816
 *         if self is other:
 *             return True
 *         if not isinstance(other, BoundArguments):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return (self.signature == other.signature and
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2818
 *         if not isinstance(other, BoundArguments):
 *             return NotImplemented
 *         return (self.signature == other.signature and             # <<<<<<<<<<<<<<
 *                 self.arguments == other.arguments)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_signature_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2818, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2818, __pyx_L1_error)
  if (__pyx_t_1) {
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_3 = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2819
 *             return NotImplemented
 *         return (self.signature == other.signature and
 *                 self.arguments == other.arguments)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_arguments); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2813
 *         self.arguments = dict(new_arguments)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2821
 *                 self.arguments == other.arguments)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._signature = state['_signature']
 *         self.arguments = state['arguments']
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_13__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_13__setstate__ = {"__setstate__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_13__setstate__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_13__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, 1); __PYX_ERR(0, 2821, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate__") < 0)) __PYX_ERR(0, 2821, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2821, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_12__setstate__(__pyx_self, __pyx_v_self, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_12__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2822
 * 
 *     def __setstate__(self, state):
 *         self._signature = state['_signature']             # <<<<<<<<<<<<<<
 *         self.arguments = state['arguments']
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_u_signature_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_signature_3, __pyx_t_1) < 0) __PYX_ERR(0, 2822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2823
 *     def __setstate__(self, state):
 *         self._signature = state['_signature']
 *         self.arguments = state['arguments']             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_u_arguments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_arguments, __pyx_t_1) < 0) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2821
 *                 self.arguments == other.arguments)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._signature = state['_signature']
 *         self.arguments = state['arguments']
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2825
 *         self.arguments = state['arguments']
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_15__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_15__getstate__ = {"__getstate__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_15__getstate__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_15__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_14__getstate__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_14__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2826
 * 
 *     def __getstate__(self):
 *         return {'_signature': self._signature, 'arguments': self.arguments}             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_signature_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_signature_3, __pyx_t_2) < 0) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_arguments, __pyx_t_2) < 0) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2825
 *         self.arguments = state['arguments']
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2828
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for arg, value in self.arguments.items():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_17__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_17__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_17__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_17__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_16__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_16__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2829
 * 
 *     def __repr__(self):
 *         args = []             # <<<<<<<<<<<<<<
 *         for arg, value in self.arguments.items():
 *             args.append('{}={!r}'.format(arg, value))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2830
 *     def __repr__(self):
 *         args = []
 *         for arg, value in self.arguments.items():             # <<<<<<<<<<<<<<
 *             args.append('{}={!r}'.format(arg, value))
 *         return '<{} ({})>'.format(self.__class__.__name__, ', '.join(args))
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_arguments); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2830, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, &__pyx_t_5, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2831
 *         args = []
 *         for arg, value in self.arguments.items():
 *             args.append('{}={!r}'.format(arg, value))             # <<<<<<<<<<<<<<
 *         return '<{} ({})>'.format(self.__class__.__name__, ', '.join(args))
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_arg, __pyx_v_value};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2831, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_arg, __pyx_v_value};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2831, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_arg);
      __Pyx_GIVEREF(__pyx_v_arg);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_v_arg);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_args, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 2831, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2832
 *         for arg, value in self.arguments.items():
 *             args.append('{}={!r}'.format(arg, value))
 *         return '<{} ({})>'.format(self.__class__.__name__, ', '.join(args))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__77, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__41, __pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_9, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_9, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_4, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_4, __pyx_t_6);
    __pyx_t_9 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2828
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for arg, value in self.arguments.items():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.BoundArguments.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_201__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_annotation, __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_return_annotation) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2866
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,
 *                  __validate_parameters__=True):             # <<<<<<<<<<<<<<
 *         """Constructs Signature from the given list of Parameter
 *         objects and 'return_annotation'.  All arguments are optional.
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_validate_parameters, ((PyObject *)Py_True)) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature___init__[] = "Constructs Signature from the given list of Parameter\n        objects and 'return_annotation'.  All arguments are optional.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature___init__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parameters = 0;
  PyObject *__pyx_v_return_annotation = 0;
  PyObject *__pyx_v___validate_parameters__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_parameters,&__pyx_n_s_return_annotation,&__pyx_n_s_validate_parameters,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults5 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self);
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = __pyx_dynamic_args->__pyx_arg_return_annotation;
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parameters);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 2; index < 4 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2865, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parameters = values[1];
    __pyx_v_return_annotation = values[2];
    __pyx_v___validate_parameters__ = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2865, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature___init__(__pyx_self, __pyx_v_self, __pyx_v_parameters, __pyx_v_return_annotation, __pyx_v___validate_parameters__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/contracts/inspection.py":2914
 *                     params[name] = param
 *             else:
 *                 params = OrderedDict((param.name, param) for param in parameters)             # <<<<<<<<<<<<<<
 * 
 *         self._parameters = types.MappingProxyType(params)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2914, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___2generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Signature___init___locals_genexp, __pyx_n_s_Aspidites__vendor_contracts_insp); if (unlikely(!gen)) __PYX_ERR(0, 2914, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2914, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parameters)) { __Pyx_RaiseClosureNameError("parameters"); __PYX_ERR(0, 2914, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parameters)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parameters)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_parameters; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parameters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2914, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2914, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2914, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2914, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2914, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2914, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_param);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_param, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_param);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_param);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_cur_scope->__pyx_v_param);
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2914, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parameters, PyObject *__pyx_v_return_annotation, PyObject *__pyx_v___validate_parameters__) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_params = NULL;
  PyObject *__pyx_v_top_kind = NULL;
  int __pyx_v_kind_defaults;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___2generator3 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2865, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_parameters = __pyx_v_parameters;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_parameters);

  /* "Aspidites/_vendor/contracts/inspection.py":2871
 *         """
 * 
 *         if parameters is None:             # <<<<<<<<<<<<<<
 *             params = OrderedDict()
 *         else:
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_parameters == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2872
 * 
 *         if parameters is None:
 *             params = OrderedDict()             # <<<<<<<<<<<<<<
 *         else:
 *             if __validate_parameters__:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_params = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2871
 *         """
 * 
 *         if parameters is None:             # <<<<<<<<<<<<<<
 *             params = OrderedDict()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2874
 *             params = OrderedDict()
 *         else:
 *             if __validate_parameters__:             # <<<<<<<<<<<<<<
 *                 params = OrderedDict()
 *                 top_kind = _POSITIONAL_ONLY
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v___validate_parameters__); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2874, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/contracts/inspection.py":2875
 *         else:
 *             if __validate_parameters__:
 *                 params = OrderedDict()             # <<<<<<<<<<<<<<
 *                 top_kind = _POSITIONAL_ONLY
 *                 kind_defaults = False
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_params = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2876
 *             if __validate_parameters__:
 *                 params = OrderedDict()
 *                 top_kind = _POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 *                 kind_defaults = False
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_top_kind = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2877
 *                 params = OrderedDict()
 *                 top_kind = _POSITIONAL_ONLY
 *                 kind_defaults = False             # <<<<<<<<<<<<<<
 * 
 *                 for param in parameters:
 */
      __pyx_v_kind_defaults = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2879
 *                 kind_defaults = False
 * 
 *                 for param in parameters:             # <<<<<<<<<<<<<<
 *                     kind = param.kind
 *                     name = param.name
 */
      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_parameters)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_parameters)) {
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_parameters; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_parameters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2879, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2879, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2879, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2879, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2879, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_7(__pyx_t_3);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 2879, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2880
 * 
 *                 for param in parameters:
 *                     kind = param.kind             # <<<<<<<<<<<<<<
 *                     name = param.name
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2881
 *                 for param in parameters:
 *                     kind = param.kind
 *                     name = param.name             # <<<<<<<<<<<<<<
 * 
 *                     if kind < top_kind:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2881, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2883
 *                     name = param.name
 * 
 *                     if kind < top_kind:             # <<<<<<<<<<<<<<
 *                         msg = (
 *                             'wrong parameter order: {} parameter before {} '
 */
        __pyx_t_4 = PyObject_RichCompare(__pyx_v_kind, __pyx_v_top_kind, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2883, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2883, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(__pyx_t_2)) {

          /* "Aspidites/_vendor/contracts/inspection.py":2885
 *                     if kind < top_kind:
 *                         msg = (
 *                             'wrong parameter order: {} parameter before {} '             # <<<<<<<<<<<<<<
 *                             'parameter'
 *                         )
 */
          __Pyx_INCREF(__pyx_kp_u_wrong_parameter_order_parameter);
          __pyx_v_msg = __pyx_kp_u_wrong_parameter_order_parameter;

          /* "Aspidites/_vendor/contracts/inspection.py":2888
 *                             'parameter'
 *                         )
 *                         msg = msg.format(top_kind.description,             # <<<<<<<<<<<<<<
 *                                          kind.description)
 *                         raise ValueError(msg)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_top_kind, __pyx_n_s_description); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2888, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);

          /* "Aspidites/_vendor/contracts/inspection.py":2889
 *                         )
 *                         msg = msg.format(top_kind.description,
 *                                          kind.description)             # <<<<<<<<<<<<<<
 *                         raise ValueError(msg)
 *                     elif kind > top_kind:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_kind, __pyx_n_s_description); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2889, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "Aspidites/_vendor/contracts/inspection.py":2888
 *                             'parameter'
 *                         )
 *                         msg = msg.format(top_kind.description,             # <<<<<<<<<<<<<<
 *                                          kind.description)
 *                         raise ValueError(msg)
 */
          __pyx_t_8 = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyUnicode_Type_format, __pyx_v_msg, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2888, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2890
 *                         msg = msg.format(top_kind.description,
 *                                          kind.description)
 *                         raise ValueError(msg)             # <<<<<<<<<<<<<<
 *                     elif kind > top_kind:
 *                         kind_defaults = False
 */
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2890, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_Raise(__pyx_t_8, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 2890, __pyx_L1_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2883
 *                     name = param.name
 * 
 *                     if kind < top_kind:             # <<<<<<<<<<<<<<
 *                         msg = (
 *                             'wrong parameter order: {} parameter before {} '
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2891
 *                                          kind.description)
 *                         raise ValueError(msg)
 *                     elif kind > top_kind:             # <<<<<<<<<<<<<<
 *                         kind_defaults = False
 *                         top_kind = kind
 */
        __pyx_t_8 = PyObject_RichCompare(__pyx_v_kind, __pyx_v_top_kind, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2891, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2891, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/contracts/inspection.py":2892
 *                         raise ValueError(msg)
 *                     elif kind > top_kind:
 *                         kind_defaults = False             # <<<<<<<<<<<<<<
 *                         top_kind = kind
 * 
 */
          __pyx_v_kind_defaults = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2893
 *                     elif kind > top_kind:
 *                         kind_defaults = False
 *                         top_kind = kind             # <<<<<<<<<<<<<<
 * 
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):
 */
          __Pyx_INCREF(__pyx_v_kind);
          __Pyx_DECREF_SET(__pyx_v_top_kind, __pyx_v_kind);

          /* "Aspidites/_vendor/contracts/inspection.py":2891
 *                                          kind.description)
 *                         raise ValueError(msg)
 *                     elif kind > top_kind:             # <<<<<<<<<<<<<<
 *                         kind_defaults = False
 *                         top_kind = kind
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2895
 *                         top_kind = kind
 * 
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):             # <<<<<<<<<<<<<<
 *                         if param.default is _empty:
 *                             if kind_defaults:
 */
        __Pyx_INCREF(__pyx_v_kind);
        __pyx_t_8 = __pyx_v_kind;
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_8, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_1) {
        } else {
          __pyx_t_2 = __pyx_t_1;
          goto __pyx_L9_bool_binop_done;
        }
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyObject_RichCompare(__pyx_t_8, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_2 = __pyx_t_1;
        __pyx_L9_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_1 = (__pyx_t_2 != 0);
        if (__pyx_t_1) {

          /* "Aspidites/_vendor/contracts/inspection.py":2896
 * 
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):
 *                         if param.default is _empty:             # <<<<<<<<<<<<<<
 *                             if kind_defaults:
 *                                 # No default for this parameter, but the
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_default); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2896, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2896, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_1 = (__pyx_t_8 == __pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_2 = (__pyx_t_1 != 0);
          if (__pyx_t_2) {

            /* "Aspidites/_vendor/contracts/inspection.py":2897
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):
 *                         if param.default is _empty:
 *                             if kind_defaults:             # <<<<<<<<<<<<<<
 *                                 # No default for this parameter, but the
 *                                 # previous parameter of the same kind had
 */
            __pyx_t_2 = (__pyx_v_kind_defaults != 0);
            if (unlikely(__pyx_t_2)) {

              /* "Aspidites/_vendor/contracts/inspection.py":2901
 *                                 # previous parameter of the same kind had
 *                                 # a default
 *                                 msg = 'non-default argument follows default ' \             # <<<<<<<<<<<<<<
 *                                       'argument'
 *                                 raise ValueError(msg)
 */
              __Pyx_INCREF(__pyx_kp_u_non_default_argument_follows_def);
              __pyx_v_msg = __pyx_kp_u_non_default_argument_follows_def;

              /* "Aspidites/_vendor/contracts/inspection.py":2903
 *                                 msg = 'non-default argument follows default ' \
 *                                       'argument'
 *                                 raise ValueError(msg)             # <<<<<<<<<<<<<<
 *                         else:
 *                             # There is a default for this parameter.
 */
              __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2903, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_Raise(__pyx_t_5, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __PYX_ERR(0, 2903, __pyx_L1_error)

              /* "Aspidites/_vendor/contracts/inspection.py":2897
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):
 *                         if param.default is _empty:
 *                             if kind_defaults:             # <<<<<<<<<<<<<<
 *                                 # No default for this parameter, but the
 *                                 # previous parameter of the same kind had
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":2896
 * 
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):
 *                         if param.default is _empty:             # <<<<<<<<<<<<<<
 *                             if kind_defaults:
 *                                 # No default for this parameter, but the
 */
            goto __pyx_L11;
          }

          /* "Aspidites/_vendor/contracts/inspection.py":2906
 *                         else:
 *                             # There is a default for this parameter.
 *                             kind_defaults = True             # <<<<<<<<<<<<<<
 * 
 *                     if name in params:
 */
          /*else*/ {
            __pyx_v_kind_defaults = 1;
          }
          __pyx_L11:;

          /* "Aspidites/_vendor/contracts/inspection.py":2895
 *                         top_kind = kind
 * 
 *                     if kind in (_POSITIONAL_ONLY, _POSITIONAL_OR_KEYWORD):             # <<<<<<<<<<<<<<
 *                         if param.default is _empty:
 *                             if kind_defaults:
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2908
 *                             kind_defaults = True
 * 
 *                     if name in params:             # <<<<<<<<<<<<<<
 *                         msg = 'duplicate parameter name: {!r}'.format(name)
 *                         raise ValueError(msg)
 */
        __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_params, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2908, __pyx_L1_error)
        __pyx_t_1 = (__pyx_t_2 != 0);
        if (unlikely(__pyx_t_1)) {

          /* "Aspidites/_vendor/contracts/inspection.py":2909
 * 
 *                     if name in params:
 *                         msg = 'duplicate parameter name: {!r}'.format(name)             # <<<<<<<<<<<<<<
 *                         raise ValueError(msg)
 * 
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_duplicate_parameter_name_r, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2909, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_name);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2909, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_v_msg = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":2910
 *                     if name in params:
 *                         msg = 'duplicate parameter name: {!r}'.format(name)
 *                         raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *                     params[name] = param
 */
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2910, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_Raise(__pyx_t_5, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __PYX_ERR(0, 2910, __pyx_L1_error)

          /* "Aspidites/_vendor/contracts/inspection.py":2908
 *                             kind_defaults = True
 * 
 *                     if name in params:             # <<<<<<<<<<<<<<
 *                         msg = 'duplicate parameter name: {!r}'.format(name)
 *                         raise ValueError(msg)
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":2912
 *                         raise ValueError(msg)
 * 
 *                     params[name] = param             # <<<<<<<<<<<<<<
 *             else:
 *                 params = OrderedDict((param.name, param) for param in parameters)
 */
        if (unlikely(PyObject_SetItem(__pyx_v_params, __pyx_v_name, __pyx_v_param) < 0)) __PYX_ERR(0, 2912, __pyx_L1_error)

        /* "Aspidites/_vendor/contracts/inspection.py":2879
 *                 kind_defaults = False
 * 
 *                 for param in parameters:             # <<<<<<<<<<<<<<
 *                     kind = param.kind
 *                     name = param.name
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2874
 *             params = OrderedDict()
 *         else:
 *             if __validate_parameters__:             # <<<<<<<<<<<<<<
 *                 params = OrderedDict()
 *                 top_kind = _POSITIONAL_ONLY
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":2914
 *                     params[name] = param
 *             else:
 *                 params = OrderedDict((param.name, param) for param in parameters)             # <<<<<<<<<<<<<<
 * 
 *         self._parameters = types.MappingProxyType(params)
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_params = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L4:;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":2916
 *                 params = OrderedDict((param.name, param) for param in parameters)
 * 
 *         self._parameters = types.MappingProxyType(params)             # <<<<<<<<<<<<<<
 *         self._return_annotation = return_annotation
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MappingProxyType); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_v_params) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_params);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parameters_2, __pyx_t_3) < 0) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2917
 * 
 *         self._parameters = types.MappingProxyType(params)
 *         self._return_annotation = return_annotation             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation_2, __pyx_v_return_annotation) < 0) __PYX_ERR(0, 2917, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF(__pyx_v_top_kind);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8__init___2generator3);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2920
 * 
 *     @classmethod
 *     def from_function(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given python function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_3from_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_2from_function[] = "Constructs Signature for the given python function.\n\n        Deprecated since Python 3.5, use `Signature.from_callable()`.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_3from_function = {"from_function", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_3from_function, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_2from_function};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_3from_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_func = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_func_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("from_function", 1, 2, 2, 1); __PYX_ERR(0, 2920, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "from_function") < 0)) __PYX_ERR(0, 2920, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_func = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2920, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_2from_function(__pyx_self, __pyx_v_cls, __pyx_v_func);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_2from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_function", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2926
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_function() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2928
 *         warnings.warn("inspect.Signature.from_function() is deprecated since "
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *         return _signature_from_function(cls, func)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2928, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2926
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_function() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__79, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2929
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 *         return _signature_from_function(cls, func)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_signature_from_function); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cls, __pyx_v_func};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cls, __pyx_v_func};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_func);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2920
 * 
 *     @classmethod
 *     def from_function(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given python function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2932
 * 
 *     @classmethod
 *     def from_builtin(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given builtin function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_5from_builtin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_4from_builtin[] = "Constructs Signature for the given builtin function.\n\n        Deprecated since Python 3.5, use `Signature.from_callable()`.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_5from_builtin = {"from_builtin", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_5from_builtin, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_4from_builtin};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_5from_builtin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_func = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_builtin (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_func_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("from_builtin", 1, 2, 2, 1); __PYX_ERR(0, 2932, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "from_builtin") < 0)) __PYX_ERR(0, 2932, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_func = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_builtin", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2932, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_builtin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_4from_builtin(__pyx_self, __pyx_v_cls, __pyx_v_func);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_4from_builtin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_builtin", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2938
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_builtin() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2940
 *         warnings.warn("inspect.Signature.from_builtin() is deprecated since "
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *         return _signature_from_builtin(cls, func)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2940, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2938
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_builtin() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__80, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2941
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 *         return _signature_from_builtin(cls, func)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_signature_from_builtin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cls, __pyx_v_func};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cls, __pyx_v_func};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_func);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2932
 * 
 *     @classmethod
 *     def from_builtin(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given builtin function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_builtin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2944
 * 
 *     @classmethod
 *     def from_callable(cls, obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_7from_callable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_6from_callable[] = "Constructs Signature for the given callable object.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_7from_callable = {"from_callable", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_7from_callable, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_6from_callable};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_7from_callable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_follow_wrapped = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_callable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_obj,&__pyx_n_s_follow_wrapped,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("from_callable", 1, 2, 2, 1); __PYX_ERR(0, 2944, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 2;
        PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
        if (value) { values[index] = value; kw_args--; }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "from_callable") < 0)) __PYX_ERR(0, 2944, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_obj = values[1];
    __pyx_v_follow_wrapped = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_callable", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2944, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_6from_callable(__pyx_self, __pyx_v_cls, __pyx_v_obj, __pyx_v_follow_wrapped);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_6from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapped) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_callable", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2946
 *     def from_callable(cls, obj, *, follow_wrapped=True):
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,             # <<<<<<<<<<<<<<
 *                                         follow_wrapper_chains=follow_wrapped)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_signature_from_callable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_obj);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sigcls, __pyx_v_cls) < 0) __PYX_ERR(0, 2946, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2947
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,
 *                                         follow_wrapper_chains=follow_wrapped)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_follow_wrapper_chains, __pyx_v_follow_wrapped) < 0) __PYX_ERR(0, 2946, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2946
 *     def from_callable(cls, obj, *, follow_wrapped=True):
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,             # <<<<<<<<<<<<<<
 *                                         follow_wrapper_chains=follow_wrapped)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2944
 * 
 *     @classmethod
 *     def from_callable(cls, obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2950
 * 
 *     @property
 *     def parameters(self):             # <<<<<<<<<<<<<<
 *         return self._parameters
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_9parameters(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_9parameters = {"parameters", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_9parameters, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_9parameters(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parameters (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8parameters(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_8parameters(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parameters", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2951
 *     @property
 *     def parameters(self):
 *         return self._parameters             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parameters_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2950
 * 
 *     @property
 *     def parameters(self):             # <<<<<<<<<<<<<<
 *         return self._parameters
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2954
 * 
 *     @property
 *     def return_annotation(self):             # <<<<<<<<<<<<<<
 *         return self._return_annotation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11return_annotation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11return_annotation = {"return_annotation", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11return_annotation, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11return_annotation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("return_annotation (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_10return_annotation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_10return_annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("return_annotation", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2955
 *     @property
 *     def return_annotation(self):
 *         return self._return_annotation             # <<<<<<<<<<<<<<
 * 
 *     def replace(self, *, parameters=_void, return_annotation=_void):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2954
 * 
 *     @property
 *     def return_annotation(self):             # <<<<<<<<<<<<<<
 *         return self._return_annotation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.return_annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2957
 *         return self._return_annotation
 * 
 *     def replace(self, *, parameters=_void, return_annotation=_void):             # <<<<<<<<<<<<<<
 *         """Creates a customized copy of the Signature.
 *         Pass 'parameters' and/or 'return_annotation' arguments
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_203__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_parameters, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_parameters) < 0) __PYX_ERR(0, 2957, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_annotation, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_return_annotation) < 0) __PYX_ERR(0, 2957, __pyx_L1_error)
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 0, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_13replace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_12replace[] = "Creates a customized copy of the Signature.\n        Pass 'parameters' and/or 'return_annotation' arguments\n        to override them in the new copy.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_13replace = {"replace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_13replace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_12replace};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_13replace(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parameters = 0;
  PyObject *__pyx_v_return_annotation = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_parameters,&__pyx_n_s_return_annotation,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults6 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_parameters;
    values[2] = __pyx_dynamic_args->__pyx_arg_return_annotation;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 1; index < 3 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "replace") < 0)) __PYX_ERR(0, 2957, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
    __pyx_v_parameters = values[1];
    __pyx_v_return_annotation = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("replace", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2957, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_12replace(__pyx_self, __pyx_v_self, __pyx_v_parameters, __pyx_v_return_annotation);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_12replace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parameters, PyObject *__pyx_v_return_annotation) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 0);
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_INCREF(__pyx_v_return_annotation);

  /* "Aspidites/_vendor/contracts/inspection.py":2963
 *         """
 * 
 *         if parameters is _void:             # <<<<<<<<<<<<<<
 *             parameters = self.parameters.values()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_parameters == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":2964
 * 
 *         if parameters is _void:
 *             parameters = self.parameters.values()             # <<<<<<<<<<<<<<
 * 
 *         if return_annotation is _void:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parameters); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_parameters, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2963
 *         """
 * 
 *         if parameters is _void:             # <<<<<<<<<<<<<<
 *             parameters = self.parameters.values()
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2966
 *             parameters = self.parameters.values()
 * 
 *         if return_annotation is _void:             # <<<<<<<<<<<<<<
 *             return_annotation = self._return_annotation
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_return_annotation == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2967
 * 
 *         if return_annotation is _void:
 *             return_annotation = self._return_annotation             # <<<<<<<<<<<<<<
 * 
 *         return type(self)(parameters,
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_return_annotation, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2966
 *             parameters = self.parameters.values()
 * 
 *         if return_annotation is _void:             # <<<<<<<<<<<<<<
 *             return_annotation = self._return_annotation
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2969
 *             return_annotation = self._return_annotation
 * 
 *         return type(self)(parameters,             # <<<<<<<<<<<<<<
 *                           return_annotation=return_annotation)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_v_parameters);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_parameters);

  /* "Aspidites/_vendor/contracts/inspection.py":2970
 * 
 *         return type(self)(parameters,
 *                           return_annotation=return_annotation)             # <<<<<<<<<<<<<<
 * 
 *     def _hash_basis(self):
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return_annotation, __pyx_v_return_annotation) < 0) __PYX_ERR(0, 2970, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2969
 *             return_annotation = self._return_annotation
 * 
 *         return type(self)(parameters,             # <<<<<<<<<<<<<<
 *                           return_annotation=return_annotation)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)Py_TYPE(__pyx_v_self)), __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2957
 *         return self._return_annotation
 * 
 *     def replace(self, *, parameters=_void, return_annotation=_void):             # <<<<<<<<<<<<<<
 *         """Creates a customized copy of the Signature.
 *         Pass 'parameters' and/or 'return_annotation' arguments
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parameters);
  __Pyx_XDECREF(__pyx_v_return_annotation);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_15_hash_basis(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_15_hash_basis = {"_hash_basis", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_15_hash_basis, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_15_hash_basis(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_basis (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_14_hash_basis(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/contracts/inspection.py":2973
 * 
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2973, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Signature__hash_basis_locals_gen, __pyx_n_s_Aspidites__vendor_contracts_insp); if (unlikely(!gen)) __PYX_ERR(0, 2973, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._hash_basis.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 2973, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_parameters); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(0, 2973, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_values, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, NULL, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_param);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_param, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":2974
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)             # <<<<<<<<<<<<<<
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2974, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 2974, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/contracts/inspection.py":2973
 * 
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_param);
      __pyx_r = __pyx_cur_scope->__pyx_v_param;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2973, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":2974
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)             # <<<<<<<<<<<<<<
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "Aspidites/_vendor/contracts/inspection.py":2973
 * 
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_14_hash_basis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *__pyx_cur_scope;
  PyObject *__pyx_v_params = NULL;
  PyObject *__pyx_v_kwo_params = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_2generator4 = 0;
  PyObject *__pyx_9genexpr11__pyx_v_param = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_basis", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(__pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2972, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/contracts/inspection.py":2973
 * 
 *     def _hash_basis(self):
 *         params = tuple(param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_params = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2976
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                       if param.kind == _KEYWORD_ONLY}
 * 
 */
  { /* enter inner scope */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2976, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_parameters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2976, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
      __PYX_ERR(0, 2976, __pyx_L5_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_values, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2976, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_4, &__pyx_t_3, NULL, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 2976, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_param, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":2977
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()
 *                       if param.kind == _KEYWORD_ONLY}             # <<<<<<<<<<<<<<
 * 
 *         return params, kwo_params, self.return_annotation
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr11__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2977, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2977, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2977, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 2977, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_10) {

        /* "Aspidites/_vendor/contracts/inspection.py":2976
 *                        if param.kind != _KEYWORD_ONLY)
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()             # <<<<<<<<<<<<<<
 *                       if param.kind == _KEYWORD_ONLY}
 * 
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr11__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2976, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_t_9, (PyObject*)__pyx_9genexpr11__pyx_v_param))) __PYX_ERR(0, 2976, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":2977
 * 
 *         kwo_params = {param.name: param for param in self.parameters.values()
 *                       if param.kind == _KEYWORD_ONLY}             # <<<<<<<<<<<<<<
 * 
 *         return params, kwo_params, self.return_annotation
 */
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_param); __pyx_9genexpr11__pyx_v_param = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_param); __pyx_9genexpr11__pyx_v_param = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_kwo_params = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2979
 *                       if param.kind == _KEYWORD_ONLY}
 * 
 *         return params, kwo_params, self.return_annotation             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_return_annotation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_params);
  __Pyx_GIVEREF(__pyx_v_params);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_params);
  __Pyx_INCREF(__pyx_v_kwo_params);
  __Pyx_GIVEREF(__pyx_v_kwo_params);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_kwo_params);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._hash_basis", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF(__pyx_v_kwo_params);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11_hash_basis_2generator4);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_param);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2981
 *         return params, kwo_params, self.return_annotation
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_17__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_17__hash__ = {"__hash__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_17__hash__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_17__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_16__hash__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_16__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_params = NULL;
  PyObject *__pyx_v_kwo_params = NULL;
  PyObject *__pyx_v_return_annotation = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  Py_hash_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2982
 * 
 *     def __hash__(self):
 *         params, kwo_params, return_annotation = self._hash_basis()             # <<<<<<<<<<<<<<
 *         kwo_params = frozenset(kwo_params.values())
 *         return hash((params, kwo_params, return_annotation))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_hash_basis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 2982, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 2982, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 2982, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_params = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_kwo_params = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_return_annotation = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2983
 *     def __hash__(self):
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())             # <<<<<<<<<<<<<<
 *         return hash((params, kwo_params, return_annotation))
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwo_params, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyFrozenSet_New(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_kwo_params, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2984
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())
 *         return hash((params, kwo_params, return_annotation))             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_params);
  __Pyx_GIVEREF(__pyx_v_params);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_params);
  __Pyx_INCREF(__pyx_v_kwo_params);
  __Pyx_GIVEREF(__pyx_v_kwo_params);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_kwo_params);
  __Pyx_INCREF(__pyx_v_return_annotation);
  __Pyx_GIVEREF(__pyx_v_return_annotation);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_return_annotation);
  __pyx_t_7 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_7 == ((Py_hash_t)-1))) __PYX_ERR(0, 2984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromHash_t(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2981
 *         return params, kwo_params, self.return_annotation
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF(__pyx_v_kwo_params);
  __Pyx_XDECREF(__pyx_v_return_annotation);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2986
 *         return hash((params, kwo_params, return_annotation))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_19__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_19__eq__ = {"__eq__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_19__eq__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_19__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, 1); __PYX_ERR(0, 2986, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__eq__") < 0)) __PYX_ERR(0, 2986, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2986, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_18__eq__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_18__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2987
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, Signature):
 */
  __pyx_t_1 = (__pyx_v_self == __pyx_v_other);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/contracts/inspection.py":2988
 *     def __eq__(self, other):
 *         if self is other:
 *             return True             # <<<<<<<<<<<<<<
 *         if not isinstance(other, Signature):
 *             return NotImplemented
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2987
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         if not isinstance(other, Signature):
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2989
 *         if self is other:
 *             return True
 *         if not isinstance(other, Signature):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._hash_basis() == other._hash_basis()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Signature); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/contracts/inspection.py":2990
 *             return True
 *         if not isinstance(other, Signature):
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return self._hash_basis() == other._hash_basis()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":2989
 *         if self is other:
 *             return True
 *         if not isinstance(other, Signature):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._hash_basis() == other._hash_basis()
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":2991
 *         if not isinstance(other, Signature):
 *             return NotImplemented
 *         return self._hash_basis() == other._hash_basis()             # <<<<<<<<<<<<<<
 * 
 *     def _bind(self, args, kwargs, *, partial=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_hash_basis); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_hash_basis); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2991, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2986
 *         return hash((params, kwo_params, return_annotation))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":2993
 *         return self._hash_basis() == other._hash_basis()
 * 
 *     def _bind(self, args, kwargs, *, partial=False):             # <<<<<<<<<<<<<<
 *         """Private method. Don't use directly."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_21_bind(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_20_bind[] = "Private method. Don't use directly.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_21_bind = {"_bind", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_21_bind, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_9Signature_20_bind};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_21_bind(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_v_partial = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bind (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_partial,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bind", 1, 3, 3, 1); __PYX_ERR(0, 2993, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bind", 1, 3, 3, 2); __PYX_ERR(0, 2993, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 3;
        PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
        if (value) { values[index] = value; kw_args--; }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_bind") < 0)) __PYX_ERR(0, 2993, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_partial = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bind", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2993, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_20_bind(__pyx_self, __pyx_v_self, __pyx_v_args, __pyx_v_kwargs, __pyx_v_partial);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_20_bind(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_partial) {
  PyObject *__pyx_v_arguments = NULL;
  PyObject *__pyx_v_parameters = NULL;
  PyObject *__pyx_v_parameters_ex = NULL;
  PyObject *__pyx_v_arg_vals = NULL;
  PyObject *__pyx_v_arg_val = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_kwargs_param = NULL;
  PyObject *__pyx_v_param_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bind", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":2996
 *         """Private method. Don't use directly."""
 * 
 *         arguments = {}             # <<<<<<<<<<<<<<
 * 
 *         parameters = iter(self.parameters.values())
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_arguments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2998
 *         arguments = {}
 * 
 *         parameters = iter(self.parameters.values())             # <<<<<<<<<<<<<<
 *         parameters_ex = ()
 *         arg_vals = iter(args)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parameters); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_parameters = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2999
 * 
 *         parameters = iter(self.parameters.values())
 *         parameters_ex = ()             # <<<<<<<<<<<<<<
 *         arg_vals = iter(args)
 * 
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_parameters_ex = __pyx_empty_tuple;

  /* "Aspidites/_vendor/contracts/inspection.py":3000
 *         parameters = iter(self.parameters.values())
 *         parameters_ex = ()
 *         arg_vals = iter(args)             # <<<<<<<<<<<<<<
 * 
 *         while True:
 */
  __pyx_t_3 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_arg_vals = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3002
 *         arg_vals = iter(args)
 * 
 *         while True:             # <<<<<<<<<<<<<<
 *             # Let's iterate through the positional arguments and corresponding
 *             # parameters
 */
  while (1) {

    /* "Aspidites/_vendor/contracts/inspection.py":3005
 *             # Let's iterate through the positional arguments and corresponding
 *             # parameters
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = next(arg_vals)
 *             except StopIteration:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":3006
 *             # parameters
 *             try:
 *                 arg_val = next(arg_vals)             # <<<<<<<<<<<<<<
 *             except StopIteration:
 *                 # No more positional arguments
 */
        __pyx_t_3 = __Pyx_PyIter_Next(__pyx_v_arg_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3006, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_arg_val, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":3005
 *             # Let's iterate through the positional arguments and corresponding
 *             # parameters
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = next(arg_vals)
 *             except StopIteration:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":3047
 *             else:
 *                 # We have a positional argument to process
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
      /*else:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":3048
 *                 # We have a positional argument to process
 *                 try:
 *                     param = next(parameters)             # <<<<<<<<<<<<<<
 *                 except StopIteration:
 *                     raise TypeError('too many positional arguments') from None
 */
            __pyx_t_3 = __Pyx_PyIter_Next(__pyx_v_parameters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3048, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":3047
 *             else:
 *                 # We have a positional argument to process
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
          }

          /* "Aspidites/_vendor/contracts/inspection.py":3052
 *                     raise TypeError('too many positional arguments') from None
 *                 else:
 *                     if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                         # Looks like we have no parameter for this positional
 *                         # argument
 */
          /*else:*/ {
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (!__pyx_t_11) {
            } else {
              __pyx_t_10 = __pyx_t_11;
              goto __pyx_L22_bool_binop_done;
            }
            __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3052, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_10 = __pyx_t_11;
            __pyx_L22_bool_binop_done:;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_11 = (__pyx_t_10 != 0);
            if (unlikely(__pyx_t_11)) {

              /* "Aspidites/_vendor/contracts/inspection.py":3055
 *                         # Looks like we have no parameter for this positional
 *                         # argument
 *                         raise TypeError(             # <<<<<<<<<<<<<<
 *                             'too many positional arguments') from None
 * 
 */
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3055, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_3);

              /* "Aspidites/_vendor/contracts/inspection.py":3056
 *                         # argument
 *                         raise TypeError(
 *                             'too many positional arguments') from None             # <<<<<<<<<<<<<<
 * 
 *                     if param.kind == _VAR_POSITIONAL:
 */
              __Pyx_Raise(__pyx_t_3, 0, 0, Py_None);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __PYX_ERR(0, 3055, __pyx_L15_except_error)

              /* "Aspidites/_vendor/contracts/inspection.py":3052
 *                     raise TypeError('too many positional arguments') from None
 *                 else:
 *                     if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):             # <<<<<<<<<<<<<<
 *                         # Looks like we have no parameter for this positional
 *                         # argument
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3058
 *                             'too many positional arguments') from None
 * 
 *                     if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                         # We have an '*args'-like argument, let's fill it with
 *                         # all positional arguments we have left and move on to
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3058, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3058, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3058, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3058, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_11) {

              /* "Aspidites/_vendor/contracts/inspection.py":3062
 *                         # all positional arguments we have left and move on to
 *                         # the next phase
 *                         values = [arg_val]             # <<<<<<<<<<<<<<
 *                         values.extend(arg_vals)
 *                         arguments[param.name] = tuple(values)
 */
              __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3062, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_INCREF(__pyx_v_arg_val);
              __Pyx_GIVEREF(__pyx_v_arg_val);
              PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_arg_val);
              __pyx_v_values = ((PyObject*)__pyx_t_2);
              __pyx_t_2 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3063
 *                         # the next phase
 *                         values = [arg_val]
 *                         values.extend(arg_vals)             # <<<<<<<<<<<<<<
 *                         arguments[param.name] = tuple(values)
 *                         break
 */
              __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_values, __pyx_v_arg_vals); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 3063, __pyx_L15_except_error)

              /* "Aspidites/_vendor/contracts/inspection.py":3064
 *                         values = [arg_val]
 *                         values.extend(arg_vals)
 *                         arguments[param.name] = tuple(values)             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
              __pyx_t_2 = PyList_AsTuple(__pyx_v_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3064, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3064, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_1);
              if (unlikely(PyDict_SetItem(__pyx_v_arguments, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 3064, __pyx_L15_except_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3065
 *                         values.extend(arg_vals)
 *                         arguments[param.name] = tuple(values)
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:
 */
              goto __pyx_L18_try_break;

              /* "Aspidites/_vendor/contracts/inspection.py":3058
 *                             'too many positional arguments') from None
 * 
 *                     if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                         # We have an '*args'-like argument, let's fill it with
 *                         # all positional arguments we have left and move on to
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3067
 *                         break
 * 
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_kwargs, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_13 = (__pyx_t_10 != 0);
            if (__pyx_t_13) {
            } else {
              __pyx_t_11 = __pyx_t_13;
              goto __pyx_L26_bool_binop_done;
            }
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3067, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_11 = __pyx_t_13;
            __pyx_L26_bool_binop_done:;
            if (unlikely(__pyx_t_11)) {

              /* "Aspidites/_vendor/contracts/inspection.py":3069
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(             # <<<<<<<<<<<<<<
 *                                 arg=param.name)) from None
 * 
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_multiple_values_for_argument_arg, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3069, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_3);

              /* "Aspidites/_vendor/contracts/inspection.py":3070
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(
 *                                 arg=param.name)) from None             # <<<<<<<<<<<<<<
 * 
 *                     arguments[param.name] = arg_val
 */
              __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3070, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3070, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_arg, __pyx_t_2) < 0) __PYX_ERR(0, 3070, __pyx_L15_except_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3069
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(             # <<<<<<<<<<<<<<
 *                                 arg=param.name)) from None
 * 
 */
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3069, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3068
 * 
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:
 *                         raise TypeError(             # <<<<<<<<<<<<<<
 *                             'multiple values for argument {arg!r}'.format(
 *                                 arg=param.name)) from None
 */
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3070
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(
 *                                 arg=param.name)) from None             # <<<<<<<<<<<<<<
 * 
 *                     arguments[param.name] = arg_val
 */
              __Pyx_Raise(__pyx_t_1, 0, 0, Py_None);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __PYX_ERR(0, 3068, __pyx_L15_except_error)

              /* "Aspidites/_vendor/contracts/inspection.py":3067
 *                         break
 * 
 *                     if param.name in kwargs and param.kind != _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                         raise TypeError(
 *                             'multiple values for argument {arg!r}'.format(
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3072
 *                                 arg=param.name)) from None
 * 
 *                     arguments[param.name] = arg_val             # <<<<<<<<<<<<<<
 * 
 *         # Now, we iterate through the remaining parameters to process
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3072, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (unlikely(PyDict_SetItem(__pyx_v_arguments, __pyx_t_1, __pyx_v_arg_val) < 0)) __PYX_ERR(0, 3072, __pyx_L15_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L20_try_end;
          __pyx_L13_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":3049
 *                 try:
 *                     param = next(parameters)
 *                 except StopIteration:             # <<<<<<<<<<<<<<
 *                     raise TypeError('too many positional arguments') from None
 *                 else:
 */
          __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
          if (__pyx_t_14) {
            __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 3049, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GOTREF(__pyx_t_3);

            /* "Aspidites/_vendor/contracts/inspection.py":3050
 *                     param = next(parameters)
 *                 except StopIteration:
 *                     raise TypeError('too many positional arguments') from None             # <<<<<<<<<<<<<<
 *                 else:
 *                     if param.kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 */
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3050, __pyx_L15_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_Raise(__pyx_t_15, 0, 0, Py_None);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __PYX_ERR(0, 3050, __pyx_L15_except_error)
          }
          goto __pyx_L15_except_error;
          __pyx_L15_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":3047
 *             else:
 *                 # We have a positional argument to process
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L7_except_error;
          __pyx_L18_try_break:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L10_try_break;
          __pyx_L20_try_end:;
        }
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3007
 *             try:
 *                 arg_val = next(arg_vals)
 *             except StopIteration:             # <<<<<<<<<<<<<<
 *                 # No more positional arguments
 *                 try:
 */
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
      if (__pyx_t_14) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 3007, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Aspidites/_vendor/contracts/inspection.py":3009
 *             except StopIteration:
 *                 # No more positional arguments
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_7);
          /*try:*/ {

            /* "Aspidites/_vendor/contracts/inspection.py":3010
 *                 # No more positional arguments
 *                 try:
 *                     param = next(parameters)             # <<<<<<<<<<<<<<
 *                 except StopIteration:
 *                     # No more parameters. That's it. Just need to check that
 */
            __pyx_t_15 = __Pyx_PyIter_Next(__pyx_v_parameters); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3010, __pyx_L32_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_15);
            __pyx_t_15 = 0;

            /* "Aspidites/_vendor/contracts/inspection.py":3009
 *             except StopIteration:
 *                 # No more positional arguments
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
          }

          /* "Aspidites/_vendor/contracts/inspection.py":3016
 *                     break
 *                 else:
 *                     if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                         # That's OK, just empty *args.  Let's start parsing
 *                         # kwargs
 */
          /*else:*/ {
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3016, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3016, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_17 = PyObject_RichCompare(__pyx_t_15, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_17); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3016, __pyx_L34_except_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3016, __pyx_L34_except_error)
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (__pyx_t_11) {

              /* "Aspidites/_vendor/contracts/inspection.py":3019
 *                         # That's OK, just empty *args.  Let's start parsing
 *                         # kwargs
 *                         break             # <<<<<<<<<<<<<<
 *                     elif param.name in kwargs:
 *                         if param.kind == _POSITIONAL_ONLY:
 */
              goto __pyx_L37_try_break;

              /* "Aspidites/_vendor/contracts/inspection.py":3016
 *                     break
 *                 else:
 *                     if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                         # That's OK, just empty *args.  Let's start parsing
 *                         # kwargs
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3020
 *                         # kwargs
 *                         break
 *                     elif param.name in kwargs:             # <<<<<<<<<<<<<<
 *                         if param.kind == _POSITIONAL_ONLY:
 *                             msg = '{arg!r} parameter is positional only, ' \
 */
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3020, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_17, __pyx_v_kwargs, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3020, __pyx_L34_except_error)
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_13 = (__pyx_t_11 != 0);
            if (__pyx_t_13) {

              /* "Aspidites/_vendor/contracts/inspection.py":3021
 *                         break
 *                     elif param.name in kwargs:
 *                         if param.kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                             msg = '{arg!r} parameter is positional only, ' \
 *                                   'but was passed as a keyword'
 */
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3021, __pyx_L34_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3021, __pyx_L34_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_15 = PyObject_RichCompare(__pyx_t_17, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_15); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3021, __pyx_L34_except_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_15); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3021, __pyx_L34_except_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(__pyx_t_13)) {

                /* "Aspidites/_vendor/contracts/inspection.py":3022
 *                     elif param.name in kwargs:
 *                         if param.kind == _POSITIONAL_ONLY:
 *                             msg = '{arg!r} parameter is positional only, ' \             # <<<<<<<<<<<<<<
 *                                   'but was passed as a keyword'
 *                             msg = msg.format(arg=param.name)
 */
                __Pyx_INCREF(__pyx_kp_u_arg_r_parameter_is_positional_o);
                __pyx_v_msg = __pyx_kp_u_arg_r_parameter_is_positional_o;

                /* "Aspidites/_vendor/contracts/inspection.py":3024
 *                             msg = '{arg!r} parameter is positional only, ' \
 *                                   'but was passed as a keyword'
 *                             msg = msg.format(arg=param.name)             # <<<<<<<<<<<<<<
 *                             raise TypeError(msg) from None
 *                         parameters_ex = (param,)
 */
                __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3024, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3024, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_16);
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3024, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_arg, __pyx_t_17) < 0) __PYX_ERR(0, 3024, __pyx_L34_except_error)
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_empty_tuple, __pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3024, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_17);
                __pyx_t_17 = 0;

                /* "Aspidites/_vendor/contracts/inspection.py":3025
 *                                   'but was passed as a keyword'
 *                             msg = msg.format(arg=param.name)
 *                             raise TypeError(msg) from None             # <<<<<<<<<<<<<<
 *                         parameters_ex = (param,)
 *                         break
 */
                __pyx_t_17 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_v_msg); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3025, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_Raise(__pyx_t_17, 0, 0, Py_None);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __PYX_ERR(0, 3025, __pyx_L34_except_error)

                /* "Aspidites/_vendor/contracts/inspection.py":3021
 *                         break
 *                     elif param.name in kwargs:
 *                         if param.kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                             msg = '{arg!r} parameter is positional only, ' \
 *                                   'but was passed as a keyword'
 */
              }

              /* "Aspidites/_vendor/contracts/inspection.py":3026
 *                             msg = msg.format(arg=param.name)
 *                             raise TypeError(msg) from None
 *                         parameters_ex = (param,)             # <<<<<<<<<<<<<<
 *                         break
 *                     elif (param.kind == _VAR_KEYWORD or
 */
              __pyx_t_17 = PyTuple_New(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3026, __pyx_L34_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_INCREF(__pyx_v_param);
              __Pyx_GIVEREF(__pyx_v_param);
              PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v_param);
              __Pyx_DECREF_SET(__pyx_v_parameters_ex, ((PyObject*)__pyx_t_17));
              __pyx_t_17 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3027
 *                             raise TypeError(msg) from None
 *                         parameters_ex = (param,)
 *                         break             # <<<<<<<<<<<<<<
 *                     elif (param.kind == _VAR_KEYWORD or
 *                           param.default is not _empty):
 */
              goto __pyx_L37_try_break;

              /* "Aspidites/_vendor/contracts/inspection.py":3020
 *                         # kwargs
 *                         break
 *                     elif param.name in kwargs:             # <<<<<<<<<<<<<<
 *                         if param.kind == _POSITIONAL_ONLY:
 *                             msg = '{arg!r} parameter is positional only, ' \
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3028
 *                         parameters_ex = (param,)
 *                         break
 *                     elif (param.kind == _VAR_KEYWORD or             # <<<<<<<<<<<<<<
 *                           param.default is not _empty):
 *                         # That's fine too - we have a default value for this
 */
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3028, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3028, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = PyObject_RichCompare(__pyx_t_17, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_15); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3028, __pyx_L34_except_error)
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_15); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3028, __pyx_L34_except_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (!__pyx_t_11) {
            } else {
              __pyx_t_13 = __pyx_t_11;
              goto __pyx_L42_bool_binop_done;
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3029
 *                         break
 *                     elif (param.kind == _VAR_KEYWORD or
 *                           param.default is not _empty):             # <<<<<<<<<<<<<<
 *                         # That's fine too - we have a default value for this
 *                         # parameter.  So, lets start parsing `kwargs`, starting
 */
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_default); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3029, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_empty); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3029, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_11 = (__pyx_t_15 != __pyx_t_16);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_10 = (__pyx_t_11 != 0);
            __pyx_t_13 = __pyx_t_10;
            __pyx_L42_bool_binop_done:;

            /* "Aspidites/_vendor/contracts/inspection.py":3028
 *                         parameters_ex = (param,)
 *                         break
 *                     elif (param.kind == _VAR_KEYWORD or             # <<<<<<<<<<<<<<
 *                           param.default is not _empty):
 *                         # That's fine too - we have a default value for this
 */
            if (__pyx_t_13) {

              /* "Aspidites/_vendor/contracts/inspection.py":3033
 *                         # parameter.  So, lets start parsing `kwargs`, starting
 *                         # with the current parameter
 *                         parameters_ex = (param,)             # <<<<<<<<<<<<<<
 *                         break
 *                     else:
 */
              __pyx_t_16 = PyTuple_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3033, __pyx_L34_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_INCREF(__pyx_v_param);
              __Pyx_GIVEREF(__pyx_v_param);
              PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_param);
              __Pyx_DECREF_SET(__pyx_v_parameters_ex, ((PyObject*)__pyx_t_16));
              __pyx_t_16 = 0;

              /* "Aspidites/_vendor/contracts/inspection.py":3034
 *                         # with the current parameter
 *                         parameters_ex = (param,)
 *                         break             # <<<<<<<<<<<<<<
 *                     else:
 *                         # No default, not VAR_KEYWORD, not VAR_POSITIONAL,
 */
              goto __pyx_L37_try_break;

              /* "Aspidites/_vendor/contracts/inspection.py":3028
 *                         parameters_ex = (param,)
 *                         break
 *                     elif (param.kind == _VAR_KEYWORD or             # <<<<<<<<<<<<<<
 *                           param.default is not _empty):
 *                         # That's fine too - we have a default value for this
 */
            }

            /* "Aspidites/_vendor/contracts/inspection.py":3038
 *                         # No default, not VAR_KEYWORD, not VAR_POSITIONAL,
 *                         # not in `kwargs`
 *                         if partial:             # <<<<<<<<<<<<<<
 *                             parameters_ex = (param,)
 *                             break
 */
            /*else*/ {
              __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_partial); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3038, __pyx_L34_except_error)
              if (likely(__pyx_t_13)) {

                /* "Aspidites/_vendor/contracts/inspection.py":3039
 *                         # not in `kwargs`
 *                         if partial:
 *                             parameters_ex = (param,)             # <<<<<<<<<<<<<<
 *                             break
 *                         else:
 */
                __pyx_t_16 = PyTuple_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3039, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_16);
                __Pyx_INCREF(__pyx_v_param);
                __Pyx_GIVEREF(__pyx_v_param);
                PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_param);
                __Pyx_DECREF_SET(__pyx_v_parameters_ex, ((PyObject*)__pyx_t_16));
                __pyx_t_16 = 0;

                /* "Aspidites/_vendor/contracts/inspection.py":3040
 *                         if partial:
 *                             parameters_ex = (param,)
 *                             break             # <<<<<<<<<<<<<<
 *                         else:
 *                             msg = 'missing a required argument: {arg!r}'
 */
                goto __pyx_L37_try_break;

                /* "Aspidites/_vendor/contracts/inspection.py":3038
 *                         # No default, not VAR_KEYWORD, not VAR_POSITIONAL,
 *                         # not in `kwargs`
 *                         if partial:             # <<<<<<<<<<<<<<
 *                             parameters_ex = (param,)
 *                             break
 */
              }

              /* "Aspidites/_vendor/contracts/inspection.py":3042
 *                             break
 *                         else:
 *                             msg = 'missing a required argument: {arg!r}'             # <<<<<<<<<<<<<<
 *                             msg = msg.format(arg=param.name)
 *                             raise TypeError(msg) from None
 */
              /*else*/ {
                __Pyx_INCREF(__pyx_kp_u_missing_a_required_argument_arg);
                __pyx_v_msg = __pyx_kp_u_missing_a_required_argument_arg;

                /* "Aspidites/_vendor/contracts/inspection.py":3043
 *                         else:
 *                             msg = 'missing a required argument: {arg!r}'
 *                             msg = msg.format(arg=param.name)             # <<<<<<<<<<<<<<
 *                             raise TypeError(msg) from None
 *             else:
 */
                __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3043, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_16);
                __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3043, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3043, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_arg, __pyx_t_17) < 0) __PYX_ERR(0, 3043, __pyx_L34_except_error)
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_empty_tuple, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3043, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_17);
                __pyx_t_17 = 0;

                /* "Aspidites/_vendor/contracts/inspection.py":3044
 *                             msg = 'missing a required argument: {arg!r}'
 *                             msg = msg.format(arg=param.name)
 *                             raise TypeError(msg) from None             # <<<<<<<<<<<<<<
 *             else:
 *                 # We have a positional argument to process
 */
                __pyx_t_17 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_v_msg); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3044, __pyx_L34_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_Raise(__pyx_t_17, 0, 0, Py_None);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __PYX_ERR(0, 3044, __pyx_L34_except_error)
              }
            }
          }
          __pyx_L32_error:;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":3011
 *                 try:
 *                     param = next(parameters)
 *                 except StopIteration:             # <<<<<<<<<<<<<<
 *                     # No more parameters. That's it. Just need to check that
 *                     # we have no `kwargs` after this while loop
 */
          __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
          if (__pyx_t_14) {
            __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_17, &__pyx_t_15, &__pyx_t_16) < 0) __PYX_ERR(0, 3011, __pyx_L34_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GOTREF(__pyx_t_16);

            /* "Aspidites/_vendor/contracts/inspection.py":3014
 *                     # No more parameters. That's it. Just need to check that
 *                     # we have no `kwargs` after this while loop
 *                     break             # <<<<<<<<<<<<<<
 *                 else:
 *                     if param.kind == _VAR_POSITIONAL:
 */
            goto __pyx_L45_except_break;
            __pyx_L45_except_break:;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            goto __pyx_L37_try_break;
          }
          goto __pyx_L34_except_error;
          __pyx_L34_except_error:;

          /* "Aspidites/_vendor/contracts/inspection.py":3009
 *             except StopIteration:
 *                 # No more positional arguments
 *                 try:             # <<<<<<<<<<<<<<
 *                     param = next(parameters)
 *                 except StopIteration:
 */
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
          goto __pyx_L7_except_error;
          __pyx_L37_try_break:;
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
          goto __pyx_L30_except_break;
        }
        __pyx_L30_except_break:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L10_try_break;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":3005
 *             # Let's iterate through the positional arguments and corresponding
 *             # parameters
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = next(arg_vals)
 *             except StopIteration:
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L10_try_break:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L4_break;
      __pyx_L12_try_end:;
    }
  }
  __pyx_L4_break:;

  /* "Aspidites/_vendor/contracts/inspection.py":3076
 *         # Now, we iterate through the remaining parameters to process
 *         # keyword arguments
 *         kwargs_param = None             # <<<<<<<<<<<<<<
 *         for param in itertools.chain(parameters_ex, parameters):
 *             if param.kind == _VAR_KEYWORD:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_kwargs_param = Py_None;

  /* "Aspidites/_vendor/contracts/inspection.py":3077
 *         # keyword arguments
 *         kwargs_param = None
 *         for param in itertools.chain(parameters_ex, parameters):             # <<<<<<<<<<<<<<
 *             if param.kind == _VAR_KEYWORD:
 *                 # Memorize that we have a '**kwargs'-like parameter
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_chain); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_14 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_14 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parameters_ex, __pyx_v_parameters};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parameters_ex, __pyx_v_parameters};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_16 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parameters_ex);
    __Pyx_GIVEREF(__pyx_v_parameters_ex);
    PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_14, __pyx_v_parameters_ex);
    __Pyx_INCREF(__pyx_v_parameters);
    __Pyx_GIVEREF(__pyx_v_parameters);
    PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_14, __pyx_v_parameters);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_18 = 0;
    __pyx_t_19 = NULL;
  } else {
    __pyx_t_18 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_19 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 3077, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_19)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 3077, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 3077, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_19(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3077, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3078
 *         kwargs_param = None
 *         for param in itertools.chain(parameters_ex, parameters):
 *             if param.kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 # Memorize that we have a '**kwargs'-like parameter
 *                 kwargs_param = param
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3078, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3078, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":3080
 *             if param.kind == _VAR_KEYWORD:
 *                 # Memorize that we have a '**kwargs'-like parameter
 *                 kwargs_param = param             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __Pyx_INCREF(__pyx_v_param);
      __Pyx_DECREF_SET(__pyx_v_kwargs_param, __pyx_v_param);

      /* "Aspidites/_vendor/contracts/inspection.py":3081
 *                 # Memorize that we have a '**kwargs'-like parameter
 *                 kwargs_param = param
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             if param.kind == _VAR_POSITIONAL:
 */
      goto __pyx_L47_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":3078
 *         kwargs_param = None
 *         for param in itertools.chain(parameters_ex, parameters):
 *             if param.kind == _VAR_KEYWORD:             # <<<<<<<<<<<<<<
 *                 # Memorize that we have a '**kwargs'-like parameter
 *                 kwargs_param = param
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3083
 *                 continue
 * 
 *             if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # Named arguments don't refer to '*args'-like parameters.
 *                 # We only arrive here if the positional arguments ended
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_13) {

      /* "Aspidites/_vendor/contracts/inspection.py":3087
 *                 # We only arrive here if the positional arguments ended
 *                 # before reaching the last parameter before *args.
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             param_name = param.name
 */
      goto __pyx_L47_continue;

      /* "Aspidites/_vendor/contracts/inspection.py":3083
 *                 continue
 * 
 *             if param.kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # Named arguments don't refer to '*args'-like parameters.
 *                 # We only arrive here if the positional arguments ended
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3089
 *                 continue
 * 
 *             param_name = param.name             # <<<<<<<<<<<<<<
 *             try:
 *                 arg_val = kwargs.pop(param_name)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3089, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_param_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3090
 * 
 *             param_name = param.name
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = kwargs.pop(param_name)
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":3091
 *             param_name = param.name
 *             try:
 *                 arg_val = kwargs.pop(param_name)             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 # We have no value for this parameter.  It's fine though,
 */
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_pop); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3091, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_16);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_16, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_2, __pyx_v_param_name) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_param_name);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3091, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_arg_val, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/contracts/inspection.py":3090
 * 
 *             param_name = param.name
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = kwargs.pop(param_name)
 *             except KeyError:
 */
      }

      /* "Aspidites/_vendor/contracts/inspection.py":3103
 * 
 *             else:
 *                 if param.kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                     # This should never happen in case of a properly built
 *                     # Signature object (but let's have this check here
 */
      /*else:*/ {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3103, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3103, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_16, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3103, __pyx_L53_except_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3103, __pyx_L53_except_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(__pyx_t_13)) {

          /* "Aspidites/_vendor/contracts/inspection.py":3108
 *                     # to ensure correct behaviour just in case)
 *                     raise TypeError('{arg!r} parameter is positional only, '
 *                                     'but was passed as a keyword'. \             # <<<<<<<<<<<<<<
 *                                     format(arg=param.name))
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_arg_r_parameter_is_positional_o, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3108, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "Aspidites/_vendor/contracts/inspection.py":3109
 *                     raise TypeError('{arg!r} parameter is positional only, '
 *                                     'but was passed as a keyword'. \
 *                                     format(arg=param.name))             # <<<<<<<<<<<<<<
 * 
 *                 arguments[param_name] = arg_val
 */
          __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3109, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3109, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_arg, __pyx_t_1) < 0) __PYX_ERR(0, 3109, __pyx_L53_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3109, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":3107
 *                     # Signature object (but let's have this check here
 *                     # to ensure correct behaviour just in case)
 *                     raise TypeError('{arg!r} parameter is positional only, '             # <<<<<<<<<<<<<<
 *                                     'but was passed as a keyword'. \
 *                                     format(arg=param.name))
 */
          __pyx_t_16 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3107, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_16, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __PYX_ERR(0, 3107, __pyx_L53_except_error)

          /* "Aspidites/_vendor/contracts/inspection.py":3103
 * 
 *             else:
 *                 if param.kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                     # This should never happen in case of a properly built
 *                     # Signature object (but let's have this check here
 */
        }

        /* "Aspidites/_vendor/contracts/inspection.py":3111
 *                                     format(arg=param.name))
 * 
 *                 arguments[param_name] = arg_val             # <<<<<<<<<<<<<<
 * 
 *         if kwargs:
 */
        if (unlikely(PyDict_SetItem(__pyx_v_arguments, __pyx_v_param_name, __pyx_v_arg_val) < 0)) __PYX_ERR(0, 3111, __pyx_L53_except_error)
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L58_try_end;
      __pyx_L51_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3092
 *             try:
 *                 arg_val = kwargs.pop(param_name)
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 # We have no value for this parameter.  It's fine though,
 *                 # if it has a default value, or it is an '*args'-like
 */
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_14) {
        __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_16, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 3092, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Aspidites/_vendor/contracts/inspection.py":3097
 *                 # parameter, left alone by the processing of positional
 *                 # arguments.
 *                 if (not partial and param.kind != _VAR_POSITIONAL and             # <<<<<<<<<<<<<<
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 */
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_partial); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3097, __pyx_L53_except_error)
        __pyx_t_11 = ((!__pyx_t_10) != 0);
        if (__pyx_t_11) {
        } else {
          __pyx_t_13 = __pyx_t_11;
          goto __pyx_L63_bool_binop_done;
        }
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3097, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3097, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_20 = PyObject_RichCompare(__pyx_t_15, __pyx_t_17, Py_NE); __Pyx_XGOTREF(__pyx_t_20); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3097, __pyx_L53_except_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_20); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3097, __pyx_L53_except_error)
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        if (__pyx_t_11) {
        } else {
          __pyx_t_13 = __pyx_t_11;
          goto __pyx_L63_bool_binop_done;
        }

        /* "Aspidites/_vendor/contracts/inspection.py":3098
 *                 # arguments.
 *                 if (not partial and param.kind != _VAR_POSITIONAL and
 *                         param.default is _empty):             # <<<<<<<<<<<<<<
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 *                                     format(arg=param_name)) from None
 */
        __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_default); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3098, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_empty); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3098, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_11 = (__pyx_t_20 == __pyx_t_17);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_10 = (__pyx_t_11 != 0);
        __pyx_t_13 = __pyx_t_10;
        __pyx_L63_bool_binop_done:;

        /* "Aspidites/_vendor/contracts/inspection.py":3097
 *                 # parameter, left alone by the processing of positional
 *                 # arguments.
 *                 if (not partial and param.kind != _VAR_POSITIONAL and             # <<<<<<<<<<<<<<
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 */
        if (unlikely(__pyx_t_13)) {

          /* "Aspidites/_vendor/contracts/inspection.py":3099
 *                 if (not partial and param.kind != _VAR_POSITIONAL and
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \             # <<<<<<<<<<<<<<
 *                                     format(arg=param_name)) from None
 * 
 */
          __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_missing_a_required_argument_arg, __pyx_n_s_format); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3099, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_17);

          /* "Aspidites/_vendor/contracts/inspection.py":3100
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 *                                     format(arg=param_name)) from None             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
          __pyx_t_20 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3100, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_arg, __pyx_v_param_name) < 0) __PYX_ERR(0, 3100, __pyx_L53_except_error)
          __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_empty_tuple, __pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3100, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":3099
 *                 if (not partial and param.kind != _VAR_POSITIONAL and
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \             # <<<<<<<<<<<<<<
 *                                     format(arg=param_name)) from None
 * 
 */
          __pyx_t_20 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3099, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "Aspidites/_vendor/contracts/inspection.py":3100
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 *                                     format(arg=param_name)) from None             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
          __Pyx_Raise(__pyx_t_20, 0, 0, Py_None);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          __PYX_ERR(0, 3099, __pyx_L53_except_error)

          /* "Aspidites/_vendor/contracts/inspection.py":3097
 *                 # parameter, left alone by the processing of positional
 *                 # arguments.
 *                 if (not partial and param.kind != _VAR_POSITIONAL and             # <<<<<<<<<<<<<<
 *                         param.default is _empty):
 *                     raise TypeError('missing a required argument: {arg!r}'. \
 */
        }
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L52_exception_handled;
      }
      goto __pyx_L53_except_error;
      __pyx_L53_except_error:;

      /* "Aspidites/_vendor/contracts/inspection.py":3090
 * 
 *             param_name = param.name
 *             try:             # <<<<<<<<<<<<<<
 *                 arg_val = kwargs.pop(param_name)
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L52_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      __pyx_L58_try_end:;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3077
 *         # keyword arguments
 *         kwargs_param = None
 *         for param in itertools.chain(parameters_ex, parameters):             # <<<<<<<<<<<<<<
 *             if param.kind == _VAR_KEYWORD:
 *                 # Memorize that we have a '**kwargs'-like parameter
 */
    __pyx_L47_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3113
 *                 arguments[param_name] = arg_val
 * 
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             if kwargs_param is not None:
 *                 # Process our '**kwargs'-like parameter
 */
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 3113, __pyx_L1_error)
  if (__pyx_t_13) {

    /* "Aspidites/_vendor/contracts/inspection.py":3114
 * 
 *         if kwargs:
 *             if kwargs_param is not None:             # <<<<<<<<<<<<<<
 *                 # Process our '**kwargs'-like parameter
 *                 arguments[kwargs_param.name] = kwargs
 */
    __pyx_t_13 = (__pyx_v_kwargs_param != Py_None);
    __pyx_t_10 = (__pyx_t_13 != 0);
    if (likely(__pyx_t_10)) {

      /* "Aspidites/_vendor/contracts/inspection.py":3116
 *             if kwargs_param is not None:
 *                 # Process our '**kwargs'-like parameter
 *                 arguments[kwargs_param.name] = kwargs             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError(
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs_param, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(PyDict_SetItem(__pyx_v_arguments, __pyx_t_3, __pyx_v_kwargs) < 0)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3114
 * 
 *         if kwargs:
 *             if kwargs_param is not None:             # <<<<<<<<<<<<<<
 *                 # Process our '**kwargs'-like parameter
 *                 arguments[kwargs_param.name] = kwargs
 */
      goto __pyx_L67;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3118
 *                 arguments[kwargs_param.name] = kwargs
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'got an unexpected keyword argument {arg!r}'.format(
 *                         arg=next(iter(kwargs))))
 */
    /*else*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":3119
 *             else:
 *                 raise TypeError(
 *                     'got an unexpected keyword argument {arg!r}'.format(             # <<<<<<<<<<<<<<
 *                         arg=next(iter(kwargs))))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_got_an_unexpected_keyword_argume, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "Aspidites/_vendor/contracts/inspection.py":3120
 *                 raise TypeError(
 *                     'got an unexpected keyword argument {arg!r}'.format(
 *                         arg=next(iter(kwargs))))             # <<<<<<<<<<<<<<
 * 
 *         return self._bound_arguments_cls(self, arguments)
 */
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyObject_GetIter(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_16 = __Pyx_PyIter_Next(__pyx_t_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_arg, __pyx_t_16) < 0) __PYX_ERR(0, 3120, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3119
 *             else:
 *                 raise TypeError(
 *                     'got an unexpected keyword argument {arg!r}'.format(             # <<<<<<<<<<<<<<
 *                         arg=next(iter(kwargs))))
 * 
 */
      __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3118
 *                 arguments[kwargs_param.name] = kwargs
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'got an unexpected keyword argument {arg!r}'.format(
 *                         arg=next(iter(kwargs))))
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 3118, __pyx_L1_error)
    }
    __pyx_L67:;

    /* "Aspidites/_vendor/contracts/inspection.py":3113
 *                 arguments[param_name] = arg_val
 * 
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             if kwargs_param is not None:
 *                 # Process our '**kwargs'-like parameter
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3122
 *                         arg=next(iter(kwargs))))
 * 
 *         return self._bound_arguments_cls(self, arguments)             # <<<<<<<<<<<<<<
 * 
 *     # def bind(self, /, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bound_arguments_cls); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_3 = NULL;
  __pyx_t_14 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
      __pyx_t_14 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_arguments};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_arguments};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_14, __pyx_v_self);
    __Pyx_INCREF(__pyx_v_arguments);
    __Pyx_GIVEREF(__pyx_v_arguments);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_14, __pyx_v_arguments);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":2993
 *         return self._hash_basis() == other._hash_basis()
 * 
 *     def _bind(self, args, kwargs, *, partial=False):             # <<<<<<<<<<<<<<
 *         """Private method. Don't use directly."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arguments);
  __Pyx_XDECREF(__pyx_v_parameters);
  __Pyx_XDECREF(__pyx_v_parameters_ex);
  __Pyx_XDECREF(__pyx_v_arg_vals);
  __Pyx_XDECREF(__pyx_v_arg_val);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_kwargs_param);
  __Pyx_XDECREF(__pyx_v_param_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3138
 *     #     return self._bind(args, kwargs, partial=True)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_23__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_23__reduce__ = {"__reduce__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_23__reduce__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_23__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_22__reduce__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_22__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3139
 * 
 *     def __reduce__(self):
 *         return (type(self),             # <<<<<<<<<<<<<<
 *                 (tuple(self._parameters.values()),),
 *                 {'_return_annotation': self._return_annotation})
 */
  __Pyx_XDECREF(__pyx_r);

  /* "Aspidites/_vendor/contracts/inspection.py":3140
 *     def __reduce__(self):
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),             # <<<<<<<<<<<<<<
 *                 {'_return_annotation': self._return_annotation})
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parameters_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3141
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),
 *                 {'_return_annotation': self._return_annotation})             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_return_annotation_2, __pyx_t_2) < 0) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3139
 * 
 *     def __reduce__(self):
 *         return (type(self),             # <<<<<<<<<<<<<<
 *                 (tuple(self._parameters.values()),),
 *                 {'_return_annotation': self._return_annotation})
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3138
 *     #     return self._bind(args, kwargs, partial=True)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3143
 *                 {'_return_annotation': self._return_annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._return_annotation = state['_return_annotation']
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_25__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_25__setstate__ = {"__setstate__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_25__setstate__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_25__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self_2,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, 1); __PYX_ERR(0, 3143, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate__") < 0)) __PYX_ERR(0, 3143, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3143, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_24__setstate__(__pyx_self, __pyx_v_self, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_24__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3144
 * 
 *     def __setstate__(self, state):
 *         self._return_annotation = state['_return_annotation']             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_u_return_annotation_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation_2, __pyx_t_1) < 0) __PYX_ERR(0, 3144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3143
 *                 {'_return_annotation': self._return_annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._return_annotation = state['_return_annotation']
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3146
 *         self._return_annotation = state['_return_annotation']
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_27__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_27__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_27__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_27__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_26__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_26__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3147
 * 
 *     def __repr__(self):
 *         return '<{} {}>'.format(self.__class__.__name__, self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__82, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3146
 *         self._return_annotation = state['_return_annotation']
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3149
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         result = []
 *         render_pos_only_separator = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_29__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_29__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_29__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_9Signature_29__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_28__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_9Signature_28__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_render_pos_only_separator;
  int __pyx_v_render_kw_only_separator;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_formatted = NULL;
  PyObject *__pyx_v_kind = NULL;
  PyObject *__pyx_v_rendered = NULL;
  PyObject *__pyx_v_anno = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3150
 * 
 *     def __str__(self):
 *         result = []             # <<<<<<<<<<<<<<
 *         render_pos_only_separator = False
 *         render_kw_only_separator = True
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3151
 *     def __str__(self):
 *         result = []
 *         render_pos_only_separator = False             # <<<<<<<<<<<<<<
 *         render_kw_only_separator = True
 *         for param in self.parameters.values():
 */
  __pyx_v_render_pos_only_separator = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3152
 *         result = []
 *         render_pos_only_separator = False
 *         render_kw_only_separator = True             # <<<<<<<<<<<<<<
 *         for param in self.parameters.values():
 *             formatted = str(param)
 */
  __pyx_v_render_kw_only_separator = 1;

  /* "Aspidites/_vendor/contracts/inspection.py":3153
 *         render_pos_only_separator = False
 *         render_kw_only_separator = True
 *         for param in self.parameters.values():             # <<<<<<<<<<<<<<
 *             formatted = str(param)
 * 
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parameters); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(0, 3153, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_values, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, NULL, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 3153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3154
 *         render_kw_only_separator = True
 *         for param in self.parameters.values():
 *             formatted = str(param)             # <<<<<<<<<<<<<<
 * 
 *             kind = param.kind
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_param); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_formatted, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3156
 *             formatted = str(param)
 * 
 *             kind = param.kind             # <<<<<<<<<<<<<<
 * 
 *             if kind == _POSITIONAL_ONLY:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_kind, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3158
 *             kind = param.kind
 * 
 *             if kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                 render_pos_only_separator = True
 *             elif render_pos_only_separator:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_kind, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":3159
 * 
 *             if kind == _POSITIONAL_ONLY:
 *                 render_pos_only_separator = True             # <<<<<<<<<<<<<<
 *             elif render_pos_only_separator:
 *                 # It's not a positional-only parameter, and the flag
 */
      __pyx_v_render_pos_only_separator = 1;

      /* "Aspidites/_vendor/contracts/inspection.py":3158
 *             kind = param.kind
 * 
 *             if kind == _POSITIONAL_ONLY:             # <<<<<<<<<<<<<<
 *                 render_pos_only_separator = True
 *             elif render_pos_only_separator:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3160
 *             if kind == _POSITIONAL_ONLY:
 *                 render_pos_only_separator = True
 *             elif render_pos_only_separator:             # <<<<<<<<<<<<<<
 *                 # It's not a positional-only parameter, and the flag
 *                 # is set to 'True' (there were pos-only params before.)
 */
    __pyx_t_8 = (__pyx_v_render_pos_only_separator != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":3163
 *                 # It's not a positional-only parameter, and the flag
 *                 # is set to 'True' (there were pos-only params before.)
 *                 result.append('/')             # <<<<<<<<<<<<<<
 *                 render_pos_only_separator = False
 * 
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_kp_u__58); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 3163, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":3164
 *                 # is set to 'True' (there were pos-only params before.)
 *                 result.append('/')
 *                 render_pos_only_separator = False             # <<<<<<<<<<<<<<
 * 
 *             if kind == _VAR_POSITIONAL:
 */
      __pyx_v_render_pos_only_separator = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3160
 *             if kind == _POSITIONAL_ONLY:
 *                 render_pos_only_separator = True
 *             elif render_pos_only_separator:             # <<<<<<<<<<<<<<
 *                 # It's not a positional-only parameter, and the flag
 *                 # is set to 'True' (there were pos-only params before.)
 */
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/contracts/inspection.py":3166
 *                 render_pos_only_separator = False
 * 
 *             if kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # OK, we have an '*args'-like parameter, so we won't need
 *                 # a '*' to separate keyword-only arguments
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_kind, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":3169
 *                 # OK, we have an '*args'-like parameter, so we won't need
 *                 # a '*' to separate keyword-only arguments
 *                 render_kw_only_separator = False             # <<<<<<<<<<<<<<
 *             elif kind == _KEYWORD_ONLY and render_kw_only_separator:
 *                 # We have a keyword-only parameter to render and we haven't
 */
      __pyx_v_render_kw_only_separator = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3166
 *                 render_pos_only_separator = False
 * 
 *             if kind == _VAR_POSITIONAL:             # <<<<<<<<<<<<<<
 *                 # OK, we have an '*args'-like parameter, so we won't need
 *                 # a '*' to separate keyword-only arguments
 */
      goto __pyx_L6;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3170
 *                 # a '*' to separate keyword-only arguments
 *                 render_kw_only_separator = False
 *             elif kind == _KEYWORD_ONLY and render_kw_only_separator:             # <<<<<<<<<<<<<<
 *                 # We have a keyword-only parameter to render and we haven't
 *                 # rendered an '*args'-like parameter before, so add a '*'
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_kind, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3170, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3170, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_10) {
    } else {
      __pyx_t_8 = __pyx_t_10;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_render_kw_only_separator != 0);
    __pyx_t_8 = __pyx_t_10;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/contracts/inspection.py":3174
 *                 # rendered an '*args'-like parameter before, so add a '*'
 *                 # separator to the parameters list ("foo(arg1, *, arg2)" case)
 *                 result.append('*')             # <<<<<<<<<<<<<<
 *                 # This condition should be only triggered once, so
 *                 # reset the flag
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_kp_u_); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 3174, __pyx_L1_error)

      /* "Aspidites/_vendor/contracts/inspection.py":3177
 *                 # This condition should be only triggered once, so
 *                 # reset the flag
 *                 render_kw_only_separator = False             # <<<<<<<<<<<<<<
 * 
 *             result.append(formatted)
 */
      __pyx_v_render_kw_only_separator = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3170
 *                 # a '*' to separate keyword-only arguments
 *                 render_kw_only_separator = False
 *             elif kind == _KEYWORD_ONLY and render_kw_only_separator:             # <<<<<<<<<<<<<<
 *                 # We have a keyword-only parameter to render and we haven't
 *                 # rendered an '*args'-like parameter before, so add a '*'
 */
    }
    __pyx_L6:;

    /* "Aspidites/_vendor/contracts/inspection.py":3179
 *                 render_kw_only_separator = False
 * 
 *             result.append(formatted)             # <<<<<<<<<<<<<<
 * 
 *         if render_pos_only_separator:
 */
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_formatted); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 3179, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3181
 *             result.append(formatted)
 * 
 *         if render_pos_only_separator:             # <<<<<<<<<<<<<<
 *             # There were only positional-only parameters, hence the
 *             # flag was not reset to 'False'
 */
  __pyx_t_8 = (__pyx_v_render_pos_only_separator != 0);
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/contracts/inspection.py":3184
 *             # There were only positional-only parameters, hence the
 *             # flag was not reset to 'False'
 *             result.append('/')             # <<<<<<<<<<<<<<
 * 
 *         rendered = '({})'.format(', '.join(result))
 */
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_kp_u__58); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 3184, __pyx_L1_error)

    /* "Aspidites/_vendor/contracts/inspection.py":3181
 *             result.append(formatted)
 * 
 *         if render_pos_only_separator:             # <<<<<<<<<<<<<<
 *             # There were only positional-only parameters, hence the
 *             # flag was not reset to 'False'
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3186
 *             result.append('/')
 * 
 *         rendered = '({})'.format(', '.join(result))             # <<<<<<<<<<<<<<
 * 
 *         if self.return_annotation is not _empty:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__83, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__41, __pyx_v_result); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_11, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_rendered = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3188
 *         rendered = '({})'.format(', '.join(result))
 * 
 *         if self.return_annotation is not _empty:             # <<<<<<<<<<<<<<
 *             anno = formatannotation(self.return_annotation)
 *             rendered += ' -> {}'.format(anno)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = (__pyx_t_1 != __pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = (__pyx_t_8 != 0);
  if (__pyx_t_10) {

    /* "Aspidites/_vendor/contracts/inspection.py":3189
 * 
 *         if self.return_annotation is not _empty:
 *             anno = formatannotation(self.return_annotation)             # <<<<<<<<<<<<<<
 *             rendered += ' -> {}'.format(anno)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_formatannotation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_return_annotation); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_11, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_anno = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3190
 *         if self.return_annotation is not _empty:
 *             anno = formatannotation(self.return_annotation)
 *             rendered += ' -> {}'.format(anno)             # <<<<<<<<<<<<<<
 * 
 *         return rendered
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__84, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_anno) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_anno);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_rendered, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_rendered, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3188
 *         rendered = '({})'.format(', '.join(result))
 * 
 *         if self.return_annotation is not _empty:             # <<<<<<<<<<<<<<
 *             anno = formatannotation(self.return_annotation)
 *             rendered += ' -> {}'.format(anno)
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3192
 *             rendered += ' -> {}'.format(anno)
 * 
 *         return rendered             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rendered);
  __pyx_r = __pyx_v_rendered;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3149
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         result = []
 *         render_pos_only_separator = False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.Signature.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_formatted);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_rendered);
  __Pyx_XDECREF(__pyx_v_anno);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3195
 * 
 * 
 * def signature(obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *     """Get a signature object for the passed callable."""
 *     return Signature.from_callable(obj, follow_wrapped=follow_wrapped)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_169signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_168signature[] = "Get a signature object for the passed callable.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_169signature = {"signature", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_169signature, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_168signature};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_169signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_follow_wrapped = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("signature (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_follow_wrapped,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 1;
        PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
        if (value) { values[index] = value; kw_args--; }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "signature") < 0)) __PYX_ERR(0, 3195, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_obj = values[0];
    __pyx_v_follow_wrapped = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("signature", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_168signature(__pyx_self, __pyx_v_obj, __pyx_v_follow_wrapped);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_168signature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_follow_wrapped) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("signature", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3197
 * def signature(obj, *, follow_wrapped=True):
 *     """Get a signature object for the passed callable."""
 *     return Signature.from_callable(obj, follow_wrapped=follow_wrapped)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Signature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_from_callable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_obj);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_follow_wrapped, __pyx_v_follow_wrapped) < 0) __PYX_ERR(0, 3197, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3195
 * 
 * 
 * def signature(obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *     """Get a signature object for the passed callable."""
 *     return Signature.from_callable(obj, follow_wrapped=follow_wrapped)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3200
 * 
 * 
 * def can_accept_exactly_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept exactly one argument
 *         using introspection.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_171can_accept_exactly_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_170can_accept_exactly_one_argument[] = " Checks that a callable can accept exactly one argument\n        using introspection.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_171can_accept_exactly_one_argument = {"can_accept_exactly_one_argument", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_171can_accept_exactly_one_argument, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_170can_accept_exactly_one_argument};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_171can_accept_exactly_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_accept_exactly_one_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_170can_accept_exactly_one_argument(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_170can_accept_exactly_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_accept_exactly_one_argument", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3204
 *         using introspection.
 *     """
 *     if ismethod(callable_thing):  # bound method             # <<<<<<<<<<<<<<
 *         f = callable_thing.__func__
 *         args = (callable_thing.__self__, 'test',)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":3205
 *     """
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__             # <<<<<<<<<<<<<<
 *         args = (callable_thing.__self__, 'test',)
 *     else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_callable_thing, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_f = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3206
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__
 *         args = (callable_thing.__self__, 'test',)             # <<<<<<<<<<<<<<
 *     else:
 *         if not isfunction(callable_thing):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_callable_thing, __pyx_n_s_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_n_u_test);
    __Pyx_GIVEREF(__pyx_n_u_test);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_test);
    __pyx_t_1 = 0;
    __pyx_v_args = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3204
 *         using introspection.
 *     """
 *     if ismethod(callable_thing):  # bound method             # <<<<<<<<<<<<<<
 *         f = callable_thing.__func__
 *         args = (callable_thing.__self__, 'test',)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3208
 *         args = (callable_thing.__self__, 'test',)
 *     else:
 *         if not isfunction(callable_thing):             # <<<<<<<<<<<<<<
 *             f = callable_thing.__call__
 *         else:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_callable_thing);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = ((!__pyx_t_4) != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":3209
 *     else:
 *         if not isfunction(callable_thing):
 *             f = callable_thing.__call__             # <<<<<<<<<<<<<<
 *         else:
 *             f = callable_thing
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_callable_thing, __pyx_n_s_call); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_f = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3208
 *         args = (callable_thing.__self__, 'test',)
 *     else:
 *         if not isfunction(callable_thing):             # <<<<<<<<<<<<<<
 *             f = callable_thing.__call__
 *         else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3211
 *             f = callable_thing.__call__
 *         else:
 *             f = callable_thing             # <<<<<<<<<<<<<<
 *         args = ('test',)
 * 
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_callable_thing);
      __pyx_v_f = __pyx_v_callable_thing;
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/contracts/inspection.py":3212
 *         else:
 *             f = callable_thing
 *         args = ('test',)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __Pyx_INCREF(__pyx_tuple__85);
    __pyx_v_args = __pyx_tuple__85;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":3214
 *         args = ('test',)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, *args)
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":3215
 * 
 *     try:
 *         getcallargs(f, *args)             # <<<<<<<<<<<<<<
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 *         # print 'Get call args exception (f=%r,args=%r): %s ' % (f, args, e)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getcallargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3215, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3215, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_f);
      __Pyx_GIVEREF(__pyx_v_f);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_f);
      __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3215, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3215, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3214
 *         args = ('test',)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, *args)
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3220
 *         return False, str(e)
 *     else:
 *         return True, None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_tuple__86);
      __pyx_r = __pyx_tuple__86;
      goto __pyx_L8_except_return;
    }
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3216
 *     try:
 *         getcallargs(f, *args)
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         # print 'Get call args exception (f=%r,args=%r): %s ' % (f, args, e)
 *         return False, str(e)
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ArityError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3216, __pyx_L7_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_builtin_TypeError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_3, __pyx_t_2);
    __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_exactly_one_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 3216, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_e = __pyx_t_3;
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":3218
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 *         # print 'Get call args exception (f=%r,args=%r): %s ' % (f, args, e)
 *         return False, str(e)             # <<<<<<<<<<<<<<
 *     else:
 *         return True, None
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3218, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3218, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(Py_False);
        __Pyx_GIVEREF(Py_False);
        PyTuple_SET_ITEM(__pyx_t_11, 0, Py_False);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_9);
        __pyx_t_9 = 0;
        __pyx_r = __pyx_t_11;
        __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L15_return;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":3216
 *     try:
 *         getcallargs(f, *args)
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         # print 'Get call args exception (f=%r,args=%r): %s ' % (f, args, e)
 *         return False, str(e)
 */
      /*finally:*/ {
        __pyx_L16_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_10 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L7_except_error;
        }
        __pyx_L15_return: {
          __pyx_t_19 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_19;
          __pyx_t_19 = 0;
          goto __pyx_L8_except_return;
        }
      }
    }
    goto __pyx_L7_except_error;
    __pyx_L7_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":3214
 *         args = ('test',)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, *args)
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3200
 * 
 * 
 * def can_accept_exactly_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept exactly one argument
 *         using introspection.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_exactly_one_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3223
 * 
 * 
 * def get_f_from_callable(callable_thing):             # <<<<<<<<<<<<<<
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_173get_f_from_callable(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_173get_f_from_callable = {"get_f_from_callable", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_173get_f_from_callable, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_173get_f_from_callable(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_f_from_callable (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_172get_f_from_callable(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_172get_f_from_callable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_f_from_callable", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3224
 * 
 * def get_f_from_callable(callable_thing):
 *     if ismethod(callable_thing):  # bound method             # <<<<<<<<<<<<<<
 *         f = callable_thing.__func__
 *         # args = (callable_thing.__self__, 'test',)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ismethod); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":3225
 * def get_f_from_callable(callable_thing):
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__             # <<<<<<<<<<<<<<
 *         # args = (callable_thing.__self__, 'test',)
 *     else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_callable_thing, __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_f = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3224
 * 
 * def get_f_from_callable(callable_thing):
 *     if ismethod(callable_thing):  # bound method             # <<<<<<<<<<<<<<
 *         f = callable_thing.__func__
 *         # args = (callable_thing.__self__, 'test',)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3228
 *         # args = (callable_thing.__self__, 'test',)
 *     else:
 *         if not isfunction(callable_thing):             # <<<<<<<<<<<<<<
 *             f = callable_thing.__call__
 *         else:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfunction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3228, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = ((!__pyx_t_4) != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/contracts/inspection.py":3229
 *     else:
 *         if not isfunction(callable_thing):
 *             f = callable_thing.__call__             # <<<<<<<<<<<<<<
 *         else:
 *             f = callable_thing
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_callable_thing, __pyx_n_s_call); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_f = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3228
 *         # args = (callable_thing.__self__, 'test',)
 *     else:
 *         if not isfunction(callable_thing):             # <<<<<<<<<<<<<<
 *             f = callable_thing.__call__
 *         else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3231
 *             f = callable_thing.__call__
 *         else:
 *             f = callable_thing             # <<<<<<<<<<<<<<
 *             # args = ('test',)
 *     return f
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_callable_thing);
      __pyx_v_f = __pyx_v_callable_thing;
    }
    __pyx_L4:;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/contracts/inspection.py":3233
 *             f = callable_thing
 *             # args = ('test',)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_f);
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3223
 * 
 * 
 * def get_f_from_callable(callable_thing):             # <<<<<<<<<<<<<<
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.get_f_from_callable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3236
 * 
 * 
 * def get_callable_fullargspec(callable_thing):             # <<<<<<<<<<<<<<
 *     f: Callable
 *     spec: FullArgSpec
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_175get_callable_fullargspec(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_175get_callable_fullargspec = {"get_callable_fullargspec", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_175get_callable_fullargspec, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_175get_callable_fullargspec(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_callable_fullargspec (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_174get_callable_fullargspec(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_174get_callable_fullargspec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_f = 0;
  PyObject *__pyx_v_spec = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_callable_fullargspec", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3239
 *     f: Callable
 *     spec: FullArgSpec
 *     f = get_f_from_callable(callable_thing)             # <<<<<<<<<<<<<<
 *     spec = getfullargspec(f)
 *     return spec
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_f_from_callable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3240
 *     spec: FullArgSpec
 *     f = get_f_from_callable(callable_thing)
 *     spec = getfullargspec(f)             # <<<<<<<<<<<<<<
 *     return spec
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getfullargspec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_f) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_f);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_spec = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3241
 *     f = get_f_from_callable(callable_thing)
 *     spec = getfullargspec(f)
 *     return spec             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_spec);
  __pyx_r = __pyx_v_spec;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3236
 * 
 * 
 * def get_callable_fullargspec(callable_thing):             # <<<<<<<<<<<<<<
 *     f: Callable
 *     spec: FullArgSpec
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.get_callable_fullargspec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_spec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3244
 * 
 * 
 * def can_accept_at_least_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept at least one argument
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_177can_accept_at_least_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_176can_accept_at_least_one_argument[] = " \n        Checks that a callable can accept at least one argument\n        using introspection.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_177can_accept_at_least_one_argument = {"can_accept_at_least_one_argument", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_177can_accept_at_least_one_argument, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_176can_accept_at_least_one_argument};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_177can_accept_at_least_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_accept_at_least_one_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_176can_accept_at_least_one_argument(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_176can_accept_at_least_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_spec = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_accept_at_least_one_argument", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3249
 *         using introspection.
 *     """
 *     spec = get_callable_fullargspec(callable_thing)             # <<<<<<<<<<<<<<
 *     return len(spec.args) > 0 or spec.varargs
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_callable_fullargspec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_spec = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3250
 *     """
 *     spec = get_callable_fullargspec(callable_thing)
 *     return len(spec.args) > 0 or spec.varargs             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 > 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_varargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3244
 * 
 * 
 * def can_accept_at_least_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept at least one argument
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_at_least_one_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_spec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3253
 * 
 * 
 * def can_accept_self(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable's first argument is self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_179can_accept_self(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_178can_accept_self[] = " \n        Checks that a callable's first argument is self\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_179can_accept_self = {"can_accept_self", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_179can_accept_self, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_178can_accept_self};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_179can_accept_self(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_accept_self (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_178can_accept_self(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_178can_accept_self(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_spec = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_accept_self", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3257
 *         Checks that a callable's first argument is self
 *     """
 *     spec = get_callable_fullargspec(callable_thing)             # <<<<<<<<<<<<<<
 * 
 *     if len(spec.args) == 0 or spec.args[0] != 'self':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_callable_fullargspec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_spec = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3259
 *     spec = get_callable_fullargspec(callable_thing)
 * 
 *     if len(spec.args) == 0 or spec.args[0] != 'self':             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((__pyx_t_5 == 0) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_self_2, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":3260
 * 
 *     if len(spec.args) == 0 or spec.args[0] != 'self':
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":3259
 *     spec = get_callable_fullargspec(callable_thing)
 * 
 *     if len(spec.args) == 0 or spec.args[0] != 'self':             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3262
 *         return False
 * 
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Aspidites/_vendor/contracts/inspection.py":3253
 * 
 * 
 * def can_accept_self(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable's first argument is self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_self", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_spec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3265
 * 
 * 
 * def can_accept_self_plus_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept exactly self plus one argument
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_181can_accept_self_plus_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_180can_accept_self_plus_one_argument[] = " \n        Checks that a callable can accept exactly self plus one argument\n        using introspection.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_181can_accept_self_plus_one_argument = {"can_accept_self_plus_one_argument", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_181can_accept_self_plus_one_argument, METH_O, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_180can_accept_self_plus_one_argument};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_181can_accept_self_plus_one_argument(PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_accept_self_plus_one_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_180can_accept_self_plus_one_argument(__pyx_self, ((PyObject *)__pyx_v_callable_thing));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_180can_accept_self_plus_one_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing) {
  PyObject *__pyx_v_spec = NULL;
  PyObject *__pyx_v_f = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_accept_self_plus_one_argument", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3271
 *     """
 * 
 *     spec = get_callable_fullargspec(callable_thing)             # <<<<<<<<<<<<<<
 *     if len(spec.args) == 0 or spec.args[0] != 'self':
 *         return False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_callable_fullargspec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_spec = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3272
 * 
 *     spec = get_callable_fullargspec(callable_thing)
 *     if len(spec.args) == 0 or spec.args[0] != 'self':             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((__pyx_t_5 == 0) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_spec, __pyx_n_s_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_self_2, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/contracts/inspection.py":3273
 *     spec = get_callable_fullargspec(callable_thing)
 *     if len(spec.args) == 0 or spec.args[0] != 'self':
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     # TODO: redo better
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Aspidites/_vendor/contracts/inspection.py":3272
 * 
 *     spec = get_callable_fullargspec(callable_thing)
 *     if len(spec.args) == 0 or spec.args[0] != 'self':             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3276
 * 
 *     # TODO: redo better
 *     f = get_f_from_callable(callable_thing)             # <<<<<<<<<<<<<<
 *     try:
 *         getcallargs(f, 'self', 'value')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_get_f_from_callable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_f = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3277
 *     # TODO: redo better
 *     f = get_f_from_callable(callable_thing)
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":3278
 *     f = get_f_from_callable(callable_thing)
 *     try:
 *         getcallargs(f, 'self', 'value')             # <<<<<<<<<<<<<<
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 *         return False
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getcallargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3278, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_f, __pyx_n_u_self_2, __pyx_n_u_value};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3278, __pyx_L6_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_f, __pyx_n_u_self_2, __pyx_n_u_value};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3278, __pyx_L6_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3278, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_f);
        __Pyx_GIVEREF(__pyx_v_f);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_f);
        __Pyx_INCREF(__pyx_n_u_self_2);
        __Pyx_GIVEREF(__pyx_n_u_self_2);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_n_u_self_2);
        __Pyx_INCREF(__pyx_n_u_value);
        __Pyx_GIVEREF(__pyx_n_u_value);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_n_u_value);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3278, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3277
 *     # TODO: redo better
 *     f = get_f_from_callable(callable_thing)
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3282
 *         return False
 *     else:
 *         return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L9_except_return;
    }
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3279
 *     try:
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         return False
 *     else:
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ArityError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3279, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_builtin_TypeError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_11);
    __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_11 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_self_plus_one_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 3279, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_e = __pyx_t_1;
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":3280
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 *         return False             # <<<<<<<<<<<<<<
 *     else:
 *         return True
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L16_return;
      }

      /* "Aspidites/_vendor/contracts/inspection.py":3279
 *     try:
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         return False
 *     else:
 */
      /*finally:*/ {
        __pyx_L16_return: {
          __pyx_t_12 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_12;
          __pyx_t_12 = 0;
          goto __pyx_L9_except_return;
        }
      }
    }
    goto __pyx_L8_except_error;
    __pyx_L8_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":3277
 *     # TODO: redo better
 *     f = get_f_from_callable(callable_thing)
 *     try:             # <<<<<<<<<<<<<<
 *         getcallargs(f, 'self', 'value')
 *     except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L9_except_return:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3265
 * 
 * 
 * def can_accept_self_plus_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept exactly self plus one argument
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.can_accept_self_plus_one_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_spec);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/contracts/inspection.py":3289
 * 
 * 
 * def check_callable_accepts_these_arguments(callable_thing, args, kwargs):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept the args and kwargs.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_183check_callable_accepts_these_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_182check_callable_accepts_these_arguments[] = " Checks that a callable can accept the args and kwargs. \n    \n        Returns either None or raises InvalidArgs. \n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_183check_callable_accepts_these_arguments = {"check_callable_accepts_these_arguments", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_183check_callable_accepts_these_arguments, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_9contracts_10inspection_182check_callable_accepts_these_arguments};
static PyObject *__pyx_pw_9Aspidites_7_vendor_9contracts_10inspection_183check_callable_accepts_these_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_callable_thing = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_callable_accepts_these_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_callable_thing,&__pyx_n_s_args,&__pyx_n_s_kwargs,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_callable_thing)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("check_callable_accepts_these_arguments", 1, 3, 3, 1); __PYX_ERR(0, 3289, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("check_callable_accepts_these_arguments", 1, 3, 3, 2); __PYX_ERR(0, 3289, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "check_callable_accepts_these_arguments") < 0)) __PYX_ERR(0, 3289, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_callable_thing = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("check_callable_accepts_these_arguments", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3289, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.check_callable_accepts_these_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_182check_callable_accepts_these_arguments(__pyx_self, __pyx_v_callable_thing, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_182check_callable_accepts_these_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callable_thing, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_f = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_bound = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_callable_accepts_these_arguments", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":3294
 *         Returns either None or raises InvalidArgs.
 *     """
 *     f = get_f_from_callable(callable_thing)             # <<<<<<<<<<<<<<
 * 
 *     # TODO: more cleanly
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_f_from_callable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_callable_thing) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_callable_thing);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3297
 * 
 *     # TODO: more cleanly
 *     try:             # <<<<<<<<<<<<<<
 *         bound = getcallargs(f, *args, **kwargs)
 *         # print('bound: %r ' % bound)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "Aspidites/_vendor/contracts/inspection.py":3298
 *     # TODO: more cleanly
 *     try:
 *         bound = getcallargs(f, *args, **kwargs)             # <<<<<<<<<<<<<<
 *         # print('bound: %r ' % bound)
 *     except (TypeError, ValueError) as e:  # @UnusedVariable
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getcallargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3298, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3298, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_f);
      __Pyx_GIVEREF(__pyx_v_f);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_f);
      __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3298, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3298, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_v_kwargs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
        __PYX_ERR(0, 3298, __pyx_L3_error)
      }
      if (likely(PyDict_CheckExact(__pyx_v_kwargs))) {
        __pyx_t_3 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3298, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_3 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_kwargs, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3298, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3298, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_bound = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/contracts/inspection.py":3297
 * 
 *     # TODO: more cleanly
 *     try:             # <<<<<<<<<<<<<<
 *         bound = getcallargs(f, *args, **kwargs)
 *         # print('bound: %r ' % bound)
 */
    }

    /* "Aspidites/_vendor/contracts/inspection.py":3305
 *         # return False
 *     else:
 *         return True             # <<<<<<<<<<<<<<
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":3300
 *         bound = getcallargs(f, *args, **kwargs)
 *         # print('bound: %r ' % bound)
 *     except (TypeError, ValueError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         # print('no!: %s' % e)
 *         raise InvalidArgs('%s does not accept %s, %s: %s' % (f, args, kwargs, e))
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.check_callable_accepts_these_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 3300, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_e = __pyx_t_3;
      /*try:*/ {

        /* "Aspidites/_vendor/contracts/inspection.py":3302
 *     except (TypeError, ValueError) as e:  # @UnusedVariable
 *         # print('no!: %s' % e)
 *         raise InvalidArgs('%s does not accept %s, %s: %s' % (f, args, kwargs, e))             # <<<<<<<<<<<<<<
 *         # return False
 *     else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_InvalidArgs); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyTuple_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = 0;
        __pyx_t_12 = 127;
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_f), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u_does_not_accept);
        __pyx_t_11 += 17;
        __Pyx_GIVEREF(__pyx_kp_u_does_not_accept);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_kp_u_does_not_accept);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_args), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u__41);
        __pyx_t_11 += 2;
        __Pyx_GIVEREF(__pyx_kp_u__41);
        PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_kp_u__41);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_kwargs), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u__37);
        __pyx_t_11 += 2;
        __Pyx_GIVEREF(__pyx_kp_u__37);
        PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_kp_u__37);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_e), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_13);
        __pyx_t_13 = 0;
        __pyx_t_13 = __Pyx_PyUnicode_Join(__pyx_t_10, 7, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3302, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 3302, __pyx_L14_error)
      }

      /* "Aspidites/_vendor/contracts/inspection.py":3300
 *         bound = getcallargs(f, *args, **kwargs)
 *         # print('bound: %r ' % bound)
 *     except (TypeError, ValueError) as e:  # @UnusedVariable             # <<<<<<<<<<<<<<
 *         # print('no!: %s' % e)
 *         raise InvalidArgs('%s does not accept %s, %s: %s' % (f, args, kwargs, e))
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __pyx_t_8 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          }
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/contracts/inspection.py":3297
 * 
 *     # TODO: more cleanly
 *     try:             # <<<<<<<<<<<<<<
 *         bound = getcallargs(f, *args, **kwargs)
 *         # print('bound: %r ' % bound)
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":3289
 * 
 * 
 * def check_callable_accepts_these_arguments(callable_thing, args, kwargs):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept the args and kwargs.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.contracts.inspection.check_callable_accepts_these_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_bound);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_metamro);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)o;
  if (p->__pyx_v_metamro) {
    e = (*v)(p->__pyx_v_metamro, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs *)o;
  tmp = ((PyObject*)p->__pyx_v_metamro);
  p->__pyx_v_metamro = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct__classify_class_attrs", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_cls);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_cls) {
    e = (*v)(p->__pyx_v_cls, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_stop);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)o;
  if (p->__pyx_v_stop) {
    e = (*v)(p->__pyx_v_stop, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap *)o;
  tmp = ((PyObject*)p->__pyx_v_stop);
  p->__pyx_v_stop = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_2_unwrap", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_all_bytecode_suffixes);
  Py_CLEAR(p->__pyx_v_filename);
  Py_CLEAR(p->__pyx_v_object);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)o;
  if (p->__pyx_v_all_bytecode_suffixes) {
    e = (*v)(p->__pyx_v_all_bytecode_suffixes, a); if (e) return e;
  }
  if (p->__pyx_v_filename) {
    e = (*v)(p->__pyx_v_filename, a); if (e) return e;
  }
  if (p->__pyx_v_object) {
    e = (*v)(p->__pyx_v_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile *)o;
  tmp = ((PyObject*)p->__pyx_v_all_bytecode_suffixes);
  p->__pyx_v_all_bytecode_suffixes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_filename);
  p->__pyx_v_filename = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_object);
  p->__pyx_v_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_3_getsourcefile", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_s);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_s);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_module);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)o;
  if (p->__pyx_v_module) {
    e = (*v)(p->__pyx_v_module, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto *)o;
  tmp = ((PyObject*)p->__pyx_v_module);
  p->__pyx_v_module = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_6_formatannotationrelativeto", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_annotations);
  Py_CLEAR(p->__pyx_v_formatannotation);
  Py_CLEAR(p->__pyx_v_formatarg);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)o;
  if (p->__pyx_v_annotations) {
    e = (*v)(p->__pyx_v_annotations, a); if (e) return e;
  }
  if (p->__pyx_v_formatannotation) {
    e = (*v)(p->__pyx_v_formatannotation, a); if (e) return e;
  }
  if (p->__pyx_v_formatarg) {
    e = (*v)(p->__pyx_v_formatarg, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec *)o;
  tmp = ((PyObject*)p->__pyx_v_annotations);
  p->__pyx_v_annotations = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_formatannotation);
  p->__pyx_v_formatannotation = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_formatarg);
  p->__pyx_v_formatarg = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_7_formatargspec", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_arg2value);
  Py_CLEAR(p->__pyx_v_assign);
  Py_CLEAR(p->__pyx_v_assigned_tuple_params);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)o;
  if (p->__pyx_v_arg2value) {
    e = (*v)(p->__pyx_v_arg2value, a); if (e) return e;
  }
  if (p->__pyx_v_assign) {
    e = (*v)(p->__pyx_v_assign, a); if (e) return e;
  }
  if (p->__pyx_v_assigned_tuple_params) {
    e = (*v)(p->__pyx_v_assigned_tuple_params, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs *)o;
  tmp = ((PyObject*)p->__pyx_v_arg2value);
  p->__pyx_v_arg2value = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_assign);
  p->__pyx_v_assign = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_assigned_tuple_params);
  p->__pyx_v_assigned_tuple_params = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_8_getcallargs", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_Parameter);
  Py_CLEAR(p->__pyx_v_RewriteSymbolics);
  Py_CLEAR(p->__pyx_v_ast);
  Py_CLEAR(p->__pyx_v_empty);
  Py_CLEAR(p->__pyx_v_invalid);
  Py_CLEAR(p->__pyx_v_kind);
  Py_CLEAR(p->__pyx_v_module_dict);
  Py_CLEAR(p->__pyx_v_parameters);
  Py_CLEAR(p->__pyx_v_parse_name);
  Py_CLEAR(p->__pyx_v_sys_module_dict);
  Py_CLEAR(p->__pyx_v_wrap_value);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)o;
  if (p->__pyx_v_Parameter) {
    e = (*v)(p->__pyx_v_Parameter, a); if (e) return e;
  }
  if (p->__pyx_v_RewriteSymbolics) {
    e = (*v)(p->__pyx_v_RewriteSymbolics, a); if (e) return e;
  }
  if (p->__pyx_v_ast) {
    e = (*v)(p->__pyx_v_ast, a); if (e) return e;
  }
  if (p->__pyx_v_empty) {
    e = (*v)(p->__pyx_v_empty, a); if (e) return e;
  }
  if (p->__pyx_v_invalid) {
    e = (*v)(p->__pyx_v_invalid, a); if (e) return e;
  }
  if (p->__pyx_v_kind) {
    e = (*v)(p->__pyx_v_kind, a); if (e) return e;
  }
  if (p->__pyx_v_module_dict) {
    e = (*v)(p->__pyx_v_module_dict, a); if (e) return e;
  }
  if (p->__pyx_v_parameters) {
    e = (*v)(p->__pyx_v_parameters, a); if (e) return e;
  }
  if (p->__pyx_v_parse_name) {
    e = (*v)(p->__pyx_v_parse_name, a); if (e) return e;
  }
  if (p->__pyx_v_sys_module_dict) {
    e = (*v)(p->__pyx_v_sys_module_dict, a); if (e) return e;
  }
  if (p->__pyx_v_wrap_value) {
    e = (*v)(p->__pyx_v_wrap_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr *)o;
  tmp = ((PyObject*)p->__pyx_v_Parameter);
  p->__pyx_v_Parameter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_RewriteSymbolics);
  p->__pyx_v_RewriteSymbolics = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_ast);
  p->__pyx_v_ast = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_empty);
  p->__pyx_v_empty = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_invalid);
  p->__pyx_v_invalid = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_kind);
  p->__pyx_v_kind = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_module_dict);
  p->__pyx_v_module_dict = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_parameters);
  p->__pyx_v_parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_parse_name);
  p->__pyx_v_parse_name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_sys_module_dict);
  p->__pyx_v_sys_module_dict = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_wrap_value);
  p->__pyx_v_wrap_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_9__signature_fromstr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_parameters);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)o;
  if (p->__pyx_v_parameters) {
    e = (*v)(p->__pyx_v_parameters, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_parameters);
  p->__pyx_v_parameters = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_10___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_param);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_param) {
    e = (*v)(p->__pyx_v_param, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_11_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_12__hash_basis", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr[--__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_param);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr[__pyx_freecount_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_param) {
    e = (*v)(p->__pyx_v_param, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.contracts.inspection.__pyx_scope_struct_13_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_inspection(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_inspection},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "inspection",
    __pyx_k_Get_useful_information_from_live, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_n_s_ABCMeta, __pyx_k_ABCMeta, sizeof(__pyx_k_ABCMeta), 0, 0, 1, 1},
  {&__pyx_kp_s_A_Signature_object_represents_th, __pyx_k_A_Signature_object_represents_th, sizeof(__pyx_k_A_Signature_object_represents_th), 0, 0, 1, 0},
  {&__pyx_kp_s_A_private_marker_used_in_Paramet, __pyx_k_A_private_marker_used_in_Paramet, sizeof(__pyx_k_A_private_marker_used_in_Paramet), 0, 0, 1, 0},
  {&__pyx_kp_u_Annotations_are_not_currently_su, __pyx_k_Annotations_are_not_currently_su, sizeof(__pyx_k_Annotations_are_not_currently_su), 0, 1, 0, 0},
  {&__pyx_n_s_ArgInfo, __pyx_k_ArgInfo, sizeof(__pyx_k_ArgInfo), 0, 0, 1, 1},
  {&__pyx_n_u_ArgInfo, __pyx_k_ArgInfo, sizeof(__pyx_k_ArgInfo), 0, 1, 0, 1},
  {&__pyx_n_s_ArgSpec, __pyx_k_ArgSpec, sizeof(__pyx_k_ArgSpec), 0, 0, 1, 1},
  {&__pyx_n_u_ArgSpec, __pyx_k_ArgSpec, sizeof(__pyx_k_ArgSpec), 0, 1, 0, 1},
  {&__pyx_n_s_Arguments, __pyx_k_Arguments, sizeof(__pyx_k_Arguments), 0, 0, 1, 1},
  {&__pyx_n_u_Arguments, __pyx_k_Arguments, sizeof(__pyx_k_Arguments), 0, 1, 0, 1},
  {&__pyx_n_s_ArityError, __pyx_k_ArityError, sizeof(__pyx_k_ArityError), 0, 0, 1, 1},
  {&__pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_k_Aspidites__vendor_contracts_insp, sizeof(__pyx_k_Aspidites__vendor_contracts_insp), 0, 0, 1, 1},
  {&__pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_k_Aspidites__vendor_contracts_insp_2, sizeof(__pyx_k_Aspidites__vendor_contracts_insp_2), 0, 0, 1, 0},
  {&__pyx_n_s_AsyncGeneratorType, __pyx_k_AsyncGeneratorType, sizeof(__pyx_k_AsyncGeneratorType), 0, 0, 1, 1},
  {&__pyx_n_s_Attribute, __pyx_k_Attribute, sizeof(__pyx_k_Attribute), 0, 0, 1, 1},
  {&__pyx_n_u_Attribute, __pyx_k_Attribute, sizeof(__pyx_k_Attribute), 0, 1, 0, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_Awaitable, __pyx_k_Awaitable, sizeof(__pyx_k_Awaitable), 0, 0, 1, 1},
  {&__pyx_n_s_BlockFinder, __pyx_k_BlockFinder, sizeof(__pyx_k_BlockFinder), 0, 0, 1, 1},
  {&__pyx_n_s_BlockFinder___init, __pyx_k_BlockFinder___init, sizeof(__pyx_k_BlockFinder___init), 0, 0, 1, 1},
  {&__pyx_n_s_BlockFinder_tokeneater, __pyx_k_BlockFinder_tokeneater, sizeof(__pyx_k_BlockFinder_tokeneater), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments, __pyx_k_BoundArguments, sizeof(__pyx_k_BoundArguments), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments___eq, __pyx_k_BoundArguments___eq, sizeof(__pyx_k_BoundArguments___eq), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments___getstate, __pyx_k_BoundArguments___getstate, sizeof(__pyx_k_BoundArguments___getstate), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments___init, __pyx_k_BoundArguments___init, sizeof(__pyx_k_BoundArguments___init), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments___repr, __pyx_k_BoundArguments___repr, sizeof(__pyx_k_BoundArguments___repr), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments___setstate, __pyx_k_BoundArguments___setstate, sizeof(__pyx_k_BoundArguments___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments_apply_defaults, __pyx_k_BoundArguments_apply_defaults, sizeof(__pyx_k_BoundArguments_apply_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments_args, __pyx_k_BoundArguments_args, sizeof(__pyx_k_BoundArguments_args), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments_kwargs, __pyx_k_BoundArguments_kwargs, sizeof(__pyx_k_BoundArguments_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_BoundArguments_signature, __pyx_k_BoundArguments_signature, sizeof(__pyx_k_BoundArguments_signature), 0, 0, 1, 1},
  {&__pyx_n_s_BuiltinFunctionType, __pyx_k_BuiltinFunctionType, sizeof(__pyx_k_BuiltinFunctionType), 0, 0, 1, 1},
  {&__pyx_n_s_BuiltinMethodType, __pyx_k_BuiltinMethodType, sizeof(__pyx_k_BuiltinMethodType), 0, 0, 1, 1},
  {&__pyx_n_u_CO, __pyx_k_CO, sizeof(__pyx_k_CO), 0, 1, 0, 1},
  {&__pyx_n_s_COMMENT, __pyx_k_COMMENT, sizeof(__pyx_k_COMMENT), 0, 0, 1, 1},
  {&__pyx_n_s_COMPILER_FLAG_NAMES, __pyx_k_COMPILER_FLAG_NAMES, sizeof(__pyx_k_COMPILER_FLAG_NAMES), 0, 0, 1, 1},
  {&__pyx_n_s_CORO_CLOSED, __pyx_k_CORO_CLOSED, sizeof(__pyx_k_CORO_CLOSED), 0, 0, 1, 1},
  {&__pyx_n_u_CORO_CLOSED, __pyx_k_CORO_CLOSED, sizeof(__pyx_k_CORO_CLOSED), 0, 1, 0, 1},
  {&__pyx_n_s_CORO_CREATED, __pyx_k_CORO_CREATED, sizeof(__pyx_k_CORO_CREATED), 0, 0, 1, 1},
  {&__pyx_n_u_CORO_CREATED, __pyx_k_CORO_CREATED, sizeof(__pyx_k_CORO_CREATED), 0, 1, 0, 1},
  {&__pyx_n_s_CORO_RUNNING, __pyx_k_CORO_RUNNING, sizeof(__pyx_k_CORO_RUNNING), 0, 0, 1, 1},
  {&__pyx_n_u_CORO_RUNNING, __pyx_k_CORO_RUNNING, sizeof(__pyx_k_CORO_RUNNING), 0, 1, 0, 1},
  {&__pyx_n_s_CORO_SUSPENDED, __pyx_k_CORO_SUSPENDED, sizeof(__pyx_k_CORO_SUSPENDED), 0, 0, 1, 1},
  {&__pyx_n_u_CORO_SUSPENDED, __pyx_k_CORO_SUSPENDED, sizeof(__pyx_k_CORO_SUSPENDED), 0, 1, 0, 1},
  {&__pyx_n_s_CO_ASYNC_GENERATOR, __pyx_k_CO_ASYNC_GENERATOR, sizeof(__pyx_k_CO_ASYNC_GENERATOR), 0, 0, 1, 1},
  {&__pyx_n_s_CO_COROUTINE, __pyx_k_CO_COROUTINE, sizeof(__pyx_k_CO_COROUTINE), 0, 0, 1, 1},
  {&__pyx_n_s_CO_GENERATOR, __pyx_k_CO_GENERATOR, sizeof(__pyx_k_CO_GENERATOR), 0, 0, 1, 1},
  {&__pyx_n_s_CO_ITERABLE_COROUTINE, __pyx_k_CO_ITERABLE_COROUTINE, sizeof(__pyx_k_CO_ITERABLE_COROUTINE), 0, 0, 1, 1},
  {&__pyx_n_s_CO_VARARGS, __pyx_k_CO_VARARGS, sizeof(__pyx_k_CO_VARARGS), 0, 0, 1, 1},
  {&__pyx_n_s_CO_VARKEYWORDS, __pyx_k_CO_VARKEYWORDS, sizeof(__pyx_k_CO_VARKEYWORDS), 0, 0, 1, 1},
  {&__pyx_n_s_Callable, __pyx_k_Callable, sizeof(__pyx_k_Callable), 0, 0, 1, 1},
  {&__pyx_n_s_ClassFinder, __pyx_k_ClassFinder, sizeof(__pyx_k_ClassFinder), 0, 0, 1, 1},
  {&__pyx_n_s_ClassFinder___init, __pyx_k_ClassFinder___init, sizeof(__pyx_k_ClassFinder___init), 0, 0, 1, 1},
  {&__pyx_n_s_ClassFinder_visit_ClassDef, __pyx_k_ClassFinder_visit_ClassDef, sizeof(__pyx_k_ClassFinder_visit_ClassDef), 0, 0, 1, 1},
  {&__pyx_n_s_ClassFinder_visit_FunctionDef, __pyx_k_ClassFinder_visit_FunctionDef, sizeof(__pyx_k_ClassFinder_visit_FunctionDef), 0, 0, 1, 1},
  {&__pyx_n_s_ClassFoundException, __pyx_k_ClassFoundException, sizeof(__pyx_k_ClassFoundException), 0, 0, 1, 1},
  {&__pyx_n_s_ClassMethodDescriptorType, __pyx_k_ClassMethodDescriptorType, sizeof(__pyx_k_ClassMethodDescriptorType), 0, 0, 1, 1},
  {&__pyx_n_s_ClassMethodWrapper, __pyx_k_ClassMethodWrapper, sizeof(__pyx_k_ClassMethodWrapper), 0, 0, 1, 1},
  {&__pyx_n_s_ClassType, __pyx_k_ClassType, sizeof(__pyx_k_ClassType), 0, 0, 1, 1},
  {&__pyx_n_s_ClosureVars, __pyx_k_ClosureVars, sizeof(__pyx_k_ClosureVars), 0, 0, 1, 1},
  {&__pyx_n_u_ClosureVars, __pyx_k_ClosureVars, sizeof(__pyx_k_ClosureVars), 0, 1, 0, 1},
  {&__pyx_n_s_CodeType, __pyx_k_CodeType, sizeof(__pyx_k_CodeType), 0, 0, 1, 1},
  {&__pyx_n_s_Constant, __pyx_k_Constant, sizeof(__pyx_k_Constant), 0, 0, 1, 1},
  {&__pyx_n_s_CoroutineType, __pyx_k_CoroutineType, sizeof(__pyx_k_CoroutineType), 0, 0, 1, 1},
  {&__pyx_n_s_DEBUG_BYTECODE_SUFFIXES, __pyx_k_DEBUG_BYTECODE_SUFFIXES, sizeof(__pyx_k_DEBUG_BYTECODE_SUFFIXES), 0, 0, 1, 1},
  {&__pyx_n_s_DEDENT, __pyx_k_DEDENT, sizeof(__pyx_k_DEDENT), 0, 0, 1, 1},
  {&__pyx_n_s_DeprecationWarning, __pyx_k_DeprecationWarning, sizeof(__pyx_k_DeprecationWarning), 0, 0, 1, 1},
  {&__pyx_n_s_DynamicClassAttribute, __pyx_k_DynamicClassAttribute, sizeof(__pyx_k_DynamicClassAttribute), 0, 0, 1, 1},
  {&__pyx_n_s_ENCODING, __pyx_k_ENCODING, sizeof(__pyx_k_ENCODING), 0, 0, 1, 1},
  {&__pyx_n_s_ERRORTOKEN, __pyx_k_ERRORTOKEN, sizeof(__pyx_k_ERRORTOKEN), 0, 0, 1, 1},
  {&__pyx_n_s_EXTENSION_SUFFIXES, __pyx_k_EXTENSION_SUFFIXES, sizeof(__pyx_k_EXTENSION_SUFFIXES), 0, 0, 1, 1},
  {&__pyx_n_s_EndOfBlock, __pyx_k_EndOfBlock, sizeof(__pyx_k_EndOfBlock), 0, 0, 1, 1},
  {&__pyx_n_u_False, __pyx_k_False, sizeof(__pyx_k_False), 0, 1, 0, 1},
  {&__pyx_n_s_FrameInfo, __pyx_k_FrameInfo, sizeof(__pyx_k_FrameInfo), 0, 0, 1, 1},
  {&__pyx_n_u_FrameInfo, __pyx_k_FrameInfo, sizeof(__pyx_k_FrameInfo), 0, 1, 0, 1},
  {&__pyx_n_s_FrameType, __pyx_k_FrameType, sizeof(__pyx_k_FrameType), 0, 0, 1, 1},
  {&__pyx_n_s_FullArgSpec, __pyx_k_FullArgSpec, sizeof(__pyx_k_FullArgSpec), 0, 0, 1, 1},
  {&__pyx_n_u_FullArgSpec, __pyx_k_FullArgSpec, sizeof(__pyx_k_FullArgSpec), 0, 1, 0, 1},
  {&__pyx_n_s_FunctionType, __pyx_k_FunctionType, sizeof(__pyx_k_FunctionType), 0, 0, 1, 1},
  {&__pyx_kp_u_Function_has_keyword_only_parame, __pyx_k_Function_has_keyword_only_parame, sizeof(__pyx_k_Function_has_keyword_only_parame), 0, 1, 0, 0},
  {&__pyx_n_s_GEN_CLOSED, __pyx_k_GEN_CLOSED, sizeof(__pyx_k_GEN_CLOSED), 0, 0, 1, 1},
  {&__pyx_n_u_GEN_CLOSED, __pyx_k_GEN_CLOSED, sizeof(__pyx_k_GEN_CLOSED), 0, 1, 0, 1},
  {&__pyx_n_s_GEN_CREATED, __pyx_k_GEN_CREATED, sizeof(__pyx_k_GEN_CREATED), 0, 0, 1, 1},
  {&__pyx_n_u_GEN_CREATED, __pyx_k_GEN_CREATED, sizeof(__pyx_k_GEN_CREATED), 0, 1, 0, 1},
  {&__pyx_n_s_GEN_RUNNING, __pyx_k_GEN_RUNNING, sizeof(__pyx_k_GEN_RUNNING), 0, 0, 1, 1},
  {&__pyx_n_u_GEN_RUNNING, __pyx_k_GEN_RUNNING, sizeof(__pyx_k_GEN_RUNNING), 0, 1, 0, 1},
  {&__pyx_n_s_GEN_SUSPENDED, __pyx_k_GEN_SUSPENDED, sizeof(__pyx_k_GEN_SUSPENDED), 0, 0, 1, 1},
  {&__pyx_n_u_GEN_SUSPENDED, __pyx_k_GEN_SUSPENDED, sizeof(__pyx_k_GEN_SUSPENDED), 0, 1, 0, 1},
  {&__pyx_n_s_GeneratorType, __pyx_k_GeneratorType, sizeof(__pyx_k_GeneratorType), 0, 0, 1, 1},
  {&__pyx_n_s_GetSetDescriptorType, __pyx_k_GetSetDescriptorType, sizeof(__pyx_k_GetSetDescriptorType), 0, 0, 1, 1},
  {&__pyx_n_u_GetSetDescriptorType, __pyx_k_GetSetDescriptorType, sizeof(__pyx_k_GetSetDescriptorType), 0, 1, 0, 1},
  {&__pyx_n_s_INDENT, __pyx_k_INDENT, sizeof(__pyx_k_INDENT), 0, 0, 1, 1},
  {&__pyx_kp_u_I_m_sorry_I_don_t_have_the_logic, __pyx_k_I_m_sorry_I_don_t_have_the_logic, sizeof(__pyx_k_I_m_sorry_I_don_t_have_the_logic), 0, 1, 0, 0},
  {&__pyx_n_s_IndentationError, __pyx_k_IndentationError, sizeof(__pyx_k_IndentationError), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_n_s_InvalidArgs, __pyx_k_InvalidArgs, sizeof(__pyx_k_InvalidArgs), 0, 0, 1, 1},
  {&__pyx_n_s_KEYWORD_ONLY, __pyx_k_KEYWORD_ONLY, sizeof(__pyx_k_KEYWORD_ONLY), 0, 0, 1, 1},
  {&__pyx_n_s_KEYWORD_ONLY_2, __pyx_k_KEYWORD_ONLY_2, sizeof(__pyx_k_KEYWORD_ONLY_2), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_Load, __pyx_k_Load, sizeof(__pyx_k_Load), 0, 0, 1, 1},
  {&__pyx_n_s_MappingProxyType, __pyx_k_MappingProxyType, sizeof(__pyx_k_MappingProxyType), 0, 0, 1, 1},
  {&__pyx_kp_s_Marker_object_for_Signature_empt, __pyx_k_Marker_object_for_Signature_empt, sizeof(__pyx_k_Marker_object_for_Signature_empt), 0, 0, 1, 0},
  {&__pyx_n_s_MemberDescriptorType, __pyx_k_MemberDescriptorType, sizeof(__pyx_k_MemberDescriptorType), 0, 0, 1, 1},
  {&__pyx_n_u_MemberDescriptorType, __pyx_k_MemberDescriptorType, sizeof(__pyx_k_MemberDescriptorType), 0, 1, 0, 1},
  {&__pyx_n_s_MethodType, __pyx_k_MethodType, sizeof(__pyx_k_MethodType), 0, 0, 1, 1},
  {&__pyx_n_s_MethodWrapper, __pyx_k_MethodWrapper, sizeof(__pyx_k_MethodWrapper), 0, 0, 1, 1},
  {&__pyx_n_s_Module, __pyx_k_Module, sizeof(__pyx_k_Module), 0, 0, 1, 1},
  {&__pyx_n_s_ModuleType, __pyx_k_ModuleType, sizeof(__pyx_k_ModuleType), 0, 0, 1, 1},
  {&__pyx_n_s_NEWLINE, __pyx_k_NEWLINE, sizeof(__pyx_k_NEWLINE), 0, 0, 1, 1},
  {&__pyx_n_s_NL, __pyx_k_NL, sizeof(__pyx_k_NL), 0, 0, 1, 1},
  {&__pyx_n_s_Name, __pyx_k_Name, sizeof(__pyx_k_Name), 0, 0, 1, 1},
  {&__pyx_n_s_NameError, __pyx_k_NameError, sizeof(__pyx_k_NameError), 0, 0, 1, 1},
  {&__pyx_n_s_NodeTransformer, __pyx_k_NodeTransformer, sizeof(__pyx_k_NodeTransformer), 0, 0, 1, 1},
  {&__pyx_n_s_NodeVisitor, __pyx_k_NodeVisitor, sizeof(__pyx_k_NodeVisitor), 0, 0, 1, 1},
  {&__pyx_n_s_NonUserDefinedCallables, __pyx_k_NonUserDefinedCallables, sizeof(__pyx_k_NonUserDefinedCallables), 0, 0, 1, 1},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_OP, __pyx_k_OP, sizeof(__pyx_k_OP), 0, 0, 1, 1},
  {&__pyx_n_s_OPTIMIZED_BYTECODE_SUFFIXES, __pyx_k_OPTIMIZED_BYTECODE_SUFFIXES, sizeof(__pyx_k_OPTIMIZED_BYTECODE_SUFFIXES), 0, 0, 1, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_PARAM_NAME_MAPPING, __pyx_k_PARAM_NAME_MAPPING, sizeof(__pyx_k_PARAM_NAME_MAPPING), 0, 0, 1, 1},
  {&__pyx_n_s_POSITIONAL_ONLY, __pyx_k_POSITIONAL_ONLY, sizeof(__pyx_k_POSITIONAL_ONLY), 0, 0, 1, 1},
  {&__pyx_n_s_POSITIONAL_ONLY_2, __pyx_k_POSITIONAL_ONLY_2, sizeof(__pyx_k_POSITIONAL_ONLY_2), 0, 0, 1, 1},
  {&__pyx_n_s_POSITIONAL_OR_KEYWORD, __pyx_k_POSITIONAL_OR_KEYWORD, sizeof(__pyx_k_POSITIONAL_OR_KEYWORD), 0, 0, 1, 1},
  {&__pyx_n_s_POSITIONAL_OR_KEYWORD_2, __pyx_k_POSITIONAL_OR_KEYWORD_2, sizeof(__pyx_k_POSITIONAL_OR_KEYWORD_2), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter, __pyx_k_Parameter, sizeof(__pyx_k_Parameter), 0, 0, 1, 1},
  {&__pyx_n_s_ParameterKind, __pyx_k_ParameterKind, sizeof(__pyx_k_ParameterKind), 0, 0, 1, 1},
  {&__pyx_n_s_ParameterKind___str, __pyx_k_ParameterKind___str, sizeof(__pyx_k_ParameterKind___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParameterKind_description, __pyx_k_ParameterKind_description, sizeof(__pyx_k_ParameterKind_description), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___eq, __pyx_k_Parameter___eq, sizeof(__pyx_k_Parameter___eq), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___hash, __pyx_k_Parameter___hash, sizeof(__pyx_k_Parameter___hash), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___init, __pyx_k_Parameter___init, sizeof(__pyx_k_Parameter___init), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___reduce, __pyx_k_Parameter___reduce, sizeof(__pyx_k_Parameter___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___repr, __pyx_k_Parameter___repr, sizeof(__pyx_k_Parameter___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___setstate, __pyx_k_Parameter___setstate, sizeof(__pyx_k_Parameter___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter___str, __pyx_k_Parameter___str, sizeof(__pyx_k_Parameter___str), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter_annotation, __pyx_k_Parameter_annotation, sizeof(__pyx_k_Parameter_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter_default, __pyx_k_Parameter_default, sizeof(__pyx_k_Parameter_default), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter_kind, __pyx_k_Parameter_kind, sizeof(__pyx_k_Parameter_kind), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter_name, __pyx_k_Parameter_name, sizeof(__pyx_k_Parameter_name), 0, 0, 1, 1},
  {&__pyx_n_s_Parameter_replace, __pyx_k_Parameter_replace, sizeof(__pyx_k_Parameter_replace), 0, 0, 1, 1},
  {&__pyx_kp_s_Provide_a_tokeneater_method_to_d, __pyx_k_Provide_a_tokeneater_method_to_d, sizeof(__pyx_k_Provide_a_tokeneater_method_to_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Represents_a_parameter_in_a_func, __pyx_k_Represents_a_parameter_in_a_func, sizeof(__pyx_k_Represents_a_parameter_in_a_func), 0, 0, 1, 0},
  {&__pyx_kp_s_Result_of_Signature_bind_call_Ho, __pyx_k_Result_of_Signature_bind_call_Ho, sizeof(__pyx_k_Result_of_Signature_bind_call_Ho), 0, 0, 1, 0},
  {&__pyx_n_s_RewriteSymbolics, __pyx_k_RewriteSymbolics, sizeof(__pyx_k_RewriteSymbolics), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SOURCE_SUFFIXES, __pyx_k_SOURCE_SUFFIXES, sizeof(__pyx_k_SOURCE_SUFFIXES), 0, 0, 1, 1},
  {&__pyx_n_s_Signature, __pyx_k_Signature, sizeof(__pyx_k_Signature), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___eq, __pyx_k_Signature___eq, sizeof(__pyx_k_Signature___eq), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___hash, __pyx_k_Signature___hash, sizeof(__pyx_k_Signature___hash), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___init, __pyx_k_Signature___init, sizeof(__pyx_k_Signature___init), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___init___locals_genexp, __pyx_k_Signature___init___locals_genexp, sizeof(__pyx_k_Signature___init___locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___reduce, __pyx_k_Signature___reduce, sizeof(__pyx_k_Signature___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___repr, __pyx_k_Signature___repr, sizeof(__pyx_k_Signature___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___setstate, __pyx_k_Signature___setstate, sizeof(__pyx_k_Signature___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_Signature___str, __pyx_k_Signature___str, sizeof(__pyx_k_Signature___str), 0, 0, 1, 1},
  {&__pyx_n_s_Signature__bind, __pyx_k_Signature__bind, sizeof(__pyx_k_Signature__bind), 0, 0, 1, 1},
  {&__pyx_n_s_Signature__hash_basis, __pyx_k_Signature__hash_basis, sizeof(__pyx_k_Signature__hash_basis), 0, 0, 1, 1},
  {&__pyx_n_s_Signature__hash_basis_locals_gen, __pyx_k_Signature__hash_basis_locals_gen, sizeof(__pyx_k_Signature__hash_basis_locals_gen), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_from_builtin, __pyx_k_Signature_from_builtin, sizeof(__pyx_k_Signature_from_builtin), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_from_callable, __pyx_k_Signature_from_callable, sizeof(__pyx_k_Signature_from_callable), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_from_function, __pyx_k_Signature_from_function, sizeof(__pyx_k_Signature_from_function), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_parameters, __pyx_k_Signature_parameters, sizeof(__pyx_k_Signature_parameters), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_replace, __pyx_k_Signature_replace, sizeof(__pyx_k_Signature_replace), 0, 0, 1, 1},
  {&__pyx_n_s_Signature_return_annotation, __pyx_k_Signature_return_annotation, sizeof(__pyx_k_Signature_return_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_SyntaxError, __pyx_k_SyntaxError, sizeof(__pyx_k_SyntaxError), 0, 0, 1, 1},
  {&__pyx_n_s_TPFLAGS_IS_ABSTRACT, __pyx_k_TPFLAGS_IS_ABSTRACT, sizeof(__pyx_k_TPFLAGS_IS_ABSTRACT), 0, 0, 1, 1},
  {&__pyx_n_s_Traceback, __pyx_k_Traceback, sizeof(__pyx_k_Traceback), 0, 0, 1, 1},
  {&__pyx_n_u_Traceback, __pyx_k_Traceback, sizeof(__pyx_k_Traceback), 0, 1, 0, 1},
  {&__pyx_n_s_TracebackType, __pyx_k_TracebackType, sizeof(__pyx_k_TracebackType), 0, 0, 1, 1},
  {&__pyx_n_u_True, __pyx_k_True, sizeof(__pyx_k_True), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UnicodeError, __pyx_k_UnicodeError, sizeof(__pyx_k_UnicodeError), 0, 0, 1, 1},
  {&__pyx_n_s_VAR_KEYWORD, __pyx_k_VAR_KEYWORD, sizeof(__pyx_k_VAR_KEYWORD), 0, 0, 1, 1},
  {&__pyx_n_s_VAR_KEYWORD_2, __pyx_k_VAR_KEYWORD_2, sizeof(__pyx_k_VAR_KEYWORD_2), 0, 0, 1, 1},
  {&__pyx_n_s_VAR_POSITIONAL, __pyx_k_VAR_POSITIONAL, sizeof(__pyx_k_VAR_POSITIONAL), 0, 0, 1, 1},
  {&__pyx_n_s_VAR_POSITIONAL_2, __pyx_k_VAR_POSITIONAL_2, sizeof(__pyx_k_VAR_POSITIONAL_2), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WrapperDescriptor, __pyx_k_WrapperDescriptor, sizeof(__pyx_k_WrapperDescriptor), 0, 0, 1, 1},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
  {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
  {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
  {&__pyx_kp_u__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 1, 0, 0},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_kp_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 0},
  {&__pyx_kp_u__26, __pyx_k__26, sizeof(__pyx_k__26), 0, 1, 0, 0},
  {&__pyx_kp_u__28, __pyx_k__28, sizeof(__pyx_k__28), 0, 1, 0, 0},
  {&__pyx_kp_u__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 1, 0, 0},
  {&__pyx_kp_u__37, __pyx_k__37, sizeof(__pyx_k__37), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__41, __pyx_k__41, sizeof(__pyx_k__41), 0, 1, 0, 0},
  {&__pyx_kp_u__44, __pyx_k__44, sizeof(__pyx_k__44), 0, 1, 0, 0},
  {&__pyx_kp_u__51, __pyx_k__51, sizeof(__pyx_k__51), 0, 1, 0, 0},
  {&__pyx_kp_u__52, __pyx_k__52, sizeof(__pyx_k__52), 0, 1, 0, 0},
  {&__pyx_kp_u__55, __pyx_k__55, sizeof(__pyx_k__55), 0, 1, 0, 0},
  {&__pyx_kp_u__56, __pyx_k__56, sizeof(__pyx_k__56), 0, 1, 0, 0},
  {&__pyx_kp_u__57, __pyx_k__57, sizeof(__pyx_k__57), 0, 1, 0, 0},
  {&__pyx_kp_u__58, __pyx_k__58, sizeof(__pyx_k__58), 0, 1, 0, 0},
  {&__pyx_kp_u__59, __pyx_k__59, sizeof(__pyx_k__59), 0, 1, 0, 0},
  {&__pyx_kp_u__73, __pyx_k__73, sizeof(__pyx_k__73), 0, 1, 0, 0},
  {&__pyx_kp_u__74, __pyx_k__74, sizeof(__pyx_k__74), 0, 1, 0, 0},
  {&__pyx_kp_u__75, __pyx_k__75, sizeof(__pyx_k__75), 0, 1, 0, 0},
  {&__pyx_kp_u__76, __pyx_k__76, sizeof(__pyx_k__76), 0, 1, 0, 0},
  {&__pyx_kp_u__77, __pyx_k__77, sizeof(__pyx_k__77), 0, 1, 0, 0},
  {&__pyx_kp_u__82, __pyx_k__82, sizeof(__pyx_k__82), 0, 1, 0, 0},
  {&__pyx_kp_u__83, __pyx_k__83, sizeof(__pyx_k__83), 0, 1, 0, 0},
  {&__pyx_kp_u__84, __pyx_k__84, sizeof(__pyx_k__84), 0, 1, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_abc, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
  {&__pyx_n_s_abspath, __pyx_k_abspath, sizeof(__pyx_k_abspath), 0, 0, 1, 1},
  {&__pyx_n_u_abstractmethods, __pyx_k_abstractmethods, sizeof(__pyx_k_abstractmethods), 0, 1, 0, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_all_bases, __pyx_k_all_bases, sizeof(__pyx_k_all_bases), 0, 0, 1, 1},
  {&__pyx_n_s_all_bytecode_suffixes, __pyx_k_all_bytecode_suffixes, sizeof(__pyx_k_all_bytecode_suffixes), 0, 0, 1, 1},
  {&__pyx_n_s_all_suffixes, __pyx_k_all_suffixes, sizeof(__pyx_k_all_suffixes), 0, 0, 1, 1},
  {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
  {&__pyx_kp_u_and_2, __pyx_k_and_2, sizeof(__pyx_k_and_2), 0, 1, 0, 0},
  {&__pyx_n_s_ann, __pyx_k_ann, sizeof(__pyx_k_ann), 0, 0, 1, 1},
  {&__pyx_n_s_anno, __pyx_k_anno, sizeof(__pyx_k_anno), 0, 0, 1, 1},
  {&__pyx_n_s_annotation, __pyx_k_annotation, sizeof(__pyx_k_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_annotation_2, __pyx_k_annotation_2, sizeof(__pyx_k_annotation_2), 0, 0, 1, 1},
  {&__pyx_n_u_annotation_2, __pyx_k_annotation_2, sizeof(__pyx_k_annotation_2), 0, 1, 0, 1},
  {&__pyx_n_s_annotations, __pyx_k_annotations, sizeof(__pyx_k_annotations), 0, 0, 1, 1},
  {&__pyx_n_s_annotations_2, __pyx_k_annotations_2, sizeof(__pyx_k_annotations_2), 0, 0, 1, 1},
  {&__pyx_n_u_annotations_2, __pyx_k_annotations_2, sizeof(__pyx_k_annotations_2), 0, 1, 0, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_apply_defaults, __pyx_k_apply_defaults, sizeof(__pyx_k_apply_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_arg, __pyx_k_arg, sizeof(__pyx_k_arg), 0, 0, 1, 1},
  {&__pyx_n_s_arg2dict, __pyx_k_arg2dict, sizeof(__pyx_k_arg2dict), 0, 0, 1, 1},
  {&__pyx_n_s_arg2value, __pyx_k_arg2value, sizeof(__pyx_k_arg2value), 0, 0, 1, 1},
  {&__pyx_n_s_arg_names, __pyx_k_arg_names, sizeof(__pyx_k_arg_names), 0, 0, 1, 1},
  {&__pyx_kp_u_arg_r_parameter_is_positional_o, __pyx_k_arg_r_parameter_is_positional_o, sizeof(__pyx_k_arg_r_parameter_is_positional_o), 0, 1, 0, 0},
  {&__pyx_n_s_arg_val, __pyx_k_arg_val, sizeof(__pyx_k_arg_val), 0, 0, 1, 1},
  {&__pyx_n_s_arg_vals, __pyx_k_arg_vals, sizeof(__pyx_k_arg_vals), 0, 0, 1, 1},
  {&__pyx_n_s_arg_value, __pyx_k_arg_value, sizeof(__pyx_k_arg_value), 0, 0, 1, 1},
  {&__pyx_n_s_argnames, __pyx_k_argnames, sizeof(__pyx_k_argnames), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_kp_u_args_varargs_keywords_defaults, __pyx_k_args_varargs_keywords_defaults, sizeof(__pyx_k_args_varargs_keywords_defaults), 0, 1, 0, 0},
  {&__pyx_kp_u_args_varargs_keywords_locals, __pyx_k_args_varargs_keywords_locals, sizeof(__pyx_k_args_varargs_keywords_locals), 0, 1, 0, 0},
  {&__pyx_kp_u_args_varargs_varkw, __pyx_k_args_varargs_varkw, sizeof(__pyx_k_args_varargs_varkw), 0, 1, 0, 0},
  {&__pyx_kp_u_args_varargs_varkw_defaults_kwon, __pyx_k_args_varargs_varkw_defaults_kwon, sizeof(__pyx_k_args_varargs_varkw_defaults_kwon), 0, 1, 0, 0},
  {&__pyx_n_u_argument, __pyx_k_argument, sizeof(__pyx_k_argument), 0, 1, 0, 1},
  {&__pyx_n_s_arguments, __pyx_k_arguments, sizeof(__pyx_k_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_arguments, __pyx_k_arguments, sizeof(__pyx_k_arguments), 0, 1, 0, 1},
  {&__pyx_n_u_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 1, 0, 1},
  {&__pyx_n_s_assign, __pyx_k_assign, sizeof(__pyx_k_assign), 0, 0, 1, 1},
  {&__pyx_n_s_assigned_tuple_params, __pyx_k_assigned_tuple_params, sizeof(__pyx_k_assigned_tuple_params), 0, 0, 1, 1},
  {&__pyx_n_s_ast, __pyx_k_ast, sizeof(__pyx_k_ast), 0, 0, 1, 1},
  {&__pyx_kp_u_at_least, __pyx_k_at_least, sizeof(__pyx_k_at_least), 0, 1, 0, 0},
  {&__pyx_kp_u_at_least_2, __pyx_k_at_least_2, sizeof(__pyx_k_at_least_2), 0, 1, 0, 0},
  {&__pyx_kp_u_at_most, __pyx_k_at_most, sizeof(__pyx_k_at_most), 0, 1, 0, 0},
  {&__pyx_n_s_atleast, __pyx_k_atleast, sizeof(__pyx_k_atleast), 0, 0, 1, 1},
  {&__pyx_n_s_attr, __pyx_k_attr, sizeof(__pyx_k_attr), 0, 0, 1, 1},
  {&__pyx_n_s_attrgetter, __pyx_k_attrgetter, sizeof(__pyx_k_attrgetter), 0, 0, 1, 1},
  {&__pyx_n_s_ba, __pyx_k_ba, sizeof(__pyx_k_ba), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_base_module, __pyx_k_base_module, sizeof(__pyx_k_base_module), 0, 0, 1, 1},
  {&__pyx_n_s_basename, __pyx_k_basename, sizeof(__pyx_k_basename), 0, 0, 1, 1},
  {&__pyx_n_s_bases, __pyx_k_bases, sizeof(__pyx_k_bases), 0, 0, 1, 1},
  {&__pyx_n_s_bind, __pyx_k_bind, sizeof(__pyx_k_bind), 0, 0, 1, 1},
  {&__pyx_n_s_bind_partial, __pyx_k_bind_partial, sizeof(__pyx_k_bind_partial), 0, 0, 1, 1},
  {&__pyx_n_s_blockfinder, __pyx_k_blockfinder, sizeof(__pyx_k_blockfinder), 0, 0, 1, 1},
  {&__pyx_n_s_body, __pyx_k_body, sizeof(__pyx_k_body), 0, 0, 1, 1},
  {&__pyx_n_s_body_col0, __pyx_k_body_col0, sizeof(__pyx_k_body_col0), 0, 0, 1, 1},
  {&__pyx_n_s_bound, __pyx_k_bound, sizeof(__pyx_k_bound), 0, 0, 1, 1},
  {&__pyx_n_s_bound_arguments_cls, __pyx_k_bound_arguments_cls, sizeof(__pyx_k_bound_arguments_cls), 0, 0, 1, 1},
  {&__pyx_n_s_builtin, __pyx_k_builtin, sizeof(__pyx_k_builtin), 0, 0, 1, 1},
  {&__pyx_n_s_builtin_ns, __pyx_k_builtin_ns, sizeof(__pyx_k_builtin_ns), 0, 0, 1, 1},
  {&__pyx_n_s_builtin_vars, __pyx_k_builtin_vars, sizeof(__pyx_k_builtin_vars), 0, 0, 1, 1},
  {&__pyx_n_s_builtinobject, __pyx_k_builtinobject, sizeof(__pyx_k_builtinobject), 0, 0, 1, 1},
  {&__pyx_n_s_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
  {&__pyx_n_u_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 1, 0, 1},
  {&__pyx_n_u_builtins_2, __pyx_k_builtins_2, sizeof(__pyx_k_builtins_2), 0, 1, 0, 1},
  {&__pyx_kp_u_but, __pyx_k_but, sizeof(__pyx_k_but), 0, 1, 0, 0},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
  {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {&__pyx_n_u_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 1, 0, 1},
  {&__pyx_n_s_call_2, __pyx_k_call_2, sizeof(__pyx_k_call_2), 0, 0, 1, 1},
  {&__pyx_kp_u_callable_r_is_not_supported_by_s, __pyx_k_callable_r_is_not_supported_by_s, sizeof(__pyx_k_callable_r_is_not_supported_by_s), 0, 1, 0, 0},
  {&__pyx_n_s_callable_thing, __pyx_k_callable_thing, sizeof(__pyx_k_callable_thing), 0, 0, 1, 1},
  {&__pyx_n_s_can_accept_at_least_one_argument, __pyx_k_can_accept_at_least_one_argument, sizeof(__pyx_k_can_accept_at_least_one_argument), 0, 0, 1, 1},
  {&__pyx_n_s_can_accept_exactly_one_argument, __pyx_k_can_accept_exactly_one_argument, sizeof(__pyx_k_can_accept_exactly_one_argument), 0, 0, 1, 1},
  {&__pyx_n_s_can_accept_self, __pyx_k_can_accept_self, sizeof(__pyx_k_can_accept_self), 0, 0, 1, 1},
  {&__pyx_n_s_can_accept_self_plus_one_argumen, __pyx_k_can_accept_self_plus_one_argumen, sizeof(__pyx_k_can_accept_self_plus_one_argumen), 0, 0, 1, 1},
  {&__pyx_n_s_cell, __pyx_k_cell, sizeof(__pyx_k_cell), 0, 0, 1, 1},
  {&__pyx_n_s_cell_contents, __pyx_k_cell_contents, sizeof(__pyx_k_cell_contents), 0, 0, 1, 1},
  {&__pyx_n_s_chain, __pyx_k_chain, sizeof(__pyx_k_chain), 0, 0, 1, 1},
  {&__pyx_n_s_check_callable_accepts_these_arg, __pyx_k_check_callable_accepts_these_arg, sizeof(__pyx_k_check_callable_accepts_these_arg), 0, 0, 1, 1},
  {&__pyx_n_s_check_class, __pyx_k_check_class, sizeof(__pyx_k_check_class), 0, 0, 1, 1},
  {&__pyx_n_s_check_instance, __pyx_k_check_instance, sizeof(__pyx_k_check_instance), 0, 0, 1, 1},
  {&__pyx_n_s_checkcache, __pyx_k_checkcache, sizeof(__pyx_k_checkcache), 0, 0, 1, 1},
  {&__pyx_n_s_children, __pyx_k_children, sizeof(__pyx_k_children), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_u_class_2, __pyx_k_class_2, sizeof(__pyx_k_class_2), 0, 1, 0, 1},
  {&__pyx_n_s_class_bases, __pyx_k_class_bases, sizeof(__pyx_k_class_bases), 0, 0, 1, 1},
  {&__pyx_n_s_class_dict, __pyx_k_class_dict, sizeof(__pyx_k_class_dict), 0, 0, 1, 1},
  {&__pyx_n_s_class_finder, __pyx_k_class_finder, sizeof(__pyx_k_class_finder), 0, 0, 1, 1},
  {&__pyx_kp_u_class_method, __pyx_k_class_method, sizeof(__pyx_k_class_method), 0, 1, 0, 0},
  {&__pyx_n_s_classes, __pyx_k_classes, sizeof(__pyx_k_classes), 0, 0, 1, 1},
  {&__pyx_n_s_classify_class_attrs, __pyx_k_classify_class_attrs, sizeof(__pyx_k_classify_class_attrs), 0, 0, 1, 1},
  {&__pyx_n_s_classify_class_attrs_locals_gene, __pyx_k_classify_class_attrs_locals_gene, sizeof(__pyx_k_classify_class_attrs_locals_gene), 0, 0, 1, 1},
  {&__pyx_n_s_classmethod, __pyx_k_classmethod, sizeof(__pyx_k_classmethod), 0, 0, 1, 1},
  {&__pyx_n_s_clean_signature, __pyx_k_clean_signature, sizeof(__pyx_k_clean_signature), 0, 0, 1, 1},
  {&__pyx_n_s_cleandoc, __pyx_k_cleandoc, sizeof(__pyx_k_cleandoc), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_closure, __pyx_k_closure, sizeof(__pyx_k_closure), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_co, __pyx_k_co, sizeof(__pyx_k_co), 0, 0, 1, 1},
  {&__pyx_n_s_co_argcount, __pyx_k_co_argcount, sizeof(__pyx_k_co_argcount), 0, 0, 1, 1},
  {&__pyx_n_s_co_filename, __pyx_k_co_filename, sizeof(__pyx_k_co_filename), 0, 0, 1, 1},
  {&__pyx_n_s_co_firstlineno, __pyx_k_co_firstlineno, sizeof(__pyx_k_co_firstlineno), 0, 0, 1, 1},
  {&__pyx_n_u_co_firstlineno, __pyx_k_co_firstlineno, sizeof(__pyx_k_co_firstlineno), 0, 1, 0, 1},
  {&__pyx_n_s_co_flags, __pyx_k_co_flags, sizeof(__pyx_k_co_flags), 0, 0, 1, 1},
  {&__pyx_n_s_co_freevars, __pyx_k_co_freevars, sizeof(__pyx_k_co_freevars), 0, 0, 1, 1},
  {&__pyx_n_s_co_kwonlyargcount, __pyx_k_co_kwonlyargcount, sizeof(__pyx_k_co_kwonlyargcount), 0, 0, 1, 1},
  {&__pyx_n_s_co_name, __pyx_k_co_name, sizeof(__pyx_k_co_name), 0, 0, 1, 1},
  {&__pyx_n_s_co_names, __pyx_k_co_names, sizeof(__pyx_k_co_names), 0, 0, 1, 1},
  {&__pyx_n_s_co_posonlyargcount, __pyx_k_co_posonlyargcount, sizeof(__pyx_k_co_posonlyargcount), 0, 0, 1, 1},
  {&__pyx_n_s_co_varnames, __pyx_k_co_varnames, sizeof(__pyx_k_co_varnames), 0, 0, 1, 1},
  {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {&__pyx_n_u_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 1, 0, 1},
  {&__pyx_n_s_code_2, __pyx_k_code_2, sizeof(__pyx_k_code_2), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
  {&__pyx_n_s_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 0, 1, 1},
  {&__pyx_n_s_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_content, __pyx_k_content, sizeof(__pyx_k_content), 0, 0, 1, 1},
  {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
  {&__pyx_n_s_convert, __pyx_k_convert, sizeof(__pyx_k_convert), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_coroutine, __pyx_k_coroutine, sizeof(__pyx_k_coroutine), 0, 0, 1, 1},
  {&__pyx_kp_u_could_not_find_class_definition, __pyx_k_could_not_find_class_definition, sizeof(__pyx_k_could_not_find_class_definition), 0, 1, 0, 0},
  {&__pyx_kp_u_could_not_find_code_object, __pyx_k_could_not_find_code_object, sizeof(__pyx_k_could_not_find_code_object), 0, 1, 0, 0},
  {&__pyx_kp_u_could_not_find_function_definiti, __pyx_k_could_not_find_function_definiti, sizeof(__pyx_k_could_not_find_function_definiti), 0, 1, 0, 0},
  {&__pyx_kp_u_could_not_get_source_code, __pyx_k_could_not_get_source_code, sizeof(__pyx_k_could_not_get_source_code), 0, 1, 0, 0},
  {&__pyx_n_s_cpos, __pyx_k_cpos, sizeof(__pyx_k_cpos), 0, 0, 1, 1},
  {&__pyx_n_s_cr_frame, __pyx_k_cr_frame, sizeof(__pyx_k_cr_frame), 0, 0, 1, 1},
  {&__pyx_n_u_cr_frame, __pyx_k_cr_frame, sizeof(__pyx_k_cr_frame), 0, 1, 0, 1},
  {&__pyx_n_s_cr_running, __pyx_k_cr_running, sizeof(__pyx_k_cr_running), 0, 0, 1, 1},
  {&__pyx_n_s_ctx, __pyx_k_ctx, sizeof(__pyx_k_ctx), 0, 0, 1, 1},
  {&__pyx_n_s_current_parameter, __pyx_k_current_parameter, sizeof(__pyx_k_current_parameter), 0, 0, 1, 1},
  {&__pyx_n_s_currentframe, __pyx_k_currentframe, sizeof(__pyx_k_currentframe), 0, 0, 1, 1},
  {&__pyx_n_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 1},
  {&__pyx_n_u_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 1, 0, 1},
  {&__pyx_n_s_decorator_list, __pyx_k_decorator_list, sizeof(__pyx_k_decorator_list), 0, 0, 1, 1},
  {&__pyx_n_s_decoratorhasargs, __pyx_k_decoratorhasargs, sizeof(__pyx_k_decoratorhasargs), 0, 0, 1, 1},
  {&__pyx_n_u_def, __pyx_k_def, sizeof(__pyx_k_def), 0, 1, 0, 1},
  {&__pyx_kp_u_def_foo, __pyx_k_def_foo, sizeof(__pyx_k_def_foo), 0, 1, 0, 0},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_default_2, __pyx_k_default_2, sizeof(__pyx_k_default_2), 0, 0, 1, 1},
  {&__pyx_n_u_default_2, __pyx_k_default_2, sizeof(__pyx_k_default_2), 0, 1, 0, 1},
  {&__pyx_n_s_default_node, __pyx_k_default_node, sizeof(__pyx_k_default_node), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_defaults_2, __pyx_k_defaults_2, sizeof(__pyx_k_defaults_2), 0, 0, 1, 1},
  {&__pyx_n_u_defaults_2, __pyx_k_defaults_2, sizeof(__pyx_k_defaults_2), 0, 1, 0, 1},
  {&__pyx_n_s_defcount, __pyx_k_defcount, sizeof(__pyx_k_defcount), 0, 0, 1, 1},
  {&__pyx_n_s_delayed_comma, __pyx_k_delayed_comma, sizeof(__pyx_k_delayed_comma), 0, 0, 1, 1},
  {&__pyx_n_u_delete, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 1, 0, 1},
  {&__pyx_n_s_description, __pyx_k_description, sizeof(__pyx_k_description), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_u_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 1, 0, 1},
  {&__pyx_kp_u_dict___is_special_don_t_want_t, __pyx_k_dict___is_special_don_t_want_t, sizeof(__pyx_k_dict___is_special_don_t_want_t), 0, 1, 0, 0},
  {&__pyx_n_s_dict_attr, __pyx_k_dict_attr, sizeof(__pyx_k_dict_attr), 0, 0, 1, 1},
  {&__pyx_n_s_dict_obj, __pyx_k_dict_obj, sizeof(__pyx_k_dict_obj), 0, 0, 1, 1},
  {&__pyx_n_s_dis, __pyx_k_dis, sizeof(__pyx_k_dis), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_doc_2, __pyx_k_doc_2, sizeof(__pyx_k_doc_2), 0, 0, 1, 1},
  {&__pyx_kp_u_does_not_accept, __pyx_k_does_not_accept, sizeof(__pyx_k_does_not_accept), 0, 1, 0, 0},
  {&__pyx_kp_u_duplicate_parameter_name_r, __pyx_k_duplicate_parameter_name_r, sizeof(__pyx_k_duplicate_parameter_name_r), 0, 1, 0, 0},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_empty_2, __pyx_k_empty_2, sizeof(__pyx_k_empty_2), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_endswith, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {&__pyx_n_s_entry, __pyx_k_entry, sizeof(__pyx_k_entry), 0, 0, 1, 1},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_erowcol, __pyx_k_erowcol, sizeof(__pyx_k_erowcol), 0, 0, 1, 1},
  {&__pyx_n_s_eval, __pyx_k_eval, sizeof(__pyx_k_eval), 0, 0, 1, 1},
  {&__pyx_n_s_ex, __pyx_k_ex, sizeof(__pyx_k_ex), 0, 0, 1, 1},
  {&__pyx_n_u_exactly, __pyx_k_exactly, sizeof(__pyx_k_exactly), 0, 1, 0, 1},
  {&__pyx_n_s_exc, __pyx_k_exc, sizeof(__pyx_k_exc), 0, 0, 1, 1},
  {&__pyx_n_s_exc_info, __pyx_k_exc_info, sizeof(__pyx_k_exc_info), 0, 0, 1, 1},
  {&__pyx_n_s_exists, __pyx_k_exists, sizeof(__pyx_k_exists), 0, 0, 1, 1},
  {&__pyx_n_s_expandtabs, __pyx_k_expandtabs, sizeof(__pyx_k_expandtabs), 0, 0, 1, 1},
  {&__pyx_n_s_expline, __pyx_k_expline, sizeof(__pyx_k_expline), 0, 0, 1, 1},
  {&__pyx_n_s_extra_args, __pyx_k_extra_args, sizeof(__pyx_k_extra_args), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_f_back, __pyx_k_f_back, sizeof(__pyx_k_f_back), 0, 0, 1, 1},
  {&__pyx_n_s_f_code, __pyx_k_f_code, sizeof(__pyx_k_f_code), 0, 0, 1, 1},
  {&__pyx_n_s_f_lasti, __pyx_k_f_lasti, sizeof(__pyx_k_f_lasti), 0, 0, 1, 1},
  {&__pyx_n_s_f_lineno, __pyx_k_f_lineno, sizeof(__pyx_k_f_lineno), 0, 0, 1, 1},
  {&__pyx_n_s_f_locals, __pyx_k_f_locals, sizeof(__pyx_k_f_locals), 0, 0, 1, 1},
  {&__pyx_n_s_f_name, __pyx_k_f_name, sizeof(__pyx_k_f_name), 0, 0, 1, 1},
  {&__pyx_n_s_fget, __pyx_k_fget, sizeof(__pyx_k_fget), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_u_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 1, 0, 1},
  {&__pyx_n_s_file_2, __pyx_k_file_2, sizeof(__pyx_k_file_2), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_filename_2, __pyx_k_filename_2, sizeof(__pyx_k_filename_2), 0, 0, 1, 1},
  {&__pyx_kp_u_filename_lineno_function_code_co, __pyx_k_filename_lineno_function_code_co, sizeof(__pyx_k_filename_lineno_function_code_co), 0, 1, 0, 0},
  {&__pyx_n_s_filesbymodname, __pyx_k_filesbymodname, sizeof(__pyx_k_filesbymodname), 0, 0, 1, 1},
  {&__pyx_n_s_fillvalue, __pyx_k_fillvalue, sizeof(__pyx_k_fillvalue), 0, 0, 1, 1},
  {&__pyx_n_s_find, __pyx_k_find, sizeof(__pyx_k_find), 0, 0, 1, 1},
  {&__pyx_n_s_findclass, __pyx_k_findclass, sizeof(__pyx_k_findclass), 0, 0, 1, 1},
  {&__pyx_n_s_finddoc, __pyx_k_finddoc, sizeof(__pyx_k_finddoc), 0, 0, 1, 1},
  {&__pyx_n_s_findsource, __pyx_k_findsource, sizeof(__pyx_k_findsource), 0, 0, 1, 1},
  {&__pyx_n_s_first_wrapped_param, __pyx_k_first_wrapped_param, sizeof(__pyx_k_first_wrapped_param), 0, 0, 1, 1},
  {&__pyx_n_s_firstdefault, __pyx_k_firstdefault, sizeof(__pyx_k_firstdefault), 0, 0, 1, 1},
  {&__pyx_n_s_flag, __pyx_k_flag, sizeof(__pyx_k_flag), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_fn_underscore, __pyx_k_fn_underscore, sizeof(__pyx_k_fn_underscore), 0, 0, 1, 1},
  {&__pyx_n_s_fname, __pyx_k_fname, sizeof(__pyx_k_fname), 0, 0, 1, 1},
  {&__pyx_n_s_follow_wrapped, __pyx_k_follow_wrapped, sizeof(__pyx_k_follow_wrapped), 0, 0, 1, 1},
  {&__pyx_n_s_follow_wrapper_chains, __pyx_k_follow_wrapper_chains, sizeof(__pyx_k_follow_wrapper_chains), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_formatannotation, __pyx_k_formatannotation, sizeof(__pyx_k_formatannotation), 0, 0, 1, 1},
  {&__pyx_n_s_formatannotation_2, __pyx_k_formatannotation_2, sizeof(__pyx_k_formatannotation_2), 0, 0, 1, 1},
  {&__pyx_n_s_formatannotationrelativeto, __pyx_k_formatannotationrelativeto, sizeof(__pyx_k_formatannotationrelativeto), 0, 0, 1, 1},
  {&__pyx_n_s_formatannotationrelativeto_local, __pyx_k_formatannotationrelativeto_local, sizeof(__pyx_k_formatannotationrelativeto_local), 0, 0, 1, 1},
  {&__pyx_n_s_formatarg, __pyx_k_formatarg, sizeof(__pyx_k_formatarg), 0, 0, 1, 1},
  {&__pyx_n_s_formatargandannotation, __pyx_k_formatargandannotation, sizeof(__pyx_k_formatargandannotation), 0, 0, 1, 1},
  {&__pyx_n_s_formatargspec, __pyx_k_formatargspec, sizeof(__pyx_k_formatargspec), 0, 0, 1, 1},
  {&__pyx_kp_u_formatargspec_is_deprecated_sin, __pyx_k_formatargspec_is_deprecated_sin, sizeof(__pyx_k_formatargspec_is_deprecated_sin), 0, 1, 0, 0},
  {&__pyx_n_s_formatargspec_locals_formatargan, __pyx_k_formatargspec_locals_formatargan, sizeof(__pyx_k_formatargspec_locals_formatargan), 0, 0, 1, 1},
  {&__pyx_n_s_formatargspec_locals_lambda, __pyx_k_formatargspec_locals_lambda, sizeof(__pyx_k_formatargspec_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_formatargvalues, __pyx_k_formatargvalues, sizeof(__pyx_k_formatargvalues), 0, 0, 1, 1},
  {&__pyx_n_s_formatargvalues_locals_convert, __pyx_k_formatargvalues_locals_convert, sizeof(__pyx_k_formatargvalues_locals_convert), 0, 0, 1, 1},
  {&__pyx_n_s_formatargvalues_locals_lambda, __pyx_k_formatargvalues_locals_lambda, sizeof(__pyx_k_formatargvalues_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_formatreturns, __pyx_k_formatreturns, sizeof(__pyx_k_formatreturns), 0, 0, 1, 1},
  {&__pyx_n_s_formatted, __pyx_k_formatted, sizeof(__pyx_k_formatted), 0, 0, 1, 1},
  {&__pyx_n_s_formatvalue, __pyx_k_formatvalue, sizeof(__pyx_k_formatvalue), 0, 0, 1, 1},
  {&__pyx_n_s_formatvarargs, __pyx_k_formatvarargs, sizeof(__pyx_k_formatvarargs), 0, 0, 1, 1},
  {&__pyx_n_s_formatvarkw, __pyx_k_formatvarkw, sizeof(__pyx_k_formatvarkw), 0, 0, 1, 1},
  {&__pyx_n_s_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 0, 1, 1},
  {&__pyx_n_u_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 1, 0, 1},
  {&__pyx_n_s_frameinfo, __pyx_k_frameinfo, sizeof(__pyx_k_frameinfo), 0, 0, 1, 1},
  {&__pyx_n_s_framelist, __pyx_k_framelist, sizeof(__pyx_k_framelist), 0, 0, 1, 1},
  {&__pyx_kp_u_from, __pyx_k_from, sizeof(__pyx_k_from), 0, 1, 0, 0},
  {&__pyx_n_s_from_builtin, __pyx_k_from_builtin, sizeof(__pyx_k_from_builtin), 0, 0, 1, 1},
  {&__pyx_n_u_from_bytes, __pyx_k_from_bytes, sizeof(__pyx_k_from_bytes), 0, 1, 0, 1},
  {&__pyx_n_s_from_callable, __pyx_k_from_callable, sizeof(__pyx_k_from_callable), 0, 0, 1, 1},
  {&__pyx_n_s_from_function, __pyx_k_from_function, sizeof(__pyx_k_from_function), 0, 0, 1, 1},
  {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
  {&__pyx_n_u_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 1, 0, 1},
  {&__pyx_n_s_func_2, __pyx_k_func_2, sizeof(__pyx_k_func_2), 0, 0, 1, 1},
  {&__pyx_n_s_func_code, __pyx_k_func_code, sizeof(__pyx_k_func_code), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_generate_tokens, __pyx_k_generate_tokens, sizeof(__pyx_k_generate_tokens), 0, 0, 1, 1},
  {&__pyx_n_s_generator, __pyx_k_generator, sizeof(__pyx_k_generator), 0, 0, 1, 1},
  {&__pyx_n_s_generic_visit, __pyx_k_generic_visit, sizeof(__pyx_k_generic_visit), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_u_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 1, 0, 1},
  {&__pyx_n_s_get_2, __pyx_k_get_2, sizeof(__pyx_k_get_2), 0, 0, 1, 1},
  {&__pyx_n_s_get_callable_fullargspec, __pyx_k_get_callable_fullargspec, sizeof(__pyx_k_get_callable_fullargspec), 0, 0, 1, 1},
  {&__pyx_n_s_get_f_from_callable, __pyx_k_get_f_from_callable, sizeof(__pyx_k_get_f_from_callable), 0, 0, 1, 1},
  {&__pyx_n_s_get_obj, __pyx_k_get_obj, sizeof(__pyx_k_get_obj), 0, 0, 1, 1},
  {&__pyx_n_s_getabsfile, __pyx_k_getabsfile, sizeof(__pyx_k_getabsfile), 0, 0, 1, 1},
  {&__pyx_n_s_getargs, __pyx_k_getargs, sizeof(__pyx_k_getargs), 0, 0, 1, 1},
  {&__pyx_n_s_getargspec, __pyx_k_getargspec, sizeof(__pyx_k_getargspec), 0, 0, 1, 1},
  {&__pyx_n_s_getargvalues, __pyx_k_getargvalues, sizeof(__pyx_k_getargvalues), 0, 0, 1, 1},
  {&__pyx_n_s_getattr, __pyx_k_getattr, sizeof(__pyx_k_getattr), 0, 0, 1, 1},
  {&__pyx_n_s_getattr_static, __pyx_k_getattr_static, sizeof(__pyx_k_getattr_static), 0, 0, 1, 1},
  {&__pyx_n_s_getattribute, __pyx_k_getattribute, sizeof(__pyx_k_getattribute), 0, 0, 1, 1},
  {&__pyx_n_s_getblock, __pyx_k_getblock, sizeof(__pyx_k_getblock), 0, 0, 1, 1},
  {&__pyx_n_s_getcallargs, __pyx_k_getcallargs, sizeof(__pyx_k_getcallargs), 0, 0, 1, 1},
  {&__pyx_n_s_getcallargs_locals_assign, __pyx_k_getcallargs_locals_assign, sizeof(__pyx_k_getcallargs_locals_assign), 0, 0, 1, 1},
  {&__pyx_n_s_getcallargs_locals_is_assigned, __pyx_k_getcallargs_locals_is_assigned, sizeof(__pyx_k_getcallargs_locals_is_assigned), 0, 0, 1, 1},
  {&__pyx_n_s_getclasstree, __pyx_k_getclasstree, sizeof(__pyx_k_getclasstree), 0, 0, 1, 1},
  {&__pyx_n_s_getclosurevars, __pyx_k_getclosurevars, sizeof(__pyx_k_getclosurevars), 0, 0, 1, 1},
  {&__pyx_n_s_getcomments, __pyx_k_getcomments, sizeof(__pyx_k_getcomments), 0, 0, 1, 1},
  {&__pyx_n_s_getcoroutinelocals, __pyx_k_getcoroutinelocals, sizeof(__pyx_k_getcoroutinelocals), 0, 0, 1, 1},
  {&__pyx_n_s_getcoroutinestate, __pyx_k_getcoroutinestate, sizeof(__pyx_k_getcoroutinestate), 0, 0, 1, 1},
  {&__pyx_n_s_getdefaultencoding, __pyx_k_getdefaultencoding, sizeof(__pyx_k_getdefaultencoding), 0, 0, 1, 1},
  {&__pyx_n_s_getdoc, __pyx_k_getdoc, sizeof(__pyx_k_getdoc), 0, 0, 1, 1},
  {&__pyx_n_s_getfile, __pyx_k_getfile, sizeof(__pyx_k_getfile), 0, 0, 1, 1},
  {&__pyx_n_s_getframe, __pyx_k_getframe, sizeof(__pyx_k_getframe), 0, 0, 1, 1},
  {&__pyx_n_u_getframe, __pyx_k_getframe, sizeof(__pyx_k_getframe), 0, 1, 0, 1},
  {&__pyx_n_s_getframeinfo, __pyx_k_getframeinfo, sizeof(__pyx_k_getframeinfo), 0, 0, 1, 1},
  {&__pyx_n_s_getfullargspec, __pyx_k_getfullargspec, sizeof(__pyx_k_getfullargspec), 0, 0, 1, 1},
  {&__pyx_n_s_getgeneratorlocals, __pyx_k_getgeneratorlocals, sizeof(__pyx_k_getgeneratorlocals), 0, 0, 1, 1},
  {&__pyx_n_s_getgeneratorstate, __pyx_k_getgeneratorstate, sizeof(__pyx_k_getgeneratorstate), 0, 0, 1, 1},
  {&__pyx_n_s_getinnerframes, __pyx_k_getinnerframes, sizeof(__pyx_k_getinnerframes), 0, 0, 1, 1},
  {&__pyx_n_s_getlineno, __pyx_k_getlineno, sizeof(__pyx_k_getlineno), 0, 0, 1, 1},
  {&__pyx_n_s_getlines, __pyx_k_getlines, sizeof(__pyx_k_getlines), 0, 0, 1, 1},
  {&__pyx_n_s_getmembers, __pyx_k_getmembers, sizeof(__pyx_k_getmembers), 0, 0, 1, 1},
  {&__pyx_n_s_getmembers_locals_lambda, __pyx_k_getmembers_locals_lambda, sizeof(__pyx_k_getmembers_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_getmodule, __pyx_k_getmodule, sizeof(__pyx_k_getmodule), 0, 0, 1, 1},
  {&__pyx_n_s_getmodulename, __pyx_k_getmodulename, sizeof(__pyx_k_getmodulename), 0, 0, 1, 1},
  {&__pyx_n_s_getmro, __pyx_k_getmro, sizeof(__pyx_k_getmro), 0, 0, 1, 1},
  {&__pyx_n_s_getouterframes, __pyx_k_getouterframes, sizeof(__pyx_k_getouterframes), 0, 0, 1, 1},
  {&__pyx_n_s_getrecursionlimit, __pyx_k_getrecursionlimit, sizeof(__pyx_k_getrecursionlimit), 0, 0, 1, 1},
  {&__pyx_n_s_getsource, __pyx_k_getsource, sizeof(__pyx_k_getsource), 0, 0, 1, 1},
  {&__pyx_n_s_getsourcefile, __pyx_k_getsourcefile, sizeof(__pyx_k_getsourcefile), 0, 0, 1, 1},
  {&__pyx_n_s_getsourcefile_locals_genexpr, __pyx_k_getsourcefile_locals_genexpr, sizeof(__pyx_k_getsourcefile_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getsourcelines, __pyx_k_getsourcelines, sizeof(__pyx_k_getsourcelines), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_gi_code, __pyx_k_gi_code, sizeof(__pyx_k_gi_code), 0, 0, 1, 1},
  {&__pyx_n_s_gi_frame, __pyx_k_gi_frame, sizeof(__pyx_k_gi_frame), 0, 0, 1, 1},
  {&__pyx_n_u_gi_frame, __pyx_k_gi_frame, sizeof(__pyx_k_gi_frame), 0, 1, 0, 1},
  {&__pyx_n_s_gi_running, __pyx_k_gi_running, sizeof(__pyx_k_gi_running), 0, 0, 1, 1},
  {&__pyx_n_s_given, __pyx_k_given, sizeof(__pyx_k_given), 0, 0, 1, 1},
  {&__pyx_kp_u_given_2, __pyx_k_given_2, sizeof(__pyx_k_given_2), 0, 1, 0, 0},
  {&__pyx_kp_u_given_3, __pyx_k_given_3, sizeof(__pyx_k_given_3), 0, 1, 0, 0},
  {&__pyx_n_s_global_ns, __pyx_k_global_ns, sizeof(__pyx_k_global_ns), 0, 0, 1, 1},
  {&__pyx_n_s_global_vars, __pyx_k_global_vars, sizeof(__pyx_k_global_vars), 0, 0, 1, 1},
  {&__pyx_n_s_globals, __pyx_k_globals, sizeof(__pyx_k_globals), 0, 0, 1, 1},
  {&__pyx_kp_u_got_an_unexpected_keyword_argum, __pyx_k_got_an_unexpected_keyword_argum, sizeof(__pyx_k_got_an_unexpected_keyword_argum), 0, 1, 0, 0},
  {&__pyx_kp_u_got_an_unexpected_keyword_argume, __pyx_k_got_an_unexpected_keyword_argume, sizeof(__pyx_k_got_an_unexpected_keyword_argume), 0, 1, 0, 0},
  {&__pyx_kp_u_got_multiple_values_for_keyword, __pyx_k_got_multiple_values_for_keyword, sizeof(__pyx_k_got_multiple_values_for_keyword), 0, 1, 0, 0},
  {&__pyx_n_s_has_code_flag, __pyx_k_has_code_flag, sizeof(__pyx_k_has_code_flag), 0, 0, 1, 1},
  {&__pyx_n_s_hash, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
  {&__pyx_n_s_hash_basis, __pyx_k_hash_basis, sizeof(__pyx_k_hash_basis), 0, 0, 1, 1},
  {&__pyx_n_s_homecls, __pyx_k_homecls, sizeof(__pyx_k_homecls), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_id_func, __pyx_k_id_func, sizeof(__pyx_k_id_func), 0, 0, 1, 1},
  {&__pyx_n_s_im_self, __pyx_k_im_self, sizeof(__pyx_k_im_self), 0, 0, 1, 1},
  {&__pyx_kp_u_implicit, __pyx_k_implicit, sizeof(__pyx_k_implicit), 0, 1, 0, 0},
  {&__pyx_kp_u_implicit_arguments_must_be_passe, __pyx_k_implicit_arguments_must_be_passe, sizeof(__pyx_k_implicit_arguments_must_be_passe), 0, 1, 0, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_importlib, __pyx_k_importlib, sizeof(__pyx_k_importlib), 0, 0, 1, 1},
  {&__pyx_n_s_importlib_machinery, __pyx_k_importlib_machinery, sizeof(__pyx_k_importlib_machinery), 0, 0, 1, 1},
  {&__pyx_n_s_inPy2, __pyx_k_inPy2, sizeof(__pyx_k_inPy2), 0, 0, 1, 1},
  {&__pyx_n_s_indecorator, __pyx_k_indecorator, sizeof(__pyx_k_indecorator), 0, 0, 1, 1},
  {&__pyx_n_s_indent, __pyx_k_indent, sizeof(__pyx_k_indent), 0, 0, 1, 1},
  {&__pyx_n_s_indentsize, __pyx_k_indentsize, sizeof(__pyx_k_indentsize), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_u_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 1, 0, 1},
  {&__pyx_n_s_init_2, __pyx_k_init_2, sizeof(__pyx_k_init_2), 0, 0, 1, 1},
  {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
  {&__pyx_kp_u_inspect_Signature_from_builtin_i, __pyx_k_inspect_Signature_from_builtin_i, sizeof(__pyx_k_inspect_Signature_from_builtin_i), 0, 1, 0, 0},
  {&__pyx_kp_u_inspect_Signature_from_function, __pyx_k_inspect_Signature_from_function, sizeof(__pyx_k_inspect_Signature_from_function), 0, 1, 0, 0},
  {&__pyx_kp_u_inspect_getargspec_is_deprecated, __pyx_k_inspect_getargspec_is_deprecated, sizeof(__pyx_k_inspect_getargspec_is_deprecated), 0, 1, 0, 0},
  {&__pyx_n_s_instance_dict, __pyx_k_instance_dict, sizeof(__pyx_k_instance_dict), 0, 0, 1, 1},
  {&__pyx_n_s_instance_result, __pyx_k_instance_result, sizeof(__pyx_k_instance_result), 0, 0, 1, 1},
  {&__pyx_n_s_invalid, __pyx_k_invalid, sizeof(__pyx_k_invalid), 0, 0, 1, 1},
  {&__pyx_kp_u_invalid_argument_type, __pyx_k_invalid_argument_type, sizeof(__pyx_k_invalid_argument_type), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_method_signature, __pyx_k_invalid_method_signature, sizeof(__pyx_k_invalid_method_signature), 0, 1, 0, 0},
  {&__pyx_n_s_is_assigned, __pyx_k_is_assigned, sizeof(__pyx_k_is_assigned), 0, 0, 1, 1},
  {&__pyx_n_s_is_duck_function, __pyx_k_is_duck_function, sizeof(__pyx_k_is_duck_function), 0, 0, 1, 1},
  {&__pyx_kp_u_is_not_a_valid_Parameter_kind, __pyx_k_is_not_a_valid_Parameter_kind, sizeof(__pyx_k_is_not_a_valid_Parameter_kind), 0, 1, 0, 0},
  {&__pyx_n_s_is_type, __pyx_k_is_type, sizeof(__pyx_k_is_type), 0, 0, 1, 1},
  {&__pyx_n_s_is_wrapper, __pyx_k_is_wrapper, sizeof(__pyx_k_is_wrapper), 0, 0, 1, 1},
  {&__pyx_n_s_isabstract, __pyx_k_isabstract, sizeof(__pyx_k_isabstract), 0, 0, 1, 1},
  {&__pyx_n_u_isabstractmethod, __pyx_k_isabstractmethod, sizeof(__pyx_k_isabstractmethod), 0, 1, 0, 1},
  {&__pyx_n_s_isasyncgen, __pyx_k_isasyncgen, sizeof(__pyx_k_isasyncgen), 0, 0, 1, 1},
  {&__pyx_n_s_isasyncgenfunction, __pyx_k_isasyncgenfunction, sizeof(__pyx_k_isasyncgenfunction), 0, 0, 1, 1},
  {&__pyx_n_s_isawaitable, __pyx_k_isawaitable, sizeof(__pyx_k_isawaitable), 0, 0, 1, 1},
  {&__pyx_n_s_isbuiltin, __pyx_k_isbuiltin, sizeof(__pyx_k_isbuiltin), 0, 0, 1, 1},
  {&__pyx_n_s_isclass, __pyx_k_isclass, sizeof(__pyx_k_isclass), 0, 0, 1, 1},
  {&__pyx_n_s_iscode, __pyx_k_iscode, sizeof(__pyx_k_iscode), 0, 0, 1, 1},
  {&__pyx_n_s_iscoroutine, __pyx_k_iscoroutine, sizeof(__pyx_k_iscoroutine), 0, 0, 1, 1},
  {&__pyx_n_s_iscoroutinefunction, __pyx_k_iscoroutinefunction, sizeof(__pyx_k_iscoroutinefunction), 0, 0, 1, 1},
  {&__pyx_n_s_isdatadescriptor, __pyx_k_isdatadescriptor, sizeof(__pyx_k_isdatadescriptor), 0, 0, 1, 1},
  {&__pyx_n_s_isdigit, __pyx_k_isdigit, sizeof(__pyx_k_isdigit), 0, 0, 1, 1},
  {&__pyx_n_s_isframe, __pyx_k_isframe, sizeof(__pyx_k_isframe), 0, 0, 1, 1},
  {&__pyx_n_s_isfunction, __pyx_k_isfunction, sizeof(__pyx_k_isfunction), 0, 0, 1, 1},
  {&__pyx_n_s_isgenerator, __pyx_k_isgenerator, sizeof(__pyx_k_isgenerator), 0, 0, 1, 1},
  {&__pyx_n_s_isgeneratorfunction, __pyx_k_isgeneratorfunction, sizeof(__pyx_k_isgeneratorfunction), 0, 0, 1, 1},
  {&__pyx_n_s_isgetsetdescriptor, __pyx_k_isgetsetdescriptor, sizeof(__pyx_k_isgetsetdescriptor), 0, 0, 1, 1},
  {&__pyx_n_s_isidentifier, __pyx_k_isidentifier, sizeof(__pyx_k_isidentifier), 0, 0, 1, 1},
  {&__pyx_n_s_islambda, __pyx_k_islambda, sizeof(__pyx_k_islambda), 0, 0, 1, 1},
  {&__pyx_n_s_ismemberdescriptor, __pyx_k_ismemberdescriptor, sizeof(__pyx_k_ismemberdescriptor), 0, 0, 1, 1},
  {&__pyx_n_s_ismethod, __pyx_k_ismethod, sizeof(__pyx_k_ismethod), 0, 0, 1, 1},
  {&__pyx_n_s_ismethoddescriptor, __pyx_k_ismethoddescriptor, sizeof(__pyx_k_ismethoddescriptor), 0, 0, 1, 1},
  {&__pyx_n_s_ismodule, __pyx_k_ismodule, sizeof(__pyx_k_ismodule), 0, 0, 1, 1},
  {&__pyx_n_s_isroutine, __pyx_k_isroutine, sizeof(__pyx_k_isroutine), 0, 0, 1, 1},
  {&__pyx_n_s_istraceback, __pyx_k_istraceback, sizeof(__pyx_k_istraceback), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_kp_u_keyword_only, __pyx_k_keyword_only, sizeof(__pyx_k_keyword_only), 0, 1, 0, 0},
  {&__pyx_n_s_keyword_only_2, __pyx_k_keyword_only_2, sizeof(__pyx_k_keyword_only_2), 0, 0, 1, 1},
  {&__pyx_n_s_keyword_only_count, __pyx_k_keyword_only_count, sizeof(__pyx_k_keyword_only_count), 0, 0, 1, 1},
  {&__pyx_n_s_keywords, __pyx_k_keywords, sizeof(__pyx_k_keywords), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kind_2, __pyx_k_kind_2, sizeof(__pyx_k_kind_2), 0, 0, 1, 1},
  {&__pyx_n_u_kind_2, __pyx_k_kind_2, sizeof(__pyx_k_kind_2), 0, 1, 0, 1},
  {&__pyx_n_s_kind_defaults, __pyx_k_kind_defaults, sizeof(__pyx_k_kind_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_klass, __pyx_k_klass, sizeof(__pyx_k_klass), 0, 0, 1, 1},
  {&__pyx_n_s_klass_result, __pyx_k_klass_result, sizeof(__pyx_k_klass_result), 0, 0, 1, 1},
  {&__pyx_n_s_kw_defaults, __pyx_k_kw_defaults, sizeof(__pyx_k_kw_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_kwarg, __pyx_k_kwarg, sizeof(__pyx_k_kwarg), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs_param, __pyx_k_kwargs_param, sizeof(__pyx_k_kwargs_param), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs_started, __pyx_k_kwargs_started, sizeof(__pyx_k_kwargs_started), 0, 0, 1, 1},
  {&__pyx_n_s_kwdefaults, __pyx_k_kwdefaults, sizeof(__pyx_k_kwdefaults), 0, 0, 1, 1},
  {&__pyx_n_u_kwdefaults, __pyx_k_kwdefaults, sizeof(__pyx_k_kwdefaults), 0, 1, 0, 1},
  {&__pyx_n_s_kwdefaults_2, __pyx_k_kwdefaults_2, sizeof(__pyx_k_kwdefaults_2), 0, 0, 1, 1},
  {&__pyx_n_s_kwo_params, __pyx_k_kwo_params, sizeof(__pyx_k_kwo_params), 0, 0, 1, 1},
  {&__pyx_n_s_kwonly, __pyx_k_kwonly, sizeof(__pyx_k_kwonly), 0, 0, 1, 1},
  {&__pyx_n_s_kwonly_given, __pyx_k_kwonly_given, sizeof(__pyx_k_kwonly_given), 0, 0, 1, 1},
  {&__pyx_n_s_kwonly_sig, __pyx_k_kwonly_sig, sizeof(__pyx_k_kwonly_sig), 0, 0, 1, 1},
  {&__pyx_n_s_kwonlyarg, __pyx_k_kwonlyarg, sizeof(__pyx_k_kwonlyarg), 0, 0, 1, 1},
  {&__pyx_n_s_kwonlyargs, __pyx_k_kwonlyargs, sizeof(__pyx_k_kwonlyargs), 0, 0, 1, 1},
  {&__pyx_n_s_kwonlydefaults, __pyx_k_kwonlydefaults, sizeof(__pyx_k_kwonlydefaults), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_u_lambda, __pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 1, 0, 1},
  {&__pyx_n_s_last, __pyx_k_last, sizeof(__pyx_k_last), 0, 0, 1, 1},
  {&__pyx_n_s_last_cls, __pyx_k_last_cls, sizeof(__pyx_k_last_cls), 0, 0, 1, 1},
  {&__pyx_n_s_last_positional_only, __pyx_k_last_positional_only, sizeof(__pyx_k_last_positional_only), 0, 0, 1, 1},
  {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
  {&__pyx_n_s_line_number, __pyx_k_line_number, sizeof(__pyx_k_line_number), 0, 0, 1, 1},
  {&__pyx_n_s_linecache, __pyx_k_linecache, sizeof(__pyx_k_linecache), 0, 0, 1, 1},
  {&__pyx_n_s_lineno, __pyx_k_lineno, sizeof(__pyx_k_lineno), 0, 0, 1, 1},
  {&__pyx_kp_u_lineno_is_out_of_bounds, __pyx_k_lineno_is_out_of_bounds, sizeof(__pyx_k_lineno_is_out_of_bounds), 0, 1, 0, 0},
  {&__pyx_n_s_lines, __pyx_k_lines, sizeof(__pyx_k_lines), 0, 0, 1, 1},
  {&__pyx_n_s_literal_eval, __pyx_k_literal_eval, sizeof(__pyx_k_literal_eval), 0, 0, 1, 1},
  {&__pyx_n_s_lnum, __pyx_k_lnum, sizeof(__pyx_k_lnum), 0, 0, 1, 1},
  {&__pyx_n_u_loader, __pyx_k_loader, sizeof(__pyx_k_loader), 0, 1, 0, 1},
  {&__pyx_kp_u_locals, __pyx_k_locals, sizeof(__pyx_k_locals), 0, 1, 0, 0},
  {&__pyx_n_s_locals_2, __pyx_k_locals_2, sizeof(__pyx_k_locals_2), 0, 0, 1, 1},
  {&__pyx_n_s_lstrip, __pyx_k_lstrip, sizeof(__pyx_k_lstrip), 0, 0, 1, 1},
  {&__pyx_n_s_machinery, __pyx_k_machinery, sizeof(__pyx_k_machinery), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_u_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 1, 0, 1},
  {&__pyx_n_s_main_2, __pyx_k_main_2, sizeof(__pyx_k_main_2), 0, 0, 1, 1},
  {&__pyx_n_s_mainobject, __pyx_k_mainobject, sizeof(__pyx_k_mainobject), 0, 0, 1, 1},
  {&__pyx_n_s_margin, __pyx_k_margin, sizeof(__pyx_k_margin), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_maxsize, __pyx_k_maxsize, sizeof(__pyx_k_maxsize), 0, 0, 1, 1},
  {&__pyx_n_s_memo, __pyx_k_memo, sizeof(__pyx_k_memo), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_metamro, __pyx_k_metamro, sizeof(__pyx_k_metamro), 0, 0, 1, 1},
  {&__pyx_n_s_meth, __pyx_k_meth, sizeof(__pyx_k_meth), 0, 0, 1, 1},
  {&__pyx_n_u_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 1, 0, 1},
  {&__pyx_n_s_method_name, __pyx_k_method_name, sizeof(__pyx_k_method_name), 0, 0, 1, 1},
  {&__pyx_n_s_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 0, 1, 1},
  {&__pyx_kp_u_missing_a_required_argument_arg, __pyx_k_missing_a_required_argument_arg, sizeof(__pyx_k_missing_a_required_argument_arg), 0, 1, 0, 0},
  {&__pyx_n_s_missing_arguments, __pyx_k_missing_arguments, sizeof(__pyx_k_missing_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_mod_dict, __pyx_k_mod_dict, sizeof(__pyx_k_mod_dict), 0, 0, 1, 1},
  {&__pyx_n_s_modname, __pyx_k_modname, sizeof(__pyx_k_modname), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_u_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 1, 0, 1},
  {&__pyx_kp_u_module_2, __pyx_k_module_2, sizeof(__pyx_k_module_2), 0, 1, 0, 0},
  {&__pyx_n_s_module_3, __pyx_k_module_3, sizeof(__pyx_k_module_3), 0, 0, 1, 1},
  {&__pyx_kp_u_module_class_method_function_tra, __pyx_k_module_class_method_function_tra, sizeof(__pyx_k_module_class_method_function_tra), 0, 1, 0, 0},
  {&__pyx_n_s_module_dict, __pyx_k_module_dict, sizeof(__pyx_k_module_dict), 0, 0, 1, 1},
  {&__pyx_n_s_module_name, __pyx_k_module_name, sizeof(__pyx_k_module_name), 0, 0, 1, 1},
  {&__pyx_n_s_modules, __pyx_k_modules, sizeof(__pyx_k_modules), 0, 0, 1, 1},
  {&__pyx_n_s_modulesbyfile, __pyx_k_modulesbyfile, sizeof(__pyx_k_modulesbyfile), 0, 0, 1, 1},
  {&__pyx_n_s_move_to_end, __pyx_k_move_to_end, sizeof(__pyx_k_move_to_end), 0, 0, 1, 1},
  {&__pyx_n_s_mro, __pyx_k_mro, sizeof(__pyx_k_mro), 0, 0, 1, 1},
  {&__pyx_n_u_mro, __pyx_k_mro, sizeof(__pyx_k_mro), 0, 1, 0, 1},
  {&__pyx_n_s_mro_2, __pyx_k_mro_2, sizeof(__pyx_k_mro_2), 0, 0, 1, 1},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_kp_u_multiple_values_for_argument_arg, __pyx_k_multiple_values_for_argument_arg, sizeof(__pyx_k_multiple_values_for_argument_arg), 0, 1, 0, 0},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_name_3, __pyx_k_name_3, sizeof(__pyx_k_name_3), 0, 0, 1, 1},
  {&__pyx_n_s_name_4, __pyx_k_name_4, sizeof(__pyx_k_name_4), 0, 0, 1, 1},
  {&__pyx_n_u_name_4, __pyx_k_name_4, sizeof(__pyx_k_name_4), 0, 1, 0, 1},
  {&__pyx_kp_u_name_is_a_required_attribute_for, __pyx_k_name_is_a_required_attribute_for, sizeof(__pyx_k_name_is_a_required_attribute_for), 0, 1, 0, 0},
  {&__pyx_kp_u_name_kind_defining_class_object, __pyx_k_name_kind_defining_class_object, sizeof(__pyx_k_name_kind_defining_class_object), 0, 1, 0, 0},
  {&__pyx_kp_u_name_must_be_a_str_not_a, __pyx_k_name_must_be_a_str_not_a, sizeof(__pyx_k_name_must_be_a_str_not_a), 0, 1, 0, 0},
  {&__pyx_n_s_name_node, __pyx_k_name_node, sizeof(__pyx_k_name_node), 0, 0, 1, 1},
  {&__pyx_n_s_named, __pyx_k_named, sizeof(__pyx_k_named), 0, 0, 1, 1},
  {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_nargs, __pyx_k_nargs, sizeof(__pyx_k_nargs), 0, 0, 1, 1},
  {&__pyx_kp_u_need_more_than, __pyx_k_need_more_than, sizeof(__pyx_k_need_more_than), 0, 1, 0, 0},
  {&__pyx_n_s_neglen, __pyx_k_neglen, sizeof(__pyx_k_neglen), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_u_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 1, 0, 1},
  {&__pyx_n_s_new_2, __pyx_k_new_2, sizeof(__pyx_k_new_2), 0, 0, 1, 1},
  {&__pyx_n_s_new_arguments, __pyx_k_new_arguments, sizeof(__pyx_k_new_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_new_param, __pyx_k_new_param, sizeof(__pyx_k_new_param), 0, 0, 1, 1},
  {&__pyx_n_s_new_params, __pyx_k_new_params, sizeof(__pyx_k_new_params), 0, 0, 1, 1},
  {&__pyx_n_s_next, __pyx_k_next, sizeof(__pyx_k_next), 0, 0, 1, 1},
  {&__pyx_n_s_nkwargs, __pyx_k_nkwargs, sizeof(__pyx_k_nkwargs), 0, 0, 1, 1},
  {&__pyx_kp_u_no_signature_found_for_builtin_f, __pyx_k_no_signature_found_for_builtin_f, sizeof(__pyx_k_no_signature_found_for_builtin_f), 0, 1, 0, 0},
  {&__pyx_kp_u_no_signature_found_for_builtin_r, __pyx_k_no_signature_found_for_builtin_r, sizeof(__pyx_k_no_signature_found_for_builtin_r), 0, 1, 0, 0},
  {&__pyx_kp_u_no_signature_found_for_builtin_t, __pyx_k_no_signature_found_for_builtin_t, sizeof(__pyx_k_no_signature_found_for_builtin_t), 0, 1, 0, 0},
  {&__pyx_kp_u_no_signature_found_for_r, __pyx_k_no_signature_found_for_r, sizeof(__pyx_k_no_signature_found_for_r), 0, 1, 0, 0},
  {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
  {&__pyx_kp_u_non_default_argument_follows_def, __pyx_k_non_default_argument_follows_def, sizeof(__pyx_k_non_default_argument_follows_def), 0, 1, 0, 0},
  {&__pyx_n_s_non_default_count, __pyx_k_non_default_count, sizeof(__pyx_k_non_default_count), 0, 0, 1, 1},
  {&__pyx_n_s_nonlocal_vars, __pyx_k_nonlocal_vars, sizeof(__pyx_k_nonlocal_vars), 0, 0, 1, 1},
  {&__pyx_kp_u_nonlocals_globals_builtins_unbou, __pyx_k_nonlocals_globals_builtins_unbou, sizeof(__pyx_k_nonlocals_globals_builtins_unbou), 0, 1, 0, 0},
  {&__pyx_n_s_normcase, __pyx_k_normcase, sizeof(__pyx_k_normcase), 0, 0, 1, 1},
  {&__pyx_n_s_num_args, __pyx_k_num_args, sizeof(__pyx_k_num_args), 0, 0, 1, 1},
  {&__pyx_n_s_num_defaults, __pyx_k_num_defaults, sizeof(__pyx_k_num_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_num_pos, __pyx_k_num_pos, sizeof(__pyx_k_num_pos), 0, 0, 1, 1},
  {&__pyx_n_s_num_required, __pyx_k_num_required, sizeof(__pyx_k_num_required), 0, 0, 1, 1},
  {&__pyx_n_s_num_total, __pyx_k_num_total, sizeof(__pyx_k_num_total), 0, 0, 1, 1},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_objclass, __pyx_k_objclass, sizeof(__pyx_k_objclass), 0, 0, 1, 1},
  {&__pyx_n_u_objclass, __pyx_k_objclass, sizeof(__pyx_k_objclass), 0, 1, 0, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_old_params, __pyx_k_old_params, sizeof(__pyx_k_old_params), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_param, __pyx_k_param, sizeof(__pyx_k_param), 0, 0, 1, 1},
  {&__pyx_n_s_param_name, __pyx_k_param_name, sizeof(__pyx_k_param_name), 0, 0, 1, 1},
  {&__pyx_n_s_parameter_cls, __pyx_k_parameter_cls, sizeof(__pyx_k_parameter_cls), 0, 0, 1, 1},
  {&__pyx_n_s_parameters, __pyx_k_parameters, sizeof(__pyx_k_parameters), 0, 0, 1, 1},
  {&__pyx_n_s_parameters_2, __pyx_k_parameters_2, sizeof(__pyx_k_parameters_2), 0, 0, 1, 1},
  {&__pyx_n_u_parameters_2, __pyx_k_parameters_2, sizeof(__pyx_k_parameters_2), 0, 1, 0, 1},
  {&__pyx_kp_u_parameters_cannot_have_default, __pyx_k_parameters_cannot_have_default, sizeof(__pyx_k_parameters_cannot_have_default), 0, 1, 0, 0},
  {&__pyx_n_s_parameters_ex, __pyx_k_parameters_ex, sizeof(__pyx_k_parameters_ex), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_parse, __pyx_k_parse, sizeof(__pyx_k_parse), 0, 0, 1, 1},
  {&__pyx_n_s_parse_name, __pyx_k_parse_name, sizeof(__pyx_k_parse_name), 0, 0, 1, 1},
  {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {&__pyx_n_s_partial_args, __pyx_k_partial_args, sizeof(__pyx_k_partial_args), 0, 0, 1, 1},
  {&__pyx_n_s_partial_keywords, __pyx_k_partial_keywords, sizeof(__pyx_k_partial_keywords), 0, 0, 1, 1},
  {&__pyx_kp_u_partial_object_r_has_incorrect_a, __pyx_k_partial_object_r_has_incorrect_a, sizeof(__pyx_k_partial_object_r_has_incorrect_a), 0, 1, 0, 0},
  {&__pyx_n_s_partialmethod, __pyx_k_partialmethod, sizeof(__pyx_k_partialmethod), 0, 0, 1, 1},
  {&__pyx_n_s_partialmethod_2, __pyx_k_partialmethod_2, sizeof(__pyx_k_partialmethod_2), 0, 0, 1, 1},
  {&__pyx_kp_u_pass, __pyx_k_pass, sizeof(__pyx_k_pass), 0, 1, 0, 0},
  {&__pyx_n_s_passline, __pyx_k_passline, sizeof(__pyx_k_passline), 0, 0, 1, 1},
  {&__pyx_n_s_pat, __pyx_k_pat, sizeof(__pyx_k_pat), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_plural, __pyx_k_plural, sizeof(__pyx_k_plural), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_pos_count, __pyx_k_pos_count, sizeof(__pyx_k_pos_count), 0, 0, 1, 1},
  {&__pyx_n_s_pos_default_count, __pyx_k_pos_default_count, sizeof(__pyx_k_pos_default_count), 0, 0, 1, 1},
  {&__pyx_n_s_positional, __pyx_k_positional, sizeof(__pyx_k_positional), 0, 0, 1, 1},
  {&__pyx_n_u_positional, __pyx_k_positional, sizeof(__pyx_k_positional), 0, 1, 0, 1},
  {&__pyx_kp_u_positional_argument, __pyx_k_positional_argument, sizeof(__pyx_k_positional_argument), 0, 1, 0, 0},
  {&__pyx_kp_u_positional_argument_s_and_d_key, __pyx_k_positional_argument_s_and_d_key, sizeof(__pyx_k_positional_argument_s_and_d_key), 0, 1, 0, 0},
  {&__pyx_kp_u_positional_only, __pyx_k_positional_only, sizeof(__pyx_k_positional_only), 0, 1, 0, 0},
  {&__pyx_kp_u_positional_or_keyword, __pyx_k_positional_or_keyword, sizeof(__pyx_k_positional_or_keyword), 0, 1, 0, 0},
  {&__pyx_n_s_posonly_count, __pyx_k_posonly_count, sizeof(__pyx_k_posonly_count), 0, 0, 1, 1},
  {&__pyx_n_s_posonly_left, __pyx_k_posonly_left, sizeof(__pyx_k_posonly_left), 0, 0, 1, 1},
  {&__pyx_n_s_predicate, __pyx_k_predicate, sizeof(__pyx_k_predicate), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_processed, __pyx_k_processed, sizeof(__pyx_k_processed), 0, 0, 1, 1},
  {&__pyx_n_s_program, __pyx_k_program, sizeof(__pyx_k_program), 0, 0, 1, 1},
  {&__pyx_n_s_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 0, 1, 1},
  {&__pyx_n_u_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 1, 0, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_qualname_2, __pyx_k_qualname_2, sizeof(__pyx_k_qualname_2), 0, 0, 1, 1},
  {&__pyx_kp_u_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 1, 0, 0},
  {&__pyx_kp_u_r_builtin_has_invalid_signature, __pyx_k_r_builtin_has_invalid_signature, sizeof(__pyx_k_r_builtin_has_invalid_signature), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_a_built_in_class, __pyx_k_r_is_a_built_in_class, sizeof(__pyx_k_r_is_a_built_in_class), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_a_built_in_module, __pyx_k_r_is_a_built_in_module, sizeof(__pyx_k_r_is_a_built_in_module), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_Python_builtin_funct, __pyx_k_r_is_not_a_Python_builtin_funct, sizeof(__pyx_k_r_is_not_a_Python_builtin_funct), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_Python_function, __pyx_k_r_is_not_a_Python_function, sizeof(__pyx_k_r_is_not_a_Python_function), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_Python_generator, __pyx_k_r_is_not_a_Python_generator, sizeof(__pyx_k_r_is_not_a_Python_generator), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_callable_object, __pyx_k_r_is_not_a_callable_object, sizeof(__pyx_k_r_is_not_a_callable_object), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_code_object, __pyx_k_r_is_not_a_code_object, sizeof(__pyx_k_r_is_not_a_code_object), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_frame_or_traceback_o, __pyx_k_r_is_not_a_frame_or_traceback_o, sizeof(__pyx_k_r_is_not_a_frame_or_traceback_o), 0, 1, 0, 0},
  {&__pyx_kp_u_r_is_not_a_valid_parameter_name, __pyx_k_r_is_not_a_valid_parameter_name, sizeof(__pyx_k_r_is_not_a_valid_parameter_name), 0, 1, 0, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_realpath, __pyx_k_realpath, sizeof(__pyx_k_realpath), 0, 0, 1, 1},
  {&__pyx_n_s_recursion_limit, __pyx_k_recursion_limit, sizeof(__pyx_k_recursion_limit), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_render_kw_only_separator, __pyx_k_render_kw_only_separator, sizeof(__pyx_k_render_kw_only_separator), 0, 0, 1, 1},
  {&__pyx_n_s_render_pos_only_separator, __pyx_k_render_pos_only_separator, sizeof(__pyx_k_render_pos_only_separator), 0, 0, 1, 1},
  {&__pyx_n_s_rendered, __pyx_k_rendered, sizeof(__pyx_k_rendered), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_u_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 1, 0, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 0, 1, 1},
  {&__pyx_n_u_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 1, 0, 1},
  {&__pyx_n_s_return_annotation, __pyx_k_return_annotation, sizeof(__pyx_k_return_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_return_annotation_2, __pyx_k_return_annotation_2, sizeof(__pyx_k_return_annotation_2), 0, 0, 1, 1},
  {&__pyx_n_u_return_annotation_2, __pyx_k_return_annotation_2, sizeof(__pyx_k_return_annotation_2), 0, 1, 0, 1},
  {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
  {&__pyx_n_s_roots, __pyx_k_roots, sizeof(__pyx_k_roots), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 1},
  {&__pyx_kp_u_s_def_s_s_async_s_def_s_w_lambd, __pyx_k_s_def_s_s_async_s_def_s_w_lambd, sizeof(__pyx_k_s_def_s_s_async_s_def_s_w_lambd), 0, 1, 0, 0},
  {&__pyx_kp_u_s_missing_i_required_s_argument, __pyx_k_s_missing_i_required_s_argument, sizeof(__pyx_k_s_missing_i_required_s_argument), 0, 1, 0, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_u_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 1, 0, 1},
  {&__pyx_n_s_self_2, __pyx_k_self_2, sizeof(__pyx_k_self_2), 0, 0, 1, 1},
  {&__pyx_n_u_self_2, __pyx_k_self_2, sizeof(__pyx_k_self_2), 0, 1, 0, 1},
  {&__pyx_n_s_self_3, __pyx_k_self_3, sizeof(__pyx_k_self_3), 0, 0, 1, 1},
  {&__pyx_n_s_self_isbound, __pyx_k_self_isbound, sizeof(__pyx_k_self_isbound), 0, 0, 1, 1},
  {&__pyx_n_s_self_ismodule, __pyx_k_self_ismodule, sizeof(__pyx_k_self_ismodule), 0, 0, 1, 1},
  {&__pyx_n_s_self_parameter, __pyx_k_self_parameter, sizeof(__pyx_k_self_parameter), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_sentinel, __pyx_k_sentinel, sizeof(__pyx_k_sentinel), 0, 0, 1, 1},
  {&__pyx_n_u_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 1, 0, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_shadowed_dict, __pyx_k_shadowed_dict, sizeof(__pyx_k_shadowed_dict), 0, 0, 1, 1},
  {&__pyx_n_s_sig, __pyx_k_sig, sizeof(__pyx_k_sig), 0, 0, 1, 1},
  {&__pyx_n_s_sig_params, __pyx_k_sig_params, sizeof(__pyx_k_sig_params), 0, 0, 1, 1},
  {&__pyx_n_s_sigcls, __pyx_k_sigcls, sizeof(__pyx_k_sigcls), 0, 0, 1, 1},
  {&__pyx_n_s_signature, __pyx_k_signature, sizeof(__pyx_k_signature), 0, 0, 1, 1},
  {&__pyx_n_u_signature, __pyx_k_signature, sizeof(__pyx_k_signature), 0, 1, 0, 1},
  {&__pyx_n_s_signature_2, __pyx_k_signature_2, sizeof(__pyx_k_signature_2), 0, 0, 1, 1},
  {&__pyx_n_s_signature_3, __pyx_k_signature_3, sizeof(__pyx_k_signature_3), 0, 0, 1, 1},
  {&__pyx_n_u_signature_3, __pyx_k_signature_3, sizeof(__pyx_k_signature_3), 0, 1, 0, 1},
  {&__pyx_n_s_signature_bound_method, __pyx_k_signature_bound_method, sizeof(__pyx_k_signature_bound_method), 0, 0, 1, 1},
  {&__pyx_n_s_signature_from_builtin, __pyx_k_signature_from_builtin, sizeof(__pyx_k_signature_from_builtin), 0, 0, 1, 1},
  {&__pyx_n_s_signature_from_callable, __pyx_k_signature_from_callable, sizeof(__pyx_k_signature_from_callable), 0, 0, 1, 1},
  {&__pyx_n_s_signature_from_callable_locals, __pyx_k_signature_from_callable_locals, sizeof(__pyx_k_signature_from_callable_locals), 0, 0, 1, 1},
  {&__pyx_n_s_signature_from_function, __pyx_k_signature_from_function, sizeof(__pyx_k_signature_from_function), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr, __pyx_k_signature_fromstr, sizeof(__pyx_k_signature_fromstr), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_Rewrit, __pyx_k_signature_fromstr_locals_Rewrit, sizeof(__pyx_k_signature_fromstr_locals_Rewrit), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_Rewrit_2, __pyx_k_signature_fromstr_locals_Rewrit_2, sizeof(__pyx_k_signature_fromstr_locals_Rewrit_2), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_Rewrit_3, __pyx_k_signature_fromstr_locals_Rewrit_3, sizeof(__pyx_k_signature_fromstr_locals_Rewrit_3), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_p, __pyx_k_signature_fromstr_locals_p, sizeof(__pyx_k_signature_fromstr_locals_p), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_parse, __pyx_k_signature_fromstr_locals_parse, sizeof(__pyx_k_signature_fromstr_locals_parse), 0, 0, 1, 1},
  {&__pyx_n_s_signature_fromstr_locals_wrap_v, __pyx_k_signature_fromstr_locals_wrap_v, sizeof(__pyx_k_signature_fromstr_locals_wrap_v), 0, 0, 1, 1},
  {&__pyx_n_s_signature_get_bound_param, __pyx_k_signature_get_bound_param, sizeof(__pyx_k_signature_get_bound_param), 0, 0, 1, 1},
  {&__pyx_n_s_signature_get_partial, __pyx_k_signature_get_partial, sizeof(__pyx_k_signature_get_partial), 0, 0, 1, 1},
  {&__pyx_n_s_signature_get_user_defined_meth, __pyx_k_signature_get_user_defined_meth, sizeof(__pyx_k_signature_get_user_defined_meth), 0, 0, 1, 1},
  {&__pyx_n_s_signature_is_builtin, __pyx_k_signature_is_builtin, sizeof(__pyx_k_signature_is_builtin), 0, 0, 1, 1},
  {&__pyx_n_s_signature_is_functionlike, __pyx_k_signature_is_functionlike, sizeof(__pyx_k_signature_is_functionlike), 0, 0, 1, 1},
  {&__pyx_n_s_signature_strip_non_python_synt, __pyx_k_signature_strip_non_python_synt, sizeof(__pyx_k_signature_strip_non_python_synt), 0, 0, 1, 1},
  {&__pyx_n_s_skip_bound_arg, __pyx_k_skip_bound_arg, sizeof(__pyx_k_skip_bound_arg), 0, 0, 1, 1},
  {&__pyx_n_s_skip_next_comma, __pyx_k_skip_next_comma, sizeof(__pyx_k_skip_next_comma), 0, 0, 1, 1},
  {&__pyx_n_s_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 0, 1, 1},
  {&__pyx_n_u_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 1, 0, 1},
  {&__pyx_n_s_slots_2, __pyx_k_slots_2, sizeof(__pyx_k_slots_2), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_kp_u_source_code_not_available, __pyx_k_source_code_not_available, sizeof(__pyx_k_source_code_not_available), 0, 1, 0, 0},
  {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {&__pyx_n_s_specs, __pyx_k_specs, sizeof(__pyx_k_specs), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_splitext, __pyx_k_splitext, sizeof(__pyx_k_splitext), 0, 0, 1, 1},
  {&__pyx_n_s_srch_cls, __pyx_k_srch_cls, sizeof(__pyx_k_srch_cls), 0, 0, 1, 1},
  {&__pyx_n_s_srch_obj, __pyx_k_srch_obj, sizeof(__pyx_k_srch_obj), 0, 0, 1, 1},
  {&__pyx_n_s_srowcol, __pyx_k_srowcol, sizeof(__pyx_k_srowcol), 0, 0, 1, 1},
  {&__pyx_n_s_stack, __pyx_k_stack, sizeof(__pyx_k_stack), 0, 0, 1, 1},
  {&__pyx_n_s_stacklevel, __pyx_k_stacklevel, sizeof(__pyx_k_stacklevel), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_started, __pyx_k_started, sizeof(__pyx_k_started), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_static_getmro, __pyx_k_static_getmro, sizeof(__pyx_k_static_getmro), 0, 0, 1, 1},
  {&__pyx_kp_u_static_method, __pyx_k_static_method, sizeof(__pyx_k_static_method), 0, 1, 0, 0},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_subarg, __pyx_k_subarg, sizeof(__pyx_k_subarg), 0, 0, 1, 1},
  {&__pyx_n_s_subvalue, __pyx_k_subvalue, sizeof(__pyx_k_subvalue), 0, 0, 1, 1},
  {&__pyx_n_s_suffix, __pyx_k_suffix, sizeof(__pyx_k_suffix), 0, 0, 1, 1},
  {&__pyx_n_s_suffixes, __pyx_k_suffixes, sizeof(__pyx_k_suffixes), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_sys_module_dict, __pyx_k_sys_module_dict, sizeof(__pyx_k_sys_module_dict), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_tail, __pyx_k_tail, sizeof(__pyx_k_tail), 0, 0, 1, 1},
  {&__pyx_kp_u_takes, __pyx_k_takes, sizeof(__pyx_k_takes), 0, 1, 0, 0},
  {&__pyx_kp_u_takes_no_arguments, __pyx_k_takes_no_arguments, sizeof(__pyx_k_takes_no_arguments), 0, 1, 0, 0},
  {&__pyx_n_s_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
  {&__pyx_n_s_tb_frame, __pyx_k_tb_frame, sizeof(__pyx_k_tb_frame), 0, 0, 1, 1},
  {&__pyx_n_s_tb_lineno, __pyx_k_tb_lineno, sizeof(__pyx_k_tb_lineno), 0, 0, 1, 1},
  {&__pyx_n_s_tb_next, __pyx_k_tb_next, sizeof(__pyx_k_tb_next), 0, 0, 1, 1},
  {&__pyx_n_u_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 1, 0, 1},
  {&__pyx_n_s_test_2, __pyx_k_test_2, sizeof(__pyx_k_test_2), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_text_sig, __pyx_k_text_sig, sizeof(__pyx_k_text_sig), 0, 0, 1, 1},
  {&__pyx_n_s_text_signature, __pyx_k_text_signature, sizeof(__pyx_k_text_signature), 0, 0, 1, 1},
  {&__pyx_n_u_text_signature, __pyx_k_text_signature, sizeof(__pyx_k_text_signature), 0, 1, 0, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_kp_u_to, __pyx_k_to, sizeof(__pyx_k_to), 0, 1, 0, 0},
  {&__pyx_kp_u_to_unpack, __pyx_k_to_unpack, sizeof(__pyx_k_to_unpack), 0, 1, 0, 0},
  {&__pyx_n_s_token, __pyx_k_token, sizeof(__pyx_k_token), 0, 0, 1, 1},
  {&__pyx_n_s_token_2, __pyx_k_token_2, sizeof(__pyx_k_token_2), 0, 0, 1, 1},
  {&__pyx_n_s_token_stream, __pyx_k_token_stream, sizeof(__pyx_k_token_stream), 0, 0, 1, 1},
  {&__pyx_n_s_tokeneater, __pyx_k_tokeneater, sizeof(__pyx_k_tokeneater), 0, 0, 1, 1},
  {&__pyx_n_s_tokenize, __pyx_k_tokenize, sizeof(__pyx_k_tokenize), 0, 0, 1, 1},
  {&__pyx_n_s_tokens, __pyx_k_tokens, sizeof(__pyx_k_tokens), 0, 0, 1, 1},
  {&__pyx_n_s_too_many, __pyx_k_too_many, sizeof(__pyx_k_too_many), 0, 0, 1, 1},
  {&__pyx_kp_u_too_many_positional_arguments, __pyx_k_too_many_positional_arguments, sizeof(__pyx_k_too_many_positional_arguments), 0, 1, 0, 0},
  {&__pyx_kp_u_too_many_values_to_unpack, __pyx_k_too_many_values_to_unpack, sizeof(__pyx_k_too_many_values_to_unpack), 0, 1, 0, 0},
  {&__pyx_n_s_top_kind, __pyx_k_top_kind, sizeof(__pyx_k_top_kind), 0, 0, 1, 1},
  {&__pyx_n_s_tp, __pyx_k_tp, sizeof(__pyx_k_tp), 0, 0, 1, 1},
  {&__pyx_n_s_trace, __pyx_k_trace, sizeof(__pyx_k_trace), 0, 0, 1, 1},
  {&__pyx_n_s_transform_to_kwonly, __pyx_k_transform_to_kwonly, sizeof(__pyx_k_transform_to_kwonly), 0, 0, 1, 1},
  {&__pyx_n_s_tree, __pyx_k_tree, sizeof(__pyx_k_tree), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_n_u_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 1, 0, 1},
  {&__pyx_kp_u_typing_2, __pyx_k_typing_2, sizeof(__pyx_k_typing_2), 0, 1, 0, 0},
  {&__pyx_n_s_unassigned, __pyx_k_unassigned, sizeof(__pyx_k_unassigned), 0, 0, 1, 1},
  {&__pyx_n_s_unbound_names, __pyx_k_unbound_names, sizeof(__pyx_k_unbound_names), 0, 0, 1, 1},
  {&__pyx_n_s_unexpected, __pyx_k_unexpected, sizeof(__pyx_k_unexpected), 0, 0, 1, 1},
  {&__pyx_kp_u_unexpected_object_r_in___signatu, __pyx_k_unexpected_object_r_in___signatu, sizeof(__pyx_k_unexpected_object_r_in___signatu), 0, 1, 0, 0},
  {&__pyx_n_s_unique, __pyx_k_unique, sizeof(__pyx_k_unique), 0, 0, 1, 1},
  {&__pyx_n_s_unwrap, __pyx_k_unwrap, sizeof(__pyx_k_unwrap), 0, 0, 1, 1},
  {&__pyx_n_s_unwrap_locals__is_wrapper, __pyx_k_unwrap_locals__is_wrapper, sizeof(__pyx_k_unwrap_locals__is_wrapper), 0, 0, 1, 1},
  {&__pyx_n_s_unwrap_partial, __pyx_k_unwrap_partial, sizeof(__pyx_k_unwrap_partial), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_validate_parameters, __pyx_k_validate_parameters, sizeof(__pyx_k_validate_parameters), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_u_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 1, 0, 1},
  {&__pyx_kp_u_value_2, __pyx_k_value_2, sizeof(__pyx_k_value_2), 0, 1, 0, 0},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_u_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 1, 0, 1},
  {&__pyx_n_s_var, __pyx_k_var, sizeof(__pyx_k_var), 0, 0, 1, 1},
  {&__pyx_n_s_vararg, __pyx_k_vararg, sizeof(__pyx_k_vararg), 0, 0, 1, 1},
  {&__pyx_n_s_varargs, __pyx_k_varargs, sizeof(__pyx_k_varargs), 0, 0, 1, 1},
  {&__pyx_kp_u_variadic_keyword, __pyx_k_variadic_keyword, sizeof(__pyx_k_variadic_keyword), 0, 1, 0, 0},
  {&__pyx_kp_u_variadic_positional, __pyx_k_variadic_positional, sizeof(__pyx_k_variadic_positional), 0, 1, 0, 0},
  {&__pyx_n_s_varkw, __pyx_k_varkw, sizeof(__pyx_k_varkw), 0, 0, 1, 1},
  {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
  {&__pyx_n_s_visit, __pyx_k_visit, sizeof(__pyx_k_visit), 0, 0, 1, 1},
  {&__pyx_n_s_visit_AsyncFunctionDef, __pyx_k_visit_AsyncFunctionDef, sizeof(__pyx_k_visit_AsyncFunctionDef), 0, 0, 1, 1},
  {&__pyx_n_s_visit_Attribute, __pyx_k_visit_Attribute, sizeof(__pyx_k_visit_Attribute), 0, 0, 1, 1},
  {&__pyx_n_s_visit_ClassDef, __pyx_k_visit_ClassDef, sizeof(__pyx_k_visit_ClassDef), 0, 0, 1, 1},
  {&__pyx_n_s_visit_FunctionDef, __pyx_k_visit_FunctionDef, sizeof(__pyx_k_visit_FunctionDef), 0, 0, 1, 1},
  {&__pyx_n_s_visit_Name, __pyx_k_visit_Name, sizeof(__pyx_k_visit_Name), 0, 0, 1, 1},
  {&__pyx_n_s_void, __pyx_k_void, sizeof(__pyx_k_void), 0, 0, 1, 1},
  {&__pyx_n_s_walktree, __pyx_k_walktree, sizeof(__pyx_k_walktree), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_u_was, __pyx_k_was, sizeof(__pyx_k_was), 0, 1, 0, 1},
  {&__pyx_n_u_weakref, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 1, 0, 1},
  {&__pyx_n_u_were, __pyx_k_were, sizeof(__pyx_k_were), 0, 1, 0, 1},
  {&__pyx_n_s_wrap_value, __pyx_k_wrap_value, sizeof(__pyx_k_wrap_value), 0, 0, 1, 1},
  {&__pyx_n_s_wrapped, __pyx_k_wrapped, sizeof(__pyx_k_wrapped), 0, 0, 1, 1},
  {&__pyx_n_u_wrapped, __pyx_k_wrapped, sizeof(__pyx_k_wrapped), 0, 1, 0, 1},
  {&__pyx_n_s_wrapped_sig, __pyx_k_wrapped_sig, sizeof(__pyx_k_wrapped_sig), 0, 0, 1, 1},
  {&__pyx_kp_u_wrapper_loop_when_unwrapping_r, __pyx_k_wrapper_loop_when_unwrapping_r, sizeof(__pyx_k_wrapper_loop_when_unwrapping_r), 0, 1, 0, 0},
  {&__pyx_kp_u_wrong_parameter_order_parameter, __pyx_k_wrong_parameter_order_parameter, sizeof(__pyx_k_wrong_parameter_order_parameter), 0, 1, 0, 0},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {&__pyx_n_s_zip_longest, __pyx_k_zip_longest, sizeof(__pyx_k_zip_longest), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 1611, __pyx_L1_error)
  __pyx_builtin_all = __Pyx_GetBuiltinName(__pyx_n_s_all); if (!__pyx_builtin_all) __PYX_ERR(0, 1794, __pyx_L1_error)
  __pyx_builtin_property = __Pyx_GetBuiltinName(__pyx_n_s_property); if (!__pyx_builtin_property) __PYX_ERR(0, 2530, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 375, __pyx_L1_error)
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 502, __pyx_L1_error)
  __pyx_builtin_classmethod = __Pyx_GetBuiltinName(__pyx_n_s_classmethod); if (!__pyx_builtin_classmethod) __PYX_ERR(0, 505, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 554, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 560, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 659, __pyx_L1_error)
  __pyx_builtin_UnicodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeError); if (!__pyx_builtin_UnicodeError) __PYX_ERR(0, 673, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 873, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 915, __pyx_L1_error)
  __pyx_builtin_IndentationError = __Pyx_GetBuiltinName(__pyx_n_s_IndentationError); if (!__pyx_builtin_IndentationError) __PYX_ERR(0, 1045, __pyx_L1_error)
  __pyx_builtin_DeprecationWarning = __Pyx_GetBuiltinName(__pyx_n_s_DeprecationWarning); if (!__pyx_builtin_DeprecationWarning) __PYX_ERR(0, 1177, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1257, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1422, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 1397, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 1509, __pyx_L1_error)
  __pyx_builtin_SyntaxError = __Pyx_GetBuiltinName(__pyx_n_s_SyntaxError); if (!__pyx_builtin_SyntaxError) __PYX_ERR(0, 2069, __pyx_L1_error)
  __pyx_builtin_reversed = __Pyx_GetBuiltinName(__pyx_n_s_reversed); if (!__pyx_builtin_reversed) __PYX_ERR(0, 2144, __pyx_L1_error)
  __pyx_builtin_eval = __Pyx_GetBuiltinName(__pyx_n_s_eval); if (!__pyx_builtin_eval) __PYX_ERR(0, 2098, __pyx_L1_error)
  __pyx_builtin_NameError = __Pyx_GetBuiltinName(__pyx_n_s_NameError); if (!__pyx_builtin_NameError) __PYX_ERR(0, 2099, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 2103, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 2698, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "Aspidites/_vendor/contracts/inspection.py":464
 *             try:
 *                 if name == '__dict__':
 *                     raise Exception("__dict__ is special, don't want the proxy")             # <<<<<<<<<<<<<<
 *                 get_obj = getattr(cls, name)
 *             except Exception as exc:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_dict___is_special_don_t_want_t); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "Aspidites/_vendor/contracts/inspection.py":546
 *     """
 *     if stop is None:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__')
 *     else:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_n_s_f); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_is_wrapper, 546, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 546, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":549
 *             return hasattr(f, '__wrapped__')
 *     else:
 *         def _is_wrapper(f):             # <<<<<<<<<<<<<<
 *             return hasattr(f, '__wrapped__') and not stop(f)
 *     f = func  # remember the original func for error reporting
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_n_s_f); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_is_wrapper, 549, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 549, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":576
 *     if cls is None:
 *         return None
 *     for name in func.__qualname__.split('.')[:-1]:             # <<<<<<<<<<<<<<
 *         cls = getattr(cls, name)
 *     if not isclass(cls):
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "Aspidites/_vendor/contracts/inspection.py":678
 *         # Find minimum indentation of any non-blank lines after first line.
 *         margin = sys.maxsize
 *         for line in lines[1:]:             # <<<<<<<<<<<<<<
 *             content = len(line.lstrip())
 *             if content:
 */
  __pyx_slice__13 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);

  /* "Aspidites/_vendor/contracts/inspection.py":741
 *     """
 *     filename = getfile(object)
 *     all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]             # <<<<<<<<<<<<<<
 *     all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
 *     if any(filename.endswith(s) for s in all_bytecode_suffixes):
 */
  __pyx_slice__14 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__14)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__14);
  __Pyx_GIVEREF(__pyx_slice__14);

  /* "Aspidites/_vendor/contracts/inspection.py":873
 *         # inspection, so let `linecache.getlines` to be called.
 *         if not (file.startswith('<') and file.endswith('>')):
 *             raise OSError('source code not available')             # <<<<<<<<<<<<<<
 * 
 *     module = getmodule(object, file)
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_source_code_not_available); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "Aspidites/_vendor/contracts/inspection.py":881
 *         lines = linecache.getlines(file)
 *     if not lines:
 *         raise OSError('could not get source code')             # <<<<<<<<<<<<<<
 * 
 *     if ismodule(object):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_could_not_get_source_code); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "Aspidites/_vendor/contracts/inspection.py":897
 *             return lines, line_number
 *         else:
 *             raise OSError('could not find class definition')             # <<<<<<<<<<<<<<
 * 
 *     if ismethod(object):
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_could_not_find_class_definition); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "Aspidites/_vendor/contracts/inspection.py":909
 *     if iscode(object):
 *         if not hasattr(object, 'co_firstlineno'):
 *             raise OSError('could not find function definition')             # <<<<<<<<<<<<<<
 *         lnum = object.co_firstlineno - 1
 *         pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_u_could_not_find_function_definiti); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "Aspidites/_vendor/contracts/inspection.py":916
 *                 line = lines[lnum]
 *             except IndexError:
 *                 raise OSError('lineno is out of bounds')             # <<<<<<<<<<<<<<
 *             if pat.match(line):
 *                 break
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u_lineno_is_out_of_bounds); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "Aspidites/_vendor/contracts/inspection.py":921
 *             lnum = lnum - 1
 *         return lines, lnum
 *     raise OSError('could not find code object')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_could_not_find_code_object); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "Aspidites/_vendor/contracts/inspection.py":937
 *         # Look for a comment block at the top of the file.
 *         start = 0
 *         if lines and lines[0][:2] == '#!':             # <<<<<<<<<<<<<<
 *             start = 1
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 */
  __pyx_slice__24 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__24)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__24);
  __Pyx_GIVEREF(__pyx_slice__24);

  /* "Aspidites/_vendor/contracts/inspection.py":941
 *         while start < len(lines) and lines[start].strip() in ('', '#'):
 *             start = start + 1
 *         if start < len(lines) and lines[start][:1] == '#':             # <<<<<<<<<<<<<<
 *             comments = []
 *             end = start
 */
  __pyx_slice__27 = PySlice_New(Py_None, __pyx_int_1, Py_None); if (unlikely(!__pyx_slice__27)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__27);
  __Pyx_GIVEREF(__pyx_slice__27);

  /* "Aspidites/_vendor/contracts/inspection.py":1086
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 *     classes.sort(key=attrgetter('__module__', '__name__'))             # <<<<<<<<<<<<<<
 *     for c in classes:
 *         results.append((c, c.__bases__))
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_u_module, __pyx_n_u_name); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "Aspidites/_vendor/contracts/inspection.py":1175
 *     Deprecated since Python 3.5, use `inspect.getfullargspec()`.
 *     """
 *     warnings.warn("inspect.getargspec() is deprecated since Python 3.0, "             # <<<<<<<<<<<<<<
 *                   "use inspect.signature() or inspect.getfullargspec()",
 *                   DeprecationWarning, stacklevel=2)
 */
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_kp_u_inspect_getargspec_is_deprecated, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "Aspidites/_vendor/contracts/inspection.py":1181
 *         getfullargspec(func)
 *     if kwonlyargs or ann:
 *         raise ValueError("Function has keyword-only parameters or annotations"             # <<<<<<<<<<<<<<
 *                          ", use inspect.signature() API which can support them")
 *     return ArgSpec(args, varargs, varkw, defaults)
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_u_Function_has_keyword_only_parame); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "Aspidites/_vendor/contracts/inspection.py":1204
 * def formatannotation(annotation, base_module=None):
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')             # <<<<<<<<<<<<<<
 *     if isinstance(annotation, type):
 *         if annotation.__module__ in ('builtins', base_module):
 */
  __pyx_tuple__34 = PyTuple_Pack(2, __pyx_kp_u_typing_2, __pyx_kp_u__19); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "Aspidites/_vendor/contracts/inspection.py":1215
 *     module = getattr(object, '__module__', None)
 * 
 *     def _formatannotation(annotation):             # <<<<<<<<<<<<<<
 *         return formatannotation(annotation, module)
 * 
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_n_s_annotation); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatannotation_2, 1215, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 1215, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1243
 *     from warnings import warn
 * 
 *     warn("`formatargspec` is deprecated since Python 3.5. Use `signature` and "             # <<<<<<<<<<<<<<
 *          "the `Signature` object directly",
 *          DeprecationWarning,
 */
  __pyx_tuple__38 = PyTuple_Pack(2, __pyx_kp_u_formatargspec_is_deprecated_sin, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "Aspidites/_vendor/contracts/inspection.py":1248
 *          stacklevel=2)
 * 
 *     def formatargandannotation(arg):             # <<<<<<<<<<<<<<
 *         result = formatarg(arg)
 *         if arg in annotations:
 */
  __pyx_tuple__39 = PyTuple_Pack(2, __pyx_n_s_arg, __pyx_n_s_result); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatargandannotation, 1248, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 1248, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1294
 *     argument is an optional function to format the sequence of arguments."""
 * 
 *     def convert(name, locals=locals,             # <<<<<<<<<<<<<<
 *                 formatarg=formatarg, formatvalue=formatvalue):
 *         return formatarg(name) + formatvalue(locals[name])
 */
  __pyx_tuple__42 = PyTuple_Pack(4, __pyx_n_s_name_2, __pyx_n_s_locals_2, __pyx_n_s_formatarg, __pyx_n_s_formatvalue); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_convert, 1294, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 1294, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1406
 *                 pass
 *             else:
 *                 raise ValueError('too many values to unpack')             # <<<<<<<<<<<<<<
 * 
 *     def is_assigned(arg):
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_u_too_many_values_to_unpack); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "Aspidites/_vendor/contracts/inspection.py":1373
 * 
 *     if kwonlyargs:
 *         raise ValueError("I'm sorry, I don't have the logic to use kwonlyargs. "             # <<<<<<<<<<<<<<
 *                          "Perhapse you can help PyContracts and implement this? Thanks.")
 * 
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_u_I_m_sorry_I_don_t_have_the_logic); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "Aspidites/_vendor/contracts/inspection.py":1383
 *     assigned_tuple_params = []
 * 
 *     def assign(arg, value):             # <<<<<<<<<<<<<<
 *         # noinspection PyUnresolvedReferences
 *         arg: cython.p_char
 */
  __pyx_tuple__47 = PyTuple_Pack(5, __pyx_n_s_arg, __pyx_n_s_value, __pyx_n_s_i, __pyx_n_s_subarg, __pyx_n_s_subvalue); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_assign, 1383, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 1383, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1408
 *                 raise ValueError('too many values to unpack')
 * 
 *     def is_assigned(arg):             # <<<<<<<<<<<<<<
 *         if isinstance(arg, str):
 *             return arg in arg2value
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_is_assigned, 1408, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 1408, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1902
 * 
 *     if not params or params[0].kind in (_VAR_KEYWORD, _KEYWORD_ONLY):
 *         raise ValueError('invalid method signature')             # <<<<<<<<<<<<<<
 * 
 *     kind = params[0].kind
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_u_invalid_method_signature); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 1902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "Aspidites/_vendor/contracts/inspection.py":1913
 *             # Unless we add a new parameter type we never
 *             # get here
 *             raise ValueError('invalid argument type')             # <<<<<<<<<<<<<<
 *         # It's a var-positional parameter.
 *         # Do nothing. '(*args[, ...])' -> '(*args[, ...])'
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_u_invalid_argument_type); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 1913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "Aspidites/_vendor/contracts/inspection.py":2093
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 *             raise ValueError("Annotations are not currently supported")             # <<<<<<<<<<<<<<
 *         return node.arg
 * 
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_u_Annotations_are_not_currently_su); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "Aspidites/_vendor/contracts/inspection.py":2090
 *     sys_module_dict = sys.modules.copy()
 * 
 *     def parse_name(node):             # <<<<<<<<<<<<<<
 *         assert isinstance(node, ast.arg)
 *         if node.annotation is not None:
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_n_s_node); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 2090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_parse_name, 2090, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 2090, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2096
 *         return node.arg
 * 
 *     def wrap_value(s):             # <<<<<<<<<<<<<<
 *         try:
 *             value = eval(s, module_dict)
 */
  __pyx_tuple__63 = PyTuple_Pack(2, __pyx_n_s_s, __pyx_n_s_value); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 2096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_wrap_value, 2096, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 2096, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2110
 * 
 *     class RewriteSymbolics(ast.NodeTransformer):
 *         def visit_Attribute(self, node):             # <<<<<<<<<<<<<<
 *             a = []
 *             n = node
 */
  __pyx_tuple__65 = PyTuple_Pack(5, __pyx_n_s_self_2, __pyx_n_s_node, __pyx_n_s_a, __pyx_n_s_n, __pyx_n_s_value); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 2110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_visit_Attribute, 2110, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 2110, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2122
 *             return wrap_value(value)
 * 
 *         def visit_Name(self, node):             # <<<<<<<<<<<<<<
 *             if not isinstance(node.ctx, ast.Load):
 *                 raise ValueError()
 */
  __pyx_tuple__67 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_node); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_visit_Name, 2122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 2122, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2127
 *             return wrap_value(node.id)
 * 
 *     def p(name_node, default_node, default=empty):             # <<<<<<<<<<<<<<
 *         name = parse_name(name_node)
 *         if name is invalid:
 */
  __pyx_tuple__69 = PyTuple_Pack(5, __pyx_n_s_name_node, __pyx_n_s_default_node, __pyx_n_s_default, __pyx_n_s_name_2, __pyx_n_s_o); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_p, 2127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 2127, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2379
 *                 sigcls=sigcls)
 * 
 *             sig = _signature_get_partial(wrapped_sig, partialmethod, (None,))             # <<<<<<<<<<<<<<
 *             first_wrapped_param = tuple(wrapped_sig.parameters.values())[0]
 *             if first_wrapped_param.kind is Parameter.VAR_POSITIONAL:
 */
  __pyx_tuple__71 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 2379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "Aspidites/_vendor/contracts/inspection.py":2596
 * 
 *         if name is _empty:
 *             raise ValueError('name is a required attribute for Parameter')             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(name, str):
 */
  __pyx_tuple__72 = PyTuple_Pack(1, __pyx_kp_u_name_is_a_required_attribute_for); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 2596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */
  __pyx_tuple__78 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "Aspidites/_vendor/contracts/inspection.py":2926
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_function() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __pyx_tuple__79 = PyTuple_Pack(2, __pyx_kp_u_inspect_Signature_from_function, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 2926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "Aspidites/_vendor/contracts/inspection.py":2938
 *         """
 * 
 *         warnings.warn("inspect.Signature.from_builtin() is deprecated since "             # <<<<<<<<<<<<<<
 *                       "Python 3.5, use Signature.from_callable()",
 *                       DeprecationWarning, stacklevel=2)
 */
  __pyx_tuple__80 = PyTuple_Pack(2, __pyx_kp_u_inspect_Signature_from_builtin_i, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 2938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "Aspidites/_vendor/contracts/inspection.py":3055
 *                         # Looks like we have no parameter for this positional
 *                         # argument
 *                         raise TypeError(             # <<<<<<<<<<<<<<
 *                             'too many positional arguments') from None
 * 
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_u_too_many_positional_arguments); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 3055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "Aspidites/_vendor/contracts/inspection.py":3212
 *         else:
 *             f = callable_thing
 *         args = ('test',)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_n_u_test); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 3212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "Aspidites/_vendor/contracts/inspection.py":3220
 *         return False, str(e)
 *     else:
 *         return True, None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__86 = PyTuple_Pack(2, Py_True, Py_None); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 3220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "Aspidites/_vendor/contracts/inspection.py":74
 * 
 * # ----------------------------------------------------------- type-checking
 * def ismodule(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a module.
 * 
 */
  __pyx_tuple__87 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_ismodule, 74, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 74, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":84
 * 
 * 
 * def isclass(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a class.
 * 
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__89, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isclass, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":93
 * 
 * 
 * def ismethod(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an instance method.
 * 
 */
  __pyx_tuple__91 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_ismethod, 93, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 93, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":104
 * 
 * 
 * def ismethoddescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a method descriptor.
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(2, __pyx_n_s_obj, __pyx_n_s_tp); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_ismethoddescriptor, 104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 104, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":125
 * 
 * 
 * def isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a data descriptor.
 * 
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_s_obj, __pyx_n_s_tp); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isdatadescriptor, 125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 125, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":142
 * if hasattr(types, 'MemberDescriptorType'):
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */
  __pyx_tuple__97 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_ismemberdescriptor, 142, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 142, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":150
 * else:
 *     # Other implementations
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */
  __pyx_tuple__99 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_ismemberdescriptor, 150, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 150, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":159
 * if hasattr(types, 'GetSetDescriptorType'):
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */
  __pyx_tuple__101 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isgetsetdescriptor, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 159, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":167
 * else:
 *     # Other implementations
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */
  __pyx_tuple__103 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isgetsetdescriptor, 167, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 167, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":175
 * 
 * 
 * def isfunction(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined function.
 * 
 */
  __pyx_tuple__105 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isfunction, 175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 175, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":189
 * 
 * 
 * def _has_code_flag(f, flag):             # <<<<<<<<<<<<<<
 *     """Return true if ``f`` is a function (or a method or functools.partial
 *     wrapper wrapping a function) whose code object has the given ``flag``
 */
  __pyx_tuple__107 = PyTuple_Pack(2, __pyx_n_s_f, __pyx_n_s_flag); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_has_code_flag, 189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 189, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":201
 * 
 * 
 * def isgeneratorfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined generator function.
 * 
 */
  __pyx_tuple__109 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__109, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isgeneratorfunction, 201, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 201, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":209
 * 
 * 
 * def iscoroutinefunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine function.
 * 
 */
  __pyx_tuple__111 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_iscoroutinefunction, 209, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 209, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":217
 * 
 * 
 * def isasyncgenfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator function.
 * 
 */
  __pyx_tuple__113 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__113, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isasyncgenfunction, 217, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 217, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":226
 * 
 * 
 * def isasyncgen(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator."""
 *     return isinstance(object, types.AsyncGeneratorType)
 */
  __pyx_tuple__115 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__115, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isasyncgen, 226, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 226, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":231
 * 
 * 
 * def isgenerator(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a generator.
 * 
 */
  __pyx_tuple__117 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isgenerator, 231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 231, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":249
 * 
 * 
 * def iscoroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine."""
 *     return isinstance(object, types.CoroutineType)
 */
  __pyx_tuple__119 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_iscoroutine, 249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(0, 249, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":254
 * 
 * 
 * def isawaitable(object):             # <<<<<<<<<<<<<<
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or
 */
  __pyx_tuple__121 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__121, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isawaitable, 254, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 254, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":262
 * 
 * 
 * def istraceback(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a traceback.
 * 
 */
  __pyx_tuple__123 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_istraceback, 262, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 262, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":273
 * 
 * 
 * def isframe(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a frame object.
 * 
 */
  __pyx_tuple__125 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isframe, 273, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(0, 273, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":288
 * 
 * 
 * def iscode(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a code object.
 * 
 */
  __pyx_tuple__127 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_iscode, 288, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 288, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":314
 * 
 * 
 * def isbuiltin(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a built-in function or method.
 * 
 */
  __pyx_tuple__129 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isbuiltin, 314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(0, 314, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":324
 * 
 * 
 * def isroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 */
  __pyx_tuple__131 = PyTuple_Pack(1, __pyx_n_s_object); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__131, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isroutine, 324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":332
 * 
 * 
 * def isabstract(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):
 */
  __pyx_tuple__133 = PyTuple_Pack(4, __pyx_n_s_object, __pyx_n_s_name_2, __pyx_n_s_value, __pyx_n_s_base); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_isabstract, 332, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(0, 332, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":357
 * 
 * 
 * def getmembers(object, predicate=None):             # <<<<<<<<<<<<<<
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 */
  __pyx_tuple__135 = PyTuple_Pack(11, __pyx_n_s_object, __pyx_n_s_predicate, __pyx_n_s_mro_2, __pyx_n_s_results, __pyx_n_s_processed, __pyx_n_s_names, __pyx_n_s_base, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getmembers, 357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_tuple__137 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "Aspidites/_vendor/contracts/inspection.py":402
 * 
 * 
 * Attribute = namedtuple('Attribute', 'name kind defining_class object')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__138 = PyTuple_Pack(2, __pyx_n_u_Attribute, __pyx_kp_u_name_kind_defining_class_object); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */
  __pyx_tuple__139 = PyTuple_Pack(23, __pyx_n_s_cls, __pyx_n_s_mro_2, __pyx_n_s_metamro, __pyx_n_s_class_bases, __pyx_n_s_all_bases, __pyx_n_s_names, __pyx_n_s_base, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_result, __pyx_n_s_processed, __pyx_n_s_name_2, __pyx_n_s_homecls, __pyx_n_s_get_obj, __pyx_n_s_dict_obj, __pyx_n_s_exc, __pyx_n_s_last_cls, __pyx_n_s_srch_cls, __pyx_n_s_srch_obj, __pyx_n_s_obj, __pyx_n_s_kind, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(1, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_classify_class_attrs, 405, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 405, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":522
 * # ----------------------------------------------------------- class helpers
 * 
 * def getmro(cls):             # <<<<<<<<<<<<<<
 *     """Return tuple of base classes (including cls) in method resolution order."""
 *     return cls.__mro__
 */
  __pyx_tuple__141 = PyTuple_Pack(1, __pyx_n_s_cls); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getmro, 522, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 522, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */
  __pyx_tuple__143 = PyTuple_Pack(9, __pyx_n_s_func_2, __pyx_n_s_stop, __pyx_n_s_is_wrapper, __pyx_n_s_is_wrapper, __pyx_n_s_is_wrapper, __pyx_n_s_f, __pyx_n_s_memo, __pyx_n_s_recursion_limit, __pyx_n_s_id_func); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(1, 1, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_unwrap, 529, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 529, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":566
 * 
 * # -------------------------------------------------- source code extraction
 * def indentsize(line):             # <<<<<<<<<<<<<<
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()
 */
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_n_s_line, __pyx_n_s_expline); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_indentsize, 566, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 566, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":572
 * 
 * 
 * def _findclass(func):             # <<<<<<<<<<<<<<
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:
 */
  __pyx_tuple__147 = PyTuple_Pack(3, __pyx_n_s_func_2, __pyx_n_s_cls, __pyx_n_s_name_2); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_findclass, 572, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(0, 572, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":583
 * 
 * 
 * def _finddoc(obj):             # <<<<<<<<<<<<<<
 *     if isclass(obj):
 *         for base in obj.__mro__:
 */
  __pyx_tuple__149 = PyTuple_Pack(8, __pyx_n_s_obj, __pyx_n_s_base, __pyx_n_s_doc_2, __pyx_n_s_name_2, __pyx_n_s_self_2, __pyx_n_s_cls, __pyx_n_s_func_2, __pyx_n_s_slots_2); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_finddoc, 583, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(0, 583, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":646
 * 
 * 
 * def getdoc(object):             # <<<<<<<<<<<<<<
 *     """Get the documentation string for an object.
 * 
 */
  __pyx_tuple__151 = PyTuple_Pack(2, __pyx_n_s_object, __pyx_n_s_doc_2); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__151, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getdoc, 646, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":666
 * 
 * 
 * def cleandoc(doc):             # <<<<<<<<<<<<<<
 *     """Clean up indentation from docstrings.
 * 
 */
  __pyx_tuple__153 = PyTuple_Pack(7, __pyx_n_s_doc_2, __pyx_n_s_lines, __pyx_n_s_margin, __pyx_n_s_line, __pyx_n_s_content, __pyx_n_s_indent, __pyx_n_s_i); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_cleandoc, 666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(0, 666, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":696
 * 
 * 
 * def getfile(object):             # <<<<<<<<<<<<<<
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 */
  __pyx_tuple__155 = PyTuple_Pack(2, __pyx_n_s_object, __pyx_n_s_module_3); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getfile, 696, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(0, 696, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":723
 * 
 * 
 * def getmodulename(path):             # <<<<<<<<<<<<<<
 *     """Return the module name for a given file, or None."""
 *     fname = os.path.basename(path)
 */
  __pyx_tuple__157 = PyTuple_Pack(6, __pyx_n_s_path, __pyx_n_s_fname, __pyx_n_s_suffixes, __pyx_n_s_neglen, __pyx_n_s_suffix, __pyx_n_s_suffix); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getmodulename, 723, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(0, 723, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */
  __pyx_tuple__159 = PyTuple_Pack(6, __pyx_n_s_object, __pyx_n_s_filename_2, __pyx_n_s_all_bytecode_suffixes, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getsourcefile, 736, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(0, 736, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":759
 * 
 * 
 * def getabsfile(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return an absolute path to the source or compiled file for an object.
 * 
 */
  __pyx_tuple__161 = PyTuple_Pack(2, __pyx_n_s_object, __pyx_n_s_filename); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__161, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getabsfile, 759, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(0, 759, __pyx_L1_error)
  __pyx_tuple__163 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "Aspidites/_vendor/contracts/inspection.py":773
 * 
 * 
 * def getmodule(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):
 */
  __pyx_tuple__164 = PyTuple_Pack(10, __pyx_n_s_object, __pyx_n_s_filename, __pyx_n_s_file_2, __pyx_n_s_modname, __pyx_n_s_module_3, __pyx_n_s_f, __pyx_n_s_main_2, __pyx_n_s_mainobject, __pyx_n_s_builtin, __pyx_n_s_builtinobject); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__164, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getmodule, 773, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_tuple__166 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "Aspidites/_vendor/contracts/inspection.py":826
 * class _ClassFinder(ast.NodeVisitor):
 * 
 *     def __init__(self, qualname):             # <<<<<<<<<<<<<<
 *         self.stack = []
 *         self.qualname = qualname
 */
  __pyx_tuple__167 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_qualname_2); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_init, 826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(0, 826, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":830
 *         self.qualname = qualname
 * 
 *     def visit_FunctionDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')
 */
  __pyx_tuple__169 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_node); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_visit_FunctionDef, 830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(0, 830, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":839
 *     visit_AsyncFunctionDef = visit_FunctionDef
 * 
 *     def visit_ClassDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):
 */
  __pyx_tuple__171 = PyTuple_Pack(3, __pyx_n_s_self_2, __pyx_n_s_node, __pyx_n_s_line_number); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__171, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_visit_ClassDef, 839, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(0, 839, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":855
 * 
 * 
 * def findsource(object):             # <<<<<<<<<<<<<<
 *     """Return the entire source file and starting line number for an object.
 * 
 */
  __pyx_tuple__173 = PyTuple_Pack(13, __pyx_n_s_object, __pyx_n_s_file_2, __pyx_n_s_module_3, __pyx_n_s_lines, __pyx_n_s_qualname_2, __pyx_n_s_source, __pyx_n_s_tree, __pyx_n_s_class_finder, __pyx_n_s_e, __pyx_n_s_line_number, __pyx_n_s_lnum, __pyx_n_s_pat, __pyx_n_s_line); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(1, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__173, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_findsource, 855, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(0, 855, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":924
 * 
 * 
 * def getcomments(object):             # <<<<<<<<<<<<<<
 *     """Get lines of comments immediately preceding an object's source code.
 * 
 */
  __pyx_tuple__175 = PyTuple_Pack(8, __pyx_n_s_object, __pyx_n_s_lines, __pyx_n_s_lnum, __pyx_n_s_start, __pyx_n_s_comments, __pyx_n_s_end, __pyx_n_s_indent, __pyx_n_s_comment); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__175, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getcomments, 924, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(0, 924, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":977
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.indent = 0
 *         self.islambda = False
 */
  __pyx_tuple__177 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__177, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_init, 977, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(0, 977, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":987
 *         self.body_col0 = None
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):             # <<<<<<<<<<<<<<
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 */
  __pyx_tuple__179 = PyTuple_Pack(6, __pyx_n_s_self_2, __pyx_n_s_type, __pyx_n_s_token, __pyx_n_s_srowcol, __pyx_n_s_erowcol, __pyx_n_s_line); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__179, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_tokeneater, 987, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(0, 987, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1038
 * 
 * 
 * def getblock(lines):             # <<<<<<<<<<<<<<
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 */
  __pyx_tuple__181 = PyTuple_Pack(4, __pyx_n_s_lines, __pyx_n_s_blockfinder, __pyx_n_s_tokens, __pyx_n_s_token_2); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__181, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getblock, 1038, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(0, 1038, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1050
 * 
 * 
 * def getsourcelines(object):             # <<<<<<<<<<<<<<
 *     """Return a list of source lines and starting line number for an object.
 * 
 */
  __pyx_tuple__183 = PyTuple_Pack(3, __pyx_n_s_object, __pyx_n_s_lines, __pyx_n_s_lnum); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getsourcelines, 1050, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(0, 1050, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1072
 * 
 * 
 * def getsource(object):             # <<<<<<<<<<<<<<
 *     """Return the text of the source code for an object.
 * 
 */
  __pyx_tuple__185 = PyTuple_Pack(3, __pyx_n_s_object, __pyx_n_s_lines, __pyx_n_s_lnum); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);
  __pyx_codeobj__186 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getsource, 1072, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__186)) __PYX_ERR(0, 1072, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1083
 * 
 * # --------------------------------------------------- class tree extraction
 * def walktree(classes, children, parent):             # <<<<<<<<<<<<<<
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 */
  __pyx_tuple__187 = PyTuple_Pack(5, __pyx_n_s_classes, __pyx_n_s_children, __pyx_n_s_parent, __pyx_n_s_results, __pyx_n_s_c); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);
  __pyx_codeobj__188 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__187, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_walktree, 1083, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__188)) __PYX_ERR(0, 1083, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1094
 * 
 * 
 * def getclasstree(classes, unique=False):             # <<<<<<<<<<<<<<
 *     """Arrange the given list of classes into a hierarchy of nested lists.
 * 
 */
  __pyx_tuple__189 = PyTuple_Pack(6, __pyx_n_s_classes, __pyx_n_s_unique, __pyx_n_s_children, __pyx_n_s_roots, __pyx_n_s_c, __pyx_n_s_parent); if (unlikely(!__pyx_tuple__189)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__189);
  __Pyx_GIVEREF(__pyx_tuple__189);
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__189, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getclasstree, 1094, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_tuple__191 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__191)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__191);
  __Pyx_GIVEREF(__pyx_tuple__191);

  /* "Aspidites/_vendor/contracts/inspection.py":1122
 * 
 * # ------------------------------------------------ argument list extraction
 * Arguments = namedtuple('Arguments', 'args, varargs, varkw')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__192 = PyTuple_Pack(2, __pyx_n_u_Arguments, __pyx_kp_u_args_varargs_varkw); if (unlikely(!__pyx_tuple__192)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__192);
  __Pyx_GIVEREF(__pyx_tuple__192);

  /* "Aspidites/_vendor/contracts/inspection.py":1125
 * 
 * 
 * def getargs(co):             # <<<<<<<<<<<<<<
 *     """Get information about the arguments accepted by a code object.
 * 
 */
  __pyx_tuple__193 = PyTuple_Pack(9, __pyx_n_s_co, __pyx_n_s_names, __pyx_n_s_nargs, __pyx_n_s_nkwargs, __pyx_n_s_args, __pyx_n_s_kwonlyargs, __pyx_n_s_step, __pyx_n_s_varargs, __pyx_n_s_varkw); if (unlikely(!__pyx_tuple__193)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__193);
  __Pyx_GIVEREF(__pyx_tuple__193);
  __pyx_codeobj__194 = (PyObject*)__Pyx_PyCode_New(1, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__193, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getargs, 1125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__194)) __PYX_ERR(0, 1125, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1152
 *     return Arguments(args + kwonlyargs, varargs, varkw)
 * 
 * ArgSpec = namedtuple('ArgSpec', 'args varargs keywords defaults')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__195 = PyTuple_Pack(2, __pyx_n_u_ArgSpec, __pyx_kp_u_args_varargs_keywords_defaults); if (unlikely(!__pyx_tuple__195)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__195);
  __Pyx_GIVEREF(__pyx_tuple__195);

  /* "Aspidites/_vendor/contracts/inspection.py":1155
 * 
 * 
 * def getargspec(func):             # <<<<<<<<<<<<<<
 *     """Get the names and default values of a function's parameters.
 * 
 */
  __pyx_tuple__196 = PyTuple_Pack(8, __pyx_n_s_func_2, __pyx_n_s_args, __pyx_n_s_varargs, __pyx_n_s_varkw, __pyx_n_s_defaults, __pyx_n_s_kwonlyargs, __pyx_n_s_kwonlydefaults, __pyx_n_s_ann); if (unlikely(!__pyx_tuple__196)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__196);
  __Pyx_GIVEREF(__pyx_tuple__196);
  __pyx_codeobj__197 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__196, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getargspec, 1155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__197)) __PYX_ERR(0, 1155, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1186
 * 
 * 
 * FullArgSpec = namedtuple('FullArgSpec', 'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations')             # <<<<<<<<<<<<<<
 * 
 * ArgInfo = namedtuple('ArgInfo', 'args varargs keywords locals')
 */
  __pyx_tuple__198 = PyTuple_Pack(2, __pyx_n_u_FullArgSpec, __pyx_kp_u_args_varargs_varkw_defaults_kwon); if (unlikely(!__pyx_tuple__198)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__198);
  __Pyx_GIVEREF(__pyx_tuple__198);

  /* "Aspidites/_vendor/contracts/inspection.py":1188
 * FullArgSpec = namedtuple('FullArgSpec', 'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations')
 * 
 * ArgInfo = namedtuple('ArgInfo', 'args varargs keywords locals')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__199 = PyTuple_Pack(2, __pyx_n_u_ArgInfo, __pyx_kp_u_args_varargs_keywords_locals); if (unlikely(!__pyx_tuple__199)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__199);
  __Pyx_GIVEREF(__pyx_tuple__199);

  /* "Aspidites/_vendor/contracts/inspection.py":1191
 * 
 * 
 * def getargvalues(frame):             # <<<<<<<<<<<<<<
 *     """Get information about arguments passed into a particular frame.
 * 
 */
  __pyx_tuple__200 = PyTuple_Pack(4, __pyx_n_s_frame, __pyx_n_s_args, __pyx_n_s_varargs, __pyx_n_s_varkw); if (unlikely(!__pyx_tuple__200)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__200);
  __Pyx_GIVEREF(__pyx_tuple__200);
  __pyx_codeobj__201 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__200, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getargvalues, 1191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__201)) __PYX_ERR(0, 1191, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1202
 * 
 * 
 * def formatannotation(annotation, base_module=None):             # <<<<<<<<<<<<<<
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 */
  __pyx_tuple__202 = PyTuple_Pack(2, __pyx_n_s_annotation, __pyx_n_s_base_module); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);
  __pyx_codeobj__203 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatannotation, 1202, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__203)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __pyx_tuple__204 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__204)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__204);
  __Pyx_GIVEREF(__pyx_tuple__204);

  /* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */
  __pyx_tuple__205 = PyTuple_Pack(4, __pyx_n_s_object, __pyx_n_s_module_3, __pyx_n_s_formatannotation_2, __pyx_n_s_formatannotation_2); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);
  __pyx_codeobj__206 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__205, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatannotationrelativeto, 1212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__206)) __PYX_ERR(0, 1212, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */
  __pyx_tuple__207 = PyTuple_Pack(23, __pyx_n_s_args, __pyx_n_s_varargs, __pyx_n_s_varkw, __pyx_n_s_defaults, __pyx_n_s_kwonlyargs, __pyx_n_s_kwonlydefaults, __pyx_n_s_annotations, __pyx_n_s_formatarg, __pyx_n_s_formatvarargs, __pyx_n_s_formatvarkw, __pyx_n_s_formatvalue, __pyx_n_s_formatreturns, __pyx_n_s_formatannotation, __pyx_n_s_warn, __pyx_n_s_formatargandannotation, __pyx_n_s_formatargandannotation, __pyx_n_s_specs, __pyx_n_s_firstdefault, __pyx_n_s_i, __pyx_n_s_arg, __pyx_n_s_spec, __pyx_n_s_kwonlyarg, __pyx_n_s_result); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(13, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatargspec, 1221, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(0, 1221, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1282
 * 
 * 
 * def formatargvalues(args, varargs, varkw, locals,             # <<<<<<<<<<<<<<
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 */
  __pyx_tuple__209 = PyTuple_Pack(12, __pyx_n_s_args, __pyx_n_s_varargs, __pyx_n_s_varkw, __pyx_n_s_locals_2, __pyx_n_s_formatarg, __pyx_n_s_formatvarargs, __pyx_n_s_formatvarkw, __pyx_n_s_formatvalue, __pyx_n_s_convert, __pyx_n_s_convert, __pyx_n_s_specs, __pyx_n_s_i); if (unlikely(!__pyx_tuple__209)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__209);
  __Pyx_GIVEREF(__pyx_tuple__209);
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(8, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_formatargvalues, 1282, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(0, 1282, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1308
 * 
 * 
 * def _missing_arguments(f_name, argnames, pos, values):             # <<<<<<<<<<<<<<
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 */
  __pyx_tuple__211 = PyTuple_Pack(9, __pyx_n_s_f_name, __pyx_n_s_argnames, __pyx_n_s_pos, __pyx_n_s_values, __pyx_n_s_names, __pyx_n_s_missing, __pyx_n_s_s, __pyx_n_s_tail, __pyx_n_s_name_2); if (unlikely(!__pyx_tuple__211)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__211);
  __Pyx_GIVEREF(__pyx_tuple__211);
  __pyx_codeobj__212 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__211, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_missing_arguments, 1308, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__212)) __PYX_ERR(0, 1308, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1325
 * 
 * 
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):             # <<<<<<<<<<<<<<
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 */
  __pyx_tuple__213 = PyTuple_Pack(14, __pyx_n_s_f_name, __pyx_n_s_args, __pyx_n_s_kwonly, __pyx_n_s_varargs, __pyx_n_s_defcount, __pyx_n_s_given, __pyx_n_s_values, __pyx_n_s_atleast, __pyx_n_s_kwonly_given, __pyx_n_s_plural, __pyx_n_s_sig, __pyx_n_s_kwonly_sig, __pyx_n_s_msg, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__213)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__213);
  __Pyx_GIVEREF(__pyx_tuple__213);
  __pyx_codeobj__214 = (PyObject*)__Pyx_PyCode_New(7, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__213, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_too_many, 1325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__214)) __PYX_ERR(0, 1325, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */
  __pyx_tuple__215 = PyTuple_Pack(29, __pyx_n_s_func_2, __pyx_n_s_positional, __pyx_n_s_named, __pyx_n_s_f_name, __pyx_n_s_assigned_tuple_params, __pyx_n_s_arg2dict, __pyx_n_s_num_args, __pyx_n_s_num_pos, __pyx_n_s_num_total, __pyx_n_s_num_defaults, __pyx_n_s_num_required, __pyx_n_s_args, __pyx_n_s_varargs, __pyx_n_s_varkw, __pyx_n_s_defaults, __pyx_n_s_kwonlyargs, __pyx_n_s_kwonlydefaults, __pyx_n_s_annotations, __pyx_n_s_arg2value, __pyx_n_s_assign, __pyx_n_s_assign, __pyx_n_s_is_assigned, __pyx_n_s_is_assigned, __pyx_n_s_im_self, __pyx_n_s_arg, __pyx_n_s_value, __pyx_n_s_unexpected, __pyx_n_s_unassigned, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__215)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__215);
  __Pyx_GIVEREF(__pyx_tuple__215);
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(1, 0, 29, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getcallargs, 1347, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(0, 1347, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1464
 * 
 * 
 * ClosureVars = namedtuple('ClosureVars', 'nonlocals globals builtins unbound')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__217 = PyTuple_Pack(2, __pyx_n_u_ClosureVars, __pyx_kp_u_nonlocals_globals_builtins_unbou); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);

  /* "Aspidites/_vendor/contracts/inspection.py":1467
 * 
 * 
 * def getclosurevars(func):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of free variables to their current values.
 */
  __pyx_tuple__218 = PyTuple_Pack(11, __pyx_n_s_func_2, __pyx_n_s_code_2, __pyx_n_s_nonlocal_vars, __pyx_n_s_global_ns, __pyx_n_s_builtin_ns, __pyx_n_s_global_vars, __pyx_n_s_builtin_vars, __pyx_n_s_unbound_names, __pyx_n_s_name_2, __pyx_n_s_var, __pyx_n_s_cell); if (unlikely(!__pyx_tuple__218)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__218);
  __Pyx_GIVEREF(__pyx_tuple__218);
  __pyx_codeobj__219 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__218, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getclosurevars, 1467, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__219)) __PYX_ERR(0, 1467, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1521
 * # -------------------------------------------------- stack frame extraction
 * 
 * Traceback = namedtuple('Traceback', 'filename lineno function code_context index')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__220 = PyTuple_Pack(2, __pyx_n_u_Traceback, __pyx_kp_u_filename_lineno_function_code_co); if (unlikely(!__pyx_tuple__220)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__220);
  __Pyx_GIVEREF(__pyx_tuple__220);

  /* "Aspidites/_vendor/contracts/inspection.py":1524
 * 
 * 
 * def getframeinfo(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get information about a frame or traceback object.
 * 
 */
  __pyx_tuple__221 = PyTuple_Pack(8, __pyx_n_s_frame, __pyx_n_s_context, __pyx_n_s_lineno, __pyx_n_s_filename_2, __pyx_n_s_start, __pyx_n_s_lines, __pyx_n_s_lnum, __pyx_n_s_index); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getframeinfo, 1524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __pyx_tuple__223 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__223)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__223);
  __Pyx_GIVEREF(__pyx_tuple__223);

  /* "Aspidites/_vendor/contracts/inspection.py":1557
 * 
 * 
 * def getlineno(frame):             # <<<<<<<<<<<<<<
 *     """Get the line number from a frame object, allowing for optimization."""
 *     # FrameType.f_lineno is now a descriptor that grovels co_lnotab
 */
  __pyx_tuple__224 = PyTuple_Pack(1, __pyx_n_s_frame); if (unlikely(!__pyx_tuple__224)) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__224);
  __Pyx_GIVEREF(__pyx_tuple__224);
  __pyx_codeobj__225 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__224, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getlineno, 1557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__225)) __PYX_ERR(0, 1557, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1563
 * 
 * 
 * FrameInfo = namedtuple('FrameInfo', ('frame',) + Traceback._fields)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__226 = PyTuple_Pack(1, __pyx_n_u_frame); if (unlikely(!__pyx_tuple__226)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__226);
  __Pyx_GIVEREF(__pyx_tuple__226);

  /* "Aspidites/_vendor/contracts/inspection.py":1566
 * 
 * 
 * def getouterframes(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a frame and all higher (calling) frames.
 * 
 */
  __pyx_tuple__227 = PyTuple_Pack(4, __pyx_n_s_frame, __pyx_n_s_context, __pyx_n_s_framelist, __pyx_n_s_frameinfo); if (unlikely(!__pyx_tuple__227)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__227);
  __Pyx_GIVEREF(__pyx_tuple__227);
  __pyx_codeobj__228 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__227, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getouterframes, 1566, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__228)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __pyx_tuple__229 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__229)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__229);
  __Pyx_GIVEREF(__pyx_tuple__229);

  /* "Aspidites/_vendor/contracts/inspection.py":1579
 * 
 * 
 * def getinnerframes(tb, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a traceback's frame and all lower frames.
 * 
 */
  __pyx_tuple__230 = PyTuple_Pack(4, __pyx_n_s_tb, __pyx_n_s_context, __pyx_n_s_framelist, __pyx_n_s_frameinfo); if (unlikely(!__pyx_tuple__230)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__230);
  __Pyx_GIVEREF(__pyx_tuple__230);
  __pyx_codeobj__231 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__230, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getinnerframes, 1579, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__231)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __pyx_tuple__232 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__232)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__232);
  __Pyx_GIVEREF(__pyx_tuple__232);

  /* "Aspidites/_vendor/contracts/inspection.py":1592
 * 
 * 
 * def currentframe():             # <<<<<<<<<<<<<<
 *     """Return the frame of the caller or None if this is not possible."""
 *     return sys._getframe(1) if hasattr(sys, "_getframe") else None
 */
  __pyx_codeobj__233 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_currentframe, 1592, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__233)) __PYX_ERR(0, 1592, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1597
 * 
 * 
 * def stack(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack above the caller's frame."""
 *     # noinspection PyProtectedMember, PyUnresolvedReferences
 */
  __pyx_tuple__234 = PyTuple_Pack(1, __pyx_n_s_context); if (unlikely(!__pyx_tuple__234)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__234);
  __Pyx_GIVEREF(__pyx_tuple__234);
  __pyx_codeobj__235 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__234, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_stack, 1597, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__235)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __pyx_tuple__236 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__236)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__236);
  __Pyx_GIVEREF(__pyx_tuple__236);

  /* "Aspidites/_vendor/contracts/inspection.py":1603
 * 
 * 
 * def trace(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack below the current exception."""
 *     # noinspection PyTypeChecker
 */
  __pyx_tuple__237 = PyTuple_Pack(1, __pyx_n_s_context); if (unlikely(!__pyx_tuple__237)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__237);
  __Pyx_GIVEREF(__pyx_tuple__237);
  __pyx_codeobj__238 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__237, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_trace, 1603, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__238)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __pyx_tuple__239 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__239)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__239);
  __Pyx_GIVEREF(__pyx_tuple__239);

  /* "Aspidites/_vendor/contracts/inspection.py":1614
 * 
 * 
 * def _static_getmro(klass):             # <<<<<<<<<<<<<<
 *     return type.__dict__['__mro__'].__get__(klass)
 * 
 */
  __pyx_tuple__240 = PyTuple_Pack(1, __pyx_n_s_klass); if (unlikely(!__pyx_tuple__240)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__240);
  __Pyx_GIVEREF(__pyx_tuple__240);
  __pyx_codeobj__241 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_static_getmro, 1614, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__241)) __PYX_ERR(0, 1614, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1618
 * 
 * 
 * def _check_instance(obj, attr):             # <<<<<<<<<<<<<<
 *     instance_dict = {}
 *     try:
 */
  __pyx_tuple__242 = PyTuple_Pack(3, __pyx_n_s_obj, __pyx_n_s_attr, __pyx_n_s_instance_dict); if (unlikely(!__pyx_tuple__242)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__242);
  __Pyx_GIVEREF(__pyx_tuple__242);
  __pyx_codeobj__243 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__242, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_check_instance, 1618, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__243)) __PYX_ERR(0, 1618, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1627
 * 
 * 
 * def _check_class(klass, attr):             # <<<<<<<<<<<<<<
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 */
  __pyx_tuple__244 = PyTuple_Pack(3, __pyx_n_s_klass, __pyx_n_s_attr, __pyx_n_s_entry); if (unlikely(!__pyx_tuple__244)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__244);
  __Pyx_GIVEREF(__pyx_tuple__244);
  __pyx_codeobj__245 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__244, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_check_class, 1627, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__245)) __PYX_ERR(0, 1627, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1637
 * 
 * 
 * def _is_type(obj):             # <<<<<<<<<<<<<<
 *     try:
 *         _static_getmro(obj)
 */
  __pyx_tuple__246 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__246)) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__246);
  __Pyx_GIVEREF(__pyx_tuple__246);
  __pyx_codeobj__247 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__246, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_is_type, 1637, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__247)) __PYX_ERR(0, 1637, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1645
 * 
 * 
 * def _shadowed_dict(klass):             # <<<<<<<<<<<<<<
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 */
  __pyx_tuple__248 = PyTuple_Pack(4, __pyx_n_s_klass, __pyx_n_s_dict_attr, __pyx_n_s_entry, __pyx_n_s_class_dict); if (unlikely(!__pyx_tuple__248)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__248);
  __Pyx_GIVEREF(__pyx_tuple__248);
  __pyx_codeobj__249 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__248, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_shadowed_dict, 1645, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__249)) __PYX_ERR(0, 1645, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1660
 * 
 * 
 * def getattr_static(obj, attr, default=_sentinel):             # <<<<<<<<<<<<<<
 *     """Retrieve attributes without triggering dynamic lookup via the
 *        descriptor protocol,  __getattr__ or __getattribute__.
 */
  __pyx_tuple__250 = PyTuple_Pack(8, __pyx_n_s_obj, __pyx_n_s_attr, __pyx_n_s_default, __pyx_n_s_instance_result, __pyx_n_s_klass, __pyx_n_s_dict_attr, __pyx_n_s_klass_result, __pyx_n_s_entry); if (unlikely(!__pyx_tuple__250)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__250);
  __Pyx_GIVEREF(__pyx_tuple__250);
  __pyx_codeobj__251 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__250, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getattr_static, 1660, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__251)) __PYX_ERR(0, 1660, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1714
 * 
 * 
 * def getgeneratorstate(generator):             # <<<<<<<<<<<<<<
 *     """Get current state of a generator-iterator.
 * 
 */
  __pyx_tuple__252 = PyTuple_Pack(1, __pyx_n_s_generator); if (unlikely(!__pyx_tuple__252)) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__252);
  __Pyx_GIVEREF(__pyx_tuple__252);
  __pyx_codeobj__253 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getgeneratorstate, 1714, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__253)) __PYX_ERR(0, 1714, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1732
 * 
 * 
 * def getgeneratorlocals(generator):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of generator local variables to their current values.
 */
  __pyx_tuple__254 = PyTuple_Pack(2, __pyx_n_s_generator, __pyx_n_s_frame); if (unlikely(!__pyx_tuple__254)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__254);
  __Pyx_GIVEREF(__pyx_tuple__254);
  __pyx_codeobj__255 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__254, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getgeneratorlocals, 1732, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__255)) __PYX_ERR(0, 1732, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1757
 * 
 * 
 * def getcoroutinestate(coroutine):             # <<<<<<<<<<<<<<
 *     """Get current state of a coroutine object.
 * 
 */
  __pyx_tuple__256 = PyTuple_Pack(1, __pyx_n_s_coroutine); if (unlikely(!__pyx_tuple__256)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__256);
  __Pyx_GIVEREF(__pyx_tuple__256);
  __pyx_codeobj__257 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__256, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getcoroutinestate, 1757, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__257)) __PYX_ERR(0, 1757, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1775
 * 
 * 
 * def getcoroutinelocals(coroutine):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of coroutine local variables to their current values.
 */
  __pyx_tuple__258 = PyTuple_Pack(2, __pyx_n_s_coroutine, __pyx_n_s_frame); if (unlikely(!__pyx_tuple__258)) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__258);
  __Pyx_GIVEREF(__pyx_tuple__258);
  __pyx_codeobj__259 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__258, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getcoroutinelocals, 1775, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__259)) __PYX_ERR(0, 1775, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1803
 * 
 * 
 * def _signature_get_user_defined_method(cls, method_name):             # <<<<<<<<<<<<<<
 *     """Private helper. Checks if ``cls`` has an attribute
 *     named ``method_name`` and returns it only if it is a
 */
  __pyx_tuple__260 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_method_name, __pyx_n_s_meth); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(0, 1803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);
  __pyx_codeobj__261 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__260, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_get_user_defined_meth, 1803, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__261)) __PYX_ERR(0, 1803, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1819
 * 
 * 
 * def _signature_get_partial(wrapped_sig, partial, extra_args=()):             # <<<<<<<<<<<<<<
 *     """Private helper to calculate how 'wrapped_sig' signature will
 *     look like after applying a 'functools.partial' object (or alike)
 */
  __pyx_tuple__262 = PyTuple_Pack(15, __pyx_n_s_wrapped_sig, __pyx_n_s_partial, __pyx_n_s_extra_args, __pyx_n_s_old_params, __pyx_n_s_new_params, __pyx_n_s_partial_args, __pyx_n_s_partial_keywords, __pyx_n_s_ba, __pyx_n_s_ex, __pyx_n_s_msg, __pyx_n_s_transform_to_kwonly, __pyx_n_s_param_name, __pyx_n_s_param, __pyx_n_s_arg_value, __pyx_n_s_new_param); if (unlikely(!__pyx_tuple__262)) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__262);
  __Pyx_GIVEREF(__pyx_tuple__262);
  __pyx_codeobj__263 = (PyObject*)__Pyx_PyCode_New(3, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__262, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_get_partial, 1819, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__263)) __PYX_ERR(0, 1819, __pyx_L1_error)
  __pyx_tuple__264 = PyTuple_Pack(1, ((PyObject*)__pyx_empty_tuple)); if (unlikely(!__pyx_tuple__264)) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__264);
  __Pyx_GIVEREF(__pyx_tuple__264);

  /* "Aspidites/_vendor/contracts/inspection.py":1894
 * 
 * 
 * def _signature_bound_method(sig):             # <<<<<<<<<<<<<<
 *     """Private helper to transform signatures for unbound
 *     functions to bound methods.
 */
  __pyx_tuple__265 = PyTuple_Pack(3, __pyx_n_s_sig, __pyx_n_s_params, __pyx_n_s_kind); if (unlikely(!__pyx_tuple__265)) __PYX_ERR(0, 1894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__265);
  __Pyx_GIVEREF(__pyx_tuple__265);
  __pyx_codeobj__266 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__265, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_bound_method, 1894, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__266)) __PYX_ERR(0, 1894, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1920
 * 
 * 
 * def _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a callable that might
 *     support Argument Clinic's __text_signature__ protocol.
 */
  __pyx_tuple__267 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__267)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__267);
  __Pyx_GIVEREF(__pyx_tuple__267);
  __pyx_codeobj__268 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__267, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_is_builtin, 1920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__268)) __PYX_ERR(0, 1920, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1932
 * 
 * 
 * def _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a duck type of FunctionType.
 *     A good example of such objects are functions compiled with
 */
  __pyx_tuple__269 = PyTuple_Pack(6, __pyx_n_s_obj, __pyx_n_s_name_2, __pyx_n_s_code_2, __pyx_n_s_defaults, __pyx_n_s_kwdefaults_2, __pyx_n_s_annotations); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 1932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__269, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_is_functionlike, 1932, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(0, 1932, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1957
 * 
 * 
 * def _signature_get_bound_param(spec):             # <<<<<<<<<<<<<<
 *     """ Private helper to get first parameter name from a
 *     __text_signature__ of a builtin method, which should
 */
  __pyx_tuple__271 = PyTuple_Pack(3, __pyx_n_s_spec, __pyx_n_s_pos, __pyx_n_s_cpos); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_get_bound_param, 1957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(0, 1957, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1980
 * 
 * 
 * def _signature_strip_non_python_syntax(signature):             # <<<<<<<<<<<<<<
 *     """
 *     Private helper function. Takes a signature in Argument Clinic's
 */
  __pyx_tuple__273 = PyTuple_Pack(18, __pyx_n_s_signature_2, __pyx_n_s_self_parameter, __pyx_n_s_last_positional_only, __pyx_n_s_lines, __pyx_n_s_generator, __pyx_n_s_token_stream, __pyx_n_s_delayed_comma, __pyx_n_s_skip_next_comma, __pyx_n_s_text, __pyx_n_s_add, __pyx_n_s_current_parameter, __pyx_n_s_OP, __pyx_n_s_ERRORTOKEN, __pyx_n_s_t, __pyx_n_s_type, __pyx_n_s_string, __pyx_n_s_clean_signature, __pyx_n_s_l); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);
  __pyx_codeobj__274 = (PyObject*)__Pyx_PyCode_New(1, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_strip_non_python_synt, 1980, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__274)) __PYX_ERR(0, 1980, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */
  __pyx_tuple__275 = PyTuple_Pack(35, __pyx_n_s_cls, __pyx_n_s_obj, __pyx_n_s_s, __pyx_n_s_skip_bound_arg, __pyx_n_s_ast, __pyx_n_s_Parameter, __pyx_n_s_clean_signature, __pyx_n_s_self_parameter, __pyx_n_s_last_positional_only, __pyx_n_s_program, __pyx_n_s_module_3, __pyx_n_s_f, __pyx_n_s_parameters, __pyx_n_s_empty_2, __pyx_n_s_invalid, __pyx_n_s_module_dict, __pyx_n_s_module_name, __pyx_n_s_sys_module_dict, __pyx_n_s_parse_name, __pyx_n_s_parse_name, __pyx_n_s_wrap_value, __pyx_n_s_wrap_value, __pyx_n_s_RewriteSymbolics, __pyx_n_s_p, __pyx_n_s_p, __pyx_n_s_args, __pyx_n_s_defaults, __pyx_n_s_iter, __pyx_n_s_kind, __pyx_n_s_i, __pyx_n_s_name_2, __pyx_n_s_default, __pyx_n_s_self_3, __pyx_n_s_self_isbound, __pyx_n_s_self_ismodule); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);
  __pyx_codeobj__276 = (PyObject*)__Pyx_PyCode_New(4, 0, 35, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__275, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_fromstr, 2052, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__276)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __pyx_tuple__277 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);

  /* "Aspidites/_vendor/contracts/inspection.py":2197
 * 
 * 
 * def _signature_from_builtin(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper function to get signature for
 *     builtin callables.
 */
  __pyx_tuple__278 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_func_2, __pyx_n_s_skip_bound_arg, __pyx_n_s_s); if (unlikely(!__pyx_tuple__278)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__278);
  __Pyx_GIVEREF(__pyx_tuple__278);
  __pyx_codeobj__279 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__278, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_from_builtin, 2197, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__279)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __pyx_tuple__280 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__280)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__280);
  __Pyx_GIVEREF(__pyx_tuple__280);

  /* "Aspidites/_vendor/contracts/inspection.py":2213
 * 
 * 
 * def _signature_from_function(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper: constructs Signature for the given python function."""
 * 
 */
  __pyx_tuple__281 = PyTuple_Pack(26, __pyx_n_s_cls, __pyx_n_s_func_2, __pyx_n_s_skip_bound_arg, __pyx_n_s_is_duck_function, __pyx_n_s_s, __pyx_n_s_Parameter, __pyx_n_s_func_code, __pyx_n_s_pos_count, __pyx_n_s_arg_names, __pyx_n_s_posonly_count, __pyx_n_s_positional, __pyx_n_s_keyword_only_count, __pyx_n_s_keyword_only_2, __pyx_n_s_annotations, __pyx_n_s_defaults, __pyx_n_s_kwdefaults_2, __pyx_n_s_pos_default_count, __pyx_n_s_parameters, __pyx_n_s_non_default_count, __pyx_n_s_posonly_left, __pyx_n_s_name_2, __pyx_n_s_kind, __pyx_n_s_annotation, __pyx_n_s_offset, __pyx_n_s_default, __pyx_n_s_index); if (unlikely(!__pyx_tuple__281)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__281);
  __Pyx_GIVEREF(__pyx_tuple__281);
  __pyx_codeobj__282 = (PyObject*)__Pyx_PyCode_New(3, 0, 26, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__281, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_from_function, 2213, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__282)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __pyx_tuple__283 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__283)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__283);
  __Pyx_GIVEREF(__pyx_tuple__283);

  /* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */
  __pyx_tuple__284 = PyTuple_Pack(17, __pyx_n_s_obj, __pyx_n_s_follow_wrapper_chains, __pyx_n_s_skip_bound_arg, __pyx_n_s_sigcls, __pyx_n_s_sig, __pyx_n_s_partialmethod_2, __pyx_n_s_wrapped_sig, __pyx_n_s_first_wrapped_param, __pyx_n_s_sig_params, __pyx_n_s_new_params, __pyx_n_s_call_2, __pyx_n_s_new_2, __pyx_n_s_init_2, __pyx_n_s_base, __pyx_n_s_text_sig, __pyx_n_s_ex, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__284)) __PYX_ERR(0, 2307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__284);
  __Pyx_GIVEREF(__pyx_tuple__284);
  __pyx_codeobj__285 = (PyObject*)__Pyx_PyCode_New(1, 3, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__284, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_from_callable, 2307, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__285)) __PYX_ERR(0, 2307, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2527
 *     VAR_KEYWORD = 4
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._name_
 * 
 */
  __pyx_tuple__286 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__286)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__286);
  __Pyx_GIVEREF(__pyx_tuple__286);
  __pyx_codeobj__287 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__286, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_str, 2527, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__287)) __PYX_ERR(0, 2527, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2531
 * 
 *     @property
 *     def description(self):             # <<<<<<<<<<<<<<
 *         return _PARAM_NAME_MAPPING[self]
 * 
 */
  __pyx_tuple__288 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__288)) __PYX_ERR(0, 2531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__288);
  __Pyx_GIVEREF(__pyx_tuple__288);
  __pyx_codeobj__289 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__288, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_description, 2531, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__289)) __PYX_ERR(0, 2531, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2572
 *     """
 * 
 *     __slots__ = ('_name', '_kind', '_default', '_annotation')             # <<<<<<<<<<<<<<
 * 
 *     POSITIONAL_ONLY = _POSITIONAL_ONLY
 */
  __pyx_tuple__290 = PyTuple_Pack(4, __pyx_n_u_name_4, __pyx_n_u_kind_2, __pyx_n_u_default_2, __pyx_n_u_annotation_2); if (unlikely(!__pyx_tuple__290)) __PYX_ERR(0, 2572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__290);
  __Pyx_GIVEREF(__pyx_tuple__290);

  /* "Aspidites/_vendor/contracts/inspection.py":2582
 *     empty = _empty
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):             # <<<<<<<<<<<<<<
 *         try:
 *             self._kind = _ParameterKind(kind)
 */
  __pyx_tuple__291 = PyTuple_Pack(6, __pyx_n_s_self_2, __pyx_n_s_name_2, __pyx_n_s_kind, __pyx_n_s_default, __pyx_n_s_annotation, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__291)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__291);
  __Pyx_GIVEREF(__pyx_tuple__291);
  __pyx_codeobj__292 = (PyObject*)__Pyx_PyCode_New(3, 2, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__291, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_init, 2582, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__292)) __PYX_ERR(0, 2582, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2622
 *         self._name = name
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (self._name, self._kind),
 */
  __pyx_tuple__293 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);
  __pyx_codeobj__294 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_reduce, 2622, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__294)) __PYX_ERR(0, 2622, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2628
 *                  '_annotation': self._annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._default = state['_default']
 *         self._annotation = state['_annotation']
 */
  __pyx_tuple__295 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_state); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 2628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);
  __pyx_codeobj__296 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__295, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_setstate, 2628, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__296)) __PYX_ERR(0, 2628, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2633
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self._name
 * 
 */
  __pyx_tuple__297 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(0, 2633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);
  __pyx_codeobj__298 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__297, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_name_2, 2633, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__298)) __PYX_ERR(0, 2633, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2637
 * 
 *     @property
 *     def default(self):             # <<<<<<<<<<<<<<
 *         return self._default
 * 
 */
  __pyx_tuple__299 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__299)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__299);
  __Pyx_GIVEREF(__pyx_tuple__299);
  __pyx_codeobj__300 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__299, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_default, 2637, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__300)) __PYX_ERR(0, 2637, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2641
 * 
 *     @property
 *     def annotation(self):             # <<<<<<<<<<<<<<
 *         return self._annotation
 * 
 */
  __pyx_tuple__301 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__301)) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__301);
  __Pyx_GIVEREF(__pyx_tuple__301);
  __pyx_codeobj__302 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__301, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_annotation, 2641, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__302)) __PYX_ERR(0, 2641, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2645
 * 
 *     @property
 *     def kind(self):             # <<<<<<<<<<<<<<
 *         return self._kind
 * 
 */
  __pyx_tuple__303 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);
  __pyx_codeobj__304 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_kind, 2645, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__304)) __PYX_ERR(0, 2645, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2648
 *         return self._kind
 * 
 *     def replace(self, *, name=_void, kind=_void,             # <<<<<<<<<<<<<<
 *                 annotation=_void, default=_void):
 *         """Creates a customized copy of the Parameter."""
 */
  __pyx_tuple__305 = PyTuple_Pack(5, __pyx_n_s_self_2, __pyx_n_s_name_2, __pyx_n_s_kind, __pyx_n_s_annotation, __pyx_n_s_default); if (unlikely(!__pyx_tuple__305)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__305);
  __Pyx_GIVEREF(__pyx_tuple__305);
  __pyx_codeobj__306 = (PyObject*)__Pyx_PyCode_New(1, 4, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__305, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_replace, 2648, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__306)) __PYX_ERR(0, 2648, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2666
 *         return type(self)(name, kind, default=default, annotation=annotation)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         kind = self.kind
 *         formatted = self._name
 */
  __pyx_tuple__307 = PyTuple_Pack(3, __pyx_n_s_self_2, __pyx_n_s_kind, __pyx_n_s_formatted); if (unlikely(!__pyx_tuple__307)) __PYX_ERR(0, 2666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__307);
  __Pyx_GIVEREF(__pyx_tuple__307);
  __pyx_codeobj__308 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__307, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_str, 2666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__308)) __PYX_ERR(0, 2666, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2688
 *         return formatted
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 */
  __pyx_tuple__309 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__309)) __PYX_ERR(0, 2688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__309);
  __Pyx_GIVEREF(__pyx_tuple__309);
  __pyx_codeobj__310 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__309, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_repr, 2688, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__310)) __PYX_ERR(0, 2688, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2691
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 */
  __pyx_tuple__311 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__311)) __PYX_ERR(0, 2691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__311);
  __Pyx_GIVEREF(__pyx_tuple__311);
  __pyx_codeobj__312 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__311, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_hash, 2691, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__312)) __PYX_ERR(0, 2691, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2694
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_tuple__313 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_other); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);
  __pyx_codeobj__314 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__313, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_eq, 2694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__314)) __PYX_ERR(0, 2694, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2722
 *     """
 * 
 *     __slots__ = ('arguments', '_signature', '__weakref__')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, signature, arguments):
 */
  __pyx_tuple__315 = PyTuple_Pack(3, __pyx_n_u_arguments, __pyx_n_u_signature_3, __pyx_n_u_weakref); if (unlikely(!__pyx_tuple__315)) __PYX_ERR(0, 2722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__315);
  __Pyx_GIVEREF(__pyx_tuple__315);

  /* "Aspidites/_vendor/contracts/inspection.py":2724
 *     __slots__ = ('arguments', '_signature', '__weakref__')
 * 
 *     def __init__(self, signature, arguments):             # <<<<<<<<<<<<<<
 *         self.arguments = arguments
 *         self._signature = signature
 */
  __pyx_tuple__316 = PyTuple_Pack(3, __pyx_n_s_self_2, __pyx_n_s_signature_2, __pyx_n_s_arguments); if (unlikely(!__pyx_tuple__316)) __PYX_ERR(0, 2724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__316);
  __Pyx_GIVEREF(__pyx_tuple__316);
  __pyx_codeobj__317 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__316, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_init, 2724, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__317)) __PYX_ERR(0, 2724, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2729
 * 
 *     @property
 *     def signature(self):             # <<<<<<<<<<<<<<
 *         return self._signature
 * 
 */
  __pyx_tuple__318 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__318)) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__318);
  __Pyx_GIVEREF(__pyx_tuple__318);
  __pyx_codeobj__319 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__318, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_2, 2729, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__319)) __PYX_ERR(0, 2729, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2733
 * 
 *     @property
 *     def args(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 */
  __pyx_tuple__320 = PyTuple_Pack(5, __pyx_n_s_self_2, __pyx_n_s_args, __pyx_n_s_param_name, __pyx_n_s_param, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__320)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__320);
  __Pyx_GIVEREF(__pyx_tuple__320);
  __pyx_codeobj__321 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__320, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_args, 2733, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__321)) __PYX_ERR(0, 2733, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2756
 * 
 *     @property
 *     def kwargs(self):             # <<<<<<<<<<<<<<
 *         kwargs = {}
 *         kwargs_started = False
 */
  __pyx_tuple__322 = PyTuple_Pack(6, __pyx_n_s_self_2, __pyx_n_s_kwargs, __pyx_n_s_kwargs_started, __pyx_n_s_param_name, __pyx_n_s_param, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__322)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__322);
  __Pyx_GIVEREF(__pyx_tuple__322);
  __pyx_codeobj__323 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__322, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_kwargs, 2756, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__323)) __PYX_ERR(0, 2756, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2785
 *         return kwargs
 * 
 *     def apply_defaults(self):             # <<<<<<<<<<<<<<
 *         """Set default values for missing arguments.
 * 
 */
  __pyx_tuple__324 = PyTuple_Pack(6, __pyx_n_s_self_2, __pyx_n_s_arguments, __pyx_n_s_new_arguments, __pyx_n_s_name_2, __pyx_n_s_param, __pyx_n_s_val); if (unlikely(!__pyx_tuple__324)) __PYX_ERR(0, 2785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__324);
  __Pyx_GIVEREF(__pyx_tuple__324);
  __pyx_codeobj__325 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__324, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_apply_defaults, 2785, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__325)) __PYX_ERR(0, 2785, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2813
 *         self.arguments = dict(new_arguments)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_tuple__326 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_other); if (unlikely(!__pyx_tuple__326)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__326);
  __Pyx_GIVEREF(__pyx_tuple__326);
  __pyx_codeobj__327 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__326, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_eq, 2813, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__327)) __PYX_ERR(0, 2813, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2821
 *                 self.arguments == other.arguments)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._signature = state['_signature']
 *         self.arguments = state['arguments']
 */
  __pyx_tuple__328 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_state); if (unlikely(!__pyx_tuple__328)) __PYX_ERR(0, 2821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__328);
  __Pyx_GIVEREF(__pyx_tuple__328);
  __pyx_codeobj__329 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__328, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_setstate, 2821, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__329)) __PYX_ERR(0, 2821, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2825
 *         self.arguments = state['arguments']
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 */
  __pyx_tuple__330 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__330)) __PYX_ERR(0, 2825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__330);
  __Pyx_GIVEREF(__pyx_tuple__330);
  __pyx_codeobj__331 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__330, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_getstate, 2825, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__331)) __PYX_ERR(0, 2825, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2828
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for arg, value in self.arguments.items():
 */
  __pyx_tuple__332 = PyTuple_Pack(4, __pyx_n_s_self_2, __pyx_n_s_args, __pyx_n_s_arg, __pyx_n_s_value); if (unlikely(!__pyx_tuple__332)) __PYX_ERR(0, 2828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__332);
  __Pyx_GIVEREF(__pyx_tuple__332);
  __pyx_codeobj__333 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__332, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_repr, 2828, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__333)) __PYX_ERR(0, 2828, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2858
 *     """
 * 
 *     __slots__ = ('_return_annotation', '_parameters')             # <<<<<<<<<<<<<<
 * 
 *     _parameter_cls = Parameter
 */
  __pyx_tuple__334 = PyTuple_Pack(2, __pyx_n_u_return_annotation_2, __pyx_n_u_parameters_2); if (unlikely(!__pyx_tuple__334)) __PYX_ERR(0, 2858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__334);
  __Pyx_GIVEREF(__pyx_tuple__334);

  /* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */
  __pyx_tuple__335 = PyTuple_Pack(13, __pyx_n_s_self_2, __pyx_n_s_parameters, __pyx_n_s_return_annotation, __pyx_n_s_validate_parameters, __pyx_n_s_params, __pyx_n_s_top_kind, __pyx_n_s_kind_defaults, __pyx_n_s_param, __pyx_n_s_kind, __pyx_n_s_name_2, __pyx_n_s_msg, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__335)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__335);
  __Pyx_GIVEREF(__pyx_tuple__335);
  __pyx_codeobj__336 = (PyObject*)__Pyx_PyCode_New(2, 2, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_init, 2865, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__336)) __PYX_ERR(0, 2865, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2920
 * 
 *     @classmethod
 *     def from_function(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given python function.
 * 
 */
  __pyx_tuple__337 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_func_2); if (unlikely(!__pyx_tuple__337)) __PYX_ERR(0, 2920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__337);
  __Pyx_GIVEREF(__pyx_tuple__337);
  __pyx_codeobj__338 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_from_function, 2920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__338)) __PYX_ERR(0, 2920, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2932
 * 
 *     @classmethod
 *     def from_builtin(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given builtin function.
 * 
 */
  __pyx_tuple__339 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_func_2); if (unlikely(!__pyx_tuple__339)) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__339);
  __Pyx_GIVEREF(__pyx_tuple__339);
  __pyx_codeobj__340 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__339, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_from_builtin, 2932, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__340)) __PYX_ERR(0, 2932, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2944
 * 
 *     @classmethod
 *     def from_callable(cls, obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,
 */
  __pyx_tuple__341 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_obj, __pyx_n_s_follow_wrapped); if (unlikely(!__pyx_tuple__341)) __PYX_ERR(0, 2944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__341);
  __Pyx_GIVEREF(__pyx_tuple__341);
  __pyx_codeobj__342 = (PyObject*)__Pyx_PyCode_New(2, 1, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__341, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_from_callable, 2944, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__342)) __PYX_ERR(0, 2944, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2950
 * 
 *     @property
 *     def parameters(self):             # <<<<<<<<<<<<<<
 *         return self._parameters
 * 
 */
  __pyx_tuple__343 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__343)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__343);
  __Pyx_GIVEREF(__pyx_tuple__343);
  __pyx_codeobj__344 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__343, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_parameters, 2950, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__344)) __PYX_ERR(0, 2950, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2954
 * 
 *     @property
 *     def return_annotation(self):             # <<<<<<<<<<<<<<
 *         return self._return_annotation
 * 
 */
  __pyx_tuple__345 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__345)) __PYX_ERR(0, 2954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__345);
  __Pyx_GIVEREF(__pyx_tuple__345);
  __pyx_codeobj__346 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__345, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_return_annotation, 2954, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__346)) __PYX_ERR(0, 2954, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2957
 *         return self._return_annotation
 * 
 *     def replace(self, *, parameters=_void, return_annotation=_void):             # <<<<<<<<<<<<<<
 *         """Creates a customized copy of the Signature.
 *         Pass 'parameters' and/or 'return_annotation' arguments
 */
  __pyx_tuple__347 = PyTuple_Pack(3, __pyx_n_s_self_2, __pyx_n_s_parameters, __pyx_n_s_return_annotation); if (unlikely(!__pyx_tuple__347)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__347);
  __Pyx_GIVEREF(__pyx_tuple__347);
  __pyx_codeobj__348 = (PyObject*)__Pyx_PyCode_New(1, 2, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__347, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_replace, 2957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__348)) __PYX_ERR(0, 2957, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */
  __pyx_tuple__349 = PyTuple_Pack(6, __pyx_n_s_self_2, __pyx_n_s_params, __pyx_n_s_kwo_params, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_param); if (unlikely(!__pyx_tuple__349)) __PYX_ERR(0, 2972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__349);
  __Pyx_GIVEREF(__pyx_tuple__349);
  __pyx_codeobj__350 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__349, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_hash_basis, 2972, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__350)) __PYX_ERR(0, 2972, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2981
 *         return params, kwo_params, self.return_annotation
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())
 */
  __pyx_tuple__351 = PyTuple_Pack(4, __pyx_n_s_self_2, __pyx_n_s_params, __pyx_n_s_kwo_params, __pyx_n_s_return_annotation); if (unlikely(!__pyx_tuple__351)) __PYX_ERR(0, 2981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__351);
  __Pyx_GIVEREF(__pyx_tuple__351);
  __pyx_codeobj__352 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__351, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_hash, 2981, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__352)) __PYX_ERR(0, 2981, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2986
 *         return hash((params, kwo_params, return_annotation))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_tuple__353 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_other); if (unlikely(!__pyx_tuple__353)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__353);
  __Pyx_GIVEREF(__pyx_tuple__353);
  __pyx_codeobj__354 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__353, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_eq, 2986, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__354)) __PYX_ERR(0, 2986, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2993
 *         return self._hash_basis() == other._hash_basis()
 * 
 *     def _bind(self, args, kwargs, *, partial=False):             # <<<<<<<<<<<<<<
 *         """Private method. Don't use directly."""
 * 
 */
  __pyx_tuple__355 = PyTuple_Pack(14, __pyx_n_s_self_2, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_partial, __pyx_n_s_arguments, __pyx_n_s_parameters, __pyx_n_s_parameters_ex, __pyx_n_s_arg_vals, __pyx_n_s_arg_val, __pyx_n_s_param, __pyx_n_s_msg, __pyx_n_s_values, __pyx_n_s_kwargs_param, __pyx_n_s_param_name); if (unlikely(!__pyx_tuple__355)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__355);
  __Pyx_GIVEREF(__pyx_tuple__355);
  __pyx_codeobj__356 = (PyObject*)__Pyx_PyCode_New(3, 1, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__355, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_bind, 2993, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__356)) __PYX_ERR(0, 2993, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3138
 *     #     return self._bind(args, kwargs, partial=True)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),
 */
  __pyx_tuple__357 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__357)) __PYX_ERR(0, 3138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__357);
  __Pyx_GIVEREF(__pyx_tuple__357);
  __pyx_codeobj__358 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_reduce, 3138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__358)) __PYX_ERR(0, 3138, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3143
 *                 {'_return_annotation': self._return_annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._return_annotation = state['_return_annotation']
 * 
 */
  __pyx_tuple__359 = PyTuple_Pack(2, __pyx_n_s_self_2, __pyx_n_s_state); if (unlikely(!__pyx_tuple__359)) __PYX_ERR(0, 3143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__359);
  __Pyx_GIVEREF(__pyx_tuple__359);
  __pyx_codeobj__360 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__359, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_setstate, 3143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__360)) __PYX_ERR(0, 3143, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3146
 *         self._return_annotation = state['_return_annotation']
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 */
  __pyx_tuple__361 = PyTuple_Pack(1, __pyx_n_s_self_2); if (unlikely(!__pyx_tuple__361)) __PYX_ERR(0, 3146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__361);
  __Pyx_GIVEREF(__pyx_tuple__361);
  __pyx_codeobj__362 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__361, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_repr, 3146, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__362)) __PYX_ERR(0, 3146, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3149
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         result = []
 *         render_pos_only_separator = False
 */
  __pyx_tuple__363 = PyTuple_Pack(9, __pyx_n_s_self_2, __pyx_n_s_result, __pyx_n_s_render_pos_only_separator, __pyx_n_s_render_kw_only_separator, __pyx_n_s_param, __pyx_n_s_formatted, __pyx_n_s_kind, __pyx_n_s_rendered, __pyx_n_s_anno); if (unlikely(!__pyx_tuple__363)) __PYX_ERR(0, 3149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__363);
  __Pyx_GIVEREF(__pyx_tuple__363);
  __pyx_codeobj__364 = (PyObject*)__Pyx_PyCode_New(1, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__363, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_str, 3149, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__364)) __PYX_ERR(0, 3149, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3195
 * 
 * 
 * def signature(obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *     """Get a signature object for the passed callable."""
 *     return Signature.from_callable(obj, follow_wrapped=follow_wrapped)
 */
  __pyx_tuple__365 = PyTuple_Pack(2, __pyx_n_s_obj, __pyx_n_s_follow_wrapped); if (unlikely(!__pyx_tuple__365)) __PYX_ERR(0, 3195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__365);
  __Pyx_GIVEREF(__pyx_tuple__365);
  __pyx_codeobj__366 = (PyObject*)__Pyx_PyCode_New(1, 1, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__365, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_signature_2, 3195, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__366)) __PYX_ERR(0, 3195, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3200
 * 
 * 
 * def can_accept_exactly_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept exactly one argument
 *         using introspection.
 */
  __pyx_tuple__367 = PyTuple_Pack(4, __pyx_n_s_callable_thing, __pyx_n_s_f, __pyx_n_s_args, __pyx_n_s_e); if (unlikely(!__pyx_tuple__367)) __PYX_ERR(0, 3200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__367);
  __Pyx_GIVEREF(__pyx_tuple__367);
  __pyx_codeobj__368 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_can_accept_exactly_one_argument, 3200, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__368)) __PYX_ERR(0, 3200, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3223
 * 
 * 
 * def get_f_from_callable(callable_thing):             # <<<<<<<<<<<<<<
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__
 */
  __pyx_tuple__369 = PyTuple_Pack(2, __pyx_n_s_callable_thing, __pyx_n_s_f); if (unlikely(!__pyx_tuple__369)) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__369);
  __Pyx_GIVEREF(__pyx_tuple__369);
  __pyx_codeobj__370 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__369, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_get_f_from_callable, 3223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__370)) __PYX_ERR(0, 3223, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3236
 * 
 * 
 * def get_callable_fullargspec(callable_thing):             # <<<<<<<<<<<<<<
 *     f: Callable
 *     spec: FullArgSpec
 */
  __pyx_tuple__371 = PyTuple_Pack(3, __pyx_n_s_callable_thing, __pyx_n_s_f, __pyx_n_s_spec); if (unlikely(!__pyx_tuple__371)) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__371);
  __Pyx_GIVEREF(__pyx_tuple__371);
  __pyx_codeobj__372 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__371, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_get_callable_fullargspec, 3236, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__372)) __PYX_ERR(0, 3236, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3244
 * 
 * 
 * def can_accept_at_least_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept at least one argument
 */
  __pyx_tuple__373 = PyTuple_Pack(2, __pyx_n_s_callable_thing, __pyx_n_s_spec); if (unlikely(!__pyx_tuple__373)) __PYX_ERR(0, 3244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__373);
  __Pyx_GIVEREF(__pyx_tuple__373);
  __pyx_codeobj__374 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__373, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_can_accept_at_least_one_argument, 3244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__374)) __PYX_ERR(0, 3244, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3253
 * 
 * 
 * def can_accept_self(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable's first argument is self
 */
  __pyx_tuple__375 = PyTuple_Pack(2, __pyx_n_s_callable_thing, __pyx_n_s_spec); if (unlikely(!__pyx_tuple__375)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__375);
  __Pyx_GIVEREF(__pyx_tuple__375);
  __pyx_codeobj__376 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__375, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_can_accept_self, 3253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__376)) __PYX_ERR(0, 3253, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3265
 * 
 * 
 * def can_accept_self_plus_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept exactly self plus one argument
 */
  __pyx_tuple__377 = PyTuple_Pack(4, __pyx_n_s_callable_thing, __pyx_n_s_spec, __pyx_n_s_f, __pyx_n_s_e); if (unlikely(!__pyx_tuple__377)) __PYX_ERR(0, 3265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__377);
  __Pyx_GIVEREF(__pyx_tuple__377);
  __pyx_codeobj__378 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__377, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_can_accept_self_plus_one_argumen, 3265, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__378)) __PYX_ERR(0, 3265, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":3289
 * 
 * 
 * def check_callable_accepts_these_arguments(callable_thing, args, kwargs):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept the args and kwargs.
 * 
 */
  __pyx_tuple__379 = PyTuple_Pack(6, __pyx_n_s_callable_thing, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_f, __pyx_n_s_bound, __pyx_n_s_e); if (unlikely(!__pyx_tuple__379)) __PYX_ERR(0, 3289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__379);
  __Pyx_GIVEREF(__pyx_tuple__379);
  __pyx_codeobj__380 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__379, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_contracts_insp_2, __pyx_n_s_check_callable_accepts_these_arg, 3289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__380)) __PYX_ERR(0, 3289, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_update.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyUnicode_Type_format.type = (PyObject*)&PyUnicode_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1048576 = PyInt_FromLong(1048576L); if (unlikely(!__pyx_int_1048576)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct__classify_class_attrs;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 434, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_2_unwrap;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile) < 0) __PYX_ERR(0, 736, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_3_getsourcefile;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 743, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_4_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_6_formatannotationrelativeto;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec) < 0) __PYX_ERR(0, 1221, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_7_formatargspec;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_8_getcallargs;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr) < 0) __PYX_ERR(0, 2052, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_9__signature_fromstr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__ = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_10___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr) < 0) __PYX_ERR(0, 2914, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_11_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis) < 0) __PYX_ERR(0, 2972, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_12__hash_basis;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr) < 0) __PYX_ERR(0, 2973, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr = &__pyx_type_9Aspidites_7_vendor_9contracts_10inspection___pyx_scope_struct_13_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initinspection(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initinspection(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_inspection(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_inspection(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_inspection(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'inspection' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_inspection(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("inspection", __pyx_methods, __pyx_k_Get_useful_information_from_live, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_Aspidites___vendor__contracts__inspection) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "Aspidites._vendor.contracts.inspection")) {
      if (unlikely(PyDict_SetItemString(modules, "Aspidites._vendor.contracts.inspection", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "Aspidites/_vendor/contracts/inspection.py":31
 * """
 * 
 * import sys             # <<<<<<<<<<<<<<
 * import inspect
 * import abc
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":32
 * 
 * import sys
 * import inspect             # <<<<<<<<<<<<<<
 * import abc
 * import ast
 */
  __pyx_t_1 = __Pyx_patch_inspect(__Pyx_Import(__pyx_n_s_inspect, 0, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inspect, __pyx_t_1) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":33
 * import sys
 * import inspect
 * import abc             # <<<<<<<<<<<<<<
 * import ast
 * import dis
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_abc, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_abc, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":34
 * import inspect
 * import abc
 * import ast             # <<<<<<<<<<<<<<
 * import dis
 * import collections.abc
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ast, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ast, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":35
 * import abc
 * import ast
 * import dis             # <<<<<<<<<<<<<<
 * import collections.abc
 * import enum
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_dis, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dis, __pyx_t_1) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":36
 * import ast
 * import dis
 * import collections.abc             # <<<<<<<<<<<<<<
 * import enum
 * import importlib.machinery
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections_abc, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_1) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":37
 * import dis
 * import collections.abc
 * import enum             # <<<<<<<<<<<<<<
 * import importlib.machinery
 * import itertools
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_enum, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum, __pyx_t_1) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":38
 * import collections.abc
 * import enum
 * import importlib.machinery             # <<<<<<<<<<<<<<
 * import itertools
 * import linecache
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_importlib_machinery, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_importlib, __pyx_t_1) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":39
 * import enum
 * import importlib.machinery
 * import itertools             # <<<<<<<<<<<<<<
 * import linecache
 * import os
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_itertools, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_1) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":40
 * import importlib.machinery
 * import itertools
 * import linecache             # <<<<<<<<<<<<<<
 * import os
 * import re
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_linecache, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_linecache, __pyx_t_1) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":41
 * import itertools
 * import linecache
 * import os             # <<<<<<<<<<<<<<
 * import re
 * import sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":42
 * import linecache
 * import os
 * import re             # <<<<<<<<<<<<<<
 * import sys
 * import tokenize
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":43
 * import os
 * import re
 * import sys             # <<<<<<<<<<<<<<
 * import tokenize
 * import token
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":44
 * import re
 * import sys
 * import tokenize             # <<<<<<<<<<<<<<
 * import token
 * import types
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_tokenize, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tokenize, __pyx_t_1) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":45
 * import sys
 * import tokenize
 * import token             # <<<<<<<<<<<<<<
 * import types
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_token, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_token, __pyx_t_1) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":46
 * import tokenize
 * import token
 * import types             # <<<<<<<<<<<<<<
 * import warnings
 * import functools
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_types, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_types, __pyx_t_1) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":47
 * import token
 * import types
 * import warnings             # <<<<<<<<<<<<<<
 * import functools
 * import builtins
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":48
 * import types
 * import warnings
 * import functools             # <<<<<<<<<<<<<<
 * import builtins
 * from operator import attrgetter
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_functools, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_functools, __pyx_t_1) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":49
 * import warnings
 * import functools
 * import builtins             # <<<<<<<<<<<<<<
 * from operator import attrgetter
 * from collections import namedtuple, OrderedDict
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_builtins, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_builtins, __pyx_t_1) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":50
 * import functools
 * import builtins
 * from operator import attrgetter             # <<<<<<<<<<<<<<
 * from collections import namedtuple, OrderedDict
 * from typing import Callable
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_attrgetter);
  __Pyx_GIVEREF(__pyx_n_s_attrgetter);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_attrgetter);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_operator, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_attrgetter, __pyx_t_1) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":51
 * import builtins
 * from operator import attrgetter
 * from collections import namedtuple, OrderedDict             # <<<<<<<<<<<<<<
 * from typing import Callable
 * from inspect import (FullArgSpec, CO_COROUTINE, CO_GENERATOR, CO_ASYNC_GENERATOR, CORO_CLOSED, CORO_CREATED,
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_namedtuple);
  __Pyx_INCREF(__pyx_n_s_OrderedDict);
  __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_OrderedDict);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":52
 * from operator import attrgetter
 * from collections import namedtuple, OrderedDict
 * from typing import Callable             # <<<<<<<<<<<<<<
 * from inspect import (FullArgSpec, CO_COROUTINE, CO_GENERATOR, CO_ASYNC_GENERATOR, CORO_CLOSED, CORO_CREATED,
 *                      CORO_RUNNING, CORO_SUSPENDED, CO_VARARGS, CO_VARKEYWORDS, CO_ITERABLE_COROUTINE)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Callable);
  __Pyx_GIVEREF(__pyx_n_s_Callable);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Callable);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Callable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Callable, __pyx_t_1) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":53
 * from collections import namedtuple, OrderedDict
 * from typing import Callable
 * from inspect import (FullArgSpec, CO_COROUTINE, CO_GENERATOR, CO_ASYNC_GENERATOR, CORO_CLOSED, CORO_CREATED,             # <<<<<<<<<<<<<<
 *                      CORO_RUNNING, CORO_SUSPENDED, CO_VARARGS, CO_VARKEYWORDS, CO_ITERABLE_COROUTINE)
 * import cython
 */
  __pyx_t_2 = PyList_New(11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_FullArgSpec);
  __Pyx_GIVEREF(__pyx_n_s_FullArgSpec);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_FullArgSpec);
  __Pyx_INCREF(__pyx_n_s_CO_COROUTINE);
  __Pyx_GIVEREF(__pyx_n_s_CO_COROUTINE);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_CO_COROUTINE);
  __Pyx_INCREF(__pyx_n_s_CO_GENERATOR);
  __Pyx_GIVEREF(__pyx_n_s_CO_GENERATOR);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_CO_GENERATOR);
  __Pyx_INCREF(__pyx_n_s_CO_ASYNC_GENERATOR);
  __Pyx_GIVEREF(__pyx_n_s_CO_ASYNC_GENERATOR);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_CO_ASYNC_GENERATOR);
  __Pyx_INCREF(__pyx_n_s_CORO_CLOSED);
  __Pyx_GIVEREF(__pyx_n_s_CORO_CLOSED);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_CORO_CLOSED);
  __Pyx_INCREF(__pyx_n_s_CORO_CREATED);
  __Pyx_GIVEREF(__pyx_n_s_CORO_CREATED);
  PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_CORO_CREATED);
  __Pyx_INCREF(__pyx_n_s_CORO_RUNNING);
  __Pyx_GIVEREF(__pyx_n_s_CORO_RUNNING);
  PyList_SET_ITEM(__pyx_t_2, 6, __pyx_n_s_CORO_RUNNING);
  __Pyx_INCREF(__pyx_n_s_CORO_SUSPENDED);
  __Pyx_GIVEREF(__pyx_n_s_CORO_SUSPENDED);
  PyList_SET_ITEM(__pyx_t_2, 7, __pyx_n_s_CORO_SUSPENDED);
  __Pyx_INCREF(__pyx_n_s_CO_VARARGS);
  __Pyx_GIVEREF(__pyx_n_s_CO_VARARGS);
  PyList_SET_ITEM(__pyx_t_2, 8, __pyx_n_s_CO_VARARGS);
  __Pyx_INCREF(__pyx_n_s_CO_VARKEYWORDS);
  __Pyx_GIVEREF(__pyx_n_s_CO_VARKEYWORDS);
  PyList_SET_ITEM(__pyx_t_2, 9, __pyx_n_s_CO_VARKEYWORDS);
  __Pyx_INCREF(__pyx_n_s_CO_ITERABLE_COROUTINE);
  __Pyx_GIVEREF(__pyx_n_s_CO_ITERABLE_COROUTINE);
  PyList_SET_ITEM(__pyx_t_2, 10, __pyx_n_s_CO_ITERABLE_COROUTINE);
  __pyx_t_1 = __Pyx_patch_inspect(__Pyx_Import(__pyx_n_s_inspect, __pyx_t_2, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_FullArgSpec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FullArgSpec, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_COROUTINE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_COROUTINE, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_GENERATOR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_GENERATOR, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_ASYNC_GENERATOR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_ASYNC_GENERATOR, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CORO_CLOSED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_CLOSED, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CORO_CREATED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_CREATED, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CORO_RUNNING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_RUNNING, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CORO_SUSPENDED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_SUSPENDED, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_VARARGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_VARARGS, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_VARKEYWORDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_VARKEYWORDS, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CO_ITERABLE_COROUTINE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CO_ITERABLE_COROUTINE, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":56
 *                      CORO_RUNNING, CORO_SUSPENDED, CO_VARARGS, CO_VARKEYWORDS, CO_ITERABLE_COROUTINE)
 * import cython
 * from inspect import getfullargspec             # <<<<<<<<<<<<<<
 * from ..fn.underscore import ArityError
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_getfullargspec);
  __Pyx_GIVEREF(__pyx_n_s_getfullargspec);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_getfullargspec);
  __pyx_t_2 = __Pyx_patch_inspect(__Pyx_Import(__pyx_n_s_inspect, __pyx_t_1, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getfullargspec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getfullargspec, __pyx_t_1) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":57
 * import cython
 * from inspect import getfullargspec
 * from ..fn.underscore import ArityError             # <<<<<<<<<<<<<<
 * 
 * inPy2 = sys.version_info[0] == 2
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ArityError);
  __Pyx_GIVEREF(__pyx_n_s_ArityError);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ArityError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fn_underscore, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ArityError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ArityError, __pyx_t_2) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":59
 * from ..fn.underscore import ArityError
 * 
 * inPy2 = sys.version_info[0] == 2             # <<<<<<<<<<<<<<
 * if inPy2:  # pragma: no cover
 *     from types import ClassType
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_version_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inPy2, __pyx_t_2) < 0) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":60
 * 
 * inPy2 = sys.version_info[0] == 2
 * if inPy2:  # pragma: no cover             # <<<<<<<<<<<<<<
 *     from types import ClassType
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_inPy2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":61
 * inPy2 = sys.version_info[0] == 2
 * if inPy2:  # pragma: no cover
 *     from types import ClassType             # <<<<<<<<<<<<<<
 * 
 * # Create constants for the compiler flags in Include/code.h
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_ClassType);
    __Pyx_GIVEREF(__pyx_n_s_ClassType);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ClassType);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_types, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ClassType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClassType, __pyx_t_2) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":60
 * 
 * inPy2 = sys.version_info[0] == 2
 * if inPy2:  # pragma: no cover             # <<<<<<<<<<<<<<
 *     from types import ClassType
 * 
 */
  }

  /* "Aspidites/_vendor/contracts/inspection.py":65
 * # Create constants for the compiler flags in Include/code.h
 * # We try to get them from dis to avoid duplication
 * mod_dict = globals()             # <<<<<<<<<<<<<<
 * for k, v in dis.COMPILER_FLAG_NAMES.items():
 *     mod_dict["CO_" + v] = k
 */
  __pyx_t_1 = __Pyx_Globals(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mod_dict, __pyx_t_1) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":66
 * # We try to get them from dis to avoid duplication
 * mod_dict = globals()
 * for k, v in dis.COMPILER_FLAG_NAMES.items():             # <<<<<<<<<<<<<<
 *     mod_dict["CO_" + v] = k
 * 
 */
  __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_dis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_COMPILER_FLAG_NAMES); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_7 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 66, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_dict_iterator(__pyx_t_7, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_5, &__pyx_t_4, &__pyx_t_2, &__pyx_t_7, NULL, __pyx_t_6);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_k, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_v, __pyx_t_7) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":67
 * mod_dict = globals()
 * for k, v in dis.COMPILER_FLAG_NAMES.items():
 *     mod_dict["CO_" + v] = k             # <<<<<<<<<<<<<<
 * 
 * # See Include/object.h
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_k); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_mod_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_v); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_n_u_CO, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_10, __pyx_t_7) < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":70
 * 
 * # See Include/object.h
 * TPFLAGS_IS_ABSTRACT = 1 << 20             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TPFLAGS_IS_ABSTRACT, __pyx_int_1048576) < 0) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":74
 * 
 * # ----------------------------------------------------------- type-checking
 * def ismodule(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a module.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_1ismodule, 0, __pyx_n_s_ismodule, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ismodule, __pyx_t_1) < 0) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":84
 * 
 * 
 * def isclass(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a class.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_3isclass, 0, __pyx_n_s_isclass, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isclass, __pyx_t_1) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":93
 * 
 * 
 * def ismethod(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an instance method.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_5ismethod, 0, __pyx_n_s_ismethod, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ismethod, __pyx_t_1) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":104
 * 
 * 
 * def ismethoddescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a method descriptor.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_7ismethoddescriptor, 0, __pyx_n_s_ismethoddescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ismethoddescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":125
 * 
 * 
 * def isdatadescriptor(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a data descriptor.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9isdatadescriptor, 0, __pyx_n_s_isdatadescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isdatadescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":140
 * 
 * 
 * if hasattr(types, 'MemberDescriptorType'):             # <<<<<<<<<<<<<<
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_u_MemberDescriptorType); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = (__pyx_t_3 != 0);
  if (__pyx_t_11) {

    /* "Aspidites/_vendor/contracts/inspection.py":142
 * if hasattr(types, 'MemberDescriptorType'):
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11ismemberdescriptor, 0, __pyx_n_s_ismemberdescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ismemberdescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":140
 * 
 * 
 * if hasattr(types, 'MemberDescriptorType'):             # <<<<<<<<<<<<<<
 *     # CPython and equivalent
 *     def ismemberdescriptor(object):
 */
    goto __pyx_L5;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":150
 * else:
 *     # Other implementations
 *     def ismemberdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a member descriptor.
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_13ismemberdescriptor, 0, __pyx_n_s_ismemberdescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ismemberdescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "Aspidites/_vendor/contracts/inspection.py":157
 *         return False
 * 
 * if hasattr(types, 'GetSetDescriptorType'):             # <<<<<<<<<<<<<<
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_u_GetSetDescriptorType); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_11 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/contracts/inspection.py":159
 * if hasattr(types, 'GetSetDescriptorType'):
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_15isgetsetdescriptor, 0, __pyx_n_s_isgetsetdescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_isgetsetdescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/contracts/inspection.py":157
 *         return False
 * 
 * if hasattr(types, 'GetSetDescriptorType'):             # <<<<<<<<<<<<<<
 *     # CPython and equivalent
 *     def isgetsetdescriptor(object):
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/contracts/inspection.py":167
 * else:
 *     # Other implementations
 *     def isgetsetdescriptor(object):             # <<<<<<<<<<<<<<
 *         """Return true if the object is a getset descriptor.
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_17isgetsetdescriptor, 0, __pyx_n_s_isgetsetdescriptor, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_isgetsetdescriptor, __pyx_t_1) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L6:;

  /* "Aspidites/_vendor/contracts/inspection.py":175
 * 
 * 
 * def isfunction(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined function.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_19isfunction, 0, __pyx_n_s_isfunction, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isfunction, __pyx_t_1) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":189
 * 
 * 
 * def _has_code_flag(f, flag):             # <<<<<<<<<<<<<<
 *     """Return true if ``f`` is a function (or a method or functools.partial
 *     wrapper wrapping a function) whose code object has the given ``flag``
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_21_has_code_flag, 0, __pyx_n_s_has_code_flag, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_has_code_flag, __pyx_t_1) < 0) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":201
 * 
 * 
 * def isgeneratorfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a user-defined generator function.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_23isgeneratorfunction, 0, __pyx_n_s_isgeneratorfunction, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isgeneratorfunction, __pyx_t_1) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":209
 * 
 * 
 * def iscoroutinefunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine function.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_25iscoroutinefunction, 0, __pyx_n_s_iscoroutinefunction, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iscoroutinefunction, __pyx_t_1) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":217
 * 
 * 
 * def isasyncgenfunction(obj):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator function.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_27isasyncgenfunction, 0, __pyx_n_s_isasyncgenfunction, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isasyncgenfunction, __pyx_t_1) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":226
 * 
 * 
 * def isasyncgen(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an asynchronous generator."""
 *     return isinstance(object, types.AsyncGeneratorType)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_29isasyncgen, 0, __pyx_n_s_isasyncgen, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isasyncgen, __pyx_t_1) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":231
 * 
 * 
 * def isgenerator(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a generator.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_31isgenerator, 0, __pyx_n_s_isgenerator, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isgenerator, __pyx_t_1) < 0) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":249
 * 
 * 
 * def iscoroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a coroutine."""
 *     return isinstance(object, types.CoroutineType)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_33iscoroutine, 0, __pyx_n_s_iscoroutine, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iscoroutine, __pyx_t_1) < 0) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":254
 * 
 * 
 * def isawaitable(object):             # <<<<<<<<<<<<<<
 *     """Return true if object can be passed to an ``await`` expression."""
 *     return (isinstance(object, types.CoroutineType) or
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_35isawaitable, 0, __pyx_n_s_isawaitable, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isawaitable, __pyx_t_1) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":262
 * 
 * 
 * def istraceback(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a traceback.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_37istraceback, 0, __pyx_n_s_istraceback, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_istraceback, __pyx_t_1) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":273
 * 
 * 
 * def isframe(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a frame object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_39isframe, 0, __pyx_n_s_isframe, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isframe, __pyx_t_1) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":288
 * 
 * 
 * def iscode(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a code object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_41iscode, 0, __pyx_n_s_iscode, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iscode, __pyx_t_1) < 0) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":314
 * 
 * 
 * def isbuiltin(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is a built-in function or method.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_43isbuiltin, 0, __pyx_n_s_isbuiltin, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isbuiltin, __pyx_t_1) < 0) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":324
 * 
 * 
 * def isroutine(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is any kind of function or method."""
 *     return (isbuiltin(object)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_45isroutine, 0, __pyx_n_s_isroutine, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isroutine, __pyx_t_1) < 0) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":332
 * 
 * 
 * def isabstract(object):             # <<<<<<<<<<<<<<
 *     """Return true if the object is an abstract base class (ABC)."""
 *     if not isinstance(object, type):
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_47isabstract, 0, __pyx_n_s_isabstract, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isabstract, __pyx_t_1) < 0) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":357
 * 
 * 
 * def getmembers(object, predicate=None):             # <<<<<<<<<<<<<<
 *     """Return all members of an object as (name, value) pairs sorted by name.
 *     Optionally, only return members that satisfy a given predicate."""
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_49getmembers, 0, __pyx_n_s_getmembers, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__137);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getmembers, __pyx_t_1) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":402
 * 
 * 
 * Attribute = namedtuple('Attribute', 'name kind defining_class object')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__138, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Attribute, __pyx_t_7) < 0) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":405
 * 
 * 
 * def classify_class_attrs(cls):             # <<<<<<<<<<<<<<
 *     """Return list of attribute-descriptor tuples.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_51classify_class_attrs, 0, __pyx_n_s_classify_class_attrs, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_classify_class_attrs, __pyx_t_7) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":522
 * # ----------------------------------------------------------- class helpers
 * 
 * def getmro(cls):             # <<<<<<<<<<<<<<
 *     """Return tuple of base classes (including cls) in method resolution order."""
 *     return cls.__mro__
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_53getmro, 0, __pyx_n_s_getmro, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getmro, __pyx_t_7) < 0) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":529
 * # -------------------------------------------------------- function helpers
 * 
 * def unwrap(func, *, stop=None):             # <<<<<<<<<<<<<<
 *     """Get the object wrapped by *func*.
 * 
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_stop, ((PyObject *)Py_None)) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_55unwrap, 0, __pyx_n_s_unwrap, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_1, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unwrap, __pyx_t_1) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":566
 * 
 * # -------------------------------------------------- source code extraction
 * def indentsize(line):             # <<<<<<<<<<<<<<
 *     """Return the indent size, in spaces, at the start of a line of text."""
 *     expline = line.expandtabs()
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_57indentsize, 0, __pyx_n_s_indentsize, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_indentsize, __pyx_t_1) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":572
 * 
 * 
 * def _findclass(func):             # <<<<<<<<<<<<<<
 *     cls = sys.modules.get(func.__module__)
 *     if cls is None:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_59_findclass, 0, __pyx_n_s_findclass, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_findclass, __pyx_t_1) < 0) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":583
 * 
 * 
 * def _finddoc(obj):             # <<<<<<<<<<<<<<
 *     if isclass(obj):
 *         for base in obj.__mro__:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_61_finddoc, 0, __pyx_n_s_finddoc, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_finddoc, __pyx_t_1) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":646
 * 
 * 
 * def getdoc(object):             # <<<<<<<<<<<<<<
 *     """Get the documentation string for an object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_63getdoc, 0, __pyx_n_s_getdoc, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getdoc, __pyx_t_1) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":666
 * 
 * 
 * def cleandoc(doc):             # <<<<<<<<<<<<<<
 *     """Clean up indentation from docstrings.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_65cleandoc, 0, __pyx_n_s_cleandoc, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cleandoc, __pyx_t_1) < 0) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":696
 * 
 * 
 * def getfile(object):             # <<<<<<<<<<<<<<
 *     """Work out which source or compiled file an object was defined in."""
 *     if ismodule(object):
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_67getfile, 0, __pyx_n_s_getfile, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getfile, __pyx_t_1) < 0) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":723
 * 
 * 
 * def getmodulename(path):             # <<<<<<<<<<<<<<
 *     """Return the module name for a given file, or None."""
 *     fname = os.path.basename(path)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_69getmodulename, 0, __pyx_n_s_getmodulename, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getmodulename, __pyx_t_1) < 0) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":736
 * 
 * 
 * def getsourcefile(object):             # <<<<<<<<<<<<<<
 *     """Return the filename that can be used to locate an object's source.
 *     Return None if no way can be identified to get the source.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_71getsourcefile, 0, __pyx_n_s_getsourcefile, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getsourcefile, __pyx_t_1) < 0) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":759
 * 
 * 
 * def getabsfile(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return an absolute path to the source or compiled file for an object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_73getabsfile, 0, __pyx_n_s_getabsfile, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__163);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getabsfile, __pyx_t_1) < 0) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":769
 * 
 * 
 * modulesbyfile = {}             # <<<<<<<<<<<<<<
 * _filesbymodname = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_modulesbyfile, __pyx_t_1) < 0) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":770
 * 
 * modulesbyfile = {}
 * _filesbymodname = {}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_filesbymodname, __pyx_t_1) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":773
 * 
 * 
 * def getmodule(object, _filename=None):             # <<<<<<<<<<<<<<
 *     """Return the module an object was defined in, or None if not found."""
 *     if ismodule(object):
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_75getmodule, 0, __pyx_n_s_getmodule, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__166);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getmodule, __pyx_t_1) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":820
 * 
 * 
 * class ClassFoundException(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_1, __pyx_n_s_ClassFoundException, __pyx_n_s_ClassFoundException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_ClassFoundException, __pyx_t_1, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClassFoundException, __pyx_t_2) < 0) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":824
 * 
 * 
 * class _ClassFinder(ast.NodeVisitor):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, qualname):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ast); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_NodeVisitor); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_1, __pyx_n_s_ClassFinder, __pyx_n_s_ClassFinder, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "Aspidites/_vendor/contracts/inspection.py":826
 * class _ClassFinder(ast.NodeVisitor):
 * 
 *     def __init__(self, qualname):             # <<<<<<<<<<<<<<
 *         self.stack = []
 *         self.qualname = qualname
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_1__init__, 0, __pyx_n_s_ClassFinder___init, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":830
 *         self.qualname = qualname
 * 
 *     def visit_FunctionDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         self.stack.append('<locals>')
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_3visit_FunctionDef, 0, __pyx_n_s_ClassFinder_visit_FunctionDef, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_n_s_visit_FunctionDef, __pyx_t_2) < 0) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":837
 *         self.stack.pop()
 * 
 *     visit_AsyncFunctionDef = visit_FunctionDef             # <<<<<<<<<<<<<<
 * 
 *     def visit_ClassDef(self, node):
 */
  __pyx_t_2 = PyObject_GetItem(__pyx_t_10, __pyx_n_s_visit_FunctionDef);
  if (unlikely(!__pyx_t_2)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_visit_FunctionDef);
  }
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_n_s_visit_AsyncFunctionDef, __pyx_t_2) < 0) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":839
 *     visit_AsyncFunctionDef = visit_FunctionDef
 * 
 *     def visit_ClassDef(self, node):             # <<<<<<<<<<<<<<
 *         self.stack.append(node.name)
 *         if self.qualname == '.'.join(self.stack):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_12_ClassFinder_5visit_ClassDef, 0, __pyx_n_s_ClassFinder_visit_ClassDef, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_n_s_visit_ClassDef, __pyx_t_2) < 0) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":824
 * 
 * 
 * class _ClassFinder(ast.NodeVisitor):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, qualname):
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_ClassFinder, __pyx_t_1, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClassFinder, __pyx_t_2) < 0) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":855
 * 
 * 
 * def findsource(object):             # <<<<<<<<<<<<<<
 *     """Return the entire source file and starting line number for an object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_77findsource, 0, __pyx_n_s_findsource, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_findsource, __pyx_t_1) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":924
 * 
 * 
 * def getcomments(object):             # <<<<<<<<<<<<<<
 *     """Get lines of comments immediately preceding an object's source code.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_79getcomments, 0, __pyx_n_s_getcomments, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getcomments, __pyx_t_1) < 0) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":971
 * 
 * 
 * class EndOfBlock(Exception): pass             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_1, __pyx_n_s_EndOfBlock, __pyx_n_s_EndOfBlock, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_EndOfBlock, __pyx_t_1, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EndOfBlock, __pyx_t_2) < 0) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":974
 * 
 * 
 * class BlockFinder:             # <<<<<<<<<<<<<<
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_BlockFinder, __pyx_n_s_BlockFinder, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_Provide_a_tokeneater_method_to_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":977
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.indent = 0
 *         self.islambda = False
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_1__init__, 0, __pyx_n_s_BlockFinder___init, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":987
 *         self.body_col0 = None
 * 
 *     def tokeneater(self, type, token, srowcol, erowcol, line):             # <<<<<<<<<<<<<<
 *         if not self.started and not self.indecorator:
 *             # skip any decorators
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_11BlockFinder_3tokeneater, 0, __pyx_n_s_BlockFinder_tokeneater, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_tokeneater, __pyx_t_7) < 0) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":974
 * 
 * 
 * class BlockFinder:             # <<<<<<<<<<<<<<
 *     """Provide a tokeneater() method to detect the end of a code block."""
 * 
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_BlockFinder, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BlockFinder, __pyx_t_7) < 0) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1038
 * 
 * 
 * def getblock(lines):             # <<<<<<<<<<<<<<
 *     """Extract the block of code at the top of the given list of lines."""
 *     blockfinder = BlockFinder()
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_81getblock, 0, __pyx_n_s_getblock, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getblock, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1050
 * 
 * 
 * def getsourcelines(object):             # <<<<<<<<<<<<<<
 *     """Return a list of source lines and starting line number for an object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_83getsourcelines, 0, __pyx_n_s_getsourcelines, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getsourcelines, __pyx_t_1) < 0) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1072
 * 
 * 
 * def getsource(object):             # <<<<<<<<<<<<<<
 *     """Return the text of the source code for an object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_85getsource, 0, __pyx_n_s_getsource, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__186)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getsource, __pyx_t_1) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1083
 * 
 * # --------------------------------------------------- class tree extraction
 * def walktree(classes, children, parent):             # <<<<<<<<<<<<<<
 *     """Recursive helper function for getclasstree()."""
 *     results = []
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_87walktree, 0, __pyx_n_s_walktree, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__188)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_walktree, __pyx_t_1) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1094
 * 
 * 
 * def getclasstree(classes, unique=False):             # <<<<<<<<<<<<<<
 *     """Arrange the given list of classes into a hierarchy of nested lists.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_89getclasstree, 0, __pyx_n_s_getclasstree, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__191);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getclasstree, __pyx_t_1) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1122
 * 
 * # ------------------------------------------------ argument list extraction
 * Arguments = namedtuple('Arguments', 'args, varargs, varkw')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__192, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Arguments, __pyx_t_7) < 0) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1125
 * 
 * 
 * def getargs(co):             # <<<<<<<<<<<<<<
 *     """Get information about the arguments accepted by a code object.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_91getargs, 0, __pyx_n_s_getargs, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__194)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getargs, __pyx_t_7) < 0) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1152
 *     return Arguments(args + kwonlyargs, varargs, varkw)
 * 
 * ArgSpec = namedtuple('ArgSpec', 'args varargs keywords defaults')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__195, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ArgSpec, __pyx_t_1) < 0) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1155
 * 
 * 
 * def getargspec(func):             # <<<<<<<<<<<<<<
 *     """Get the names and default values of a function's parameters.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_93getargspec, 0, __pyx_n_s_getargspec, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__197)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getargspec, __pyx_t_1) < 0) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1186
 * 
 * 
 * FullArgSpec = namedtuple('FullArgSpec', 'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations')             # <<<<<<<<<<<<<<
 * 
 * ArgInfo = namedtuple('ArgInfo', 'args varargs keywords locals')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__198, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FullArgSpec, __pyx_t_7) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1188
 * FullArgSpec = namedtuple('FullArgSpec', 'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations')
 * 
 * ArgInfo = namedtuple('ArgInfo', 'args varargs keywords locals')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__199, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ArgInfo, __pyx_t_1) < 0) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1191
 * 
 * 
 * def getargvalues(frame):             # <<<<<<<<<<<<<<
 *     """Get information about arguments passed into a particular frame.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_95getargvalues, 0, __pyx_n_s_getargvalues, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__201)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getargvalues, __pyx_t_1) < 0) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1202
 * 
 * 
 * def formatannotation(annotation, base_module=None):             # <<<<<<<<<<<<<<
 *     if getattr(annotation, '__module__', None) == 'typing':
 *         return repr(annotation).replace('typing.', '')
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_97formatannotation, 0, __pyx_n_s_formatannotation, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__203)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__204);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_formatannotation, __pyx_t_1) < 0) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1212
 * 
 * 
 * def formatannotationrelativeto(object):             # <<<<<<<<<<<<<<
 *     module = getattr(object, '__module__', None)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_99formatannotationrelativeto, 0, __pyx_n_s_formatannotationrelativeto, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__206)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_formatannotationrelativeto, __pyx_t_1) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1221
 * 
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,             # <<<<<<<<<<<<<<
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_101formatargspec, 0, __pyx_n_s_formatargspec, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults), 8)) __PYX_ERR(0, 1221, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1222
 * 
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},             # <<<<<<<<<<<<<<
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_kwonlydefaults = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_annotations = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1223
 * def formatargspec(args, varargs=None, varkw=None, defaults=None,
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,             # <<<<<<<<<<<<<<
 *                   formatvarargs=lambda name: '*' + name,
 *                   formatvarkw=lambda name: '**' + name,
 */
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatarg = ((PyObject *)(&PyUnicode_Type));
  __Pyx_GIVEREF((&PyUnicode_Type));

  /* "Aspidites/_vendor/contracts/inspection.py":1224
 *                   kwonlyargs=(), kwonlydefaults={}, annotations={},
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,             # <<<<<<<<<<<<<<
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_184lambda4, 0, __pyx_n_s_formatargspec_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatvarargs = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1225
 *                   formatarg=str,
 *                   formatvarargs=lambda name: '*' + name,
 *                   formatvarkw=lambda name: '**' + name,             # <<<<<<<<<<<<<<
 *                   formatvalue=lambda value: '=' + repr(value),
 *                   formatreturns=lambda text: ' -> ' + text,
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_185lambda5, 0, __pyx_n_s_formatargspec_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatvarkw = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1226
 *                   formatvarargs=lambda name: '*' + name,
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),             # <<<<<<<<<<<<<<
 *                   formatreturns=lambda text: ' -> ' + text,
 *                   formatannotation=formatannotation):
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_186lambda6, 0, __pyx_n_s_formatargspec_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatvalue = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1227
 *                   formatvarkw=lambda name: '**' + name,
 *                   formatvalue=lambda value: '=' + repr(value),
 *                   formatreturns=lambda text: ' -> ' + text,             # <<<<<<<<<<<<<<
 *                   formatannotation=formatannotation):
 *     """Format an argument spec from the values returned by getfullargspec.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_187lambda7, 0, __pyx_n_s_formatargspec_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatreturns = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1228
 *                   formatvalue=lambda value: '=' + repr(value),
 *                   formatreturns=lambda text: ' -> ' + text,
 *                   formatannotation=formatannotation):             # <<<<<<<<<<<<<<
 *     """Format an argument spec from the values returned by getfullargspec.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_formatannotation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_formatannotation = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_191__defaults__);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_formatargspec, __pyx_t_1) < 0) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1282
 * 
 * 
 * def formatargvalues(args, varargs, varkw, locals,             # <<<<<<<<<<<<<<
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_103formatargvalues, 0, __pyx_n_s_formatargvalues, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults1), 4)) __PYX_ERR(0, 1282, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1283
 * 
 * def formatargvalues(args, varargs, varkw, locals,
 *                     formatarg=str,             # <<<<<<<<<<<<<<
 *                     formatvarargs=lambda name: '*' + name,
 *                     formatvarkw=lambda name: '**' + name,
 */
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_1)->__pyx_arg_formatarg = ((PyObject *)(&PyUnicode_Type));
  __Pyx_GIVEREF((&PyUnicode_Type));

  /* "Aspidites/_vendor/contracts/inspection.py":1284
 * def formatargvalues(args, varargs, varkw, locals,
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,             # <<<<<<<<<<<<<<
 *                     formatvarkw=lambda name: '**' + name,
 *                     formatvalue=lambda value: '=' + repr(value)):
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_188lambda8, 0, __pyx_n_s_formatargvalues_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_1)->__pyx_arg_formatvarargs = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1285
 *                     formatarg=str,
 *                     formatvarargs=lambda name: '*' + name,
 *                     formatvarkw=lambda name: '**' + name,             # <<<<<<<<<<<<<<
 *                     formatvalue=lambda value: '=' + repr(value)):
 *     """Format an argument spec from the 4 values returned by getargvalues.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_189lambda9, 0, __pyx_n_s_formatargvalues_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_1)->__pyx_arg_formatvarkw = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1286
 *                     formatvarargs=lambda name: '*' + name,
 *                     formatvarkw=lambda name: '**' + name,
 *                     formatvalue=lambda value: '=' + repr(value)):             # <<<<<<<<<<<<<<
 *     """Format an argument spec from the 4 values returned by getargvalues.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_190lambda10, 0, __pyx_n_s_formatargvalues_locals_lambda, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_1)->__pyx_arg_formatvalue = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_193__defaults__);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_formatargvalues, __pyx_t_1) < 0) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1308
 * 
 * 
 * def _missing_arguments(f_name, argnames, pos, values):             # <<<<<<<<<<<<<<
 *     names = [repr(name) for name in argnames if name not in values]
 *     missing = len(names)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_105_missing_arguments, 0, __pyx_n_s_missing_arguments, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__212)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_missing_arguments, __pyx_t_1) < 0) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1325
 * 
 * 
 * def _too_many(f_name, args, kwonly, varargs, defcount, given, values):             # <<<<<<<<<<<<<<
 *     atleast = len(args) - defcount
 *     kwonly_given = len([arg for arg in kwonly if arg in values])
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_107_too_many, 0, __pyx_n_s_too_many, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__214)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_too_many, __pyx_t_1) < 0) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1347
 * 
 * 
 * def getcallargs(func, *positional, **named):             # <<<<<<<<<<<<<<
 *     f_name: str
 *     # noinspection PyUnresolvedReferences
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_109getcallargs, 0, __pyx_n_s_getcallargs, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getcallargs, __pyx_t_1) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1464
 * 
 * 
 * ClosureVars = namedtuple('ClosureVars', 'nonlocals globals builtins unbound')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__217, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClosureVars, __pyx_t_7) < 0) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1467
 * 
 * 
 * def getclosurevars(func):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of free variables to their current values.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_111getclosurevars, 0, __pyx_n_s_getclosurevars, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__219)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getclosurevars, __pyx_t_7) < 0) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1521
 * # -------------------------------------------------- stack frame extraction
 * 
 * Traceback = namedtuple('Traceback', 'filename lineno function code_context index')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__220, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Traceback, __pyx_t_1) < 0) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1524
 * 
 * 
 * def getframeinfo(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get information about a frame or traceback object.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_113getframeinfo, 0, __pyx_n_s_getframeinfo, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__223);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getframeinfo, __pyx_t_1) < 0) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1557
 * 
 * 
 * def getlineno(frame):             # <<<<<<<<<<<<<<
 *     """Get the line number from a frame object, allowing for optimization."""
 *     # FrameType.f_lineno is now a descriptor that grovels co_lnotab
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_115getlineno, 0, __pyx_n_s_getlineno, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__225)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getlineno, __pyx_t_1) < 0) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1563
 * 
 * 
 * FrameInfo = namedtuple('FrameInfo', ('frame',) + Traceback._fields)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Traceback); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_fields); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyNumber_Add(__pyx_tuple__226, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_u_FrameInfo);
  __Pyx_GIVEREF(__pyx_n_u_FrameInfo);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_u_FrameInfo);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FrameInfo, __pyx_t_7) < 0) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1566
 * 
 * 
 * def getouterframes(frame, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a frame and all higher (calling) frames.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_117getouterframes, 0, __pyx_n_s_getouterframes, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__228)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__229);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getouterframes, __pyx_t_7) < 0) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1579
 * 
 * 
 * def getinnerframes(tb, context=1):             # <<<<<<<<<<<<<<
 *     """Get a list of records for a traceback's frame and all lower frames.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_119getinnerframes, 0, __pyx_n_s_getinnerframes, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__231)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__232);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getinnerframes, __pyx_t_7) < 0) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1592
 * 
 * 
 * def currentframe():             # <<<<<<<<<<<<<<
 *     """Return the frame of the caller or None if this is not possible."""
 *     return sys._getframe(1) if hasattr(sys, "_getframe") else None
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_121currentframe, 0, __pyx_n_s_currentframe, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__233)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_currentframe, __pyx_t_7) < 0) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1597
 * 
 * 
 * def stack(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack above the caller's frame."""
 *     # noinspection PyProtectedMember, PyUnresolvedReferences
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_123stack, 0, __pyx_n_s_stack, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__235)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__236);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stack, __pyx_t_7) < 0) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1603
 * 
 * 
 * def trace(context=1):             # <<<<<<<<<<<<<<
 *     """Return a list of records for the stack below the current exception."""
 *     # noinspection PyTypeChecker
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_125trace, 0, __pyx_n_s_trace, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__238)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__239);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trace, __pyx_t_7) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1611
 * # ------------------------------------------------ static version of getattr
 * 
 * _sentinel = object()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sentinel, __pyx_t_7) < 0) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1614
 * 
 * 
 * def _static_getmro(klass):             # <<<<<<<<<<<<<<
 *     return type.__dict__['__mro__'].__get__(klass)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_127_static_getmro, 0, __pyx_n_s_static_getmro, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__241)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_static_getmro, __pyx_t_7) < 0) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1618
 * 
 * 
 * def _check_instance(obj, attr):             # <<<<<<<<<<<<<<
 *     instance_dict = {}
 *     try:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_129_check_instance, 0, __pyx_n_s_check_instance, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__243)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_instance, __pyx_t_7) < 0) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1627
 * 
 * 
 * def _check_class(klass, attr):             # <<<<<<<<<<<<<<
 *     for entry in _static_getmro(klass):
 *         if _shadowed_dict(type(entry)) is _sentinel:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_131_check_class, 0, __pyx_n_s_check_class, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__245)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_class, __pyx_t_7) < 0) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1637
 * 
 * 
 * def _is_type(obj):             # <<<<<<<<<<<<<<
 *     try:
 *         _static_getmro(obj)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_133_is_type, 0, __pyx_n_s_is_type, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__247)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_type, __pyx_t_7) < 0) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1645
 * 
 * 
 * def _shadowed_dict(klass):             # <<<<<<<<<<<<<<
 *     dict_attr = type.__dict__["__dict__"]
 *     for entry in _static_getmro(klass):
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_135_shadowed_dict, 0, __pyx_n_s_shadowed_dict, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__249)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_shadowed_dict, __pyx_t_7) < 0) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1660
 * 
 * 
 * def getattr_static(obj, attr, default=_sentinel):             # <<<<<<<<<<<<<<
 *     """Retrieve attributes without triggering dynamic lookup via the
 *        descriptor protocol,  __getattr__ or __getattribute__.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_137getattr_static, 0, __pyx_n_s_getattr_static, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__251)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults2), 1)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_sentinel); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_t_7)->__pyx_arg_default = __pyx_t_10;
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_195__defaults__);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getattr_static, __pyx_t_7) < 0) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1708
 * # ------------------------------------------------ generator introspection
 * 
 * GEN_CREATED = 'GEN_CREATED'             # <<<<<<<<<<<<<<
 * GEN_RUNNING = 'GEN_RUNNING'
 * GEN_SUSPENDED = 'GEN_SUSPENDED'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEN_CREATED, __pyx_n_u_GEN_CREATED) < 0) __PYX_ERR(0, 1708, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1709
 * 
 * GEN_CREATED = 'GEN_CREATED'
 * GEN_RUNNING = 'GEN_RUNNING'             # <<<<<<<<<<<<<<
 * GEN_SUSPENDED = 'GEN_SUSPENDED'
 * GEN_CLOSED = 'GEN_CLOSED'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEN_RUNNING, __pyx_n_u_GEN_RUNNING) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1710
 * GEN_CREATED = 'GEN_CREATED'
 * GEN_RUNNING = 'GEN_RUNNING'
 * GEN_SUSPENDED = 'GEN_SUSPENDED'             # <<<<<<<<<<<<<<
 * GEN_CLOSED = 'GEN_CLOSED'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEN_SUSPENDED, __pyx_n_u_GEN_SUSPENDED) < 0) __PYX_ERR(0, 1710, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1711
 * GEN_RUNNING = 'GEN_RUNNING'
 * GEN_SUSPENDED = 'GEN_SUSPENDED'
 * GEN_CLOSED = 'GEN_CLOSED'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEN_CLOSED, __pyx_n_u_GEN_CLOSED) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1714
 * 
 * 
 * def getgeneratorstate(generator):             # <<<<<<<<<<<<<<
 *     """Get current state of a generator-iterator.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_139getgeneratorstate, 0, __pyx_n_s_getgeneratorstate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__253)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getgeneratorstate, __pyx_t_7) < 0) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1732
 * 
 * 
 * def getgeneratorlocals(generator):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of generator local variables to their current values.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_141getgeneratorlocals, 0, __pyx_n_s_getgeneratorlocals, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__255)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getgeneratorlocals, __pyx_t_7) < 0) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1751
 * # ------------------------------------------------ coroutine introspection
 * 
 * CORO_CREATED = 'CORO_CREATED'             # <<<<<<<<<<<<<<
 * CORO_RUNNING = 'CORO_RUNNING'
 * CORO_SUSPENDED = 'CORO_SUSPENDED'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_CREATED, __pyx_n_u_CORO_CREATED) < 0) __PYX_ERR(0, 1751, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1752
 * 
 * CORO_CREATED = 'CORO_CREATED'
 * CORO_RUNNING = 'CORO_RUNNING'             # <<<<<<<<<<<<<<
 * CORO_SUSPENDED = 'CORO_SUSPENDED'
 * CORO_CLOSED = 'CORO_CLOSED'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_RUNNING, __pyx_n_u_CORO_RUNNING) < 0) __PYX_ERR(0, 1752, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1753
 * CORO_CREATED = 'CORO_CREATED'
 * CORO_RUNNING = 'CORO_RUNNING'
 * CORO_SUSPENDED = 'CORO_SUSPENDED'             # <<<<<<<<<<<<<<
 * CORO_CLOSED = 'CORO_CLOSED'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_SUSPENDED, __pyx_n_u_CORO_SUSPENDED) < 0) __PYX_ERR(0, 1753, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1754
 * CORO_RUNNING = 'CORO_RUNNING'
 * CORO_SUSPENDED = 'CORO_SUSPENDED'
 * CORO_CLOSED = 'CORO_CLOSED'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CORO_CLOSED, __pyx_n_u_CORO_CLOSED) < 0) __PYX_ERR(0, 1754, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":1757
 * 
 * 
 * def getcoroutinestate(coroutine):             # <<<<<<<<<<<<<<
 *     """Get current state of a coroutine object.
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_143getcoroutinestate, 0, __pyx_n_s_getcoroutinestate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__257)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getcoroutinestate, __pyx_t_7) < 0) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1775
 * 
 * 
 * def getcoroutinelocals(coroutine):             # <<<<<<<<<<<<<<
 *     """
 *     Get the mapping of coroutine local variables to their current values.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_145getcoroutinelocals, 0, __pyx_n_s_getcoroutinelocals, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__259)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getcoroutinelocals, __pyx_t_7) < 0) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1793
 * 
 * 
 * _WrapperDescriptor = type(type.__call__)             # <<<<<<<<<<<<<<
 * _MethodWrapper = type(all.__call__)
 * _ClassMethodWrapper = type(int.__dict__['from_bytes'])
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_call); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WrapperDescriptor, ((PyObject *)Py_TYPE(__pyx_t_7))) < 0) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1794
 * 
 * _WrapperDescriptor = type(type.__call__)
 * _MethodWrapper = type(all.__call__)             # <<<<<<<<<<<<<<
 * _ClassMethodWrapper = type(int.__dict__['from_bytes'])
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_all, __pyx_n_s_call); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MethodWrapper, ((PyObject *)Py_TYPE(__pyx_t_7))) < 0) __PYX_ERR(0, 1794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1795
 * _WrapperDescriptor = type(type.__call__)
 * _MethodWrapper = type(all.__call__)
 * _ClassMethodWrapper = type(int.__dict__['from_bytes'])             # <<<<<<<<<<<<<<
 * 
 * _NonUserDefinedCallables = (_WrapperDescriptor,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_u_from_bytes); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClassMethodWrapper, ((PyObject *)Py_TYPE(__pyx_t_10))) < 0) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1797
 * _ClassMethodWrapper = type(int.__dict__['from_bytes'])
 * 
 * _NonUserDefinedCallables = (_WrapperDescriptor,             # <<<<<<<<<<<<<<
 *                             _MethodWrapper,
 *                             _ClassMethodWrapper,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_WrapperDescriptor); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "Aspidites/_vendor/contracts/inspection.py":1798
 * 
 * _NonUserDefinedCallables = (_WrapperDescriptor,
 *                             _MethodWrapper,             # <<<<<<<<<<<<<<
 *                             _ClassMethodWrapper,
 *                             types.BuiltinFunctionType)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MethodWrapper); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/contracts/inspection.py":1799
 * _NonUserDefinedCallables = (_WrapperDescriptor,
 *                             _MethodWrapper,
 *                             _ClassMethodWrapper,             # <<<<<<<<<<<<<<
 *                             types.BuiltinFunctionType)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ClassMethodWrapper); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":1800
 *                             _MethodWrapper,
 *                             _ClassMethodWrapper,
 *                             types.BuiltinFunctionType)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BuiltinFunctionType); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1797
 * _ClassMethodWrapper = type(int.__dict__['from_bytes'])
 * 
 * _NonUserDefinedCallables = (_WrapperDescriptor,             # <<<<<<<<<<<<<<
 *                             _MethodWrapper,
 *                             _ClassMethodWrapper,
 */
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_9);
  __pyx_t_10 = 0;
  __pyx_t_7 = 0;
  __pyx_t_1 = 0;
  __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NonUserDefinedCallables, __pyx_t_2) < 0) __PYX_ERR(0, 1797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1803
 * 
 * 
 * def _signature_get_user_defined_method(cls, method_name):             # <<<<<<<<<<<<<<
 *     """Private helper. Checks if ``cls`` has an attribute
 *     named ``method_name`` and returns it only if it is a
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_147_signature_get_user_defined_method, 0, __pyx_n_s_signature_get_user_defined_meth, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__261)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_get_user_defined_meth, __pyx_t_2) < 0) __PYX_ERR(0, 1803, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1819
 * 
 * 
 * def _signature_get_partial(wrapped_sig, partial, extra_args=()):             # <<<<<<<<<<<<<<
 *     """Private helper to calculate how 'wrapped_sig' signature will
 *     look like after applying a 'functools.partial' object (or alike)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_149_signature_get_partial, 0, __pyx_n_s_signature_get_partial, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__263)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__264);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_get_partial, __pyx_t_2) < 0) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1894
 * 
 * 
 * def _signature_bound_method(sig):             # <<<<<<<<<<<<<<
 *     """Private helper to transform signatures for unbound
 *     functions to bound methods.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_151_signature_bound_method, 0, __pyx_n_s_signature_bound_method, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__266)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_bound_method, __pyx_t_2) < 0) __PYX_ERR(0, 1894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1920
 * 
 * 
 * def _signature_is_builtin(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a callable that might
 *     support Argument Clinic's __text_signature__ protocol.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_153_signature_is_builtin, 0, __pyx_n_s_signature_is_builtin, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__268)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_is_builtin, __pyx_t_2) < 0) __PYX_ERR(0, 1920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1932
 * 
 * 
 * def _signature_is_functionlike(obj):             # <<<<<<<<<<<<<<
 *     """Private helper to test if `obj` is a duck type of FunctionType.
 *     A good example of such objects are functions compiled with
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_155_signature_is_functionlike, 0, __pyx_n_s_signature_is_functionlike, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_is_functionlike, __pyx_t_2) < 0) __PYX_ERR(0, 1932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1957
 * 
 * 
 * def _signature_get_bound_param(spec):             # <<<<<<<<<<<<<<
 *     """ Private helper to get first parameter name from a
 *     __text_signature__ of a builtin method, which should
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_157_signature_get_bound_param, 0, __pyx_n_s_signature_get_bound_param, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_get_bound_param, __pyx_t_2) < 0) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1980
 * 
 * 
 * def _signature_strip_non_python_syntax(signature):             # <<<<<<<<<<<<<<
 *     """
 *     Private helper function. Takes a signature in Argument Clinic's
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_159_signature_strip_non_python_syntax, 0, __pyx_n_s_signature_strip_non_python_synt, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__274)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_strip_non_python_synt, __pyx_t_2) < 0) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2052
 * 
 * 
 * def _signature_fromstr(cls, obj, s, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper to parse content of '__text_signature__'
 *     and return a Signature based on it.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_161_signature_fromstr, 0, __pyx_n_s_signature_fromstr, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__276)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__277);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_fromstr, __pyx_t_2) < 0) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2197
 * 
 * 
 * def _signature_from_builtin(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper function to get signature for
 *     builtin callables.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_163_signature_from_builtin, 0, __pyx_n_s_signature_from_builtin, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__279)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__280);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_from_builtin, __pyx_t_2) < 0) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2213
 * 
 * 
 * def _signature_from_function(cls, func, skip_bound_arg=True):             # <<<<<<<<<<<<<<
 *     """Private helper: constructs Signature for the given python function."""
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_165_signature_from_function, 0, __pyx_n_s_signature_from_function, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__282)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__283);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_from_function, __pyx_t_2) < 0) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2307
 * 
 * 
 * def _signature_from_callable(obj, *,             # <<<<<<<<<<<<<<
 *                              follow_wrapper_chains=True,
 *                              skip_bound_arg=True,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_follow_wrapper_chains, ((PyObject *)Py_True)) < 0) __PYX_ERR(0, 2307, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_skip_bound_arg, ((PyObject *)Py_True)) < 0) __PYX_ERR(0, 2307, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_167_signature_from_callable, 0, __pyx_n_s_signature_from_callable, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__285)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_9, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_from_callable, __pyx_t_9) < 0) __PYX_ERR(0, 2307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2512
 * 
 * 
 * class _void:             # <<<<<<<<<<<<<<
 *     """A private marker - used in Parameter & Signature."""
 * 
 */
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_void, __pyx_n_s_void, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_A_private_marker_used_in_Paramet); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_void, __pyx_empty_tuple, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_void, __pyx_t_2) < 0) __PYX_ERR(0, 2512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2516
 * 
 * 
 * class _empty:             # <<<<<<<<<<<<<<
 *     """Marker object for Signature.empty and Parameter.empty."""
 * 
 */
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_empty, __pyx_n_s_empty, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_Marker_object_for_Signature_empt); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_empty, __pyx_empty_tuple, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_empty, __pyx_t_2) < 0) __PYX_ERR(0, 2516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2520
 * 
 * 
 * class _ParameterKind(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     POSITIONAL_ONLY = 0
 *     POSITIONAL_OR_KEYWORD = 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_enum); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_9, __pyx_n_s_ParameterKind, __pyx_n_s_ParameterKind, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2521
 * 
 * class _ParameterKind(enum.IntEnum):
 *     POSITIONAL_ONLY = 0             # <<<<<<<<<<<<<<
 *     POSITIONAL_OR_KEYWORD = 1
 *     VAR_POSITIONAL = 2
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_POSITIONAL_ONLY_2, __pyx_int_0) < 0) __PYX_ERR(0, 2521, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2522
 * class _ParameterKind(enum.IntEnum):
 *     POSITIONAL_ONLY = 0
 *     POSITIONAL_OR_KEYWORD = 1             # <<<<<<<<<<<<<<
 *     VAR_POSITIONAL = 2
 *     KEYWORD_ONLY = 3
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_POSITIONAL_OR_KEYWORD_2, __pyx_int_1) < 0) __PYX_ERR(0, 2522, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2523
 *     POSITIONAL_ONLY = 0
 *     POSITIONAL_OR_KEYWORD = 1
 *     VAR_POSITIONAL = 2             # <<<<<<<<<<<<<<
 *     KEYWORD_ONLY = 3
 *     VAR_KEYWORD = 4
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_VAR_POSITIONAL_2, __pyx_int_2) < 0) __PYX_ERR(0, 2523, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2524
 *     POSITIONAL_OR_KEYWORD = 1
 *     VAR_POSITIONAL = 2
 *     KEYWORD_ONLY = 3             # <<<<<<<<<<<<<<
 *     VAR_KEYWORD = 4
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_KEYWORD_ONLY_2, __pyx_int_3) < 0) __PYX_ERR(0, 2524, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2525
 *     VAR_POSITIONAL = 2
 *     KEYWORD_ONLY = 3
 *     VAR_KEYWORD = 4             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_VAR_KEYWORD_2, __pyx_int_4) < 0) __PYX_ERR(0, 2525, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2527
 *     VAR_KEYWORD = 4
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._name_
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_1__str__, 0, __pyx_n_s_ParameterKind___str, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__287)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_str, __pyx_t_7) < 0) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2531
 * 
 *     @property
 *     def description(self):             # <<<<<<<<<<<<<<
 *         return _PARAM_NAME_MAPPING[self]
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14_ParameterKind_3description, 0, __pyx_n_s_ParameterKind_description, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__289)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/contracts/inspection.py":2530
 *         return self._name_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def description(self):
 *         return _PARAM_NAME_MAPPING[self]
 */
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_description, __pyx_t_10) < 0) __PYX_ERR(0, 2531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2520
 * 
 * 
 * class _ParameterKind(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     POSITIONAL_ONLY = 0
 *     POSITIONAL_OR_KEYWORD = 1
 */
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParameterKind, __pyx_t_9, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParameterKind, __pyx_t_10) < 0) __PYX_ERR(0, 2520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2535
 * 
 * 
 * _POSITIONAL_ONLY = _ParameterKind.POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 * _POSITIONAL_OR_KEYWORD = _ParameterKind.POSITIONAL_OR_KEYWORD
 * _VAR_POSITIONAL = _ParameterKind.VAR_POSITIONAL
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_POSITIONAL_ONLY_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POSITIONAL_ONLY, __pyx_t_2) < 0) __PYX_ERR(0, 2535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2536
 * 
 * _POSITIONAL_ONLY = _ParameterKind.POSITIONAL_ONLY
 * _POSITIONAL_OR_KEYWORD = _ParameterKind.POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 * _VAR_POSITIONAL = _ParameterKind.VAR_POSITIONAL
 * _KEYWORD_ONLY = _ParameterKind.KEYWORD_ONLY
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_POSITIONAL_OR_KEYWORD_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POSITIONAL_OR_KEYWORD, __pyx_t_9) < 0) __PYX_ERR(0, 2536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2537
 * _POSITIONAL_ONLY = _ParameterKind.POSITIONAL_ONLY
 * _POSITIONAL_OR_KEYWORD = _ParameterKind.POSITIONAL_OR_KEYWORD
 * _VAR_POSITIONAL = _ParameterKind.VAR_POSITIONAL             # <<<<<<<<<<<<<<
 * _KEYWORD_ONLY = _ParameterKind.KEYWORD_ONLY
 * _VAR_KEYWORD = _ParameterKind.VAR_KEYWORD
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_VAR_POSITIONAL_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VAR_POSITIONAL, __pyx_t_2) < 0) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2538
 * _POSITIONAL_OR_KEYWORD = _ParameterKind.POSITIONAL_OR_KEYWORD
 * _VAR_POSITIONAL = _ParameterKind.VAR_POSITIONAL
 * _KEYWORD_ONLY = _ParameterKind.KEYWORD_ONLY             # <<<<<<<<<<<<<<
 * _VAR_KEYWORD = _ParameterKind.VAR_KEYWORD
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_KEYWORD_ONLY_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_KEYWORD_ONLY, __pyx_t_9) < 0) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2539
 * _VAR_POSITIONAL = _ParameterKind.VAR_POSITIONAL
 * _KEYWORD_ONLY = _ParameterKind.KEYWORD_ONLY
 * _VAR_KEYWORD = _ParameterKind.VAR_KEYWORD             # <<<<<<<<<<<<<<
 * 
 * _PARAM_NAME_MAPPING = {
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParameterKind); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_VAR_KEYWORD_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VAR_KEYWORD, __pyx_t_2) < 0) __PYX_ERR(0, 2539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2542
 * 
 * _PARAM_NAME_MAPPING = {
 *     _POSITIONAL_ONLY: 'positional-only',             # <<<<<<<<<<<<<<
 *     _POSITIONAL_OR_KEYWORD: 'positional or keyword',
 *     _VAR_POSITIONAL: 'variadic positional',
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_2, __pyx_t_9, __pyx_kp_u_positional_only) < 0) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2543
 * _PARAM_NAME_MAPPING = {
 *     _POSITIONAL_ONLY: 'positional-only',
 *     _POSITIONAL_OR_KEYWORD: 'positional or keyword',             # <<<<<<<<<<<<<<
 *     _VAR_POSITIONAL: 'variadic positional',
 *     _KEYWORD_ONLY: 'keyword-only',
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_2, __pyx_t_9, __pyx_kp_u_positional_or_keyword) < 0) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2544
 *     _POSITIONAL_ONLY: 'positional-only',
 *     _POSITIONAL_OR_KEYWORD: 'positional or keyword',
 *     _VAR_POSITIONAL: 'variadic positional',             # <<<<<<<<<<<<<<
 *     _KEYWORD_ONLY: 'keyword-only',
 *     _VAR_KEYWORD: 'variadic keyword'
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_2, __pyx_t_9, __pyx_kp_u_variadic_positional) < 0) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2545
 *     _POSITIONAL_OR_KEYWORD: 'positional or keyword',
 *     _VAR_POSITIONAL: 'variadic positional',
 *     _KEYWORD_ONLY: 'keyword-only',             # <<<<<<<<<<<<<<
 *     _VAR_KEYWORD: 'variadic keyword'
 * }
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_2, __pyx_t_9, __pyx_kp_u_keyword_only) < 0) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2546
 *     _VAR_POSITIONAL: 'variadic positional',
 *     _KEYWORD_ONLY: 'keyword-only',
 *     _VAR_KEYWORD: 'variadic keyword'             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_2, __pyx_t_9, __pyx_kp_u_variadic_keyword) < 0) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PARAM_NAME_MAPPING, __pyx_t_2) < 0) __PYX_ERR(0, 2541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2550
 * 
 * 
 * class Parameter:             # <<<<<<<<<<<<<<
 *     """Represents a parameter in a function signature.
 * 
 */
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_Parameter, __pyx_n_s_Parameter, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_Represents_a_parameter_in_a_func); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/contracts/inspection.py":2572
 *     """
 * 
 *     __slots__ = ('_name', '_kind', '_default', '_annotation')             # <<<<<<<<<<<<<<
 * 
 *     POSITIONAL_ONLY = _POSITIONAL_ONLY
 */
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_slots, __pyx_tuple__290) < 0) __PYX_ERR(0, 2572, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2574
 *     __slots__ = ('_name', '_kind', '_default', '_annotation')
 * 
 *     POSITIONAL_ONLY = _POSITIONAL_ONLY             # <<<<<<<<<<<<<<
 *     POSITIONAL_OR_KEYWORD = _POSITIONAL_OR_KEYWORD
 *     VAR_POSITIONAL = _VAR_POSITIONAL
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_POSITIONAL_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_POSITIONAL_ONLY_2, __pyx_t_9) < 0) __PYX_ERR(0, 2574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2575
 * 
 *     POSITIONAL_ONLY = _POSITIONAL_ONLY
 *     POSITIONAL_OR_KEYWORD = _POSITIONAL_OR_KEYWORD             # <<<<<<<<<<<<<<
 *     VAR_POSITIONAL = _VAR_POSITIONAL
 *     KEYWORD_ONLY = _KEYWORD_ONLY
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_POSITIONAL_OR_KEYWORD); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_POSITIONAL_OR_KEYWORD_2, __pyx_t_9) < 0) __PYX_ERR(0, 2575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2576
 *     POSITIONAL_ONLY = _POSITIONAL_ONLY
 *     POSITIONAL_OR_KEYWORD = _POSITIONAL_OR_KEYWORD
 *     VAR_POSITIONAL = _VAR_POSITIONAL             # <<<<<<<<<<<<<<
 *     KEYWORD_ONLY = _KEYWORD_ONLY
 *     VAR_KEYWORD = _VAR_KEYWORD
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_VAR_POSITIONAL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_VAR_POSITIONAL_2, __pyx_t_9) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2577
 *     POSITIONAL_OR_KEYWORD = _POSITIONAL_OR_KEYWORD
 *     VAR_POSITIONAL = _VAR_POSITIONAL
 *     KEYWORD_ONLY = _KEYWORD_ONLY             # <<<<<<<<<<<<<<
 *     VAR_KEYWORD = _VAR_KEYWORD
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_KEYWORD_ONLY); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_KEYWORD_ONLY_2, __pyx_t_9) < 0) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2578
 *     VAR_POSITIONAL = _VAR_POSITIONAL
 *     KEYWORD_ONLY = _KEYWORD_ONLY
 *     VAR_KEYWORD = _VAR_KEYWORD             # <<<<<<<<<<<<<<
 * 
 *     empty = _empty
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_VAR_KEYWORD); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_VAR_KEYWORD_2, __pyx_t_9) < 0) __PYX_ERR(0, 2578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2580
 *     VAR_KEYWORD = _VAR_KEYWORD
 * 
 *     empty = _empty             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_empty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_empty_2, __pyx_t_9) < 0) __PYX_ERR(0, 2580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2582
 *     empty = _empty
 * 
 *     def __init__(self, name, kind, *, default=_empty, annotation=_empty):             # <<<<<<<<<<<<<<
 *         try:
 *             self._kind = _ParameterKind(kind)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_1__init__, 0, __pyx_n_s_Parameter___init, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__292)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults3), 2)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_9)->__pyx_arg_default = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_9)->__pyx_arg_annotation = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_197__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2622
 *         self._name = name
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (self._name, self._kind),
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_3__reduce__, 0, __pyx_n_s_Parameter___reduce, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__294)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_reduce, __pyx_t_9) < 0) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2628
 *                  '_annotation': self._annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._default = state['_default']
 *         self._annotation = state['_annotation']
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_5__setstate__, 0, __pyx_n_s_Parameter___setstate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__296)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setstate, __pyx_t_9) < 0) __PYX_ERR(0, 2628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2633
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self._name
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_7name, 0, __pyx_n_s_Parameter_name, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__298)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2632
 *         self._annotation = state['_annotation']
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name(self):
 *         return self._name
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_name_2, __pyx_t_1) < 0) __PYX_ERR(0, 2633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2637
 * 
 *     @property
 *     def default(self):             # <<<<<<<<<<<<<<
 *         return self._default
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_9default, 0, __pyx_n_s_Parameter_default, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__300)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2636
 *         return self._name
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def default(self):
 *         return self._default
 */
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_default, __pyx_t_9) < 0) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2641
 * 
 *     @property
 *     def annotation(self):             # <<<<<<<<<<<<<<
 *         return self._annotation
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_11annotation, 0, __pyx_n_s_Parameter_annotation, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__302)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2640
 *         return self._default
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def annotation(self):
 *         return self._annotation
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_annotation, __pyx_t_1) < 0) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2645
 * 
 *     @property
 *     def kind(self):             # <<<<<<<<<<<<<<
 *         return self._kind
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_13kind, 0, __pyx_n_s_Parameter_kind, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__304)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2644
 *         return self._annotation
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kind(self):
 *         return self._kind
 */
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_kind, __pyx_t_9) < 0) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2648
 *         return self._kind
 * 
 *     def replace(self, *, name=_void, kind=_void,             # <<<<<<<<<<<<<<
 *                 annotation=_void, default=_void):
 *         """Creates a customized copy of the Parameter."""
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_15replace, 0, __pyx_n_s_Parameter_replace, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__306)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults4), 4)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_9)->__pyx_arg_name = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_9)->__pyx_arg_kind = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2649
 * 
 *     def replace(self, *, name=_void, kind=_void,
 *                 annotation=_void, default=_void):             # <<<<<<<<<<<<<<
 *         """Creates a customized copy of the Parameter."""
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_9)->__pyx_arg_annotation = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_void); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_9)->__pyx_arg_default = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_199__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_replace, __pyx_t_9) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2666
 *         return type(self)(name, kind, default=default, annotation=annotation)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         kind = self.kind
 *         formatted = self._name
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_17__str__, 0, __pyx_n_s_Parameter___str, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__308)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 2666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2688
 *         return formatted
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_19__repr__, 0, __pyx_n_s_Parameter___repr, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__310)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_repr, __pyx_t_9) < 0) __PYX_ERR(0, 2688, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2691
 *         return '<{} "{}">'.format(self.__class__.__name__, self)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_21__hash__, 0, __pyx_n_s_Parameter___hash, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__312)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_hash, __pyx_t_9) < 0) __PYX_ERR(0, 2691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2694
 *         return hash((self.name, self.kind, self.annotation, self.default))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Parameter_23__eq__, 0, __pyx_n_s_Parameter___eq, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__314)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_eq, __pyx_t_9) < 0) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2550
 * 
 * 
 * class Parameter:             # <<<<<<<<<<<<<<
 *     """Represents a parameter in a function signature.
 * 
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_Parameter, __pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Parameter, __pyx_t_9) < 0) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2705
 * 
 * 
 * class BoundArguments:             # <<<<<<<<<<<<<<
 *     """Result of `Signature.bind` call.  Holds the mapping of arguments
 *     to the function's parameters.
 */
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_BoundArguments, __pyx_n_s_BoundArguments, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_Result_of_Signature_bind_call_Ho); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/contracts/inspection.py":2722
 *     """
 * 
 *     __slots__ = ('arguments', '_signature', '__weakref__')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, signature, arguments):
 */
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_slots, __pyx_tuple__315) < 0) __PYX_ERR(0, 2722, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2724
 *     __slots__ = ('arguments', '_signature', '__weakref__')
 * 
 *     def __init__(self, signature, arguments):             # <<<<<<<<<<<<<<
 *         self.arguments = arguments
 *         self._signature = signature
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_1__init__, 0, __pyx_n_s_BoundArguments___init, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__317)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2724, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2729
 * 
 *     @property
 *     def signature(self):             # <<<<<<<<<<<<<<
 *         return self._signature
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_3signature, 0, __pyx_n_s_BoundArguments_signature, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__319)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2728
 *         self._signature = signature
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def signature(self):
 *         return self._signature
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_signature_2, __pyx_t_1) < 0) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2733
 * 
 *     @property
 *     def args(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for param_name, param in self._signature.parameters.items():
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_5args, 0, __pyx_n_s_BoundArguments_args, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__321)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2732
 *         return self._signature
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def args(self):
 *         args = []
 */
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_args, __pyx_t_9) < 0) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2756
 * 
 *     @property
 *     def kwargs(self):             # <<<<<<<<<<<<<<
 *         kwargs = {}
 *         kwargs_started = False
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_7kwargs, 0, __pyx_n_s_BoundArguments_kwargs, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__323)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2755
 *         return tuple(args)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kwargs(self):
 *         kwargs = {}
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_kwargs, __pyx_t_1) < 0) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2785
 *         return kwargs
 * 
 *     def apply_defaults(self):             # <<<<<<<<<<<<<<
 *         """Set default values for missing arguments.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_9apply_defaults, 0, __pyx_n_s_BoundArguments_apply_defaults, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__325)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_apply_defaults, __pyx_t_1) < 0) __PYX_ERR(0, 2785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2813
 *         self.arguments = dict(new_arguments)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_11__eq__, 0, __pyx_n_s_BoundArguments___eq, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__327)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_eq, __pyx_t_1) < 0) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2821
 *                 self.arguments == other.arguments)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._signature = state['_signature']
 *         self.arguments = state['arguments']
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_13__setstate__, 0, __pyx_n_s_BoundArguments___setstate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__329)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setstate, __pyx_t_1) < 0) __PYX_ERR(0, 2821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2825
 *         self.arguments = state['arguments']
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_15__getstate__, 0, __pyx_n_s_BoundArguments___getstate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__331)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getstate, __pyx_t_1) < 0) __PYX_ERR(0, 2825, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2828
 *         return {'_signature': self._signature, 'arguments': self.arguments}
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         args = []
 *         for arg, value in self.arguments.items():
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_14BoundArguments_17__repr__, 0, __pyx_n_s_BoundArguments___repr, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__333)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_repr, __pyx_t_1) < 0) __PYX_ERR(0, 2828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2705
 * 
 * 
 * class BoundArguments:             # <<<<<<<<<<<<<<
 *     """Result of `Signature.bind` call.  Holds the mapping of arguments
 *     to the function's parameters.
 */
  __pyx_t_1 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_BoundArguments, __pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BoundArguments, __pyx_t_1) < 0) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2835
 * 
 * 
 * class Signature:             # <<<<<<<<<<<<<<
 *     """A Signature object represents the overall signature of a function.
 *     It stores a Parameter object for each parameter accepted by the
 */
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_Signature, __pyx_n_s_Signature, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_kp_s_A_Signature_object_represents_th); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/contracts/inspection.py":2858
 *     """
 * 
 *     __slots__ = ('_return_annotation', '_parameters')             # <<<<<<<<<<<<<<
 * 
 *     _parameter_cls = Parameter
 */
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_slots, __pyx_tuple__334) < 0) __PYX_ERR(0, 2858, __pyx_L1_error)

  /* "Aspidites/_vendor/contracts/inspection.py":2860
 *     __slots__ = ('_return_annotation', '_parameters')
 * 
 *     _parameter_cls = Parameter             # <<<<<<<<<<<<<<
 *     _bound_arguments_cls = BoundArguments
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Parameter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_parameter_cls, __pyx_t_1) < 0) __PYX_ERR(0, 2860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2861
 * 
 *     _parameter_cls = Parameter
 *     _bound_arguments_cls = BoundArguments             # <<<<<<<<<<<<<<
 * 
 *     empty = _empty
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_BoundArguments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_bound_arguments_cls, __pyx_t_1) < 0) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2863
 *     _bound_arguments_cls = BoundArguments
 * 
 *     empty = _empty             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_empty_2, __pyx_t_1) < 0) __PYX_ERR(0, 2863, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2865
 *     empty = _empty
 * 
 *     def __init__(self, parameters=None, *, return_annotation=_empty,             # <<<<<<<<<<<<<<
 *                  __validate_parameters__=True):
 *         """Constructs Signature from the given list of Parameter
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_1__init__, 0, __pyx_n_s_Signature___init, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__336)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults5), 1)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_empty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_t_1)->__pyx_arg_return_annotation = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_201__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_1) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2920
 * 
 *     @classmethod
 *     def from_function(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given python function.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_3from_function, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_Signature_from_function, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__338)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2919
 *         self._return_annotation = return_annotation
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_function(cls, func):
 *         """Constructs Signature for the given python function.
 */
  __pyx_t_9 = __Pyx_Method_ClassMethod(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_from_function, __pyx_t_9) < 0) __PYX_ERR(0, 2920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2932
 * 
 *     @classmethod
 *     def from_builtin(cls, func):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given builtin function.
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_5from_builtin, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_Signature_from_builtin, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__340)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2931
 *         return _signature_from_function(cls, func)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_builtin(cls, func):
 *         """Constructs Signature for the given builtin function.
 */
  __pyx_t_1 = __Pyx_Method_ClassMethod(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_from_builtin, __pyx_t_1) < 0) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2944
 * 
 *     @classmethod
 *     def from_callable(cls, obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *         """Constructs Signature for the given callable object."""
 *         return _signature_from_callable(obj, sigcls=cls,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_follow_wrapped, ((PyObject *)Py_True)) < 0) __PYX_ERR(0, 2944, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_7from_callable, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_Signature_from_callable, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__342)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_9, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2943
 *         return _signature_from_builtin(cls, func)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_callable(cls, obj, *, follow_wrapped=True):
 *         """Constructs Signature for the given callable object."""
 */
  __pyx_t_1 = __Pyx_Method_ClassMethod(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_from_callable, __pyx_t_1) < 0) __PYX_ERR(0, 2944, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2950
 * 
 *     @property
 *     def parameters(self):             # <<<<<<<<<<<<<<
 *         return self._parameters
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_9parameters, 0, __pyx_n_s_Signature_parameters, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__344)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/contracts/inspection.py":2949
 *                                         follow_wrapper_chains=follow_wrapped)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parameters(self):
 *         return self._parameters
 */
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_parameters, __pyx_t_9) < 0) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2954
 * 
 *     @property
 *     def return_annotation(self):             # <<<<<<<<<<<<<<
 *         return self._return_annotation
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_11return_annotation, 0, __pyx_n_s_Signature_return_annotation, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__346)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "Aspidites/_vendor/contracts/inspection.py":2953
 *         return self._parameters
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def return_annotation(self):
 *         return self._return_annotation
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_return_annotation, __pyx_t_1) < 0) __PYX_ERR(0, 2954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2957
 *         return self._return_annotation
 * 
 *     def replace(self, *, parameters=_void, return_annotation=_void):             # <<<<<<<<<<<<<<
 *         """Creates a customized copy of the Signature.
 *         Pass 'parameters' and/or 'return_annotation' arguments
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_13replace, 0, __pyx_n_s_Signature_replace, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__348)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults6), 2)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_void); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_1)->__pyx_arg_parameters = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_void); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_1)->__pyx_arg_return_annotation = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_9Aspidites_7_vendor_9contracts_10inspection_203__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_replace, __pyx_t_1) < 0) __PYX_ERR(0, 2957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2972
 *                           return_annotation=return_annotation)
 * 
 *     def _hash_basis(self):             # <<<<<<<<<<<<<<
 *         params = tuple(param for param in self.parameters.values()
 *                        if param.kind != _KEYWORD_ONLY)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_15_hash_basis, 0, __pyx_n_s_Signature__hash_basis, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__350)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_hash_basis, __pyx_t_1) < 0) __PYX_ERR(0, 2972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2981
 *         return params, kwo_params, self.return_annotation
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         params, kwo_params, return_annotation = self._hash_basis()
 *         kwo_params = frozenset(kwo_params.values())
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_17__hash__, 0, __pyx_n_s_Signature___hash, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__352)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_hash, __pyx_t_1) < 0) __PYX_ERR(0, 2981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2986
 *         return hash((params, kwo_params, return_annotation))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_19__eq__, 0, __pyx_n_s_Signature___eq, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__354)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_eq, __pyx_t_1) < 0) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2993
 *         return self._hash_basis() == other._hash_basis()
 * 
 *     def _bind(self, args, kwargs, *, partial=False):             # <<<<<<<<<<<<<<
 *         """Private method. Don't use directly."""
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_partial, ((PyObject *)Py_False)) < 0) __PYX_ERR(0, 2993, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_21_bind, 0, __pyx_n_s_Signature__bind, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__356)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_9, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_bind, __pyx_t_9) < 0) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3138
 *     #     return self._bind(args, kwargs, partial=True)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self),
 *                 (tuple(self._parameters.values()),),
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_23__reduce__, 0, __pyx_n_s_Signature___reduce, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__358)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_reduce, __pyx_t_9) < 0) __PYX_ERR(0, 3138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3143
 *                 {'_return_annotation': self._return_annotation})
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._return_annotation = state['_return_annotation']
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_25__setstate__, 0, __pyx_n_s_Signature___setstate, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__360)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setstate, __pyx_t_9) < 0) __PYX_ERR(0, 3143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3146
 *         self._return_annotation = state['_return_annotation']
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_27__repr__, 0, __pyx_n_s_Signature___repr, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__362)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_repr, __pyx_t_9) < 0) __PYX_ERR(0, 3146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3149
 *         return '<{} {}>'.format(self.__class__.__name__, self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         result = []
 *         render_pos_only_separator = False
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_9Signature_29__str__, 0, __pyx_n_s_Signature___str, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__364)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 3149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":2835
 * 
 * 
 * class Signature:             # <<<<<<<<<<<<<<
 *     """A Signature object represents the overall signature of a function.
 *     It stores a Parameter object for each parameter accepted by the
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_Signature, __pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Signature, __pyx_t_9) < 0) __PYX_ERR(0, 2835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3195
 * 
 * 
 * def signature(obj, *, follow_wrapped=True):             # <<<<<<<<<<<<<<
 *     """Get a signature object for the passed callable."""
 *     return Signature.from_callable(obj, follow_wrapped=follow_wrapped)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_follow_wrapped, ((PyObject *)Py_True)) < 0) __PYX_ERR(0, 3195, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_169signature, 0, __pyx_n_s_signature_2, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__366)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_9, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signature_2, __pyx_t_9) < 0) __PYX_ERR(0, 3195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3200
 * 
 * 
 * def can_accept_exactly_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept exactly one argument
 *         using introspection.
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_171can_accept_exactly_one_argument, 0, __pyx_n_s_can_accept_exactly_one_argument, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__368)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_can_accept_exactly_one_argument, __pyx_t_9) < 0) __PYX_ERR(0, 3200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3223
 * 
 * 
 * def get_f_from_callable(callable_thing):             # <<<<<<<<<<<<<<
 *     if ismethod(callable_thing):  # bound method
 *         f = callable_thing.__func__
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_173get_f_from_callable, 0, __pyx_n_s_get_f_from_callable, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__370)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_f_from_callable, __pyx_t_9) < 0) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3236
 * 
 * 
 * def get_callable_fullargspec(callable_thing):             # <<<<<<<<<<<<<<
 *     f: Callable
 *     spec: FullArgSpec
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_175get_callable_fullargspec, 0, __pyx_n_s_get_callable_fullargspec, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__372)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_callable_fullargspec, __pyx_t_9) < 0) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3244
 * 
 * 
 * def can_accept_at_least_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept at least one argument
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_177can_accept_at_least_one_argument, 0, __pyx_n_s_can_accept_at_least_one_argument, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__374)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_can_accept_at_least_one_argument, __pyx_t_9) < 0) __PYX_ERR(0, 3244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3253
 * 
 * 
 * def can_accept_self(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable's first argument is self
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_179can_accept_self, 0, __pyx_n_s_can_accept_self, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__376)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_can_accept_self, __pyx_t_9) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3265
 * 
 * 
 * def can_accept_self_plus_one_argument(callable_thing):             # <<<<<<<<<<<<<<
 *     """
 *         Checks that a callable can accept exactly self plus one argument
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_181can_accept_self_plus_one_argument, 0, __pyx_n_s_can_accept_self_plus_one_argumen, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__378)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_can_accept_self_plus_one_argumen, __pyx_t_9) < 0) __PYX_ERR(0, 3265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3285
 * 
 * 
 * class InvalidArgs(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_9, __pyx_n_s_InvalidArgs, __pyx_n_s_InvalidArgs, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_InvalidArgs, __pyx_t_9, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InvalidArgs, __pyx_t_10) < 0) __PYX_ERR(0, 3285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":3289
 * 
 * 
 * def check_callable_accepts_these_arguments(callable_thing, args, kwargs):             # <<<<<<<<<<<<<<
 *     """ Checks that a callable can accept the args and kwargs.
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_9contracts_10inspection_183check_callable_accepts_these_arguments, 0, __pyx_n_s_check_callable_accepts_these_arg, NULL, __pyx_n_s_Aspidites__vendor_contracts_insp, __pyx_d, ((PyObject *)__pyx_codeobj__380)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_callable_accepts_these_arg, __pyx_t_9) < 0) __PYX_ERR(0, 3289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/contracts/inspection.py":1
 * #cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=True, initializedcheck=False             # <<<<<<<<<<<<<<
 * """Get useful information from live Python objects.
 * This is a statically compiled version of the inspect module from the standard library.
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_2, __pyx_t_9) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init Aspidites._vendor.contracts.inspection", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init Aspidites._vendor.contracts.inspection");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* pop_index */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix) {
    PyObject *r;
    if (unlikely(!py_ix)) return NULL;
    r = __Pyx__PyObject_PopIndex(L, py_ix);
    Py_DECREF(py_ix);
    return r;
}
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix) {
    return __Pyx_PyObject_CallMethod1(L, __pyx_n_s_pop, py_ix);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix) {
    Py_ssize_t size = PyList_GET_SIZE(L);
    if (likely(size > (((PyListObject*)L)->allocated >> 1))) {
        Py_ssize_t cix = ix;
        if (cix < 0) {
            cix += size;
        }
        if (likely(__Pyx_is_valid_index(cix, size))) {
            PyObject* v = PyList_GET_ITEM(L, cix);
            __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
            size -= 1;
            memmove(&PyList_GET_ITEM(L, cix), &PyList_GET_ITEM(L, cix+1), (size_t)(size-cix)*sizeof(PyObject*));
            return v;
        }
    }
    if (py_ix == Py_None) {
        return __Pyx__PyObject_PopNewIndex(L, PyInt_FromSsize_t(ix));
    } else {
        return __Pyx__PyObject_PopIndex(L, py_ix);
    }
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* RaiseKeywordRequired */
static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name) {
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() needs keyword-only argument %U", func_name, kw_name);
        #else
        "%s() needs keyword-only argument %s", func_name,
        PyString_AS_STRING(kw_name));
        #endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchInspect */
static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
